{*******************************************************************}
{ E_Io.pas
  This module copyright 1996
  Copied from T_IO.PAS and extensively modified by Peter E. Midford
  for Theodore Garland, Jr.
  Last modified 12 May 1996                                         }
{*******************************************************************}

{This unit has the input and output procedures.  It provides for I/O of a
tree to and from disk as well as the creation of a tree.}

Unit e_io;
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Interface

Uses e_utils,  {type definitions, constants, and utility functions.}
     DOS;      {definitions for file operations.}

Procedure openSimFile(Var flName :flString;
                      Var simFile :text;
                      Var succ :boolean); 

Procedure readTree(Var rt :nodeptr;
                   Var tipCount : integer;
                   Var simFile :text);

Procedure readDataSet(Var simData :nodeDataPtr;
                          tipCount : integer;
                      Var simFile :text);

Procedure readSecondTree(Var rt :nodeptr;
                         Var fl :text;
                         Var tipCount : integer);

Procedure skipTree(Var SimFile :text);

Procedure openResultFile(rootName :flString;
                         Var resultFile :text;
                         Var resultFileName :flString;
                         resultFileExt :typString);

Procedure openTweaksFile(Var tweaksFile :text;tkFileName :flString);

Procedure openParametersFile(rootName :flString;
                             Var paramFile :text;
                             Var paramFileName :flString);

Procedure writeParameters(Var paramFile :text;
                          resultFileName :flString;
                          dataFileName :flString;
                          secondTreeName :flString;
                          tweaksToSave : integer;
                          tweakFileName : flString;
                          binCount :itraits;
                          binBase :itraits;
                          doRanks :traitFlags;
                          tweak :tweakPtr;
                          xForms : xFormPtr;
                          rSeed :longint);

Procedure writeXFormResults(T1,T2 :real;
                            type1,type2 :char;
                            Var branchFile :text);

Procedure write1TreeStats(Var statFile :text;
                          simStats :statPak;
                          setNumber :integer;
                          corrA :corrArray;
                          root :nodePtr;
                          tipStats :statPak);

Procedure write2TreeStats(Var StatFile :text;
                          simStats :statPak;
                          setNumber :integer;
                          corrA :CorrArray;
                          t1,t2 :extended;
                          type1,type2 :char;
                          root1 :nodeptr;
                          root2 :nodeptr;
                          tipStats :statPak);

Procedure writeTweakedTree(rt :nodeptr;var treeFile :text);

Procedure extendWrite(Var outFile :text;num :extended);

Procedure scanForLimits(inputFLName : flString;
                        Var globalMin : traits;
                        Var globalMax : traits;
                        tipCount : integer;
                        userLimit : integer);

Implementation

Uses crt;     {for reading single keystrokes}

Const buffSize = 8192;
      largeBuffSize = 65000;

Type textBuff = array[1..largeBuffSize] of char;
     buffPtr = ^textBuff;

Var inBuff :array[1..buffSize] of char;
    outBuff :array[1..buffSize] of char;
    outBuff2 :array[1..buffSize] of char;


{*******************************************************************}
{ This procedure reads the filename from the user and opens the
  file with that name if it exists.                                 }
{*******************************************************************}
Procedure openSimFile(Var flName :flString;
                      Var simFile :text;
                      Var succ :boolean);
   {************************************************************************}
   {This procedure reads in a file name.  It is simplified from the reader
    in PDTREE.}
   {************************************************************************}
   Procedure getInFile(Var fleName :flString;
                       Var succ :boolean);
      Var exten :typString;
          ch :char;
          place :byte;
          exists :boolean;
          i :integer;
          fl :text;
      Begin
         exists := true;
         Repeat
            If not exists Then
               writeln(flename,' does not exist.');
            exists := true;
            writeln('Please enter the name of the input file containing simulated tip data,');
            writeln('including the extension.');
            write('   Or "Q" to quit program: ');
            readln(fleName);
            If ((fleName='Q') or (fleName='q')) Then halt; {Bail out}
            {$I-}
            assign(fl,fleName);
            setTextBuf(fl,inBuff);
            reset(Fl);
            close(Fl);
            {$I+}
            exists := (ioResult=0) and (fleName<>'')
         Until (exists)
      End; {procedure getinfile}
   {*****************************************************************}
   Begin {procedure opensimfile}
      succ := true;
      getInFile(flName,succ);
      writeln;
      If succ Then assign(simFile,flName)
   End; {procedure opensimfile}

{***************************************************************************}
{This procedure, lifted from PDTREE, reads the tree, specified in INP format}
{that appears at the top of the file.}
{***************************************************************************}
Procedure readTree(Var rt :nodeptr;
                   Var tipCount : integer;
                   Var SimFile :text);
   Var ch :char;
       me :nodeptr;
       fail :boolean;
       dumptr :nodeptr;
       ok, good :boolean;
       name :namestring;
       ch2 :string[2];
       chhh :char;
       ch4 :string[4];
       tempint :integer;
       trait1,trait2 :real;
       setDone :boolean;
   {*********************************************************************}
   {This just performs the menial task of reading in the data on a
    node.}
   {*********************************************************************}
   Procedure readSib(Var fl :text;sib,me :nodePtr);
      Var tempreal :real;
          ch2 :string[2];
          ch :char;
      Begin
         sib^.ancestor := me;
         read(fl,sib^.name);
         read(fl,ch);
         read(fl,sib^.length)
      End; {procedure readsib}
   {*********************************************************************}
   {This does something similar to the previous procedure, except that
   it checks the tree for duplicate names, then reads in one real piece
   of data.}
   {*********************************************************************}
   Procedure readtrait(i :integer;
                       rt :nodePtr;
                       Var fl :text;
                       ch2 :string;
                       Var fail :boolean);
      Var tempreal :extended;
          me :nodePtr;
      Begin
         fail := true;
         checkname(rt,ch2,fail,me);
         If fail Then exit;
         read(fl,tempreal);
         me^.dat[i] := tempreal
      End; {procedure readtrait}
   {***********************************************************************}
   Begin {procedure readtree}
      reset(SimFile);
      Repeat
         Read(SimFile,ch4);
         Readln(SimFile)
      Until ((ch4='(Nod') or eof(simFile));
      If eof(simFile) Then Begin
         writeln('Input file is not recognizable as a simulation or randomization file.');
         writeln('Please check the file or run the program with a different file.');
         writeln;
         writeln('Exiting...');
         halt;
      End;
      new(rt);
      initme(rt);
      rt^.ancestor := nil;
      rt^.length := 0.0;
      read(SimFile,ch2);
      read(SimFile,ch2);
      rt^.name := ch2;
      tipCount := 1;
      While (ch2<>'MU') Do
         Begin
            fail := true;
            checkName(rt,ch2,fail,me);
            If fail Then Begin
               writeln('Tree in SIM file is corrupted(1).  Please correct Tree, in  '+
                        'PDTREE, and rerun simulation.  Press any key to quit.');
               ch := readkey;
               halt           {bail out}
            End;
            new(me^.sibL);
            initme(me^.sibL); 
            read(SimFile,ch);
            readsib(SimFile,me^.sibL,me);
            read(SimFile,tempint);
            read(SimFile,ch);
            new(me^.sibR);
            initme(me^.sibR);
            readsib(SimFile,me^.sibR,me);
            readln(SimFile);
            read(SimFile,ch2);
            read(SimFile,ch2);
            inc(tipCount)
         End; {while}
      Repeat
         read(SimFile,ch);
         readln(SimFile)
      Until (ch='_')
   End; {procedure readtree}

{***************************************************************************}
{This procedure reads a set of simulated data from the sim file.  Each data
 set is a list of nodeDataPtr's.}
{***************************************************************************}
Procedure readDataSet(Var simData :nodeDataPtr;
                          tipCount : integer;
                      Var simFile :text);
   Var
      currentValues :nodeDataPtr;
      count : integer;
      setDone :boolean;
      ch :char;
   Begin
      simData := nil;
      setDone := false;
      count := 0;
      While ((not setDone) and (not eof(simFile))) Do Begin
         count := count + 1;
         If (simData = nil) Then Begin
            new(currentValues);
            initNodeData(currentValues);
            simData := currentValues
         End {then}
         Else Begin
            new(currentValues^.next);
            initNodeData(currentValues^.next);
            currentValues := currentValues^.next           
         End; {else}
         read(simFile,currentValues^.name);
         read(simFile,currentValues^.data[1]);
         read(simFile,currentValues^.data[2]);
         read(simFile,ch);
         new(currentValues^.next);
         initNodeData(currentValues^.next);
         currentValues := currentValues^.next;
         read(simFile,currentValues^.name);
         read(simFile,currentValues^.data[1]);
         read(simFile,currentValues^.data[2]);
         If (count = tipCount) Then setDone := true;
         write('');
         readln(simFile)
      End; {while}
   End; {procedure readdataset}
       
{***************************************************************************}
{Once again, an INP tree reader for reading in a separate tree.  This tree
 must have the same topology (including tip names) as the tree in the
 simulation, but we don't check any of that here.
 {***************************************************************************}
Procedure readSecondTree(Var rt :nodeptr;
                         Var fl :text;
                         Var tipCount : integer);
   Var
      me : nodePtr;
      ch : char;
      ch2 :string[2];
      chhh : char;
      ch4 : string[4];
      tempint : integer;
      nameFail : boolean;
   {*********************************************************************}
   {This just performs the menial task of reading in the data on a
    node.}
   {*********************************************************************}
   Procedure readSib(var fl: text; sib,me:nodePtr);
      Var
         tempReal : real;
         ch2 : string[2];
         ch : char;
      Begin
         sib^.ancestor:=me;
         read(fl,sib^.name);
         read(fl,ch);
         read(fl,sib^.length);
      End; {procedure readsib}
   {*********************************************************************}
   {This does something similar to the previous procedure, except that
    it checks the tree for duplicate names, then reads in one real piece
    of data.}
   {*********************************************************************}
   Procedure readTrait(i:integer;
                       rt:nodeptr;
                       Var fl:text;
                       ch2:string;
                       Var fail:boolean);
      Var
         tempReal :real;
         me :nodeptr;
      Begin
         fail:=true;
         checkName(rt,ch2,fail,me);
         If fail Then exit;
         read(fl,tempreal);
         me^.dat[i]:=tempreal;
      End; {readTrait}
   {***********************************************************************}
   Begin
      {$I-}
      reset(fl);
      {$I+}
      If (IOResult = 0) Then Begin
         Repeat
            read(fl,ch4);
            readln(fl);
         Until ch4='  of';
         new(rt);
         initme(rt);
         rt^.ancestor:=nil;
         rt^.length:=0.0;
         read(fl,ch2);
         rt^.name:=ch2;
         tipCount := 1;
         tipCount := 1;
         While not eof(fl) Do Begin
            nameFail := true;
            checkname(rt,ch2,nameFail,me);
            If nameFail Then Begin
               writeln('Tree in file is invalid(1).  Please press any key '+
                   'to enter a new filename.');
               killtree(rt);
               exit;
            End; {then}
            new(me^.sibL);
            initme(me^.sibL); 
            read(fl,ch);
            readsib(fl,me^.sibL,me);
            read(fl,tempint);
            read(fl,ch);
            new(me^.sibR);
            initme(me^.sibR);
            readsib(fl,me^.sibR,me);
            readln(fl);
            read(fl,ch2);
            inc(tipCount);
         End; {while}
      End
      Else Begin
         writeln('File does not exist.  Please press any key to enter a new file name.');
         rt := nil;
      End; {else}
   End; {procedure readSecondTree}

{***************************************************************************}
{This procedure simply skips the tree in the sim file.  The only effect of
 this procedure is open the sim file and move to the start of the data sets.}
{***************************************************************************}
Procedure skipTree(Var SimFile :text);
   Var ch :char;
   Begin
      reset(SimFile);
      Repeat
         read(SimFile,ch);
         readln(SimFile)
      Until (ch='_')
   End; {procedure skipTree}

{***************************************************************************}
{This procedure opens the results file.  The name as well as the file
 pointer are returned.}  
{***************************************************************************}
Procedure openResultFile(rootName :flString;
                         Var resultFile :text;
                         Var resultFileName :flString;
                         resultFileExt :typString);
   Begin
      resultFileName := rootName + '.' + resultFileExt;
      assign(resultFile,resultFileName);
      setTextBuf(resultFile,outBuff);
      reWrite(resultFile)
   End; {procedure openresultfile}

{***************************************************************************}
{This procedure opens a file for writing tweaked trees. The file
 pointer is returned.}  
{***************************************************************************}
Procedure openTweaksFile(Var tweaksFile :text;tkFileName :flString);
   Begin
      assign(tweaksFile,tkFileName);
      setTextBuf(tweaksFile,outBuff2);
      reWrite(tweaksFile)
   End;

{***************************************************************************}
{This procedure opens a file for writing tweaked trees. The file
 pointer is returned.}  
{***************************************************************************}
Procedure openParametersFile(rootName :flString;
                             Var paramFile :text;
                             Var paramFileName :flString);
   Begin
      paramFileName := rootName + '.' + 'PAR';
      assign(paramFile,paramFileName);
      rewrite(paramFile);
      writeln(paramFile,'File = ',paramFileName,' = parameters used in PDERROR run.');
      writeln(paramFile);
   End; { procedure openParametersFile }

{***************************************************************************}
{ This procedure writes the header and parameter choices information into   }
{ the .PAR file.                                                            }
{***************************************************************************}
Procedure writeParameters(Var paramFile :text;
                          resultFileName :flString;
                          dataFileName :flString;
                          secondTreeName :flString;
                          tweaksToSave : integer;
                          tweakFileName : flString;
                          binCount :itraits;
                          binBase :itraits;
                          doRanks :traitFlags;
                          tweak :tweakPtr;
                          xForms : xFormPtr;
                          rSeed :longint);
   Var fileTime :DateTime;                  {Holds the creation date of the results file}
       resultFileHandle :text;              {opens the result file to get creation date}
       dataFile : text;
       fileTimeNumber :longInt;             {Holds intermediate value}
       lineBuffer : string;

   Begin

      {get the creation date of the result file}
      assign(resultFileHandle,resultFileName);
      reset(resultFileHandle);
      getFTime(resultFileHandle,fileTimeNumber);
      unPackTime(fileTimeNumber,fileTime);
      close(resultFileHandle);
      write(paramFile,'PERROR results were written to ',resultFileName,' on ');
      write(paramFile,fileTime.Day,' ');
      Case fileTime.Month of
         1 : Write(paramFile, 'Jan ');
         2 : Write(paramFile, 'Feb ');
         3 : Write(paramFile, 'Mar ');
         4 : Write(paramFile, 'Apr ');
         5 : Write(paramFile, 'May ');
         6 : Write(paramFile, 'Jun ');
         7 : Write(paramFile, 'Jul ');
         8 : Write(paramFile, 'Aug ');
         9 : Write(paramFile, 'Sep ');
        10 : Write(paramFile, 'Oct ');
        11 : Write(paramFile, 'Nov ');
        12 : Write(paramFile, 'Dec ');
      End; {case month of}
      write(paramFile, fileTime.Year,' at ');
      write(paramFile,fileTime.Hour,':');
      If (fileTime.min < 10)
      Then write(paramFile,'0',fileTime.min,':')
      Else write(paramFile,fileTime.min,':');
      If (fileTime.sec < 10)
      Then writeln(paramFile,'0',fileTime.sec,'.')
      Else writeln(paramFile,fileTime.sec,'.');
      writeln(paramFile);

      writeln(paramFile,'Data were read from ',dataFileName);
      assign(dataFile,dataFileName);
      reset(dataFile);
      readln(dataFile);
      readln(dataFile,lineBuffer);
      close(dataFile);
      writeln(paramFile,lineBuffer);
      writeln(paramFile);
      writeln(paramFile);
      If (secondTreeName <> '') Then Begin
         writeln(paramFile,'Data were applied to a tree read from ',secondTreeName);
         writeln(paramFile);
      End; {then}

      If (binCount[1] > 0) Then Begin
         writeln(paramFile,'Trait 1 data were categorized into ',
                           binCount[1]:2,
                           ' bins, which were numbered from ',
                           binBase[1]:1,
                           '.');
      End;
      If (binCount[2] > 0) Then Begin
         writeln(paramFile,'Trait 2 data were categorized into ',
                           binCount[2]:2,
                           ' bins, which were numbered from ',
                           binBase[2]:1,
                           '.');
      End;

      If (binCount[1]+binCount[2] = 0)
      Then writeln(paramFile,'Data were not categorized.');
      writeln(paramFile);


      If (1 in doRanks)
      Then writeln(paramFile,'Trait 1 data were ranked.');
      If (2 in doRanks)
      Then writeln(paramFile,'Trait 2 data were ranked.');
      If (doRanks = [])
      Then writeln(paramFile, 'Data were not ranked.');
      writeln(paramFile);


      If (tweak <> nil) Then Begin
         If (tweak^.myType = FIXED) Then Begin
            writeln(paramFile,'Branch lengths were modifed by either adding or subtracting');
            writeln(paramFile,'the quantity (', tweak^.deviation:5:3,' times the length).');
            writeln(paramFile);
            If (tweak^.deviation >= 1.0)
            Then Begin
               write(paramFile,'Any branches tweaked to nonpositive lengths had their lengths set to ');
               writeln(paramFile,tweak^.zeroLength:9:6,'.');
               writeln(paramFile);
            End {then}
         End
         Else Begin
            writeln(paramFile,'Branch lengths were modified by adding a relative normal fraction.');
            write(paramFile,'The normal number was drawn from a distribution with mean = 0 and CV = ');
            writeln(paramFile,tweak^.width*100:4:1,'%.');
            writeln(paramFile);
            write(paramFile,'Any branches tweaked to nonpositive lengths had their lengths set to ');
            writeln(paramFile,tweak^.zeroLength:9:6,'.');
            writeln(paramFile);
         End;
         writeln(paramFile,'The random number seed used was ',rSeed,'.');
      End
      Else
         writeln(paramFile,'Branch lengths were not modified.');
      writeln(paramFile);

      If (xForms <> NIL) Then Begin
         writeln(paramFile);
         case(xForms^.cost) of

            SdcBL: Begin
                      writeln(paramFile,'Branch lengths were transformed to minimize the correlation between ');
                      writeln(paramFile,'the absolute value of standardized contrasts and their standard deviations.');
                      writeln(paramFile,'These are the correlations on PDTREE diagnostic screens 1 and 2.');
                      writeln(paramFile,'This is option B.');
                      writeln(paramFile,'This diagnostic was suggested by Garland et al. (1992) and studied by');
                      writeln(paramFile,'Diaz-Uriarte and Garland (1996,1998) and Garland and Diaz-Uriarte (1999).');
                   End; {case SdcNV}
            SdcNV: Begin
                      writeln(paramFile,'Branch lengths were transformed to minimize the correlation between the');
                      writeln(paramFile,'absolute value of standardized contrasts and the estimated value of the');
                      writeln(paramFile,'base node.');
                      writeln(paramFile,'This is option N.');
                      writeln(paramFile,'These are the correlations on PDTREE diagnostic screens 3 and 4.');
                   End; {case SdcNV}

            SdcHt: Begin
                      writeln(paramFile,'Branch lengths were transformed to minimize the correlation between the');
                      writeln(paramFile,'absolute value of standardized contrasts and the corrected height of the');
                      writeln(paramFile,'base node.');
                      writeln(paramFile,'This is option H.');
                      writeln(paramFile,'These are the correlations on PDTREE diagnostic screens 5 and 6.');
                   End; {case SdcHt}

            NvHt: Begin
                     writeln(paramFile,'Branch lengths were transformed to minimize the correlation between the');
                     writeln(paramFile,'estimated value of the base node and its corrected height.');
                     writeln(paramFile,'These are the correlations on PDTREE diagnostic screens 7 and 8.');
                     writeln(paramFile,'This is option V.');
                  End; {case NvHt}
         End; {case xForms^.cost}
         writeln(paramfile);
         If (xForms^.doExps) Then Begin
            writeln(paramFile,'  Exponential transforms were explored from ',
                              xForms^.expStart:2:1,
                              ' to ',
                              xForms^.expEnd:2:1,'.');
            writeln(paramFile);
         End; {then}
         If xForms^.doLog Then Begin
            writeln(paramFile,'  The log10 transform was considered.');
            writeln(paramFile);
         End {then}
         Else Begin
            writeln(paramFile,'  The log10 transform was not considered.');
            writeln(paramFile);
         End;
         If xForms^.doRho Then Begin
            writeln(paramFile,'  Grafen''s (1989) rho transforms were explored from ',
                              xForms^.rhoStart:2:1,
                              ' to ',
                              xForms^.rhoEnd:2:1,'.');
            writeln(paramFile);
         End; {then}

         If xForms^.doGrafen Then Begin 
            writeln(paramFile,
                        '  Grafen''s (1989) arbitrary branch length transformation was considered.');
            writeln(paramFile);
         End; {then}

         If xForms^.doGrafen Then Begin 
            writeln(paramFile,
                        '  Nee''s arbitrary branch length transformation (cited in Purvis 1995)');
            writeln(paramFile,'         was considered.');
            writeln(paramFile);
         End; {then}

         If xForms^.doPagel Then Begin 
            writeln(paramFile,
                        '  Pagel''s (1992) arbitrary branch length transformation was considered.');
            writeln(paramFile);
         End; {then}
         
         If xForms^.doOne Then Begin 
            writeln(paramFile,
                        '  The transform setting all branch lengths to unity was considered.');
            writeln(paramFile);
         End; {then}

         If xForms^.doSelf Then Begin
            writeln(paramFile,
                        '  The self (identity) transform was considered.');
            writeln(paramFile);
         End; {then}

      End {then}
      Else
         writeln(paramFile,'Branch length transformations were not explored.');
   End; {procedure writeParameters}

{************************************************************************}
{This procedure writes coprocessor extended reals to a file without the
extra exponent and mantessa digits.

Warning: although this extendWrite looks similar that found elsewhere
in PDAP, it pads the output with a terminal space.  If you cut and paste
an update from another program, make sure the terminal space is still
present, or you'll have to add the space after every call to extendwrite!}
{************************************************************************}
Procedure extendWrite(Var outFile :text;num :extended);
   Var outStr :string[19];
       userChar : char;
   Begin
      If (abs(num) < 9.9E99) Then Begin
         If ((abs(num) > 1E-99) Or (Abs(num) = 0)) Then Begin
            str(num:19:-11,outStr);
            delete(outStr,16,2);
            write(outFile,outStr,' ')
         End
         Else Begin
            clrscr;
            writeln('Error:  PDERROR tried to write a number with an absolute value smaller');
            writeln('        than 1E-99.  The program will now exit.  You should go into');
            writeln('        PDTREE and rescale your tree, for example, if the height of your');
            writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
            writeln('        you may need to rescale tip values.');
            writeln;
            write('Press a key to exit the program');
            userChar := readkey;
            halt
         End; {else}
      End
      Else Begin
         clrscr;
         writeln('Error:  PDERROR tried to write a number with an absolute value larger than');
         writeln('        9.9999999999E+99.  The program will now exit.  You should go into');
         writeln('        PDTREE and rescale your tree, for example, if the height of your');
         writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
         writeln('        you may need to rescale tip values.');
         writeln;
         write('Press a key to exit the program');
         userChar := readkey;
         halt
      End; {else}
   End; {procedure extendwrite}

{***************************************************************************}
{This procedure writes a line of data to the branch length transforms file.}
{***************************************************************************}
Procedure writeXFormResults(T1,T2 :real;
                            type1,type2 :char;
                            Var branchFile :text);
   Begin
      write(branchFile,type1,'   ',T1:4:2);
      If ((T1 >= 0) And (T1 < 10.0)) Then write(branchFile,' ');
      writeln(branchFile,'    ',type2,'   ',T2:4:2)
   End; {procedure writexformresults}

{***************************************************************************}
{This procedure writes a line of statistical values to the file.}
{***************************************************************************}
Procedure write1TreeStats(Var statFile :text;
                          simStats :statPak;
                          setNumber :integer;
                          corrA :corrArray;
                          root :nodePtr;
                          tipStats :statPak);
   {**********************************************************************}
   {Add extra space for small integers.}
   {**********************************************************************}
   Procedure padInteger(Var outFile :text;
                        iValue :integer);
      Begin
         If iValue<10 Then write(outFile,' ');
         If iValue<100 Then write(outFile,' ');
      End; {procedure padInteger}
   {**********************************************************************}
   Begin
      Write(statFile,setNumber,' ');              {Column 1 is set number}
      If (setNumber < 100)
      Then write(statFile,' ');
      If (setNumber < 10)
      Then write(statFile,' ');
      If not(simStats.badStats) Then Begin
         extendWrite(statFile,simStats.Corr);        {Column 2 is correlation}
         extendWrite(statFile,simStats.ls);          {Column 3 is OLS slope}
         extendWrite(statFile,simStats.rma);         {Column 4 is reduced MA slope}
         extendWrite(statFile,simStats.ma);          {Column 5 is major axis slope}
         extendWrite(statFile,simStats.variance[1]); {Column 6 is Var(x)}
         extendWrite(statFile,simStats.variance[2]); {Column 7 is Var(y)}
         extendWrite(statFile,simStats.f);           {Column 8 is F statistic}
      End
      Else Begin
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
      End; {else}
      padInteger(statFile,simStats.conypos);
      write(statFile,simStats.conypos,' ');       {Column 9 is # of Y SDC > 0}
      padInteger(statFile,simStats.conyneg);
      write(statFile,simStats.conyneg,' ');       {Column 10 is # of Y SDC < 0}
      padInteger(statFile,simStats.cony0);
      write(statFile,simStats.cony0,' ');         {Column 11 is # of Y SDC = 0}
      extendWrite(statFile,corrA[11]); {Column 12 is ABS residuals SDC X on Y vs Height (Tr 1) (screen 11)}
      extendWrite(statFile,corrA[1]);  {Column 13 is corr T1 vs SQRT(Sum BL) (screen 1)}
      extendWrite(statFile,corrA[2]);  {Column 14 is corr T2 vs SQRT(Sum BL) (screen 2)}
      extendWrite(statFile,corrA[3]);  {Column 15 is corr SDC(T1) vs base values (screen 3)}
      extendWrite(statFile,corrA[4]);  {Column 16 is corr SDC(T2) vs base values (screen 4)}
      extendWrite(statFile,corrA[5]);  {Column 17 is corr SDC(T1) vs Height (screen 5)}
      extendWrite(statFile,corrA[6]);  {Column 18 is corr SDC(T2) vs Height (screen 6)}
      extendWrite(statFile,corrA[7]);  {Column 19 is corr Est(Node(T1)) vs Height (screen 7)}
      extendWrite(statFile,corrA[8]);  {Column 20 is corr Est(Node(T2)) vs Height (screen 8)}
      extendWrite(statFile,root^.dat[1]);   {Column 21 is root value of trait 1 after best xform}
      extendWrite(statFile,root^.dat[2]);   {Column 22 is root value of trait 2 after best xform}
      extendWrite(statFile,tipStats.mean[1]); {Column 23 is simple tip mean of trait 1}
      extendWrite(statFile,tipStats.mean[2]); {Column 24 is simple tip mean of trait 2}
      writeln(statFile)
   End; {procedure write1treestats}

{***************************************************************************}
{ Write a line of "two tree" statistical results to an output file.        }
{***************************************************************************}
Procedure write2TreeStats(Var StatFile :text;
                          simStats :statPak;
                          setNumber :integer;
                          corrA :CorrArray;
                          t1,t2 :extended;
                          type1,type2 :char;
                          root1 :nodeptr;
                          root2 :nodeptr;
                          tipStats :statPak);
   {*********************************************************************}
   {Add extra space for small integers.}
   {*********************************************************************}
   Procedure padInteger(Var outFile :text;
                        iValue :integer);
      Begin
         If iValue<10 Then write(outFile,' ');
         If iValue<100 Then write(outFile,' ');
      End; {procedure padInteger}
   {*********************************************************************}
   Begin {procedure write2TreeStats}
      write(statFile,setNumber,' ');              {Column 1 is set number}
      If (setNumber < 100)
      Then write(statFile,' ');
      If (setNumber < 10)
      Then write(statFile,' ');
      If not(simStats.badStats) Then Begin
         extendWrite(statFile,simStats.Corr);        {Column 2 is correlation}
         extendWrite(statFile,simStats.ls);          {Column 3 is OLS slope}
         extendWrite(statFile,simStats.rma);         {Column 4 is reduced MA slope}
         extendWrite(statFile,simStats.ma);          {Column 5 is major axis slope}
         extendWrite(statFile,simStats.variance[1]); {Column 6 is Var(x)}
         extendWrite(statFile,simStats.variance[2]); {Column 7 is Var(y)}
         extendWrite(statFile,simStats.f);           {Column 8 is F statistic}
      End
      Else Begin
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
         write(statFile,'-9                ');
      End; {else}
      padInteger(statFile,simStats.conypos);
      write(statFile,simStats.conypos,' ');       {Column 9 is # of Y SDC > 0}
      padInteger(statFile,simStats.conyneg);
      write(statFile,simStats.conyneg,' ');       {Column 10 is # of Y SDC < 0}
      padInteger(statFile,simStats.cony0);
      write(statFile,simStats.cony0,' ');         {Column 11 is # of Y SDC = 0}
      extendWrite(statFile,corrA[11]); {Column 12 is ABS residuals SDC X on Y vs Height (Tr 1) (screen 11)}
      extendWrite(statFile,corrA[1]);  {Column 13 is corr T1 vs SQRT(Sum BL) (screen 1)}
      extendWrite(statFile,corrA[2]);  {Column 14 is corr T2 vs SQRT(Sum BL) (screen 2)}
      extendWrite(statFile,corrA[3]);  {Column 15 is corr SDC(T1) vs base values (screen 3)}
      extendWrite(statFile,corrA[4]);  {Column 16 is corr SDC(T2) vs base values (screen 4)}
      extendWrite(statFile,corrA[5]);  {Column 17 is corr SDC(T1) vs Height (screen 5)}
      extendWrite(statFile,corrA[6]);  {Column 18 is corr SDC(T2) vs Height (screen 6)}
      extendWrite(statFile,corrA[7]);  {Column 19 is corr Est(Node(T1)) vs mean(daughter height) (screen 7)}
      extendWrite(statFile,corrA[8]);  {Column 20 is corr Est(Node(T2)) vs mean(daughter height) (screen 8)}
      write(statFile,type1,' ');       {Column 21 is best transform type for trait 1}
      If T1 <> -9 
      Then Begin                       {Column 22 is best transform factor for trait 1}
         write(statFile,T1:4:2,' ');
         If ((T1 >= 0) And (T1 < 10.0))   {Right pad column 22 if needed}
         Then write(statFile,' ');
      End {then}
      Else write(statFile,'-9.0  ');
      write(statFile,type2,' ');       {Column 23 is best transform type for trait 2}
      If T2 <> -9
      Then Begin                       {Column 24 is best transform factor for trait 2}
         write(statFile,T2:4:2,' ');      
         If ((T2 >= 0) And (T2 < 10.0))
         Then write(statFile,' ');           {Left pad column 24 if needed}
      End {then}
      Else write(statFile,'-9.0  ');
      extendWrite(statFile,root1^.dat[1]);   {Column 25 is root value of trait 1 after all operations}
      extendWrite(statFile,root2^.dat[2]);   {Column 26 is root value of trait 2 after all operations}
      extendWrite(statFile,tipStats.mean[1]); {Column 27 is simple tip mean of trait 1 after all operations}
      extendWrite(statFile,tipStats.mean[2]); {Column 28 is simple tip mean of trait 2 after all operations}
      writeln(statFile)
   End; {procedure write2TreeStats}

{***************************************************************************}
{This procedure writes a tree to an open file in .PDI format.
It only requires the tree and the file pointer.}
{***************************************************************************}
Procedure writeTweakedTree(rt :nodeptr;var treeFile :text);
   Var
      num :integer;
   {**************************************************************************}
   {This recursive procedure returns the number of tips descended from the node
   it is given.  It calls nothing but itself.}
   {**************************************************************************}
   Procedure numtps(nod:nodeptr; var num:integer);
      Begin
      If nod^.sibL<>nil Then Begin
         numtps(nod^.sibL,num);
         numtps(nod^.sibR,num);
      End {begin}
      Else inc(num);
   End; {procedure writeTweakedtree}
   {*********************************************************************}
   {This procedure writes the tips data to the file in the PDI format
   recursively.}
   {*********************************************************************}
   Procedure wrttips(nd:nodeptr;var fl:text);
      Begin
         If nd^.sibL=nil Then Begin
            write(fl,nd^.name,'  ');
            extendWrite(fl,nd^.dat[1]);
            write(fl,'  ');
            extendWrite(fl,nd^.dat[2]);
            writeln(fl);
         End
         Else Begin
            wrttips(nd^.sibL,fl);
            wrttips(nd^.sibR,fl);
         End {else}
      End; {procedure wrttips}
   {************************************************************************}
   {You may have noticed that this series of procedures is analogous to the
   ones in the read file block.  This one is in charge of writing the tree
   to the file.  One difference is that this uses recursive procedures of
   its own to do a lot of the work.}
   {************************************************************************}
   {*********************************************************************}
   {This procedure writes the tree structure to a file in the PDI form
   recursively.}
   {*********************************************************************}
   Procedure wrtTree(nd:nodePtr; Var fl:text);
      Begin
         write(fl,nd^.name,'  ');
         If nd^.ancestor=nil
         Then write(fl,'$$')
         Else write(fl,nd^.ancestor^.name);
         write(fl,'  ');
         extendWrite(fl,nd^.length);
         writeln(fl);
         If nd^.sibL<>nil Then Begin
            wrttree(nd^.sibL,fl);
            wrttree(nd^.sibR,fl);
         End {then}
      End; {procedure wrtTree}
      {*********************************************************************}
   Begin {writetweakedtree}
      num:=0;
      numtps(rt,num);
      writeln(treeFile,num);
      writeln(treeFile);
      wrttree(rt,treeFile);
      writeln(treeFile);
      wrttips(rt,treeFile);

      {add a blank line between trees}
      writeln(treeFile);
   End; {procedure writetweakedtree}

{***************************************************************************}
{This procedure opens up the sim file, and scans ALL the simulated tip      }
{values to determine the maximum and minimum for each trait.  This ugly     }
{operation is necessary for setting category bin widths automatically.      }
{Even if the user wants to set widths manually, they still need to see      }
{this information.  This scanning occurs after the tree is read, but        }
{before any sets of simulated data have been processed.  Because the tree   }
{should be small relative to the heap, I took a chance that the large       }
{file buffer (65K) this procedure uses could be heap allocated.  If this    }
{assumption blows up in the future, you can either make the buffer smaller  }
{or try allocating it somewhere else (e.g., the stack).  I sincerely hope   }
{neither option will be necessary as this procedure is already a disk       }
{abuser.                                                                    }
{***************************************************************************}
Procedure scanForLimits(inputFLName : flString;
                        Var globalMin : traits;
                        Var globalMax : traits;
                        tipCount : integer;
                        userLimit : integer);
   Const smallValue = -9.999E99;
         largeValue = 9.999E99;
   Var inFile : text;
       i : longInt;
       lineLimit : longInt;
       fileBuff : buffPtr;
       ch4 : string[4];
       tempName : nameString;
       tempVal1 : extended;
       tempVal2 : extended;
       ch : char;
   Begin
      lineLimit := tipCount;
      lineLimit := lineLimit*userLimit;
      globalMin[1] := largeValue;
      globalMax[1] := smallValue;
      globalMin[2] := largeValue;
      globalMax[2] := smallValue;
      new(fileBuff);
      assign(inFile,inputFLName);
      setTextBuf(infile,fileBuff^);
      reset(inFile);
      Repeat
         read(inFile,ch4);
         readln(inFile)
      Until ch4='(Nod';
      Repeat
         read(inFile,ch4);
         readln(inFile);
      Until ch4='SIMU';
      Repeat
         read(inFile,ch);
         readln(inFile)
      Until ch='_';
      i := 1;
      While ((not eof(inFile)) And (i <= lineLimit)) Do Begin
         read(inFile,tempName);
         read(inFile,tempVal1);
         read(inFile,tempVal2);
         readln(inFile);
         If (tempVal1 < globalMin[1]) Then globalMin[1] := tempVal1;
         If (tempVal1 > globalMax[1]) Then globalMax[1] := tempVal1;
         If (tempVal2 < globalMin[2]) Then globalMin[2] := tempVal2;
         If (tempVal2 > globalMax[2]) Then globalMax[2] := tempVal2;
         inc(i);
      End; {while}
      close(inFile);
      dispose(fileBuff);
   End; {procedure scanForLimits}

End. {module e_io.pas}
