{********************************************************************}
{E_UTILS.PAS
 utilities for PDERROR.PAS (mostly copied and modified from
 T_UTILS.PAS (for PDTREE).
 This module copyright 1996.
 Written by Peter E. Midford for Theodore Garland, Jr., with
 extensive copying from PDTREE.PAS.
 Last modified 18 November 1999.}
{********************************************************************}

unit e_utils;
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Interface

   Const maxTips = 256;

   Type
      flstring = string[30];
      typstring = string[3];
      namestring = string[2];
      traits = array[1..2] of extended;
      itraits = array[1..2] of integer;
      traitFlags = set of 1..2;

      statpak = Record
         min      :traits;
         max      :traits;
         mean     :traits;
         variance :traits;
         cov      :extended;
         corr     :extended;
         rma      :extended;
         rmay     :extended;
         ma       :extended;
         may      :extended;
         ls       :extended;
         lsy      :extended;
         r2       :extended;
         sumresid :extended;
         ts       :extended;
         f        :extended;
         n1       :integer;      {which trait is on horizontal axis}
         n2       :integer;      {which trait is on vertical axis}
         number   :integer;
         df       :integer;
         conypos  :integer;
         cony0    :integer;
         conyneg  :integer;
         badStats :boolean;
      End; {record statpak}

      nodeptr = ^node;
      node = Record
         Ancestor :nodeptr;
         dat      :traits;
         contrast :traits;
         temp     :extended;
         temp2    :extended;
         sdcon    :extended;
         variance :extended;
         length   :extended;
         SibL     :nodeptr;
         SibR     :nodeptr;
         tipL     :nodeptr;
         tipR     :nodeptr;
         tipnum   :integer;
         name     :namestring;
         tempx    :integer;
         tempy    :integer;
      End; {record node}


      nodedataptr = ^nodedata;
      nodedata = Record            {Simulation sets are loaded into lists}
         name     :nameString;     {The name of the node}
         data     :traits;         {The simulated data for that node}
         ansName  :nameString;
         ansData  :traits;
         rank     :traits;
         mark     :boolean;        {used in ranking process}
         next     :nodeDataPtr     {next element in the list}
      End; {record}

      CorrArray = array[1..11] of extended; {This is shared by main and i/o }

      ConArray = Array[1..maxTips] of extended; {Array for sorting and KS test}

      tweakType = (FIXED,         {Ways to modify branch lengths: fixed}
                   NORMAL);       {or drawn from a normal distribution}

      tweakRecord = Record        {Holds all the information for modifying lengths}
         deviation   :extended;   {Amount or mean amount of change}
         width       :extended;   {Coefficient of variance for Normal tweaks}
         zeroLength  :extended;   {Value to set nonpositive branch lengths to}
         tipLength   :extended;   {Value to set zero tip branch lengths to}
         mytype      :tweakType   {FIXED=uniform dist.; NORMAL=normal dist.}
      End; {type tweakrecord}

      tweakPtr = ^tweakRecord;

                                  {This type indicates what to minimize in the
                                   search for branch lenght transforms.  Since
                                   PDERROR seeks the minimum value, it's a cost.}

      costType = (SdcBl,          {ABS(SDC) vs sqrt(BL1+BL2); PDTREE screens 1&2, original cost}
                  SdcNv,          {ABS(SDC) vs est Nodal value; PDTREE screens 3&4}
                  SdcHt,          {ABS(SDC) vs base node height; PDTREE screen 5&6}
                  NvHt);          {Node value vs base node height; PDTREE screen 7&8}

      {the ordering of the following may be counterintuitive, but the compiler}
      {seems have an alignment problem sometimes, so it's best to keep fields in}
      {decreasing order.}

      xFormRecord = Record     {Holds information about user transform choices}
         rhoStart  :extended;  {Bracket values for the Rho transform}
         rhoEnd    :extended;
         expStart  :extended;  {Bracket values for the exponential transform}
         expEnd    :extended;
         cost      :costType;  {what to minimize}
         doExps    :boolean;   {Flag for finding optimum exp xform}
         doLog     :boolean;   {Flag for including log in exp xforms}
         doRho     :boolean;   {Flag for finding optimum rho xform}
         doGrafen  :boolean;   {Flag for trying Grafen's arbitrary bl transform}
         doNee     :boolean;   {Flag for trying Nee's arbitrary bl transform}
         doPagel   :boolean;   {Flag for trying Pagel's arbitrary bl transform}
         doSelf    :boolean;   {Flag for trying no bl transform}
         doOne     :boolean;   {Flag for trying all branches set to one}
         doBLStats :boolean;   {Flag for saving statistics based on best bl xform to file}
      End; {record}

      xFormPtr = ^xFormRecord;

Procedure initMe(me :nodePtr);                  {initialize a node}
Procedure InitNodeData(theData :nodeDataPtr);   {initialize a data pair}
Procedure killTree(Var nod :nodePtr);           {recursively unlink and remove a node}
Procedure killData(theData :nodeDataPtr);       {unlink and remove a data pair}
Procedure finishStat(Var stpk :statPak);        {do regression stats on sums in a statPak}
Procedure finishstatorg(var stpk:statpak);      {as above except regress through origin}
Procedure CheckName(nod :nodeptr;               {check for duplicate names}
                    name :namestring;
                    Var notFound :boolean;
                    Var want :nodePtr);
Procedure GetUnivariates(Var z1,z2 :extended);
Procedure NormalPair(Var g, h :extended;        {a pair of random #'s}
                     MeanA, MeanB :extended;    {the means of the distributions}
                     Cov, VarA, VarB :extended); {the covariance and
                                                  variances of the distributions}
Procedure printTree(var nod:nodeptr);          {a debugging procedure}

Implementation


{***********************************************************************}
{This procedure initializes some of the fields of a node to 0.0 or nil.
 It is called just about whenever the program allocates a node.  It calls
 nothing.}
{***********************************************************************}
Procedure initMe(me:nodeptr);
   Begin
      me^.dat[1] := 0.0;
      me^.dat[2] := 0.0;
      me^.sibL := nil;
      me^.sibR := nil;
      me^.ancestor := nil  {added PEM 10 Oct 1995}
   End; {procedure initme}

{***********************************************************************}
{This recursive procedure deallocates the linked list implemented tree.
This is done to keep the heap neat.  It calls nothing but itself.}
{***********************************************************************}
Procedure killtree(var nod:nodeptr);
   Begin
      If nod^.sibL<>nil Then killTree(nod^.sibL);
      If nod^.sibR<>nil Then killTree(nod^.sibR);
      dispose(nod)
   End; {procedure killtree}

{***********************************************************************}
{This procedure is used for debugging.}
{***********************************************************************}
Procedure printTree(var nod:nodeptr);
   Begin
      If nod^.sibL<>nil Then Begin
         write(' ');
         printTree(nod^.sibL);
      End; {then}
      If nod^.sibR<>nil Then Begin
         write(' ');
         printTree(nod^.sibR);
      End; {then}
      writeln(nod^.name);
   End; {procedure printTree}

{***********************************************************************}
{This initializes some of the fields of the element holding simulated
data from a single tree node.  It should be called whenever one of these
elements is allocated.  It calls nothing.}
{***********************************************************************}
Procedure initNodeData(thedata :nodedataptr);
   Begin
      thedata^.next := NIL;
      thedata^.data[1] := 0.0;
      thedata^.data[2] := 0.0;
      thedata^.name := '  ';
   End; {procedure initnodedata}

{***********************************************************************}
{This recursive procedure deallocates the Simulated data list.
This is done to keep the heap neat.  It calls nothing.}
{***********************************************************************}
Procedure KillData(theData :NodeDataPtr);
   Var current : nodedataptr;
       temp : nodedataptr;
   Begin
      current := theData;
      While (current <> NIL) Do Begin
         temp := current^.next;
         dispose(current);
         current := temp;
      End {while}
   End; {Procedure KillData}

{*************************************************************************}
{This recursive procedure checks a name that is passed with the rest of
the tree to see if it is in there.  If it is, it sets notfound to false.
It calls nothing but itself.}
{*************************************************************************}
Procedure CheckName(nod :nodePtr;
                    name :nameString;
                    Var notFound :boolean;
                    Var want :nodeptr);
   Begin
      If nod^.name=name Then Begin
         notfound := false;
         want := nod
      End
      Else Begin
         If (nod^.sibL<>nil) and notfound then
            checkname(nod^.sibL,name,notfound,want);
         If (nod^.sibR<>nil) and notfound then
            checkname(nod^.sibR,name,notfound,want)
      End; {else}
   End; {procedure checkname}

{*************************************************************************}
{This calculates the rma and the y intercept for rma called rmay, ma and
the y intercept for the ma called may, r squared called r2, t statistic
called ts,the f statistic, and the degrees of freedom called df.}
{*************************************************************************}
Procedure finishStat(Var stpk :statpak);
   Begin
      If stpk.variance[stpk.n1]>0
      Then stpk.rma:=sqrt(stpk.variance[stpk.n2]/stpk.variance[stpk.n1])
      Else stpk.rma:=0;
      If ((stpk.corr>0) And (stpk.rma<0)) Or
           ((stpk.corr<0) And (stpk.rma>0))
      Then stpk.rma:=stpk.rma*-1;
      stpk.rmay:=stpk.mean[stpk.n2]-stpk.rma*stpk.mean[stpk.n1];

      stpk.ma:=sqr(stpk.variance[1]+stpk.variance[2]) -
           4*(stpk.variance[2]*stpk.variance[1] - sqr(stpk.cov));
      stpk.ma:=0.5*(stpk.variance[2]+stpk.variance[1]+
           sqrt(stpk.ma));
      if (stpk.ma-stpk.variance[stpk.n2]<>0)
      Then stpk.ma:=stpk.cov/(stpk.ma-stpk.variance[stpk.n2])
      Else stpk.ma:=0;
      stpk.may:=stpk.mean[stpk.n2]-stpk.ma*stpk.mean[stpk.n1];

      stpk.r2:=sqr(stpk.corr);

      stpk.ts:=stpk.corr*sqrt((stpk.number-2)/(1-stpk.r2));
      stpk.f:=sqr(stpk.ts);
      stpk.df:=stpk.number-2;
   End; {procedure finishstat}

{*************************************************************************}
{This is similar to the procedure above, except the rma and ma are forced
to go through the origin.
This calculates the rma and the y intercept for rma called rmay, ma and
the y intercept for the ma called may, r squared called r2, t statistic
called ts, f statistic, and the degrees of freedom called df.}
{*************************************************************************}
Procedure finishStatOrg(Var stpk :statpak);
   Begin
      If (stpk.corr <> 0)
      Then stpk.rma:=stpk.ls/stpk.corr
      Else stpk.rma := 0;
      If ((stpk.corr>0) And (stpk.rma<0)) Or
          ((stpk.corr<0) And (stpk.rma>0))
      Then stpk.rma:=stpk.rma*-1;
      stpk.rmay:=0.0;

      stpk.ma:=sqr(stpk.variance[1]+stpk.variance[2]) -
           4*(stpk.variance[2]*stpk.variance[1] - sqr(stpk.cov));
      If stpk.ma < 0 Then stpk.ma := 0;                  {bogus data check}
      stpk.ma:=0.5*(stpk.variance[2]+stpk.variance[1]+
           sqrt(stpk.ma));
      If (stpk.ma = stpk.variance[stpk.n2])
      Then stpk.ma := -9.9999999999E-9
      Else stpk.ma:=stpk.cov/(stpk.ma-stpk.variance[stpk.n2]);
      stpk.may:=0.0;

      stpk.r2:=sqr(stpk.corr);

      If (stpk.r2 >= 1.0) Then Begin
         stpk.ts := -9.9999999999E-9;
         stpk.f := -9.9999999999E-9
      End
      Else Begin
         stpk.ts:=stpk.corr*sqrt((stpk.number-1)/(1-stpk.r2));
         stpk.f:=sqr(stpk.ts)
      End;
      stpk.df:=stpk.number-2
   End; {procedure finishstatorg}

{********************************************************************}
{ This is the pair of normal random numbers from CMSIMUL, that allow
{ specifying means, variances and covariance. (Name was GetChanges.) }
{********************************************************************}
Procedure GetUnivariates(Var z1,z2:extended);
{Draws two normal numbers from N(0,1)}
   Var x,y,Theta,R: extended;
   Begin
      y := Random;
      x := Random;
      Theta := 2*PI*x;
      R := sqrt(-2*(ln(y)));
      z1 := R*Cos(Theta);
      z2 := R*Sin(Theta)
   End; {Procedure GetUnivariates}

Procedure NormalPair(var g,h:extended;MeanA,MeanB,Cov,VarA,VarB:extended);
   Var Z1,Z2,rho :extended;
   Begin
      GetUnivariates(Z1,Z2);
      G := sqrt(VarA)*Z1 + MeanA;
      if (Cov <> 0.0) then begin
         rho := Cov/(sqrt(VarA)*Sqrt(VarB));
         H := sqrt(VarB)*((rho*Z1) + ((sqrt(1-(rho*rho))) * Z2)) + MeanB
      end {then begin}
      else H := sqrt(VarB)*Z2 + MeanB
   End; {Procedure NormalPair}


End. {Module E_Utils}
