{This unit handles reading the nexus file and writing a pdi file.  If     }
{the capabilty to write nexus files is added, it will be here as well     }

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}


Unit n_io1;

Interface

Uses n_util1;  {nexus specific type defintitions}


   Procedure ReadNexusFile(Var contents: nexusPtr;
                           Var flname :flstring;
                           Var ext :typstring;
                           Var rootName :flString;
                           Var succ:boolean);

   Procedure saveNexusToPdi(contents :nexusPtr;
                            Var fleName: flString;
                            Var rootName: flString;
                            characterBlockIndex : integer;
                            treeIndex : integer;
                            ok: boolean);


Implementation

uses crt;      {Borland low-res text screen utils}

Const
   tabChar         = chr(9);
   eofChar         = chr(26);
   crChar          = chr(10);
   lfChar          = chr(13);
   spaceChar       = ' ';
   semicolonChar   = ';';
   colonChar       = ':';
   commaChar       = ',';
   leftParenChar    = '(';
   rightParenChar   = ')';
   leftBracketChar  = '[';
   rightBracketChar = ']';
   quoteChar        = '''';
   doubleQuoteChar  = '"';
   minusChar        = '-';
   equalsChar       = '=';
   digitChar        = ['0','1','2','3','4','5','6','7','8','9'];
   whiteSpace = [spaceChar,lfChar,crChar,tabChar];

   terminator = [semicolonChar,
                 colonChar,
                 rightParenChar,
                 leftParenChar,
                 commaChar,
                 equalsChar];
   quoteChars = [quoteChar,doubleQuoteChar];

Type
     tokenType = (NOTOKEN,
                  NEXUS,
                  NCOMMENT,
                  NENDCOMMENT,
                  NBEGIN,
                  NEND,
                  NDATA,
                  NTAXA,
                  NSEMICOLON,
                  NDIMENSIONS,
                  NEQUALS,
                  NNTAXA,
                  NNCHAR,
                  NFORMAT,
                  NSYMBOLS,
                  NMISSING,
                  NGAP,
                  NMATRIX,
                  NCOLON,
                  NTAXLABELS,
                  NASSUMPTIONS,
                  NOPTIONS,
                  NDEFTYPE,
                  NCHARACTERS,
                  NEXSET,
                  NCHARSET,
                  NTAXSET,
                  NTREES,
                  NTRANSLATE,
                  NCOMMA,
                  NTREE,
                  NUTREE,
                  NROOTED,
                  NUNROOTED,
                  NLEFTPAREN,
                  NRIGHTPAREN,
                  NFLOAT,
                  NINTEGER,
                  NSTRING,
                  NID);
     stateType = (NONEXUS,
                  NOBLOCK,
                  BLOCKNAME,
                  INDATA,
                  INTAXA,
                  INDIMENSIONS,
                  INNTAXA,
                  INNCHAR,
                  INFORMAT,
                  INSYMBOLS,
                  INMATRIX,
                  INTNTAXA,
                  INTDIMENSIONS,
                  INTAXLABELS,
                  INEXSET,
                  INCHARSET,
                  INTAXSET,
                  INTREES,
                  INTREE,
                  INUTREE,
                  TREELABEL,
                  UTREELABEL,
                  BRANCHLENGTH,
                  INTRANSLATE,
                  INTRANSLATEPAIR);

     treeReadState = (READLEFT,READRIGHT);

     statePtr = ^parserState;
     parserState = Record
                      translateList : translatePairPtr;
                      currentTreeBlock : treeBlockPtr;
                      currentTree : treeEntryPtr;
                      currentNode : nnodePtr;
                      lastNode : nnodePtr;
                      lastInteger : integer;
                      lastFloat : extended;
                      entry_counter : integer;   {general purpose counter}
                      curState : stateType;
                      token : tokenType;
                      nextString: string;
                      treeState : treeReadState;
                      rootName : String;
                   End;



{************************************************}
{ Ersatz constructor function.                   }
{************************************************}
Function parserState_new(rootName : string) : statePtr;
   Var
      result : statePtr;
   Begin
      result := new(statePtr);
      result^.curState := NONEXUS;
      parserState_new := result;
      result^.translateList := nil;
      result^.rootName := rootName;
   End; {function parserState_new}

{***********************************************************************}
{ A disposal function for the parser.                                   }
{***********************************************************************}
Procedure parserState_dispose(toKill : statePtr);
   Begin
      dispose(toKill);
   End; {procedure parserState_dispose}

{***********************************************************************}
{ This procedure scans the symbol list.  The scanning removes whitespace}
{ and prompts the user for mapping options based on what kind of        }
{ characters are in the symbol set.                                     }
{***********************************************************************}
Procedure processSymbolList(Var contents: nexusPtr;
                            Var parser : statePtr);
  Var
    data : charBlockPtr;
    symbolList : string;
    digitFlag : boolean;
    i : integer;
    choice : char;
    foo : set of char;
    raw : string;
  Begin
     data := contents^.characters;
     digitFlag := true;
     raw := parser^.nextString;
     foo := [];
     For i := 2 to length(raw)-1 Do Begin
        If not(raw[i] in whiteSpace)
        Then Begin
           data^.symbols := data^.symbols + raw[i];
           If not(raw[i] in digitChar)
           Then digitFlag := false;
        End;
     End;
     If (digitFlag)
     Then Begin
        writeln('While processing a character or data block, PDNEXUS encountered');
        writeln('a definition for a set of symbols.  All symbols in the set are');
        writeln('single digits.  For purposes of assigning (provisional) tip values');
        writeln('PDNEXUS can code these (directly) as numeric values or (ordinally)');
        writeln('as their position in the set listing.');
        writeln;
        write('Do you want to code them directly or ordinally?(O/D(default)): ');
        readln(choice);
        If (choice = 'O') Or (choice = 'o')
        Then data^.mapDigits := true
        Else data^.mapDigits := false;
     End; {if digitFlag}
     If (not (digitFlag) or data^.mapDigits)
     Then Begin
        writeln('Symbols will be coded ordinally.  Shall the first symbol be coded');
        write('as 0 or 1? (0/1=default): ');
        readln(choice);
        If (choice = '0')
        Then data^.zeroBased := true
        Else data^.zeroBased := false;
     End; {then}
  End; {procedure processSymbolList}

{***********************************************************************}
{ This procedure finishes up a translate block by allocating the data   }
{ structure for the block, assigning the list of translations to the    }
{ block and connecting the new block to the file's nexus structure.     }
{***********************************************************************}
Procedure finishTranslate(Var nexusBlock: nexusPtr;
                          Var parser : statePtr);
   Var newTranslate : translateBlockPtr;
   Begin
      newTranslate := translateBlock_new;
      newTranslate^.tlength := parser^.entry_counter;
      newTranslate^.data := parser^.translateList;
      newTranslate^.next := nexusBlock^.translations;
      nexusBlock^.translations := newTranslate;
      nexusBlock^.selectedTranslateBlock := newTranslate;
   End; {procedure finishTranslate}

{************************************************************************}
{ A translate pair represents a single integer-taxon mapping in the      }
{ translate block.  This creates one and adds it to the list that        }
{ represents the block.                                                  }
{************************************************************************}
Procedure addTranslatePair(Var parser: statePtr;
                               context : nexusPtr;
                               index : integer;
                               id : string);
   Var newpair : translatePairPtr;
   Begin
      new(newpair);
      newpair^.index := index;
      newPair^.curTaxa := context^.taxa;
      If (newPair^.curTaxa = NIL)
      Then Begin
         writeln('Id = ', id, '; index = ', index);
         pdnexus_error('Null taxa block for translate pair');
      End;
      newpair^.taxonindex := lookupTaxon(newpair^.curTaxa,id);
      newpair^.next := parser^.translateList;
      parser^.translateList := newpair;
      inc(parser^.entry_counter);
   End; {procedure addTranslatePair}

{************************************************************************}
{This creates a new tree block to serve as a container for a set of trees}
{that should appear in the file.                                         }
{************************************************************************}
Procedure startTreeBlock(parser : statePtr;
                         id : string;
                         contents : nexusPtr);
   Var newTreeBlock : treeBlockPtr;
   Begin
      newTreeBlock := treeBlock_new;
      newTreeBlock^.name := id;
      newTreeBlock^.next := contents^.trees;
      contents^.trees := newTreeBlock;
      parser^.currentTreeBlock := newTreeBlock;
   End;

{************************************************************************}
{This sets up the tree block and allocates and attaches the root node.   }
{************************************************************************}
Procedure startTree(parser : statePtr;
                    id : string;
                    block : treeBlockPtr;
                    rooted : boolean);
   Var newTree : treeEntryPtr;
   Begin
      treeBlock_addTree(parser^.currentTreeBlock,
                        id,
                        rooted,
                        parser^.currentTree);
      parser^.currentTree^.tree^.length := 0;
      parser^.currentNode := parser^.currentTree^.tree;
   End;

{************************************************************************}
{Creating a node in the tree and attaching it appropriately.             }
{************************************************************************}
Procedure addNode(parser : statePtr);
   Var newNode : nnodePtr;
   Begin
      If (parser^.currentTree^.tree^.ancestor <> nil)
      Then pdnexus_error('Bug check AddNode - root ancestor is non-nil');
      If (parser^.currentTree = nil)
      Then pdnexus_error('Tried to add a node to an empty tree')
      Else Begin
         newNode := nnode_new;
         parser^.currentNode^.sibL := newNode;
         newNode^.ancestor := parser^.currentNode;
      End;
      parser^.currentNode := newNode;
   End; {procedure addNode}
      
{************************************************************************}
{ Lookup the index of a taxon (in a taxon block) from the index in a     }
{ translate block.  These are not necessarily the same.                  }
{************************************************************************}
   Function findTaxon(transBlock : translateBlockPtr;
                      key : integer) : integer;
      Var
         current : translatePairPtr;
      Begin
         current := transBlock^.data;
         While ((current <> NIL) And
                (current^.index <> key)) Do current := current^.next;
         If (current <> NIL)
         Then findTaxon := current^.taxonIndex
         Else findTaxon := -1;
      End;

{************************************************************************}
{ This function extracts the nth 'dark' character from the data string.  }
{ It isn't just an index because there may be whitespace in the string.  }
{************************************************************************}
   Function extractSymbol(data : string; num : integer) : char;
     Var count : integer;
         index : integer;
     Begin
        count := 0;
        index := 0;
        While (index <= length(data)) And (count < num) Do Begin
          inc(index);
          If not (data[index] in whitespace)
          Then inc(count);
        End; {while}
        If (count = num)
        Then extractSymbol := data[index]
        Else extractSymbol := chr(127);
     End; {function extractSymbol}

{************************************************************************}
{ This function just wraps the pos library function.                     }
{************************************************************************}
   Function extractPosition(symbols : string; token : char) : extended;
     Var key : String[1];
         index : integer;
     Begin
        key := ''+ token;
        index := pos(key,symbols);
        If index = 0
        Then extractPosition := -9.9
        Else extractPosition := 1.0*index;
     End; {function extractPosition}

{************************************************************************}
{ This returns a trait value for a tip, using the trait index, the data  }
{ string from the row of the character matrix, and the symbols/format    }
{ information in the character block.                                    }
{************************************************************************}
   Function extractValue(chars : charBlockPtr;
                         data : String;
                         traitnum : integer) : extended;
      Var
         result : extended;
         symbolChar : char;
      Begin
         symbolChar := extractSymbol(data,traitnum);
         If not(chars^.mapDigits)
         Then result := 1.0*(ord(symbolChar)-48)
         Else Begin
            result := extractPosition(chars^.symbols,symbolChar);
            If (result >= 0)     {avoid float problems}
            Then If (chars^.zeroBased)
                 Then result := result-1.0;
         End; {else}
         extractValue := result;
      End; {function extractValue}

{************************************************************************}
{************************************************************************}
   Procedure fillTip(parser : statePtr; contents : nexusPtr);
      Var newNode : nnodePtr;
         taxonIndex : integer;
         row : charMatrixRowPtr;
         chars : charBlockPtr;
         dataString : string;
      Begin
         parser^.currentNode^.index := parser^.lastInteger;
         taxonIndex := findTaxon(contents^.translations,parser^.lastInteger);
         If (taxonIndex > -1) Then Begin
             chars := contents^.characters;
             row := chars^.matrix[taxonIndex];
             dataString := row^.dataString;
             parser^.currentNode^.dat[1] := extractValue(chars,dataString,1);
             parser^.currentNode^.dat[2] := extractValue(chars,dataString,2);
         End; {then}
         parser^.lastNode := parser^.currentNode;
         parser^.currentNode := parser^.currentNode^.ancestor;
      End; {procedure fillTip}

{************************************************************************}
{ Fills in tip information when noTranslate block is defined for the     }
{ current tree.                                                          }
{************************************************************************}
   Procedure fillTipNoTranslate(parser : statePtr; contents : nexusPtr);
      Var newNode : nnodePtr;
         taxonIndex : integer;
         row : charMatrixRowPtr;
         chars : charBlockPtr;
         dataString : string;
      Begin
         taxonIndex := lookupTaxon(contents^.taxa,parser^.nextString);
         If (taxonIndex > -1) Then Begin
            chars := contents^.characters;
            row := chars^.matrix[taxonIndex];
            dataString := row^.dataString;
            parser^.currentNode^.dat[1] := extractValue(chars,dataString,1);
            parser^.currentNode^.dat[2] := extractValue(chars,dataString,2);
         End {then}
         Else Begin {something, or many thins are missing}
            If (contents^.taxa = NIL)  Then Begin {no block}
               contents^.taxa := taxaBlock_new;
               contents^.taxa^.limit := 946;  {just a guess}
            End;
            taxonIndex := taxaBlock_addTaxon(contents^.taxa,
                                             parser^.nextString);
         End;
         parser^.currentNode^.index := taxonIndex;
         parser^.lastNode := parser^.currentNode;
         parser^.currentNode := parser^.currentNode^.ancestor;
      End; {procedure fillTipNoTranslate}

{************************************************************************}
{This handles adding a node to a parent whose left daughter is already   }
{occupied.  Note that expressions such as (A,B,C,D) represent polytomies }
{and must generate additional nodes and zero length branches as daughters}
{are added to nodes that already have two - the else clause handles this.}
{************************************************************************}
   Procedure moveToSister (parser : statePtr);
      Var newNode : nnodePtr;
      Begin
         With Parser^ Do Begin
            If (currentNode^.sibR = NIL) Then Begin
               newNode := nnode_new;
               currentNode^.sibR := newNode;
               newNode^.ancestor := currentNode;
            End
            Else Begin {polytomy}
               newNode := nnode_new;
               newNode^.sibL := currentNode^.sibR;
               currentNode^.sibR^.ancestor := newNode;
               newNode^.ancestor := currentNode;
               newNode^.length := 0.0;
               currentNode^.sibR := newNode;
               currentNode := newNode;
               newNode := nnode_new;
               currentNode^.sibR := newNode;
               newNode^.ancestor := currentNode;
            End; {else begin}
            currentNode := currentNode^.sibR;
         End; {with parser^}
      End; {moveToSister}

{************************************************************************}
{ This gets called while reading a tree when the right (close)           }
{ parenthesis is read to make sure that both siblings are filled in.     }
{************************************************************************}
   Procedure finishNode(parser : statePtr);
      Begin
         If (parser^.currentNode <> parser^.currentTree^.tree)
         Then Begin
            If ((parser^.currentNode^.sibR = NIL) And
                (parser^.currentNode^.index = UNINITIALIZED))
            Then Begin
               pdnexus_warning('encountered unbranched internal node');
            End; {then - may add something to deal with this perhaps splice it out}
            parser^.lastNode := parser^.currentNode;  {maybe unnessary but easier}
            parser^.currentNode := parser^.currentNode^.ancestor; {pop up a level}
            If (parser^.currentTree^.tree^.ancestor <>NIL)
            Then pdnexus_error('Bug check in Finish - root ancestor non-NIL');
            If (parser^.lastnode^.length = 0)
            Then finishNode(parser)
         End {then}
         Else {just need to set lastnode so a name can be set}
            parser^.lastNode := parser^.currentNode;
      End; {procedure finishNode}

{**********************************************************************}
{This procedure suggests a name for a tip node.                        }
{**********************************************************************}
   Procedure suggestName(Var nm: pdinamestr;
                             nd : nnodePtr);
      Begin
           {currently a no-op}
      End; {procedure suggestName}


{**********************************************************************}
{This procedure returns a 2 character node name.  It only uses upper
case characters and the digits 0-9.  It is deterministic and simply
returns the first unused name after the "num"th name in a sequence
that starts with A and ends with 9.  It increments num as necessary so
it points to an unexplored name when the procedure returns.}
{***********************************************************************}
   Procedure genName(Var nm : pdinamestr;
                         nd : nnodeptr;
                         rt:nnodeptr;
                     Var num:integer);
      Var
         n1,n2:integer;
         where:nnodeptr;
         ch1:char;
         ch2:char;
         nm2: pdinamestr;
         ok:boolean;
      Begin
         Repeat
            n1:=num div 36;
            n2:=num mod 36;
            ch1 := chr(65+n1);
            If (n2<= 9)
            Then ch2 := chr(48+n2)
            Else ch2 := chr(55+n2);
            nm2 := ch1+ch2;
            inc(num);
            ok:=true;
            {CheckPDIName(rt,nm2,ok,where);}{Uncomment if heuristic names added}
         Until ok;
         nm:=nm2;
      End; {genName}

{***********************************************************************}
{Version of genName for internal nodes, where lowercase is ok.  Split   }
{from genName 21-Feb-2004 PEM.                                          }
{***********************************************************************}
   Procedure genIName(Var nm: pdinamestr; rt:nnodeptr; Var num:integer);
      Var
         n1,n2:integer;
         where:nnodeptr;
         ch1:char;
         ch2:char;
         nm2: pdinamestr;
         ok:boolean;
      Begin
         Repeat
            n1:=num div 36;
            n2:=num mod 36;
            ch1 := chr(97+n1);
            If (n2<= 9)
            Then ch2 := chr(48+n2)
            Else ch2 := chr(87+n2);
            nm2 := ch1+ch2;
            inc(num);
            ok:=true;
            {CheckPDIName(rt,nm2,ok,where);}{no need to check if all generated}
         Until ok;
         nm:=nm2;
      End; {genIName}

{***********************************************************************}
{ This function walks through the tree, generating names for tips and   }
{ internal nodes as appropriate.                                        }
{***********************************************************************}
   Procedure fillPDINames(rt: nnodeptr;
                          nd: nnodeptr;
                          Var base :integer;
                          Var ibase :integer);
      Begin
         If (nd^.sibL <> NIL) Then Begin
            genIName(nd^.pdiname,rt,ibase);
            fillPDINames(rt,nd^.sibL,base,ibase);
            fillPDINames(rt,nd^.sibR,base,ibase);
         End
         Else genName(nd^.pdiname,nd,rt,base);
      End; {procedure fillPDINames}

{***********************************************************************}
{This sets things up for the recursive filePDINames, and sets the       }
{completed tree as the selected tree for the file's contents record.    }
{***********************************************************************}
   Procedure finishTree(Var contents: nexusPtr;
                        Var parser : statePtr);
      Var base : integer;
          ibase : integer;
          fname : flString;
          typn  : typString;
          tree  : nnodePtr;
      Begin
         base := 0;
         ibase := 0;
         tree := parser^.currentTree^.tree;
         fillPDINames(tree,tree,base,ibase);
         {parser^.currentTree^.next := contents.trees;
         contents^.trees := parser^.currentTree;
         contents^.selectedTree := parser^.currentTree;}
      End; {procedure finishTree}

{************************************************************************}
{This procedure reads in a file name.  That is simple, but the task gets
more complex when you require that the program not crash when invalid
junk is typed in by the user.  That is the reason that this routine is
a bit of a mess.  Hopefully you won't have to mess with it.}
{************************************************************************}
   Procedure getInFile(Var fleName: flstring;
                       Var exten:typstring;
                       Var rootName: flString;
                       Var succ:boolean);
      Var fl :text;
          i :integer;
          ch :char;
          place : byte;
          exists, valid:boolean;
          returnToMenu : boolean;
      Begin
         Repeat
            exists:=true;
            valid:=true;
            writeln('Please input the name of the input file with ');
            writeln('the .NEX ending.');
            writeln('   Or enter "R" to exit the program');
            readln(flename);
            returnToMenu := ((fleName ='R') or (fleName='r'));
            If (not returnToMenu) Then Begin
               place:=pos('.',fleName);
               If place = 0          {If no . is found in the filename}
               Then valid:= false
               Else Begin
                  exten:=copy(flename,place+1,3);
                  For i:=1 To 3 Do
                     exten[i]:=upcase(exten[i]);
                  If (exten<>'NEX')
                  Then valid := false              {If extension bad}
                  Else Begin                 {check to see if the file exists}
                     {$I-}
                     assign(fl,flename);
                     reset(Fl);
                     close(Fl);
                     {$I+}
                     exists:= (IOResult=0) and (flename<>'');
                  End; {else}
               End; {else}
               If not exists
               Then writeln(fleName,' does not exist.');
               If not valid
               Then writeln(fleName,' does not have a required .NEX extension.');
            End; {then}
         Until (exists And valid) Or returnToMenu;
         If returnToMenu
         Then succ := false {If user wants to abort}
         Else Begin
            place := pos('.',fleName);
            rootName:=copy(fleName,1,place-1);
         End; {else}
      End; {procedure getInFile}

{************************************************************************}
{Gobbles up whitespace at the end of the line.                           }
{************************************************************************}
   Procedure skipWhiteSpace(Var fl : text; Var ch : char);
      Begin
         While ((ch in whiteSpace) And not(eof(fl))) Do
            read(fl,ch)
         End;  {procedure skipWhiteSpace}

{************************************************************************}
{This reads the text of the next token and copies it into state's        }
{nextString field.                                                       }
{************************************************************************}
   Procedure getNextToken(Var fl : text; state: statePtr; Var ch : char);
      Var
         matchChar : char;
      Begin
         If (ch in terminator)
         Then Begin
            state^.nextString := ch;   {single character, done!}
            read(fl,ch);            {move along to the next character}
         End {then}
         Else Begin
            If ((ch in quoteChars) or (ch = leftBracketChar)) Then Begin
               If (ch=leftBracketChar)
               Then matchChar := rightBracketChar
               Else matchChar := ch;
               state^.nextString := ch;
               Repeat
                  read(fl,ch);
                  state^.nextString := state^.nextString + ch;
               Until (eof(fl) OR (ch = matchChar));
               If (eof(fl))
               Then pdnexus_error('end of file occurred in middle of string or comment');
               read(fl,ch);
            End {if ch=quoteChar or commentChar}
            Else Begin
               skipWhiteSpace(fl,ch);
               While (not (eof(fl) OR
                      (ch in whiteSpace) OR
                      (ch in terminator))) Do Begin
                  state^.nextString := state^.nextString + ch;
                  read(fl,ch)
               End {while}
            End; {else}
         End; {not special char}
      End; {procedure getNextToken}

{************************************************************************}
{After a token is read, this classifies the string in nextString and sets}
{the token field appropriately.                                          }
{************************************************************************}
   Procedure analyzeToken(state: statePtr);
      Var valCode : integer;
          i : integer;
          valString : string;
          upCaseToken : string;
          firstChar : char;
      Begin
         {writeln('Token is ',state^.nextString);
         readln; }
         state^.token := NID;  {default}
         If length(state^.nextString) > 0
         Then Begin
            firstChar := upcase(state^.nextString[1]);
            If (firstChar in digitChar) OR      {doubt this can be CASE'd}
                ((firstChar = '-') AND
                 (length(state^.nextString) > 1) AND
                 (state^.nextString[2] in digitChar))
            Then Begin
               If (pos('.',state^.nextString) = 0)
               Then Begin {read as integer}
                  {MacClade 4 has some bit field data we should try to ignore}
                  If length(state^.nextString)<=11 Then Begin
                     val(state^.nextString,state^.lastInteger,valCode);
                     If (valCode <> 0) Then Begin
                        str(valCode,valString);
                        {pdnexus_warning('Problem in integer conversion of: ' +
                                          state^.nextString + ' at position: ' + valString);}
                     End {then}
                     Else state^.token := NINTEGER;
                  End {ignoring MacClade bit data}
               End {reading as integer}
               Else Begin {read as extended}
                  val(state^.nextString,state^.lastFloat,valCode);
                  If (valCode <> 0) Then Begin
                     str(valCode,valString);
                     {pdnexus_warning('Problem in floating conversion of: ' +
                                       state^.nextString + ' at position: ' + valString)}
                  End {then}
                  Else state^.token := NFLOAT;
               End {reading as integer}
            End {if ...digitChar...}
            Else Begin
               For i := 0 to length(state^.nextString)
                 Do upCaseToken[i] := upcase(state^.nextString[i]);
               {pdnexus_warning('Check - upCaseToken is '+upCaseToken);}
               Case firstChar of

                    '"' : state^.token := NSTRING;

                    '#' : state^.token := NEXUS;

                    '[' : Begin
                             If (upCaseToken = '[&R]')
                             Then state^.token := NROOTED
                             Else If (upCaseToken = '[&U]')
                                  Then state^.token := NUNROOTED
                                  Else state^.token := NCOMMENT;
                          END; {case '['}


                    ';' : state^.token := NSEMICOLON;

                    '=' : state^.token := NEQUALS;

                    ':' : state^.token := NCOLON;

                    ',' : state^.token := NCOMMA;

                    '(' : state^.token := NLEFTPAREN;

                    ')' : state^.token := NRIGHTPAREN;

                    'A' : Begin
                             If (upCaseToken = 'ASSUMPTIONS')
                             Then state^.token := NASSUMPTIONS;
                          End; {case 'A'}

                    'B' : Begin
                             If (upCaseToken = 'BEGIN')
                             Then state^.token := NBEGIN;
                          End; {case 'B'}

                    'C' : Begin
                             If (upCaseToken = 'CHARACTERS')
                             Then state^.token := NCHARACTERS;
                             If (upCaseToken = 'CHARSET')
                             Then state^.token := NCHARSET;
                          End; {case 'C'}

                    'D' : Begin
                             If (upCaseToken = 'DATA')
                             Then state^.token := NDATA;
                             If (upCaseToken = 'DIMENSIONS')
                             Then state^.token := NDIMENSIONS;
                             If (upCaseToken = 'DEFTYPE')
                             Then state^.token := NDEFTYPE;
                          End; {case 'D'}

                    'E' : Begin
                             If (upCaseToken = 'END')
                             Then state^.token := NEND;
                             {Treat (nonstandard) ENDBLOCK as a synonym for END}
                             If (upCaseToken = 'ENDBLOCK')
                             Then state^.token := NEND;
                             If (upCaseToken = 'EXSET')
                             Then state^.token := NEXSET;
                          End; {case 'E'}

                    'F' : Begin
                             If (upCaseToken = 'FORMAT')
                             Then state^.token := NFORMAT;
                          End; {case 'F'}

                    'G' : Begin
                             If (upCaseToken = 'GAP')
                             Then state^.token := NGAP;
                          End; {case 'G'}

                    'M' : Begin
                             If (upCaseToken = 'MISSING')
                             Then state^.token := NMISSING;
                             If (upCaseToken = 'MATRIX')
                             Then state^.token := NMATRIX;
                          End; {case 'M'}

                    'N' : Begin
                             If (upCaseToken = 'NTAX')
                             Then state^.token := NNTAXA;
                             If (upCaseToken = 'NCHAR')
                             Then state^.token := NNCHAR;
                          End; {case 'N'}

                    'O' : Begin
                             If (upCaseToken = 'OPTIONS')
                             Then state^.token := NOPTIONS;
                          End; {case 'O'}

                    'S' : Begin
                             If (upCaseToken = 'SYMBOLS')
                             Then state^.token := NSYMBOLS;
                          End; {case 'S'}

                    'T' : Begin
                             If (upCaseToken = 'TREES')
                             Then state^.token := NTREES;
                             If (upCaseToken = 'TRANSLATE')
                             Then state^.token := NTRANSLATE;
                             If (upCaseToken = 'Translate')
                             Then state^.token := NTRANSLATE;
                             If (upCaseToken = 'TREE')
                             Then state^.token := NTREE;
                             If (upCaseToken = 'TAXA')
                             Then state^.token := NTAXA;
                             If (upCaseToken = 'TAXLABELS')
                             Then state^.token := NTAXLABELS;
                             If (upCaseToken = 'TAXSET')
                             Then state^.Token := NTAXSET;
                          End; {case 'T'}

                    'U' : Begin
                             If (upCaseToken = 'UTREE')
                             Then state^.token := NUTREE;
                          End; {case 'U'}

               End; {Case}
            End; {else}
         End;
      End; {procedure analyzeToken}

{************************************************************************}
{This is the main loop for the parsing process.  Note that the main case }
{statement is switches on the parser state rather than the type of the   }
{current token.                                                          }
{************************************************************************}
Procedure parserLoop(state: statePtr;Var fl : text; Var contents : nexusPtr);
   Var
      ch : char;
      count : integer;
      index : integer;
      readName : boolean;
      fakeTranslate : boolean;
   Begin
      ch := ' ';  {safe initial state}
      Repeat
         state^.nextString := '';
         state^.token := NOTOKEN;
         skipWhiteSpace(fl,ch);
         getNextToken(fl,state,ch);
         analyzeToken(state);
         Case state^.curState of
            NONEXUS: Begin
                        If (state^.token = NEXUS) Then Begin
                           contents := nexusRecord_new;
                           state^.curState := NOBLOCK;
                           fakeTranslate := false;   {horrible place for this initializer}
                        End; {then}
                        {else ignore the token}
                      End; {case NONEXUS}
            NOBLOCK: Begin
                        If (state^.token = NBEGIN) Then Begin
                           state^.curState := BLOCKNAME;
                        End; {then}
                        {else ignore the token}
                     End; {case NONEXUS}
            BLOCKNAME:
               Begin
                  If (state^.token = NDATA)
                  Then Begin
                     state^.curState := INDATA;
                  End; {then}

                  If (state^.token = NTREES)
                  Then Begin
                     startTreeBlock(state,'',contents);
                     state^.curState := INTREES;
                  End; {then}

                  If (state^.token = NTAXA)
                  Then state^.curState := INTAXA;

                  If (state^.token = NUTREE)
                  Then state^.curState := INUTREE;
                  {else ignore the token}
               End; {case NONEXUS}

            INDATA:
               Begin
                  If (state^.token = NEND)
                  Then Begin
                     state^.curState := NOBLOCK;
                  End; {then}

                  If (state^.token = NDIMENSIONS)
                  Then state^.curState := INDIMENSIONS;

                  If (state^.token = NFORMAT)
                  Then Begin
                     state^.curState := INFORMAT;
                  End; {then}

                  If (state^.token = NMATRIX)
                  Then Begin
                     state^.curState := INMATRIX;
                     count := 0;
                     readName := true;
                  End {then}
               End; {case INDATA}

            INDIMENSIONS:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then state^.curState := INDATA;
                  If (state^.token = NNTAXA)
                  Then state^.curState := INNTAXA;
                  If (state^.token = NNCHAR)
                  Then state^.curState := INNCHAR;
               End; {case INDIMENSIONS}

            INNTAXA:
               Begin
                  If (state^.token = NINTEGER)
                  Then Begin
                     If (contents^.characters = nil)
                     Then contents^.characters := charBlock_new;
                     If (contents^.characters^.taxa = nil)
                     Then Begin
                        contents^.taxa := taxaBlock_new;
                        contents^.characters^.taxa := contents^.taxa;
                     End; {then}
                     contents^.characters^.taxa^.limit := state^.lastInteger;
		     If (contents^.translations <> NIL)
                     Then updateTranslateBlocks(contents^.translations,contents^.taxa);
                     state^.curState := INDIMENSIONS;
                  End; {then}
               End; {case INNTAXA}

            INNCHAR:
               Begin
                  If (state^.token = NINTEGER)
                  Then Begin
                     If (contents^.characters = nil)
                     Then contents^.characters := charBlock_new;
                     contents^.characters^.nchar := state^.lastInteger;
                     state^.curState := INDIMENSIONS;
                  End; {then}
               End; {case INNCHAR}

            INFORMAT:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then Begin
                     state^.curState := INDATA;
                  End; {then}
                  If (state^.token = NSYMBOLS)
                  Then state^.curState := INSYMBOLS;
               End; {INFORMAT}

            INSYMBOLS:
               Begin
                  If (state^.token = NSTRING)
                  Then Begin
                     processSymbolList(contents,state);
                     state^.curState := INFORMAT;
                  End; {then}
               End;

            INMATRIX:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then Begin
                     state^.curState := INDATA;
                  End
                  Else If (state^.token <> NCOMMENT)
                       Then Begin
                          If readName
                          Then Begin
                             inc(count);
                             index := dataBlock_addTaxon(contents^.characters,
                                                         state^.nextString);
                             readName := false;
                          End
                          Else Begin
                             dataBlock_addCharData(contents^.characters,
                                                   index,
                                                   state^.nextString);
                             readName := true;
                          End;
                       End;

               End;

            INEXSET:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then Begin
                     state^.curState := NOBLOCK;
                  End; {then}
               End; {case INEXSET}

            INCHARSET:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then Begin
                     state^.curState := NOBLOCK;
                  End; {then}
               End; {case INCHARSET}

            INTAXSET:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then Begin
                     state^.curState := NOBLOCK;
                  End; {then}
               End; {case INTAXSET}

            INTAXA:
               Begin
                  If (state^.token = NEND)
                  Then Begin
                     state^.curState := NOBLOCK;
                  End; {then}

                  If (state^.token = NDIMENSIONS)
                  Then state^.curState := INTDIMENSIONS;

                  If (state^.token = NTAXLABELS)
                  Then state^.curState := INTAXLABELS;

               End; {case INTAXA}

            INTDIMENSIONS:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then state^.curState := INTAXA;
                  If (state^.token = NNTAXA)
                  Then state^.curState := INTNTAXA;
               End; {case INDIMENSIONS}

            INTNTAXA:
               Begin
                  If (state^.token = NINTEGER)
                  Then Begin
                     If (contents^.taxa = nil)
                     Then contents^.taxa := taxaBlock_new;
                     contents^.taxa^.limit := state^.lastInteger;
		     If (contents^.translations <> NIL)
                     Then updateTranslateBlocks(contents^.translations,contents^.taxa);
                     state^.curState := INTDIMENSIONS;
                  End; {then}
               End; {case INNTAXA}

            INTAXLABELS:
               Begin
                  If (state^.token = NSEMICOLON)
                  Then state^.curState := INTAXA
                  Else If (state^.token <> NCOMMENT)
                       Then Begin
                          inc(count);
                          index := taxaBlock_addTaxon(contents^.taxa,
                                                      state^.nextString);
                       End;
               End;

            INTREES:
               Begin
                  Case state^.token of
                       NEND: state^.curState := NOBLOCK;

                       NUTREE:
                          Begin
                             state^.curState := UTREELABEL;
                             {pdnexus_warning('UTREE keyword suggests a tree in the file is unrooted');}
                          End; {case NUTREE}

                       NTREE: state^.curState := TREELABEL;

                       NTRANSLATE:
                          Begin
                             state^.curState := INTRANSLATE;
                             state^.entry_counter := 0;
                          End; {case NTRANSLATE}
                  End; {case stat^.token of}
               End; {case INTREES}

            TREELABEL:
               Begin
                  If (state^.token = NID)
                  Then If (state^.currentTreeBlock^.trees^.tree^.sibL = NIL)  {avoid multiple name problem}
                       Then Begin
                           state^.currentTree^.name := state^.nextString;
                           {writeln('Updating tree name to ', state^.nextString);}
                       End {then}
                       Else Begin
                          startTree(state,state^.nextString,contents^.trees,true);
                          {writeln('Creating new tree called ', state^.nextString);}
                       End; {else}

                  If (state^.token = NEQUALs)
                  Then state^.curState := INTREE;


               End; {case TREELABEL}

            UTREELABEL:
               Begin
                  If (state^.token = NID)
                  Then If (state^.currentTreeBlock^.trees^.tree^.sibL = NIL)  {avoid multiple name problem}
                       Then Begin
                           state^.currentTree^.name := state^.nextString;
                       End {then}
                       Else Begin
                           startTree(state,state^.nextString,contents^.trees,false);
                       End; {else}
                  If (state^.token = NEQUALs)
                  Then state^.curState := INTREE;
               End; {case TREELABEL}


            INTREE:
               Begin

                  If (state^.token = NLEFTPAREN)
                  Then addNode(state);

                  If (state^.token = NRIGHTPAREN)
                  Then finishNode(state);

                  If (state^.token = NCOLON)
                  Then state^.curState := BRANCHLENGTH;

                  If (state^.token = NINTEGER)
                  Then Begin
                     {writeln(state^.nextString);}
                     fillTip(state,contents);
                  End;

                  If (state^.token = NCOMMA)
                  Then moveToSister(state);

                  If (state^.token = NSEMICOLON)
                  Then Begin
                     If (fakeTranslate) Then Begin
                        finishTranslate(contents,state);
                        fakeTranslate := false;
                     End; {then}
                     finishTree(contents,state);
                     state^.curState := INTREES
                  End;

                  If (state^.token = NROOTED)
                  Then state^.currentTreeBlock^.trees^.rooted := true;

                  If (state^.token = NUNROOTED)
                  Then state^.currentTreeBlock^.trees^.rooted := false;

                  If (state^.token = NID)   {no translate?}
                  Then Begin
                     If ((state^.currentNode^.sibL = NIL) And
                         (state^.currentNode^.sibR = NIL))
                          Then fillTipNoTranslate(state,contents)
                          Else Begin
                             state^.lastNode^.pdiname[1] := state^.nextString[1];
                             state^.lastNode^.pdiname[2] := state^.nextString[2];
                             state^.lastNode^.pdiname[0] := chr(2);
                          End; {else}
                  End;
               End;

            BRANCHLENGTH:
                Begin
                   If (state^.lastNode <> NIL) Then Begin
                      If state^.token = NINTEGER
                      Then state^.lastNode^.length := 1.0*state^.lastInteger
                      Else state^.lastNode^.length := state^.lastFloat;
                   End {then}
                   Else pdnexus_error('Colon without a branch detected');
                   state^.curState := INTREE;
               End; {case INTREE}

            INTRANSLATE:
               Begin
                  If (contents^.taxa = NIL)
                  Then Begin
                     {pdnexus_warning('Adding taxon block for translate block');}
                     contents^.taxa := taxaBlock_new;
                     contents^.taxa^.limit := -1;
		     If (contents^.translations <> NIL)
                     Then updateTranslateBlocks(contents^.translations,contents^.taxa)
                     Else Begin
                     End;
                  End; {then}
                  If (state^.token = NSEMICOLON)
                  Then Begin
                     finishTranslate(contents,state);
                     state^.curState := INTREES
                  End;
                  If (state^.token = NEND)
                  Then pdnexus_error('Tree block Ended before translate!');
                  If (state^.token = NCOMMA)
                  Then pdnexus_error('Comma out of phase in translate block');
                  If (state^.token = NINTEGER)
                  Then Begin
                     state^.curState := INTRANSLATEPAIR;
                  End; {then}
               End; {case INTRANSLATE}

            INTRANSLATEPAIR:
               Begin
                  If (state^.token = NID)
                  Then Begin
                     {writeln('Last integer: ',state^.lastInteger,
                             ' taxa block ntaxa is ',contents^.taxa^.ntaxa);}
                     If (state^.lastInteger>contents^.taxa^.ntaxa)
                     Then index := taxaBlock_addTaxon(contents^.taxa,
                                                      state^.nextString);
                     addTranslatePair(state,
                                      contents,
                                      state^.lastInteger,
                                      state^.nextString);
                  End;

                  If (state^.token = NCOMMA)
                  Then state^.curState := INTRANSLATE;

                  If (state^.token = NSEMICOLON)
                  Then Begin
                     finishTranslate(contents,state);
                     state^.curState := INTREES
                  End;
                  If (state^.token = NEND)
                  Then pdnexus_error('Tree block Ended before translate!');
               End; {case INTRANSLATEPAIR}

            End; {case}
      Until(eof(fl)); {case}
   End; {procedure parserLoop}


{************************************************************************}
{************************************************************************}
Procedure ReadNexusFile(Var contents: nexusPtr;
                        Var flname :flstring;
                        Var ext :typstring;
                        Var rootName :flString;
                        Var succ:boolean);
   Var
      state : statePtr;
      fl : text;
   Begin
       writeln;
       contents := NIL;
       succ := true;   {until proven otherwise}
       getInFile(flName,ext,rootName,succ);
       If(succ) Then Begin
          init_errors;
          state := parserState_new(rootName);
          assign(fl,flName);
          Reset(fl);
          parserLoop(state,fl,contents);
          parserState_dispose(state);
          close(fl)
       End; {then}
   End; {procedure ReadNexusFile}


{***************************************************************************}
{This procedure again is long, but only saves the tree to disk.  It requires
the root, the current filename, the current extension, and ok which is true
if there is a tree to save.  It is only called by the main program.}
{***************************************************************************}
Procedure saveNexusToPdi(contents :nexusPtr;
                         Var fleName: flString;
                         Var rootName: flString;
                         characterBlockIndex : integer;
                         treeIndex : integer;
                         ok: boolean);
   Var
      fl            : text;
      rt            : nnodeptr;
      userTreeBlock : treeBlockPtr;
      userTree      : treeEntryPtr;
      num           : extended;
      firstTip      : nnodePtr;
      count         : integer;
      treeName      : string;
      typ           : typString;
      ans           : char;
      userChar      : char;
      doFloats      : boolean;
      writeError    : boolean;
   {************************************************************************}
   {This procedure gets the filename from the user.  That's all.  It is
   longer and messier than necessary because we try not to let the program
   crash when the user does silly things.}
   {************************************************************************}
   Procedure getFlnm(Var flnm:flstring;
                     Var typ:typString;
                     treeName : string);
      Var
         good, valid, again : boolean;
         ch:char;
         i :integer;  
         place :byte;
       Begin
         writeln('The nexus file you loaded was: ',flnm);
         writeln;
         writeln('And the tree you will be saving is named: ',treeName);
         writeln;
         writeln('You need to provide the name of a PDI file for this tree.');
         good:=true;
         valid:=true;
         again:=false;
         Repeat
            writeln;
            If Not valid
            Then writeln(flnm,' is not of a valid type.');
            If Not good
            Then writeln(flnm,' is write locked or not a legal filename.');
            writeln('What shall be the new name?  Do not forget the ',
                     '.PDI extension.');
            writeln('This program only saves in .PDI format!');
            writeln;
            readln(flnm);
            good:=true;
            valid:=true;
            again:=false;
            place:=pos('.',flnm);
            If place = 0
            Then valid:=false
            Else Begin
               typ:=copy(flnm,place+1,3);
               For i:=1 To 3 Do
                  typ[i]:=upcase(typ[i]);
               If (typ<>'PDI')
               Then valid:=false
               Else Begin
                  {$I-}
                  assign(fl,flnm);
                  reset(fl);
                  close(fl);
                  {$I+}
                  If (IOresult=0) and (flnm<>'')
                  Then Begin
                     writeln;
                     writeln(flnm,' already exits.');
                     writeln('Do you want to delete it and ',
                               'save the current tree in its ',
                               'place? (Y/N)');
                     Repeat
                        ch:=UpCase(readkey);
                     Until ch in['Y','N'];
                     If ch='N'
                     Then again:=true
                     Else Begin
                        {$I-}
                        Assign(fl,flnm);
                        Rewrite(fl);
                        close(fl);
                        {$I+}
                        If (IOresult<>0) or (flnm='') Then good:=false;
                     End; {else}
                  End {then}
                  Else Begin
                     {$I-}
                     Assign(fl,flnm);
                     Rewrite(fl);
                     close(fl);
                     {$I+}
                     If (IOresult<>0) or (flnm='') Then good:=false;
                  End; {else}
               End; {else}
            End; {else}
         Until good And valid And Not again
      End; {procedure getflnm}
   {********************************************************************}
   {This function sweeps through the tree, checking branch lengths and  }
   {tip values for values out of bounds.  It returns false if it finds  }
   {an out of bounds value.                                             }
   {********************************************************************}
   Function boundsSweep(nd,rt : nnodePtr) : boolean;
      Var temp : boolean;
      Begin
         temp := false;
         If nd^.sibL = nil Then Begin {It's a tip; check the trait values.}
            If ((abs(nd^.dat[1]) >= 1E15) Or
                ((abs(nd^.dat[1]) <= 1E-15) And
                 (abs(nd^.dat[1]) > 1E-99)))
            Then temp := true;
            If ((abs(nd^.dat[2]) >= 1E15) Or
                ((abs(nd^.dat[2]) <= 1E-15) And
                 (abs(nd^.dat[2]) > 1E-99)))
            Then temp := true;
            If ((abs(nd^.length)>1E15) Or     {check the length}
                ((abs(nd^.length)<1E-15) And
                 (abs(nd^.length) > 1E-99)))
               Then temp := true;
         End {then}
         Else Begin       {It's an internal node: check length and recurse.}
            If nd<>rt Then Begin
               If ((abs(nd^.length)>1E15) Or
                   ((abs(nd^.length)<1E-15) And
                    (abs(nd^.length) > 1E-99)))
               Then temp := true;
            End; {then}
            If not(temp) Then temp := boundsSweep(nd^.sibL,rt);
            If not(temp) Then temp := boundsSweep(nd^.sibR,rt);
         End; {else}
         boundsSweep := temp;
      End; {function boundsSweep}

   {********************************************************************}
   {This procedure writes an extended value to the output file.
    Depending on the value of doFloats, either an extended precision
    value (with exponents) or a floating point number will be written.  }
   {********************************************************************}
   Procedure ew(Var outFile :text;num :extended);
      Var outStr :string;
      Begin
         If doFloats
         Then Begin
            If ((abs(num)<1E15) And
                ((abs(num)>1E-15) Or
                 (abs(num)=0))) Then Begin
               str(num:17:15,outStr);
               If (length(outStr)>16)
               Then delete(outStr,16,length(outStr)-16);
               write(outFile,' ',outStr);
            End {then}
            Else writeln('Output value out of bounds for floating point format')
         End {then}
         Else extendwrite(outFile,num);
      End; {procedure ew}

   {*********************************************************************}
   {This procedure writes the tips data to the file in the PDI format
   recursively.}
   {*********************************************************************}
   Procedure wrttips(nd:nnodePtr;Var fl:text);
     Begin
        If nd^.sibL=nil Then Begin   {this is a tip}
           write(fl,nd^.pdiname,'  ');
           ew(fl,nd^.dat[1]);
           write(fl,'  ');
           ew(fl,nd^.dat[2]);
           writeln(fl)
        End {then}
        Else Begin                 {this is not a tip, ascend recursively}
           wrttips(nd^.sibL,fl);
           wrttips(nd^.sibR,fl);
        End {else}
     End; {procedure wrttips}
   {************************************************************************}
   {You may have noticed that this series of procedures is analogous to the
   ones in the read file block.  This one is in charge of writing the tree
   to the file.  One difference is that this uses recursive procedures of
   its own to do a lot of the work.}
   {************************************************************************}
   Procedure writePDI(rt:nnodePtr;flname:flstring;treesize:integer);
      Var
         fl : text;
         num: integer;
      {*********************************************************************}
      {This procedure writes the tree structure to a file in the PDI form
      recursively.}
      {*********************************************************************}
      Procedure wrtTree(nd:nnodePtr; var fl:text);
         Begin
            write(fl,nd^.pdiname,'  ');
            If nd^.ancestor=nil
            Then write(fl,'$$')                {write dummy ancestor for root}
            Else write(fl,nd^.ancestor^.pdiname);
            write(fl,'  ');
            ew(fl,nd^.length); {write length of branch leading to nd}
            writeln(fl);
            If nd^.sibL<>nil Then Begin
                  wrttree(nd^.sibL,fl); {ascend tree recursively}
                  wrttree(nd^.sibR,fl);
            End
         End;
      {*********************************************************************}
      Begin
         assign(fl,flname);
         rewrite(fl);
         writeln(fl,treeSize); {write tip count at top of file}
         writeln(fl);
         wrttree(rt,fl);
         writeln(fl);
         wrttips(rt,fl);
         close(fl)
      End; {writePDI}

   {************************************************************************}
   {This procedure writes node names into Nexus files after performing      }
   {several checks and conversions.  It is also used in writing PDAP        }
   {formats derived from the Nexus specification (Maddison et. al. 1997)    }
   {(e.g.,BRK,TIP,NAM).                                                     }
   {************************************************************************}
   Procedure writeNexusName(Var fl : text; nam : pdinamestr);
      Var
         i : integer;
         dummy : pdinamestr;
         quotesNeeded : boolean;
      Begin
         quotesNeeded := false;
         For i := 1 To 2 Do Begin
            If nam[i] in ['[',']','(',')','{','}','=',',',';',':','-','\',
                          '/','*','"','''']
            Then quotesNeeded := true;
         End; {for i}
         dummy := nam;
         If (nam[1] = ' ') Then dummy[1] := '_';
         If ((length(nam) > 1) And (nam[2] = ' '))
         Then dummy[2] := '_';
         If quotesNeeded
         Then write(fl,'''',dummy,'''')
         Else write(fl,dummy);
      End; {procedure writeNexusName}

   {************************************************************************}
   {This function takes a name and returns the number of characters required}
   {to write the name according to the rules implemented by the preceeding  }
   {procedure.                                                              }
   {************************************************************************}
   Function nexusNameLength(nam : pdinamestr) : integer;
      Var
         i : integer;
         quotesNeeded : boolean;
         result : integer;
      Begin
         quotesNeeded := false;
         For i := 1 To 2 Do Begin
            If nam[i] in ['[',']','(',')','{','}','=',',',';',':','-','\',
                          '/','*','"','''']
            Then quotesNeeded := true;
         End; {for i}
         If quotesNeeded
         Then nexusNameLength := length(nam)+3
         Else nexusNameLength := length(nam)+1;
      End; {function nexusnamelength}


   {************************************************************************}
   {************************************************************************}
   Function findPdiName(start : nNodePtr; key : integer) : pdiNameStr;
      Var current : nNodePtr;
      Begin
         current := start;
         While ((current <> NIL) And
                (current^.index <> key)) Do current := current^.tipR;
         If (current <> NIL)
         Then findPdiName := current^.pdiName
         Else findPdiName := '';
      End; {function findPdiName}


   {************************************************************************}
   {************************************************************************}
   Function findTrans(transBlock : translateBlockPtr; key : integer) : string;
      Var
         current : translatePairPtr;
      Begin
         current := transBlock^.data;
         While ((current <> NIL) And
                (current^.index <> key)) Do current := current^.next;
         If (current <> NIL)
         Then Begin
            findTrans := current^.curtaxa^.taxa[current^.index]^.name;
         End {then}
         Else findTrans := '';
      End;

   {************************************************************************}
   { This procedure actually writes the table to the file, assuming a       }
   { translate block was loaded from the nexus file.                        }
   {************************************************************************}
   Procedure writeTranslateTable(Var fl : text;
                                     rt : nnodeptr;
                                     start : nnodeptr;
                                     transBlock : translateBlockPtr);
      Var 
          i : integer;
      Begin
         writeln(fl,'Translate Index',tabchar,'Pdi Name',tabchar,'Nexus Name');
         For i := 1 to transBlock^.tlength Do
            writeln(fl,i:4,tabchar,
                        findPdiName(start,i),tabchar,
                        findTrans(transBlock,i));
      End; {procedure writeTranslateTable}

   {************************************************************************}
   { This procedure writes a table when no translate block and possibly no  }
   { taxon block is available.                                              }
   {************************************************************************}
   Procedure writeNoTranslateTable(Var fl : text;
                                       rt : nnodePtr;
                                       start : nnodeptr;
                                       taxBlock : taxaBlockPtr;
                                       count : integer);
      Var
         i : integer;
      Begin
         writeln(fl,'Taxon Index',tabchar,'Pdi Name',tabChar,'Nexus Name');
         If (taxBlock <> NIL) Then
            For i := 1 to count Do
               writeln(fl,i:4,tabchar,
                          findPdiName(start,i),tabchar,
                          taxBlock^.taxa[i]^.name)
         Else {no translate or taxa - try to handle a mangled file}
            For i := 1 to count Do
               writeln(fl,i:4,tabchar,
                          findPdiName(start,i),tabchar,
                          i:4)   {just write the number again}
      End; {procedure writeNoTranslateTable}

   {************************************************************************}
   {This writes the TSV (tab separated values) file that can be loaded into }
   {MS Excel to see the correspondences between taxon names in the Nexus    }
   {file and the generated PDI tip names.                                   }
   {************************************************************************}
   Procedure writeTSV(rt : nNodePtr;
                      contents : nexusPtr;
                      count : integer;
                      start : nNodePtr;
                      flname:flstring);
      Var
         fl : text;
         place : integer;
         transBlock : translateBlockPtr;
         taxBlock : taxaBlockPtr;
      Begin
        transBlock := contents^.selectedTranslateBlock;
        If (transBlock <> nil) Then Begin
           If (count > transBlock^.tlength)
           Then Begin
              writeln ('Error - tree has more tips than translate');
              writeln ('        No TSV file can be written');
              readln;
           End {then}
           Else Begin
              place:=pos('.',flname);
              flname:=copy(flname,1,place)+'TSV';
              assign(fl,flName);
              rewrite(fl);
              writeTranslateTable(fl,rt,start,transBlock);
              close(fl);
           End; {else}
        End {then}
        Else Begin
              place:=pos('.',flname);
              flname:=copy(flname,1,place)+'TSV';
              assign(fl,flName);
              rewrite(fl);
              taxBlock := contents^.taxa;
              writeNoTranslateTable(fl,rt,start,taxBlock,count);
              close(fl);
        End; {else}
      End; {procedure writeTSV}

   {*************************************************************************}
   {Helper function - retrieves the character Block with the right index     }
   {*************************************************************************}
   Function retrieveCharBlock(contents : nexusPtr;
                              index : integer) : charBlockPtr;
      Var block : charBlockPtr;
          counter : integer;
      Begin
         counter := 1;
         block := contents^.characters;
         While ((counter < index) And (block <> NIL)) Do Begin
            inc(counter);
            block := block^.next;
         End; {while}
         retrieveCharBlock := block
      End; {function retrievecharblock}

   {*************************************************************************}
   {Helper function - retrieves the treeEntry with the right index.  Note    }
   {that this function searches through tree indices and returns the         }
   {treeBlock as well as the specified treeEntry.                            }
   {*************************************************************************}
   Procedure retrieveTree(contents : nexusPtr;
                          index : integer;
                          Var block : treeBlockPtr;
                          Var tree : treeEntryPtr);
      Var counter : integer;
      Begin
         counter := 1;
         block := contents^.trees;
         tree := block^.trees;
         While ((counter < index) And (block <> NIL)) Do Begin
            While ((counter < index) And (tree <> NIL)) Do Begin
               tree := tree^.next;
               inc(counter);
            End; {while}
            If (counter < index) Then Begin
               block := block^.next;
               tree := block^.trees;
               inc(counter);
            End;
         End; {while}
      End; {procedure retrievetree}

   {************************************2************************************}
   Begin {saveTree}
      clrscr;
      writeln;
      With contents^ Do Begin
         selectedCharBlock := retrieveCharBlock(contents,characterBlockIndex);
         retrieveTree(contents,treeIndex,userTreeBlock,userTree);
         rt := userTree^.tree;
         treename := userTree^.name;
      End; {with}
      If not ok
      Then writeln('You must have a tree in memory to save it to disk.')
      Else Begin
         getflnm(flename,typ,treename);
         writeln;
         writeln('Would you like to write values in floating point format (no exponents)?');
         write('(Y/N=Enter): ');
         userChar := readKey;
         writeln;
         doFloats := userChar in ['Y','y'];
         If doFloats Then Begin
            writeError := boundsSweep(rt,rt);
            If writeError Then Begin
               writeln('The tree contains values that are too large or small to write in floating');
               writeln('point format.');
               writeln;
               write('Would you like save the tree using exponential format (Y/N)? ');
               userChar := readKey;
               writeln;
               If userChar in ['Y','y'] Then Begin
                  doFloats := false;
                  writeError := false;
               End; {then}
            End; {then}
         End {then}
         Else writeError := false;
         If doFloats Then Begin
            writeln;
            writeln('Warning - Values in your PDI file may include roundoff errors.  These will not');
            writeln('          affect the values used internally when the file is reloaded.');
            writeln;
         End; {then}
         If not(writeError) Then Begin
            prepareTips(rt,count,firstTip);
         End;
         If not(writeError) Then Begin
            writePDI(rt,flename,count);
            writeTSV(rt,contents,count,firstTip,flename);
            writeln;
	    writeln('Tree ',treename,' successfully saved to ',flename,'.');
         End; {then}
      End; {else}
      {ret('Please press any key to return to the Main Menu.');}
   End; {procedure saveNexusToPdi}



End. {unit N_IO1}
