{****************************************************************************}
{This is a stripped down version of PDTREE, which will read trees and tip
data from nexus (Maddison, Swofford & Maddison 1997 Systematic Biology
46:590-621) files.  It will save the tree in PDI format and will produce
a three column report mapping the long taxon names and translate block
numbers to the two character PDAP names.

This program consists of the following compilation units:


PDNEXUS: The unit you are in now. (Title screen)

N_IO1: The input and output unit.  It mostly calls functions in T_IO1,
       but also supports reading of Nexus files.  Reading Nexus files
       results in the creation of translation tables.

N_UTIL1: Types and utility functions for nexus files.  Similar to T_UTIL1,
         but some structures are more compact (so bigger files with
         multiple trees will fit!)


{****************************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}


Program PDNexus;
   uses
      crt,
      n_util1,      {Nexus file related types}
      n_io1;        {This unit has the input/output tree from disk procedures
                      as well as the create tree procedure.}
   Var
      filename : flstring;
      rootName : flString;
      extension: typstring;
      contents : nexusPtr;
      ch : char;
      liveFile :boolean;
      cBlockIndex : integer;    {specifies the user's choice of character block}
      tIndex      : integer;    {specifies the user's choice of tree within a block}

{***************************************************************************}
{Here the program just writes a title page for the program.  It is only
called by the main program and only once.}
{***************************************************************************}
Procedure doIntro;
   Begin
      clrscr;
      Writeln(' ********* Phenotypic Diversity Analysis Program - PDNEXUS Module *********');
      Writeln(' Version 1.0 Copyright 29 September 2004 by Theodore Garland, Jr., ');
      Writeln('                      and Peter E. Midford');
      Writeln('Questions to T.G. at Dept. of Biology, Univ. of Calif., Riverside, CA 92521');
      Writeln('    Phone 951-827-3524; Fax 951-787-4286; E-mail tgarland@ucr.edu');
      Writeln;
   End; {procedure doIntro}
{***************************************************************************}
{            The main program is basically a menu and little more.          }
{***************************************************************************}
   Begin                 { THE MAIN PROGRAM BEGINS HERE!!! }
      doIntro;
      ch := ' ';
      Repeat
         liveFile:=false;            {No file in memory}
         Repeat
            ReadNexusFile(contents,filename,extension,rootName,livefile);
         Until (liveFile = false) OR (contents <> NIL);
         If liveFile Then Begin
            AnalyzeContents(contents,rootName,cBlockIndex,tIndex);
            If (tIndex>0)
            Then SaveNexusToPDI(contents,
                                filename,
                                rootName,
                                cBlockIndex,
                                tIndex,
                                livefile);
            If(contents <> NIL)
            Then nexusRecord_dispose(contents);
            writeln;
            writeln;
            write('Translate another NEXUS file? (Y/N(=default): ');
            ch := readkey;
         End; {then}
      Until ((ch <> 'Y') And (ch <> 'y'));
   End.  {program PDNEXUS}
