{ PDRANDOM.PAS                                                         }
{ This is the main program module for PDRANDOM.                        }
{                                                                      }
{ It uses the following units:                                         }
{                                                                      }
{ RANDUTIL - Utility functions for working with trees and matrices     }
{ RANDPERM - Performs permutation operations and calculates distances  }
{            and probabilities for trees                               }
{ RANDIO  - High level I/O                                             }
{ RAND_IO1 - Reads trees, matrices, etc.                               }
{ crt - Borland supplied utilities used for clearing the screen and    }
{       reading keystrokes.                                            }
{                                                                      }
{ ******************************************************************** }
{ Modification History:                                                }
{                                                                      }
{   11-Mar-96: Put in extra line reads so messages don't disappear     }
{              before returning to the main menu.  PEM                 }
{                                                                      }
{   13-Mar-96: Put permutation choices in header of permutation and    }
{              distance files.                                         }
{                                                                      }
{   14-Mar-96: Incorporated functionality of PDRNDTEST and included    }
{              distance and permutation matrices into the .TST file    }
{              which is always written.                                }
{                                                                      }
{   22-Mar-96: Added row and column totals and other tst file goodies  }
{              from the column normalizing version.                    }
{                                                                      }
{                                                                      }
{   26-Mar-96: Added matrix style output to the TST file. PEM          }
{                                                                      }
{   28-Mar-96: Added transform option.  PEM                            }
{                                                                      }
{   12-Apr-96: Minor changes to output formating.  PEM                 }
{                                                                      }
{   23 Aug 97: Fixed number of minor formatting typos and a problem    }
{              with the .TST matrix.                                   }
{                                                                      }
{   25 Aug 97: Added option to read probability matrix directly.       }
{                                                                      }
{    4 Sep 97: Added option to use Random number generator from Manley }
{              1991.                                                   }
{                                                                      }
{   29 Sep 97: Minor improvements to format of .TST file.              }
{                                                                      }
{   13 Oct 97: Completed modifications for extended precision          }
{              probability calculations.                               }
{                                                                      }
{   15 Oct 97: Changed the counts matrix to type WORD so we get full   }
{              range; counts could never be negative anyway.           }
{                                                                      }
{                                                                      }
{    4 Dec 97: Flushed old X,E,M,S options; added new (M)ethod option  }
{              to change the method of calculating .PRB probabilities. }
{              If new (similarity) method is selected (default) then   }
{              a K option (for specifying the K) value is selected;    }
{              this replaces the F option used for the reciprocal      }
{              distance method. PEM                                    }
{                                                                      }
{    8 Dec 97: Finished adding code for the similarity method. PEM     }
{                                                                      }
{   15 Dec 97: Replaced code for checking for factor = 0 in the        }
{              distance method (.PRB and randomization do not work).   }
{              Also replaced message about exclusions to be listed in  }
{              a later version with blank lines.  PEM                  }
{                                                                      }
{                                                                      }
{    9 Feb 98: Fixed erroneous reporting of file extension after       }
{              loading the .PRB containing tip data for a tree in      }
{              matrix form.                                            }
{                                                                      }
{   25 Jun 98: Fixed a bug that caused the wrong (unscaled) maxDist    }
{              value to be used during permutation.  I did this by     }
{              removing the code that rescaled the tree to a height    }
{              of 1.                                                   }
{                                                                      }
{   13 Jul 98: Added option to allow equiprobable permuations.         }
{                                                                      }
{   13 Nov 99: Added permutation set count to the randomized output    }
{              (the .PRM file).  Also added improved .PDI file reader  }
{              to Randio.  PEM                                         }
{                                                                      }
{   10 Jul 01: Added title screen and released by T.G.                 }
{                                                                      }
{ Last Modified:  18 May 1999 P. E. Midford                            }

Program PDRANDOM;
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {small stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

uses RANDUTIL,RANDPERM,RANDIO,RAND_IO1,crt;

   Var data1         : phyloPtr;       { structures containing tree or }
       data2         : phyloPtr;       { matrix information}
       tipCount      : integer;        { tips or matrix size (same for both!) }
       i             : integer;        { general purpose loop counter}
       ix,iy,iz      : integer;        { for testing Manley's generator}

{*********************************************************************}
{ This function prevents an error from being signaled when the heap is}
{ exhausted.  Instead the call to new just returns NIL and the        }
{ function that's doing the allocation has to deal with it.  This     }
{ makes the output caching stuff in RANDIO possible.                  }
{*********************************************************************} 
   Function heapfunc(size:word): integer; far;
      Begin
         heapfunc:=1;
      End; {heapfunc}

{*********************************************************************}
{ This function just displays the information on the help screen.     }
{ Update the program revision date here.  If you want to make a batch }
{ version of this program, you'll need to remove the call to clrscr.  }
{*********************************************************************}
   Procedure doTitleScreen;
      Var name : string;
      Begin
         getProgramName(name);
         clrscr;
         writeln('       ********* Phenotypic Diversity Analysis Program *********');
         writeln;
         writeln('                             PDRANDOM Module');
         writeln;
         writeln('V. 1.0 Copyright 10 July 2001 by Peter E. Midford and Theodore Garland, Jr.');
         writeln;
         writeln('     Contact T.G. at Department of Biology, University of California,');
         writeln('                         Riverside, CA 92521  U.S.A.');
         writeln;
         writeln('This programs performs conventional (equiprobable) and phylogenetic');
         writeln('randomizations (permutations) for the analysis of comparative data,');
         writeln('as described in:');
         writeln;
         writeln('Lapointe, F. J., and T. Garland, Jr. 2001. A generalized permutation model');
         writeln('     for the analysis of cross-species data.');
         writeln('     Journal of Classification 18:109-127.');
         writeln;
         writeln('The randomized data sets, output in a RND file, can be treated similarly to');
         writeln('data that have been simulated phylogenetically, e.g., by PDSIMUL.');
         writeln;
         writeln;
         writeln;
         writeln;
         writeln('Press the enter key to continue');
         readln;
      End; {procedure doTitleScreen}

{*********************************************************************}
{ This procedure might someday display a summary screen after the user}
{ quits from the menu, but for now it does nothing.                   }
{*********************************************************************}
   Procedure doSummary;
      Begin
      End; {doSummary}

{*********************************************************************}
{ This procedure displays the main menu as well as the current setting}
{ of several important values and switches.  Look at the main menu    }
{ screen to see what it does.                                         }
{*********************************************************************}
   Procedure showMenu(userOptions : options);
      Var matrixFlag : boolean;
      Begin
         If (data1^.matrix <> nil) Or (data2^.matrix <> nil)
         Then matrixFlag := true
         Else matrixFlag := false;
         clrscr;
         gotoxy(1,1);
         write('********************************* Main Menu **********************************');
         If ((data1^.tree = nil) and (data2^.tree = nil) and
             (data1^.matrix = nil) and (data2^.matrix = nil))
         Then Begin
            gotoxy(1,2);
            write('No Trees or Matrices in memory')
         End {then}
         Else Begin
            gotoxy(1,2);
            If matrixFlag = false
            Then write('Tree(s) in memory')
            Else write('Matrix(s) in memory');
            If ((data1^.tree <> nil) or (data1^.matrix <> nil))
            Then Begin
               gotoxy(1,3);
               write('        trait 1: ',data1^.sourceName)
            End;
            If ((data2^.tree <> nil) or (data2^.matrix <> nil))
            Then Begin
               gotoxy(1,4);
               write('        trait 2: ',data2^.sourceName)
            End
         End;
         gotoxy(35,2);
         If matrixFlag
         Then write('**************')
         Else
            If (userOptions.myMethod = similarity)
            Then write('(M) = similarity')
            Else write('(M) = reciprocal distance');
         gotoxy(35,3);
         If matrixFlag
         Then write('**************')
         Else
            If (userOptions.myMethod = similarity)
            Then write('(K) = ',userOptions.k:7:2)
            Else write('(F)actor = ',userOptions.factor:7:2);
         gotoxy(35,4);
         If matrixFlag
         Then write('(T)ransform = XXX')
         Else Begin
            write('(T)ransform = ');
            If userOptions.logTrans
            Then write('Log base 10')
            Else write(userOptions.blPower:5:2)
         End; {else}
         gotoxy(35,5);
         write('(N)umber = ',userOptions.seed);
         gotoxy(35,6);
         {write('(M)in/Max = ');}
         gotoxy(1,7);
         writeln('******************************************************************************');
         writeln('Which of the following would you like to do?');
         If matrixFlag Then Begin
            writeln(' (I)nput a tree for either trait 1 or trait 2');
            writeln(' (IPRB) = Input a .PRB matrix instead of a tree - 80x80 Maximum size');
            writeln(' (N)umber = specify or check seed for the pseudorandom number generator');
            writeln(' (C)hange random number generator (Turbo Pascal or Manley (1991))');
            writeln(' (P)robability = write file(s) of (row normalized) transition probabilities');
            writeln(' (R)andomize and save a set of data');
            writeln(' (E) = randomize equiprobably and save a set of set');
            writeln(' (Q)uit');
            writeln('******************************************************************************');
            writeln('Please choose (I/IPRB/N/C/P/R/E/Q)')
         End
         Else
            If (userOptions.myMethod = similarity) Then Begin
               writeln(' (I)nput a tree for either trait 1 or trait 2');
               writeln(' (T)ransform branch lengths = specify exponent or log (base 10) transform');
               writeln(' (M)ethod of calculating .PRB matrix (similarity or reciprocal distance)');
               writeln(' (K) = specify constant for similarity calculation');
               writeln(' (IPRB) = input a .PRB matrix instead of a tree - 80 X 80 maximum size');
               writeln(' (N)umber = specify or check seed for the pseudorandom number generator');
               writeln(' (C)hange random number generator (Turbo Pascal default or Manley, 1991)');
               writeln(' (D)istance = write one or two distance matrices (*.DST)');
               writeln(' (P)robability = write one or two transition probability matrices (*.PRB)');
               writeln(' (R)andomize and save sets of data (no .TST file for tree larger than 179)');
               writeln(' (E) = randomize equiprobably and save sets of data');
               writeln(' (Q)uit');
               writeln('******************************************************************************');
               writeln('Please choose (I/T/M/K/IPRB/N/C/D/P/R/E/Q)')
            End {then}
            Else Begin
               writeln(' (I)nput a tree for either trait 1 or trait 2');
               writeln(' (T)ransform branch lengths = specify exponent or log (base 10) transform');
               writeln(' (M)ethod of calculating .PRB matrix (similarity or reciprocal distance)');
               writeln(' (F)actor = factor for computing self distance from nearest ancestor distance');
               writeln(' (IPRB) = input a .PRB matrix instead of a tree - 80 X 80 maximum size');
               writeln(' (N)umber = specify or check seed for the pseudorandom number generator');
               writeln(' (C)hange random number generator (Turbo Pascal default or Manley, 1991)');
               writeln(' (D)istance = write one or two distance matrices (*.DST)');
               writeln(' (P)robability = write one or two transition probability matrices (*.PRB)');
               writeln(' (R)andomize and save sets of data (no .TST file for tree larger than 179)');
               writeln(' (E) = randomize equiprobably and save a set of set');
               writeln(' (Q)uit');
               writeln('******************************************************************************');
               writeln('Please choose (I/T/M/F/IPRB/N/C/D/P/R/E/Q)')
            End {else}
      End; { procedure showmenu }

{****************************************************************************}
{This procedure handles the dialog for loading trees into memory, loads the  }
{trees and sets up the permuation arrays.                                    }
{****************************************************************************}
   Procedure treeDialog(Var succ1 : boolean;
                        Var succ2 : boolean);
      Var menuChar : char;
      Begin
         writeln('For which trait do you want to read a tree?');
         writeln('  Remember that the trees need to have the same topology, though');
         writeln('  branch lengths may differ.');
         writeln;
         write('1');
         If succ1
         Then writeln('  (tree ',data1^.sourceName,' already loaded)')
         Else writeln;
         write('2');
         If succ2
         Then writeln('  (tree ',data2^.sourceName,' already loaded)')
         Else writeln;
         Repeat
            writeln('Enter either 1 or 2 to read one tree');
            writeln('Enter 3 to read the same tree for both traits');
            write('Enter 0 to return to the main menu: ');
            readln(menuChar);
            writeln;
            Case menuChar of
               '1' : Begin {read the trait 1 tree file}
                        readTree(data1,succ1);
                        If (data1^.tree <> nil)
                        Then Begin
                           prepareTips(data1,tipCount);
                           setUpPermArray(data1,tipCount);
                           checkTipBranchLengths(data1,tipCount);
                        End {then}
                     End; {case of 1}

               '2' : Begin {read the trait 2 tree file}
                        readTree(data2,succ2);
                        If (data2^.tree <> nil)
                        Then Begin
                           prepareTips(data2,tipCount);
                           setUpPermArray(data2,tipCount);
                           checkTipBranchLengths(data2,tipCount);
                        End {then}
                     End; {case of 2}

               '3' : Begin {read the same *file* for both traits}
                        readTree(data1,succ1);
                        If (data1^.tree <> nil)
                        Then Begin
                           prepareTips(data1,tipCount);
                           setUpPermArray(data1,tipCount);
                           checkTipBranchLengths(data1,tipCount); 
                           data2^.sourceName := data1^.sourceName;
                           data2^.sourceExt := data1^.sourceExt;
                           rereadTree(data2,succ2);
                           prepareTips(data2,tipCount);
                           setUpPermArray(data2,tipCount);
                           checkTipBranchLengths(data2,tipCount); 
                        End {then}
                     End {case of 3}
            End; {Case}
         Until(menuChar in ['0','1','2','3']);
         ret('Please press any key to return to the Main Menu.');
      End; {procedure treedialog}

{****************************************************************************}
{This procedure handles the dialog for loading matrices into memory, loads   }
{matrices and sets up the permuation arrays.                                 }
{****************************************************************************}
   Procedure matrixDialog(Var msucc1 : boolean;
                          Var msucc2 : boolean);
      Var menuChar : char;
      Begin

         writeln('  Remember that the matrices need to have the same size, though');
         writeln('  specific entries may differ.');
         writeln;
         write('1');
         If msucc1
         Then writeln('  (matrix ',data1^.sourceName,' already loaded)')
         Else writeln;
         write('2');
         If msucc2
         Then writeln('  (matrix ',data2^.sourceName,' already loaded)')
         Else writeln;
         Repeat
            writeln('Enter either 1 or 2 to read one matrix');
            writeln('Enter 3 to read the same matrix for both traits');
            write('Enter 0 to return to the main menu: ');
            readln(menuChar);
            writeln;
            Case menuChar of
               '1' : Begin {read the trait 1 matrix file}
                        readMatrix(data1,msucc1,false);
                        If (data1^.matrix <> nil)
                        Then readMatTips(data1,1,tipCount,false)
                        Else tipCount := 0;
                        If (tipCount > 0)
                        Then setUpPermArray(data1,tipCount)
                        Else Begin
                           dispose(data1^.matrix);
                           data1^.matrix := nil
                        End {else}
                     End; {case of 1}

               '2' : Begin {read the trait 2 matrix file}
                        readMatrix(data2,msucc2,false);
                        If (data2^.matrix <> nil)
                        Then readMatTips(data2,2,tipCount,false)
                        Else tipCount := 0;
                        If (tipCount > 0)
                        Then setUpPermArray(data2,tipCount)
                        Else Begin
                           dispose (data2^.matrix);
                           data2^.matrix := nil
                        End {else}
                     End; {case of 2}

               '3' : Begin {read the same *file* for both traits}
                        readMatrix(data1,msucc1,false);
                        If (data1^.matrix <> nil)
                        Then readMatTips(data1,1,tipCount,false)
                        Else tipCount := 0;
                        If (tipCount > 0)
                        Then Begin
                           setUpPermArray(data1,tipCount);
                           data2^.sourceName := data1^.sourceName;
                           data2^.sourceExt := data1^.sourceExt;
                           data2^.matrixTipFile := data1^.matrixTipFile;
                           writeln;
                           writeln('Re-Reading matrix for trait 2');
                           writeln;
                           readMatrix(data2,msucc2,true);
                           readMatTips(data2,2,tipCount,true);
                           setUpPermArray(data2,tipCount)
                        End {then}
                        Else Begin
                           msucc1 := false;
                           msucc2 := false;
                           If (data1^.matrix <> nil)
                           Then dispose(data1^.matrix);
                           data1^.matrix := nil;
                           data2^.matrix := nil;
                        End {else}
                     End {case of 3}
            End; {Case}
         Until(menuChar in ['0','1','2','3']);
         writeln;
         ret('Please press any key to return to the Main Menu.');
      End; {procedure matrixDialog}

{*********************************************************************}
{ This procedure is the heart of PDRANDOM (and anything based on it). }
{ The procedure implements the main menu.  It calls show menu and     }
{ reads a single character from the user.  The dispatching is also    }
{ coded here.  Once the user choses 'Q' the loop in this procedure    }
{ exits.                                                              }
{*********************************************************************}
   Procedure doMenu;
      Var diskbuffernum       : longint;  { Used as part of the output caching scheme}
          diskbuffermax       : longint;  { Used as part of the output caching scheme}
          diskbufferptr       : datptr;   { Used as part of the output caching scheme}
          diskbufferstart     : datptr;   { Used as part of the output caching scheme}
          permFile            : text;     { points to the permutation output file}
          tempReal            : extended; { holds a real (extended) number read from the user}
          tree1Height         : extended; { calculated height of tree 1}
          tree2Height         : extended; { calculated height of tree 2}
          maxPermEst          : longint;  { suggested maximum value for permutationCount}
          permutationCount    : longint;  { holds number of permutations to perform}
          pCount              : longint;  { actual counter }
          valCode             : integer; { receives status from string->number conversions}
          userChar            : char;    { holds anyother character read from the user}
          done                : boolean; { true when it's time to exit}
          succ1               : boolean; { true when tree 1 was successfully read}
          succ2               : boolean; { tree when tree 2 was successfully read}
          msucc1              : boolean; { true when matrix 1 was successfully read}
          msucc2              : boolean; { tree when matrix 2 was successfully read}
          savedMethod         : pMethod; { used for saving old method when doing equiprobable permutations}
          seedStr             : string;  { holds string from the user specifying a random # seed}
          factorStr           : string;  { holds string from the user specifying a self distance factor}
          powerStr            : string;  { holds string from the user specifying a branch length power}
          menuStr             : string;  { the user's menu choice }
          userOptions         : options; { holds the current option values }


      Procedure permutationMenuProc;
         Var pcount : integer;
             lastSet : integer;
         Begin
            clrscr;
            If tipCount < 180 Then Begin
               initCountTable(data1^.testTable);
               initCountTable(data2^.testTable);
               If ((data1^.testTable = nil) or (data2^.testTable = nil))
               Then Begin
                  writeln('Fatal Error - heap exhausted');
                  writeln('              Press the enter key to exit program');
                  readln;
                  halt;
               End; {then}
            End {then}
            Else Begin
               data1^.testTable := nil;
               data2^.testTable := nil;
               writeln;
               writeln('Because your tree has more than 179 tips, no test file will be written.');
            End; {else}
            If (data1^.tree <> nil) Then linkTree(data1^.tree);
            If (data2^.tree <> nil) Then linkTree(data2^.tree);
            openPermFile(data1^.sourceName,permFile,data1^.tstFileName);
            heapError := @heapfunc;
            userOptions.Seed := RandSeed;
            writePermFileHeader(data1,
                                data2,
                                permFile,
                                tipCount,
                                userOptions,
                                diskbuffernum,
                                diskbuffermax,
                                diskbufferptr,
                                diskbufferstart);
            maxDistance(data1,tipCount,userOptions); {as good a place as any}
            maxDistance(data2,tipCount,userOptions);
            {maxDistance must be called no later than this point}
            {maxPermEst := maxPermEstimate(data1,tipCount,userOptions);}
            write('How many sets of permutated data should be generated (Max = 32767)?');
            readln(permutationCount);
            While ((permutationCount < 0) Or (permutationCount > 32767)) Do Begin
               writeln;
               If (permutationCount <= 0)
               Then writeln(permutationCount,' is not positive.')
               Else writeln(permutationCount,' is too large.');
               write('Enter a positive number < 32768: ');
               readln(permutationCount);
            End; {while}
            clrscr;
            pCount := 1;
            lastSet := 0;
            Repeat
               gotoxy(20,13);
               clreol;
               write('Generating permutation set #',pCount);
               If (pCount > 1)
               Then Begin
                  doPermutations(data1,tipCount,1,userOptions);
                  doPermutations(data2,tipCount,2,userOptions);
               End; {then}
               If (pCount = permutationCount)
               Then lastSet := 1;
               If (data1^.tree <> nil)
               Then writeTrees(permFile,
                               data1^.tree,
                               data2^.tree,
                               tipCount,
                               lastSet,
                               pCount,
                               diskbuffernum,
                               diskbuffermax,
                               diskbufferptr,
                               diskbufferstart)
               Else writeTips(permFile,
                              data1^.permArray,
                              data2^.permArray,
                              tipCount,
                              lastSet,
                              pCount,
                              diskbuffernum,
                              diskbuffermax,
                              diskbufferptr,
                              diskbufferstart);
               If (pCount > 1) Then Begin
                  backPermute(tipCount,1,data1^.permArray);
                  backPermute(tipCount,2,data2^.permArray);
               End; {then}
               inc(pCount)
            Until (lastSet=1);
            close(permFile);
            freeheap(diskbufferptr);
            If data1^.testTable <> nil
            Then Begin
               gotoxy(1,13);
               clreol;
               gotoxy(30,13);
               write('Writing .TST file.');
               writeTestStatistics(data1,
                                   data2,
                                   tipCount,
                                   userOptions,
                                   permutationCount);
            End {then}
            Else Begin
               gotoxy(20,13);
               clreol;
               write('Tree too big, so no .TST file will be written.');
               ret('Press Enter to return to main menu.')
            End; {else }
         End; {procedure permutationMenuProc}


      Begin
         done := false;
         succ1 := false;
         succ2 := false;
         msucc1 := false;
         msucc2 := false;
         userOptions.seed := RandSeed;        { Get the system seed to display as a default }
         userOptions.exclude := false;        { allow self permutations }
         userOptions.generator := intrinsic;  { Use the system random # generator by default }
         userOptions.myMethod := similarity;  { use similarity algorithm for probabilities }
         userOptions.k := 2.0;                { default value for k }
         userOptions.blPower := 1.0;          { Default Transform value }
         userOptions.logTrans := false;       { Default log Transform value }
         tree1Height := 0;          { necessary initialization }
         tree2Height := 0;          { likewise necessary }
         While (done = false) Do Begin
            showMenu(userOptions);
            write(': ');
            readln(menuStr);
            writeln;
            Case menuStr[1] Of

               'C',      {change the choice of random number generator }
               'c': Begin
                       If (userOptions.generator = manley)
                       Then Begin
                          writeln('The program currently uses the random number generator from Manley (1991).');
                          write('Do you want to switch to Turbo Pascal''s intrinsic generator? (Y/N): ');
                          readln(menuStr);
                          writeln;
                          If ((menuStr[1] = 'Y') or (menuStr[1] = 'y'))
                          Then Begin
                             userOptions.generator := intrinsic;
                             randomize;
                             writeln('Press the enter key to return to the menu');
                             readln;
                          End {then}
                       End {then}
                       Else Begin
                          writeln('The program currently uses Turbo Pascal''s intrinsic random number generator.');
                          write('Do you want to switch to the generator from Manley (1991)? (Y/N): ');
                          readln(menuStr);
                          writeln;
                          If ((menuStr[1] = 'Y') or (menuStr[1] = 'y'))
                          Then Begin
                             userOptions.generator := manley;
                             {manRandomize;}
                             ix := 6788;
                             iy := 6788;
                             iz := 6788;
                             writeln('Press the enter key to return to the menu');
                             readln;
                          End {then}
                       End {else}
                    End; {case C,c}

               'T',      {specify the branch length transformation}
               't': Begin
                       write('Do you want a log10 transform (as opposed to a exponential transform)? (Y/N) ');
                       readln(userChar);
                       writeln;
                       If ((userChar = 'Y') or (userChar = 'y'))
                       Then userOptions.logTrans := true
                       Else Begin
                          userOptions.logTrans := false;
                          write('Enter the power you want to transform branch length sums by: ');
                          readln(powerStr);
                          writeln;
                          If (length(powerStr) = 0)
                          Then writeln('The current power will not be changed.')
                          Else Begin
                             Val(powerstr,tempReal,valCode);
                             If (valCode <> 0)
                             Then writeln('Error - Invalid entry, transform not changed.')
                             Else userOptions.blPower := tempReal;
                          End; { else }
                       End; { else }
                       write('Press enter to return to the main menu');
                       readln
                    End; {case of T,t }

               'M',
               'm': Begin {change the method of building probability matrix}
                       If userOptions.myMethod = similarity
                       Then Begin
                          writeln('The program currently calculates probabilities using a similarity');
                          writeln('score: probability = k-(dist/maxdist).  Would you like to change');
                          writeln('to using using probabilities based on reciprocals of the distance?');
                          write('(Y/N): ');
                          readln(menuStr);
                          writeln;
                          If (menuStr[1] in ['Y','y']) Then Begin
                             userOptions.myMethod := distance;
                             userOptions.factor := 2.0;
                             write('Self distance factor defaulted to ');
                             extendWrite(output,userOptions.factor);
                             writeln;
                             writeln('Press the enter key to return to the main menu');
                             readln;
                          End {then}
                       End {then}
                       Else Begin
                          writeln('The program currently calculates probabilities based on the reciprocal of');
                          writeln('the distance between two tips.  Would you like to change to using a');
                          write('similarity score: probability = k-(dist/maxdist) (Y/N)?');
                          readln(menuStr);
                          writeln;
                          If (menuStr[1] in ['Y','y']) Then Begin
                             userOptions.myMethod := similarity;
                             userOptions.k := 2.0;
                             write('Similarity constant (K) defaulted to ');
                             extendWrite(output,userOptions.k);
                             writeln;
                             writeln('Press the enter key to return to the main menu');
                             readln;
                          End {then}
                       End
                    End; {case M,m }

               'F',      {specify the self distance factor }
               'f': Begin
                       If (userOptions.myMethod = distance) Then Begin
                          writeln('Enter the multiplier to use to calculate the distance from a node to itself');
                          write('from its length to closest ancestor (or enter to see current value): ');
                          readln(factorStr);
                          writeln;
                          If (length(factorStr) = 0)
                          Then writeln('The current multiplier for the distance to ancestor = ',userOptions.Factor)
                          Else Begin
                             Val(factorStr,tempReal,valCode);
                             If (valCode <> 0)
                             Then writeln('Error - Invalid entry, self distance factor not changed')
                             Else Begin
                                If tempReal = 0
                                Then Begin
                                   writeln('Warning - A multiplier of 0 makes calculating a PRB matrix or');
                                   writeln('          generating permutations impossible.  For 0 self');
                                   writeln('          distances try the similarity method.');
                                End;
                                userOptions.Factor := tempReal;
                             End; {else}
                          End; { else }
                          write('Press enter to return to the main menu');
                          readln
                       End {then}
                    End; {case of F,f }

               'K',      {specify the constant in the similarity formula}
               'k': Begin
                       If (userOptions.myMethod = similarity) Then Begin
                          writeln('Enter the constant to use to calculate the similarity index between');
                          write('two tips (or enter to see current value): ');
                          readln(factorStr);
                          writeln;
                          If (length(factorStr) = 0)
                          Then writeln('The current value of k = ',userOptions.k)
                          Else Begin
                             Val(factorStr,tempReal,valCode);
                             If (valCode <> 0)
                             Then writeln('Error - Invalid entry, K not changed')
                             Else userOptions.k := tempReal;
                          End; { else }
                          write('Press enter to return to the main menu');
                          readln
                       End {then}
                    End; {case of K,k}

               'I',      { read tree or matrix files }
               'i': Begin
                       If ((menuStr = 'iprb') or (menuStr = 'IPRB'))
                       Then Begin
                          matrixDialog(msucc1,msucc2);
                       End {then}
                       Else Begin
                          treeDialog(succ1,succ2);
                       End; {else}
                    End; {case I,i }


               'D',      {write the distance matrices to a file }
               'd': Begin
                       If ((data1^.tree = nil) Or (data2^.tree = nil))
                       Then Begin
                          If ((data1^.matrix <> nil) or (data2^.matrix <> nil))
                          Then Begin
                             writeln('Error - Can''t write a distance matrix from a .PRB matrix')
                          End {then}
                          Else Begin
                             If (data1^.tree = nil)
                             Then writeln('Error - no tree for trait 1 specified.')
                             Else writeln('Error - no tree for trait 2 specified.')
                          End {else}
                       End {then}
                       Else Begin
                          userOptions.showExtended := false;
                          {$IFOPT N+}  {only ask if we are compiling with coprocessor}
                          write('Do you want to write the values in full precision? (Y/N): ');
                          readln(userChar);
                          If (upcase(userChar) = 'Y')
                          Then userOptions.showExtended := true
                          Else userOptions.showExtended := false;
                          {$ENDIF}
                          If (data1^.sourceName <> data2^.sourceName)
                          Then Begin  { don't write the same file twice!}
                             showDistances(data1,tipCount,userOptions);
                             showDistances(data2,tipCount,userOptions);
                          End {then}
                          Else Begin
                             writeln('Writing only one distance file for both traits.');
                             showDistances(data1,tipCount,userOptions);
                          End {else}
                       End; { else }
                       writeln;
                       writeln('Press the enter key to return to the main menu');
                       readln
                    End; {case of D,d }

               'P',      {write the probability matrices to a file}
               'p': Begin
                       If (((data1^.tree = nil) or (data2^.tree = nil)) And
                           ((data1^.matrix = nil) or (data2^.matrix = nil)) Or
                           ((userOptions.myMethod = distance) And
                            (userOptions.Factor = 0)))
                       Then Begin
                          If ((userOptions.myMethod = distance) And
                              (userOptions.Factor = 0))
                          Then writeln('Error - Self factor = 0; can''t calculate .PRB matrix');
                          If ((data1^.tree = nil) And (data1^.matrix = nil))
                          Then Begin
                             writeln('Error - no tree or matrix for trait 1 specified');
                          End; {then}
                          If ((data2^.tree = nil) And (data2^.matrix = nil))
                          Then Begin
                             writeln('Error - no tree or matrix for trait 2 specified');
                          End; {then}
                       End {then}
                       Else Begin
                          userOptions.showExtended := false;
                          {$IFOPT N+}  {only ask if we are compiling with coprocessor}
                          write('Do you want to write the values in full precision? (Y/N): ');
                          readln(userChar);
                          If (upcase(userChar) = 'Y')
                          Then userOptions.showExtended := true
                          Else userOptions.showExtended := false;
                          {$ENDIF}
                          If (data1^.sourceName <> data2^.sourceName)
                          Then Begin  { don't write the same file twice!}
                             maxDistance(data1,tipCount,userOptions);
                             maxDistance(data2,tipCount,userOptions);
                             showProbabilities(data1,tipCount,userOptions);
                             showProbabilities(data2,tipCount,userOptions);
                          End {then}
                          Else Begin
                             maxDistance(data1,tipCount,userOptions);
                             writeln('Writing only one probability file for both traits.');
                             showProbabilities(data1,tipCount,userOptions);
                          End {else}
                       End; { else }
                       writeln;
                       writeln('Press the enter key to return to the main menu');
                       readln
                    End; {case of P,p }

               'N',      {specify the random number seed}
               'n': Begin
                       write('Enter a seed number or blank to display current seed: ');
                       readln(seedStr);
                       writeln;
                       If (length(seedStr) = 0)
                       Then writeln('Current seed is: ',RandSeed)
                       Else Begin
                          Val(seedStr,userOptions.Seed,valCode);
                          If (valCode <> 0)
                          Then Begin
                             writeln('Error - Invalid entry, seed not changed.');
                             userOptions.Seed := RandSeed
                          End {then}
                          Else RandSeed := userOptions.Seed
                       End; { else }
                       write('Press enter to return to the main menu');
                       readln
                    End; {case of N,n }



               'R',      {Randomize and save a set of data }
               'r': Begin
                       If (((data1^.tree = nil) or (data2^.tree = nil)) And
                           ((data1^.matrix = nil) or (data2^.matrix = nil)) Or
                           ((userOptions.myMethod = distance) And
                            (userOptions.Factor = 0)))
                       Then Begin
                          If ((userOptions.myMethod = distance) And
                              (userOptions.Factor = 0))
                          Then writeln('Error - Self factor = 0; can''t randomize');
                          If ((data1^.tree = nil) And (data1^.matrix = nil))
                          Then Begin
                             writeln('Error - no tree or matrix for trait 1 specified');
                          End; {then}
                          If ((data2^.tree = nil) And (data2^.matrix = nil))
                          Then Begin
                             writeln('Error - no tree or matrix for trait 2 specified');
                          End; {then}
                          writeln;
                          writeln('Press the enter key to return to the main menu');
                          readln;
                       End {then}
                       Else permutationMenuProc;
                       writeln;
                    End; {case of R,r }

               'E',      {Randomize equiprobably and save a set of data }
               'e': Begin
                       savedMethod := userOptions.myMethod;
                       userOptions.myMethod := equiprobable;
                       If (((data1^.tree = nil) or (data2^.tree = nil)) And
                           ((data1^.matrix = nil) or (data2^.matrix = nil)))
                       Then Begin
                          If ((data1^.tree = nil) And (data1^.matrix = nil))
                          Then Begin
                             writeln('Error - no tree or matrix for trait 1 specified');
                          End; {then}
                          If ((data2^.tree = nil) And (data2^.matrix = nil))
                          Then Begin
                             writeln('Error - no tree or matrix for trait 2 specified');
                          End; {then}
                          writeln;
                          writeln('Press the enter key to return to the main menu');
                          readln;
                       End {then}
                       Else permutationMenuProc;
                       writeln;
                       userOptions.myMethod := savedMethod;
                    End; {case of E,e }

               'Q',      {exit this loop}
               'q': Begin
                       done := true;
                    End; {case of Q,q }
            End; {case}
         End; {while}
      End; {procedure doMenu}


Begin { main program }
   randomize;          {initialize both generators}
   {manRandomize;}
   ix := 6788;         {initialize Manley generator to known state}
   iy := 6788;
   iz := 6788;
   new(data1);         {allocate and initialize structures for phylogeny data}
   new(data2);
   initData(data1);
   initData(data2);
   initProgramName('PDRANDOM');   {set up program name for title screen}
   doTitleScreen;
   doMenu;                        {run the menu}
   doSummary;                     {display a summary of the session (someday...)}
End. { program pdrandom}

