{This program reads a file of simulated data (.SIM) and a file containing
 a tree (.PDI) and replaces the simulated data for one of the two traits
 with the corresponding data from the tree file.  The replacement occurs
 for each simulated data set.

 Modification history:

    20 March 1998 - Coding begun P. Midford

     1 April 1998 - First version completed PEM

    26 May 1998 - Second version completed PEM

    15 June 1998 - Additional enhancements PEM

    20 August 2004 - Released by TG
}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R+} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Program PDSIMMOD;
   uses DOS,
        CRT;
   Const maxSets = 15;

   {* Because sets are used, the program is limited to trees with 255 tips *}
   Type speciesSet = set of 1..255;  
                                      
        setInfo = Record
                      species : speciesSet;
                      Case tweakType: char of
                           'M' : (Mean : extended;
                                  SD :  extended);
                           'C' : (ConstValue : extended);
                      End; {type setInfo}
        setArray = array[0..maxSets] of setInfo;

        traitNode = Record
                       value1 : extended;
                       value2 : extended;
                       name   : string[2];
                    End; {type traitnode}
        traitArray = array[1..255] of traitNode;

        {If the format of SIM files ever changes, the length of this
         line may need to be expanded. !!!! PEM 15 June 1998}

        linePtr = ^lineList;
        lineList = Record
                      line : string[80];
                      next : linePtr;
                   End;


   Var simDataName : pathStr; {name of file of simulated data}
       treeName : pathStr;    {name of tree file with replacement values}
       outName : pathStr;     {name of file to receive modified data set}
       IDStr : string;        {holds program's name and time the run started}
       fileBuffer : linePtr;  {}
       sourceTrait : integer; {which trait (1 or 2) to use in replacement}
       destTrait : integer;   {which trait (1 or 2) to replace}
       setCount1 : integer;   {number of replacement sets for trait 1}
       allSets1 : setArray;   {array of sets for trait 1}
       setCount2 : integer;   {number of replacement sets for trait 2}
       allSets2 : setArray;   {array of sets for trait 2}
       tipCount : integer;    {tips in the tree}
       tips : traitArray;     {trait values and names for each tip}
       doReplacements : boolean;   {true to do substitutions from a PDI file}
       doTipMods : boolean;        {true to add values to tip traits}

   {*******************************************************************}
   {This function displays the intro screen and lets the user decide
    to terminate the program after seeing the screen. }
   {*******************************************************************}
   Procedure intro;
      Var userChar : char;
      Begin
         Clrscr;
         Writeln(' ********* Phenotypic Diversity Analysis Program - PDSIMMOD Module *********');
         Writeln;
         Writeln(' V. 1.0 Copyright 20 Aug. 2004 by Peter E. Midford and Theodore Garland, Jr.');
         Writeln('        Direct all questions to T.G. at Department of Biology,');
         Writeln('         University of California, Riverside, CA 92521  U.S.A.');
         Writeln('Telephone 951-787-3524; Facsimile 951-787-4286; E-mail tgarland@ucr.edu');
         Writeln;
         Writeln('This program works in combination with PDSIMUL.  It reads a file of simulated');
         Writeln('character (tip) data (*.SIM) from PDSIMUL, then modifies the tip data.  These');
         Writeln('files can then be analyzed with such programs as PDANOVA, PDERROR or PDTIPS.');
         Writeln;
         Writeln('Two kinds of modifications can be performed, in the following sequence.');
         Writeln('  1. Replace the data for one of the two traits in each set of simulated');
         Writeln('     data with the corresponding trait value from a *.PDI file (from PDTREE).');
         Writeln('  2. Change data by adding constants or normalized errors to specified subsets');
         Writeln('     of tips.  Different treatments can be applied to different subsets.');
         Writeln;
         Writeln('The modified simulation sets are then written to a file with the same name as');
         Writeln('the original *.SIM file, but with an extension of either .SR2 or .S1R depending');
         Writeln('on whether trait 1 or trait 2 was replaced (in step 1.), respectively.');
         Writeln('If neither was replaced in step 1, then the output file will be named *.S12.');
         Writeln;
         Writeln('This program is limited to trees of 255 or fewer tips.');
         Writeln;
         Write('Please press any key to continue');
         userChar := readKey;
         clrscr;
      End; {procedure intro}

   {************************************************************************}
   {This function is installed so that when the heap is exhausted, nil is
    returned by a call to new, rather than signaling an error.  See p. 217
    of the Turbo Pascal 6.0 Programmer's Guide.}
   {************************************************************************}
   Function heapFunc(size: word) : integer; Far;
      Begin
         heapFunc := 1;
      End; {function heapFunc}

   {************************************************************************}
   {This procedure writes coprocessor extended reals to a file without the
    extra exponent and mantessa digits.  See p. 194 of the Turbo Pascal 6.0
    Programmer's Guide.}
   {************************************************************************}
   Procedure extendWrite(Var outFile :text;num :real);
      Var outStr :string[19];
      Begin
         str(num:19:-11,outStr);
         delete(outStr,16,2);
         write(outFile,outStr)
      End; {procedure extendwrite}

   {*************************************************************************}
   {This procedure outputs the tree name, the date, and the time.  It figures
    out the date and time from dos calls.}
   {*************************************************************************}
   Procedure dateAndTime(Var timeStr : string);
      Var
         yr,mo,day,dow,h,m,s,s100:word;   {holds the values returned from DOS}
         yr2,mo2,day2,h2,m2,s2:string[5]; {text string fragments}
      Begin
         getdate(yr,mo,day,dow);
         gettime(h,m,s,s100);
         str(yr,yr2);
         Case mo Of
            1: mo2:='January';
            2: mo2:='February';
            3: mo2:='March';
            4: mo2:='April';
            5: mo2:='May';
            6: mo2:='June';
            7: mo2:='July';
            8: mo2:='August';
            9: mo2:='September';
           10: mo2:='October';
           11: mo2:='November';
           12: mo2:='December';
         End; {case mo}
         str(day,day2);
         str(h,h2);
         str(m,m2);
         If m<10 Then m2 := '0'+m2;
         str(s,s2);
         If s<10 Then s2 := '0'+s2;
         timeStr := day2+' '+mo2+' '+yr2+' '+h2+':'+m2+':'+s2
      End; {procedure dateAndTime}

   {*******************************************************************}
   {This function  builds the list of lineList nodes.                  }
   {*******************************************************************}
   Procedure buildBuffer(Var buffList : linePtr);
      Var cur : linePtr;
      Begin
         new(cur);
         buffList := cur;
         While (cur <> NIL) Do Begin
            new (cur^.next);
            cur := cur^.next;
         End; {while}
      End; {procedure buildBuffer}

   {*******************************************************************}
   {This function frees the list of lineList nodes.                    }
   {*******************************************************************}
   Procedure freeBuffer(buffList : linePtr);
      Var i : integer;
          cur : linePtr;
          next : linePtr;
      Begin
         cur := buffList;
         While (cur <> NIL) Do Begin
            next := cur^.next;
            dispose(cur);
            cur := next;
         End; {while}
      End; {procedure buildBuffer}

   {*******************************************************************}
   {This procedure reads the pathnames of the input and output files.
    It also does some simple checking of presumptive .SIM and .PDI
    files, and determines whether the user wants to do replacements.   }
   {*******************************************************************}
   Procedure getFiles(Var sName : pathStr;
                      Var pName : pathStr;
                      Var oName : pathStr;
                      Var sourceTrait : integer;
                      Var destTrait : integer;
                      Var doReplacements : boolean);
      Var
         rawName : pathStr;   {holds name entered by user}
         checkDir : dirStr;   {holds dir string returned by FSplit}
         checkName : nameStr; {holds name string returned by FSplit}
         checkExt : extStr;   {holds extension string returned by FSplit}
         outExt : extStr;     {holds new extension for output file}
         userChar : char;     {for receiving single character input}
         ok : boolean;        {true when a filename checks out}

      {**************************************************************}
      {This procedure checks for an empty extension field in name and}
      {replaces such an empty extension with the extension specified }
      {in ext.  This extension should include a '.' as the first     }
      {character.                                                    }
      {**************************************************************}
      Procedure makeDefault(Var name : pathStr; ext : extStr);
         Begin
            FSplit(name,checkDir,checkName,checkExt);
            If checkExt = '' Then Begin
               checkExt := ext;
               name := checkDir+checkName+checkExt
            End {then}
         End; {procedure makeDefault}

      {***************************************************************}
      {This function checks that a specified name corresponds to a
       file in .PDI format.  This function performs very limited
       format checking, to wit, that the first line contains an
       integer.                                                       }
      {***************************************************************}
      Function checkPDIFile (name : pathStr) : boolean;
         Var dummy : text;        {actually opened to check the file}
             valCheck : string;   {holds the first line of the file}
             sizeCheck : integer; {holds the presumptive tree size}
             valCode : integer;   {if this is not 0, file can't be good}
             fileCheck : boolean; {true if the file exists}
         Begin
            {$I-}
            assign(dummy,name);
            reset(dummy);
            {$I+}
            fileCheck := (IOResult = 0);
            If fileCheck Then Begin
               readln(dummy,valCheck);
               val(valCheck,sizeCheck,valCode);
               checkPDIFile := (valCode = 0);
               close(dummy);
            End {then}
            Else checkPDIFile := false;
         End; {function checkFile}
      {***************************************************************}
      {This function checks that a specified named corresponds to a
       file in .SIM format.  This function performs very limited
       format checking, to wit, that first character in the file is
       an upcase 'O' (which is the start of "Output =").              }
      {***************************************************************}
      Function checkSIMFile(name : pathStr) : boolean;
         Var dummy : text;         {actually opened to check the file}
             valCode : integer;    {holds code returned by val}
             i : integer;          {counter for first three lines}
             sizeCheck : integer;  {holds integer value returned by val}
             fileCheck : boolean;  {true if the file exists}
             checkString : string; {holds the fourth line of the file}
         Begin
            {$I-}
            assign(dummy,name);
            reset(dummy);
            {$I+}
            fileCheck := (IOResult = 0);
            If fileCheck Then Begin
               For i := 1 To 3 Do readln(dummy);
               readln(dummy,checkString);
               val(copy(checkString,24,3),sizeCheck,valCode);
               checkSIMFile := (valCode = 0);
               close(dummy);
            End
            Else checkSIMFile := false;
         End; {function checkSIMFile}
      {***************************************************************}
      Begin {procedure getFiles}
         writeln('Enter the filename and extension of the simulated data file');
         write('(typically *.SIM from PDSIMUL): ');
         readln(rawName);
         ok := checkSIMFile(rawName);
         While not(ok) Do Begin
            writeln('File ',rawName,' was either not found or not a .SIM file;');
            write('please enter another name: ');
            readln(rawName);
            makeDefault(rawName,'.SIM');
            ok := checkSIMFile(rawName);
         End; {while}
         sName := rawName;
            writeln;
            writeln('******************************** Replace Module ******************************');
            writeln;
         writeln('Do you want to replace simulated values with ');
         write('values from a tree (.PDI) file? (Y/N): ');
         userChar := chr(32);
         readln(userChar);
         userChar := upCase(userChar);
         While Not(userChar in ['Y','N']) Do Begin
            write('Please enter Y or N: ');
            readln(userChar);
            userChar := upCase(userChar);
         End; {while}
         If userChar = 'N'
         Then Begin
            doReplacements := false;
            FSplit(sName,checkDir,checkName,checkExt);
            outName := checkDir+checkName+'.S12';
            sourceTrait := 0;
            destTrait := 0; {set for sanity and to avoid duplicating SIM file}
         End {then}
         Else Begin
            doReplacements := true;
            Repeat
               write('Enter 1 to replace the first trait or 2 for the second trait (1,2): ');
               readln(destTrait);
            Until destTrait in [1,2];
            If destTrait = 1 Then outExt := '.SR2' Else outExt := '.S1R';
            FSplit(sName,checkDir,checkName,checkExt);
            outName := checkDir+checkName+outExt;
            writeln;
            writeln('Modified file will be written to ',outName);
            writeln;
            writeln('Enter the root name of the tree file, which must be in .PDI format: ');
            readln(rawName);
            ok := checkPDIFile(rawName);
            makeDefault(rawName,'.PDI');
            ok := checkPDIFile(rawName);
            While not(ok) Do Begin
               writeln('File ',rawName,' was either not found or not a .PDI file;');
               write('please enter another name: ');
               readln(rawName);
               makeDefault(rawName,'.PDI');
               ok := checkPDIFile(rawName);
            End; {while}
            pName := rawName;
            Repeat
               writeln('Enter 1 to use the first trait for replacement or 2 to');
               write('use second trait (1,2): ');
               readln(sourceTrait);
            Until sourceTrait in [1,2];
         End {else}
      End; { procedure getFiles }

   {******************************************************************}
   {This function scans the tree file (in either .PDI or the top of a
    .SIM file) and fills an array of trait values used to do the
    replacement or modification.                                      }
   {******************************************************************}
   Procedure fillTipTable(pName : pathStr;         {PDI filename}
                          Var result : traitArray; {array of tips}
                          Var tipCount : integer;  {number of tips}
                          PDIFile : boolean);      {true=.PDI;false=.SIM}
      Var
         treeFile : text;    {the file}
         valCode : integer;  {holds status value returned from val}
         lineBuff : string;  {holds line read from the file}
      {*************************************************************}
      {This procedure scans a file in .PDI format.                  }
      {*************************************************************}
      Procedure scanPDIFile;
         Var
            i : integer;   {counter for reading lines with tip data}
         Begin
            readln(treeFile,tipCount); {first line contains the # of tips}
            readln(treeFile);
            For i := 1 to 2*tipCount-1 Do readln(treeFile,lineBuff);
            readln(treeFile,lineBuff);
            For i := 1 to tipCount Do Begin
               readln(treeFile,lineBuff);
               result[i].name := copy(lineBuff,1,2);
               val(copy(lineBuff,4,18),result[i].value1,valCode);
               val(copy(lineBuff,24,18),result[i].value2,valCode);
            End; {for i}
         End; {local procedure scanPDIFile}

      {*************************************************************}
      {This procedure scans a file in .SIM format.                  }
      {*************************************************************}
      Procedure scanSIMFile;
         Var
            i : integer;   {counter for reading lines with tip data}
         Begin
            For i := 1 To 3 Do readln(treeFile);
            readln(treeFile,lineBuff);  {read the line with tipCount}
            val(copy(lineBuff,24,3),tipCount,valCode);
            For i := 1 To tipCount+17 Do readln(treeFile);
            For i := 1 To tipCount Do Begin
               readln(treeFile,lineBuff);
               result[i].name := copy(lineBuff,1,2);
               result[i].value1 := 0;
               result[i].value2 := 0;
            End; {for i}
         End; {local procedure scanSIMFile}

      Begin {*******************************************************}
         assign(treeFile,pname);
         reset(treeFile);
         If PDIFile
         Then scanPDIFile
         Else scanSIMFile;
         close(treeFile);
      End; {function fillTipTable}

   {********************************************************************}
   {This function takes a species index number and an array of sets and }
   {returns the index of the first set that contains the species.       }
   {********************************************************************}
   Function findSet(namePos : integer;          {species index}
                    sets : setArray;            {all the sets}
                    setCount : integer;         {how many sets}
                    start : integer) : integer; {set to start looking at}
      Var i : integer;
          done : boolean;
      Begin
         If ((namePos>255) Or (namePos < 1))
         Then findSet := -1
         Else Begin
            i := start+1;
            done := false;
            Repeat
               If (namePos In sets[i].species)
               Then done := true
               Else inc(i);
            Until (done Or (i > setCount));
            If (i > setCount)
            Then findSet := -1
            Else findSet := i;
         End; {else}
      End; {function findSet}

   {********************************************************************}
   {This function fills the file buffer with as many lines as it can.
    It returns true if it encounters end of file.                       }
   {********************************************************************}
   Function fillBuff(Var source : text;
                     buff : linePtr;
                     Var bufEnd : linePtr) : boolean;
      Var cur : linePtr;
      Begin
         cur := buff;
         readln(source,cur^.line);
         While not(eof(source) Or (cur^.next = nil)) Do Begin
            cur := cur^.next;
            readln(source,cur^.line);
         End; {while}
         bufEnd := cur^.next;
         fillBuff := eof(source);
      End; {function fillBuff}

   {********************************************************************}
   {This function empties the file buffer into the output file.         }
   {********************************************************************}
   Procedure dumpBuff(Var sink : text;
                     buff : linePtr;
                     bufEnd : linePtr);
      Var cur : linePtr;
      Begin
         cur := buff;
         While (cur <> bufEnd) Do Begin
            writeln(sink,cur^.line);
            cur := cur^.next;
         End; {while}
      End; {procedure dumpBuff}

   {********************************************************************}
   {This procedure opens the files and performs the conversions.        }
   {********************************************************************}
   Procedure processFiles(sName   : pathStr;    {name of .SIM file}
                          pName   : pathStr;    {name of .PDI file}
                          outName : pathStr;    {name of output file}
                          traits : traitArray;  {holds trait values of tips}
                          sourceTrait : integer; {trait use for replacing}
                          destTrait : integer;  {trait to be replaced}
                          doTipMods : boolean;  {true to perform modifications}
                          setCount1 : integer;  {number of sets in trait 1}
                          allSets1 : setArray;  {sets for trait 1}
                          setCount2 : integer;  {number of sets in trait 2}
                          allSets2 : setArray); {sets for trait 2}
      Var
         trait1 : extended;     {holds trait value from simulation file}
         trait2 : extended;     {holds trait value from simulation file}
         errorVal : extended;   {holds normally distributed "error" value}
         z1,z2 : extended;      {hold two N[0,1] distributed numbers}
         sFile : text;          {.SIM file}
         pFile : text;          {.PDI file}
         newFile : text;        {result/output file}
         curBuff : linePtr;
         buffEnd : linePtr;
         lineLength : integer;  {holds the length of the line, for copying}
         traitSet1 : integer;   {index of set for trait 1 for this tip}
         traitSet2 : integer;   {index of set for trait 2 for this tip}
         namePos : integer;     {holds index of tip returned by lookupName}
         sCount : integer;      {number of tips in SIM file (error checking)}
         valCode : integer;     {holds error codes returned by val}
         i : integer;           {general purpose counter}
         simCount : integer;    {counts simulations processed}
         lineCount : integer;   {used to "guess" number of simulations processed}
         name : string[2];      {holds the name of a tip}
         valBuf : string[20];   {buffer used for parsing numbers}
         lineBuff : string;     {holds current line from .SIM file}
         str1 : string[20];     {holds converted trait 1 value string}
         str2 : string[20];     {holds converted trait 2 value string}
         done : boolean;        {true sim file completely processed}
      {******************************************************************}
      {This function finds the record in the trait list with a name field
       that matches name and returns the pointer to that record.  Note
       that as currently implemented, not finding the name is fatal.
       It would be nice to cache the results from this function.         }
      {******************************************************************}
      Function lookupName(name : string;             {the name to lookup}
                          tArray : traitArray;       {array of tips}
                          last : integer) : integer; {number of tips}
         Var current : integer;            {counter for search}
             result : integer;             {location of matching name}
         Begin
            current := 1;
            result := -1;
            Repeat
               If tArray[current].name = name
               Then result := current
               Else inc(current)     {inc is supposed to be faster...}
            Until ((current > last) Or (result > 0));
            If result = 0 Then Begin
               writeln('Name ',name,' not found.  Exiting...');
               halt;
            End; {then}
            lookupName := result;
         End; {function lookupName}

      {******************************************************************}
      {This procedure provides two independent variables from a normal
       distribution.  Borrowed from PDSIMUL.                             }
      {******************************************************************}
      Procedure normalPair (Var X, Y : extended);
         Var
            A, B, R, Theta : extended;
         Begin
            a := Random;
            b := Random;
            Theta := 2*PI*A;
            R := sqrt(-2*(ln(B)));
            X := R*(cos(Theta));
            Y := R*(sin(Theta))
         End;  {normalpair}
      {******************************************************************}
      Begin {procedure process files}
         assign(sFile,sName);
         assign(newFile,outName);
         reset(sFile);
         rewrite(newFile);
         readln(sFile,lineBuff);    {copy the first list}
         writeln(newFile,lineBuff);
         readln(sFile,lineBuff);
         lineLength := length(lineBuff);
         lineBuff := copy(lineBuff,1,lineLength-1);
         lineBuff := lineBuff + '; modified by ' + IDStr + '.';
         writeln(newFile,lineBuff);  {write modified second line}
         For i := 3 to 4 Do Begin   {copy third and fourth lines}
            readln(sFile,lineBuff);
            writeln(newFile,lineBuff)
         End; {for i}
         valBuf := copy(lineBuff,24,3);
         val(valBuf,sCount,valCode);
         If (sCount <> tipCount) Then Begin
            writeln('SIM file and PDI file have different numbers of tips.');
            writeln('Exiting...');
            halt;
         End;
         While (lineBuff[1] <> '_') Do Begin
            readln(sFile,lineBuff);
            writeln(newFile,lineBuff);
         End; {while}
         done := false;
         lineCount := 0;
         buildBuffer(fileBuffer);
         clrscr;
         gotoxy(25,12);
         writeln('Processing simulation set 1');
         simCount := 1;
         While (not(done)) Do Begin
            done := fillBuff(sFile,fileBuffer,buffEnd);
            curBuff := fileBuffer;
            While (curBuff <> buffEnd) Do Begin
               If lineCount < tipCount
               Then inc(lineCount)
               Else Begin
                  inc(simCount);
                  lineCount := 1;
                  gotoxy(25,12);
                  writeln('Processing simulation set ',simCount)
               End; {else}
               name := copy(curBuff^.line,1,2);
               namePos := lookupName(name,traits,tipCount);
               valBuf := copy(curBuff^.line,4,17);
               val(valBuf,Trait1,valCode);
               valBuf := copy(curBuff^.line,22,17);
               val(valBuf,trait2,valCode);
               Case sourceTrait Of
                    1: Case destTrait Of
                            1: trait1 := traits[NamePos].value1;
                            2: trait2 := traits[NamePos].value1;
                       End; {case destTrait}
                    2: Case destTrait Of
                            1: trait1 := traits[NamePos].value2;
                            2: trait2 := traits[NamePos].value2;
                       End; {case destTrait}
               End; {case sourcetrait}
               If doTipMods Then Begin
                  If setCount1 > 0 Then Begin
                     traitSet1 := findSet(namePos,allSets1,setCount1,0);
                     While (traitSet1 <> -1) Do Begin
                        If (allSets1[traitSet1].tweakType = 'M') Then Begin
                           normalPair(z1,z2);
                           With allSets1[traitSet1] Do errorVal := z1*SD+mean;
                           trait1 := trait1 + errorVal;
                        End
                        Else trait1 := trait1 + allSets1[traitSet1].constValue;
                        traitSet1 := findSet(namePos,allSets1,setCount1,traitSet1);
                     End; {while}
                  End {then}
                  Else If SetCount1 = 0 Then Begin
                     If (allSets1[0].tweakType = 'M') Then Begin
                        normalPair(z1,z2);
                        With allSets1[0] Do errorVal := z1*SD+mean;
                        trait1 := trait1 + errorVal;
                     End
                     Else trait1 := trait1 + allSets1[0].constValue;
                  End; {else}
                  If setCount2 > 0 Then Begin
                     traitSet2 := findSet(namePos,allSets2,setCount2,0);
                     While (traitSet2 <> -1) Do Begin
                        If (allSets2[traitSet2].tweakType = 'M') Then Begin
                           normalPair(z1,z2);
                           With allSets2[traitSet2] Do errorVal := z1*SD+mean;
                           trait2 := trait2 + errorVal;
                        End
                        Else trait2 := trait2 + allSets2[traitSet2].constValue;
                        traitSet2 := findSet(namePos,allSets2,setCount2,traitSet2);
                     End; {while}
                  End {then}
                  Else If SetCount2 = 0 Then Begin
                     If (allSets2[0].tweakType = 'M') Then Begin
                        normalPair(z1,z2);
                        With allSets2[0] Do errorVal := z1*SD+mean;
                        trait2 := trait2 + errorVal;
                     End
                     Else trait2 := trait2 + allSets2[0].constValue;
                  End; {else}
               End; {then}
               str(trait1:19:-11,str1);
               delete(str1,16,2);
               str(trait2:19:-11,str2);
               delete(str2,16,2);
               curBuff^.line := copy(curBuff^.line,1,2) + ' ' + str1 +
                                   ' ' + str2 + copy(curBuff^.line,39,71);
               curBuff := curBuff^.next;
            End; {while}
            dumpBuff(newFile,fileBuffer,buffEnd);
         End; {while not(done)}
         freeBuffer(fileBuffer);
         close(sFile);
         close(newFile);
         writeln;
         writeln('Done:  modified file written as ',outName,'.');
         writeln;
      End; {procedure processFiles}

   {********************************************************************}
   {This procedure lists all the species by number in an array so the
    user can enter the number codes.                                    }
   {********************************************************************}
   Procedure showTips(tips : traitArray;     {the tips (includes names)}
                      tipCount : integer);   {the number of tips}
      Const tipsLine = 18;
            linesPage = 7;
      Var i : integer;   {counter for outer loop (total tips)}
          j : integer;   {counter for inner loop (tips on one line)}
          userChar : char;
      Begin
         i := 0;
         While (tipsLine*i <= tipCount) Do Begin
            j := 1;
            write(' ');
            While ((j <= tipsLine) And (tipsLine*i+j <= tipCount)) Do Begin
               write(tips[tipsLine*i+j].name,'  ');
               j := j+1;
            End; {while...}
            writeln;
            j := 1;
            While ((j <= tipsLine) And (tipsLine*i+j <= tipCount)) Do Begin
               write(tipsLine*i+j:3,' ');
               j := j+1;
            End; {while...}
            writeln;
            i := i + 1;
            If (tipsLine*i <= tipCount) Then writeln;
            If ((i mod linesPage) = (linesPage-1)) Then Begin
               writeln;
               writeln('Press any key to continue list of tips');
               userChar := readKey;
               clrscr;
            End; {then}
         End; {while i}
      End; {procedure showTips}

   {********************************************************************}
   {This procedure reads the species for a single set.                  }
   {********************************************************************}
   Procedure readSet(tips : traitArray;        {array of tips in the tree}
                     tipCount : integer;       {number of tips}
                     myIndex : integer;        {indicates set to specify}
                     Var dataFile : text;      {file of commands or input}
                     Var listFile : text;      {file to receive messages}
                     Var sets : setArray;      {array of all sets of tips}
                     Var foundError : boolean; {true if an error occured} {file to receive messages}
                     verbose : boolean);       {true if running interactive}

      Var dupSet : speciesSet;  {holds codes found in multiple sets}
          number : integer;     {specifies a tip to add to the set}
          hits : integer;       {used for listing species to a log file}
          checkValue : integer; {used to check for tips in multiple sets}
          i : integer;          {loop counter for non-duplicate report}
          userChar : char;      {holds a character typed by the user}
          reDisplay : boolean;  {true to redisplay the list of species}
          foundDup : boolean;   {true when a species is in a previous set}
          done : boolean;       {true when the set is complete (user entered 0)}
          foo : string;
      {****************************************************************}
      {This procedure displays the instructions for specifying which
       species are to be modified.                                     }
      {****************************************************************}
      Procedure showInstructions;
         Begin
            writeln('******************************************************************************');
            writeln('Enter species for set ', myIndex,' by above numbering.  Enter -1 to redisplay.');
            writeln;
            writeln('Separate numbers with spaces.  Press the "enter" key <return>');
            writeln('*** BEFORE *** you reach the edge of the screen.  After you have');
            writeln('typed in all species for the set, type 0 and "enter" <return> to end.');
         End; {local procedure showInstructions}
      {*****************************************************************}
      Begin {readSet}
         If verbose Then showInstructions;
         sets[myIndex].species := [];
         reDisplay := false;
         done := false;
         foundDup := false;
         dupSet := [];
         Repeat
            Repeat
               read(dataFile,number);
               If number = 0 Then done := true;
               If (eof(dataFile) And Not(Done)) Then Begin
                  writeln('Error, premature end of file');
                  If not(verbose)
                  Then writeln(listFile,'Error, premature end of file');
                  foundError := true;
               End; {then}
               If number = -1 Then redisplay := true;
               checkValue := findSet(number,sets,myIndex-1,0);
               If checkValue>-1 Then Begin
                  writeln(listFile,'Warning: ',number,' is already in set ',checkValue);
                  dupSet := dupSet + [number];
                  foundDup := true;
               End; {then}
               If ((number <-1) Or (number > tipCount))
               Then writeln(listFile,'Error: ',number,' does not correspond to a species code; IGNORED')
               Else If (number > 0)
                    Then sets[myIndex].species := sets[myIndex].species + [number];
            Until (eoln(dataFile) or done);
            readln(dataFile);
            If foundDup Then Begin
               If verbose Then Begin
                  write(listFile,'Do you want to include these species in both sets (Y/N)? ');
                  readln(dataFile,userChar);
                  userChar := upcase(userChar);
                  While not (userChar in ['Y','N']) Do Begin
                     write(listFile,'Enter either Y or N: ');
                     readln(dataFile,userChar);
                     userChar := upcase(userChar);
                  End; {while}
                  If userChar = 'Y'
                  Then sets[myIndex].species := sets[myIndex].species + dupSet
                  Else Begin
                     For i := 1 to tipCount Do
                       If (i in dupSet)
                       Then writeln('Species ',i,' removed from set ',myindex,'.');
                     If (sets[myIndex].species = dupSet)
                     Then writeln('Set ',myIndex,' is now empty');
                  End; {else}
               End; {then}
               foundDup := false;
               dupSet := [];
            End;
            If (redisplay And verbose) Then Begin
               clrscr;
               showTips(tips,tipCount);
               showInstructions;
               redisplay := false;
            End; {then}
         Until done;
         If not(verbose) Then Begin
            writeln(listFile,'Set ',myIndex:2,' contains the following species: ');
            hits := 0;
            For number := 1 To tipCount Do Begin
               If number In sets[myIndex].species Then Begin
                  write(listFile,tips[number].name,' ');
                  inc(hits);
                  If hits > 25 Then Begin
                     writeln(listFile);
                     hits := 0;
                  End; {then}
               End; {then}
            End; {for}
            writeln(listFile);
         End; {then}
         If verbose Then Begin
            writeln('Do you want to add a (C)onstant or');
            write('a (M)ean and standard deviation to this set? ');
         End; {then}
         readln(dataFile,userChar);
         userChar := upCase(userChar);
         If verbose Then Begin
            While Not(userChar in ['M','C']) Do Begin
               write('Enter C for Constant, M for Mean: ');
               readln(dataFile,userChar);
               userChar := upCase(userChar)
            End; {while}
         End
         Else Begin
            If Not(userChar in ['M','C']) Then Begin
               writeln('Error in dataFile C or M expected, found: ',userChar);
               writeln(listFile,'Error in dataFile: C or M expected, found: ',userChar);
               foundError := true;
            End {then}
         End; {else}
         sets[myIndex].tweakType := userChar;
         If sets[myIndex].tweakType = 'M' Then Begin
            If verbose Then write('Enter the mean of the added error: ');
            readln(dataFile,sets[myIndex].mean);
            If verbose Then write('Enter the standard deviation of the added error: ');
            readln(dataFile,sets[myIndex].SD);
            If verbose Then Begin
               While sets[myIndex].SD <= 0 Do Begin
                  writeln('Your standard deviation is negative or zero;');
                  write('please enter a positive value: ');
                  readln(sets[myIndex].SD);
               End; {while}
            End
            Else Begin
               If sets[myIndex].SD <= 0 Then Begin
                  writeln('Error in dataFile, non-positive standard deviation: ',sets[0].SD);
                  writeln(listFile,'Error in dataFile, non-positive standard deviation: ',sets[0].SD);
                  foundError := true;
               End {then}
            End; {else}
            If (Not(verbose) And Not(foundError)) Then Begin
               writeln(listFile,'The set was modified by adding a random value to the trait.');
               writeln(listFile,'The value was drawn from a normal distribution with a mean of ');
               extendWrite(listFile,sets[myIndex].mean);
               write(listFile,' and a standard deviation of ');
               extendWrite(listFile,sets[myIndex].SD);
               writeln(listFile,'.');
               writeln(listFile);
            End; {then}
         End; {then}
         If sets[myIndex].tweakType = 'C' Then Begin
            If verbose Then write('Enter the constant to add to each value: ');
            readln(dataFile,sets[myIndex].constValue);
            If not(verbose) Then Begin
               write(listFile,'The trait was modified by adding ');
               extendWrite(listFile,sets[myIndex].constValue);
               writeln(listFile,' for all species.');
            End; {then}
         End
      End; {procedure readSet}

   {******************************************************************}
   {This procedure fills set 0 if the user specifies that the same
    change should be performed to all species.                        }
   {******************************************************************}
   Procedure fillSet0(tipCount : integer;        {number of tips}
                      Var sets : setArray;       {array of all sets of tips}
                      Var dataFile : text;       {file of commands or input}
                      Var listFile : text;       {file to receive messages}
                      Var foundError : boolean;  {true when an error occurs}
                      verbose : boolean);        {true if running interactive}
      Var i : integer;             {counter}
          userChar : char;         {holds a number typed by the user}
      Begin
         {first initialize set 0 to all species}
         sets[0].species := [];
         If not(verbose)
         Then writeln(listFile,'All species were modified the same way.');
         For i := 1 to tipCount Do sets[0].species := sets[0].species + [i];
         If verbose Then Begin
            writeln('Do you want to add a (C)onstant or');
            write('a (M)ean and standard deviation to these species? ');
         End;
         readln(dataFile,userChar);
         userChar := upCase(userChar);
         If verbose Then Begin
            While Not(userChar in ['M','C']) Do Begin
               write('Enter C for Constant, M for Mean: ');
               readln(dataFile,userChar);
               userChar := upCase(userChar)
            End; {while}
         End
         Else Begin
            If Not(userChar in ['M','C']) Then Begin
               writeln('Error in dataFile C or M expected, found: ',userChar);
               writeln(listFile,'Error in dataFile: C or M expected, found: ',userChar);
               foundError := true;
            End {then}
         End; {else}
         sets[0].tweakType := userChar;
         If sets[0].tweakType = 'M' Then Begin
            If verbose Then write('Enter the mean of the added error: ');
            readln(dataFile,sets[0].mean);
            If verbose Then write('Enter the standard deviation of the added error: ');
            readln(dataFile,sets[0].SD);
            If verbose Then Begin
               While sets[0].SD <= 0 Do Begin
                  writeln('Your standard deviation is negative or zero;');
                  write('please enter a positive value: ');
                  readln(sets[0].SD);
               End; {while}
            End
            Else Begin
               If sets[0].SD <= 0 Then Begin
                  writeln('Error in dataFile, non-positive standard deviation: ',sets[0].SD);
                  writeln(listFile,'Error in dataFile, non-positive standard deviation: ',sets[0].SD);
                  foundError := true;
               End {then}
            End; {else}
            If (Not(verbose) And Not(foundError)) Then Begin
               writeln(listFile,'The trait was modified by adding a random value.');
               writeln(listFile,'The value was drawn from a normal distribution with a mean of ');
               extendWrite(listFile,sets[0].mean);
               write(listFile,' and a standard deviation of ');
               extendWrite(listFile,sets[0].SD);
               writeln(listFile,'.');
               writeln(listFile)
            End; {then}
         End; {then}
         If sets[0].tweakType = 'C' Then Begin
            If verbose Then write('Enter the constant to add to each value: ');
            readln(dataFile,sets[0].constValue);
            If not(verbose) Then Begin
               write(listFile,'All species were modified by adding ');
               extendWrite(listFile,sets[0].constValue);
               writeln(listFile,' to the trait.');
               writeln(listFile);
            End
         End
      End; {procedure fillSet0}

   {********************************************************************}
   {This procedure asks whether the tip value modifications should be
    performed.  If the user wants the modifications, they are prompted
    for a series of sets of tips and the specification of how to modify
    each set.                                                           }
   {********************************************************************}
   Procedure promptForTipMods (Var doIt : boolean;       {true to do modifications}
                               Var setCount1 : integer;  {number of sets for trait 1}
                               Var allSets1 : setArray;  {array of sets for trait 1}
                               Var setCount2 : integer;  {number of sets for trait 2}
                               Var allSets2 : setArray;  {array of sets for trait 2}
                               simFileName : pathStr;    {name of file of simulated data}
                               pdiFileName : pathStr;    {name of tree file}
                               outputFileName : pathStr; {name of file to received modified data}
                               tips : traitArray;        {array of all the tips in the tree}
                               tipCount : integer);      {number of tips in the tree}
      Var optionFile     : text;      {file of user options and commands}
          listFile       : text;      {file to receive messages}
          i              : integer;   {general purpose counter}
          optionFileName : pathStr;   {name of the file of options, commands}
          listFileName   : pathStr;   {name of file to receive message}
          optionDirName  : dirStr;    {holds directory component of option file path}
          optionFName    : nameStr;   {holds file name component of option file path}
          optionExtName  : extStr;    {holds extension component of option file path}
          userStr        : string;    {holds a string entered by the user}
          userChar       : char;      {holds a single character entered by the user}
          gotOptions     : boolean;   {true when options were successfully specified}
          fileGood       : boolean;   {true when option file name is good}
          doTrait1       : boolean;   {true to modify trait 1}
          doTrait2       : boolean;   {true to modify trait 2}
          foundError     : boolean;   {true when an error has occurred}
      Begin
         foundError := false;  {assume everything is ok at the start}
         writeln;
         writeln('********************************* Change Module ******************************');
         writeln;
         writeln('Would you like to change the data for each of a set of ');
         write('specified species? (Y/N): ');
         readln(userChar);
         If userChar in ['Y','y']
         Then doIt := true
         Else doIt := false;
         If doIt Then Begin
            gotOptions := false;
            writeln('Do you want to read options from a file? ');
            writeln('(If your tree is > 100 species, it will be inconvenient to specify');
            writeln('species sets on the screen.  See PDINSTR.DOC for how to use a batch file.)');
            Repeat
               write('Please enter Y or N: ');
               readln(userChar);
            Until (userChar in ['Y','y','N','n']);
            If userChar in ['Y','y'] Then Begin
               write('Enter the name of the file (press enter to cancel choice): ');
               fileGood := false;
               Repeat
                  readln(optionFileName);
                  {$I-}
                  assign(optionFile,optionFileName);
                  reset(optionFile);
                  {$I+}
                  If optionFileName = ''
                  Then Begin
                     fileGood := true;
                     writeln('Do you want to:');
                     writeln('1  Enter options from the keyboard');
                     writeln('2  Skip specifying changes (proceed to processing)');
                     write(': ');
                     userChar := ' ';
                     Repeat
                        readln(userChar);
                        If Not(userChar in ['1','2'])
                        Then write('Enter either 1 or 2: ');
                     Until userChar in ['1','2'];
                     doIt := (userChar = '1');
                  End
                  Else Begin
                     If IOResult <> 0
                     Then Begin
                        writeln('File ',optionFileName,' does not exist.');
                        writeln;
                        write('Enter a new name or press enter to skip options file: ');
                        fileGood := false
                     End
                     Else fileGood := true;
                  End; {else}
               Until fileGood;
               If (optionFileName <> '') Then Begin
                  fSplit(optionFileName,optionDirName,optionFName,optionExtName);
                  listFileName := optionDirName+optionFName+'.MOD';
                  assign(optionFile,optionFileName);
                  reset(optionFile);
                  assign(listFile,listFileName);
                  rewrite(listFile);
                  readln(optionFile,userStr);
                  writeln(listFile,'Summary file of batch modifications from ',IDStr);
                  writeln(listFile);
                  writeln(listFile,'Options read from ',optionFileName,'.');
                  writeln(listFile);
                  writeln(listFile,'SIM File = ',simFileName);
                  writeln(listFile);
                  writeln(listFile,'PDI File = ',pdiFileName);
                  writeln(listFile);
                  writeln(listFile,'Modified simulations written to ',outputFileName,'.');
                  writeln(listFile);
                  If Pos('1',userStr) > 0
                  Then doTrait1 := true
                  Else doTrait1 := false;
                  If Pos('2',userStr) > 0
                  Then doTrait2 := true
                  Else doTrait2 := false;
                  If doTrait1 Then Begin
                     writeln(listFile,'*** Trait 1 ***');
                     writeln(listFile);
                     readln(optionFile,setCount1);
                     If (setCount1 > 0) Then Begin
                        writeln(listFile,setCount1:2,' sets were defined.');
                        For i := 1 to setCount1 Do Begin
                           readSet(tips,
                                   tipCount,
                                   i,
                                   optionFile,
                                   listFile,
                                   allSets1,
                                   foundError,
                                   false);
                        End {for}
                     End {then}
                     Else fillSet0(tipCount,
                                   allSets1,
                                   optionFile,
                                   listFile,
                                   foundError,
                                   false);
                  End
                  Else setCount1 := -1;
                  If doTrait2 Then Begin
                     writeln(listFile,'*** Trait 2 ***');
                     writeln(listFile);
                     readln(optionFile,setCount2);
                     If (setCount2 > 0) Then Begin
                        writeln(listFile,setCount2:2,' sets were defined.');
                        For i := 1 to setCount2 Do Begin
                           readSet(tips,
                                   tipCount,
                                   i,
                                   optionFile,
                                   listFile,
                                   allSets2,
                                   foundError,
                                   false);
                        End {for}
                     End {then}
                     Else fillSet0(tipCount,
                                   allSets2,
                                   optionFile,
                                   listFile,
                                   foundError,
                                   false);
                  End
                  Else setCount2 := -1;
                  close(optionFile);
                  close(listFile);
                  writeln('Summary of modifications will be written to ',listFileName,'.');
                  writeln;
                  writeln('Press a key to continue');
                  userChar := readKey;
                  gotOptions := true;
               End;
            End; {then}
            If (doIt And not(gotOptions)) Then Begin
               write('Which trait do you want to change? Enter 1, 2 or 12 for both: ');
               readln(userStr);
               While ((pos('1',userStr) = 0) And (pos('2',userStr) = 0)) Do Begin
                  write('Enter either 1, 2 or 12: ');
                  readln(userStr);
               End; {while}
               If pos('1',userStr) > 0
               Then doTrait1 := true
               Else doTrait1 := false;
               If pos('2',userStr) > 0
               Then doTrait2 := true
               Else doTrait2 := false;
               If doTrait1 Then Begin
                  writeln('How many different sets do you want to define for trait 1?');
                  writeln('     You may enter 0 to change all tips the same way: ');
                  readln(setCount1);
                  If (setCount1 > 0) Then Begin
                     clrscr;
                     writeln('The species in your file are as follows:');
                     writeln;
                     showTips(tips,tipCount);
                     For i := 1 to setCount1 Do readSet(tips,
                                                        tipCount,
                                                        i,
                                                        input,
                                                        output,
                                                        allSets1,
                                                        foundError,
                                                        true)
                  End {then}
                  Else fillSet0(tipCount,
                                allSets1,
                                input,
                                output,
                                foundError,
                                true);
               End {then}
               Else setCount1 := -1;
               If doTrait2 Then Begin
                  writeln('How many different sets do you want to define for trait 2?');
                  writeln('You may enter 0 to modify all tips the same way: ');
                  readln(setCount2);
                  If (setCount2 > 0) Then Begin
                     If (setCount1 <= 0) Then Begin
                        clrscr;
                        writeln('The species in your file are as follows:');
                        writeln;
                        showTips(tips,tipCount);
                     End; {then}
                     For i := 1 to setCount2 Do readSet(tips,
                                                        tipCount,
                                                        i,
                                                        input,
                                                        output,
                                                        allSets2,
                                                        foundError,
                                                        true)
                  End {then}
                  Else fillSet0(tipCount,
                                allSets2,
                                input,
                                output,
                                foundError,
                                true);
               End {then}
               Else setCount2 := -1;
               gotOptions := true;
            End; {then}
            If not(gotOptions) Then doIt := false;
         End; {then}
         If foundError Then doIt := false;
      End; {function promptfortipmods}

   {********************************************************************}
   Begin {main program}
      intro;
      heapError := @heapFunc;       {install the alternate heap function}
      dateAndTime(IDStr);
      IDStr := 'PDSIMMOD ' + IDStr;
      getFiles(simDataName,
               treeName,
               outName,
               sourceTrait,
               destTrait,
               doReplacements);
      If doReplacements
      Then fillTipTable(treeName,tips,tipCount,doReplacements)
      Else fillTipTable(simDataName,tips,tipCount,doReplacements);
      promptForTipMods(doTipMods,
                       setCount1,
                       allSets1,
                       setCount2,
                       allSets2,
                       simDataName,
                       treeName,
                       outName,
                       tips,
                       tipCount);
      If (doTipMods or (sourceTrait <> 0)) Then Begin
         processFiles(simDataName,
                      treeName,
                      outName,
                      tips,
                      sourceTrait,
                      destTrait,
                      doTipMods,
                      setCount1,
                      allSets1,
                      setCount2,
                      allSets2);
      End
      Else Begin
         writeln;
         writeln('Error:  No changes selected, no output file will be written!!!');
         writeln;
      End; {else}
   End. {program PDSIMMOD}