{*************************************************************************}
{PDSINGLE.PAS
 mean, variance, min, max -- for all spp for 2 characters
 correlation
 mean, variance, min, max -- for each specified group
 sum of squares among, within = Sokal and Rohlf, 1981, p. 213
     for each specified group

 Modification history:

 18 August 1998 ?? PEM

 3 November 1999 Added block of compiler switches. PEM  

 15 November 1999 Enhanced getTree to accept .PDI files with longer
                  (>2 char) species names (it still throws out the extra
                  characters).  Also fixed bug where the number of seconds
                  was written to the screen rather than the output file.

 16 November 1999 Added tip count to top of .SGL file. PEM

 11 August 2000 Repeats list of species after each set is defined.  PEM

 11 January 2001 Allows user to choose species with 2-letter names,
                 instead of numbers.  PEM

 Last Modified 15 January 2001 PEM}
{*************************************************************************}
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 32768,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
Program PDSINGLE; {perform ANOVA on two characters}
{-------------------------------------------------------------------------}

   Uses Dos; {for getting the date and time}

   Const
      MaxNoSpp = 255; {number of terminal taxa}
      MaxNoSets = 20;

   Type
      TraitArType       = Array [1..MaxNoSpp] of extended;
      SpValType         = Array[1..2] of TraitArType;
      SpSetType         = Set of 1..MaxNoSpp;
      SpSetArType       = Array[0..MaxNoSets] of SpSetType;
      SetAttributesType = Array[0..MaxNoSets] of extended;
      SetPairType       = Array[1..2] of setattributestype;
      NameType          = Array[1..MaxNoSpp] of String[2];
      TraitNameType     = Array[1..2] of String[7];
      SetNameType       = Array[0..MaxNoSets] of String[7];

      SumSquaresType = Record
                          w,            {sum of squares within a set}
                          a,            {sum of squares among sets}
                          t : extended; {total sum of squares}
                       End; {record}
      DfType = Record
                  w,           {degrees of freedom within a set}
                  a,           {degrees of freedom among sets}
                  t : integer; {total degrees of freedom}
               End; {record}

   Var
      Infile          : text;           {input file .PDI format !}
      Outfile         : text;           {output file (text)}
      SmS, SXY        : sumsquarestype;
      TipAr           : SpValType;      {holds actual trait values}
      AbsDev          : SpValType;      {holds relative deviations from
                                         median or mean for Levene's test}
      df              : dftype;         {record of degrees of freedom}
      SetNo           : integer;
      NoSpp           : integer;
      NoSets          : integer;
      Sp              : integer;
      Trait           : integer;
      Ecode           : Integer;
      Name            : NameType;       {directory of two-character
                                         species names}
      Residuals       : TraitArType;    {holds regression-corrected values}

      Mean, Median    : SetAttributesType; {values for each set}

      SpSet           : SpSetArType;   {array of sets of taxa}
      T               : TraitNameType; {names for traits 1 and 2}
      setNames        : setNameType;   {directory of 7 character set names}
      Answer          : Char;          {used to read specification of covariate}
      CX              : byte;          {independent variable in ANCOVA}
      CY              : byte;          {dependent variable in ANCOVA}
      InfileName      : String;        {name of input file}
      OutFileName     : String;        {name of output file}
      MeanZero        : boolean;       {true when set's mean is zero}
      MedianZero      : boolean;       {true when set's median is zero}
      badResiduals    : boolean;       {true if one or more species has a bad
                                        residual value}

   {************************************************************************}
   {This procedure writes coprocessor extended reals to a file without the
    extra exponent and mantessa digits.}
   {************************************************************************}
   Procedure extendWrite(Var outFile :text;num :extended);
      Var outStr :string[19];
      Begin
         str(num:19:-11,outStr);
         delete(outStr,16,2);
         write(outFile,outStr)
      End; {procedure extendwrite}

   {************************************************************************}
   {This procedure writes a line of asterisks.                              }
   {************************************************************************}
   Procedure asterisks(Var outFile : text);
      Begin
         writeln(Outfile, '***************************************' +
                    '***************************************');
      End; {procedure asterisks}

   {************************************************************************}
   {This function returns the mean for one subset for one trait.            }
   {************************************************************************}
   Function GetMean (OneSet : SpSetType;
                     Nospp : Integer;
                     Value : TraitArType) : extended;
      Var Total : extended;
          Sp, N : Integer;
      Begin
         N := 0;                             {counts species in one set}
         Total := 0;                         {accumulates values in one set}
         For Sp := 1 to NoSpp Do             {loop through all species}
            If Sp In OneSet Then Begin       {if species[i] is in the set}
               inc(N);
               Total := Total + Value[Sp];
            End; {if sp in spset}
         GetMean := Total/N                 {calculate mean for the set}
      End; {func getmean}

   {************************************************************************}
   {This function calculates the sum of squares within one subset}
   {************************************************************************}
   Function GetSS (oneSet : SpSetType;
                   noSpp : integer;
                   Value : TraitArType) : extended;
      Var Sp : Integer;
          mean, SS : extended;

      Begin 
         mean := GetMean(oneSet, Nospp, Value);
         ss := 0;
         For Sp := 1 to NoSpp Do    {calculate squared deviations}
            If Sp In OneSet
            Then SS := SS + (Sqr(mean - Value[Sp]));
         getSS := ss;
      End; {func getsumsquares}

   {************************************************************************}
   {This function calculates the sum of cross products for two traits for   }
   {species in one subset                                                   }
   {************************************************************************}
   Function GetSCP (OneSet : SpSetType;
                    Nospp : integer;
                    Value1, Value2 : TraitArType) : extended;
      Var Sp : Integer;
          Mean1, Mean2, SCP : extended;

      Begin
         Mean1 := GetMean(Oneset, Nospp, Value1);
         Mean2 := GetMean(Oneset, Nospp, Value2);
         SCP := 0;
         For Sp := 1 to NoSpp Do    {calculate cross products for species in set}
            If Sp In OneSet
            Then SCP := SCP + (Mean1 - Value1[Sp]) * (Mean2 - Value2[Sp]);
         GetSCP := SCP;
      End; {func getsumsquares}

   {************************************************************************}
   {This procedure totals the various sums of squares for the analysis of
    variance, which is printed in the following procedure.                  }
   {************************************************************************}
   Procedure DoAnova(SpSet : SpSetArType;
                     Nosets, Nospp : Integer;
                     Value : TraitArType;
                     Var SS : SumSquaresType);
      Var setno, sp, n : integer;
          grandmean : extended;
      Begin
         With ss Do Begin
            w := 0;       {initialize all the accumulators}
            a := 0;
            t := 0;
            grandmean := getmean(spset[0], nospp, value); {set 0 is everyone}
            For setno := 1 To nosets Do Begin
               n := 0;
               For sp := 1 to nospp Do
                  If sp in spset[setno]
                  Then inc(n);
               w := w + getss(spset[setno], nospp, value);
               a := a + sqr(grandmean - getmean(spset[setno], nospp, value)) * n;
            End; {for setno}
            t := getss(spset[0], nospp, value);
         End; {with ss}
      End; {procedure doanova}

   {************************************************************************}
   {This procedure writes the ANOVA table to the outFile.                   }
   {************************************************************************}
   Procedure PrintAnova (SS : SumSquaresType;
                         Df : Dftype;
                         T : TraitNameType);
      Begin
         Writeln(Outfile,T[Trait]);
         Writeln(Outfile,
            'Source of variation  df          SS               MS            F-ratio');
         Writeln(Outfile,
            '-------------------------------------------------------------------------------');
         Write(Outfile, '   Among groups  ', Df.a:6, '  ');
         extendWrite(Outfile,SmS.a);
         write(outfile,'  ');
         extendWrite(outFile,SmS.a/df.a);
         write(Outfile,' ');
         extendWrite(outfile,(SmS.a/df.a)/(SmS.w/DF.w));
         writeln(outFile);
         write(Outfile, '   Within groups ', DF.w:6, '  ');
         extendWrite(outFile,SmS.w);
         write(outFile,'  ');
         extendWrite(outFile,SmS.w/DF.w);
         writeln(outFile);
         Write(Outfile, '   Total         ', DF.t:6, '  ');
         extendWrite(outFile,SmS.t);
         write(outFile,'  ');
         extendWrite(outFile,SmS.t/DF.t);
         writeln(outFile);
         Writeln(Outfile);
      End; {proc printanova}

   {************************************************************************}
   {This procedure calculates sums of cross product terms.                  }
   {************************************************************************}
   Procedure GetSXY(spSet : SpSetArType;
                    noSets, noSpp : Integer;
                    TipArray : SpValType;
                    Var SXY : SumSquaresType);
      Var setNo, sp, n : integer;
          grandMeanx, grandMeany : extended;

      Begin
         With sxy Do Begin
            w := 0;
            a := 0;
            t := 0;
            grandmeanx := getmean(spset[0], nospp, tipar[1]);
            grandmeany := getmean(spset[0], nospp, tipar[2]);
            For setno := 1 To nosets Do Begin
               n := 0;
               For sp := 1 To nospp Do
                  If sp In spset[setno] Then inc(n);
               w := w + getscp(spset[setno], noSpp, tipAr[1], tipAr[2]);
               a := a + (grandMeanx - getMean(spSet[setno], noSpp, tipAr[1])) *
                      (grandMeany - getMean(spSet[setNo], noSpp, tipAr[2])) * n;
            End; {for setno}
            t := getscp(spset[0], nospp, tipAr[1], tipAr[2]);
         End; {with ss}
      End; {proc getsxy}

   {************************************************************************}
   {This procedure calculates the analysis of covariance, including the
    test for equality of regression lines.                                  }
   {************************************************************************}
   Procedure DoAncova (Spset : SpsetArType;
                       NoSets : Integer;
                       Tipar : SpValType;
                       Cy, Cx : Byte;
                       T : TraitNameType);
      Var SS : Array[1..2] of SumSquaresType;
          SXY : SumSquaresType;
          Devyx : Array[1..MaxNoSets] of extended;
          DEVyxw, DEVyxa, DEVyxt, MeanDevyx : extended;
          groupSS : extended;
          groupSS0 : boolean;

      Begin
         DoAnova(SpSet, NoSets, NoSpp, TipAr[Cx], SS[cx]);
         DoAnova(SpSet, NoSets, NoSpp, TipAr[Cy], SS[cy]);
         GetSXY(SpSet, NoSets, NoSpp, TipAr, SXY);

         DEVyxt := SS[CY].t - sqr(SXY.t)/SS[CX].t;
         {sum of squares of deviations from regression lines (= error)}

         DEVyxw := SS[CY].w - sqr(SXY.w)/SS[CX].w;
         DEVyxa := DEVyxt - DEVyxw;
         MeanDevyx := 0;
         groupSS0 := false;
         For SetNo := 1 to NoSets Do Begin
            groupSS := GetSS(SpSet[SetNo], NoSpp, TipAr[CX]);
            If groupSS = 0 Then groupSS0 := true
            Else Begin
               DEVyx[SetNo] := GetSS(SpSet[SetNo], NoSpp, TipAr[Cy])
                                -sqr(GetSCP(SpSet[SetNo],NoSpp,TipAr[Cy],TipAr[Cx]))/groupSS;
               MeanDevyx := MeanDevyx + DEVyx[SetNo];
            End {else}
         End; {for setno, calculating slopes and mean squared deviations from
               regression for each group}
         Writeln(Outfile);
         asterisks(outfile);
         Writeln(Outfile, 'ANCOVA to test for differences in ',T[Cy],' after ');
         Writeln(outfile, '  correcting for ',t[cx],' as covariate:');
         Writeln(Outfile);
         Writeln(Outfile,'Pooled within-group slope of ',T[CY],' on ',T[CX],' = ');
         extendWrite(OutFile,SXY.w/SS[CX].w);
         writeln(outFile);
         Writeln(Outfile);
         Writeln(Outfile,
                   'Source of variation  df          SS               MS             F-ratio');
         Writeln(Outfile,
'-------------------------------------------------------------------------------');
         Write(Outfile, 'Covariate (',T[CX],')','':7-Length(T[CX]),'1':4,'  ');
         extendWrite(outFile,SS[CY].w-Devyxw);
         Write(Outfile,'  ');
         extendWrite(outFile,(SS[CY].w-DEVyxw));
         write(outFile,' ');
         extendWrite(outFile,(SS[CY].w-DEVyxw)/(DEVyxw/(DF.w-1)));
         writeln(outFile);
         Write(Outfile, 'Among groups     ', NoSets-1:6,'  ');
         extendWrite(Outfile, DEVyxa);
         write(outFile,'  ');
         extendWrite(outFile,DEVyxa/(NoSets-1));
         write(outFile,' ');
         extendWrite(Outfile, DEVyxa/(NoSets-1) / (DEVyxw/(DF.w-1)));
         writeln(outFile);
         Write(Outfile,   'Explained        ',NoSets:6,'  ');
         extendWrite(outFile,SS[CY].t-Devyxw);
         write(outFile,'  ');
         extendWrite(outFile,(SS[CY].t-DEVyxw)/NoSets);
         Write(Outfile,' ');
         extendWrite(outFile,((SS[CY].t-DEVyxw)/NoSets)/(DEVyxw/(DF.w-1)));
         writeln(outFile);
         Write(Outfile, 'Error            ',DF.w-1:6,'  ');
         extendWrite(outFile,DEVyxw);
         write(outFile,'  ');
         extendWrite(outFile,DEVyxw/(DF.w-1));
         writeln(outFile);
         Write(Outfile, 'Total            ',DF.w+NoSets-1:6,'  ');
         extendWrite(outFile,SS[CY].t);
         write(outFile,'  ');
         extendWrite(outFile,SS[CY].t/(DF.w+NoSets-1));
         writeln(outFile);
         Writeln(outfile);
         If groupSS0 Then Begin
            writeln(outfile);
            writeln(outFile,'The variance in the independent variable for at least one subset was 0,');
            writeln(outFile,'therefore slopes were not computed and -9 will appear in the following table.');
         End; {then}
         asterisks(outfile);
         Writeln(Outfile, 'Test of identity of slopes among groups:');
         Writeln(Outfile, 'SSamong = Error from pooled slope minus error from',
                   ' individual slopes.');
         Writeln(Outfile);
         Writeln(Outfile,
                   'Source of variation  df          SS               MS             F-ratio');
         Writeln(Outfile,
'-------------------------------------------------------------------------------');
         If groupSS0 Then Begin
            writeln;
            asterisks(output);
            writeln;
            writeln('Warning - Because the slope of one or more groups was undefined, the');
            writeln('          test of identity of slopes could not be performed.  The value');
            writeln('          -9 will be written to the file.');
            writeln;
            asterisks(output);
            writeln;
            Write(Outfile,'Among slopes     ',NoSets-1:6,'    ');
            write(outFile,'      -9               -9                -9');
            writeln(outFile);
            Write(Outfile, 'Error            ',df.w-NoSets:6,'    ');
            write(outFile,'      -9               -9');
         End
         Else Begin
            Write(Outfile,'Among slopes     ',NoSets-1:6,'  ');
            extendWrite(outFile,DEVyxw-MeanDevyx);
            write(outFile,'  ');
            extendWrite(Outfile, (DEVyxw-MeanDevyx)/(NoSets-1));
            write(outFile,' ');
            extendWrite(Outfile,((DEVyxw-MeanDevyx)/(NoSets-1))/(MeanDevyx/(df.w-NoSets)));
            writeln(outFile);
            Write(Outfile, 'Error            ',df.w-NoSets:6,'  ');
            extendWrite(outFile,MeanDevyx);
            write(Outfile,'  ');
            extendWrite(outFile,MeanDevyx/(df.w-NoSets));
         End; {else}
         Writeln(Outfile);
         Writeln(Outfile);
      End; {proc doancova}

   {************************************************************************}
   {This procedure counts up the within, among, and total degrees of        }
   {freedom.                                                                }
   {************************************************************************}
   Procedure GetDF (SpSet : SpSetArType;
                    Nosets, Nospp : Integer;
                    Var Df : DfType);
      Var SetNo, Sp : Integer;
      Begin
         With Df Do Begin
            w := 0; {initialize all the counts}
            a := 0;
            t := 0;
            For SetNo := 1 To NoSets Do Begin
               inc(a); {df among groups}
               For Sp := 1 to NoSpp Do
                 If Sp In SpSet[SetNo]
                 Then Begin
                    inc(w);
                    inc(t);
                 End; {then}
               dec(w); {adjust to species count - 1}
            End;
            dec(t); {adjust to total species (in sets) - 1}
            dec(a); {adjust to total sets - 1}
         End; {with df}
      End;{proc getdf}

   {***********************************************************************}
   {This function returns true if the file exists and might contain a PDI  }
   {format tree, false otherwise.                                          }
   {***********************************************************************}
   Function checkFile(fname : string) : boolean;
      Var testFile : text;
          testSize : integer;
      Begin
         {$I-}   {turn off i/o checking}
         assign(testFile,fname);
         reset(testFile);
         If (IOResult = 0) Then Begin
            testSize := 0;
            readln(testFile,testSize);
            If (testSize = 0) Then Begin
               writeln('The file ',fname, ' does not appear to be in .PDI format; try again.');
               checkFile := false;
            End
            Else checkFile := true;
            close(testFile);
         End
         Else Begin
            writeln('The file ',fname, ' does not exist; try again.');
            checkFile := false;
         End; {else}
         {$I+}
      End; {function checkFile}

   {***********************************************************************}
   {This procedure handles the filename dialog that follows the intro      }
   {screen.                                                                }
   {***********************************************************************}
   Procedure getFiles(Var inFile : text;
                      Var outFile : text;
                      Var inFileName : string;
                      Var outFileName : string);
      Begin
         Writeln('Please type input file name with path, but without the .PDI extension.');
         Readln(InfileName);
         If Pos('.',InfileName) > 0
         Then InfileName := Copy(InfileName, 1, Pos('.',InfileName)-1);
         inFileName := inFileName+'.PDI';
         While not(checkFile(inFileName)) Do Begin
            writeln;
            write('Please enter a new file name: ');
            Readln(InfileName);
            If Pos('.',InfileName) > 0
            Then InfileName := Copy(InfileName, 1, Pos('.',InfileName)-1);
            inFileName := inFileName+'.PDI';
         End;
         Assign(Infile, infileName);
         Writeln('Input file = ',infileName);
         outFilename := copy(inFileName,1,pos('.',infilename)-1)+'.SGL';
         Writeln('Output file = ',outFileName);
         writeln;
         writeln;   { purely cosmetic }
      End; {procedure getFiles}

   {************************************************************************}
   {This procedure reads the tree from the file.  Since the file is a .PDI  }
   {file, there is only one tree.                                           }
   {************************************************************************}
   Procedure GetTree (Var Name : NameType; Var TipAr : SpValType);
      Var  Sp : Integer;
           ch : char;
      Begin {proc gettree}
         Reset(Infile);
         Readln(Infile, NoSpp);
         Readln(Infile);
         Writeln('NoSpp = ',NoSpp);
         For Sp := 1 to 2*NoSpp - 1 Do Readln(Infile); {ignore tree structure}
         Readln(Infile);
         For Sp := 1 to NoSpp Do Begin
           read(inFile,name[Sp]);
           Repeat                  {skip extra characters in name}
              read(inFile,ch)
           Until (ch = ' ');
           Readln(Infile, TipAr[1,Sp], TipAr[2,Sp]);
         End; {for}
         Close(Infile);
      End; {proc gettree}

   {************************************************************************}
   { This procedure just lists the species codes.  It should be called      }
   { before the user is prompted for each set of species.                   }
   {************************************************************************}
   Procedure ListSpecies;
      Var sp,j : integer;
      Begin
         For j := 1 To 20 Do writeln;
         For j := 1 To 10 Do Write(j:4);
         Writeln;
         For sp := 1 To NoSpp Do
           If sp Mod 10 = 0 Then Begin
              writeln('  ', Name[Sp]);
              writeln;
              For j := sp + 1 To sp + 10 Do
                If J <= NoSpp Then write(j:4);
              writeln;
           End {then}
           Else write('  ',Name[Sp]);
         Writeln;
      End; {procedure listspecies}

   {************************************************************************}
   { This function looks up the two character name provided by the user and }
   { returns the appropriate species index.  If the name is not found, a    }
   { value greater than the number of species is returned.                                                           }
   {************************************************************************}
   Function nameLookup (nam : String) : integer;
      Var test : integer;
      Begin
         test := 1;
         While ((Name[test] <> nam) And (test <= noSpp)) Do
           inc(test);
         nameLookup := test;
      End; { nameLookup}

   {************************************************************************}
   { This procedure prompts user through the process of defining sets of    }
   { species which will be compared.                                        }
   {************************************************************************}
   Procedure GetSet (Var SpSet : SpSetArType);
      Var SetNo, Sp, J, K : Integer;
          SN : Array[0..MaxNoSets] of integer;
          spName : String[2];
          userChar : char;
          c1 : char;
          readNames : boolean;
      Begin
         setNames[0] := 'Overall';
         SN[0] := 0;
         SpSet[0] := [];
         Writeln('How many sets will be defined? ');
         Readln(NoSets);
         write('Do you want to specify species by two character code or number? (C/N=Enter): ');
         userChar := ' ';
         Repeat
            readln(userChar);
            If (ord(userChar) = 13) Then userChar := 'N';
            userChar := upcase(userChar);
            If Not(userChar in ['C','N'])
            Then write('Please enter either C or N:');
         Until (userChar in ['C','N']);
         readNames := (userChar = 'C');
         For j := 1 To 24 Do Writeln;
         For SetNo := 1 to NoSets Do Begin
            writeln('Enter a seven letter name for set ',setno,' or ENTER for none.');
            readln(setNames[setno]);
            writeln;
            Repeat
               SpSet[SetNo] := [];
               ListSpecies;
               If readNames Then Begin
                  writeln('Enter species for set ',setno,' ("', setNames[SetNo],'") by above coding.');
                  writeln('Separate two character names with spaces.  Enter a <return>')
               End {then}
               Else Begin
                  writeln('Enter species for set ',setno,' ("', setNames[SetNo],'") by above numbering.');
                  writeln('Separate numbers with spaces.  Enter a <return>');
               End; {else}
               Writeln('BEFORE you reach the edge of the screen.  After you have');
               Writeln('typed in all the species for the set, type 0 <return> to end.');
               If readNames Then Begin
                  While (spSet[setNo] = []) Do Begin
                     Repeat
                        Repeat
                           read(c1)
                        Until (ord(c1) > 32);
                        spName := '' + c1;
                        read(c1);
                        If (ord(c1) > 32) Then Begin
                           spName := spName + c1;
                           Repeat
                              read(c1)
                           Until (ord(c1) <= 32)
                        End; {then}
                        j := nameLookup(spName);
                        If j > 0
                        Then spSet[SetNo] := spSet[setNo] + [j];
                     Until (spName = '0');
                     If spSet[setNo] = [0] Then Begin
                        writeln('Warning: All sets must be nonempty; please reenter!');
                        writeln;
                        spSet[setNo] := []
                     End {then}
                  End {while}
               End {then}
               Else Begin
                  J := 300;
                  While (spSet[setNo] = []) Do Begin
                     Repeat
                        Read(j);
                        If J <= NoSpp Then
                        SpSet[SetNo] := SpSet[SetNo] + [J];
                     Until J = 0;
                     If spSet[setNo] = [0] Then Begin
                        writeln('Warning: All sets must be nonempty; please reenter!');
                        writeln;
                        spSet[setNo] := []
                     End {then}
                  End; {while}
               End; {else}
               If (SpSet[SetNo] * SpSet[0] <> []) Then Begin
                  writeln;
                  asterisks(output);
                  writeln;
                  Writeln('There is overlap between this set and the previously',
                            ' defined sets involving');
                  Write('species ');
                  If readNames Then
                    For k := 1 to NoSpp Do
                      If k in (SpSet[SetNo] * SpSet[0]) Then write(name[k])
                  Else
                    For k := 1 to NoSpp Do
                      If k in (SpSet[SetNo] * SpSet[0]) Then write(k:4);
                  Writeln;
                  writeln('You will need to re-enter your species list for this set.');
                  writeln;
                  asterisks(output);
                  writeln;
                  writeln('Press enter to proceed');
                  readln;  { flush remainder of previous entry line }
                  readln;
               End; {if spset[SetNo] * spset[0] <> []}
            Until SpSet[SetNo] * SpSet[0] = [];
            Write('Subset ',setNo,' ("',setNames[SetNo],'") = ');
            SN[SetNo] := 0;
            For J := 1 to NoSpp Do
            If J In SpSet[SetNo] Then Begin
               Write(Name[J],' ');
               SN[SetNo] := SN[SetNo] + 1;
               SpSet[0] := SpSet[0] + [J];
            End; {if j in spset}
            Writeln;
            Writeln('N = ',SN[SetNo]);
            Writeln;
            SN[0] := SN[0] + SN[SetNo];
            If Not(readNames)
            Then readln   { flush remainder of line }
            Else Begin    { flush a different way }
               read(c1);
               While (ord(c1) <> 10) Do read(c1)
            End {else}
         End; {for setno}
         For SetNo := 0 to NoSets Do Begin
            If setNo=0 Then Begin
               Write('Subset 0 (Overall) = ');
               Write(Outfile,'Subset 0 (Overall) = ');
            End
            Else Begin
               Write('Subset ',setNo,' ("',setNames[SetNo],'") = ');
               Write(Outfile,'Subset ',setNo,' ("',setNames[SetNo],'") = ');
            End;
            For J := 1 to NoSpp Do
              If J In SpSet[SetNo] Then Begin
                 Write(Name[J],' ');
                 Write(Outfile, Name[J],' ');
              End; {if j in spset}
            Writeln(Outfile);
            Writeln(Outfile);
            Writeln;
            Writeln;
         End; {for SetNo}
      End; {proc getset}

   {************************************************************************}
   {This function returns the median for one subset.}
   {************************************************************************}
   Function GetMedian (OneSet : SpSetType;
                       Nospp : Integer;
                       Value : TraitArType) : extended;

      Var Sp, K : Byte;
          MedianMinus, MedianPlus : extended;
          Ties, NoHigher, NoLower  : Integer;

      Begin
         For Sp := 1 to NoSpp Do
           If Sp In OneSet Then Begin  {count members of set that are lower}
              NoHigher := 0;
              NoLower := 0;
              Ties := 0;
              For K := 1 to NoSpp Do
                If K In OneSet Then Begin
                   If Value[K] > Value[Sp] Then inc(NoHigher);
                   If Value[K] < Value[Sp] Then inc(NoLower);
                   If Value[K] = Value[Sp] Then inc(Ties);
                End; {if k in oneset}
              If Abs(NoHigher - NoLower) < Ties Then Begin
                 MedianMinus := Value[Sp];
                 MedianPlus := Value[Sp];
              End
              Else If Abs(NoHigher - NoLower) = Ties
                   Then If NoLower > NoHigher
                        Then MedianMinus := Value[Sp]
                        Else MedianPlus := Value[Sp];
           End; {if Sp in oneset}
           GetMedian := (MedianPlus + MedianMinus)/2;
      End; {getmedian}

   {************************************************************************}
   {This procedure calculates (into AbsDev), for one trait, the absolute
    value of the relative deviation from their within-subgroup central value
    (median or mean)  =   |(x-cent)/cent|     }
   {************************************************************************}
   Procedure GetAbsDev (SpSet : SpSetArType;
                        NoSets, NoSpp : Integer;
                        Cent : setattributestype;
                        Value : TraitArType;
                        Var AbsDev : TraitArType);
      Var Sp, SetNo, Trait : integer;
      Begin
         For Sp := 1 to NoSpp Do
           For SetNo := 1 to NoSets Do
             If Sp In SpSet[SetNo]
             Then AbsDev[Sp] := Abs((Value[Sp] - Cent[SetNo])/Cent[SetNo]);
      End; {proc getabsdev}

   {************************************************************************}
   {This procedure finds the lowest and highest values for each trait in
    a subset of species.}
   {************************************************************************}
   Procedure GetLoHi (OneSet: SpSetType;
                      Nospp : integer;
                      Value : TraitArType;
                      Var Lo, Hi : extended);
      Var Sp : Integer;
      Begin
         For Sp := 1 to NoSpp Do
         If Sp In OneSet Then Begin
            Hi := Value[Sp];
            Lo := Value[Sp];
         End; {initializing arrays to lie within range of species in oneset}
         For Sp := 1 to NoSpp Do {calculate means and find lo and hi values}
         If Sp In OneSet Then Begin
            If Value[Sp] < Lo
            Then Lo := Value[Sp];
            If Value[Sp] > Hi
            Then Hi := Value[Sp];
         End; {if sp in oneset}
      End; {proc lowhi}

   {************************************************************************}
   {This procedure transforms the independent variable to a residual from a
    regression of the covariate within the species' subset.}
   {************************************************************************}
   Procedure GetResiduals(SpSet : SpSetArType;
                          NoSets, NoSpp : Integer;
                          Dependent, Covariate : TraitArType;
                          Var NeuValue : TraitArType;
                          Var badResiduals : boolean);


       Var MeanX, MeanY, Slope : extended;
           SetNo, Sp : Integer;
           groupSS : extended;
       Begin
          badResiduals := false;
          For SetNo := 1 to Nosets Do Begin
             MeanX := GetMean(SpSet[SetNo], NoSpp, Covariate);
             MeanY := GetMean(SpSet[SetNo], NoSpp, Dependent);
             groupSS := GetSS(SpSet[SetNo], NoSpp, Covariate);
             If groupSS <> 0  {check for a bad slope}
             Then Begin  {slope is OK, calculate residual}
                Slope := GetSCP(SpSet[SetNo], NoSpp, Dependent, Covariate) /
                          groupSS;
                Write(Outfile, '  Set ',setNo,' ("',setNames[SetNo],'") ',
                              '':7-length(setNames[setNo]),'slope = ');
                extendWrite(outFile,slope);
                writeln(outFile);
                For Sp := 1 to NoSpp Do
                  If Sp In SpSet[SetNo]
                  Then NeuValue[Sp] :=
                         Dependent[Sp]-(MeanY+((Covariate[Sp]-MeanX)*Slope));
             End {then}
             Else Begin  {slope is broken, so the residual is too}
                badResiduals := true;
                writeln(Outfile, '  Set ',setNames[SetNo],' slope =               -9');
                For Sp := 1 to NoSpp Do
                  If Sp In SpSet[SetNo]
                  Then NeuValue[Sp] := -9;
             End;
          End; {for setno}
       End; {proc getresiduals}

   {************************************************************************}
   {This procedure just writes out the current date and time to the outfile.}
   {************************************************************************}
   Procedure DateTime;
      Var Year, Month, Day, DayofWeek, Hour, Minute, Second, Sec100 : word;
      Begin
         Write(Outfile, 'Output from PDSINGLE run on ');
         GetDate(Year, Month, Day, DayofWeek);
         Write(Outfile, Day,' ');
         Case Month of
             1 : Write(Outfile, 'Jan ');
             2 : Write(Outfile, 'Feb ');
             3 : Write(Outfile, 'Mar ');
             4 : Write(Outfile, 'Apr ');
             5 : Write(Outfile, 'May ');
             6 : Write(Outfile, 'Jun ');
             7 : Write(Outfile, 'Jul ');
             8 : Write(Outfile, 'Aug ');
             9 : Write(Outfile, 'Sep ');
            10 : Write(Outfile, 'Oct ');
            11 : Write(Outfile, 'Nov ');
            12 : Write(Outfile, 'Dec ');
         End; {case month of}
         Write(Outfile, Year,' at ');
         GetTime(Hour, Minute, Second, Sec100);
         Write(Outfile, Hour, ':');
         If minute < 10 Then write(outFile,'0');
         write(outFile, Minute,':');
         If second < 10 Then write(outFile,'0');
         Writeln(outFile,Second,'  ');
         Writeln(Outfile, 'Input data file for this run was ', InfileName,
                        ', which contains ', noSpp, ' tips.');
         Writeln(Outfile);
      End; {proc datetime}

   {************************************************************************}
   {This procedure displays the intro screen.                               }
   {************************************************************************}
   Procedure DoIntro;
      Var I : Byte;
      Begin
         For I := 1 to 5 Do Writeln;
         Writeln('       ********* Phenotypic Diversity Analysis Program *********');
         Writeln;
         Writeln('                            PDSINGLE Module');
         Writeln;
         Writeln('V. 2.0 Copyright 12 Jan 2001 by Allan W. Dickerman and Theodore Garland, Jr.');
         Writeln;
         Writeln('     Contact T.G. at the Department of Zoology, 430 Lincoln Drive,');
         Writeln('                University of Wisconsin, Madison, WI 53706');
         Writeln('            Telephone:  608-262-4437; Facsimile:  608-265-6320');
         Writeln('                   E-Mail: tgarland@facstaff.wisc.edu');
         Writeln;
         Writeln('This program performs analysis of variance and covariance on a single data');
         Writeln('set for two characters from a file in the *.PDI format.');
         writeln('  Use the PDTREE program to create data files or to convert files from');
         writeln('the *.INP format of the Comparative Method Analysis Program of ');
         writeln('Martins and Garland (1991, Evolution 45:534-557).');
         Writeln;
         Writeln('Procedures are described in the following paper:');
         Writeln;
         Writeln('Garland, T., Jr., A. W. Dickerman, C. M. Janis, and J. A. Jones.');
         Writeln('     1993. Phylogenetic analysis of covariance by computer');
         Writeln('     simulation. Systematic Biology 42:265-292.');
         Writeln;
      End; {proc dointro}

   {************************************************************************}
   {This procedure calculates the descriptive statistics (mean, median,
    min, max, variance, regression, etc.                                    }
   {************************************************************************}
   Procedure DescriptiveStatistics (SpSet : SpSetArType;
                                    NoSets, NoSpp : Integer;
                                    TipAr : SpValType;
                                    T : TraitNameType);
      Var N, Sp, SetNo : Integer;
          SSx, SSy, SCP, MeanX, MeanY, Lo1, Lo2, Hi1, Hi2,
          MedianX, MedianY : extended;


      Begin
         Writeln(Outfile, 'Summary statistics:');
         writeln(outFile);
         For SetNo := 0 to NoSets Do Begin
            N := 0;
            For Sp := 1 to Nospp Do
            If Sp In SpSet[SetNo] Then Inc(N);
            MeanX := GetMean(SpSet[SetNo], NoSpp, TipAr[1]);
            MeanY := GetMean(SpSet[SetNo], NoSpp, TipAr[2]);
            GetLoHi(SpSet[SetNo], NoSpp, TipAr[1], Lo1, Hi1);
            GetLoHi(SpSet[SetNo], NoSpp, TipAr[2], Lo2, Hi2);
            MedianX := GetMedian(SpSet[SetNo], NoSpp, TipAr[1]);
            MedianY := GetMedian(SpSet[SetNo], NoSpp, TipAr[2]);
            SSx := GetSS(SpSet[SetNo], NoSpp, TipAr[1]);
            SSy := GetSS(SpSet[SetNo], NoSpp, TipAr[2]);
            SCP := GetSCP(SpSet[SetNo], NoSpp, TipAr[1], TipAr[2]);
            If SetNo = 0
            Then Writeln(Outfile, 'Overall (union of all subsets) N =', N:6)
            Else Writeln(Outfile, 'Subset ', setNo,' = "',setNames[SetNo], '"',
                                   '':7-Length(setNames[setNo]),'     N =', N:6);
            Writeln(Outfile,'              ',T[1],'':7-Length(T[1]),'         ',T[2]);
            Writeln(Outfile,'           ----------------------------------');
            Write(Outfile, '   Means :');
            extendWrite(Outfile, MeanX);
            write(outFile,' ');
            extendWrite(outFile,MeanY);
            writeln(outFile);
            Write(Outfile, ' Medians :');
            extendWrite(Outfile,MedianX);
            write(outFile,' ');
            extendWrite(outFile,MedianY);
            writeln(outFile);
            Write(Outfile, '  Lowest :');
            extendWrite(outFile,Lo1);
            write(outFile,' ');
            extendWrite(outFile,Lo2);
            writeln(outFile);
            Write(outFile, ' Highest :');
            extendWrite(outFile,Hi1);
            write(outFile,' ');
            extendWrite(outFile,Hi2);
            writeln(outFile);
            If (n <= 1) Then Begin
               writeln;
               asterisks(output);
               writeln;
               writeln('Warning - Species set ',setNames[setNo],' has only one member, so a variance can not be');
               writeln('          be calculated.  The value -9 will be written to the file.');
               writeln;
               asterisks(output);
               writeln;
               write(outFile,'Variance :               -9                -9');
               writeln(outFile)
            End
            Else Begin
               Write(Outfile,   'Variance :');
               extendWrite(outFile,SSx/(n-1));
               write(outFile,' ');
               extendWrite(outFile,SSy/(n-1));
               writeln(outFile);
            End; {else}
            Writeln(outfile);
            If (SSy = 0) Then Begin
               writeln;
               asterisks(output);
               writeln;
               writeln('Warning - Set ',setNames[setNo],' has no variance in ',t[2],', so no regression can be');
               writeln('          calculated.  The value -9 will be written for the slope and');
               writeln('          intercept.');
               writeln;
               asterisks(output);
               writeln;
               Writeln(outfile, '  Regression of ',t[1],' on ',t[2],':');
               Write(outfile, '    slope = -9              ');
               Write(outfile, '  intercept = -9');
               writeln(outFile);
            End
            Else Begin
               Writeln(outfile, '  Regression of ',t[1],' on ',t[2],':');
               Write(outfile, '    slope = ');
               extendWrite(outFile, SCP/SSy);
               Write(outfile, '  intercept = ');
               extendWrite(outFile, MeanX - MeanY * SCP/SSy);
               writeln(outFile);
            End; {else}
            Writeln(outfile);
            If (SSx = 0) Then Begin
               writeln;
               asterisks(output);
               writeln;
               writeln('Warning - Set ',setNames[setNo],' has no variance in ',t[1],', so no regression can be');
               writeln('          calculated.  The value -9 will be written for the slope and');
               writeln('          intercept.');
               writeln;
               asterisks(output);
               writeln;
               Writeln(outfile, '  Regression of ',t[2],' on ',t[1],':');
               Write(outfile, '    slope = -9              ');
               Write(outfile, '  intercept = -9');
               writeln(outFile);
            End
            Else Begin
               Writeln(outfile, '  Regression of ',t[2],' on ',t[1],':');
               Write(outfile, '    slope = ');
               extendWrite(outFile, SCP/SSx);
               Write(outfile, '  intercept = ');
               extendWrite(outFile, MeanY - MeanX * SCP/SSx);
               writeln(outFile);
            End; {else}
            Writeln(outfile);
            If ((SSx <> 0) And (SSy <> 0)) Then Begin 
               Write(Outfile, '  Correlation = ');
               extendWrite(outFile,SCP/sqrt(SSx*SSY));
               writeln(outFile)
            End
            Else Begin
               writeln;
               asterisks(output);
               writeln;
               writeln('Warning - Correlation for set ',setNames[setNo],' can not be calculated.  The value -9');
               writeln('          will be written to the file.');
               writeln;
               asterisks(output);
               writeln;
               writeln(Outfile, '  Correlation = -9');
            End;
            Writeln(Outfile);
            Writeln(Outfile);
         End; {for setno := 1 to nosets}
      End; {proc descriptivestatistics}

   {***********************************************************************}
   Begin { main }
      DoIntro;
      getFiles(inFile,outFile,inFileName,outFileName);
      GetTree(Name, TipAr);
      Assign(Outfile, outFileName);
      Rewrite(Outfile);
      DateTime;
      GetSet(SpSet);
      Writeln('Type seven letter abbreviation for trait 1, or ENTER for none.');
      Readln(T[1]);
      If T[1] = '' Then T[1] := 'Trait 1';
      Writeln('Type seven letter abbreviation for trait 2, or ENTER for none.');
      Readln(T[2]);
      If T[2] = '' Then T[2] := 'Trait 2';
      Answer  := ' ';
      While Not (Answer In ['1', '2']) Do Begin
         writeln('Which variable shall be the covariate? (1 or 2)');
         If T[1] <> 'Trait 1' Then writeln('Trait 1 = ',T[1]);
         If T[2] <> 'Trait 2' Then writeln('Trait 2 = ',T[2]);
         Readln(Answer);
      End; {while not cx in [...]}
      Val(Answer, CX, Ecode);
      CY := CX Mod 2 + 1;
      Writeln(Outfile);
      write(outFile,'Trait 1 = ');
      If T[1] = 'Trait 1'
      Then writeln(outFile,'[no name given]')
      Else writeln(outFile,'"',T[1],'"');
      write(outFile,'Trait 2 = ');
      If T[2] = 'Trait 2'
      Then writeln(outFile,'[no name given]')
      Else writeln(outFile,'"',T[2],'"');
      writeln(outFile);
      writeln(outFile);

      DescriptiveStatistics(SpSet, NoSets, NoSpp, TipAr, T);

      GetDf(SpSet, NoSets, NoSpp, Df);

      asterisks(outfile);
      Writeln(Outfile, 'ANOVA to test for differences among group means:   ');
      Writeln(Outfile);

      For Trait := 1 To 2 Do Begin
         DoAnova(SpSet, NoSets, NoSpp, TipAr[Trait], SmS);
         PrintAnova(SmS, Df, T);
      End; {for Trait, writing anova}

      DoAncova(SpSet, NoSets, TipAr, Cy, Cx, T);

      Writeln(Outfile);
      Writeln(Outfile);


      {Levene's test follows; this test requires dividing by either}
      {the median or the mean, so first the program must test for 0}
      {values}
      MeanZero := False;
      MedianZero := False;
      For SetNo := 0 To NoSets Do  {look for zero medians or means}
        For Trait := 1 To 2 Do Begin
           Median[SetNo] := GetMedian(SpSet[SetNo], NoSpp, TipAr[Trait]);
           If abs(Median[SetNo]) < 1E-10 Then MedianZero := True;
           Mean[SetNo] := GetMean(SpSet[SetNo], NoSpp, TipAr[Trait]);
           If abs(Mean[SetNo]) < 1E-10
           Then MeanZero := True;
        End; {testing for zero medians or means}

      If Not(MedianZero)
      Then For Trait := 1 to 2 Do
        GetAbsDev (SpSet, NoSets, NoSpp, Median, TipAr[Trait], AbsDev[Trait])

      Else If Not(MeanZero)
           Then Begin
              Writeln(Outfile,'The median for at least one subset is near zero.');
              Writeln(Outfile,'Consequently, Levene''s test will be performed using means.');
              For Trait := 1 to 2 Do
                GetAbsDev (SpSet, NoSets, NoSpp, Mean, TipAr[Trait], AbsDev[Trait])
           End
           Else Begin
              Writeln(Outfile, '***** Warning *****');
              Writeln(Outfile, 'The median for at least 1 subset is near zero, and');
              Writeln(Outfile, 'the mean for at least one subset is near zero.');
              Writeln(Outfile, 'Consequently, Levene''s test will not be performed.');
           End; {else, both median and mean had zero values}


      If Not(MedianZero And MeanZero) Then Begin  {do Levene's test on medians or means}
         asterisks(outfile);
         If Not(MedianZero)
         Then Writeln(Outfile,'Levene''s median-ratio test for differences in variability among subsets.')
         Else Writeln(Outfile,'Levene''s mean-ratio test for differences in variability among subsets.');
         Writeln(Outfile, '  ANOVA performed on transformed variables.');
         If Not(MedianZero)
         Then Writeln(Outfile, '  Transformation = |[Variable(i,j) - Median(i)]/Median(i)|')
         Else Writeln(Outfile, '  Transformation = |[Variable(i,j) - Mean(i)]/Mean(i)| ');
         Writeln(Outfile, '  where i indexes subsets and j species within subsets.');
         Writeln(Outfile, 'For description of Levene''s test see: ');
         Writeln(Outfile, '  Schultz, B. B. (1985) Systematic Zoology 34:449-456.');
         Writeln(Outfile);
         For Trait := 1 to 2 Do Begin
            DoAnova(SpSet, NoSets, NoSpp, AbsDev[Trait], SmS);
            PrintAnova(SmS, Df, T);
         End; {for trait}
      End; {doing regular Levene's test}


      asterisks(outfile);
      Writeln(Outfile,'Calculating residuals of ',T[CY],' by subtracting within-group regression');
      Writeln(Outfile, 'of ',T[CY],' on ',T[CX],':');
      GetResiduals(SpSet, NoSets, NoSpp, TipAr[Cy], TipAr[Cx], Residuals,badResiduals);
      Writeln(Outfile);
      Writeln(Outfile, 'Residuals of ',T[CY],' after within-group regressions on ',T[CX],':');
      For Sp := 1 to NoSpp Do
        If Sp In SpSet[0] Then Begin
           Write(Outfile, '  ',Name[Sp],'  ');
           If (Residuals[Sp] <> -9)
           Then extendWrite(outFile,Residuals[Sp])
           Else write(outFile,'               -9');
           writeln(outFile);
        End; {then}
      Writeln(Outfile);

      If badResiduals Then Begin
         writeln(outFile,'Warning - The preceding table contains uncomputable residuals indicated by');
         writeln(outFile,'          the value -9.');
         writeln(outFile);
         writeln(outFile,'Because residuals for one or more species were uncomputable, the ANOVA');
         writeln(outFile,'tests for the absolute values and squares of residuals were skipped.');
         writeln(outFile);
         writeln;
         asterisks(output);
         writeln;
         writeln('Warning - Because residuals for one or more species were uncomputable, the');
         writeln('          ANOVA tests for the absolute values and squares of residuals were');
         writeln('          skipped.');
         writeln;
         asterisks(output);
         writeln;
      End {then}
      Else Begin
         For Sp := 1 to NoSpp Do
         If Sp In SpSet[0] {one of the species being considered}
         Then Residuals[Sp] := Abs(Residuals[Sp]);

         Writeln(Outfile);
         asterisks(outfile);
         Writeln(Outfile, 'ANOVA of absolute value of residuals of ',t[cy]);
         Writeln(outfile, '  after regression on ',t[cx],' using separate regressions for each group.');

         DoAnova(SpSet, NoSets, NoSpp, Residuals, SmS);
         PrintAnova(SmS, Df, T);
         For Sp := 1 to NoSpp Do
           If Sp In SpSet[0] {one of the species being considered}
           Then Residuals[Sp] := Sqr(Residuals[Sp]);

         Writeln(Outfile);
         asterisks(outfile);
         Write(Outfile, 'ANOVA of squared residuals of ',t[cy]);
         writeln(outfile, ' after regression on ',t[cx]);
         Writeln(outfile, '  using separate regressions for each group.');

         DoAnova(SpSet, NoSets, NoSpp, Residuals, SmS);
         PrintAnova(SmS, Df, T);
         Writeln(Outfile,'--------------------------- end of output ---------------------------------');
      End; {else}
      Close(Outfile);
      writeln;
      Writeln('Output has been written to ', outFileName);
   End. {program pdsingle}
