
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Program PDSQCHP;
{This program will "reconstruct" the value of an ancestral trait using
squared-change parsimony.  It can handle polytomies and fix values for
particular nodes.}

{This program originally came from SQCHP2, which in turn was derived from
CMSINGLE.  Much unnecessary code was deleted, procedures changed and added,
type definitions changed, etc.  Many of the procedures that are left have
the same name as those in CMSINGLE but they are NOT necessarily equal:
many things have been changed!  Written by Ramon Diaz-Uriarte with
modifications by P. E. Midford and T. Garland, Jr.}

{The program was checked for estimation of root node against several
phylogenies, whose root node was estimated with PDTREE;
for more detailed testing, the program SQCHECK was used by R. D.-U.}

   uses DOS, crt;

   const   LIMIT = 579;     {Maximum number of tips and nodes on the tree= (number of tips*2)-1.}
        TWICELIMIT = 1158; {Limit & twicelimit are now the optima for the 254 species phylogeny}
        BLMin=1E-11; {branch lengths smaller than BLMin are considered polytomies}
   type  TipName = string[2];
         Equates = record
         {Stores information concerning a double node on a phylogeny, in which:
             Z = two character name of the node being described.
          Y, W = names of the nodes or tips resulting from Z.
             U = name of the most recent ancestor of Z.
                    (Given a value of "**" for the root node.)
             ZIndex = index of node in the group1 array. (cached from findIndex)
             YIndex,
             WIndex = indices of Z's descendents in the group1 array
             UIndex = index of Z's most recent ancestor in the group1 array
             X = expected variance of change from Y to Z.
             V = expected variance of change from W to Z.
             T = expected variance of change from Z to U.
                     (Given a value of 0 for the root node.)
             S = total number of tips resulting from change in Y.
             R = total number of tips resulting from change in W.}
           Z,Y,W,U:TipName;
           ZIndex,YIndex,WIndex,UIndex,S,R : integer;
           X,V,T:double;
            end;
         Typol = array [1..LIMIT] of Equates;
         {Combines a series of records concerning each node of a double phylogeny.}
      Value = record
         {Stores the name and trait value for a double point
                  on a phylogeny.}
            Name:TipName;
            Valup:double; {for "ghost" nodes in polytomies}
            Valdown:double; {as above}
            divisdown:double; {indicates # of ghost levels}
            ghostnode:boolean; {is this a "ghost" node?}
            end;
      Value1 = record
         {Stores the name and trait value for a double point
                  on a phylogeny.}
            Val:double; {for "ghost" nodes in polytomies}
            Name:TipName;
            end;
      Group= array [1..limit] of Value;
      Group1 = array [1..limit] of Value1;
         {Combines an ordered series of trait values,
            e.g., the names and values for all points on a phylogeny.}
      List = array [0..Limit] of double;
         {Stores a list of real numbers for use in calculating squared-change parsimony}
      NN=string[2];{for entering the name of the node whose value can be fixed}
      difNodePtr = ^difNode;
      difNode = record
                 diff:  double;     {the node value difference}
                 next: difNodePtr;  {next difNode in the list}
                 Name1,Name2: tipName; {the names of the two nodes}
             End;
      bigCharArray = array[1..32768] of char; {for output buffering}
      bigBuff = ^bigCharArray;     {allocate those buffers from the heap}


var     Equate:Typol; {An array of records containing complete information
                    on the phylogeny of interest.}
    TraitA:group1; {used in the input data process}
    TraitB:group1; {also used in reading data from the .inp file}
    METraitA : group; {Arrays of tip names and values}
    ValueFixed:double; {for node with fixed value}
    Nodefixed:NN; {for node with fixed value}
    FixNode:boolean; {controls whether any node will have a fixed value}
    donep:boolean; {to control for repeated runs of the program}
    ch:char; {as above}
    SquaredTotal1:double; {to compute squared-change parsimony}
    valueofinterest:double; {holds the value of node to be fixed}
    traitIndex:integer; {1 to use trait1, 2 to use trait2}
    numiterat:integer; {counter for number of iterations}
    ChangesA:List; {the value of the change between nodes or node-tip: to compute sq. cahnge pars}
    NodeFile:string; {used for assigning names to putput files}
    MEGNodes:text;  {more output file stuff}
    outs:string; {to convert double precision number to strings so that output is "nice"}
    difList : difNodePtr;
    NDBuff : bigBuff;
    TipNumber,    {Total number of tips on the phylogeny.}
     iterations   {Number of iterations used for Squared-Change Parsimony algorithm.}
         : integer;
    done : boolean;
    RootName,     {Base name of files to be used and created.}
    DataName,    {RootName + '.inp'.  Name of the file containing input data.}
    ResultName1, {RootName + 'ME1' or 'ME2';  Name of file to receive iteration results}
    ResultName2, {RootName + 'VAR'; Name of file to receive list of node names}
    ResultName3  {RootName + 'SC1' or 'SC2'; Name of file to receive branch lengths}
         : string;

{***************************************************************************}
{This function simply displays the title screen.}
{***************************************************************************}
Procedure doIntro;
   Var userChar : char;
   Begin
      clrscr;
      writeln('********** Phenotypic Diversity Analysis Programs - PDSQCHP Module **********');
      writeln;
      writeln('  Version 2.0 Copyright 8 May 1997 by Ramon Diaz-Uriarte, Peter E. Midford,');
      writeln('                          and Theodore Garland, Jr.');
      writeln('  Direct all questions to T.G. at Department of Zoology, 430 Lincoln Drive,');
      writeln('           University of Wisconsin, Madison, WI 53706-1381  U.S.A.');
      writeln('Telephone 608-262-4437; Facsimile 608-262-9083; E-mail tgarland@macc.wisc.edu');
      writeln;
      writeln('     This program will "reconstruct" hypothetical ancestral values of a');
      writeln('trait using squared-change parsimony.  The program reconstructs values for');
      writeln('one trait at a time from a .INP file.  It can handle hard polytomies, and');
      writeln('allows fixing of the value of one internal node.  It can also be used to');
      writeln('compute an evolutionary correlation between two traits (see PDINSTR.DOC).');
      writeln('     The input file must be in *.INP format.  If you have a file in a');
      writeln('different format, then use PDTREE to convert it to the .INP format.');
      writeln('     Please cite the following paper, which also presents examples:');
      writeln('Garland, T., Jr., K. L. M. Martin, and R. Diaz-Uriarte. 1997.');
      writeln('     Reconstructing ancestral trait values using squared-change');
      writeln('     parsimony: plasma osmolarity at the origin of amniotes.');
      writeln('     Pages 425-501 in S. S. Sumida and K. L. M. Martin, eds.');
      writeln('     Amniote Origins: Completing the Transition to Land.');
      writeln('     Academic Press, San Diego.');
      writeln;
      writeln('                      Press any key to continue.');
      userChar := readkey;
   End; {procedure doIntro}


{***************************************************************************}
{This procedure prompts the user for the root name.  The root name is the
name of the input file and the default name for all the output files.  Note
that if the user (foolishly) specfies an extension other than .INP, the
program will handle it, by saving the name in dataName (since the file
will be copied into the .BL_ file, the program needs to keep the actual
name.  In any case, this procedure checks whether the file actually exists.
If it does not, the program will keep prompting until the user enters a
file that exists or the user presses control-break.}
{***************************************************************************}
Procedure getRootName(Var rootName:string;Var dataName:string);
   Var dataTemp : text;
   Begin
      clrscr;
      writeln;
      writeln('NOTE:  Your file must be in the .INP format!');
      writeln;
      write ('Please enter the root name (including path) now: ');
      readln (RootName);
      If (pos('.',RootName) = 0)
      Then dataName:=RootName+'.inp'
      Else Begin
         dataName:=RootName;
         rootName:=copy(rootName,0,pos('.',rootName)-1);
      End;
      {$I-}
      assign(dataTemp,dataName);
      reset(dataTemp);
      close(dataTemp);
      {$I+}
      While (IOREsult <> 0) Do Begin
         writeln;
         writeln('Error - The file name you entered is invalid or does not exist.');
         write('Please reenter the root name (including path): ');
         readln(rootName);
         If (pos('.',RootName) = 0)
         Then dataName:=RootName+'.inp';
         {$I-}
         assign(dataTemp,dataName);
         reset(dataTemp);
         close(dataTemp);
         {$I+}
      End;
   End;

procedure GetData (var TipNumber:integer; var Equate:Typol;
                       var TraitA:group1; var TraitB:group1;
                       DataName:string);
{Reads input data on tip values and phylogeny from a text file named *.inp}
var data:text;
    Space:char;
    i:integer;
begin
  Assign (data, DataName);  Reset (data);
  for i:=1 to 3 do readln (data);
  for i:=1 to 10 do read (data, Space);
  readln (data, TipNumber);
  for i:= 1 to 4 do readln (data);
  for i:=1 to TipNumber Do Begin
    read (data, TraitA[i].Name,TraitA[i].Val);
    read(data, Space, Space ,Space);
    readln(data, TraitB[i].Name,TraitB[i].Val);
  End;
  for i:=1 to 3 do readln (data);
  for i:=1 to TipNumber-1 do
    with Equate[i] do Begin
      readln (data,Z,space,Y,X,S,space,W,V,R,space,U,T);
    End; {with}
    Close (data);
end;

{***************************************************************************}
{This function searches the array start, looking for a value record that
has a name that matches letter.  When it finds the record, it returns the
array index of the record.  If it fails, it returns 0.
{***************************************************************************}
Function findIndex (Start:group; TipNumber:integer;Letter:TipName) : integer;
   Var done:boolean;
      i:integer;
      A:integer;
      maxVal:integer;
   Begin
      i:=1;
      A:=0;
      maxVal:=2*tipNumber-1;
      done:=false;
      While done=false Do Begin
         If Start[i].Name=Letter Then Begin
            A:=i;
            done:=true;
         End
         Else
            If i<maxVal
            Then i:=i+1
            Else done:=true;
      End; {while}
      findIndex := A;
   End; {findIndex}

{***************************************************************************}
{This function calculates the mean of the values in Start}
{***************************************************************************}
Function Mean (Number: integer; Start:group1):double;
var sum: double;
     i:integer;
begin
 sum:=0;
 for i:=1 to Number do begin
   sum:=sum + Start[i].Val;
 end;
 Mean:= sum/Number;
end;

{***************************************************************************}
{This procedure transfers the tip values from Trait into FLTrait
for use in the Squared-Change methods.  Then it sets all of the internal
nodes equal to the mean of the tip values.  Finally, it sets up the
index fields in equate using findIndex, so the program doesn't have to
keep calling findIndex (which is slow!) in the middle of loops.
{***************************************************************************}
Procedure GetLists (Trait:group1; var FLTrait:Group;
                   TipNumber:integer; var Equate:Typol);
  {Transfers the tip values from Trait into FLTrait
    for use in the Squared-Change methods.
    Sets all of the nodes equal to the mean of the tip values.}
   Var i:integer;
      tipMean : double;
   Begin
      writeln;
      writeln('The program is now building node tables ... ');
      writeln;
      For i:=1 To TipNumber Do Begin
         FLTrait[i].Name:=Trait[i].Name;
         FLTrait[i].Valdown:=Trait[i].Val;
         FLTrait[i].Valup:=Trait[i].Val;
         FLTrait[i].ghostnode:=False;
      End; {for}
      tipMean := Mean(tipNumber,Trait);
      For i:=(TipNumber+1) To (2*TipNumber-1) Do Begin
         FLTrait[i].Name:=Equate[i-TipNumber].Z;
         FLTrait[i].Valup:=tipMean;
         FLTrait[i].Valdown:=FLTrait[i].Valup;
         FLTrait[i].ghostnode:=False;
      End; {for}
      For i:=1 to TipNumber Do Begin
         Equate[i].ZIndex := findIndex(FLTrait,tipNumber,Equate[i].Z);
         Equate[i].YIndex := findIndex(FLTrait,tipNumber,Equate[i].Y);
         Equate[i].WIndex := findIndex(FLTrait,tipNumber,Equate[i].W);
         Equate[i].UIndex := findIndex(FLTrait,tipNumber,Equate[i].U);
      End;
   End; {procedure getlists}

{***************************************************************************}
{ This Function formats coprocessor double reals without the extra exponent
  and mantessa digits, returning the result as a string.
  It is modified from a routine by P. E. Midford. }
{***************************************************************************}
Function extendWrite2(num :double):string;
   Var outStr :string[19];
   Begin
      str(num:19:-11,outStr);
      delete(outStr,16,2);
      extendWrite2:=outStr;
   End; {procedure extendwrite}

{***************************************************************************}
{This procedure is key to the handling of polytomies.  It does two things:
1) it finds which nodes are "ghost nodes" (i.e., have b.l=0);
if a node is not a ghost, it assigns the value of divisdown; 2) it finds
how many levels of "ghostness" are above the node of interest; this step
is necessary to divide by the right number when computing the value of a
node.  More details in PDSQCHP.DOC}
{***************************************************************************}
Procedure ghost (var METraitA: group; Equate: Typol);
   Var i,j:integer;
       ghostCount : integer;
   Begin
      For i:=1 To (TipNumber-1) Do Begin {this affects branch lengths of tips}
         With Equate[i] Do Begin
             If (X >= BLMin) Then Begin
                METraitA[YIndex].Ghostnode:=False;
                METraitA[YIndex].divisdown:=1/X;
             End {then}
             Else {if X < BLMin} Begin
                METraitA[YIndex].Ghostnode:=True;
                METraitA[YIndex].divisdown:=0;
             End; {else}
             If (V >= BLMin) Then Begin
                METraitA[WIndex].Ghostnode:=False;
                METraitA[WIndex].divisdown:=1/V;
             End
             Else {if V < BLMin} Begin
                METraitA[WIndex].Ghostnode:=True;
                METraitA[WIndex].divisdown:=0;
             End;
          End; {with}
      End;
      For i:=2 To (TipNumber-1) Do Begin {all other internal nodes}
         With Equate[i] Do Begin
            If (T < BLMin) Then Begin
               {D:=findINdex(METraitA,TipNumber,Z);}
               METraitA[ZIndex].Ghostnode:=True;
               METraitA[ZIndex].divisdown:=0;
            End;
            If (T >= BLMin) Then Begin
               METraitA[ZIndex].Ghostnode:=False;
               METraitA[ZIndex].divisdown:=1/T;
            End;
         End; {with}
      End;  {End of part 1}

      {**Here comes part 2***}
      {In all cases examined, the program runs OK without a loop for j;
      there is a chance, though, that a very weird input file might
      require 2 runs of the loop to get the divisdown value correct.
      It is left here just in case.}
      ghostCount := 0;
      For j:=1 to 2 do begin
         For i:=1 to (TipNumber-2) do begin
            With Equate[TipNumber-i] Do Begin
               If METraitA[ZIndex].Ghostnode=True Then Begin
                  METraitA[ZIndex].divisdown:=METraitA[YIndex].divisdown
                                               +METraitA[WIndex].divisdown;
                  If (j=1) Then ghostCount := ghostCount+1;
               End; {then}
            End; {with}
         End; {for i}
      End; {for j}
      writeln;
      writeln('Tree contains ',ghostCount,' ghost nodes.');
      writeln;
   End; {procedure ghost}

{***************************************************************************}
{This procedure allocates enough difNodePtr structures to hold all the nodes
in the tree and links them into a list.
{***************************************************************************}

Procedure setupDifList(var dList :difNodePtr;tipNumber :integer);
   Var i : integer;        {loop counter}
       temp : difNodePtr;  {holds the newly allocated node}
   Begin
      dlist := NIL;
      For i := 1 to 2*(tipNumber-1) Do Begin
          new(temp);
          temp^.name1 := '';
          temp^.name2 := '';
          temp^.diff := 0.0;
          temp^.next := dlist;
          dlist := temp;
      End; {for i}
   End; {procedure setupDifList}

{***************************************************************************}
{This procedure calculates differences between tips or nodes for use in
Squared-Change Parsimony methods.  It has the same name as in similar
procedures in CMSINGLE and SQCHP2, but the formulas for computing the
changes are slightly different; they use valup, which is the correct one.
See SQCHP1.TXT}
{***************************************************************************}
procedure getChanges (Var ChangeA : List;
                      Trait : group;
                      TipNumber : integer;
                      Equat : Equates;
                      Var j : integer;
                      Var node : difNodePtr);

   Begin
      With Equat Do Begin
         ChangeA[j]:=(Trait[YIndex].Valup - Trait[ZIndex].Valup);
         ChangeA[j+1]:=(Trait[WIndex].Valup - Trait[ZIndex].Valup);
         node^.name1 := Trait[YIndex].Name;
         node^.name2 := Trait[ZIndex].Name;
         node^.diff := ChangeA[j];
         node := node^.next;
         node^.name1 := Trait[WIndex].Name;
         node^.name2 := Trait[ZIndex].Name;
         node^.diff := ChangeA[j+1];
         node := node^.next;
         j:=j+1;
      End; {with equat}
   End; {procedure}

{***************************************************************************}
{This procedure calls getChanges to get differences used in Squared-Change
methods}
{***************************************************************************}
Procedure calculateChanges (TipNumber:integer;
                            var ChangeA:List;
                            METraitA:group;
                            Equate:Typol;
                            dList : difNodePtr);
   Var i,j :integer;
       done:boolean;
       nextNode : difNodePtr;
   Begin
      j:=1;
      nextNode := dList;
      for i:=1 to (TipNumber-1) do
         GetChanges (ChangeA,METraitA,TipNumber,Equate[i],j,nextNode);
   End; {procedure calculatechanges}

{***************************************************************************}

{gets name and value for node whose value will be fixed, and the node in
the output}
{***************************************************************************}
Procedure FixedNodes (var anynode: boolean;
                      var nodefixed:nn;
                      var valuefixed:double;
                      tipNumber : integer;
                      nameList:typol);
   Var userChar : char; {holds the character typed by the user}
       done : boolean;
   {************************************************************************}
   {This function is NOT the same as findIndex.  It searches through the
    array of equates (which only includes internal nodes) and returns
    true, rather than an index, if it matches.}
   {************************************************************************}
   Function checkName(name:nn) : boolean;
      var i : integer;
          doneval : boolean;
      Begin
         i := 1;
         doneVal := false;
         While((i <= tipNumber-1) and (doneVal = false)) Do Begin
            If nameList[i].Z = name
            Then doneVal := true
            Else i := i+1;
         End; {while}
         checkName := doneVal;
      End; { function fixedNodes }
   {************************************************************************}
   Begin {procedure fixednodes}
      writeln;
      writeln('*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*');
      writeln;
      writeln('NOTE:  it does NOT make any sense to fix "ghost" nodes.');
      writeln;
      write('Do you want to fix the value of a node? (Y/N/Enter=N): ');
      userChar := upcase(readkey);
      If (userChar = chr(13))
      Then writeln('N')
      Else writeln(userChar);
      While((userChar <> 'Y') And
            (userChar <> 'N') And
            (userChar <> chr(13))) do Begin;
         write('Please press Y or N: ');
         userChar := upcase(readkey);
         If (userChar = chr(13))
         Then writeln('N')
         Else writeln(userChar);
      End; {while}
      If ((userChar = 'Y') Or (userChar = '')) Then Begin
         writeln;
         write('Enter the two letter name of the node to fix (case counts): ');
         readln(Nodefixed);
         done := checkName(nodeFixed);
         If done Then anyNode := true;
         While (done = false) Do Begin
            writeln('No interior node named "',NodeFixed,'" could be found in the tree.');
            writeln;
            write('Would you like to try another name? (Y/N/Enter=Y): ');
            userChar := upcase(readkey);
            If (userChar = chr(13))
            Then writeln('Y')
            Else writeln(userChar);
            If (userChar <> 'N')
            Then Begin
               write('Enter another two letter name: ');
               readln(nodeFixed);
               done := checkName(nodeFixed);
               If done Then anyNode := true;
            End {then}
            Else Begin
               done := true;
               anyNode := false;
            End; {else}
         End; {while}
         If anyNode Then Begin
            writeln;
            write('Enter the value at which you want to fix node "',Nodefixed,'": ');
            readln(valuefixed);
         End; {then}
      End {then}
      Else anynode := false;
      writeln;
   End; {procedure fixednodes}

{****************************************************************************}
{This function prompts the user to specify which trait to reconstruct values
 for.}
{****************************************************************************}
Function traitChoice : integer;
   Var userStr: String;
   Begin
      writeln('*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*');
      writeln;
      writeln('Enter 1 to use trait 1');
      write('Enter 2 to use trait 2                      ');
      readln(userStr);
      While ((userStr <> '1') And (userStr <> '2')) do Begin
         write('Enter either 1 or 2: ');
         readln(userStr);
      End;
      If userStr = '1'
      Then traitChoice := 1
      Else TraitChoice := 2;
      writeln;
   End; {function traitChoice}


{****************************************************************************}
{This procedure allows the user to change the root names of the output files.}
{****************************************************************************}
Procedure outputName(Var rootName: string;traitCode : integer);
   Var
      newName: string;
      ch: char;
   Begin
      writeln('*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*');
      writeln;
      writeln('By default, the output file has the same name as the input file,');
      writeln('but with the extension .ND_, where _ is either 1 or 2, depending');
      writeln('on which trait you chose.');
      writeln;
      If (traitCode = 1)
      Then writeln('You chose trait 1, so the extension will be .ND1')
      Else writeln('You chose trait 2, so the extension will be .ND2');
      writeln;
      write('Do you want to change the root name of the output file? (Y/N/Enter=N): ');
      ch := upcase(readkey);
      If (ch = chr(13))
      Then writeln('N')
      Else writeln(ch);
      While((ch <> 'Y') And
            (ch <> 'N') And
            (ch <> chr(13))) do Begin;
         write('Please press Y or N: ');
         ch := upcase(readkey);
         If (ch = chr(13))
         Then writeln('N')
         Else writeln(ch);
      End; {while}
      If ((ch = 'Y') Or (ch = '')) Then Begin
         write('Please enter the root name for the output file: ');
         readln (newname);
         rootname:=newname;
      End; {then}
   End; {procedure outputname}

{****************************************************************************}
{This procedure writes the "heading" file.  The heading file is the list of
all the internal node names.}
{****************************************************************************}
Procedure writeheading(Rootname:string;
                       METraitA:group;
                       Equate:Typol;
                       Tipnumber:integer);
   Var i,j:integer;
       headf:text;
       headingfile:string;
   Begin
      headingfile:=RootName+'.VAR';
      assign(headf,headingfile);
      rewrite(headf);
      write(headf, 'iter.   ');
      For i:=1 To (TipNumber-1) Do
      With Equate[i] Do Begin
          write (headf, 'v',i);
          If i < 100 Then write(headf,' '); {add one space if <100}
          If i < 10 Then write(headf,' ');  {add two spaces if <10}
          write(headf,' =',METraitA[ZIndex].Name,'            ');
      End;
      If tipNumber<100 Then write(headf,' ');
      If tipNumber<10 Then write(headf,' ');
      write(headf,'v',i+1,' ','=Sum of Sq. Change');
      writeln(headf);
      close(headf);
   End;

{****************************************************************************}
{This procedure performs a "means" algorithm by iteratively setting each
node's trait to the weighted mean of its three nearest neighbors.  Means are
weighted by branch length.  Nodal values obtained at each iteration of the
algorithm are stored in the file ending in either ".ND1" or ".ND2" depending
on which trait the user selected.  If a node is to be fixed, then the
procedure just (re)assigns the node its fixed value.}
{****************************************************************************}
Procedure weightedNodes (Var METraitA : group;
                         Equate : Typol;
                         tipNumber : integer;
                         Var MEGFile: text;
                         Var valueOfInterest:double;
                         numiterat: integer);
   Var i :integer;  {loop counter}
       outs:string; {receives formatted string from extendWrite2}
       MEA,MEB,MEC:double; {Mean evolution intermediate values??? XX}
   Begin
      With Equate[1] Do Begin
         If fixNode And (METraitA[ZIndex].Name = NodeFixed) Then Begin
            MEtraitA[ZIndex].Valup:=ValueFixed;
            MEtraitA[ZIndex].Valdown:=ValueFixed;
            outs:=extendWrite2(METraitA[ZIndex].Valup);
         End {METrair = NodeFixed}
         Else Begin {if METraitA[ZIndex].Name <> NodeFixed}
            If METraitA[YIndex].ghostnode=false
            Then MEA:=METraitA[YIndex].Valdown*METraitA[YIndex].divisdown
            Else MEA:=METraitA[YIndex].Valdown;
            If METraitA[WIndex].ghostnode=false
            Then MEB:=METraitA[WIndex].Valdown*METraitA[WIndex].divisdown
            Else MEB:=METraitA[WIndex].Valdown;
            METraitA[ZIndex].Valup:=(MEA+MEB)/
                                     (METraitA[YIndex].divisdown+
                                       METraitA[WIndex].divisdown);
            METraitA[ZIndex].Valdown:=METraitA[ZIndex].Valup;
            outs:=extendWrite2(METraitA[ZIndex].Valup);
         End;
         write (MEGFile,numiterat,'  ',outs,' ');
      End; {with equate[1]}
      For i:=2 To (tipNumber-1) do begin
         With Equate[i] Do Begin
            If fixNode And (METraitA[ZIndex].Name = nodeFixed) Then Begin
               MEtraitA[ZIndex].Valup:=ValueFixed;
               MEtraitA[ZIndex].Valdown:=ValueFixed;
               outs:=extendWrite2(METraitA[ZIndex].Valup);
            End {METrair = NodeFixed}
            Else {if METraitA[ZIndex].Name <> NodeFixed} Begin
               If METraitA[ZIndex].ghostnode=true Then Begin
                  If METraitA[YIndex].ghostnode=false
                  Then MEA:=METraitA[YIndex].Valdown*
                             METraitA[YIndex].divisdown
                  Else MEA:=METraitA[YIndex].Valdown;
                  If METraitA[WIndex].ghostnode=false
                  Then MEB:=METraitA[WIndex].Valdown*
                             METraitA[WIndex].divisdown
                  Else MEB:=METraitA[WIndex].Valdown;
                  METraitA[ZIndex].Valdown:=MEA+MEB;
                  METraitA[ZIndex].Valup:=METraitA[UIndex].Valup;
                  outs:=extendWrite2(METraitA[ZIndex].Valup);
               End {then}
               Else {if (METraitA[ZIndex].ghostnode=false} Begin
                  If METraitA[YIndex].ghostnode=false
                  Then MEA:=METraitA[YIndex].Valdown*
                              METraitA[YIndex].divisdown
                  Else MEA:=METraitA[YIndex].Valdown;
                  If METraitA[WIndex].ghostnode=false
                  Then MEB:=METraitA[WIndex].Valdown*
                             METraitA[WIndex].divisdown
                  Else MEB:=METraitA[WIndex].Valdown;
                  MEC:=METraitA[UIndex].Valup*METraitA[ZIndex].divisdown;
                  METraitA[ZIndex].Valup:=(MEA+MEB+MEC)/
                   (METraitA[YIndex].divisdown+METraitA[WIndex].divisdown+METraitA[ZIndex].divisdown);
                  METraitA[ZIndex].Valdown:=METraitA[ZIndex].Valup;
                  outs:=extendWrite2(METraitA[ZIndex].Valup);
               End; {ghostnode=false}
            End; {if METrait <>}
         End; {with equate[i]}
         write (MEGFile,outs,' ');
      End;{for i}
   End; {procedure weightednodes}


{****************************************************************************}
{This procedure prints the details of Squared-Change analysis, including
inferred nodal values and differences between points on the phylogeny.
It is here where the actual value for squared-change parsimony is computed.}
{****************************************************************************}
Procedure PrintMEResults (tipNumber : integer;      {number of tips}
                          Var squaredTotal1:double; {total sum of squared change}
                          Var outs:string;          {received formated string from extendWrite2}
                          dList : difNodePtr);      {list of differences (used to be a file)}
   Type differences = array [1..TWICELIMIT] of double;
   Var ndiff:text;
       node :difNodePtr;             {difference node to add}
       SquaredChange1:differences;
       i : integer;                  {loop counter}
   Begin
      node := dList;
      squaredTotal1:=0;
      For i:=1 To TWICELIMIT Do 
        SquaredChange1[i]:=0;
      node := dList;
      For i:=1 To (2*(TipNumber-1)) Do Begin
         SquaredChange1[i]:=sqr(node^.diff);
         SquaredTotal1:=SquaredTotal1+SquaredChange1[i];
         node:=node^.next;
      End; {for i}
      outs:=extendWrite2(SquaredTotal1);
   End; {procedure printmerresults}


{****************************************************************************}
{This procedure returns all the difNodes on dList to the heap.  It should be
called after processing each file since different files will have different
numbers of nodes and therefore require different sized dLists.}
{****************************************************************************}
Procedure flushTipList(var dList : difNodePtr);
   Var temp : difNodePtr;
   Begin
      While dlist <> NIL Do Begin
         temp := dList;
         dList := temp^.next;
         dispose(temp);
      End; {while}
   End; {procedure flushTipList}

{****************************************************************************}
{This function is called by writeBLFile to format the current time in a
format similar to that in CMSINGLE (but improved by not writing things like
16:4:7 for 16:04:07).
{****************************************************************************}
Function timeStr : string;
   Var year : word;      {holds the integer year value; doesn't work as integer}
       month : word;     
       day : word;
       ignore : word;
       hour : word;
       minute : word;
       second : word;
       sec100 : word;
       result : string;  {Holds final string as it is built}
       tmpStr : string;  {used for formating each field one at a time}
   Begin
      result := '';
      getDate(Year,Month,Day,Ignore);
      getTime(hour,minute,second,sec100);
      str(day:2,tmpStr);
      result := result + tmpStr + ' ';
      case month of
         1:  result := result + 'Jan ';
         2:  result := result + 'Feb ';
         3:  result := result + 'Mar ';
         4:  result := result + 'Apr ';
         5:  result := result + 'May ';
         6:  result := result + 'Jun ';
         7:  result := result + 'Jul ';
         8:  result := result + 'Aug ';
         9:  result := result + 'Sep ';
         10: result := result + 'Oct ';
         11: result := result + 'Nov ';
         12: result := result + 'Dec ';
         Else result := result + 'XXX ';
      End; {case}
      str(Year:4,tmpStr);
      result := result + tmpStr + ' at ';
      str(hour,tmpStr);
      result := result + tmpStr + ':';
      If minute<10
      Then result := result + '0';
      str(minute,tmpStr);
      result := result + tmpStr + ':';
      If second<10
      Then result := result + '0';
      str(second,tmpStr);
      {result := result + tmpStr + '.'; 
      str(sec100:2,tmpStr);}            {hundredths don't seem to fit in}
      timeStr := result + tmpStr;       {80 columns, so don't write them}
   End; {function timeStr}

{****************************************************************************}
{This procedure is responsible for writing the third results file, named BL_.
This file, modeled on the .RES file of CMSINGLE, contains the .INP file,
squared-change parsimony node values, and (branch-length weighted)
distances between nodes.
Note:
  this function may be recalculating values that can be obtained elsewhere,
  but the loops are straightforward enough that there isn't much of a
  performance hit.}
{****************************************************************************}
Procedure writeBLengthFile(RootName : string;
                           var BLName : string;
                           dataName : string;
                           METraitA : group;
                           Equate : typol;
                           iterCount : integer;
                           tipNumber : integer;
                           traitIndex : integer);
   Var 
       BLFile : text;
       DataFile : text;
       sqsum : double;
       percent : double;
       percentSum : double;
       i : integer;                        {loop counter}
       lineBuf : string;                   {.INP->lineBuf->.BL_}
       fileBuf1 : bigBuff; {buffer for results output}
       fileBuf2 : bigBuff; {buffer for .INP file input}
   Begin
      If (traitIndex = 1)
      Then BLName:=RootName+'.BL1'
      Else BLName:=RootName+'.BL2';
      assign(BLFile,BLName);
      new(fileBuf1);
      setTextBuf(BLFile,fileBuf1^);
      rewrite(BLFile);
      write(BLFile,'This is the output of PDSQCHP.EXE for ');
      write(BLFile,iterCount);
      writeln(BLFile,' iterations on ',timestr);
      writeln(BLFile);
      writeln(BLFile,dataName,':');
      assign(dataFile,dataName);
      new(fileBuf2);
      setTextBuf(dataFile,fileBuf2^);
      reset(dataFile);
      Repeat
         readln(dataFile,lineBuf);
         writeln(BLFile,lineBuf);
      Until eof(dataFile);
      close(dataFile);
      writeln(BLFile);

      writeln(BLFile,'***********************************************************************');
      writeln(BLFile,'     Squared-change parsimony nodes, computed with weighting by branch');
      writeln(BLFile,'lengths, for trait ',traitIndex,':');
      writeln(BLFile);
      writeln(BLFile,'Node       Value');
      writeln(BLFile,'----------------------');
      For i:=1 to tipNumber-1 Do
      With equate[i] Do Begin
         outs := extendWrite2(METraitA[ZIndex].Valup);
         writeln(BLFile,Z,'   ',outs);
      End; {with}
      writeln(BLFile);
      {calculate percentages before writing the next (long) header}
      sqsum := 0;
      percentSum := 0;
      For i:=1 to tipNumber-1 Do
      With equate[i] do Begin
         sqsum := sqsum + sqr(METraitA[YIndex].Valup-METraitA[ZIndex].Valup);
         sqsum := sqsum + sqr(METraitA[WIndex].Valup-METraitA[ZIndex].Valup);
      End; {with}
      {write the long header}
      writeln(BLFile,'***********************************************************************');
      writeln(BLFile,'     The following columns indicate the computed differences between');
      writeln(BLFile,'the above nodal values, i.e., the amount of phenotypic evolution that');
      writeln(BLFile,'is estimated to have occurred along each branch segment.  Below, these');
      writeln(BLFile,'differences are labeled as "Length."');
      writeln(BLFile,'     These differences can be used to compute an evolutionary');
      writeln(BLFile,'correlation between two traits, termed ME1G by:');
      writeln(BLFile,'Martins, E. P., and T. Garland, Jr. 1991. Phylogenetic analyses of the');
      writeln(BLFile,'     correlated evolution of continuous characters: a simulation study.');
      writeln(BLFile,'     Evolution 45:534-557.');
      writeln(BLFile,'Garland, T., Jr., R. B. Huey, and A. F. Bennett. 1991. Phylogeny and');
      writeln(BLFile,'     thermal physiology in lizards: a reanalysis.');
      writeln(BLFile,'     Evolution 45:1969-1975.');
      writeln(BLFile,'To do this, you need to run PDSQCHP twice, once for each of the two');
      writeln(BLFile,'traits in your .INP file.  Then, compute an ordinary Pearson product-');
      writeln(BLFile,'moment correlation coefficient between the "lengths" for the two');
      writeln(BLFile,'traits.  To compute the ME1P correlation, you need to first use PDTREE');
      writeln(BLFile,'to change all branch lengths to be equal to one, save the new .INP');
      writeln(BLFile,'file, then rerun PDSQCHP, etc.');
      writeln(BLFile,'     These differences ("Lengths") can also be used as branch lengths');
      writeln(BLFile,'on a phylogenetic tree.  To do this, you will need to use a word');
      writeln(BLFile,'processor or text editor to replace the branch lengths in a .PDI');
      writeln(BLFile,'version of your .INP file (you can create the former with PDTREE).');
      writeln(BLFile,'	All of the foregoing is explained in the file PDINSTR.DOC.');
      writeln(BLFile);
      writeln(BLFile,'For "Branch,"');
      writeln(BLFile,'1st column = Descendant');
      writeln(BLFile,'2nd column = Ancestor');
      writeln(BLFile,'Branch      "Length"          Percentage');
      writeln(BLFile,'--------------------------------------------');
      For i:=1 to tipNumber-1 Do
      With equate[i] do Begin
         outs := extendWrite2(METraitA[YIndex].Valup-METraitA[ZIndex].Valup);
         write(BLFile,Y,'  ',Z,'  ',outs,'  ');
         percent := 100*sqr(METraitA[YIndex].Valup-METraitA[ZIndex].Valup)/sqsum;
         percentSum := percentSum + percent; 
         outs := extendWrite2(percent);
         writeln(BLFile,outs);
         outs := extendWrite2(METraitA[WIndex].Valup-METraitA[ZIndex].Valup);
         write(BLFile,W,'  ',Z,'  ',outs,'  ');
         percent := 100*sqr(METraitA[WIndex].Valup-METraitA[ZIndex].Valup)/sqsum;
         percentSum := percentSum + percent;
         outs := extendWrite2(percent);
         writeln(BLFile,outs);
      End; {with}
      outs := extendWrite2(percentSum);
      writeln(BLFile,'                   Total = ',outs);
      writeln(BLFile);
      outs := extendWrite2(sqsum);
      writeln(BLFile,'Sum of squared changes = ',outs);
      close(BLFile);
      dispose(fileBuf1);
      dispose(fileBuf2);
   End;

{***************************************************************************}
Begin {main program}
   donep:=false;
   doIntro;
   while(donep=false) do begin
      getRootName(rootName,dataName);
      writeln; writeln;
      traitIndex := TraitChoice;
      GetData(TipNumber,Equate,TraitA,TraitB,DataName);
      outputname(RootName,traitIndex);
      FixedNodes(FixNode,NodeFixed,ValueFixed,tipNumber,Equate);
      If (traitIndex = 1)
      Then NodeFile:=RootName+'.ND1'
      Else NodeFile:=RootName+'.ND2';
      new(NDBuff);
      Assign(MEGNodes,NodeFile);
      setTextBuf(MEGNodes,NDBuff^);
      Rewrite (MEGNodes);
      If (traitIndex = 1)
      Then GetLists(TraitA,METraitA,TipNumber,Equate)
      Else GetLists(TraitB,METraitA,TipNumber,Equate);
      writeln ('How many iterations of the Squared-Change Parsimony algorithm should be');
      writeln ('  completed to analyze the present set of data?');
      readln (iterations);
      ghost(METraitA,Equate);
      writeheading(RootName,METraitA,Equate,Tipnumber);
      setupDifList(difList,tipNumber);
      For numiterat:=1 to iterations do begin
         writeln('Processing iteration number','  ',numiterat);
         WeightedNodes (METraitA,Equate,TipNumber,MEGNodes,valueofinterest,numiterat);
         CalculateChanges (TipNumber,ChangesA,METraitA,Equate,difList);
         PrintMEResults (TipNumber,SquaredTotal1,outs,difList);
         write(MEGNodes,' ',outs);
         writeln(MEGNodes);
      End; {for numiterat}
      close(MEGNodes);
      dispose(NDBuff);
      resultname1:=NodeFile;
      resultName2:=RootName+'.VAR';
      writeBLengthFile(rootName,
                       resultName3,
                       dataName,
                       METraitA,
                       Equate,
                       numiterat,
                       tipNumber,
                       traitIndex);
      writeln;
      writeln('*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*');
      writeln;
      writeln('Iteration results are stored in the file ',ResultName1,'.');
      writeln('The correspondence between columns and node names can be found in ',ResultName2,'.');
      writeln('Final results, differences, and the sum of squared changes can be found in');
      writeln('the file ',resultName3,', which is organized like the .RES file produced by CMSINGLE.');
      flushTipList(difList);
      writeln;
      write('Do you want to run the program again? (Y/N/Enter=N): ');
      ch := upCase(readKey);
      writeln(ch);
      If(ch<>'Y') then donep:=true
   End; {while}
End. {program}
