{**************************************************************************}
{ PDTIPS.PAS module of Phenotypic Diversity Analysis Programs
{ This program copyright 1993, 1994, 1996, 1997
{ Written by Peter E. Midford for Theodore Garland, Jr.,
{ with extensive copying from PDTREE.PAS
{ Introductory screen added 16 Feb. 1997 by T.G.
{ Last modified 6 August 1998 PEM.
{**************************************************************************}

{ This program computes statistics on data values for the tips of a tree}

Program PDTips(input,output);

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Uses tp_utils,tp_io,dos,crt;

Var simData :nodeDataPtr;       {A simulated data set}
    simStats :statPak;          {Statistics for current data}
    bcStats :statPak;           {Statistics for transformed data}
    simFile :text;              {The input (.SIM) file}
    statFile :text;             {The output (.TPS) file}
    boxFile :text;              {The Box-Cox (.BOX) file}
    newSimFile :text;           {The transformed SIM file}
    median1 :extended;          {The median value of trait 1}
    median2 :extended;          {The median value of trait 2}
    globalMin1 : extended;      {The lower bound for all trait 1 data}
    globalMax1 : extended;      {The upper bound for all trait 1 data}
    globalMin2 : extended;      {The lower bound for all trait 2 data}
    globalMax2 : extended;      {The upper bound for all trait 2 data}
    i :integer;                 {General purpose index variable}
    userLimit :integer;         {Sets limit on number of data sets to read}
    binCount :integer;          {The number of categories}
    binBase :integer;           {Starting value for categories}
    tipCount :integer;          {Holds the number of tips}
    rawTips1 : tipArray;        {}
    rawTips2 : tipArray;        {}
    transTips1 :tipArray;       {Best transform of trait 1 data}
    transTips2 :tipArray;       {Best transform of trait 2 data}
    inputFlName :pathStr;       {The name of the input file}
    inputExt :typstring;        {The extension of the input file ('.SIM')}
    statFileName :pathStr;      {The name of the output file}
    boxFileName :pathStr;       {The name of the skew file}
    newSimFileName :pathStr;    {The name of the transformed SIM file}
    newExtStr : extStr;         {holds generated of Sim file extension for messages}
    userStr : string[4];        {Used for multiple character responses}
    userChar :char;             {Used for user single character responses}
    doBCStats :boolean;         {Flag for writing skewness and kurtosis file}
    doBCFile : boolean;         {flag for doing Box-Cox statistics}
    doNewSim :boolean;          {Flag for writing transformed SIM file}
    doStats : boolean;          {Flag for writing a statistics (.TPS) file}
    doFullStats : boolean;      {Flag for including medians, KS, BC stats}
    doCategories : boolean;     {True when user wants categorize data}
    doRanks : boolean;          {True when user wants to rank data}
    done :boolean;              {Used to signal reading another input file}
    succ :boolean;              {Indicates successful file open operation}
    writeNewTrait1 : boolean;
    writeNewTrait2 : boolean;
    foundBCError : boolean;
    showCategories :boolean;    {Flag for displaying categorization results}


{*********************************************************************}
{This procedure undoes the reassignment of input and output to the
{crt device.  See page 199 of the Turbo Pascal programmer's manual.   }
{*********************************************************************}
Procedure resetCrtAssignments;
   Begin
      Assign(input,'');
      reset(input);
      assign(output,'');
      rewrite(output);
   End;

{*********************************************************************}
{This procedure returns}
{*********************************************************************}
Procedure restoreCrtAssignments;
   Begin
      assignCrt(input);
      reset(input);
      assignCrt(output);
      rewrite(output);
   End;

{*********************************************************************}
{ This function categorizes a list of data according to the values of
{ binCount and binBase.
{*********************************************************************}
Procedure categorizeData(data     :nodeDataPtr;
                         binCount :integer;
                         binBase  :integer;
                         globalMin1 : extended;
                         globalMax1 : extended;
                         globalMin2 : extended;
                         globalMax2 : extended);
   Var maxData : extended;
       minData : extended;
       temp :nodeDataPtr;

   {*******************************************************************}
   { This local function figures out which bin "value" goes into.      }
   {*******************************************************************}
   Function computeBin(value    :extended;
                       min      :extended;
                       max      :extended;
                       binCount :integer) : integer;
      Begin
         computeBin := round((binCount-1)*(value-min)/(max-min));
      End; { computeBin }
   {*******************************************************************}
   Begin { procedure categorizeData }
      If binCount > 0
      Then Begin
         If (showCategories) {debugging breakout}
         Then Begin
            writeln('* Categorization Results for Trait 1 *');
            writeln('Name      Raw Data      Categorized Data');
         End; {then}
         temp := data;
         While (temp <> nil) Do Begin   {loop through list of data Nodes}
            If (showCategories)
            Then Begin
               write(temp^.name,'      ');
               extendwrite(output,temp^.data[1]);
            End; {then}
            temp^.data[1] := binBase + computeBin(temp^.data[1],
                                                  globalMin1,
                                                  globalMax1,
                                                  binCount);
            If (showCategories)
            Then writeln('           ',temp^.data[1]:2:0);
            temp := temp^.next;
         End; { while }
         If (showCategories) Then Begin  {pause to let user see results}
            writeln('Press enter to continue');
            readln;
            writeln;
         End; {then}
         If (showCategories) {debugging breakout}
         Then Begin
            writeln('* Categorization Results for Trait 2 *');
            writeln('Name      Raw Data      Categorized Data');
         End; {then}
         temp := data;
         While (temp <> nil) Do Begin  {loop again through list of data}
            If (showCategories)
            Then Begin
               write(temp^.name,'      ');
               extendWrite(output,temp^.data[2]);
            End; {then}
            temp^.data[2] := binBase + computeBin(temp^.data[2],
                                                  globalMin2,
                                                  globalmax2,
                                                  binCount);
            If (showCategories)
            Then writeln('           ',temp^.data[2]:2:0);
            temp := temp^.next;
         End; { while }
         If (showCategories) {pause to let user see results}
         Then Begin
            writeln('Press enter to continue');
            readln;
            writeln;
         End; {then}
      End; { then }
   End; { procedure categorizeData}

{*********************************************************************}
{ This function ranks a list of data in ascending order.  The
  algorithm is o(n^2) (so it could be improved), but it does what is
  required: the sum of the assigned ranks is always equal to the
  sum of the integers from 1 to tipCount.
{*********************************************************************}
Procedure rankData(data :nodeDataPtr;tipCount : integer);
   Var orderArray : array [1..maxTips] of nodeDataPtr;
       maxData : extended;
       minData : extended;
       rank : extended;
       index : integer;
       start : integer;
   {********************************************************************}
   {This local function resets the marks of all the nodes on the list.  }
   {These marks are used to pull each node out of the list only once,   }
   {when the node has the lowest trait value among the unmarked nodes   }
   {on the list.                                                        }
   {********************************************************************}
   Procedure resetMarks(data : nodeDataPtr);
      Var current : nodeDataPtr;
      Begin
         current := data;
         While (current <> nil) Do Begin
            current^.mark := false;
            current := current^.next;
         End; {while}
      End; {procedure resetMarks}
   {********************************************************************}
   {This function returns the node with the lowest trait value among the}
   {unmarked nodes in the list                                          }
   {********************************************************************}
   Function smallestData(data :nodeDataPtr;
                         index :integer)   {selects trait 1 or 2}
                                        : nodeDataPtr;
      Const largeValue = 9.999E99;
      Var temp :nodeDataPtr;
          min : extended;
          winner : nodeDataPtr;
      Begin
         temp := data;
         min := largeValue;
         While (temp <> nil) Do Begin
            If (Not(temp^.mark) And (temp^.data[index] < min))
            Then Begin
               min := temp^.data[index];
               winner := temp;
            End; {then}
            temp := temp^.next;
         End; { while}
         winner^.mark := true;
         smallestData := winner;
      End; { dataMinMax }
   {********************************************************************}
   {This local procedure assigns rank values to the index-th node in the}
   {orderArray.                                                         }
   {********************************************************************}
   Procedure assignRank(trait : integer);
      Var rank : extended;
          i : integer;
      Begin
         If (start = 0)
         Then orderArray[index]^.rank[trait] := index
         Else Begin
            rank := (start+index)/2.0;
            For i := start To index Do orderArray[i]^.rank[trait] := rank
         End; {else}
         start := 0;
      End; {procedure assignRank}
   {********************************************************************}
   Begin { procedure rankData }
      resetMarks(data);
      For index := 1 To tipCount Do orderArray[index] := smallestData(data,1);
      start := 0;
      For index := 1 To tipCount Do Begin
         If ((index < tipCount) And
             (orderArray[index+1]^.data[1] = orderArray[index]^.data[1]))
         Then Begin
            If (start = 0) Then start := index;
         End
         Else assignRank(1);
      End; {for i}
      resetMarks(data);
      For index := 1 To tipCount Do orderArray[index] := smallestData(data,2);
      start := 0;
      For index := 1 To tipCount Do Begin
         If ((index < tipCount) And
             (orderArray[index+1]^.data[2] = orderArray[index]^.data[2]))
         Then Begin
            If (start = 0) Then start := index;
         End
         Else assignRank(2);
      End; {for i}
   End; { procedure rankData}

{**************************************************************************}
{This function assigns each node rank values to their data fields.         }
{**************************************************************************}
Procedure assignRanks(data : nodeDataPtr);
   Var temp : nodeDataPtr;
   Begin
      temp := data;
      While (temp <> nil) Do Begin
         temp^.data[1] := temp^.rank[1];
         temp^.data[2] := temp^.rank[2];
         temp := temp^.next;
      End; {while}
   End; {assignRanks}

{**************************************************************************}
{ This function returns the median for one trait scanning the tip list
{ set up by PrepareTips.  Algorithm loosely based on GetMedian in PDANOVA.
{**************************************************************************}
Function tipMedian(data :nodeDataPtr; traitNum :integer) :extended;
  Var medianMinus,medianPlus :extended;
      current, kTip :nodeDataPtr;
      ties, noHigher, noLower :integer; {counts values relative to current}
  Begin
     current := data;
     While (current <> nil) Do Begin { loop through list of tips }
        noHigher := 0;
        noLower := 0;
        ties := 0;
        kTip := data;            {loop through list and count values}
        While (kTip <> nil) Do Begin {that are higher and lower}
           If (kTip^.data[traitNum] > current^.data[traitNum]) Then
              noHigher := noHigher + 1;
           If (KTip^.data[TraitNum] < current^.data[TraitNum]) Then
              noLower := noLower + 1;
           If (KTip^.data[TraitNum] = current^.data[TraitNum]) Then
              ties := ties + 1;
           kTip := kTip^.next;
        End; {while}
        If (abs(noHigher - noLower) < ties) Then Begin
           medianMinus := current^.data[traitNum];
           medianPlus := current^.data[traitNum]
        End {then begin}
        Else If (abs(noHigher - noLower) = ties) Then
                If (noLower > noHigher)
                   Then medianMinus := current^.data[traitNum]
                   Else medianPlus := current^.data[traitNum];
        current := current^.next
     End; {while}
     tipMedian := (medianPlus + medianMinus) / 2
  End; {function tipmedian}

{**************************************************************************}
{ This procedure calculates the tip statistics, using the tipStat procedure
{ (TP_UTILS).                                                              }
{**************************************************************************}
Procedure computeStatistics(Var stats :statPak;
                            data :nodeDataPtr);
   Begin
      tipStat(data,stats);
   End; {computeStatistics}

{*************************************************************************}
{This procedure calls the additional procedures to add medians and KS
 normality tests.                                                         }
{*************************************************************************}
Procedure computeFullStatistics(Var stats :statPak;
                                    data :nodeDataPtr);
   Begin
      stats.median[1] := tipMedian(data,1);
      stats.median[2] := tipMedian(data,2);
      KolSmir(data,stats);
   End;

{**************************************************************************}
{ This procedure calculates the Kolmogorov-Smirnov and Box-Cox
{ statistics using procedures in TP_UTILS and the median routine above.
{**************************************************************************}
Procedure computeBoxCoxStatistics(Var stats :statPak;
                                  Var bcstats :statPak;
                                  Var trans1 :tipArray;
                                  Var trans2 :tipArray;
                                  data :nodeDataPtr);
   Var lambda1, lambda2 : extended;
   Begin
      preBCStats(data,stats);
      BoxCox(data,lambda1,lambda2,trans1,trans2);
      stats.lambda[1] := lambda1;
      stats.lambda[2] := lambda2;
      If lambda1<>-999.0 Then Begin 
         arraySK(trans1,stats.number,1,bcstats);
         ksNorm(trans1,stats.number,1,bcstats)
      End {then}
      Else Begin
         If not(foundBCError) Then writeln;
         foundBCError := true;
         writeln('Trait 1 contains non-positive data; Box-Cox lambda set to -999');
         bcStats.ks[1] := -999.0;
         bcStats.skewness[1] := -999.0;
         bcStats.kurtosis[1] := -999.0;
      End; {else}
      If lambda2<>-999.0 Then Begin
         arraySK(trans2,stats.number,2,bcstats);
         ksNorm(trans2,stats.number,2,bcstats)
      End {then}
      Else Begin
         If not(foundBCError) Then writeln;
         foundBCError := true;
         writeln('Trait 2 contains non-positive data; Box-Cox lambda set to -999');
         bcStats.ks[2] := -999.0;
         bcStats.skewness[2] := -999.0;
         bcStats.kurtosis[2] := -999.0;
      End {else}
   End;{procedure computefullstatistics}

{**************************************************************************}
{ Displays a screen explaining the column numbers.                         }
{**************************************************************************}
Procedure displaySummary(statFlag: boolean;
                         statFileName : pathStr;
                         fullFlag: boolean;
                         boxFlag :boolean;
                         boxFileName : pathStr);
   Begin
      clrscr;
      If statFlag Then Begin
         writeln('Columns for Tip Statistics file: ',statFileName);
         writeln;
         writeln('Column    Description');
         writeln;
         writeln('1         Simulation Number');
         writeln('2         Pearson product-moment correlation of trait 1 and trait 2');
         writeln('3         Spearman rank correlation of trait 1 and trait 2');
         writeln('4,5       Minimum value of traits 1,2');
         writeln('6,7       Maximum value of traits 1,2');
         writeln('8,9       Mean of traits 1,2');
         writeln('10,11     Variance of traits 1,2');
         writeln('12        F statistic for ordinary least squares regression');
         writeln('13        Ordinary least squares slope');
         writeln('14        Ordinary least squares intercept');
         writeln('15        Reduced major axis slope');
         writeln('16        Reduced major axis intercept');
         writeln('17        Major axis slope');
         writeln('18        Major axis intercept');
         If fullFlag Then Begin
            writeln('19,20     Median of traits 1,2');
            writeln('21,22     Kolmogorov-Smirnov normality test p-value of traits 1,2');
         End; {then}
         writeln;
         writeln('Press return to continue');
         readln;
         writeln;
      End; {then}
      If boxFlag Then Begin
         clrscr;
         writeln('Columns for Box-Cox Skewness/Kurtosis file: ',boxFileName);
         writeln;
         writeln('Column    Description');
         writeln;
         writeln('1         Simulation Number');
         writeln('2         Box-Cox lambda score for trait 1');
         writeln('3         Box-Cox lambda score for trait 2');
         writeln('4         KS normality test p-value of trait 1 after Box-Cox transform');
         writeln('5         KS normality test p-value of trait 2 after Box-Cox transform');
         writeln('6         Skewness of trait 1 (before Box-Cox)');
         writeln('7         Skewness of trait 2 (before Box-Cox)');
         writeln('8         Kurtosis of trait 1 (before Box-Cox)');
         writeln('9         Kurtosis of trait 2 (before Box-Cox)');
         writeln('10        Skewness of trait 1 (after Box-Cox)');
         writeln('11        Skewness of trait 2 (after Box-Cox)');
         writeln('12        Kurtosis of trait 1 (after Box-Cox)');
         writeln('13        Kurtosis of trait 2 (after Box-Cox)');
         writeln;
         writeln('Press return to continue');
         readln;
         writeln;
      End
   End; {procedure displaysummary}
   {************************************************************************}
   Begin {program pdtips}
      resetCrtAssignments;
      clrscr;
      writeln('********* Phenotypic Diversity Analysis Program - PDTIPS Module *********');
      writeln;
      writeln('V. 1.0 Copyright 22 December 1999 by Peter E. Midford and Theodore Garland, Jr.');
      writeln;
      writeln('Direct all questions to T.G. at Department of Zoology, 430 Lincoln Drive,');
      writeln('           University of Wisconsin, Madison, WI 53706  U.S.A.');
      writeln('           Telephone:  608-262-4437; Facsimile:  608-262-6320');
      writeln('           Email:  tgarland@macc.wisc.edu or tgarland@facstaff.wisc.edu');
      writeln;
      writeln('This program is used to analyze simulated data (e.g., from PDSIMUL of');
      writeln('Garland et al., 1993, Syst. Biol. 42:265-292).  It will output a variety');
      writeln('of conventional (nonphylogenetic) statistics, including an ordinary Pearson');
      writeln('product-moment correlation coefficient, a Spearman rank correlation, means');
      writeln('and variances, regression slopes and intercepts, etc.');
      writeln('     PDTIPS can also categorize or rank the tip data prior to computing all');
      writeln('of the statistics.  As well, it can perform a Box-Cox transformation before');
      writeln('computing statistics.');
      writeln('     Finally, it can write a modified *.SIM file in which the tip data have');
      writeln('been categorized or ranked, and/or Box-Cox transformed.  This modified');
      writeln('file can then be analyzed by other modules, such as PDANOVA or PDERROR.');
      writeln;
      openSimFile(inputFlname,inputExt,simFile,succ);
      If (succ = true) Then Begin
         scanTree(tipCount,simFile);
         If (tipCount > maxTips) Then Begin
            writeln('Error -- This tree is too large; maximum size is ',maxTips);
            writeln('         Exiting...');
            halt;
         End; {then}
         openStatFile(inputFlname,statFile,statFileName)
      End; {then}

      writeln('Do you want to set an upper limit on the number of simulation');
      write('sets to analyze (Y/N = enter)? ');
      readln(userChar);
      writeln;
      If (upcase(userChar) = 'Y') Then Begin
         write('Enter the maximum number of simulation sets to analyze: ');
         readln(userLimit);
         writeln;
      End {then}
      Else userLimit := maxint;

      binCount := 0;
      clrscr;
      writeln('**************************************************************************');
      writeln('Do you want to preprocess the data by:');
      writeln;
      writeln('1  categorizing simulated tip values into equal-width bins starting at');
      writeln('   the minimum value and ending at the maximum value;');
      writeln('   the tip values will be replaced by their bin numbers.');
      writeln;
      writeln('2  ranking tip values in ascending order, with ties assigned');
      writeln('   an average value so that the sum of the ranks is always the same;');
      writeln('   the tip values will be replaced by their ranks.');
      writeln;
      writeln('3  No preprocessing (default if no choice number is entered)');
      writeln;
      Repeat
         write('Please enter 1, 2 or 3 (= enter): ');
         readln(userChar);
         If (userChar = chr(13)) Then userChar := '3';
         writeln;
      Until (userChar in ['1','2','3']);
      writeln;
      Case userChar of
         '1' : Begin
                  doCategories := true;
                  doRanks := false;
               End; {case 1}
         '2' : Begin
                  doCategories := false;
                  doRanks := true;
               End; {case 2}
         '3' : Begin
                  doCategories := false;
                  doRanks := false;
               End; {case 3}
      End; {case userChar}

      If doCategories Then Begin
         writeln('This option categorizes simulated tip values into equal-width bins');
         writeln('starting at the minimum value and ending at the maximum value.');
         writeln('The tip values are then replaced by their bin numbers.');
         writeln;
         restoreCrtAssignments;
         clrscr;
         gotoxy(1,1);
         textColor(LightGray+Blink);
         write('Scanning ',inputFLName,' for minimum and maximum values');
         scanForLimits(inputFLName,
                       globalMin1,
                       globalMax1,
                       globalMin2,
                       globalMax2,
                       tipCount,
                       userLimit);
         textColor(LightGray);
         clrscr;
         resetCrtAssignments;
         write('The maximum value of trait 1 is ');
         extendWrite(output,globalMax1);
         writeln;
         write('The minimum value of trait 1 is ');
         extendWrite(output,globalMin1);
         writeln;
         write('The maximum value of trait 2 is ');
         extendWrite(output,globalMax2);
         writeln;
         write('The minimum value of trait 2 is ');
         extendWrite(output,globalMin2);
         writeln;
         writeln;
         writeln('*****************************************************************');
         write('Do you want to specify limits (M)anually or (A)utomatically (M/A = enter)? ');
         readln(userChar);
         If (userChar in ['M','m']) Then Begin
            write('Enter the lower bound value for trait 1: ');
            readln(globalMin1);
            writeln;
            write('Enter the upper bound value for trait 1: ');
            readln(globalMax1);
            writeln;
            write('Enter the lower bound value for trait 2: ');
            readln(globalMin2);
            writeln;
            write('Enter the upper bound value for trait 2: ');
            readln(globalMax2);
            writeln;
         End
         Else Begin
            writeln('Bins for the two traits will be set from the minimum and maximum');
            writeln('values reported above.');
            writeln;
         End; {else}
         write('How many bins do you want? (0 cancels option): ');
         readln(binCount);
         writeln;
         If (binCount > 0) Then Begin
            writeln('Category numbers can be assigned starting from 0 or from 1.');
            writeln;
            writeln('Warning - If you start category numbers from 0 the Box-Cox transform');
            writeln('          will not work.  Also, PDANOVA frequently fails to work with ');
            writeln('          simulation files created by this program if categories are');
            writeln('          started from 0.');
            writeln;
            writeln('Would you like to number the bins (categories) from 0 to ',binCount-1:2);
            write('or 1 to ',binCount:2,'? (0/1 = enter): ');
            readln(userStr);
            writeln;
            If (pos('0',userStr) > 0)
            Then binBase := 0
            Else binBase := 1;
            showCategories := false;
         End; {then}
      End; { then begin }

      If (binBase <> 0) Then Begin
         clrscr;
         writeln('****************************************************************************');
         If doRanks Then Begin
            writeln('Do you want to preprocess the data with the Box-Cox transform following');
            write('ranking (Y/N = enter)? ');
         End
         Else If doCategories
              Then Begin
                 writeln('Do you want to preprocess the data with the Box-Cox transform following');
                 write('categorization (Y/N = enter)? ');
              End
              Else Begin
                 writeln('Do you want to preprocess the data with the Box-Cox transform');
                 write('(Y/N = enter)? ');
              End; {else}
         readln(userChar);
         If (userChar in ['Y','y'])
         Then doBCStats := true
         Else doBCStats := false;
      End
      Else Begin
         doBCStats := false;
         writeln('You will not be able to do Box-Cox transformations.  If you create a new file');
         writeln('of input data, use caution if you use PDANOVA.'); 
         writeln;
         write('Press enter to continue');
         readln;
      End; {Else}

      newSimExt(inputFlName,newExtStr,doBCStats,doCategories,doRanks);

      clrscr;

      If doBCStats
      Then Begin
         writeln('**************************************************************************');
         writeln('Do you want to write a file containing:');
         writeln;
         If doRanks Then Begin
            writeln('1  General statistics after ranking and Box-Cox transforms');
            writeln('        (default if no choice number is entered)');
         End {then}
         Else If doCategories
              Then Begin
                 writeln('1  General statistics after categorization and Box-Cox transforms');
                 writeln('        (default if no choice number is entered)');
              End
              Else writeln('1  General statistics (default if no choice number is entered)');
         writeln('        file .TPS');
         writeln;
         If doRanks Then Begin
            writeln('2  Skewness, kurtosis, normality test after ranking and before and ');
            writeln('        after Box-Cox transformation');
         End {then}
         Else If doCategories
              Then Begin
                 writeln('2  Skewness, kurtosis, normality test after categorization and ');
                 writeln('        before and after Box-Cox transformation');
              End
              Else writeln('2  Skewness, kurtosis, normality test before and after Box-Cox transformation');
         writeln('        file .BOX');
         writeln;
         writeln('3  Both files');
         writeln;
         Repeat
            write('Please enter 1 (= enter), 2, 3: ');
            readln(userChar);
            If (userChar = chr(13)) Then userChar := '1';
            writeln;
         Until (userChar in ['1','2','3']);
         Case userChar of
            '1' : Begin
                     doStats := true;
                     doBCFile := false;
                  End; {case 1}
            '2' : Begin
                     doBCFile := true;
                     doStats := false;
                  End; {case 2}
            '3' : Begin
                     doStats := true;
                     doBCFile := true;
                  End; {case 3}
         End; {case userChar}
      End {Then}
      Else Begin
         doBCFile := false;
         writeln('**************************************************************************');
         writeln('Do you want to write a file containing general statistics with ');
         write(' extension .TPS (Y = enter / N)? ');
         readln(userChar);
         writeln;
         If userChar in ['N','n']
         Then doStats := false
         Else doStats := true;
      End; {else}

      If doStats Then Begin
         writeln('Do you want to include medians and Kolmogorov-Smirnov normality tests in');
         write('in the .TPS file (Y/N = enter)? ');
         readln(userChar);
         writeln;
         If (upcase(userChar) = 'Y')
         Then doFullStats := true
         Else doFullStats := false;
      End {then}
      Else doFullStats := false;

      If (doBCStats Or doCategories Or doRanks) Then Begin
         userChar := 'n';
         newExtStr := '';
         newSimExt(inputFlName,newExtStr,doBCStats,doCategories,doRanks);
         If doBCStats Then Begin
            If doCategories Then Begin
               writeln('Would you like a modified input file (with extension ',newExtStr,')');
               writeln('containing the data after the categorization and Box-Cox ');
               write('transformations (Y/N = enter)? ');
               readln(userChar);
               writeln;
            End {then}
            Else Begin
               If doRanks Then Begin
                  writeln('Would you like a modified input file (with extension ',newExtStr,')');
                  writeln('containing the data after the ranking and Box-Cox ');
                  write('transformations (Y/N = enter)? ');
                  readln(userChar);
                  writeln;
               End
               Else Begin
                  writeln('Would you like a modified input file (with extension ',newExtStr,')');
                  write('containing the data after the Box-Cox transformation (Y/N = enter)? ');
                  readln(userChar);
                  writeln;
               End; {else}
            End; {else}
         End {then}
         Else Begin
            If doCategories Then Begin
               writeln('Would you like a modified input file (with extension ',newExtStr,')');
               write('containing the data after the categorization operation (Y/N = enter)? ');
               readln(userChar);
               writeln;
            End
            Else Begin
               If doRanks Then Begin
                  writeln('Would you like a modified input file (with extension ',newExtStr,')');
                  write('containing the data after the ranking operation (Y/N = enter)? ');
                  readln(userChar);
                  writeln;
               End
               Else userChar := 'N';
            End; {else}
         End;
         doNewSim := userChar in ['Y','y']
      End {then}
      Else doNewSim := false;

      If not (doNewSim Or doStats) Then Begin
         Writeln('You have requested neither a statistics file nor a modified input');
         writeln('file.  Are you sure you don''t want a statistics file? (Y / N = enter)? ');
         readln(userChar);
         If userChar in ['Y','y'] Then Begin
            writeln('PDTIPS has nothing to offer you with these options.  Exiting...');
            halt;
         End
         Else Begin
           writeln('Ok, PDTIPs will write a statistics file.');
           writeln;
           writeln('Press enter to continue');
           readln;
           doStats := true;
         End; {else}
      End; {then}

      If doNewSim Then Begin
         clrscr;
         writeln('**************************************************************************');
         writeln('Shall the modified input file include modified values for both traits?');
         writeln;
         writeln('1  write modified trait 1 tip values, unmodified trait 2 tip values');
         writeln;
         writeln('2  write modified trait 2 tip values, unmodified trait 1 tip values');
         writeln;
         writeln('3  write modified values for both traits (default if no choice is entered)');
         writeln;
         write('Enter 1, 2, or (3 = enter): ');
         readln(userStr);
         writeNewTrait1 := false;
         writeNewTrait2 := false;
         If length(userStr) = 0 Then userStr := '12';
         If (pos('1',userStr)>0) Then writeNewTrait1 := true;
         If (pos('2',userStr)>0) Then writeNewTrait2 := true;
         If (pos('3',userStr)>0) Then Begin
            writeNewTrait1 := true;
            writeNewTrait2 := true;
         End; {then}
      End; {then}

      If doNewSim Then Begin
         opennewSimFile(inputFlName,
                        newSimFile,
                        newSimFileName,
                        doBCStats,
                        doCategories,
                        doRanks);
         copySimHeader(simFile,newSimFile,inputFlName)
      End; {then begin}
      If doBCFile Then openboxFile(inputFlName,boxFile,boxFileName);

      readDataSet(simData,tipCount,simFile);
      loadTipArray(simData,1,rawTips1,tipCount);  {stash raw data}
      loadTipArray(simData,2,rawTips2,tipCount);

      rankData(simData,tipCount);
      If doCategories
      Then categorizeData(simData,
                          binCount,
                          binBase,
                          globalMin1,
                          globalMax1,
                          globalMin2,
                          globalMax2)
      Else If doRanks Then assignRanks(simData);

      i := 1;

      clrscr;
      foundBCError := false;
      Repeat
         if foundBCError
         Then writeln('Processing Set ',i)
         Else Begin
            gotoxy(30,12);
            write('Processing Set ',i);
         End; {else}
         If doStats Then Begin
            computeStatistics(simStats,simData);
            If doFullStats Then Begin
               computeFullStatistics(simStats,simData);
               writeFullStatistics(simStats,i,statFile);
            End {then}
            Else writeStatistics(simStats,i,statFile);
         End; {then}
         If doBCStats Then Begin
             computeBoxCoxStatistics(simStats,
                                bcStats,
                                transTips1,
                                transTips2,
                                simData);
             If doBCFile Then writeBCStats(simStats,bcStats,i,boxFile);
         End; {then}
         If doNewSim Then Begin
            If doBCStats
            Then Begin
               If writeNewTrait1
               Then If writeNewTrait2
                    Then writeQSet(newSimFile,transTips1,transTips2,simData)
                    Else writeQSet(newSimFile,transTips1,rawTips2,simData)
               Else If writeNewTrait2
                    Then writeQSet(newSimFile,rawTips1,transTips2,simData)
                    Else writeQSet(newSimFile,rawTips1,rawTips2,simData);  
            End
            Else Begin
               loadTipArray(simData,1,transTips1,tipCount);
               loadTipArray(simData,2,transTips2,tipCount);
               If writeNewTrait1
               Then If writeNewTrait2
                    Then writeQSet(newSimFile,transTips1,transTips2,simData)
                    Else writeQSet(newSimFile,transTips1,rawTips2,simData)
               Else If writeNewTrait2
                    Then writeQSet(newSimFile,rawTips1,transTips2,simData)
                    Else writeQSet(newSImFile,rawTips1,rawTips2,simData);
            End; {else}
         End; {then}

         killData(simData);
         inc(i);
         readDataSet(simData,tipCount,simFile);
         If (simData <> nil) Then Begin

            loadTipArray(simData,1,rawTips1,tipCount);  {stash raw data}
            loadTipArray(simData,2,rawTips2,tipCount);

            rankData(simData,tipCount);
            If doCategories
            Then categorizeData(simData,
                                binCount,
                                binBase,
                                globalMin1,
                                globalMax1,
                                globalMin2,
                                globalMax2)
            Else If doRanks Then assignRanks(simData);
         End; {then}
      Until ((simData = nil) or (i > userLimit));

      sound(880);
      delay(750);
      nosound;

      close(simFile);
      close(statFile);
      If doBCStats Then close(boxFile);
      If doNewSim Then close(newSimFile);

      clrscr;
      writeln('**************************************************************************');
      writeln;
      If doStats
      Then writeln('The name of the statistics file is ',StatFileName,'.');
      writeln;
      If doBCStats
      Then writeln('The name of the Box-Cox statistics file is ',boxFileName,'.');
      writeln;
      If doNewSim
      Then writeln('The name of the transformed SIM file is ',newSimFileName,'.');
      writeln;
      If (doStats Or doBCStats) Then Begin
         write('Do you want to see a list of output file columns (Y/N = enter)? ');
         readln(userChar);
         writeln;
         If (upcase(userChar) = 'Y')
         Then displaySummary(doStats,statFileName,doFullStats,doBCStats,boxFileName);
      End; {then}
   End. {program pdtips}

