{ RANDIO.PAS                                                           }
{ This unit contains functions for opening files, writing distance and }
{ probability data and stubs for reading trees (as well as re-reading  }
{ a tree without prompting for the filename again.                     }
{                                                                      }
{ It uses the following units:                                         }
{                                                                      }
{ RANDUTIL - Utility functions and type definitions for trees and      }
{            matrices.                                                 }
{ RAND_IO1 - The functions for reading and writing trees.              }
{ CRT - Provides clrscr for clearing the screen.                       }
{ DOS - Provides some types and functions for working with directory   } 
{       and file names.                                                }
{**********************************************************************}
{ Modification History:                                                }
{                                                                      }
{   13-Mar-96: Added info to headers of distance and probability files.}
{              PEM                                                     }
{                                                                      }
{   25 Aug 97: Added capability to read matrix files.  PEM             }
{                                                                      }
{    4 Dec 97: Changes due to addition of options record.  PEM         }
{                                                                      }
{                                                                      }
{   15 Dec 97: Removed verbage about exclusions to be listed in a      }
{              later version (since this will almost certainly never   }
{              happen.  PEM                                            }
{                                                                      }
{ Last Modified: 15 Dec 1997 P. E. Midford                             }


Unit RANDIO;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {small stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Interface

Uses randutil;


Procedure readTree(data: phyloPtr;
                   Var succ:boolean);

Procedure reReadTree(data: phyloPtr;
                     Var succ: boolean);

Procedure openDistanceFile(inputName :flstring;
                           Var distanceFile : text;
                           Var distanceFileName : flString);

Procedure writeMatFileHeader(perms        : permArrayType;
                             outFileName  : flstring;
                             Var outFile  : text;
                             tipCount     : integer;
                             treeFileName : flstring;
                             userOptions  : options;
                             includeSums  : boolean;
                             matrixData   : boolean);



Procedure writeDistanceData(perms            : permArrayType;
                            index            : integer;
                            tipCount         : integer;
                            Var distanceFile : text;
                            userOptions      : options);

Procedure openProbFile(inputName :flstring;
                       Var probFile : text;
                       Var probFileName : flString);

Procedure openRProbFile(inputName :flstring;
                        Var rProbFile : text;
                        Var rProbFileName : flString);

Procedure writeProbData(perms        : permArrayType;
                        index        : integer;
                        tipCount     : integer;
                        Var probFile : text;
                        userOptions  : options);

Procedure writeColumnSums(perms        : permArrayType; { needed for column sums}
                          tipCount     : integer;
                          Var probFile : text;
                          userOptions  : options);

Procedure linkTree(rt : nodeptr);

Procedure openPermFile(inputName    : flString;
                       Var permFile : text;
                       Var tstFileName : flString);

Procedure writePermFileHeader(data1,data2         : phyloPtr;
                              Var permFile        : text;
                              tipCount            : integer;
                              userOptions         : options;
                              Var diskBufferNum   : longint;
                              Var diskBufferMax   : longint;
                              Var diskBufferPtr   : datptr;
                              Var diskBufferStart : datptr);

Procedure writeTrees(Var permFile        : text;
                     tree1,tree2         : nodePtr;
                     tipCount            : integer;
                     done                : integer;
                     setNum              : longint;
                     var diskbuffernum   : longint;
                     var diskbuffermax   : longint;
                     var diskbufferptr   : datptr;
                     var diskbufferstart : datptr);

Procedure writeTips(Var permFile     : text;
                    permArray1       : permArrayType;
                    permArray2       : permArrayType;
                    tipCount         : integer;
                    done             : integer;
                    setNum           : longint;
                    Var diskBufNum   : longint;
                    Var diskBufMax   : longint;
                    Var diskBufPtr   : datptr;
                    Var diskBufStart : datptr);

Procedure freeHeap(Var ptr : datPtr);

Procedure openTestFile(testFileName : flstring;Var testFile : text);

Procedure writeTestFileHeader(testFileName : flString;
                              tree         : nodeptr;
                              Var testFile : text);

Procedure writeTestMatrixTitles(perms       : permArrayType;
                                outFileName : flstring;
                                Var outFile : text;
                                tipCount    : integer);

Implementation

uses rand_io1,crt,dos;

{***************************************************************************}
{ This procedure is basically a stub to the pdtree procedure inputTree. It  }
{ prompts the user for the name of the file, which is returned in flname.   }
{***************************************************************************}
Procedure readTree(data     : phyloPtr;
                   Var succ : boolean);
   Begin
      inputTree(data,succ,false); { reread is false PEM 3 May 96 }
   End; { procedure readTree}

{***************************************************************************}
{ This procedure is basically a stub to the pdtree procedure inputTree.     }
{ It re-reads the file specified by flname.                                 }
{***************************************************************************}
Procedure reReadTree(data     : phyloPtr;
                     Var succ : boolean);
   Begin
      inputTree(data,succ,true); { reread is true here PEM 3 May 96 }
   End;


{**************************************************************************}
{ This procedure is like extendWrite except it prints more digits.         }
{**************************************************************************}
Procedure longWrite(var outFile : text;num:extended);
   Var outStr : string[25];
   Begin
       str(num:25:-17,outStr);
       delete(outStr,22,2);
       write(outFile,outStr);
   End; {procedure longWrite);

{**************************************************************************}
{ This procedure allows the user to change the name of the output file.    }
{**************************************************************************}
Procedure getfilename(Var name2 : nameStr);
   Var
     fl : file;
     good,again : boolean;
     ch : char;
     dir : dirStr;
     nm : nameStr;
     ext : extStr;
     name : flString;
   Begin
      clrscr;
      write('The current base for the randomized trees filename is: ');
      writeln('"',name2,'".');
      writeln('To this base is added the extension ".rnd".');
      writeln;
      writeln('Would you like to change the base for the permutation ',
                                         'results (Y/N)?');
      writeln;
      writeln('NOTE:  the C.M.A.P. programs of Martins and Garland (1991) only');
      writeln('       allow 5 characters in the root name.');
      good := true;
      ch := Upcase(readkey);
      If ch='Y' Then Repeat
         writeln;
         If (good = false)
         Then writeln('The file is write-locked or invalid.');
         writeln('Please enter the 8 or fewer digit file name.');
         writeln;
         readln(name2);
         name:=name2+'.rnd';
         good:=true;
         again:=false;
         {$I-}              {Turn off I/O checking momentarily}
         assign(fl,name);   {try to open the file}
         reset(fl);         {if an error occurs, set IOresult}
         close(fl);         {close the file}
         {$I+}              {Turn I/O checking back on}
         If (IOresult=0) And (name<>'') Then Begin
            writeln;
            writeln(name,' already exits.');
            writeln('Do you want to delete it and ',
                    'save a new set of randomized trees in its place? (Y/N)');
            Repeat
               ch:=upCase(readkey);
            Until ch in['Y','N'];
            If ch='N'
            Then again := true
            Else Begin
               {$I-}              {Turn off I/O checking momentarily}
               assign(fl,name);   {try to open the file for output}
               rewrite(fl);       {if an error occurs, set IOresult}
               close(fl);         {close the file}
               {$I+}              {Turn I/O checking back on}
               If (IOresult<>0) Or (name='') Then good := false;
            End; {else}
         End {then}
         Else Begin
            {$I-}              {Turn off I/O checking momentarily}
            assign(fl,name);   {try to open the file for output}
            rewrite(fl);       {if an error occurs, set IOresult}
            close(fl);         {close the file}
            {$I+}              {Turn I/O checking back on}
            If (IOresult<>0) Or (name='') Then good := false;
         End {else}
      Until good and not again;
      clrscr;
   End; {procedure getFileName}

{***************************************************************************}
{ This procedure opens the .DST file to which the matrix of distances is    }
{ written.                                                                  }
{***************************************************************************}
Procedure openDistanceFile(inputName            : flstring;
                           Var distanceFile     : text;
                           Var distanceFileName : flString);
   Var
      place : integer;
   Begin
      place := pos('.',inputName);
      distanceFileName := copy(inputName,1,place-1);
      distanceFileName := distanceFileName + '.DST';
      assign(distanceFile,distanceFileName);
      rewrite(distanceFile);
      writeln('Writing Distance data to: ',distanceFileName)
   End; {opendistancefile}

{***************************************************************************}
{This procedure writes the header information for the various output files. }
{***************************************************************************}
Procedure writeMatFileHeader(perms        : permArrayType;
                             outFileName  : flstring;
                             Var outFile  : text;
                             tipCount     : integer;
                             treeFileName : flstring;
                             userOptions  : options;
                             includeSums  : boolean;
                             matrixData   : boolean);
   Var
      i : integer;
      name : string;
      year, month, day, dayofWeek, hour, minute, second, sec100 : word;
   Begin
      getProgramName(name);
      writeln(outfile,'File = ',outFileName);
      writeln(outfile);
      writeln(outfile,'Output from ',name,': ',outFileName);
      write(outFile,'This file written: ');
      getDate(Year, Month, Day, DayofWeek);
      write(outFile,' ',Day,' ');
      Case month of
         1 : Write(outFile, 'Jan ');
         2 : Write(outFile, 'Feb ');
         3 : Write(outFile, 'Mar ');
         4 : Write(outFile, 'Apr ');
         5 : Write(outFile, 'May ');
         6 : Write(outFile, 'Jun ');
         7 : Write(outFile, 'Jul ');
         8 : Write(outFile, 'Aug ');
         9 : Write(outFile, 'Sep ');
        10 : Write(outFile, 'Oct ');
        11 : Write(outFile, 'Nov ');
        12 : Write(outFile, 'Dec ');
      End; {case month}
      write(outFile, Year,' at ');
      getTime(Hour, Minute, Second, Sec100);
      write(outFile,Hour,':');
      If (minute < 10)
      Then write(outFile,'0',Minute,':')
      Else write(outFile,Minute,':');
      If (second < 10)
      Then writeln(outFile,'0',Second,'.')
      Else writeln(outFile,Second,'.');
      writeln(outFile,'The tree was read from ',treeFileName);
      If (matrixData)
      Then
         If (name = 'PDRANDOM')
         Then write(outFile,'Random seed is: ',userOptions.Seed)
         Else write(outFile,'No randomization used in this program')
      Else
         If (userOptions.myMethod = similarity)
         Then Begin
              If (name = 'PDRANDOM')
              Then write(outFile,'Random seed is: ',userOptions.Seed, '. K is : ')
              Else write(outFile,'K is: ');
              extendWrite(outFile,userOptions.k);
         End
         Else Begin
            If (name = 'PDRANDOM')
            Then write(outfile,'Random seed is: ',userOptions.Seed, '.  Self distance scaling factor is: ')
            Else write(outfile,'Self distance scaling factor is: ');
            extendWrite(outFile,userOptions.factor);
         End; {else}
      writeln(outFile,'.');
      If (matrixData)
      Then Begin
         writeln(outFile,'No transformations were performed on matrix data.');
         writeln(outFile,'Self permutation probabilities were not altered in matrix data.');
      End
      Else Begin
         If userOptions.logTrans
         Then writeln(outFile,'A log10 transform of branch length sums is currently specified.')
         Else writeln(outFile,'Branch length sums are transformed by being raised to the ',
                               userOptions.blPower:4:2,' power.');
      End; {else}
      writeln(outfile);
      write(outFile,' Destin.   ');
      For i := 1 to tipCount Do Begin
         write(outFile,perms[i]^.tip^.name);
         {$IFOPT N+}
         If userOptions.showExtended
         Then write(outFile,'                      ')
         Else write(outFile,'                ');
         {$ELSE}
         write(outFile,'                ');
         {$ENDIF}
      End;
      If includeSums
      Then writeln(outFile,'Row Sum')
      Else writeln(outFile);
      writeln(outFile,'Source     ');
   End; {procedure writematrixtitles}

{***************************************************************************}
{ This procedure writes a single line of distance data preceeded by the     }
{ name of the source tip.                                                   }
{***************************************************************************}
Procedure writeDistanceData(perms : permArrayType;
                            index : integer;
                            tipCount : integer;
                            Var distanceFile : text;
                            userOptions : options);
   Var
      i : integer;
   Begin
      write(distanceFile,perms[index]^.tip^.name,' ');
      For i := 1 to tipCount Do Begin
         If userOptions.showExtended
         Then longWrite(distanceFile,perms[i]^.distance)
         Else extendWrite(distanceFile,perms[i]^.distance);
         write(distanceFile,' ');
      End;
      writeln(distanceFile);
   End; { procedure writedistancedata}

{***************************************************************************}
{ This procedure opens the .PRB file to which the matrix of permutation     }
{ probabilities is written.                                                 }                 
{***************************************************************************}
Procedure openProbFile(inputName :flstring;
                       var probFile : text;
                       var probFileName : flString);
   Var
      place : integer;
   Begin
      place := pos('.',inputName);
      probFileName := copy(inputName,1,place-1);
      probFileName := probFileName + '.PRB';
      Assign(probFile,probFileName);
      Rewrite(probFile);
      writeln('Writing probability data to: ',probFileName)
   End; {procedure openprobfile}

{***************************************************************************}
{ This procedure opens the .RPB file to which the matrix of realized        }
{ permutation probabilities is written.                                     }            
{***************************************************************************}
Procedure openRProbFile(inputName :flstring;
                        var rProbFile : text;
                        var rProbFileName : flString);
   Var
      place : integer;
   Begin
      place := pos('.',inputName);
      rProbFileName := copy(inputName,1,place-1);
      rProbFileName := rProbFileName + '.RPB';
      Assign(rProbFile,rProbFileName);
      Rewrite(rProbFile);
      writeln('Writing probability data to: ',rProbFileName)
   End; {procedure openrprobfile}

{***************************************************************************}
{ This procedure writes a single line of probability data preceeded by the  }
{ name of the source tip.                                                   }
{***************************************************************************}
Procedure writeProbData(perms : permArrayType;
                        index : integer;
                        tipCount : integer;
                        var probFile : text;
                        userOptions : options);
   Var
      i : integer;
   Begin
      write(probFile,perms[index]^.tip^.name,' ');
      For i := 1 to tipCount Do Begin
         If userOptions.showExtended
         Then longWrite(probFile,perms[i]^.probability)
         Else extendWrite(probFile,perms[i]^.probability);
         write(probFile,' ');
      End; {for i}
      If userOptions.showExtended
      Then longwrite(probFile,perms[index]^.rowSum)
      Else extendWrite(probFile,perms[index]^.rowSum);
      writeln(probFile);
   End;
{***************************************************************************}
{ This procedure writes the row of column sums below the matrix (of any     }
{ variety).                                                                 }
{***************************************************************************}
Procedure writeColumnSums(perms        : permArrayType; { needed for column sums}
                          tipCount     : integer;
                          Var probFile : text;
                          userOptions  : options);
   Var
      column : integer;
   Begin
      writeln(probFile);
      write(probFile,'Sum');
      For column := 1 to tipCount Do Begin
         If userOptions.showExtended
         Then longWrite(probFile,perms[column]^.columnSum)
         Else extendWrite(probFile,perms[column]^.columnSum);
         write(probFile,' ');
      End; {for i}
      writeln(probFile);
   End; { procedure writeColumnSums}

{**************************************************************************}
{ This is a debugging procedure can be used to check that the next         }
{ procedure linked up the tree correctly.                                  }
{**************************************************************************}
Procedure showLinks(nd : nodeptr);
   Var
      ch:char;
   Begin
      writeln(nd^.name,' ');
      ch:=readkey;
      If nd^.tipL<>NIL Then showlinks(nd^.tipL);
   End; {procedure showlinks}

{**************************************************************************}
{This procedure links up the nodes in the tree so s_simul1 has an easier
time traversing the tree to write out the results.  The format is a left
first traversal for both the tips and the nodes, but they must alternate.
So I used the tipL and tipR fields in the tree that are not used by this
program for any other purpose to do it.  rt^.tipR points to the first tip.
then the tipL of that tip points to the first node whose tipL points to the
next tip and so on until nil is reached.  Again, this is just done so that
when the values of the nodes and tips must be written to the file it is a
quick traversal rather than a longer more complex one.}
{**************************************************************************}
Procedure linkTree(rt : nodeptr);
   Var tip, node, oldnode :nodeptr;
       ch :char;
   {***********************************************************************}
   { This procedure clears the tempx and tempy marks from the tree so the  }
   { next two procedures can keep track of which nodes they have visited   }
   { with them.                                                            }
   {***********************************************************************}
   Procedure clearMarks(nd: nodeptr);
      Begin
         nd^.tempx:=0;
         nd^.tempy:=0;
         If nd^.sibL<>NIL Then Begin
            clearmarks(nd^.sibL);
            clearmarks(nd^.sibR);
         End; {then}
      End; {procedure clearmarks}

   {***********************************************************************}
   { This procedure returns a pointer to the next tip in a left first      }
   { traversal of the tree.  It requires tempy to be cleared for it before }
   { it is called the first time.  The first time it is called, it needs   }
   { to be passed rt.  The second time it is called, it needs to be passed }
   { the results it passed back last (the last tip returned).  It then can }
   { find the next by looking at the markings it left in tempy.            }
   {***********************************************************************}
   Procedure getTip(rt: nodeptr; Var tip:nodeptr);
      Begin
         If tip^.tempy=0 Then Begin
            tip^.tempy:=1;
            If tip^.sibL<>nil Then Begin
               tip:=tip^.sibL;
               getTip(rt,tip)
            End {then}
         End {then}
         Else Begin
            If tip^.sibL=NIL Then Begin
               tip:=tip^.ancestor;
               getTip(rt,tip);
            End {then}
            Else If tip^.sibR^.tempy=0
                 Then Begin
                    tip:=tip^.sibR;
                    gettip(rt,tip);
                 End {then}
                 Else If rt<>tip
                      Then Begin
                         tip:=tip^.ancestor;
                         getTip(rt,tip)
                      End {then}
                 Else tip:=nil;
         End; {else}
      End; {procedure getTip}

   {***********************************************************************}
   { This procedure has the same function and behavior as the last one     }
   { except that it works with nodes not tips.                             }
   {***********************************************************************}
   Procedure getNode(Var node:nodePtr);
      Begin
         If (node^.tempx=0) Then node^.tempx:=1
         Else Begin
            If (node^.sibL^.sibL<>nil) And (node^.sibL^.tempx=0)
            Then Begin
               node:=node^.sibL;
               getnode(node)
            End {then}
            Else If (node^.sibR^.sibL<>nil) and (node^.sibR^.tempx=0)
                 Then Begin
                    node:=node^.sibR;
                    getNode(node)
                 End {then}
                 Else If node^.ancestor<>nil
                      Then Begin
                         node:=node^.ancestor;
                         getNode(node)
                      End {Then}
         End {else}
      End; {procedure getnode}
   {***********************************************************************}
   Begin {procedure linkTree}
      tip:=rt;
      node:=rt;
      clearmarks(rt);
      gettip(rt,tip);
      rt^.tipR:=tip;
      getnode(node);
      rt^.tipR^.tipL:=node;
      oldnode:=node;
      Repeat
         gettip(rt,tip);
         getnode(node);
         oldnode^.tipL:=tip;
         If node<>rt Then tip^.tipL:=node
         Else tip^.tipL:=nil;
         oldnode:=node;
      Until tip^.tipL=nil;
   End; {procedure linkTree}

{***************************************************************************}
{ This procedure opens the file that will receive the trees after the tips  }
{ have been permuted.  Note the extension is set to 'RND'.                  }
{***************************************************************************}
Procedure openPermFile(inputName    : flString;
                       Var permFile : text;
                       Var tstFileName : flString);

   Var dir   : dirStr;
       name  : nameStr;
       ext   : extStr;
   Begin
      fsplit(inputName,dir,name,ext);
      getfilename(name);
      tstFileName := dir+name+'.tst';    
      assign(permFile,dir+name+'.rnd');
      rewrite(permFile);
   End;

{**************************************************************************}
{ This procedure initializes the buffer by taking all the available space  }
{ on the heap and allocating it to the disk buffer.  Again, note that the  }
{ disk buffer is a linked list.                                            }
{**************************************************************************}
Procedure fillheap(var num:longint; var ptr:datptr);
   Var
      last :datptr;
   Begin
      new(ptr);
      last:=ptr;
      num:=1;
      Repeat
         new(last^.next); {will set last^.next to NIL when heap's exhusted}
         last:=last^.next;
         If last <> NIL Then num:=num+1;
      Until last=nil;
   End; {procedure fillheap}

{**************************************************************************}
{ This procedure puts the tree structure into the output file.             }
{**************************************************************************}
Procedure dumptree(Var outFile : text; nd1,rt1,nd2,rt2 : nodePtr);
   Var
      number : integer;
   Begin
      If nd1^.sibL<>NIL Then Begin { write data from a node on the first tree } 
         write(outFile,'  ');
         write(outFile,nd1^.name,' ');
         write(outFile,nd1^.sibL^.name,' ');
         extendWrite(outFile,nd1^.sibL^.length);
         write(outFile,' ');
         number:=0;
         numtps(nd1^.sibL,number); 
         write(outFile,' ',number:2,' ');
         write(outFile,nd1^.sibR^.name,' ');
         extendWrite(outFile,nd1^.sibR^.length);
         write(outFile,' ');
         number:=0;
         numtps(nd1^.sibR,number); 
         write(outFile,' ',number:2,' ');
         If nd1<>rt1 Then Begin
            write(outFile,nd1^.ancestor^.name,' ');
            extendWrite(outFile,nd1^.length);
         End {then}
         Else write(outFile,'**  0.0000000000E+00');
         { Now the same for the second tree - on the same line }
         write(outFile,'  ');
         write(outFile,nd2^.name,' ');
         write(outFile,nd2^.sibL^.name,' ');
         extendWrite(outFile,nd2^.sibL^.length);
         write(outFile,' ');
         number:=0;                        {these lines may be unnecessary}
         numtps(nd2^.sibL,number); 
         write(outfile,' ',number:2,' ');
         write(outfile,nd2^.sibR^.name,' ');
         extendWrite(outfile,nd2^.sibR^.length);
         write(outFile,' ');
         number:=0;
         numtps(nd2^.sibR,number); 
         write(outfile,' ',number:2,' ');
         If nd2<>rt2 Then Begin
            write(outfile,nd2^.ancestor^.name,' ');
            extendWrite(outFile,nd2^.length);
            writeln(outFile);
         End {then}
         Else writeln(outfile,'**  0.0000000000E+00');
         dumpTree(outfile,nd1^.sibL,rt1,nd2^.sibL,rt2);
         dumpTree(outfile,nd1^.sibR,rt1,nd2^.sibR,rt2);
      End; {then}
   End; {procedure dumptree}

{**************************************************************************}
{ This procedure writes a header for the permutation results into the      }
{ results (.RND) file.  The purpose of the this header is to resemble the  }
{ output of PDSIMUL so that programs like PDANOVA could use these results. }
{**************************************************************************}
Procedure dumpDescription(Var outfile : text;
                          userOptions : options;
                          matrixFlag  : boolean);
   Begin
      writeln(outfile,'SIMULATED (Note: this line inserted for compatibilty with PDANOVA and friends');
      writeln(outfile,'RANDOMIZATION PARAMETERS: ');
      writeln(outfile);
      If matrixFlag
      Then writeln(outFile,'Random seed is: ',userOptions.Seed, '.')
      Else Begin
         If (userOptions.myMethod = similarity)
         Then Begin
            write(outfile,'Random seed is: ',userOptions.Seed, '.  K is: ');
            extendWrite(outFile,userOptions.k)
         End {then}
         Else Begin
            write(outfile,'Random seed is: ',userOptions.Seed, '.  Self distance scaling factor is: ');
            extendWrite(outfile,userOptions.factor)
         End; {else}
         writeln(outFile,'.');
      End; {else}
      If matrixFlag
      Then Begin
         writeln(outFile,'No transformation was applied to matrix data');
         writeln(outFile);
      End {then}
      Else If userOptions.logTrans
           Then writeln(outFile,'A log10 transform of branch length sums is currently specified.')
           Else writeln(outFile,'Branch length sums are transformed by being raised to the ',
                                 userOptions.blPower:4:2,' power.');
      If matrixFlag
      Then Begin
         writeln(outFile,'Self permutations in matrix data were not modified.');
         writeln(outFile);
      End {then}
      Else Begin
         If userOptions.exclude
         Then Begin
            writeln(outfile,'Permutations from a tip to itself were excluded (except original tree).');
            writeln(outfile);
         End {then}
         Else Begin
            writeln(outfile,'Permutations from a tip to itself were allowed.');
            write(outFile,'Self probabilities were not changed');
            writeln(outFile);
         End; {else}
      End; {else}
      writeln(outFile);
      writeln(outFile);
      writeln(outfile);
      writeln(outfile);
      writeln(outFile);
      writeln(outfile);
{      writeln(outfile); } {excess line?}
   End; {procedure dumpdescription}

{***************************************************************************}
{ This procedure writes the main header for the results (.RND) file.  It    }
{ includes a call to the dumpdescription above so a header for PDANOVA gets }
{ written as well.                                                          }
{***************************************************************************}
Procedure writePermFileHeader(data1,data2         : phyloPtr;
                              Var permFile        : text;
                              tipCount            : integer;
                              userOptions         : options;
                              Var diskBuffernum   : longint;
                              Var diskBuffermax   : longint;
                              Var diskBufferptr   : datptr;
                              Var diskBufferstart : datptr);
   Var
      year, month, day, dayofWeek, hour, minute, second, sec100 : word;
      row,column : integer;
      matrixFlag : boolean;
   Begin
      matrixFlag := data1^.tree=nil;
      writeln(permFile, 'Output from PDRANDOM: randomly permuted tip values.');
      If (not matrixFlag)
      Then Begin
         write(permFile, 'Data from ', data1^.sourceName,' (tree1) and ');
         write(permFile, data2^.sourceName, ' (tree2)');
      End {then}
      Else Begin
         Write(permFile, 'Data from ', data1^.sourceName,' (prb matrix1) and ');
         write(permFile, data2^.sourceName, ' (prb matrix2)');
      End; {else}
      getDate(Year, Month, Day, DayofWeek);
      write(permfile,'; This file written ',Day,' ');
      Case Month of
         1 : Write(permFile, 'Jan ');
         2 : Write(permFile, 'Feb ');
         3 : Write(permFile, 'Mar ');
         4 : Write(permFile, 'Apr ');
         5 : Write(permFile, 'May ');
         6 : Write(permFile, 'Jun ');
         7 : Write(permFile, 'Jul ');
         8 : Write(permFile, 'Aug ');
         9 : Write(permFile, 'Sep ');
        10 : Write(permFile, 'Oct ');
        11 : Write(permFile, 'Nov ');
        12 : Write(permFile, 'Dec ');
      End; {case month}
      write(permFile, Year,' at ');
      getTime(Hour, Minute, Second, Sec100);
      write(permFile,Hour,':');
      If (minute < 10)
      Then write(permFile,'0',Minute,':')
      Else write(permFile,Minute,':');
      If (second < 10)
      Then writeln(permFile,'0',Second,'.')
      Else writeln(permFile,Second,'.');
      writeln(permFile);
      If (Not matrixFlag) Then Begin
         writeln(permFile,'    Number of Tips  =  ',tipCount);
         write(permFile,'Phylogenetic tree from input files: ',data1^.sourceName);
         writeln(permFile,' and ', data2^.sourceName);
         writeln(permFile,'(Node name- Left child, Right child, Ancestor: ',
                      'name, branch-length, number descendants.)',
                      '(Node name- Left child, Right child, Ancestor: ',
                      'name, branch-length, number descendants.)');
         dumptree(permFile,data1^.tree,data1^.tree,data2^.tree,data2^.tree);
      End {then}
      Else Begin
         writeln(permFile,' Array Dimension = ',tipCount,' x ',tipCount);
         writeln(permFile);
         write(permFile,'    ');
         For column := 1 to tipCount Do Begin
            write(permFile,data1^.permArray[column]^.tip^.name);
            write(permFile,'               ');
         End;
         writeln(permFile,'Row Sum');
         writeln(permFile,'Source     ');
         For column := 1 to tipCount Do
            data1^.permArray[column]^.columnSum := 0;

         For row := 1 to tipCount Do Begin
            For column := 1 to tipCount Do Begin
               data1^.permArray[column]^.probability :=
                  data1^.matrix^.data[row,column];
               data1^.permArray[column]^.columnSum :=
                  data1^.permArray[column]^.columnSum + data1^.matrix^.data[row,column];
            End; {for column}
            writeProbData(data1^.permArray,row,tipCount,permFile,userOptions);
         End; {for i}
         writeColumnSums(data1^.permArray,tipCount,permFile,userOptions);  { needed for column sums}
      End; {else}
      dumpdescription(permFile,userOptions,matrixFlag);
      If (data1^.tree <> nil) Then 
         writeln(permFile,'                  Tips',
                         '                               Nodes',
                         '                     Set');
      If (data1^.tree <> nil) Then
         writeln(permFile,'       Trait 1           Trait 2',
                     '              Trait 1           Trait 2',
                     '       Number')
      Else writeln(permFile,'       Trait 1           Trait 2',
                     '       Number');

      If (data1^.tree <> nil) Then
         writeln(permFile,'_________________________________________',
                         '___________________________________________')
      Else writeln(permFile,'_______________________________________________');

      diskbuffernum:=0;
      diskbuffermax:=0;
      fillheap(diskbuffermax,diskbufferptr);
      diskbufferstart:=diskbufferptr;
   End; {procedure writepermfileheader}

{**************************************************************************}
{ Here we buffer trait data from the tree into a previously allocated list }
{ of structures.  The program will continue to fill members of the list    }
{ until either there is not enough space left for a complete set of tips   }
{ or the list data set has been copied to the list.  See dumpresults for   }
{ details of what happens next.  Diskbufptr points to the next available   }
{ list member.  Diskbufnum counts one more than the number of filled       }
{ elements in the list.  This procedure copies the data from nd1 and nd2   }
{ to the structure addressed by diskbufptr and updates diskbufptr and      }
{ diskbufnum.                                                              }
{**************************************************************************}
Procedure puttrees(nd1,rt1         : nodePtr;
                   nd2,rt2         : nodePtr;
                   Var diskBufPtr  : datPtr;
                   Var diskBufNum  : longint;
                   setNum          : longint);
   Var
      ch :char;
   Begin
      diskBufNum:=diskBufNum+1;
      diskBufPtr^.name1:=nd1^.name;
      diskBufPtr^.dat1[1]:=nd1^.dat[1];
      diskBufPtr^.dat1[2]:=nd2^.dat[2];
      nd1:=nd1^.tipL;
      nd2:=nd2^.tipL;
      If nd1<>nil Then Begin
         diskBufPtr^.name2:=nd1^.name;
         diskBufPtr^.dat2[1]:=nd1^.dat[1];
         diskBufPtr^.dat2[2]:=nd2^.dat[2];
         diskBufPtr^.setCount := setNum;
         diskBufPtr:=diskBufPtr^.next;
         putTrees(nd1^.tipL,rt1,nd2^.tipL,rt2,diskBufptr,diskBufNum,setNum)
      End {then}
      Else Begin  {no more nodes to write, so write a dummy}
         diskBufPtr^.name2:='**';
         diskBufPtr^.dat2[1]:= 0.0;
         diskBufPtr^.dat2[2]:= 0.0;
         diskBufPtr^.setCount := setNum;
         diskBufPtr:=diskBufPtr^.next;
      End; {else}
   End; {procedure puttrees}

{**************************************************************************}
{ This procedure is in charge of taking the tree and extracting the        }
{ results.  If there is room in the buffer for the data from this          }
{ permutation and one more then the data is just put in the buffer.  Else  }
{ it is put in the buffer and the buffer is dumped to disk and reset.      }
{**************************************************************************}
Procedure dumpResults(Var outFile    : text;
                      rt1            : nodePtr;
                      rt2            : nodePtr;
                      tipnum         : integer;
                      now            : integer;
                      done           : integer;
                      setNum         : longint;
                      Var diskBufNum : longint;
                      Var diskBufPtr : datptr;
                      diskBufMax     : longint;
                      diskBufStart   : datPtr);
   Var
      tip,node  : nodePtr;
      i,j,where : integer;
      current   : datPtr;
   Begin
      putTrees(rt1^.tipR,rt1,rt2^.tipR,rt2,diskbufptr,diskbufnum,setNum);
      If (now=done) Or ((diskbufnum + tipnum) > diskbufmax) Then Begin
         gotoxy(20,14);
         write('Writing results to file.');   {announce that we are writing}
         current:=diskbufstart;
         For i:=1 To diskBufNum Do Begin
            write(outfile,current^.name1);
            write(outfile,' ');
            extendWrite(outfile,current^.dat1[1]);
            write(outfile,' ');
            extendWrite(outfile,current^.dat1[2]);
            write(outfile,' ');
            write(outfile,current^.name2);
            write(outfile,' ');
            extendWrite(outfile,current^.dat2[1]);
            write(outfile,' ');
            extendWrite(outfile,current^.dat2[2]);
            write(outfile,' ');
            writeln(outFile,current^.setCount);
            current:=current^.next;
         End; {for i}
         diskbufnum:=0;
         diskbufptr:=diskbufstart;
         gotoxy(20,14);                         {clear the screen message}
         clreol;
      End; {then}
   End; {procedure dumpResults}

{***************************************************************************}
{ This procedure is currently just a hook into dumpResults...               }
{***************************************************************************}
Procedure writeTrees(Var permFile        : text;
                     tree1,tree2         : nodePtr;
                     tipCount            : integer;
                     done                : integer;
                     setNum              : longint;
                     var diskbuffernum   : longint;
                     var diskbuffermax   : longint;
                     var diskbufferptr   : datptr;
                     var diskbufferstart : datptr);
   Begin
      dumpresults(permFile,
                  tree1,
                  tree2,
                  tipCount,
                  1,
                  done,
                  setNum,
                  diskbuffernum,
                  diskbufferptr,
                  diskbuffermax,
                  diskbufferstart);
   End;

{***************************************************************************}
{ This procedure both buffers and writes tip values (which are the          }
{ permutation results.                                                      }
{***************************************************************************}
Procedure writeTips(Var permFile     : text;
                    permArray1       : permArrayType;
                    permArray2       : permArrayType;
                    tipCount         : integer;
                    done             : integer;
                    setNum           : longint;
                    Var diskBufNum   : longint;
                    Var diskBufMax   : longint;
                    Var diskBufPtr   : datptr;
                    Var diskbufstart : datptr);
   Var ch : char;
       i : integer;
       current : datptr;
   Begin
      For i := 1 To tipCount Do Begin
         diskBufPtr^.name1:=permArray1[i]^.tip^.name;
         diskBufPtr^.dat1[1]:=permArray1[i]^.tip^.dat[1];
         diskBufPtr^.dat1[2]:=permArray2[i]^.tip^.dat[2];
         diskBufPtr^.setCount:=setNum;
         inc(diskBufNum);
         diskBufPtr:=diskBufPtr^.next;
      End; {for}
      If (done=1) or ((diskbufnum + tipCount) > diskBufMax) Then Begin
         gotoxy(20,14);
         write('Writing results to file.'); {announce that we are writing}
         current:=diskBufStart;
         For i:=1 To diskBufNum Do Begin
            write(permFile,current^.name1);
            write(permFile,' ');
            write(permFile,current^.dat1[1]);
            write(permFile,' ',current^.dat1[2]);
            write(permFile,' ',current^.setCount);
            writeln(permFile);
            current:=current^.next;
         End; {for i}
         diskBufNum:=0;
         diskBufPtr:=diskBufStart;
         gotoxy(20,14);                          {remove announcement}
         clreol;
      End; {then}
   End; {procedure writetips}

{***************************************************************************}
{ This procedure frees all the space on the heap taken up by the results    }
{ buffer.                                                                   }
{***************************************************************************}
Procedure freeHeap(Var ptr : datPtr);
   Var next :datptr;
   Begin
      Repeat
         next:=ptr^.next;
         dispose(ptr);
         ptr:=next;
      Until next=nil;
   End; {procedure freeheap}

{***************************************************************************}
{ This procedure opens the test (.TST) file to which the counts of observed }
{ permutations are written.                                                 }
{***************************************************************************}
Procedure openTestFile(testFileName : flstring; Var testFile : text);
   Begin
      assign(testFile,testFileName);
      rewrite(testFile);
   End; {procedure opentestfile}

{***************************************************************************}
{ This procedure writes the header information for the test (.TST) file.    }
{***************************************************************************}
Procedure writeTestFileHeader(testFileName : flString;
                              tree         : nodePtr;
                              Var testFile : text);
   Var
      year, month, day, dayofWeek, hour, minute, second, sec100 : word;
      distName,probName : flString;
      place : integer;
   Begin
      writeln(testFile,'File = ',testFileName);
      write(testFile,'This file written: ');
      getDate(Year, Month, Day, DayofWeek);
      write(testFile,' ',Day,' ');
      Case month Of
         1 : write(testFile, 'Jan ');
         2 : write(testFile, 'Feb ');
         3 : write(testFile, 'Mar ');
         4 : write(testFile, 'Apr ');
         5 : write(testFile, 'May ');
         6 : write(testFile, 'Jun ');
         7 : write(testFile, 'Jul ');
         8 : write(testFile, 'Aug ');
         9 : write(testFile, 'Sep ');
        10 : write(testFile, 'Oct ');
        11 : write(testFile, 'Nov ');
        12 : write(testFile, 'Dec ');
      End; {case month}
      write(testFile, year,' at ');
      getTime(hour, minute, second, sec100);
      write(testFile,hour,':');
      If (minute < 10)
      Then write(testFile,'0',minute,':')
      Else write(testFile,minute,':');
      If (second < 10)
      Then writeln(testFile,'0',Second)
      Else writeln(testFile,Second);
      place := pos('.',testFileName);
      distName := copy(testFileName,1,place-1);
      distName := distName + '.DST';
      probName := copy(testFileName,1,place-1);
      probName := probName + '.PRB';
      writeln(testFile,'This file contains data produced by PDRANDOM');
      writeln(testFile,'Contents are:');
      If (tree <> nil)
      Then writeln(testFile,'     ',distName,' = pathlength distance matrix (either 1 or 2)');
      writeln(testFile,'     ',probName,' = permutation probability matrix (either 1 or 2)');
      writeln(testFile,'     cell counts indicating outcome of permutations');
      writeln(testFile);
      writeln(testFile);
      writeln(testFile,'Note that .DST and .PRB files can also be produced separately in PDRANDOM');
      writeln(testFile);
      writeln(testFile,'******************************************************************************');
      writeln(testFile);
   End; {writetestfileheader}

{***************************************************************************}
{ This procedure writes the row and column titles for the matrix of counts  }
{ that appears in the test (.TST) file.                                     }
{***************************************************************************}
Procedure writeTestMatrixTitles(perms : permArrayType;
                                outFileName : flstring;
                                Var outFile : text;
                                tipCount : integer);
   Var
      i : integer;
   Begin
      writeln(outFile);
      write(outFile,' Dest. ');
      For i := 1 To tipCount Do Begin
         write(outFile,'  ');
         write(outFile,perms[i]^.tip^.name);
         write(outFile,'  ');
      End; {for i}
      writeln(outFile,'Row Sum');
      writeln(outFile,'Source');
   End; { procedure writeTestMatrixTitles }

End. {unit randio}