{ RANDPERM.PAS                                                         }
{                                                                      }
{ Modification history:                                                }
{    1 March 1996 - P. Midford                                         }
{                                                                      }
{   15 March 1996 - Fixed counting function; P. Midford                }
{                                                                      }
{    2 April 1996 - Fixed (SPSS) column spacing for large (>9999)      }
{                   permutation runs.                                  }
{                                                                      }
{   23 Aug 1997 - minor formating improvements; P. Midford             }
{                                                                      }
{ Last modified:  23 Aug 1997 P. E. Midford                            }

Unit RANDPERM;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {small stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}


Interface

uses randutil,randio,crt,dos;

Procedure setupPermArray(data : phyloPtr;tipCount : integer);

Procedure checkTipBranchLengths(data : phyloPtr; tipCount : integer);

Procedure maxDistance(data : phyloPtr;
                      tipCount : integer;
                      userOptions : options);


Procedure fillPermRow(data : phyloPtr;
                      tipCount : integer;
                      row : integer;
                      userOptions : options);

Function maxPermEstimate(data : phyloPtr;
                         tipCount : integer;
                         userOptions : options) : longint;

Procedure showDistances(data : phyloPtr;
                        tipCount : integer;
                        userOptions : options);

Procedure showProbabilities(data : phyloPtr;
                            tipCount : integer;
                            userOptions : options);


Procedure doPermutations(data : phyloPtr;
                         tipCount : integer;
                         trait : integer;
                         userOptions : options);


Procedure backPermute(tipCount : integer;
                      trait : integer;
                      permArray : permArrayType);

Procedure writeTestStatistics(data1 : phyloPtr;
                              data2 : phyloPtr;
                              tipCount : integer;
                              userOptions : options;
                              permutationCount : integer);



Implementation

{**********************************************************************}
{ This procedure copies the list of tips for the tree into the         }
{ permArray.                                                           }
{**********************************************************************}
Procedure setupPermArray(data : phyloPtr;
                         tipCount : integer);
   Var curTip : nodePtr;
       i : integer;
   Begin
      curTip := data^.tipList;
      i := 1;
      While (curTip <> nil) Do Begin
         If (curTip^.tipNum <> i)
         Then Begin
            write('Internal Index error; Name: ');
            write(curTip^.name);
            write(', tipNum: ');
            write(curTip^.tipNum);
            write(', index: ');
            writeln(i);
            halt
         End; { then }
         data^.permArray[i]^.tip := curTip;
         i := i + 1;
         curTip := curTip^.tipR
      End {While}
  End; {procedure setupPermArray}

{**********************************************************************}
{ This procedure checks that each branch that leads to a tip is non-   }
{ zero.  If a zero branch length is found, the user can either add     }
{ a small value to each branch length, or quit and re-edit the tree in }
{ PDTREE.                                                              }
{**********************************************************************}
Procedure checkTipBranchLengths(data : phyloPtr;tipCount : integer);
   Var curTip : nodePtr;
       shortest : extended;
       tweakSize : extended;
       valCode : integer;
       zeroFound : boolean;
       tweakBuff : string[80];
       userChar : char;

   {*******************************************************************}
   { This procedure adds a short distance to the length of each branch.}
   {*******************************************************************}
   Procedure addTweak(node : nodePtr; tweakSize : extended);
      Begin
         If (node^.sibL <> nil)
         Then addTweak(node^.sibL, tweakSize);
         If (node^.sibR <> nil)
         Then addTweak(node^.sibR, tweakSize);
         node^.length := node^.length + tweakSize;
      End;

   {*******************************************************************}
   { This function searchest the tree for the shortest tip branch.     }
   {*******************************************************************}
   Function findShortestBranch(node     : nodePtr;
                               shortest : extended) : extended;
      Var tmp : extended;
      Begin
         tmp := shortest;
         If (node^.sibL <> nil)
         Then Begin
            tmp := findShortestBranch(node^.sibL,tmp);
            If (tmp > shortest)
            Then tmp := shortest;
         End;
         If (node^.sibR <> nil)
         Then Begin
            tmp := findShortestBranch(node^.sibR,tmp);
            If (tmp > shortest)
            Then tmp := shortest;
         End;
         If ((node^.length < tmp) and (node^.length > 0))
         Then tmp := node^.length;
         findShortestBranch := tmp;
      End;
   {*******************************************************************}
   Begin {procedure checkTipBranchLengths}
      zeroFound := false;
      curTip := data^.tipList;
      While ((curTip <> nil) and not(zeroFound)) Do
         If (curTip^.length = 0.0)
         Then zeroFound := true
         Else curTip := curTip^.tipR;
      If zeroFound
      Then Begin
         clrscr;
         writeln('Warning -- a tip with zero branch length was detected.');
         writeln('           The tip has a name of: ',curTip^.name);
         writeln('           Would you like to correct this by adding a ');
         writeln('           small value to each branch? (Y/N): ');
         readln(userChar);
         If ((userChar = 'Y') or (userChar = 'y'))
         Then Begin
            shortest := findShortestBranch(data^.tree,1.0E30);
            writeln('What small value would you like to add');
            writeln('(Shortest branch length = ',shortest,' default choice = ',0.001*shortest,')');
            readln(tweakBuff);
            If (length(tweakBuff) > 0)
            Then val(tweakBuff,tweakSize,valCode)
            Else tweakSize := 0.001*shortest;
         End {then}
         Else Begin
            writeln('You need to edit this tree with PDTREE before using it in PDRANDOM');
            writeln('Press enter to exit PDRANDOM.');
            readln;
            halt
         End; {else}
         addTweak(data^.tree,tweakSize);
      End {then}
   End; { procedure checkTipBranchLengths }


{**********************************************************************}
{ This function calculates the distance between two nodes on the tree. }
{ If the two nodes are the same, the function returns the product of   }
{ the node's distance to its ancestor and the selfDistance factor.     }
{**********************************************************************}
Function calcDistance(root,sourceNode,destNode : nodePtr;
                      userOptions : options) : extended;
   Var
     sourceAncestor : nodePtr;
     destAncestor : nodePtr;
     sourceDist : extended;
     destDist : extended;
     totalDist : extended;
   Begin
      If (sourceNode = destNode)
      Then
         If (userOptions.myMethod = similarity)
         Then calcDistance := 0
         Else calcDistance := userOptions.factor*sourceNode^.length
      Else Begin
         totalDist := -1.0;
         sourceAncestor := sourceNode^.ancestor;
         sourceDist := sourceNode^.length;
         If (sourceAncestor <> root)
         Then Begin
            While((sourceAncestor <> root) and (totalDist < 0.0)) Do Begin
               destAncestor := destNode^.ancestor;
               destDist := destNode^.length;
               While ((destAncestor <> root) and (totalDist < 0.0)) Do Begin
                  If (destAncestor = sourceAncestor)
                  Then totalDist := sourceDist + destDist
                  Else Begin
                     destDist := destDist + destAncestor^.length;
                     destAncestor := destAncestor^.ancestor
                  End
               End;
               If (totalDist < 0.0)
               Then Begin
                  sourceDist := sourceDist + sourceAncestor^.length;
                  sourceAncestor := sourceAncestor^.ancestor;
                  If sourceAncestor = root
                  Then totalDist := sourceDist + destDist
               End {then}
            End {while}
         End {then}
         Else Begin
            destAncestor := destNode^.ancestor;
            destDist := destNode^.length;
            While ((destAncestor <> root) and (totalDist < 0.0)) Do Begin
               If (destAncestor = sourceAncestor)
               Then totalDist := sourceDist + destDist
               Else Begin
                  destDist := destDist + destAncestor^.length;
                  destAncestor := destAncestor^.ancestor
               End {else}
            End; {while}
            If (totalDist < 0.0)
            Then Begin
               destDist := destDist + destAncestor^.length;
               totalDist := sourceDist + destDist
            End {then}
         End; {else}
         If (sourceNode = destNode)
         Then calcDistance := totalDist/2.0
         Else calcDistance := totalDist
      End {else}
   End; { function calcDistance }

{**********************************************************************}
{ This procedure calculates the distance between each of the tipnum^2  }
{ pairs of tips to find the largest distance.  Because this is         }
{ expensive, it should only be done once before we start calling       }
{ fillPermRow (below).  Since this value is only used for the          }
{ similarity method of calculating probabilities, this function just   }
{ returns if another method is selected.  This procedure assumes that  }
{ the permarray has already been set up.                               }
{**********************************************************************}
Procedure maxDistance(data : phyloPtr;
                      tipCount : integer;
                      userOptions : options);
   Var dist : extended;
       xdist : extended;
       max : extended;
       ln10 : extended;
       row,column : integer;
   Begin
      max := -1.0;  {impossibily small value}
      ln10 := 2.30258509299;
      For row := 1 to tipCount Do
         For column := 1 to tipCount Do
            If row <> column Then With data^ Do Begin
               {don't waste time for known 0's}
               dist := calcDistance(data^.tree,
                                    permArray[row]^.tip,
                                    permArray[column]^.tip,
                                    userOptions);
               If userOptions.logTrans
               Then
                  xdist := ln(1/dist)/ln10
               Else
                  xdist := exp(userOptions.blPower*ln(dist));
               If xdist > max
               Then max := xdist;
            End; {then with}
      data^.maxDist := max;
   End; {procedure maxDistance}


{**********************************************************************}
{ This procedure calculates one row of the permutation matrix.  At the }
{ present (sept 1997), this is done by calculating distances from the  }
{ source node to each target node, taking reciporicals, and normalizing}
{ so that the reciporicals sum to one.                                 }
{**********************************************************************}
Procedure fillPermRow(data : phyloPtr;
                      tipCount : integer;
                      row : integer;
                      userOptions : options);
   Var probSum : extended;
       ln10 : extended;
       min : extended;
       max : extended;
       dist : extended;
       xdist : extended;
       prob : extended;
       column : integer;
   Begin
      ln10 := 2.30258509299;
      max := -1.0;
      min := 1e6;
      probSum := 0;
      For column := 1 to tipCount Do With Data^ Do Begin
         If (data^.tree <> nil) Then Begin
            Case userOptions.myMethod Of
               similarity :
                  Begin
                     If (row <> column) Then Begin
                        dist := calcDistance(tree,
                                             permArray[row]^.tip,
                                             permArray[column]^.tip,
                                             userOptions);
                        If userOptions.logTrans
                        Then xdist := ln(1/dist)/ln10
                        Else xdist := exp(userOptions.blPower*ln(dist));
                        permArray[column]^.probability :=
                           userOptions.k - (xdist/data^.maxDist);
                        probSum := probSum + permArray[column]^.probability;
                     End {then}
                     Else Begin
                        permArray[column]^.probability := userOptions.k;
                        probSum := probSum + userOptions.k;
                     End; {else}
                  End; {similarity}
               distance :
                  Begin
                     If ((row <> column) or not(userOptions.exclude))
                     Then Begin
                        dist := calcDistance(tree,
                                             permArray[row]^.tip,
                                             permArray[column]^.tip,
                                             userOptions);
                        permArray[row]^.distance := dist;
                        write;
                        If userOptions.logTrans
                        Then xDist := ln(1/dist)/ln10
                        Else xDist := exp(userOptions.blPower*ln(dist));
                        {If (prob<min)
                        Then min := prob;
                        If (prob>max)
                        Then max := prob;}
                        probSum := probSum + xDist;
                        permArray[column]^.probability := xDist;
                     End {then}
                     Else permArray[column]^.probability := 0
                  End; {distance}
               equiprobable :
                  Begin
                     permArray[column]^.probability := 1/tipCount;
                     probSum := probSum + 1/tipCount;
                  End; {equiprobable}
               End; {case userOptions.myMethod}
         End {then}
         Else Begin
            permArray[column]^.probability :=
               matrix^.data[row,column];
            probSum := probSum + matrix^.data[row,column];
         End {else}
      End; {for column / with data^}
{      If (minMax < 0) Then Begin
         probSum := probSum - data^.permArray[row]^.probability + min;
         data^.permArray[row]^.probability := min;
      End;} {then}
{      If (minMax > 0) Then Begin
         probSum := probSum - data^.permArray[row]^.probability + max;
         data^.permArray[row]^.probability := max;
      End;} {then}
      data^.permArray[row]^.rowSum := 0;
      For column := 1 to tipCount Do With data^ Do Begin
        permArray[column]^.probability := permArray[column]^.probability/probSum;
        permArray[column]^.columnSum :=
           permArray[column]^.columnSum + permArray[column]^.probability;
        permArray[row]^.rowSum :=
           permArray[row]^.rowSum + permArray[column]^.probability;
      End; {for column / with data^}
   End; {procedure fillpermrow}

{**********************************************************************}
{ This function estimates the maximum number of permutations the user
  should run.  It takes the highest calculated probability in the PRB
  matrix (calculated with fillpermrow), halves it, and divides it 
  into 65536.  The reason for doubling is that the prb matrix only
  estimates the observed probability.                                  }
{**********************************************************************}
Function maxPermEstimate(data : phyloPtr;
                         tipCount : integer;
                         userOptions : options) : longint;
   Var i,j : integer;
       maxValue : real;
   Begin
      maxValue := 0;
      With data^ Do Begin
         For i := 1 To tipCount Do Begin
            fillPermrow(data,tipCount,i,userOptions);
            For j := 1 To tipCount Do Begin
               If permArray[j]^.probability > maxValue
               Then maxValue := permArray[j]^.probability;
            End; {for j}
         End; {for i}
         maxPermEstimate := trunc(32768.0/maxValue);
      End; {with data^}
   End; { function maxPermEstimate}


{**********************************************************************}
{ This procedure writes the distance (.DST) file, which consists of a  }
{ header, row and column titles, and the distance between pairs of     }
{ tips.                                                                }
{**********************************************************************}
Procedure showDistances(data : phyloPtr;
                        tipCount : integer;
                        userOptions : options);
   Var i,j: integer;
       distanceFile : text;
       distanceFileName : flString;
   Begin
      openDistanceFile(data^.sourceName,distanceFile,distanceFileName);

      writeMatFileHeader(data^.permArray,
                         distanceFileName,
                         distanceFile,
                         tipCount,
                         data^.sourceName,
                         userOptions,
                         false, {no row sums for the distance matrix}
                         false); {matrix data never writes .DST files}

      For i := 1 to tipCount Do Begin
         For j := 1 to tipCount Do With data^ Do
           permArray[j]^.distance := calcDistance(tree,
                                                  permArray[i]^.tip,
                                                  permArray[j]^.tip,
                                                  userOptions);
         writeDistanceData(data^.permArray,
                           i,
                           tipCount,
                           distanceFile,
                           userOptions);
      End; {for i}
      close(distanceFile);
   End; { procedure showDistances }

{**********************************************************************}
{ This procedure writes the probability (.PRB) file, which consists of }
{ a header, row and column titles, and the probabilities between pairs }
{ of tips.                                                             }
{**********************************************************************}
Procedure showProbabilities(data        : phyloPtr;
                            tipCount    : integer;
                            userOptions : options);
   Var probSum : extended;
       probFileName : flString;
       probFile : text;
       i,j: integer;
       matFlag : boolean;
   Begin
      openProbFile(data^.sourceName,probFile,probFileName);
      If data^.matrix <> nil
      Then matFlag := true
      Else matFlag := false;
      writeMatFileHeader(data^.permArray,     {write headers and titles}
                         probFileName,
                         probFile,
                         tipCount,
                         data^.sourceName,
                         userOptions,
                         true,  {we DO want row sums for the prob file}
                         matFlag);
      For i := 1 to tipCount Do    {set all column sums to zero}
         data^.permArray[i]^.columnSum := 0;

      For i := 1 to tipCount Do Begin
         fillPermRow(data,tipCount,i,userOptions);
         writeProbData(data^.permArray,i,tipCount,probFile,userOptions);
      End; {for i}
      writeColumnSums(data^.permArray,tipCount,probFile,userOptions);  { needed for column sums}
      close(probFile);
   End; { procedure showProbabilities }


{**********************************************************************}
{ This procedure inserts a distance matrix into the test (.TST) file.  }
{**********************************************************************}
Procedure insertDistances(data         : phyloPtr;
                          tipCount     : integer;
                          userOptions  : options;
                          Var testFile : text);

   Var i,j: integer;
       place : integer;
       distanceFileName : flString;

   Begin
      With Data^ Do Begin
         place := pos('.',sourceName);
         distanceFileName := copy(sourceName,1,place-1);
         distanceFileName := distanceFileName + '.DST';

         writeMatFileHeader(permArray,
                            distanceFileName,
                            testFile,
                            tipCount,
                            sourceName,
                            userOptions,
                            false,  {no row sum for distance file}
                            false); {no DST info for matrix data}
         For i := 1 to tipCount Do Begin

            For j := 1 to tipCount Do
               permArray[j]^.distance := calcDistance(tree,
                                                      permArray[i]^.tip,
                                                      permArray[j]^.tip,
                                                      userOptions);
            writeDistanceData(permArray,i,tipCount,testFile,userOptions);
         End; {for i}
         writeln(testFile);
         writeln(testFile,'******************************************************************************');
         writeln(testFile);
      End; { with data}
   End; { procedure insertDistances }

{**********************************************************************}
{ This procedure inserts a probability matrix into the test (.TST)     }
{ file.                                                                }
{**********************************************************************}
Procedure insertProbabilities(data         : phyloPtr;
                              tipCount     : integer;
                              userOptions  : options;
                              Var testFile : text);

   Var probSum : extended;
       probFileName : flString;
       i,j: integer;
       place : integer;
       matFlag : boolean;

   Begin
      If data^.matrix <> nil
      Then matFlag := true
      Else matFlag := false;
      With data^ Do Begin
         place := pos('.',sourceName);
         probFileName := copy(sourceName,1,place-1);
         probFileName := probFileName + '.PRB';
         writeMatFileHeader(permArray,
                            probFileName,
                            testFile,
                            tipCount,
                            sourceName,
                            userOptions,
                            true,   {want row sums here}
                            matFlag);
         For i := 1 to tipCount Do
            permArray[i]^.columnSum := 0;

         For i := 1 to tipCount Do Begin
            fillPermRow(data,tipCount,i,userOptions);
            writeProbData(permArray,i,tipCount,testFile,userOptions); {don't insert long values}
         End; {for i}
         writeColumnSums(permArray,tipCount,testFile,userOptions);  { needed for column sums}
      End; {with data^}
      writeln(testFile);
      writeln(testFile,'******************************************************************************');
      writeln(testFile);
   End; { procedure insertProbabilities }

{**********************************************************************}
{ This procedure is the heart of PDRANDOM; it actually performs the    }
{ permutations and counts the outcomes in the testTable.               }
{**********************************************************************}
Procedure doPermutations(data : phyloPtr;
                         tipCount : integer;
                         trait : integer;
                         userOptions : options);

   Var
      probSum : extended;
      draw : extended;
      drawSum : extended;
      curTip : nodePtr;
      ancestor : nodePtr;
      i : integer;
      j : integer;
      k : integer;
      temp : integer;
      source : integer;
      target : integer;
      targets : array [1..permMax] of integer;
      sources : array [1..permMax] of integer;

   Begin
      With data^ Do Begin
         For i := 1 to tipCount Do Begin
            permArray[i]^.usedSource := false;
            permArray[i]^.usedTarget := false;
            permArray[i]^.temp := permArray[i]^.tip^.dat[trait];
         End; {for i}

         For i := 1 to tipCount - 1 Do Begin

            Repeat
               If (userOptions.generator = manley)
               Then source := trunc(manRand*tipCount) + 1
               Else source := random(tipCount) + 1;
            Until (permArray[source]^.usedSource = false);
            sources[i] := source;
            fillPermRow(data,tipCount,source,userOptions);

            permArray[source]^.usedSource := true;    { mark this as used }
            probSum := 0;
            For j := 1 to tipCount Do
               If (permArray[j]^.usedTarget = false) Then
                  probSum := probSum + permArray[j]^.probability;
            drawSum := 0;
            If (userOptions.generator = manley)
            Then draw := manRand * probSum
            Else draw := Random * probSum;
            j := 1;
            target := 0;
            While (target = 0) Do Begin
               If (permArray[j]^.usedTarget = false)
               Then Begin
                  drawSum := drawSum + permArray[j]^.probability;
                  If (draw < drawSum) Then target := j;
               End; { Then }
               j := j + 1
            End; { while }
            {write(audit,'(',target,') ');}
            targets[source] := target;
            permArray[target]^.usedTarget := true;
         End; {for i}
         { finally assign the one remaining source to the remaining target }
         source := 0;
         target := 0;
         For i := tipCount downto 1 Do Begin

            If (permArray[i]^.usedSource = false)
            Then source := i;
            If (permArray[i]^.usedTarget = false)
            Then target := i

         End; {for i}
         sources[tipCount] := source;

         targets[source] := target;
         For i := 1 to tipCount Do Begin
            permArray[targets[i]]^.tip^.dat[trait] := permArray[i]^.temp;
            If testTable <> nil
            Then testTable^[targets[i],i] := testTable^[targets[i],i]+1;
         End; { for i}
      End; {with data}
   End; {procedure do Permutations}

{**********************************************************************}
{ This procedure undoes a permutation; that is, it returns tip values  }
{ to their original tips.                                              }
{**********************************************************************}
Procedure backPermute(tipCount : integer;
                      trait : integer;
                      permArray : permArrayType);
   Var i : integer;
   Begin
      For i := 1 to tipCount Do
         permArray[i]^.tip^.dat[trait] := permArray[i]^.temp;
   End; {procedure backPermute}

{**********************************************************************}
{ This procedure writes the entire test (.TST) file.  This file        }
{ consists of a header, the distance and probability matrices, and     }
{ counts of the permutation outcomes (number of times tip i was mapped }
{ to tip j) in both matrix and three column (source,dest,count)        }
{ formats.                                                             }
{**********************************************************************}
Procedure writeTestStatistics(data1 : phyloPtr;
                              data2 : phyloPtr;
                              tipCount : integer;
                              userOptions : options;
                              permutationCount : integer);

   Var
      value,rowSum : extended;
      colSum : array [1..permMax] of extended;
      ivalue,rowiSum : longint;
      coliSum : array [1..permMax] of longint;
      testFile : text;
      i,j : integer;
      extendedOption : boolean;

   Begin
      extendedOption := userOptions.showExtended; {save state of user choice}
      userOptions.showExtended := false;     {no extended values in TST file}
      openTestFile(data1^.tstFileName,testFile);
      writeTestFileHeader(data1^.tstFileName,data1^.tree,testFile);
      If (data1^.tree <> nil)  {write distances if data is a tree}
      Then Begin
         If (data1^.sourceName <> data2^.sourceName)
         Then Begin  { don't write the information twice!}
            insertDistances(data1,tipCount,userOptions,testFile);
            insertDistances(data2,tipCount,userOptions,testFile)
         End {then}
         Else insertDistances(data1,tipCount,userOptions,testFile)
      End; {then}
      { write probabilities }
      If (data1^.sourceName <> data2^.sourceName)
      Then Begin  { don't write the same information twice!}
         insertProbabilities(data1,tipCount,userOptions,testFile);
         insertProbabilities(data2,tipCount,userOptions,testFile);
      End {then}
      Else Begin
         insertProbabilities(data1,tipCount,userOptions,testFile);
         For i := 1 to tipCount Do
            data2^.permArray[i]^.columnSum := 0;

         For i := 1 to tipCount Do
            fillPermRow(data2,tipCount,i,userOptions);
      End; {else}

      { dump matrix results for tree1 }
      With data1^ Do Begin
         writeln(testFile,'Test results for trait 1 in matrix form');
         writeln(testFile);
         writeTestMatrixTitles(permArray,data1^.tstFileName,testFile,tipCount);
         For j := 1 to tipCount Do colisum[j] := 0;
         For i := 1 to tipCount Do Begin
            write(testFile,permArray[i]^.tip^.name,'    ');
            rowiSum := 0;
            For j := 1 to tipCount Do Begin
               ivalue := data1^.testTable^[j,i];
               rowiSum := rowiSum + ivalue; 
               coliSum[j] := coliSum[j] + ivalue; 
               write(testFile,ivalue:5,' ');
            End;
            writeln(testFile,' ',rowiSum:5);
         End; { for }
         writeln(testFile);
         write(testFile,'Sum   ');
         For j := 1 to tipCount Do write(testFile,coliSum[j]:5,' ');
         writeln(testFile)
      End; {with data1^}

      { dump matrix results for tree2 }
      With data2^ Do Begin
         writeln(testFile,'******************************************************************************');
         writeln(testFile);
         writeln(testFile,'Test results for trait 2 in matrix form');
         writeln(testFile);
         writeTestMatrixTitles(permArray,data1^.tstFileName,testFile,tipCount);
         For j := 1 to tipCount Do colisum[j] := 0;
         For i := 1 to tipCount Do Begin
            write(testFile,permArray[i]^.tip^.name,'    ');
            rowiSum := 0;
            For j := 1 to tipCount Do Begin
               ivalue := testTable^[j,i];
               rowISum := rowISum + ivalue;
               coliSum[j] := coliSum[j] + ivalue; 
               write(testFile,ivalue:5,' ');
            End; { for }
            writeln(testFile,' ',rowiSum:5);
         End; {for i}
         writeln(testFile);
         write(testFile,'Sum   ');
         For j := 1 to tipCount Do write(testFile,coliSum[j]:5,' ');
         writeln(testFile);
      End; {with data2^}

      {dump results in four column format }
      writeln(testFile,'******************************************************************************');
      writeln(testFile);
      writeln(testFile,'Row=    Column=');
      writeln(testFile,'Source  Target  Counts(1)  Counts(2) ');
      writeln(testFile);
      For i := 1 to tipCount Do Begin
         fillPermRow(data1,tipCount,i,userOptions);
         fillPermRow(data2,tipCount,i,userOptions);
         For j := 1 To tipCount Do Begin
            write(testFile,i:3,' ');
            write(testFile,data1^.permArray[i]^.tip^.name,' ');
            write(testFile,j:3,' ');
            write(testFile,data2^.permArray[j]^.tip^.name,'    ');
            write(testFile,data1^.testTable^[i,j]:5,'      ');
            writeln(testFile,data2^.testTable^[i,j]:5);
         End; {for j}
      End; {for i}
      writeln(testFile,'******************************************************************************');
      close(testFile);
      userOptions.showExtended := extendedOption;     {restore user option}
   End; {procedure writeteststatistics}
End. {unit randperm}