{ RANDUTIL.PAS                                                         }
{                                                                      }
{ This unit contains most of the program's type definitions (tree,     }
{ probability matrix, count matrix, output buffer records, etc) as well}
{ as utility functions for dealing with these structures.              }
{                                                                      }
{ ******************************************************************** }
{ Last modified 22 Sep 1997 P. Midford                                 }

Unit RANDUTIL;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {small stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}


Interface

Const permMax = 300;          {maximum (potential) size of tree}
      countTableSize = 180;   {Current operating maximum}
      realSize = 3;           {maximum size of tree we can calculate realized
                              {probabilities for}
      matSize = 80;           {maximum size of matrix we can explicitly read}


Type

      flstring = string[30];    {string to hold a file name}

      typstring = string[3];    {string to hold a file's extension}

      namestring = string[2];   {string to hold the name of a node}

      traits = array[1..2] of double; {a pair of trait values}

      nodeptr = ^node;

      node = Record          { **simplifed from PDTREE** }
         Ancestor : nodeptr; {the node's ancestor (nil for the root)}
         dat      : traits;  {Tip data}
         length   : extended;    {distance to ancestor}
         SibL     : nodeptr; {Children of node}
         SibR     : nodeptr;
         tipL     : nodeptr; {Implements doubly linked list of tree's tips}           
         tipR     : nodeptr;
         tipnum   : integer;  {Number assigned to tips; left to right order}
         name     : namestring; {Two character name of the node}
         tempx    : integer;   {used in linking the tree for buffered output}
         tempy    : integer;
      End; {type node}

      permPtr = ^permRecord;

      { permRecord holds the permutation information for one tip.  These
        structures are allocated dynamically to avoid running out of static
        variable space and (potentially) so we only allocate as many as
        needed. }

      permRecord =
      Record
         tip : nodePtr;     {points to the tip (to get data values)}
         distance : extended;   {the distance from this tip to the tip whose row it is}
         probability : extended; {the permutation probability for the two tips}
         temp : extended;       {used in the actual permutation of values}
         columnSum : extended;  {holds the sum of the i-th column (i = tip number)}
         rowSum : extended;     {holds the sum of the i-th row}
         usedSource : boolean; {used in permutation to mark tip as a used source}
         usedTarget : boolean; {used in permutation to mark tip as a used target}
      End;


      { PermArrayType effectively represents one row of the permutation
        matrix.  The distance and probability fields of each permRecord
        hold the appropriate values, which are recomputed using fillPermRow.
        FillPermRow is called once for each row of the matrix. }

      permArrayType = array [1..permMax] of permPtr;


      {Exclude lists may some day hold lists of source - destination pairs
       that are not to be permuted.  Note that this is a one-way relation.
       The exclusion list feature is not currently implemented.}

      excludePtr = ^excludeList;
      excludeList = Record
         source : nodePtr;   {the source of the excluded pair}
         dest : nodePtr;     {the destination}
         next : excludePtr;
      End; {excludeList}

      {MatType holds explicitly read matrices}

      matPtr = ^matType;
      matType = Record
         data : array[1..matSize,1..matSize] of extended;  {the probabilities}
         tips : array[1..matSize] of double;             {single tip values}
         names : array[1..matSize] of namestring;      {tip names}
         size : integer;                               {amount of data}
      End; {matType}

      {a list of these structures are used by the output caching procedures
       (in randio1) to hold chunks of generated data.  The chunk is written
       out when it fills. }

      datptr = ^bigData;     {used by randio1 to keep permutation results
                              on the heap so it can print many permutation 
                              sets at once.}
      bigData = Record
         name1    :namestring;  {node name}
         dat1     :traits;      {node values}
         name2    :namestring;  {next node (not used for matrix data) }
         dat2     :traits;
         setCount :longint;
         next     :datptr;
      End; {bigData}

      {These types hold the counts of permutations written to the .TST file}
      countTableType = array [1..countTableSize,1..countTableSize] of word;
      countTablePtr = ^countTableType;


      { This array is used for calculating realized probabilities. }
      rArray = array[1..realSize,1..realSize] Of extended;

      {This structure (may be it should be called something else) holds
       either a tree or a probability matrix along with some associated
       data.  This structure will hopefully make some of the function
       calls clearer (at least there should be fewer arguments passed). }
      phylogeny = Record
         tree : nodePtr;               {tree representation}
         matrix : matPtr;              {matrix representation}
         excludeList : excludePtr;     {a list of excluded permutations}
         tipList : nodePtr;            {a list of the tree's tips}
         testTable : countTablePtr;    {table of observed count numbers}
         permArray : permArrayType;    {list of nodes and permutations}
         maxDist : extended;           {the largest distance}
         sourceName : flString;        {name of the primary data file}
         sourceExt : typString;        {extension of the primary data file}
         matrixTipFile : flString;     {name of the PDI file with tip data}
         tstFileName : flString;
      End; {record phylogeny}

      phyloPtr = ^phylogeny;

      {This structure holds all the user menu options in one place.}

      pMethod = (similarity,distance,equiprobable);
      generatorType = (intrinsic,manley);


      options = Record
                   seed : longInt; { This holds the random # seed used by the program}
                   blPower : real; { Power to raise branch lengths to}
                   logTrans : boolean;
                   exclude : boolean;
                   showExtended : boolean; {display numbers in extended mode?}
                   generator : generatorType;
                   case myMethod : pMethod of
                      similarity: (k : extended);
                      distance: (factor : extended; { multiplier for nearest ancestor distance to get self distance}
                                 minMax : integer);  { the minimize / maximize flag}
                End; {record options}


      Var ix,iy,iz : integer;   {the three seeds used by the Manley generator}

      {*********************************************************************}

      {For procedure comments, see their bodies in the implementation section.}

      Procedure initProgramName(name : string);

      Procedure getProgramName(Var name : string);

      Procedure initPerm(p : permPtr);

      Procedure initCountTable(Var t : countTablePtr);

      Procedure findHeight(rt,nod:nodeptr;height:extended; Var max:extended);

      Procedure scl(rt,nd:nodeptr; mult:extended);

      Procedure prepareTips(data :phyloPtr; var total:integer);

      Procedure initMe(me:nodePtr);

      Procedure initData(data:phyloPtr);

      Procedure killTree(Var nod:nodePtr);

      Procedure ret(message:string);

      Procedure CheckName(nod:nodeptr;
                          name:namestring;
                          Var notfound:boolean;
                          Var want:nodeptr);

      Procedure numtps(nod:nodeptr; var num:integer);

      Function ManRand : extended;

      Procedure manRandSeeds(var newX,newY,newZ : integer);

      Procedure manRandomize;

{****************************************************************************}

Implementation

uses crt,  { provides the gotoxy function for moving the text cursor }
     dos;  { provides time functions (for initializing the Manley generator }

Var programName : string[8];   {holds the name of the program in a
                                single place so that multiple programs
                                can use this module.}

{*********************************************************************}
{ This procedure initializes a newly allocated permRecord.            }
{*********************************************************************}
Procedure initPerm(p : permPtr);
   Begin
      p^.tip := nil;
      p^.temp := 0;
      p^.distance := 0.0;
      p^.probability := 0.0;
      p^.columnSum := 0.0;
      p^.rowSum := 0.0;
      p^.usedSource := false;
      p^.usedTarget := false;
      p^.distance := 0.0;
   End; {procedure initPerm}

{*********************************************************************}
{ This procedure initializes a newly allocated countTable.  Note that }
{ the diagonal is filled with 1's.  That's because the first          }
{ permutation is conventionally the orginal ordering.                 }
{*********************************************************************}
Procedure initCountTable(Var t : countTablePtr);
   Var i : integer;
       j : integer;
   Begin
      If (t = nil) Then new(t);
      If (t <> nil) Then
         For i := 1 to countTableSize do
            For j := 1 to CountTableSize do
               If (i = j)
               Then t^[i,j] := 1
               Else t^[i,j] := 0;
   End; { Procedure initCountTable }

{*********************************************************************}
{ This recursive procedure finds the distance from the node it is     }
{ given initially to the farthest tip that is a descendant of the     }
{ initial node.  It includes the length of the initial node in the    }
{ answer.  PDRANDOM uses this function to scale the tree to a height  }
{ of 1.0.                                                             }
{*********************************************************************}
Procedure findHeight(rt,nod :nodePtr;
                     height :extended;
                     Var max :extended);
   Begin
      If nod<>rt
      Then height:=height+nod^.length;
      If nod^.sibL<>nil
      Then Begin
         findheight(rt,nod^.sibL,height,max);
         findheight(rt,nod^.sibR,height,max);
      End {then}
      Else
         If height>max
         Then max := height;
   End; {findHeight}

{*********************************************************************}
{ This recursive procedure scales a tree.  It is given the scaling    }
{ factor, and then traverses the tree, multiplying all the branch     }
{ lengths by the correct amount.                                      }
{*********************************************************************}
Procedure scl(rt,nd :nodePtr; mult:extended);
   Begin
      If nd<>rt
      Then nd^.length:=nd^.length*mult;
      If nd^.sibL<>nil
      Then Begin
         scl(rt,nd^.sibL,mult);
         scl(rt,nd^.sibR,mult);
      End; {then}
   End; {procedure scl}

{*********************************************************************}
{ This procedure sets the program name, which is used in the display  }
{ routines.  It exists so that multiple programs can share this and   }
{ other modules.                                                      }
{*********************************************************************}
Procedure initProgramName(name : string);
   Begin
      programName := name;
   End; {procedure initprogramname}

{*********************************************************************}
{ This procedure retrieves the saved program name.                    }
{*********************************************************************}
Procedure getProgramName(Var name : string);
   Begin
      name := programName;
   End; {procedure getprogramname}

{***********************************************************************}
{ This procedure initializes some of the fields of a node to 0.0 or nil.}
{ It is called just about whenever a node is allocated.                 }
{***********************************************************************}
Procedure initMe(me :nodePtr);
   Begin
      me^.dat[1]:=0.0;
      me^.dat[2]:=0.0;
      me^.sibL:=nil;
      me^.sibR:=nil;
   End; {procedure initme}

{***********************************************************************}
{ This recursive procedure deallocates nodes of the tree.               }
{***********************************************************************}
Procedure killTree(Var nod :nodePtr);
   Begin
      If nod^.sibL<>nil Then killTree(nod^.sibL);
      If nod^.sibR<>nil Then killTree(nod^.sibR);
      dispose(nod);
   End; {procedure killtree}

{**************************************************************************}
{ This procedure initializes a newly allocated phylogeny (data) structure. }
{**************************************************************************}
Procedure initData(data :phyloPtr);
   Var i : integer;
   Begin
      data^.tree := nil;
      data^.matrix := nil;
      data^.excludeList := nil;
      data^.tipList := nil;
      data^.testTable := nil;
      data^.sourceName := '';
      data^.sourceExt := '';
      data^.matrixTipFile := '';
      For i := 1 to permMax Do Begin {allocate and initialize permutation rows }
         new(data^.permArray[i]);
         initPerm(data^.permArray[i]);
      End; {for}
   End; {procedure initData}

{************************************************************************}
{ This just prints a message to the bottom of the screen and then        }
{ reads a keypress.                                                      }
{************************************************************************}
Procedure ret(message :string);
   Var
      ans :char;
   Begin
      gotoxy(1,23);
      writeln;
      writeln(message);
      ans:=readkey;
   End; {procedure ret}

{*************************************************************************}
{ This recursive procedure checks a name that is passed with the rest of  }
{ the tree to see if it is already in the tree.  If it is, it sets        }
{ notfound to false.                                                      }
{*************************************************************************}
Procedure checkName(nod :nodeptr;
                    name :namestring;
                    Var notFound :boolean;
                    Var want :nodeptr);
   Begin
      If nod^.name=name Then Begin
         notfound:=false;
         want:=nod;
      End {then}
      Else Begin
         If (nod^.sibL<>nil)
         Then checkname(nod^.sibL,name,notfound,want);
         If (nod^.sibR<>nil) And notFound
         Then checkname(nod^.sibR,name,notfound,want);
      End {else}
   End; {procedure checkname}

{**************************************************************************}
{ This recursive procedure counts the number of tips descended from a node }
{ it is given.                                                             }
{**************************************************************************}
Procedure numtps(nod:nodeptr; Var num :integer);
   Begin
      If nod^.sibL<>nil Then Begin
         numtps(nod^.sibL,num);
         numtps(nod^.sibR,num);
      End
      Else num:=num+1;
   End; {procedure numTps}

{*******************************************************************}
{ This procedure creates a list of the tree's tips in a right to    }
{ left order.  It also sets the tip node's tipnum field.            }
{*******************************************************************}
Procedure prepareTips(data :phyloPtr; Var total :integer);
   Var
      num :integer;
      previous,first :nodePtr;

   {****************************************************************}
   { This local procedure traverses the tree and sets the tip       }
   { pointers tipR,tipL, and the tip number tipnum.                 }
   {****************************************************************}
   Procedure traverse(nd :nodeptr;
                      Var prev,first :nodeptr;
                      Var num :integer);
      Begin
         If (nd^.SibL <> NIL) Then Begin
            traverse(nd^.SibL,prev,first,num);
            traverse(nd^.SibR,prev,first,num);
         End {then}
         Else Begin
            num:=num+1;
            If (num=1) Then first:= nd;
            nd^.tipnum:=num;
            nd^.tipL:=prev;
            If (prev<>NIL) Then prev^.tipR:=nd;
            prev:=nd;
         End; {else}
      End; {procedure traverse}
   {****************************************************************}
   Begin {procedure preparetips}
      num:=0;
      previous:=NIL;
      traverse(data^.tree,previous,first,num);
      previous^.tipR:=NIL;
      total:= previous^.tipnum;
      data^.tipList:=first
   End; {procedure preparetips}

{*******************************************************************}
{ This function implements the random number generator described in }
{ (Manley 1991), originally from Wichmann and Hill 1982 Applied     }
{ Statistics 31:188-90.                                             }
{*******************************************************************}
Function ManRand : extended;
   Var tmp : real;
   Begin
      Ix := 171*(Ix Mod 177)-2*(Ix Div 177);
      Iy := 172*(Iy Mod 176)-35*(Iy Div 176);
      Iz := 170*(Iz Mod 178)-63*(Iz Div 178);
      If (Ix < 0) Then Ix := Ix + 30269;
      If (Iy < 0) Then Iy := Iy + 30307;
      If (Iz < 0) Then Iz := Iz + 30323;
      tmp := ((1.0*Ix)/30269.0 + (1.0*Iy)/30307.0 + (1.0*Iz)/30323.0);
      ManRand := tmp - trunc(tmp);
   End; {function manRand}

{*******************************************************************}
{ This procedure initializes the three variables that represent the }
{ state of the random number generator in the previous function.    }
{*******************************************************************}
Procedure manRandomize;
   Var hour,minute,sec,sec100 : word;
   Begin
      getTime(hour,minute,sec,sec100);
      Ix := (sec+1)*100+sec100;
      Iy := (minute+1)*60+sec;
      Iz := (hour+1)*60+minute;
   End; {procedure manRandomize}

{*******************************************************************}
{ This procedure allows the user to set or inspect the three random }
{ seeds used by the Manley random number generator.                 }
{*******************************************************************}
Procedure manRandSeeds(var newX,newY,newZ : integer);
   Var tmp : integer;
   Begin
      If newX <> 0 Then Begin
         tmp := Ix;
         Ix := newX;
         newX := tmp
      End {then}
      Else newX := Ix;
      If newY <> 0 Then Begin
         tmp := Iy;
         Iy := newY;
         newY := tmp
      End {then}
      Else newY := Iy;
      If newZ <> 0 Then Begin
         tmp := Iz;
         Iz := newZ;
         newZ := tmp
      End {then}
      Else newZ := Iz;
   End; {procedure manRandSeeds}


End.