{ RAND_IO1.PAS                                                         }
{                                                                      }
{ This module contains most of the I/O code for PDRANDOM.  Most of the }
{ code here has been lifted from PDTREE (input side) or PDSIMUL        }
{ (output side).  I have attempted to minimize the amount of           }
{ modification to the copied code (especially on the input side) to    }
{ make it easier to keep the copied code up to date with PDTREE and    }
{ PDSIMUL.                                                             }
{                                                                      }
{ It uses the following units:                                         }
{                                                                      }
{ RANDUTIL - Utility functions and type definitions for trees and      }
{            matrices                                                  }
{ CRT - Provides clrscr for clearing the screen.                       }
{                                                                      }
{**********************************************************************}
{ Modification history:                                                }
{                                                                      }
{ 13 Nov 1999 - Added the improved PDI file reader, that accepts       }
{               longer node names and "free field" spacing.            }
{ Last modified 22 Sept 1997 P. E. Midford                             }


Unit rand_io1;

Interface

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {small stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Uses randutil; {type definitions, constants, and utility functions.}

Procedure extendWrite(Var outFile :text;num :extended);

Procedure InputTree(Var data : phyloPtr;
                    Var succ : boolean;
                    reread   : boolean); { new re-read feature PEM 28 Apr 1996}

Procedure SaveTree(rt          : nodePtr;
                   Var flename : flString;
                   Var typ     : typString;
                   ok          : boolean);

Procedure readMatrix(Var data : phyloPtr;
                     Var succ : boolean;
                     reread   : boolean);

Procedure readMatTips(Var data     : phyloPtr;
                      trait        : integer;
                      Var tipCount : integer;
                      reread       : boolean);


Implementation

Uses crt;

{***************************************************************************}
{ This procedure writes coprocessor extended reals to a file without the    }
{ extra exponent and mantessa digits.                                       }
{***************************************************************************}
Procedure extendWrite(Var outFile :text;num :extended);
   Var outStr : string[19];
   Begin
      str(num:19:-11,outStr);
      delete(outStr,16,2);
      write(outFile,outStr);
   End; {procedure extendwrite}


{***************************************************************************}
{ The procedure handles the entire process of reading a tree.  It prompts   }
{ the filename, checks that the file exists, reads the tree and performs    }
{ some basic checks (for example that no two nodes have the same name).  It }
{ returns a pointer to the root, the filename, the extension component of   }
{ the filename, and succ if it did input a file.                            }
{***************************************************************************}
Procedure inputTree(Var data : phyloPtr;
                    Var succ : boolean;
                    reread   : boolean); { new re-read feature PEM 28 Apr 1996}
   Var
      fl       : text;
      ch       : char;
      me       : nodePtr;
      dumptr   : nodePtr;
      ok, good : boolean;
      name     : namestring;
   {************************************************************************}
   { This procedure reads in a file name.  The complexity is due to all the }
   { error checking required.                                               }
   {************************************************************************}
   Procedure getInFile(Var flename : flstring;
                       Var exten   : typstring;
                       Var succ    : boolean);
      Var
         ch            : char;
         place         : byte;
         exists, valid : boolean;
         i             : integer;
      Begin
         exists := true;
         valid := true;
         Repeat
            If Not exists
            Then writeln(flename,' does not exist.');
            If Not valid
            Then writeln(flename,' is not of a valid type.');
            exists := true;
            valid := true;
            writeln('Please input the name of the input file with ');
            writeln('the .INP, .PDI, .BRK, or .DIS ending.');
            writeln('   Or "R" for return to Main Menu');
            readln(flename);
            place:=pos('.',flename);
            If place = 0    {If no . is found in the filename}
            Then valid:=false
            Else Begin
               exten := copy(flename,place+1,3);
               For i := 1 To 3 Do exten[i] := upcase(exten[i]);
               If (exten<>'INP') And (exten<>'PDI') And (exten<>'BRK')
               Then valid:=false {If extension bad}
               Else Begin                 {check to see if the file exists}
                  {$I-}
                  assign(fl,flename);
                  reset(Fl);
                  close(Fl);
                  {$I+}
                  exists := (IOResult=0) and (flename<>'');
               End {else}
            End {else}
         Until (exists and valid) or (flename ='R') or (flename='r');
         If (flename ='R') Or (flename='r')    {If user wants to abort}
         Then Begin
            succ := false;
         End; {then}
      End; {local procedure getinfile}

   {************************************************************************}
   {This small function handles part of the error reporting for the file
    verification procedures.  PEM 2 Nov 1999.                               }
   {************************************************************************}
   Function badFileReport(typStr,firstLine,secondLine : string) : boolean;
     Var userChar : char;
     Begin
        writeln('This does not seem to be a valid ' + typStr + ' file.');
        writeln('The first line was:');
        writeln(firstLine);
        writeln;
        writeln('The second line was:');
        writeln(secondLine);
        writeln;
        writeln('Do you want to try to read it as a ' + typStr + ' file anyway (Y/N=Enter)? ');
        userChar := readKey;
        badFileReport := userChar in ['Y','y']
     End; {function badFileReport}
   {************************************************************************}
   { This procedure will read the tree into memory if it is in a PDI file.  }
   { It requires the file text variable and returns rt, the pointer to the  }
   { root of the tree it just read in.  It also sets fail to true if an     }
   { an error occured.                                                      }
   {************************************************************************}
   Procedure ReadPDI(Var fl :text;Var rt :nodePtr;var fail:boolean);
      Var
         me :nodeptr;
         valCode : integer;
         nextName : string;
         ancestorName : string;
         lengthStr : string;

     {************************************************************************}
     { VerifyPDI validates a presumed PDI file.  It expects an integer on the }
     { first line, followed by a blank line.                                  }
     {************************************************************************}
      Function verifyPDI(Var fl : text) : boolean;
         Var i : integer;
             firstLine : string;
             secondLine : string;
             userChar : char;
             badVal : boolean;
         Begin
            reset(fl);
            readln(fl,firstLine);
            readln(fl,secondLine);
            i := 1;
            badVal := false;
            Repeat
               If (ord(firstLine[i]) < ord('0')) or
                   (ord(firstLine[i]) > ord('9'))
               Then badVal := true
               Else inc(i)
            Until badVal or (i > length(firstLine));
            If badVal or (length(secondLine)>0)
            Then verifyPDI := badFileReport('PDI',firstLine,secondLine)
            Else verifyPDI := True;
         End; {local function verifyPDI}
      {****************************************************************}
      {This procedure reads a space delimited token from source.  The  }
      {token is returned in tok, which is set to a maximum length of   }
      {tlength.                                                        }
      {****************************************************************}
      Procedure getToken(Var source:text; Var tok:string; tlength:byte);
         Var
            count : byte;
            curChar : char;
            done : boolean;
         Begin
            count := 0;
            done := eoln(source);
            While not(done) Do Begin
               read(source,curChar);
               If (curChar = ' ') Then Begin
                  If (count > 0) Then Begin {we have a token, finish}
                     done := true;
                  End
               End
               Else Begin
                  inc(count);
                  tok[count]:= curChar;
               End;
               done := done Or eoln(source);
            End; {while}
            If count<tlength
            Then tok[0] := chr(count)
            Else tok[0] := chr(tlength);
         End; {local procedure getToken}
      {********************************************************************}
      Begin
         If verifyPDI(fl) Then Begin
            reset(fl);
            readln(fl);
            readln(fl);
            initme(rt);
            rt^.ancestor:=nil;
            rt^.length:=0.0;
            readln(fl,rt^.name);       {only read the root name, rest is trash}
            getToken(fl,nextName,2);
            While nextName<>'' Do Begin    {reading the tree block}
               getToken(fl,ancestorName,2);
               fail:=true;
               checkName(rt,ancestorName,fail,me);
               If fail Then Begin
                  gotoxy(1,21);
                  writeln('Ancestor name "'+ancestorName+'" not found in tree block. ');
                  writeln('File is invalid.');
                  killTree(rt);
                  rt := nil;   {dispose doesn't set the pointer to NIL!}
                  exit;
               End; {then}
               If me^.sibL=nil Then Begin
                  new(me^.sibL);
                  initMe(me^.sibL);
                  me^.sibL^.ancestor:=me;
                  me:=me^.sibL;
               End {then}
               Else Begin
                  new(me^.sibR);
                  initMe(me^.sibR);
                  me^.sibR^.ancestor:=me;
                  me:=me^.sibR;
               End; {else}
               me^.name:=nextName;
               me^.length := 0.0;
               getToken(fl,lengthStr,32);  {length doesn't really matter here}
               val(lengthStr,me^.length,valCode);
               If valCode <> 0 Then Begin
                  writeln('Error - Bad length for branch leading to ', me^.name, '.');
                  writeln('        Length string read was: ',lengthStr);
                  writeln('        Will set branchlength to zero; you should fix the .PDI file');
                  me^.length := 0;
               End; {then}
               readln(fl);              {flush the rest of the line}
               getToken(fl,nextName,2);
            End; {while}
            readln(fl);
            getToken(fl,nextName,2);
            While nextName<>'' Do Begin {reading the tip data block}
               fail:=true;
               checkName(rt,nextName,fail,me);
               If fail Then Begin
                  gotoxy(1,21);
                  writeln('Tip name "'+nextName+'" not found in tip value block. ');
                  writeln('File is invalid.');
                  killTree(rt);
                  rt := nil;   {dispose doesn't set the pointer to NIL!}
                  exit;
               End;
               readln(fl,me^.dat[1],me^.dat[2]);  {keep this for now}
               getToken(fl,nextName,2);
            End; {while}
         End {then}
         Else fail := true;
         close(fl)
      End; {procedure readPDI}
   {***********************************************************************}
   { This procedure reads a tree, but from a file in the .INP format.      }
   {***********************************************************************}
   Procedure ReadINP(Var fl :text; Var rt :nodeptr; Var fail :boolean);
      Var
         me      : nodePtr;
         ch      : char;
         ch2     : string[2];
         chhh    : char;
         ch4     : string[4];
         tempint : integer;
      {*********************************************************************}
      { This procedure reads the name and length of the sibling of a node.  }
      {*********************************************************************}
      Procedure readSib(Var fl: text; sib,me:nodeptr);
         Var
            tempReal : extended;
            ch2      : string[2];
            ch       : char;
         Begin
            sib^.ancestor:=me;
            read(fl,sib^.name);
            read(fl,ch);
            read(fl,sib^.length);
         End; {local procedure readsib}
      {*********************************************************************}
      { This procedure checks a node's name for duplication, then reads a   }
      { real number value into one of the node's trait fields.              }
      {*********************************************************************}
      Procedure readTrait(i        : integer;
                          rt       : nodePtr;
                          Var fl   : text;
                          ch2      : string;
                          Var fail : boolean);
         Var
            tempReal :extended;
            me :nodeptr;
         Begin
            fail:=true;
            checkname(rt,ch2,fail,me);
            If fail Then exit;  {found a problem, bail out}
            read(fl,tempReal);
            me^.dat[i]:=tempReal;
         End; {readtrait}
      {********************************************************************}
      Begin {procedure readinp}
         reset(fl);
         Repeat
            read(fl,ch4);
            readln(fl);
         Until ch4 = '  of';
         initMe(rt);
         rt^.ancestor:=nil;
         rt^.length:=0.0;
         read(fl,ch2);
         rt^.name:=ch2;
         While Not eof(fl) Do Begin
            fail:=true;
            checkName(rt,ch2,fail,me);
            If fail Then Begin    
               ret('Tree in file is invalid(1).  Please press any key '+
                     'to return to the Main Menu.');
               killTree(rt);
               exit; { bail out }
            End; {then}
            new(me^.sibL);
            initMe(me^.sibL); 
            read(fl,ch);
            readSib(fl,me^.sibL,me);
            read(fl,tempint);
            read(fl,ch);
            new(me^.sibR);
            initMe(me^.sibR);
            readSib(fl,me^.sibR,me);
            readln(fl);
            read(fl,ch2);
         End; {while}
         close(fl);
         reset(fl);
         Repeat
            read(fl,ch);
            readln(fl);
         Until ch='-';
         read(fl,ch2);
         While Not eoln(fl) Do Begin
            readTrait(1,rt,fl,ch2,fail);
            If fail Then Begin
               ret('Tree in file is invalid(2).  Please press any key '+
                     'to return to the Main Menu.');
               exit; {bail out}
            End;
            read(fl,ch);
            read(fl,ch2);
            read(fl,ch2);
            readTrait(2,rt,fl,ch2,fail);
            If fail Then Begin  
               ret('Tree in file is invalid(3).  Please press any key '+
                    'to return to the Main Menu.');
               exit; {bail out to menu}
            End; {then}
            readln(fl);
            read(fl,ch2);
         End; {while not eoln}
         close(fl);
      End; {local procedure readinp}
   {***********************************************************************}
   {This procedure reads in a tree of the .BRK format.  It has 2 major
   problems.  The .BRK format does not necessarily have a space after a
   number so I had to write the procedure readnum (below) to read a number
   character by character.  Second, while this program must have names
   associated with internal nodes, the .BRK format does not save them.
   therefore, I had to write the procedure getname (below) that simply
   gives me a 2 character name.  A lesser problem is that .BRK supports
   tip names longer than 2 characters while PDTREE and PDSIMUL support only
   2 characters.  This procedure is much different in character than the last
   two because it simply parses the input, reading it like a set of
   instructions one at a time.  Each of the characters ( ) : , is
   viewed as an order to do something so my algorithm simply inputs
   characters one at a time doing what each character says.  It ignores
   white space, and will skip over any number of lines in the beginning until
   a ( is found at the start of a line.  To get a feel for how this format
   works, print out a .BRK file, the tree from the edit module and see how
   they correleate.

   The .BRK format has 2 files that are saved along with it.  The .TIP file
   has the Tip data, and the .NAM file has the tip names followed by zeroes.
   If you do not see a need for the .NAM file ignore it, it is produced as a
   convenience to people who use MacClade.  This program only reads in the
   .BRK and the .TIP files.}
   {***********************************************************************}
   Procedure readBRK(Var fl : text;Var rt : nodePtr;Var fail : boolean);
      Var
         nd,temp    : nodeptr;
         ch         : char;
         nm         : namestring;
         namenum    : integer;
         n          : integer;
         pleaseRead : boolean;
         ok         : boolean;
      {********************************************************************}
      { This procedure reads in characters one at a time, skipping white   }
      { space until a number is input, then it collects the characters of  }
      { the number until a non number character is input.  It finally,     }
      { converts the number's characters into the number and returns it in }
      { the varaible num and returns the last character input.             }
      {********************************************************************}
      Procedure readNum(Var fl:text; Var num:extended; Var ch:char);
         Var
            st    : string;
            n,dum : integer;
            done  : boolean;
         Begin
            st:='';
            Repeat
               read(fl,ch);
               n:=ord(ch);
            Until((ch<>' ') And (n<>9) And (n<>10) And (n<>13));  {whitespace}
            done:=false;
            Repeat
               If (((n>47) And (n<58)) Or (ch='.') Or (ch='+') Or (ch='-') Or
                    (ch='e') Or (ch='E'))   {chr(47)=0, chr(58)=9}
               Then Begin
                  st:=st+ch;
                  read(fl,ch);
                  n:=ord(ch);
               End {then}
               Else done:=true;
            Until done;
            val(st,num,dum);
         End; {local procedure readnum}
      {********************************************************************}
      { This procedure generates a 2 character node name.  It only uses    }
      { upper case characters and the digits 0-9.  It is deterministic and }
      { simply returns the first unused name after the "num"th name in a   }
      { sequence that starts with A and ends with 9.  It increments num as }
      { necessary so it points to an unexplored name when the procedure    }
      { returns.                                                           }
      {********************************************************************}
      Procedure getname(var nm:namestring; rt:nodeptr; var num:integer);
         Var
            n1,n2 : integer;
            where : nodeptr;
            ch    : char;
            nm2   : namestring;
         Begin
            nm:='';
            n1:=num div 36;
            n2:=num mod 36;
            If n1<26
            Then nm2:=chr(65+n1)       {First char will be a letter}
            Else nm2:=chr(48+n1-26);   {First char will be a number}
            If n2<26
            Then nm2:=nm2 + chr(65+n2)     {Second char will be a letter}
            Else nm2:=nm2 + chr(48+n2-26); {Second char will be a number}
            num:=num+1;
            ok:=true;      
            checkName(rt,nm2,ok,where); {check if the name already exists}
            nm:=nm2;
            If (not ok) {try again if it does}
            Then getName(nm,rt,num); {turbo pascal doesn't optimize tail recursion!}
         End; {local procedure getname}
      {********************************************************************}
      Begin {procedure readbrk}     
         initMe(rt);
         nameNum:=0;
         pleaseRead:=true;
         reset(fl); 
         read(fl,ch);
         If ch<>'(' Then Repeat  {Ignore the meaningless lines at the top}
            If (ord(ch)<>13) And (ord(ch)<>10)  {if line has characters left}
            Then readln(fl);  {throw it away}
            read(fl,ch);
         Until((ch='(') Or (ord(ch)=26));
         If (ord(ch)=26) Then Begin    {chr(26)= end of file}
            fail:=true;
            ret('Tree file is invalid. Please press any key');
            exit;
         End; {then}
         rt^.ancestor:=nil;
         rt^.length:=0;
         rt^.name:='rt';
         nd:=rt;     {nd points to the node that we are working on.}
         Repeat
            If pleaseRead
            Then read(fl,ch)
            Else pleaseRead:=true;
            If ch='(' Then Begin  {Add interior node above the current node, nd.}
               If nd^.sibL=nil Then Begin {Do the left one unless it is already occupied.}
                  new(nd^.sibL);
                  initMe(nd^.sibL);
                  nd^.sibL^.ancestor:=nd;
                  nd:=nd^.sibL;
                  getName(nd^.name,rt,nameNum);
                  nd^.length:=0;
               End {then}
               Else Begin   {If the left is not free, then the right one is.}
                  new(nd^.sibR);
                  initMe(nd^.sibR);
                  nd^.sibR^.ancestor:=nd;
                  nd:=nd^.sibR;
                  getName(nd^.name,rt,namenum);
                  nd^.length:=0
               End {else}
            End {then}
            Else
               If ch=',' Then Begin    {Drop down one node in the tree.}
                  nd:=nd^.ancestor;
                  If nd^.sibR<>nil
                  Then     {If there is a trifurcation or more.}
                     If nd=rt Then Begin  {Add new root, put tree in sibL.}
                        new(nd^.ancestor);
                        initMe(nd^.ancestor);
                        nd^.ancestor^.sibL:=nd;
                        nd:=nd^.ancestor;
                        rt:=nd;
                        getName(nd^.sibL^.name,rt,namenum);
                        nd^.name:='rt';
                        nd^.length:=0;
                        nd^.ancestor:=nil;
                     End {then}
                     Else Begin {Put new node inbetween nodes, and put clade in sibL.}
                        new(temp);
                        initMe(temp);
                        temp^.ancestor:=nd^.ancestor;
                        If(nd^.ancestor^.sibL=nd)
                        Then nd^.ancestor^.sibL:=temp
                        Else nd^.ancestor^.sibR:=temp;
                        nd^.ancestor:=temp;
                        temp^.sibL:=nd;
                        temp^.length:=nd^.length;
                        nd^.length:=0;
                        getName(temp^.name,rt,namenum);
                        nd:=temp;
                     End {else}
                  End {then}
                  Else
                     If ch=')' Then Begin  {Drop down a level.}
                        If nd^.ancestor<>nil Then nd:=nd^.ancestor;
                     End
                     Else
                        If ch=':' Then Begin {input a number and interpret as branch lenght}
                           readNum(fl,nd^.length,ch);
                           pleaseRead := false;
                        End
                        Else Begin
                           n:=ord(ch);
                           If (n<>9) And (n<>10) And (n<>13) And (n<>32) And
                               (ch<>';')
                           Then Begin  {If not whitespace, must be the name of a tip node}
                              pleaseRead:=false;
                              nm:='';
                              Repeat
                                 If (length(nm)<2) And (((n>47) And (n<58)) Or
                                            ((n>64) And (n<91)) Or
                                          ((n>96) And (n<123)))
                                 Then nm := nm+ch;
                                 If (length(nm)=2) Then Begin
                                    checkName(rt,nm,ok,temp);
                                    If (Not ok) Then nm := copy(nm,1,1);
                                 End; {then}
                                 read(fl,ch);
                                 n:=ord(ch);
                              Until(ch=':');
                              If (length(nm)=2) Then Begin   
                                 checkName(rt,nm,ok,temp);
                                 If (Not ok) Then getName(nm,rt,namenum);
                           End; {then}
                           If nd^.sibL=nil Then Begin {Do the left one unless it is already occupied.}
                              new(nd^.sibL);
                              initMe(nd^.sibL);
                              nd^.sibL^.ancestor:=nd;
                              nd:=nd^.sibL;
                              nd^.name:=nm;
                              nd^.length:=0;
                           End {then}
                           Else Begin {If the left is not free, then the right one is.}
                              new(nd^.sibR);
                              initMe(nd^.sibR);
                              nd^.sibR^.ancestor:=nd;
                              nd:=nd^.sibR;
                              nd^.name:=nm;
                              nd^.length:=0;
                           End; {else}
                        End; {else}
               End; {then}
         Until ch=';';
         fail:=false;
         close(fl);
      End; {procedure readbrk}
   {***********************************************************************}
   { This procedure attempts to read the .TIP file to get the tip data     }
   { for the tree just read in in the .BRK format.  If the file does not   }
   { exist a warning is issued and all the data remains zero.              }
   {***********************************************************************}
   Procedure tryreadTIP(flnm : flString; rt : nodePtr);
      Var
         name : nameString;
         fl   : text;
         good : boolean;
      {********************************************************************}
      { This procedure is passed a tip name.  It recursively searches the  }
      { tree for a node with a name that matches the name.  When it finds  }
      { the tip, it reads a pair of trait values for the tip from the      }
      { file.                                                              }
      {********************************************************************}
      Procedure inputData(nd       : nodePtr;
                          name     : flString;
                          Var fl   : text;
                          Var good : boolean);
         Begin
            If (nd^.sibL=nil) And (nd^.name=name) Then Begin
               good := true;
               readln(fl,nd^.dat[1],nd^.dat[2]);
            End; {then}
            If (nd^.sibL<>nil) Then Begin
               inputData(nd^.sibL,name,fl,good);
               If Not good Then inputData(nd^.sibR,name,fl,good)
            End {then}
         End; {local procedure inputData}
      {********************************************************************}
      Begin {procedure tryreadtip}
         flnm:=flnm+'TIP';
         {$I-}     {check to see if the file exists}
         assign(fl,flnm);
         reset(Fl);
         close(Fl);
         {$I+}
         If ((IOResult=0) And (flnm<>'')) Then Begin
            reset(fl);
            Repeat
               read(fl,name);
               good:=false;
               inputData(rt,name,fl,good);
               If Not good Then Begin   {If the tip not found, data not read}
                  readln(fl);
                  writeln('Tip ',name,' not found.  Its data was skipped.');
               End {then}
            Until Eof(fl);
         End {then}
         Else Begin
            writeln('The .TIP file was not found.  The .TIP file should have',
                     ' had the tip data.');
            writeln('Therefore, all the tip data of the tree just read in ',
                     'remains zero.');
         End; {else}
      End; {procedure tryreadtip}
   {************************************************************************}
   { The purpose of this procedure is to make sure that no two nodes in the }
   { tree have the same name.  If this problem is found, the error is       }
   { reported and the user will be prompted for a new name for the node     }
   { where the duplication was detected.  Passing this procedure            }
   { successfully does not guarrantee the tree is correct, but this is the  }
   { only checking done after a tree has been successfully parsed.          }
   {************************************************************************}
   Procedure doubleCheckNames(rt,nd        : nodePtr;
                              Var notfound : boolean;
                              Var bdnd     : nodePtr);
      Var
         nm : string[2];
         dumPtr :nodeptr;
      Begin
         nm:=nd^.name;
         nd^.name:='  ';
         checkName(rt,nm,notfound,dumptr);
         nd^.name:=nm;
         If Not notfound Then bdnd:=nd;
         If notfound And (nd^.sibL<>nil)
         Then doublechecknames(rt,nd^.sibL,notfound,bdnd);
         If notfound And (nd^.sibR<>nil)
         Then doublechecknames(rt,nd^.sibR,notfound,bdnd);
      End; {procedure doublechecknames}
   {***************************************************************************}
   { Here we are at the body of the input procedure.  This procedure is        }
   { responsible for checking whether a file is already in memory, choosing    }
   { the correct loading procedure (based on the user supplied file extension),}
   { and checking that the newly loaded tree does not have nodes with          }
   { duplicate names.                                                          }
   {***************************************************************************}
   Procedure promptAndRead;
      Begin
         clrscr;
         writeln('******************* Input Module ******************');
         writeln;
         writeln;
         ch:='Y';
         If succ=true Then Begin
            Repeat
               writeln('You already have a tree in memory.');
               writeln('It is called: ',data^.sourceName);
               writeln('If you choose to input a different tree, then ',
                         data^.sourceName);
               writeln('will be cleared from memory.');
               writeln;
               writeln('Do you want to input a new tree and remove the ',
                         'current');
               writeln('tree from memory? (Y/N)');
               ch:=upCase(readkey);
               writeln;
            Until ch in ['Y','N'];
            If ch='Y' Then Begin    {added 13-May-95 to fix minor bug PEM}
               killtree(data^.tree);
               data^.tree:=nil
            End {then}
         End; {then}
         If ch='Y' Then Begin
            succ:=true;
            getInFile(data^.sourceName,data^.sourceExt,succ);
            If succ Then Begin
               assign(fl,data^.sourceName);
               new(data^.tree);
               If data^.sourceExt='INP'
               Then readINP(fl,data^.tree,succ);
               If data^.sourceExt='PDI'
               Then readPDI(fl,data^.tree,succ);
               If (data^.sourceExt='BRK') Then Begin
                  readBRK(fl,data^.tree,succ);
                  tryreadTIP(copy(data^.sourceName,1,length(data^.sourceName)-3),
                              data^.tree);
                  writeln;
                  writeln('NOTE: the .BRK format does not have the ',
                            'internal node names stored.  They are');
                  writeln('     generated anew each time the tree is read in.');
               End; {then}
               If (data^.tree <> nil)
               Then writeln(data^.sourceExt,' file read.');
               succ:=not succ;
               If succ Then Begin
                  Repeat
                     succ:=true;
                     doublechecknames(data^.tree,data^.tree,succ,me);
                     If Not succ Then Begin
                        writeln;
                        writeln('A duplicate node name was found.');
                        writeln('Please input a new name for: ',me^.name,'.');
                        ok:=true;
                        good:=true;
                        Repeat
                           If Not ok
                           Then writeln(name,' is not unique.  Try again.');
                           If Not good
                           Then writeln(name,' is not 2 characters.  Please try again.');
                           readln(name);
                           ok:=true;
                           good:=true;
                           If length(name)<2 Then good:=false;
                           checkName(data^.tree,name,ok,dumptr);
                        Until good And ok;
                        me^.name:=name;
                     End; {then}
                  Until succ;
                  writeln;
                  writeln;
                  writeln('File successfully loaded.')
               End {then}
            End {then}
         End; {then}
      End; {procedure promptAndRead }
   {************************************************************************}
   { This is a simplified version of promptAndRead that just loads the file }
   { and checks for duplicate names.  It is intended for re-reading the     }
   { same file. Duplicate names should have been caught when the file was   }
   { read the first time, but I've left the (harmless) check in for now.    }
   {************************************************************************}
   Procedure justRead;
      Begin
         succ := true;
         assign(fl,data^.sourceName);
         new(data^.tree);
         If data^.sourceExt='INP'
         Then readINP(fl,data^.tree,succ);
         If data^.sourceExt='PDI'
         Then readPDI(fl,data^.tree,succ);
         If data^.sourceExt='BRK' Then Begin
            readBRK(fl,data^.tree,succ);
            tryreadTIP(copy(data^.sourceName,1,length(data^.sourceName)-3),data^.tree);
         End; {then}
         {writeln(data^.sourceExt,' file read.');}
         succ:= Not succ;
         If succ Then Repeat
            succ:=true;
            doubleCheckNames(data^.tree,data^.tree,succ,me);
            If not succ Then Begin
               writeln;
               writeln('A duplicate node name was found.');
               writeln('Please input a new name for: ',me^.name,'.');
               ok:=true;
               good:=true;
               Repeat
                  If Not ok Then writeln(name,' is not unique.  ', 'Try again.');
                  If Not good Then writeln(name,' is not 2 characters.  ',
                                            'Please try again.');
                  readln(name);
                  ok:=true;
                  good:=true;
                  If length(name)<2 Then good:=false;
                  checkName(data^.tree,name,ok,dumptr);
               Until good and ok;
               me^.name:=name;
            End; {then}
         Until succ;
         writeln;
         writeln('File successfully loaded.');
         writeln;
      End; { local procedure justread }
    {***********************************************************************}
    Begin {procedure inputTree}
       If (reRead)
       Then justRead
       Else promptAndRead;
    End; {procedure inputtree}

{***************************************************************************}
{ This procedure again is long, but only saves the tree to disk.  It        }
{ requires the root, the current filename, the current extension, and ok    }
{ which is true if there is a tree to save.  It is only called by the main  }
{ program.                                                                  }
{***************************************************************************}
Procedure saveTree(rt :nodePtr;
                   Var fleName :flstring;
                   Var typ :typString;
                   ok :boolean);
   Var
      fl :text;
      ans :char;
      num :integer;
   {************************************************************************}
   { This procedure gets the output filename from the user.  It is          }
   { longer so the program won't crash when the user does silly things.     }
   {************************************************************************}
   Procedure getflnm(Var flnm :flstring; Var typ :typString);
      Var
         good, valid, again :boolean;
         ch :char;
         i :integer;
         place :byte;
      Begin
         writeln('The current name for this tree is: ',flnm);
         writeln;
         writeln('Would you like to change its name before it is saved');
         writeln('so that the old file will not be destroyed? (Y/N)');
         Repeat
            ch:=upCase(readkey);
         Until ch in ['Y','N'];
         good:=true;
         valid:=true;
         again:=false;
         If ch='Y' Then Repeat
            writeln;
            If Not valid Then writeln(flnm,' is not of a valid type.');
            If Not good Then writeln(flnm,' is not a legal filename.');
            writeln('What shall be the new name?  Do not forget the ',
                      '.INP, .PDI, .BRK, or .DIS.');
            writeln('Remember that the extension determines ',
                     'the format in which the tree is saved!');
            writeln;
            writeln('Be aware that .BRK does not save',
                      ' non-tip node names, and puts the tip');
            writeln('data it its own file with a .TIP extension.');
            writeln('In addition, the .BRK format has a third file ',
                      'with a .NAM extention that is');
            writeln('rather special purpose.  Just be aware that it exists.');
            readln(flnm);
            good:=true;
            valid:=true;
            again:=false;
            place:=pos('.',flnm);
            If place = 0
            Then valid:=false
            Else Begin
               typ:=copy(flnm,place+1,3);
               For i:=1 To 3 Do typ[i]:=upcase(typ[i]);
               If (typ<>'INP') And (typ<>'PDI') And (typ<>'BRK')
               Then valid:=false
               Else Begin
                  {$I-}
                  assign(fl,flnm);
                  reset(fl);
                  close(fl);
                  {$I+}
                  If (IOresult=0) And (flnm<>'') Then Begin
                     writeln;
                     writeln(flnm,' already exits.');
                     writeln('Do you want to delete it and ',
                               'save the current tree in its ',
                               'place? (Y/N)');
                     Repeat
                        ch:=UpCase(readkey);
                     Until ch in['Y','N'];
                     If ch='N' Then again:=true
                  End; {then}
                  {$I-}
                  assign(fl,flnm);
                  rewrite(fl);
                  close(fl);
                  {$I+}
                  If (IOresult<>0) Or (flnm='') Then good:=false
               End {else}
            End; {else}
         Until good and valid and not again
      End; {local procedure getflnm}
   {*********************************************************************}
   { This procedure writes the tips data to the file in the PDI format   }
   { recursively.                                                        }
   {*********************************************************************}
   Procedure wrtTips(nd :nodeptr; Var fl :text);
      Begin
         If nd^.sibL=nil
         Then writeln(fl,nd^.name,'  ',nd^.dat[1],'  ',nd^.dat[2])
         Else Begin
            wrtTips(nd^.sibL,fl);
            wrtTips(nd^.sibR,fl)
         End {else}
      End; {procedure wrttips}
   {*********************************************************************}
   { You may have noticed that this series of procedures is analogous to }
   { the ones in the read file block.  This one is in charge of writing  }
   { the tree to the file.  One difference is that this uses recursive   }
   { procedures of its own to do a lot of the work.                      }
   {*********************************************************************}
   Procedure writePDI(rt :nodePtr; flName :flString);
      Var
         num :integer;
         fl  :text;
      {******************************************************************}
      { This procedure writes the tree structure to a file in the PDI    }
      { format recursively.                                              }
      {******************************************************************}
      Procedure wrtTree(nd :nodePtr; Var fl :text);
         Begin
            write(fl,nd^.name,'  ');
            If nd^.ancestor=nil
            Then write(fl,'$$')
            Else write(fl,nd^.ancestor^.name);
            write(fl,'  ');
            extendWrite(fl,nd^.length);
            writeln(fl);
            If nd^.sibL<>nil Then Begin
               wrtTree(nd^.sibL,fl);
               wrtTree(nd^.sibR,fl)
            End {then}
         End; {local procedure wrtTree}
      {*********************************************************************}
      Begin
         Assign(fl,flname);
         rewrite(fl);
         num:=0;
         numtps(rt,num); 
         writeln(fl,num);
         writeln(fl);
         wrttree(rt,fl);
         writeln(fl);
         wrttips(rt,fl);
         close(fl);
      End; {local procedure writePDI}
   {************************************************************************}
   { This one writes the tree in the INP format.  Again, it uses two        }
   { recursive procedures to help it.                                       }
   {************************************************************************}
   Procedure writeINP(rt :nodePtr; flName :flString);
      Var num :integer;
      {********************************************************************}
      { This recursive procedure writes the tips data for INP format.      }
      {********************************************************************}
      Procedure writeTips(nd :nodePtr; Var fl :text);
         Begin
            If nd^.sibl=nil
            Then writeln(fl,nd^.name,' ',nd^.dat[1],'   ',
                           nd^.name,' ',nd^.dat[2])
            Else Begin
               writeTips(nd^.sibL,fl);
               writeTips(nd^.sibR,fl)
            End {else}
         End; {local procedure writetips}
      {********************************************************************}
      { This one writes the tree structure in the INP format.  Yeah, it is }
      { recursive.                                                         }
      {********************************************************************}
      Procedure writeTree(nd :nodePtr; Var fl :text);
         Var
            num :integer;
         Begin
            If nd^.sibL<>nil Then Begin
               write(fl,nd^.name,' ',nd^.sibL^.name,' ',nd^.sibL^.length);
               num:=0;
               numTps(nd^.sibL,num);
               write(fl,' ',num,' ',nd^.sibR^.name,' ',nd^.sibR^.length,' ');
               num:=0;
               numTps(nd^.sibR,num);
               write(fl,num,' ');
               If nd^.ancestor<>nil
               Then writeln(fl,nd^.ancestor^.name,' ',nd^.length)
               Else writeln(fl,'**  0.0000000000E+00');
               writetree(nd^.sibL,fl);
               writetree(nd^.sibR,fl)
            End {then}
         End; {local procedure writetree}
      {*********************************************************************}
      Begin {procedure wrtinp}
         assign(fl,flname);
         rewrite(fl);
         writeln(fl,'This is a file of descriptive and input data to',
                                    ' be used in the');
         writeln(fl,'  comparative method analysis programs CMSINGLE,',
                              ' CMSIMUL, CMTPANAL');
         writeln(fl,'  CMFLANAL, CMMEANAL, and CMTRANAL.');
         num:=0;
         numTps(rt,num);
         writeln(fl,'There are ',num,' tip values on this phylogeny.  Their');
         writeln(fl,'  measured values are :');
         writeln(fl,'    Trait A             Trait B');
         writeln(fl,'Name     Value      Name     Value');
         writeln(fl,'-------------------------------------------');
         writetips(rt,fl);
         writeln(fl);
         writeln(fl,'The phylogeny structure may be described by',
                      ' the following sequences');
         writeln(fl,'  of tips, nodes and branch lengths:');
         writeTree(rt,fl);
         close(fl)
      End; {procedure wrtinp}
   {************************************************************************}
   { This procedure writes the .BRK file out to disk.  (See the             }
   { documentation on the readBRK procedure above).  It is much easier to   }
   { write than to read.                                                    }
   {************************************************************************}
   Procedure writeBRK(rt :nodePtr; flName :flString);
      Var
         fl:text;
      {*********************************************************************}
      { This recursive procedure does all the work.  It traverses the tree, }
      { and writes out the .BRK representation of its topology with branch  }
      { lengths and tip node names.                                         }
      {*********************************************************************}
      Procedure writeme2(nd,rt :nodeptr; Var fl :text);
         Begin
            If nd^.sibL=nil
            Then write(fl,nd^.name,':',nd^.length,' ')
            Else Begin
               If (nd^.length<>0) Or (nd=rt) Then Begin
                  writeln(fl,'(');
                  writeme2(nd^.sibL,rt,fl);
                  write(fl,',');
                  writeme2(nd^.sibR,rt,fl);
                  writeln(fl,')',':',nd^.length)
               End {then}
               Else Begin
                  writeme2(nd^.sibL,rt,fl);
                  write(fl,',');
                  writeme2(nd^.sibR,rt,fl)
               End {else}
            End {else}
         End; {local procedure writeme2}
      {********************************************************************}
      Begin {procedure writebrk}
         assign(fl,flname);
         rewrite(fl);
         writeln(fl);
         writeme2(rt,rt,fl);
         writeln(fl,';');
         close(fl)
      End; {local procedure writebrk}
   {************************************************************************}
   { This procedure writes the tips names to the file in the PDI format. It }
   { follows the names with a tab and a zero.                               }
   {************************************************************************}
   Procedure writeNAM(rt :nodePtr; flname :flString);
      Var
         fl :text;
         place :integer;
      {*********************************************************************}
      { This recursive procedure helps the one it is in do its job.         }
      {*********************************************************************}
      Procedure wrtname(nd :nodePtr; Var fl :text);
         Begin
         If nd^.sibL=nil Then writeln(fl,nd^.name,chr(9),'0')
         Else Begin
            wrtname(nd^.sibL,fl);
            wrtname(nd^.sibR,fl)
         End {else}
      End; {local procedure wrtname}
      {********************************************************************}
      Begin {procedure writeNam}
         place:=pos('.',flname);
         flname:=copy(flname,1,place)+'NAM';
         assign(fl,flname);
         rewrite(fl);
         wrtname(rt,fl);
         close(fl);
      End; {local procedure writenam}
   {************************************************************************}
   { This procedure writes the .TIP file that is saved along with the .BRK  }
   { and .NAM files.  The .TIP file holds the tip data in the same format   }
   { as the .PDI file saves the tip data.                                   }
   {************************************************************************}
   Procedure writeTIP(rt :nodePtr; flName :flString);
      Var
         fl :text;
         place :integer;
      Begin
         place:=pos('.',flname);
         flname:=copy(flname,1,place)+'TIP';
         assign(fl,flname);
         rewrite(fl);
         wrtTips(rt,fl);
         close(fl);
      End; {local procedure writetip}
   {************************************************************************}
   Begin {procedure savetree}
      clrscr;
      writeln('********************* Save Module *******************');
      writeln;
      writeln;
      If Not ok
      Then writeln('You must have a tree in memory to save it to disk.')
      Else Begin
         getflnm(flename,typ);
         If typ='INP'
         Then writeINP(rt,flename);
         If typ='PDI'
         Then writePDI(rt,flename);
         If typ='BRK' Then Begin
            writeNAM(rt,flename);
            writeTIP(rt,flename);
            writeBRK(rt,flename)
         End; {then}
         writeln('The Save was successful.')
      End; {else}
      ret('Please press any key to return to the Main Menu.')
   End; {procedure savetree}

{***************************************************************************}
{ This procedure reads a matrix from a .PRB file.  The structure is similar }
{ to the code for reading a tree, but since there is only one possible      }
{ matrix file format, instead of three, things are simpler.                 }
{***************************************************************************}
Procedure readMatrix(Var data : phyloPtr;
                     Var succ : boolean;
                     reread : boolean);
   Var
      fl : text;
      ch : char;
      me : nodePtr;
      dumptr : nodePtr;
      ok, good : boolean;
      name : nameString;
   {************************************************************************}
   { This procedure reads the file name.  Although similar to the procedure }
   { of the same name in readTree, it is simpler because there is only one  }
   { format to support.                                                     }
   {************************************************************************}
   Procedure getInFile(Var flename :flString;
                       Var exten :typString;
                       Var succ :boolean);
      Var
         ch :char;
         place :byte;
         exists, valid :boolean;
         i :integer;
      Begin
         exists:=true;
         valid:=true;
         Repeat
            If Not exists Then writeln(flename,' does not exist.');
            If Not valid Then writeln(flename,' is not of a valid type.');
            exists:=true;
            valid:=true;
            writeln('Please input the name of the input file with ');
            writeln('the .PRB ending.');
            writeln('   Or "R" for return to Main Menu');
            readln(flename);
            place:=pos('.',flename);
            If place = 0              {If no . is found in the filename}
            Then valid:=false
            Else Begin
               exten:=copy(flename,place+1,3);
               For i:=1 To 3 Do exten[i]:=upcase(exten[i]);
               If (exten<>'PRB')           {If extension bad}
               Then valid:=false
               Else Begin                  {check to see if the file exists}
                  {$I-}
                  assign(fl,flename);
                  reset(Fl);
                  close(Fl);
                  {$I+}
                  exists:= (IOResult=0) and (flename<>'')
               End {else}
            End; {else}
         Until (exists And valid) Or (fleName ='R') Or (fleName='r');
         If (flename ='R') or (flename='r')     {If user wants to abort}
         Then Begin
            succ:=false;
            ret('Please press any key to return to the Main Menu.')
         End; {then}
      End; {local procedure getinfile}
   {***********************************************************************}
   { This Procedure deals with matrix files usually .PRB files which       }
   { contain permutation probabilities.  It should also be able to read    }
   { .DST files, if this is needed in the future.                          }
   {***********************************************************************}
   Procedure readMAT(Var fl :text;
                     Var mat :matPtr;
                     Var fail :boolean);
      Var
         me : nodePtr;
         ch : char;
         ch2 :string[2];
         chhh : char;
         ch4 : string[4];
         tempInt : integer;
         i : integer;
         j : integer;
         myCount : integer;
         lineBuf : string;
         done : boolean;
      Begin
         reset(fl);
         Repeat
            readln(fl,lineBuf);
            ch4 := copy(lineBuf,1,4);
         Until ch4='Sour';
         {Initialize the matrix}
         For i := 1 to matSize Do
             For j := 1 to matSize Do
                 mat^.data[i,j] := 0.0;
         For i := 1 to matSize Do mat^.tips[i] := 0.0;
         mat^.size := 0;

         {read once to find out how big it is (count the lines)}
         myCount := 1;
         done := false;
         readln(fl,lineBuf);
         While ((myCount <= matSize) and not(done)) Do Begin
            mat^.names[myCount] := copy(lineBuf,1,2);
            readln(fl,lineBuf);
            If length(lineBuf) < 2
            Then done := true
            Else myCount := myCount + 1;
         End;
         mat^.size := myCount;
         close(fl);
         { now that we know how big the matrix is, start over to read the }
         { elements.                                                      }
         reset(fl);
         Repeat
            readln(fl,lineBuf);
            ch4 := copy(lineBuf,1,4);
         Until ch4='Sour';
         For i := 1 to mat^.size Do Begin
            read(fl,ch4);
            For j := 1 to mat^.size Do Begin
               read(fl,mat^.data[i,j]);
            End; {for j}
            readln(fl);
         End; {for i}
         close(fl);
      End; {local procedure readmat}

   {************************************************************************}
   { This is similar to the procedure of the same name in readTree, except  }
   { that it asks about loading a matrix rather than a tree.                }
   {************************************************************************}
   Procedure promptAndRead;
      Begin
         clrscr;
         writeln('******************* Input Module ******************');
         writeln;
         writeln;
         ch:='Y';
         If succ=true Then Begin
            Repeat
               If (data^.tree <> nil)
               Then writeln('You already have a tree in memory.')
               Else writeln('You already have a matrix in memory.');
               writeln('It is called: ',data^.sourceName);
               If (data^.tree <> nil)
               Then write('If you choose to input a different tree, then ')
               Else write('If you choose to input a different matrix, then ');
               writeln(data^.sourceName);
               writeln('will be cleared from memory.');
               writeln;
               If (data^.tree <> nil)
               Then Begin
                  writeln('Do you want to input a new tree and remove the current');
                  writeln('tree from memory? (Y/N)');
               End {then}
               Else Begin
                  writeln('Do you want to input a new matrix and remove the current');
                  writeln('matrix from memory? (Y/N)');
               End; {else}
               ch:=UpCase(readkey);
               writeln
            Until ch in ['Y','N'];
            If ch='Y' Then     {added 13-May-95 to fix minor bug PEM}
               If (data^.tree <> nil) Then Begin
                  dispose(data^.tree);
                  data^.tree := nil
               End
               Else Begin
                  dispose(data^.matrix);
                  data^.matrix := nil
               End {else}
         End; {then}
         If ch='Y' Then Begin
            succ:=true;
            getinfile(data^.sourceName,data^.sourceExt,succ);
            If succ Then Begin
               assign(fl,data^.sourceName);
               new(data^.matrix);
               readMAT(fl,data^.matrix,succ);
               writeln(data^.sourceExt,' file read.');
               succ:=not succ;
               If succ Then Begin
                  writeln;
                  writeln;
                  writeln('File successfully loaded.')
               End {then}
            End {then}
         End {then}
      End; {local procedure promptAndRead }

   {************************************************************************}
   { This is similar to the procedure of the same name in readTree, except  }
   { that it loads a matrix rather than a tree.  It also doesn't check for  }
   { duplicate node names.                                                  }
   {************************************************************************}
   Procedure justRead;
      Begin
         succ := true;
         assign(fl,data^.sourceName);
         new(data^.matrix);
         readMat(fl,data^.matrix,succ);
         writeln(data^.sourceExt,' file read.');
      End; {local procedure promtAndRead }
   {************************************************************************}
   Begin {procedure inputMat}
      If (reRead)
      Then justRead
      Else promptAndRead;
   End; {procedure inputMat}

{****************************************************************************}
{ Matrix files do not include tip values.  Therefore if PDRANDOM is going to }
{ have something to permute, it needs to get tip values from somewhere.      }
{ This procedure reads tip values from a .PDI file (most sensibly the .PDI   }
{ file that gave rise to the .PRB file, at least prior to the .PRB file      }
{ being editted.                                                             }
{****************************************************************************}
Procedure readMatTips(Var data : phyloPtr;
                      trait : integer;
                      var tipCount : integer;
                      reread : boolean);
   Var fl : text;
       firstTip : nodePtr;
       thisTrait1 : extended;
       thisTrait2 : extended;
       curTip : nodePtr;
       blanks : integer;
       i : integer;
       tipFileExt : typstring;
       lineBuf : string;
       thisName : nameString;
       foundName : boolean;
       succ : boolean;

   {************************************************************************}
   { This procedure reads in a file name.  This file name must have the     }
   { extension .PDI since that is the only tree format that readmattips     }
   { knows how to read.                                                    }
   {************************************************************************}
   Procedure getInFile(Var flename : flString;
                       Var exten :typString;
                       Var succ :boolean);
      Var
         ch :char;
         place : byte;
         exists, valid : boolean;
         i :integer;
      Begin
         exists:=true;
         valid:=true;
         Repeat
            If Not exists Then writeln(flename,' does not exist.');
            If Not valid Then writeln(flename,' is not of a valid type.');
            exists:=true;
            valid:=true;
            writeln('Please enter the name of the .PDI file that contains the tip data');
            writeln('   Or "R" to cancel matrix file loading');
            readln(flename);
            place:=pos('.',flename);
            If place = 0             {If no . is found in the filename}
            Then valid:=false
            Else Begin
               exten:=copy(flename,place+1,3);
               For i:=1 To 3 Do exten[i]:=upcase(exten[i]);
               If (exten<>'PDI')          {If extension bad}
               Then valid:=false
               Else Begin                 {check to see if the file exists}
                  {$I-}
                  assign(fl,flename);
                  reset(Fl);
                  close(Fl);
                  {$I+}
                  exists:= (IOResult=0) and (flename<>'')
               End {else}
            End {else}
         Until (exists and valid) or (flename ='R') or (flename='r');
         If (flename ='R') or (flename='r')    {If user wants to abort}
         Then Begin
            succ:=false;
            ret('Please press any key to return to the Main Menu.')
         End {then}
      End; {local procedure getinfile}
   {*************************************************************************}
   Begin {procedure readMatTips}
      If Not reread
      Then getInFile(data^.matrixTipFile,tipFileExt,succ);
      If succ Then Begin
         With data^.matrix^ Do Begin
            tipCount := size;
            new(firstTip);
            curTip := firstTip;
            initMe(curTip);
            curTip^.tipR := nil;
            curTip^.name := names[1];
            curTip^.dat[trait] := 0.0;
            curTip^.tipNum := 1;
            For i := 2 to size Do Begin
               new(curTip^.tipR);
               curTip := curTip^.tipR;
               initMe(curTip);
               curTip^.tipR := nil;
               curTip^.name := names[i];
               curTip^.dat[trait] := 0.0;
               curTip^.tipNum := i
            End {for i}
         End; {with data^.matrix^}
         assign(fl,data^.matrixTipFile);
         reset(fl);
         blanks := 0;
         Repeat
            readln(fl,lineBuf);
            If length(lineBuf) = 0
            Then blanks := blanks + 1
         Until blanks = 2;
         read(fl,thisName);
         While(thisName <> '') Do Begin
            foundName := false;
            readln(fl,thisTrait1,thisTrait2);
            curtip := firstTip;
            While (curTip <> nil) Do Begin
               If curTip^.name = thisName
               Then Begin
                  If trait = 1
                  Then curTip^.dat[1] := thisTrait1
                  Else curTip^.dat[2] := thisTrait2;
                  curTip := nil;
                  foundName := true
               End {then}
               Else curTip := curTip^.tipR
            End; {while}
            If not(foundName)
            Then writeln('Tip Error: ',thisName);
            read(fl,thisName)
         End; {while}
         close(fl);
         data^.tipList := firstTip
      End {then}
      Else tipCount := 0   {no Good}
   End; {procedure readMatTips}


End. {Unit rand_io1}
