{**************************************************************************}
{This is unit s_io1.  It is used by PDSIMUL to input a tree and report some
of that tree's characteristics.  It uses t_io1 heavily for the actual
reading-the-tree-from-disk part.  This is basically a wrapper for the
procedures in t_io1.}
{**************************************************************************}
unit s_io1;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Interface
   Uses s_util1;

   Procedure doRead(Var parms:parameter);

Implementation
  uses t_io1,
       t_util1,
       te_util1,
       crt,
       graph;

{**************************************************************************}
{This procedure finds the mean of each trait across the tips of the tree.
{**************************************************************************}
Procedure findMean(Var parm:parameter);
   Var
      total : traits;
{**************************************************************************}
{This procedure is used by findmean to find the total of each trait across
the tips.  Findmean then divides by the number of tips to get the mean.}
{**************************************************************************}
   Procedure findTotal(nd :nodeptr; Var total:traits);
      Begin
         If nd^.sibL<>NIL Then Begin
            findtotal(nd^.sibL,total);
            findtotal(nd^.sibR,total);
         End
         Else Begin
            total[1]:=total[1]+nd^.dat[1];
            total[2]:=total[2]+nd^.dat[2];
         End; {else}
      End; {local procedure findtotal}
{***************************************************************************}
   Begin
      total[1]:=0;
      total[2]:=0;
      findtotal(parm.root,total);
      parm.mean[1]:=total[1]/parm.tipnumber;
      parm.mean[2]:=total[2]/parm.tipnumber;
   End; {procedure findmean}

{**************************************************************************}
{This procedure finds the maximum and minimum trait values and returns
them.}
{**************************************************************************}
   Procedure findMaxMin(nd :nodeptr; Var max:traits; Var min:traits);
      Begin
         If nd^.sibL<>NIL Then Begin
            findmaxmin(nd^.sibL,max,min);
            findmaxmin(nd^.sibR,max,min);
         End
         Else Begin
            If nd^.dat[1]>max[1] Then
              max[1]:=nd^.dat[1];
            If nd^.dat[2]>max[2] Then
              max[2]:=nd^.dat[2];
            If nd^.dat[1]<min[1] Then
              min[1]:=nd^.dat[1];
            If nd^.dat[2]<min[2] Then
              min[2]:=nd^.dat[2];
         End; {else}
      End; {procedure findMaxMin}

{**************************************************************************}
{This procedure is used to find the variance of each trait across the tips.
It also reports the covariance.}
{**************************************************************************}
   Procedure findVariance(Var parm:parameter);
      Var
         total : traits;
         cor :real;
         ch :char;

{**************************************************************************}
{This procedure figures out the sum of the squared deviations across the
tips.  This information is then used to find the variance and covariance.}
{**************************************************************************}
      Procedure findDeviations(nd :nodeptr; Var total:traits;Var cor:real;
                               means:traits);
         Begin
            If nd^.sibL<>NIL Then Begin
               finddeviations(nd^.sibL,total,cor,means);
               finddeviations(nd^.sibR,total,cor,means);
            End
            Else Begin
               total[1]:=total[1]+sqr((nd^.dat[1]-means[1]));
               total[2]:=total[2]+sqr((nd^.dat[2]-means[2]));
               cor:=(means[1] - nd^.dat[1])*(means[2] - nd^.dat[2]) + cor;
            End; {else}
         End; {local procedure findDeviations}
      {********************************************************************}
      Begin
         total[1]:=0;
         total[2]:=0;
         cor:=0;
         findDeviations(parm.root,total,cor,parm.mean);
         parm.variance[1]:=total[1]/(parm.tipnumber-1);
         parm.variance[2]:=total[2]/(parm.tipnumber-1);
         if (parm.variance[1]<>0) And (parm.variance[2]<>0) Then
           parm.correlation:=cor/((parm.tipnumber -1)*sqrt(parm.variance[1]*
                            parm.variance[2]))
         Else
           ret('A variance is 0.  We assume correlation = 0.  Press any key to '+
                                'continue.');
      End; {procedure findVariance}

{**************************************************************************}
{This procedure uses the procedure contemp to report if the tips across the
tree are contemporaneous.  Note that if branch length transformations are
used in pdtree, small rounding errors may occur which will cause this
procedure to report that tips are not contemporaneous, even if they
actually are to a large number of decimal places.}
{**************************************************************************}
   Procedure iscontemp(Var p :parameter);
      Var
         dum :real;
{**************************************************************************}
{This procedure checks to see of the tips are contemporaneous.}
{**************************************************************************}
      Procedure contemp(rt,nd :nodeptr; height:real; current:real;
                        Var contemperaneous:boolean);
         Begin
            If nd<>rt Then
              current:=current+nd^.length;
            If nd^.sibL<>nil Then Begin
               contemp(rt,nd^.sibL,height,current,contemperaneous);
               contemp(rt,nd^.sibR,height,current,contemperaneous);
            End
            Else
              If ((current-height) > 1e-10) Then
                contemperaneous:=false;
         End; {local procedure iscontemp}
      {*********************************************************************}
      Begin
         p.contemp:=true;
         dum:=0;
         contemp(p.root,p.root,p.height,dum,p.contemp);
      End; {procedure contemp}

{**************************************************************************}
{This procedure is the one that is used to read in the tree.  It uses
InputTree from t_io1 to read in the tree.  It then uses the procedures above
to fill in some information about the tree into the parameter variable.
Note that you must be in graphics mode to call this procedure.}
{**************************************************************************}
   Procedure doRead(var parms:parameter);
      Var
         ch :char;
         succ :boolean;
         rt :nodeptr;
         flname:flstring;
         ext:typstring;
         dum:integer;
      Begin
         rt:=parms.root;     {always initialize parms.root to something}
         if (parms.root = nil) Then 
           succ:=False
         Else Begin
            succ:=True;
            flname:=parms.filename;
         End;
         closegraph;
         InputTree(rt,flname,ext,succ);
         If succ Then Begin
            parms.filename:=flname;
            parms.root:=rt;
            dum:=0;
            numtps(parms.root,dum);
            parms.tipnumber:=dum;
            parms.height:=0;
            calculate(parms);
            findmean(parms);
            findvariance(parms);
            iscontemp(parms);
            parms.max[1]:=-2.8E37;
            parms.min[1]:=2.8E37; 
            parms.max[2]:=-2.8E37;
            parms.min[2]:=2.8E37;
            findmaxmin(parms.root,parms.max,parms.min);
            clrscr;
            If ((ext = 'BRK') and
                (parms.mean[1] = 0) and
                (parms.mean[2] = 0) and
                (parms.variance[1] = 0) and
                (parms.variance[2] = 0)) Then Begin
               writeln('Warning -- A tip file was not loaded, you will need to manually set ');
               writeln('           simulation means and variances to get meaningful results.');
               writeln;
               writeln('           Or you can quit PDSIMUL and either create a .TIP file or');
               writeln('           or fill in tip values with PDTREE.');
               writeln;
               writeln('Press any key to continue.');
               ch := readkey;
               ch:='Y';      {and overwrite}
            End
            Else Begin
               writeln('Do you want the default starting values, variances, and');
               writeln('correlation entered for you now? (Y/N) ');
               ch:=upcase(readkey);
            End; {else}
            If ch= 'Y' Then Begin
               parms.starting[1]:=parms.mean[1];
               parms.finalmean[1]:=parms.mean[1];
               parms.starting[2]:=parms.mean[2];
               parms.finalmean[2]:=parms.mean[2];
               parms.selvariance[1]:=parms.variance[1];
               parms.selvariance[2]:=parms.variance[2];
               parms.selcorr:=parms.correlation;
               parms.peak:=parms.starting;
               parms.decay[1]:=1/parms.height;
               parms.decay[2]:=1/parms.height;
            End; {then}
         End; {then}
         graphicson(mx,my);
      End; {procedure doRead}

   End. {module s_io1}
