{***************************************************************************}
{This is unit s_menu1.  It was made for PDSimul.  It is in charge of
running the menu that allows the user to input the parameters that the
program uses later for the simulations.  It uses the graphics screen
heavily.

One thing merits explaining.  The way I use the graphics screen may not be
immediately obvious.  I use procedures in s_util1 to draw boxes on the screen.
The coordinates of the upper left hand corners of these boxes are kept in a
global data structure box of type boxes defined in s_util1.  I then draw
text in the boxes with a routine called "out".  This routine takes as a
parameter the box to be drawn in (an integer 1-10) and the offset within
that box.  Also, for convenience the variables mx and my are global.  They
are the maximum x value on the screen and the maximum y value on the screen.}
{***************************************************************************}
unit s_menu1;

interface
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

   uses s_util1;

   Procedure runmenu(parm :parameter);

implementation
  uses crt,
       graph,
       t_util1,
       s_io1,
       te_util1,
       s_sim1;

{***************************************************************************}
{This procedure clears the dialogue portion of the menu screen and the small
space under it.  It happens to be the top of the screen now.}
{***************************************************************************}
Procedure clearbtm;
begin
  rect(wborder,thigh,0,0,79,9,7);     {dialogue Box}
end;

{***************************************************************************}
{This procedure draws the background portion of the screen.  It clears the
graphics screen by drawing a big filled-in rectangle over it with rect2.  It
then creates all the boxes used.  Note: rect and rect2 are in s_util1.  They
share a global variable called box.  This simply holds the position of the
upper left hand corner of each box drawn to the screen.  The procedure OUT
then uses this to output text in the box.}
{***************************************************************************}
Procedure DrawBackground(parms :parameter);
   Var
      st :string;
   Procedure OutReal(boxnum,x,y : integer;val : real);
      Begin
         str(val:19:-11,st);
         delete(st,16,2);
         out(boxnum,x,y,st);
      End; {procedure realout}
   {***********************************************************************}
   Begin {drawbackground}
      rect2(background);                  {Entire Screen}
      clearbtm;
      rect(background,background,0,9,79,11,8); {Small space}
      rect(wborder,wback,0,10,79,17,1);       {Top Tree Info Box}
      rect(wborder,wback,0,28,26,43,5);     {Left bounds box}
      rect(wborder,wback,0,19,79,22,2);      {Top General parameter box}
      rect(wborder,wback,0,23,70,27,3);     {Middle variance box}
      rect(wborder,wback,27,36,79,46,6);    {Bottom OU box}
      rect(wborder,wback,27,28,79,35,4);    {Bottom brownian box}
      setcolor(wbackhi);
      out(10,12,18,'Tree Characteristics in Box Above, Variables Below');
      setcolor(tback);
      out(1,0,0,'Current Tree:');
      out(1,55,0,'Contemporaneous Tips?');
      out(1,2,1,'Number of Tips:   Ave. Height:                   Nordheim:');
      out(1,2,2,'Mean of     Trait1:                    Trait2:');
      out(1,2,3,'Variance of Trait1:                    Trait2:');
      out(1,2,4,'PPMCC:                   Min[1]:                   Max[1]:');
      out(1,2,5,'                         Min[2]:                   Max[2]:');
      setcolor(tnormal);
      out(1,15,0,parms.filename);
      If parms.contemp
      Then out(1,77,0,'Y')
      Else out(1,77,0,'N');

      str(parms.tipnumber,st);   {integer, don't use outReal}
      out(1,17,1,st);
      outReal(1,32,1,parms.height);
      outReal(1,60,1,parms.nordheim);
      outReal(1,21,2,parms.mean[1]);
      outReal(1,48,2,parms.mean[2]);
      outReal(1,21,3,parms.variance[1]);
      outReal(1,48,3,parms.variance[2]);
      outReal(1,8,4,parms.correlation);
      outReal(1,34,4,parms.min[1]);
      outReal(1,60,4,parms.max[1]);
      outReal(1,34,5,parms.min[2]);
      outReal(1,60,5,parms.max[2]);

      setcolor(tbhigh);
      out(2,9,0,'*(R)ead in a new tree.       *(J)ust do it.'+
                '           *(Q)uit.');
      setcolor(tback);
      out(2,0,1,'(S)eed for Random Numbers:           (N)umber of'+
                ' Simulations:');
      out(3,0,0,'(I)nitial Values    Trait1:                  Trait2:');
      out(4,0,2,'(X)- Final Means');
      out(4,0,3,' Trait1:                  Trait2:');
      out(3,0,1,'(V)ariances-Tip     Trait1:                  Trait2:');
      out(4,0,4,'(W)- Variances-Mean');
      out(4,0,5,' Trait1:                  Trait2:');
      out(3,0,2,'(C)orrelation of Input Distributions:');
      out(5,0,0,'(U)pper Bounds');
      out(5,0,1,'Trait1:');
      out(5,0,2,'Trait2:');
      out(5,0,3,'  (0,)) Unbounded');
      out(5,0,4,'  (1,!) "Throw Out"');
      out(5,0,5,'  (2,@) "Replace"');
      out(5,0,6,'  (3,#) "Flip"');
      out(5,0,7,'  (4,$) "Hard Bounce"');
      out(5,0,8,'  (5,%) "Soft Bounce"');
      out(5,0,9,'  (6,^) "Truncate Change"');
      out(5,0,10,'  (7,&) "Combination"');
      out(5,0,11,'(L)ower Bounds');
      out(5,0,12,'Trait1:');
      out(5,0,13,'Trait2:');
      out(4,2,0,'(G)radual Brownian');
      out(4,24,0,'(B)- Speciational Brownian');
      out(4,2,1,'(P)unctuated Equilibrium');
      out(6,0,1,'(D)ecay Constants');
      out(6,0,2,' Trait1:                  Trait2:');
      out(6,0,3,'(A)daptive peak');
      out(6,0,4,' Trait1:                  Trait2:');
      out(6,0,5,'(Y)- Velocity of the peak');
      out(6,0,6,' Trait1:                  Trait2:');
      out(6,0,7,'(T)- Variance of peak movement');
      out(6,0,8,' Trait1:                  Trait2:');
      out(6,2,0,'(O)- Gradual OU');
      out(6,24,0,'(Z)- Speciational OU');
      out(4,2,1,'(P)unctuated Equilibrium');
      setcolor(thigh);
   End; {procedure drawbackground}


{***************************************************************************}
{This procedure is called whenever help text is put in the dialogue box
to tell the user to press return when he or she is done reading.}
{***************************************************************************}
Procedure promptreturn(parm:parameter);
var
  ch:char;
begin       
  rect(background,background,0,9,79,11,8); {Small space}
  setcolor(wbackhi);
  out(8,20,0,'Press any key to Continue.');
  ch:=readkey;         
  clearbtm;
  drawbackground(parm);
end;

{***************************************************************************}
{This procedure fills in the '>' things on the menu screen.  It relies on
information in the big parameter data structure to do so.}
{***************************************************************************}
procedure drawtoggledat(parms:parameter);
  var
    i:integer;
  begin
   setcolor(wback);
   out(4,1,0,'>');
   out(4,1,1,'>');
   out(4,23,0,'>');
   for i:=3 to 10 do
     out(5,1,i,'>');
   out(6,1,0,'>');
   out(6,23,0,'>');
   setcolor(tchange);
   out(5,1,parms.bounded+3,'>');
   if(parms.brownian=0) then
     begin
       if (parms.gradual=0) then
         out(4,1,0,'>')
       else
         if  (parms.gradual=1) then
           out(4,23,0,'>')
         else
           out(4,1,1,'>');
     end
   else
     begin 
       if (parms.gradual=0) then
         out(6,1,0,'>')
       else
         if  (parms.gradual=1) then
           out(6,23,0,'>');
     end;
  end;

{***************************************************************************}
{This procedure fills in all the data onto the menu screen.  It uses the
procedure above for some of this work.  It uses the big parameter data
structure for the source of the parameters.}
{***************************************************************************}
Procedure drawData(parms:parameter);
   Var
      st:string;
   Procedure OutReal(boxnum,x,y : integer;val : real);
      Begin
         str(val:19:-11,st);
         delete(st,16,2);
         out(boxnum,x,y,st);
      End; {procedure outReal}
   {***********************************************************************}
   Begin
      setcolor(tchange);
      str(parms.seed,st);
      out(2,26,1,st);
      str(parms.simnum,st);
      out(2,61,1,st);
      outReal(3,27,0,parms.starting[1]);
      outReal(3,52,0,parms.starting[2]);
      outReal(3,27,1,parms.selvariance[1]);
      outReal(3,52,1,parms.selvariance[2]);
      outReal(3,38,2,parms.selcorr);
      outReal(5,7,1,parms.upperb[1]);
      outReal(5,7,2,parms.upperb[2]);
      outReal(5,7,12,parms.lowerb[1]);
      outReal(5,7,13,parms.lowerb[2]);
      If (parms.brownian=1) Then Begin
         outReal(6,8,2,parms.decay[1]);
         outReal(6,33,2,parms.decay[2]);
         outReal(6,8,4,parms.peak[1]);
         outReal(6,33,4,parms.peak[2]);
         outReal(6,8,6,parms.peakvelocity[1]);
         outReal(6,33,6,parms.peakvelocity[2]);
         outReal(6,8,8,parms.peakvar[1]);
         outReal(6,33,8,parms.peakvar[2]);
         setcolor(tback);
         st:=' xx';
         out(4,8,3,st);
         out(4,33,3,st);
         out(4,8,5,st);  
         out(4,33,5,st);
      End {then}
      Else Begin
         outReal(4,8,3,parms.finalmean[1]);
         outReal(4,33,3,parms.finalmean[2]);
         outReal(4,8,5,parms.meanvariance[1]);
         outReal(4,33,5,parms.meanvariance[2]);
         setcolor(tback);
         st:=' xx';
         out(6,8,2,st);
         out(6,33,2,st); 
         out(6,8,4,st);     
         out(6,33,4,st);
         out(6,8,6,st);              
         out(6,33,6,st);
         out(6,8,8,st);  
         out(6,33,8,st);
      End; {else}
      drawToggleDat(parms);
   End; {drawData}

{***************************************************************************}
{This procedure displays the proper help message on the screen.  The char
holds the users selection.  Note that the help information is held in a
global variable called helparray.  I do this only because helparray is so
large and Turbo Pascal only allows 64k maximum for the stack and if
helparray were local it would take up too much space on the stack.  All the
global data structures go in the "Data Segment" which is also 64k, but I
can afford to fill it up with helparray.}
{***************************************************************************}
procedure displayhelp(ch:char);
var
  helpnum, i :integer;
begin
  setcolor(tback);
    case ch of
          'r' :  helpnum:=1;
          'q' :  helpnum:=2;
          'c' :  helpnum:=3;
          'p' :  helpnum:=4;
          'g' :  helpnum:=5;
          'i' :  helpnum:=6;
          'a' :  helpnum:=7;
          'b' :  helpnum:=8;
          'd' :  helpnum:=9;
          'j' :  helpnum:=10;
          'l' :  helpnum:=11;
          'n' :  helpnum:=12;
          'o' :  helpnum:=13;
          's' :  helpnum:=14;
          'u' :  helpnum:=15;
          'v' :  helpnum:=16;
          '0' :  helpnum:=17;
          '1' :  helpnum:=18;
          '2' :  helpnum:=19;
          '3' :  helpnum:=20;
          '4' :  helpnum:=21;
          '5' :  helpnum:=22;
          '6' :  helpnum:=23;
          '7' :  helpnum:=24;
          'x' :  helpnum:=25;
          'w' :  helpnum:=26;
          'z' :  helpnum:=27;
          'y' :  helpnum:=28;
          't' :  helpnum:=29;
    end;
  for i:=1 to 8 do
    out(7,0,i-1,helparray[helpnum,i]);
end;

{***************************************************************************}
{This procedure reads in one real number.  It uses readreal which is in
the unit T_Util1.}
{***************************************************************************}
  Procedure readit(wind,x,y:integer; var num:real);
  var
    succ,l :boolean;
    dummy : extended;

  begin
    window(3,5,79,5);
    repeat
      rect(wborder,wback,box[wind,1]-1,box[wind,2]+y,79,
                                      box[wind,2]+y+4,9);
      setcolor(tback);
      out(9,0,0,'>');
      gotoxy(1,1);
      readreal(dummy,succ,l);
    until (succ and (not l));
    num := dummy;
    window(1,1,80,25);
  end;

{***************************************************************************}
{This procedure writes a message to the dialogue box then uses readit to
read in a real number.}
{***************************************************************************}
  Procedure readone(var rl:real);
  begin
    setcolor(tback);
    out (7,0,3,' Please enter the new value.');
    readit(7,0,4,rl);
  end;

{***************************************************************************}
{This procedure is like the one above, except it reads in two reals.}
{***************************************************************************}
  Procedure readtwo(var reals:traits);
  Var dummy : real;
  begin
    setcolor(tback);
    out (7,0,3,' What shall the new value for trait 1 be?');
    readit(7,0,4,dummy);
    reals[1] := dummy;
    rect(thigh,thigh,box[7,1],box[7,2]+2,78,box[7,2]+4,9);
    setcolor(tback);
    out(7,0,3,' And what shall the new value for trait 2 be?');
    readit(7,0,4,dummy);
    reals[2] := dummy;
  end;

{***************************************************************************}
{The following obscene number of procedures simply prompt the user to
change various parameters.  The one directly below changes the initial
value for each trait.}
{***************************************************************************}
  Procedure doinitial(var parameters:parameter);
  begin
    out(7,0,0,'Please input the new initial value for each trait or "s" '+
                               'for no change.');
    out(7,0,1,'Note: the OU peak values will be set to the initial values.');
    readtwo(parameters.starting);
    parameters.peak:=parameters.starting;
  end;

  Procedure dovariance(var parameters:parameter);
  begin
    out(7,0,0,'Please input the new tip variance for each trait or "s" '+
                          'for no change.');
    readtwo(parameters.selvariance);
    calculate(parameters);
  end;

  Procedure doupperb(var parameters:parameter);
  begin
    out(7,0,0,'Please input the new Upper Bound for each trait or "s" '+
                          'for no change.');
    out(7,0,1,'You should enter a number greater than the starting value.');
    readtwo(parameters.upperb);
  end;

  Procedure dolowerb(var parameters:parameter);
  begin
    out(7,0,0,'Please input the new Lower Bound for each trait or "s" '+
                          'for no change.');
    out(7,0,1,'You should enter a number less than the starting value.');
    readtwo(parameters.lowerb);
  end;

  Procedure dodecay(var parameters:parameter);
  begin
    if (parameters.brownian=1) then
      begin
        out(7,0,0,'Please input the new decay constant for each trait'+
             ' or "s" for no change.');
        readtwo(parameters.decay);
      end;
  end;

  Procedure dopeak(var parameters:parameter);
  begin
    if (parameters.brownian=1) then
      begin
        out(7,0,0,'Please input the new Adaptive Peak for each trait or "s" '+
                          'for no change.');
      readtwo(parameters.peak);
    end;
  end;

  Procedure dofinalm(var parameters:parameter);
  begin
    if (parameters.brownian=0) then
      begin
        out(7,0,0,'Please input the new Final Mean for each trait or "s" '+
                           'for no change.');
        readtwo(parameters.finalmean);
      end;
  end;

  Procedure dovelocityp(var parameters:parameter);
  begin
    if (parameters.brownian=1) then
      begin
        out(7,0,0,'Please input the new peak Velocities for each trait or "s" '+
                           'for no change.');
        readtwo(parameters.peakvelocity);
      end;
  end;

  Procedure dovariancep(var parameters:parameter);
  begin
    if (parameters.brownian=1) then
      begin
        out(7,0,0,'Please input the new peak variances for each trait or "s" '+
                           'for no change.');
        readtwo(parameters.peakvar);
      end;
  end;

  Procedure dovariancem(var parameters:parameter);
  begin
    if(parameters.brownian=0) then
      begin
        out(7,0,0,'Please input the new mean variances for each trait or "s" '+
                           'for no change.');
        readtwo(parameters.meanvariance);
      end;
  end;

  Procedure docorrelation(var parameters:parameter);
  var
    realnum :real;
  begin
    out(7,0,0,'Please input the new Correlation.');
    out(7,0,1,'Note the correltaion must be in the range -1,1.');
    repeat
      realnum:=10;
      readone(realnum);
    until (realnum >= -1.0) and (realnum <= 1.0);
    parameters.selcorr:=realnum;
  end;

{***************************************************************************}
{This one and the next must check that the number input is in the correct
range and is an integer.}
{***************************************************************************}
  Procedure donumber(var parameters:parameter);
  var
    realnum:real;
  begin
    out(7,0,0,'Please input the number of simulations desired.');
    out(7,0,1,'   The number entered will be rounded to the nearest'+
             ' integer.');
    out(7,0,2,'   The number must be in the interval -32768 to 32767.');
    repeat
      readone(realnum);
    until (-32768.0<=realnum) and (32767.0>=realnum);
    parameters.simnum:=round(realnum);
  end;

{***************************************************************************}
{This one must check that the number input is an integer in the correct
range.}
{***************************************************************************}
  Procedure doseed(var parameters:parameter);
  var
    realnum:real;
  begin
    out(7,0,0,'Please enter the seed number desired.  Enter Zero for the '+
               'Computer to choose.');
    out(7,0,1,'   The number entered will be rounded to the nearest'+
             ' integer.');
    out(7,0,2,'   The number must be in the range - 2147483648 to '+
                               '2147483647.');
    repeat
      readone(realnum);
    until (-2147483648.0<=realnum) and (2147483647.0>=realnum);
    parameters.seed:=round(realnum);
  end;

{***************************************************************************}
{Here is the body of the procedure.  It is simply a menu implemented on the
graphics screen.  It draws the background and the interior, then waits for
the user to select something.  If it is lower case, the user is given a help
message.  If it is upper case, the user is allowed to change a Parameter.
The screen is updated when necessary along the way.
Note: when the user selects 'R' the s_io1 unit is used to read in a new
tree.  When the user selects 'J' the s_sim1 and s_simul1 units are used to
actually do the simulations.  If the parameters selected are inconsistent,
then s_sim1 notices and returns to here.}
{***************************************************************************}
Procedure runmenu(parm :parameter);
   Const prefixCode = chr(0);
   Var
      ch:char;
   Begin
      loadgraphics;
      box[10,1]:=0;      {This sets the 10th box to be the whole screen}
      box[10,2]:=0;
      graphicson(mx,my);
      doRead(parm);            {read in a tree}
      drawbackground(parm);
      Repeat
         drawdata(parm);
         setcolor(tback);
         out(7,0,0,'To obtain information on a parameter, enter the LOWER CASE '+
                        'version');
         out(7,0,1,'    of the letter or number representing that parameter.');
         out(7,0,3,'To change a parameter, enter the UPPER CASE version of the');
         out(7,0,4,'    letter or number representing that parameter.');
         ch:=readkey;
         clearbtm;
         If (ch In ['a','b','c','d','g','i','j','l','n','o','p','q','r','s',
                't','u','v','w','x','y','z','0','1','2','3','4',
                        '5','6','7'])
         Then Begin            {if selection valid lower case display help}
            displayhelp(ch);
            promptreturn(parm);
         End
         Else Begin
            If (ch In ['A','C','D','I','J','L','N','R','S','U','V','X','Y',
                       'T','W'])
            Then Begin {else if uppercase selection associated with a procedure}
               setcolor(tback);
               Case ch of
                  'J' :doit(parm);
                  'R' :doRead(parm);
                  'I' :doInitial(parm);
                  'V' :doVariance(parm);
                  'U' :doUpperB(parm);
                  'L' :doLowerB(parm);
                  'A' :doPeak(parm);
                  'D' :doDecay(parm);
                  'C' :doCorrelation(parm);
                  'S' :doseed(parm);
                  'N' :donumber(parm);
                  'X' :dofinalm(parm);
                  'Y' :dovelocityp(parm);
                  'T' :dovariancep(parm);
                  'W' :dovariancem(parm);
               End;
               drawbackground(parm);
               drawdata(parm);
            End
            Else Begin  {else if valid uppercase of a toggle parameter}
               If (ch in [')','!','@','#','$','%','^','&','O','P','G','B','Z'])
               Then Begin
                  case ch of
                     ')' :parm.bounded:=0;
                     '!' :parm.bounded:=1;
                     '@' :parm.bounded:=2;
                     '#' :parm.bounded:=3;
                     '$' :parm.bounded:=4;
                     '%' :parm.bounded:=5;
                     '^' :parm.bounded:=6;
                     '&' :parm.bounded:=7;
                     'G' :Begin
                             parm.gradual:=0;
                             parm.brownian:=0;
                          End; {case G}
                     'Z' :Begin
                             parm.gradual:=1;
                             parm.brownian:=1;
                          End; {case Z}
                     'P' :Begin
                             parm.gradual:=2;
                             parm.brownian:=0;
                          End; {case P}
                     'B' :Begin
                             parm.brownian:=0;
                             parm.gradual:=1;
                          End; {case B}
                     'O' :Begin
                             parm.brownian:=1;
                             parm.gradual:=0;
                          End; {case O}
                  End; {case}
                  If ch in ['G','Z','P','O','B']
                  Then Begin
                      calculate(parm);
                      drawbackground(parm);
                      drawdata(parm);
                  End
                  Else drawtoggledat(parm);
               End {Then}
               Else If (ch = prefixCode)
                    Then Begin
                       ch := readkey;  {throw away second character}
                       ch := ' ';      {overwrite second char (so ch <> 'Q')}
                    End; {then}
            End; {else}
         End;
      Until ch='Q';  {keep going until the user selects 'Q'}
  closegraph;
  clrscr;
End; {runmenu}

End. {module s_menu1}
