{**************************************************************************}
{This is unit s_sim1.  It is used by s_menu1 to check the simulation
parameters for consistency, open and initialize the output file, and call
s_simul1 to do the simulations.}
{**************************************************************************}
unit s_sim1;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

interface

uses s_util1;

   procedure doit(p:parameter);

implementation

uses   crt,
       dos,
       graph,
       t_util1,
       te_util1,
       s_simul1,
       s_stat2;

{**************************************************************************}
{This procedure just outputs text to the screen to let the user know what
is happening before the user is confronted by the simulation summary screen.}
{**************************************************************************}
procedure displaynotes(p:parameter);
begin
  clrscr;
  writeln('                 A NOTE ON THE PARAMETERS');
  writeln;
  writeln('With any given set of parameters chosen, some will',
            ' exclude others.');
  writeln('The next screen will check to see if the set you have chosen',
           ' is adequate');
  writeln('to perform a set of simulations.  If they are not, you will get',
          ' a "FATAL ERROR"');
  writeln('message.  If they are adequate you will see all of the relevant',
                   ' parameters for');
  writeln('your simulations.  Those parameters not shown will have',
                 ' no effect.');
  writeln;
  writeln('For example, if "(0) Unbounded" is chosen the upper and lower',
            ' bounds will');
  writeln('be ignored.  For a complete description of which parameters are',
          ' compatible,');
  writeln('see the documentation.');
  writeln;
  if p.brownian=0 then
    begin
      if p.gradual=0 then
        writeln('You have selected Gradual Brownian simulations.')
      else
        if p.gradual=1 then
          writeln('You have selected Speciational Brownian simulations.')
        else
            writeln('You have selected Punctuated Equilibrium simulations.');
      writeln('Therefore, the "(D)ecay Constants", "(A)daptive peak", ');
      writeln('"(Y)- Velocity of the peak", and "(T)- Variance of the ',
                               'peak movement"');
      writeln('parameters will be ignored.');
      If p.gradual=2 then
        begin
          writeln;
            writeln('Be aware that Polytomies are disallowed in Punctuated ',
             'Equilibrium.');
        end;
    end
  else
    begin                              
      if p.gradual=0 then
        writeln('You have selected Gradual OU simulations.')
      else
        writeln('You have selected Speciational OU simulations.');
      writeln('Therefore, the "(X)- Final Means" and "(W)- Variances-Mean"',
       ' parameters');
      writeln('will be ignored.');
    end;
  ret('Please press any key for the simulation summary screen.');
end;

{**************************************************************************}
{This procedure checks the parameters for consistency and also displays
them to the screen for the user to check one last time.  The user is then
asked if the simulations should start.}
{**************************************************************************}
procedure displaycheck(p:parameter; var procede:boolean);
var
  ch :char;
  bad :boolean;

{**************************************************************************}
{This procedure is used to check for a branch length=0 anywhere in the tree.
If one or more are found then bad=true.}
{**************************************************************************}
  Procedure checkzerolength(nd,rt :nodeptr; var bad:boolean);
    begin
      if (nd<>rt) and (nd^.length=0) then
        bad:=true;
      if nd^.sibL<>NIL then
        begin
          checkzerolength(nd^.sibL,rt,bad);
          checkzerolength(nd^.sibR,rt,bad);
        end;
    end;

{**************************************************************************}
{This procedure is used to report fatal errors.}
{**************************************************************************}
  Procedure fatalerr(var p:boolean; st:string);
    begin
      p:=false;
      writeln('         *****   FATAL ERROR   *****');
      writeln(st);
    end;

{*************************************************************************}
begin
  procede:=true;
  clrscr;

  writeln(p.simnum,' simulations to be performed on: ',p.filename,'.');
  if (p.simnum<1) then
    fatalerr(procede,'The number of simulations to be performed must be'+
       ' greater than zero.');
  if (p.root = nil) then
      fatalerr(procede,'No tree in memory!!!!');

  writeln;
  write('TREE CHARACTERISTICS: ');
  write('Number of Tips: ',p.tipnumber,' Average Height: ');
  extendWrite(output,p.height);
  writeln;
  write('Tip mean      Trait1: ');
  extendWrite(output,p.mean[1]);
  write(' Trait2: ');
  extendWrite(output,p.mean[2]);
  writeln;
  write('Tip variance  Trait1: ');
  extendWrite(output,p.variance[1]);
  write(' Trait2: ');
  extendWrite(output,p.variance[2]);
  writeln;
  write('Tip Pearson product-moment correlation: ');
  extendWrite(output,p.correlation);
  writeln;

  writeln;
  write('WHAT YOU CHOSE: ');
  If (p.seed=0) then
    begin
      randomize;
      writeln('Random seed chosen by computer to be: ',RandSeed);
    end
  else
    begin
       RandSeed:=p.seed;
       writeln('Random seed chosen by user to be: ',RandSeed);
    end;

    write('Starting values are      Trait1: ');
  extendWrite(output,p.starting[1]);
  write(' Trait2: ');
  extendWrite(output,p.starting[2]);
  writeln;
  If p.brownian=0 Then Begin
      write('Expected mean at tips    Trait1: ');
      extendWrite(output,p.finalmean[1]);
      write(' Trait2: ');
      extendWrite(output,p.finalmean[2]);
      writeln;
      write('Variance of mean change  Trait1: ');
      extendWrite(output,p.meanvariance[1]);
      write(' Trait2: ');
      extendWrite(output,p.meanvariance[2]);
      writeln;
  End; {then}
  write('Expected variance at tip Trait1: ');
  extendWrite(output,p.selvariance[1]);
  write(' Trait2: ');
  extendWrite(output,p.selvariance[2]);
  writeln;
  write('Correlation of evolution: ');
  extendWrite(output,p.selcorr);
  writeln;


  writeln;
  if (p.gradual=0) and (p.brownian=0) then
    writeln('The Gradual Brownian model will be used.'); 
  if (p.gradual=0) and (p.brownian=1) then
    writeln('The Gradual OU model will be used.');
  if (p.gradual=1) and (p.brownian=0) then
    writeln('The Speciational Brownian model will be used.');
  if (p.gradual=1) and (p.brownian=1) then
    writeln('The Speciational OU model will be used.');
  if p.gradual=2 then
    begin
      writeln('The Punctuated Equilibrium model will be used.');
      bad:=false;
      checkzerolength(p.root,p.root,bad);
      if bad then
        fatalerr(procede,'There can be no zero length branch lengths with'+
         ' Punctuated Equilibrium.');
    end;
  if p.brownian=1 then
    begin
      write('Decay constants         Trait1: ');
      extendWrite(output,p.decay[1]);
      write(' Trait2: ');
      extendWrite(output,p.decay[2]);
      writeln;
      If (p.decay[1]<=0) or (p.decay[2]<=0) then
        fatalerr(procede,'Decay constants must be greater than 0.');
      write('Selective peaks         Trait1: ');
      extendWrite(output,p.peak[1]);
      write(' Trait2: ');
      extendWrite(output,p.peak[2]);
      writeln;
      write('Peak velocities         Trait1: ');
      extendWrite(output,p.peakvelocity[1]);
      write(' Trait2: ');
      extendWrite(output,p.peakvelocity[2]);
      writeln;
      write('Peak movement variances Trait1: ');
      extendWrite(output,p.peakvar[1]);
      write(' Trait2: ');
      extendWrite(output,p.peakvar[2]);
      writeln;
      If p.gradual=2 then
        fatalerr(procede,'Punctualtional Equilibrium is incompatible with '+
                            'the OU process.');
    end;
    if (((p.finalmean[1]<>p.starting[1]) or (p.finalmean[2]<>p.starting[2]))
           and (p.gradual=3)) then
      fatalerr(procede,'With Punctuated Equilibrium the starting values '+
               'must equal the final means.');

  writeln;
  if p.bounded=0 then
    writeln('Unbounded simulations will be done.')
  else
    begin
      write('Bounded simulations will be done with the ');
      case p.bounded of
         1 : write('"Throw Out"');
         2 : write('"Replace"');
         3 : write('"Flip"');
         4 : write('"Hard Bounce"');
         5 : write('"Soft Bounce"');
         6 : write('"Truncate Change"');
         7 : write('"Combination"');
      end;
      writeln(' algorithm.');
      write('Upper Bounds, Trait1: ');
      extendWrite(output,p.upperb[1]);
      write(' Trait2: ');
      extendWrite(output,p.upperb[2]);
      writeln;
      write('Lower Bounds, Trait1: ');
      extendWrite(output,p.lowerb[1]);
      write(' Trait2: ');
      extendWrite(output,p.lowerb[2]);
      writeln;
      if (((p.upperb[1]<=p.starting[1]) or (p.upperb[2]<=p.starting[2])) or
        ((p.lowerb[1]>=p.starting[1]) or (p.lowerb[2]>=p.starting[2]))) then
        fatalerr(procede,'The starting values must be in bounds.');
  end;
    if procede then
      begin
        writeln;
        writeln;
        writeln('Do you wish to proceed with the simulations? (Y/N)');
        ch:=Upcase(readkey);
        if ch<>'Y' then
           procede:=false;
      end
    else
      ret('Simulations aborted, press any key to return to the '+
                                              'parameter screen.');
end;

{************************************************************************}
{This procedure writes generic information to the output file, including
the Date and Time which it gets from DOS.}
{************************************************************************}
   Procedure DateTime(var outfile: text; filename:flstring; parm:parameter);
   Var
      Year, Month, Day, DayofWeek, Hour, Minute, Second, Sec100 : word;
   Begin
      Writeln(Outfile, 'Output from PDSIMUL: continuous characters ',
               'simulated on a phylogenetic tree.');
      Write(Outfile, 'Data from ', Filename);
      GetDate(Year, Month, Day, DayofWeek);
      Write(Outfile,' ',Day,' ');
      Case Month of
         1 : Write(Outfile, 'Jan ');
         2 : Write(Outfile, 'Feb ');
         3 : Write(Outfile, 'Mar ');
         4 : Write(Outfile, 'Apr ');
         5 : Write(Outfile, 'May ');
         6 : Write(Outfile, 'Jun ');
         7 : Write(Outfile, 'Jul ');
         8 : Write(Outfile, 'Aug ');
         9 : Write(Outfile, 'Sep ');
        10 : Write(Outfile, 'Oct ');
        11 : Write(Outfile, 'Nov ');
        12 : Write(Outfile, 'Dec ');
      End; {case month of}
      Write(Outfile, Year,' at ');
      GetTime(Hour, Minute, Second, Sec100);
      Writeln(Outfile, Hour, ':', Minute,':',Second,'.');
      writeln(outfile);
      writeln(outfile,'    Number of Tips  =  ',parm.tipnumber);
      writeln(outfile,'Phylogenetic tree from input file: ',parm.filename);
      writeln(outfile,'(Node name- Left child, Right child, Ancestor: ',
                   'name, branch-length, number descendants.)');
   End; {proc datetime}

{**************************************************************************}
{This procedure puts the tree structure into the output file.}
{**************************************************************************}
procedure dumptree(var outfile:text; nd,rt:nodeptr);
var
   number : integer;
begin
  if nd^.sibL<>NIL then
    begin
      write(outfile,'  ');
      write(outfile,nd^.name,' ');
      write(outfile,nd^.sibL^.name,' ');
      extendWrite(outfile,nd^.sibL^.length);
      number:=0;
      numtps(nd^.sibL,number);
      write(outfile,' ',number:2,' ');
      write(outfile,nd^.sibR^.name,' ');
      extendWrite(outfile,nd^.sibR^.length);
      number:=0;
      numtps(nd^.sibR,number);
      write(outfile,' ',number:2,' ');
      if nd<>rt then begin
        write(outfile,nd^.ancestor^.name,' ');
        extendWrite(outfile,nd^.length);
        writeln(outfile);
      end
      else
        writeln(outfile,'**  0.0000000000E+00');

      dumptree(outfile,nd^.sibL,rt);
      dumptree(outfile,nd^.sibR,rt);
    end;
end;

{**************************************************************************}
{This procedure puts information about the simulations into the output
file.}
{**************************************************************************}
procedure dumpdescription(var outfile:text; p:parameter);
begin
  write(outfile,'SIMULATION PARAMETERS: ');


  if (p.gradual=0) and (p.brownian=0) then
    writeln(outfile,' The Gradual Brownian model was used.');
  if (p.gradual=0) and (p.brownian=1) then
    writeln(outfile,' The Gradual OU model was used.');
  if (p.gradual=1) and (p.brownian=0) then
    writeln(outfile,' The Speciational Brownian model was used.');
  if (p.gradual=1) and (p.brownian=1) then
    writeln(outfile,' The Speciational OU model was used.');
  if p.gradual=2 then
    writeln(outfile,'The Punctuated Equilibrium model was used.');

  writeln(outfile,'                                 Trait 1           ',
                                 'Trait 2          Correlation');
  write(outfile,'Initial Values:           ');
  extendWrite(outfile,p.starting[1]);
  write(outfile,' ');
  extendWrite(outfile,p.starting[2]);
  write(outfile,' ');
  extendWrite(outfile,p.selcorr);
  writeln(outfile);
  write(outfile,'Selected Tip Variance:    ');
  extendWrite(outfile,p.selvariance[1]);
  write(outfile,' ');
  extendWrite(outfile,p.selvariance[2]);
  writeln(outfile);
  if p.brownian=0 then
    begin
      write(outfile,'Selected Final Means:     ');
      extendWrite(outFile,p.finalmean[1]);
      write(outFile,' ');
      extendWrite(outFile,p.finalmean[2]);
      writeln(outfile);
      write(outfile,'Variance of Mean Movement:');
      extendWrite(outFile,p.meanvariance[1]);
      write(outFile,' ');
      extendWrite(outFile,p.meanvariance[2]);
      writeln(outFile);
      writeln(outfile);
      writeln(outfile);
    end
  else
    begin
      write(outfile,'OU Decay Constants:        ');
      extendWrite(outFile,p.decay[1]);
      write(outFile,' ');
      extendWrite(outFile,p.decay[2]);
      writeln(outFile);
      write(outfile,'OU starting peak value:    ');
      extendWrite(outFile,p.peak[1]);
      write(outFile,' ');
      extendWrite(outFile,p.peak[2]);
      writeln(outFile);
      write(outfile,'Speed of change of peak:   ');
      extendWrite(outFile,p.peakvelocity[1]);
      write(outFile,' ');
      extendWrite(outFile,p.peakvelocity[2]);
      writeln(outFile);
      write(outfile,'Variance of peak change:   ');
      extendWrite(outFile,p.peakvar[1]);
      write(outFile,' ');
      extendWrite(outFile,p.peakvar[2]);
      writeln(outFile);
    end;
  If p.bounded=0 then
    begin
      writeln(outfile,'No bounds are used.');
      writeln(outfile);
      writeln(outfile);
    end
  else
    begin
      write(outfile,'The simulations were bounded by the ');
      case p.bounded of
        1 : writeln(outfile,'"Throw Out" algorithm.');
        2 : writeln(outfile,'"Replace " algorithm.');
        3 : writeln(outfile,'"Flip" algorithm.');
        4 : writeln(outfile,'"Hard Bounce" algorithm.');
        5 : writeln(outfile,'"Soft Bounce" algorithm.');
        6 : writeln(outfile,'"Truncate Change" algorithm.');
        7 : writeln(outfile,'"Combination" algorithm.');
      end;
      write(outfile,'Upper Bounds:              ');
      extendWrite(outFile,p.UpperB[1]);
      write(outFile,' ');
      extendWrite(outFile,p.UpperB[2]);
      writeln(outFile);
      write(outfile,'Lower Bounds:              ');
      extendWrite(outFile,p.LowerB[1]);
      write(outFile,' ');
      extendWrite(outFile,p.LowerB[2]);
      writeln(outFile);
    end;
  writeln(outfile,'The Pseudo-random Number Generator Seed = ',
                         RandSeed,'.');
  writeln(outfile,'Trait values from ',p.simnum,' simulations follow.');
end;

{**************************************************************************}
{This procedure links up the nodes in the tree so s_simul1 has an easier
time traversing the tree to write out the results.  The format is a left
first traversal for both the tips and the nodes, but they must alternate.
So I used the tipL and tipR fields in the tree that are not used by this
program for any other purpose to do it.  rt^.tipR points to the first tip.
then the tipL of that tip points to the first node whose tipL points to the
next tip and so on until nil is reached.  Again, this is just done so that
when the values of the nodes and tips must be written to the file it is a
quick traversal rather than a longer more complex one.}
{**************************************************************************}
procedure linktree(rt : nodeptr);
var
  tip, node, oldnode :nodeptr;
  ch:char;

{**************************************************************************}
{This procedure clears the tempx and tempy marks from the tree so the next
two procedures can keep track of which nodes they have visited with them.}
{**************************************************************************}
procedure clearmarks(nd: nodeptr);
begin
  nd^.tempx:=0;
  nd^.tempy:=0;
  if nd^.sibL<>NIL then
    begin
      clearmarks(nd^.sibL);
      clearmarks(nd^.sibR);
    end;
end;

{**************************************************************************}
{This procedure returns a pointer to the next tip in a left first traversal
of the tree.  It requires tempy to be cleared for it before it is called the
first time.  The first time it is called, it needs to be passed rt.  The
second time it is called, it needs to be passed the results it passed back
last (the last tip returned).  It then can find the next by looking at the
markings it left in tempy.}
{**************************************************************************}
  procedure gettip(rt: nodeptr; var tip:nodeptr);
  begin
    if tip^.tempy=0 then
      begin
        tip^.tempy:=1;
        if tip^.sibL<>nil then
          begin
            tip:=tip^.sibL;
            gettip(rt,tip);
          end;
      end
    else
      begin
        if tip^.sibL=NIL then
          begin
            tip:=tip^.ancestor;
            gettip(rt,tip);
          end
        else
          if tip^.sibR^.tempy=0 then
            begin
              tip:=tip^.sibR;
              gettip(rt,tip);
            end
          else
            if rt<>tip then
              begin
                tip:=tip^.ancestor;
                gettip(rt,tip);
              end
            else
              tip:=nil;
      end;
  end;

{**************************************************************************}
{This procedure has the same function and behavior as the last one
except that it works with nodes not tips.}
{**************************************************************************}
  procedure getnode(var node:nodeptr);
  begin
    if (node^.tempx=0) then
      node^.tempx:=1
    else
      begin
        if (node^.sibL^.sibL<>nil) and (node^.sibL^.tempx=0) then
          begin
            node:=node^.sibL;
            getnode(node);
          end
        else
          if (node^.sibR^.sibL<>nil) and (node^.sibR^.tempx=0) then
            begin
              node:=node^.sibR;
              getnode(node);
            end
          else
            if node^.ancestor<>nil then
              begin
                node:=node^.ancestor;
                getnode(node);
              end;
      end;
  end;

{**************************************************************************}
begin
  tip:=rt;
  node:=rt;
  clearmarks(rt);
  gettip(rt,tip);
  rt^.tipR:=tip;
  getnode(node);
  rt^.tipR^.tipL:=node;
  oldnode:=node;
  repeat
    gettip(rt,tip);
    getnode(node);
    oldnode^.tipL:=tip;
    if node<>rt then
      tip^.tipL:=node
    else
      tip^.tipL:=nil;
    oldnode:=node;
  until tip^.tipL=nil;
end;

{**************************************************************************}
{This is a debugging procedure I used to check that the last procedure linked
up the tree correctly.}
{**************************************************************************}
procedure showlinks(nd:nodeptr);
var
  ch:char;
begin
  writeln(nd^.name,' ');
  ch:=readkey;
  if nd^.tipL<>NIL then
    showlinks(nd^.tipL);
end;
 
{**************************************************************************}
{This procedure allows the user to change the name of the output file.}
{**************************************************************************}
procedure getfilename(var name2:namestr);
var
  good, again:boolean;
  ch:char;
  dir:dirstr;
  nm:namestr;
  ext:extstr;
  fl:file;
  name:flstring;
begin
  clrscr;
  write('The current base for the simulation results filename is: ');
  writeln('"',name2,'".');
  writeln('To this base is added the extension ".sim".');
  writeln;
  writeln('Would you like to change the base for the simulation ',
                                         'results (Y/N)?');
  writeln;
  writeln('NOTE:  the C.M.A.P. programs of Martins and Garland (1991) only');
  writeln('       allow 5 characters in the root name.');
  ch:=Upcase(readkey);
  good := true;
  if ch='Y' then
    begin
      repeat
        writeln;
        if (good = false)
        then writeln('The file is write-locked or invalid.');
        writeln('Please enter the 8 or fewer digit file name.');
        writeln;
        readln(name2);
        name:=name2+'.sim';
        good:=true;
        again:=false;
        {$I-}
        Assign(fl,name);
        Reset(fl);
        close(fl);
        {$I+}
        If (IOresult=0) and (name<>'') then
          begin
            writeln;
            writeln(name,' already exits.');
            writeln('Do you want to delete it and ',
            'save a new simulation in its ',
            'place? (Y/N)');
            Repeat
              ch:=UpCase(readkey);
            Until ch in['Y','N'];
            If ch='N' then again:=true
            else begin
               {$I-}
               Assign(fl,name);
               Rewrite(fl);
               close(fl);
               {$I+}
               If (IOresult<>0) or (name='') then
                 good:=false;
            end;
          end
          else begin
              {$I-}
              Assign(fl,name);
              Rewrite(fl);
              close(fl);
              {$I+}
              If (IOresult<>0) or (name='') then
                good:=false;
          End;
      Until good and not again
    end;
  clrscr;
end;
{**************************************************************************}
{Here is the main procedure of the unit.  It shows the information screen
to the user, displays the parameters selected, and aborts if the user
has changed his or her mind or if the parameters are bad.  Otherwise, it
prepares the output file and some last minute parameters and links up the
tree.  In short it prepares everything for the simulations to actually
take place.  Note that it changes some things in the parameter variable
that the rest of the program would not like, however the parameter variable
is local so the changes don't get passed back up.}
{**************************************************************************}
procedure doit(p:parameter);
var
  continue:boolean;
  ch :char;
  dum, height :real;
  nord :traits;
  mx,my:integer;
  outfile:text;
  dir :dirstr;
  name :namestr;
  ext : extstr;
begin
  closegraph;
  displaynotes(p);
  displaycheck(p,continue); 
  if continue then
    begin        
    fsplit(p.filename,dir,name,ext);
    getfilename(name);    
      assign(outfile,dir+name+'.sim');
      rewrite(outfile);
      datetime(outfile,p.filename,p);
      dumptree(outfile,p.root,p.root);
      dumpdescription(outfile,p);
      writeln(outfile,'                  Tips',
                      '                               Nodes',
                      '                   Simulation');
      writeln(outfile,'       Trait 1           Trait 2',
                     '           Trait 1           Trait 2',
                     '           Count');
      writeln(outfile,'_________________________________________',
                      '___________________________________________');
      if p.brownian=0 then
        begin
          dum:=0;
          height:=p.height;
          p.meanvelocity[1]:=(p.finalmean[1]-p.starting[1])/height;
          p.meanvelocity[2]:=(p.finalmean[2]-p.starting[2])/height;
          p.meanvariance[1]:=sqrt(p.meanvariance[1])/p.nordheim;
          p.meanvariance[2]:=sqrt(p.meanvariance[2])/p.nordheim;
        end
      else
        begin
          p.peakvar[1]:=p.peakvar[1]/p.nordheim;
          p.peakvar[2]:=p.peakvar[2]/p.nordheim;
        end;
      p.nordvar[1]:=sqrt(p.selvariance[1])/p.nordheim;
      p.nordvar[2]:=sqrt(p.selvariance[2])/p.nordheim;
      if p.bounded = 5 then
        begin
          p.region[1]:=(p.upperb[1]-p.lowerb[1])/5.0;
          p.region[2]:=(p.upperb[2]-p.lowerb[2])/5.0;
        end;
      linktree(p.root);
 {     showlinks(p.root^.tipR);        }
      simulate(outfile,p);
      close(outfile);
      clrscr;
      gotoxy(1,10);
      write('Do you wish for me to calculate some statistics on',
             ' the results (Y/N)');
      ch:=upcase(readkey);
      if ch='Y' then
        statme(outfile,p);
    end;
  graphicson(mx,my);
end;


end.
