{*************************************************************************}
{This is unit s_stat1.  This unit is used by PDSIMUL.  It does a statistical
analysis of the results of the simulations.  It reports the mean mean, mean
variance, min min, and max max.  It also displays a histogram for each
trait.  It works as follows.  It first reads in the tip data from the
file that was just written by s_simul1.  It stops reading if reading the
tip data from one more simulation would give it more than 5000 tips in
memory.  It then calculates some statistics from these data and dumps them
to the screen.}
{*************************************************************************}
unit s_stat1;

interface

uses s_util1;

procedure statme(var infile:text; p :parameter);

implementation

uses   crt,
       t_util1;
{**************************************************************************}
{This procedure, whose body is at the end of this file, uses the following
procedures to do what is described at the beginning of the file.
{**************************************************************************}
procedure statme(var infile:text; p :parameter);
var
  n,m :longint;
  dat :data;
  mean,min,max,vari :traits;
  cor : real;
{**************************************************************************}
{This procedure makes one pass through the data and catches the max max and
the min min as well as calculating the mean mean.
{**************************************************************************}
procedure meanminmax(m, tipnumber: integer; var mean,min,max: traits);
var
  i,j,k :integer;
  temp :traits;
  num :real;
begin
  for j:=1 to 2 do
    begin
      min[j]:=dat[j,j];
      max[j]:=dat[j,j];
      temp[j]:=0;
      for i:=0 to (m-1) do
        begin
          mean[j]:=0;
          for k:=1 to tipnumber do
            begin
              num:= dat[i*tipnumber+k,j];
              mean[j]:=mean[j]+num;
              if num>max[j] then
                max[j]:=num;
              if num<min[j] then
                min[j]:=num;
            end;
          temp[j]:=temp[j]+mean[j]/tipnumber;
        end;
    end;
  mean[1]:=temp[1]/m;
  mean[2]:=temp[2]/m;
end;
{**************************************************************************}
{This procedure reads in the tip data from the file.  The text variable
infile must be associated with the correct file name and passed to this
procedure.
{**************************************************************************}
procedure readdat(var infile:text; var n:longint );
var
  st :string;
  ch :char;
  i : integer;
  t1,t2 :longint;
begin
  reset(infile);
  n:=p.tipnumber+21;
  for i:=1 to n do
    begin
      readln(infile,st);
      writeln(st);
    end;
  writeln('Reading in the rest of the data.  It may take a few seconds.');
  t1:=p.tipnumber;
  t2:=p.simnum;
  n:=t1*t2;
  if n > 5000 then
    n:=5000;
  for i:=1 to n do
    begin
      read(infile,ch);
      read(infile,ch);
      read(infile,ch);
      read(infile,dat[i,1]);
      readln(infile,dat[i,2]);
    {  writeln(i,'   ',dat[i,1],'   ',dat[i,2]);        }
    end;
  writeln;
  ret('Please press Return to continue');
  close(infile);
end;

{**************************************************************************}
{This procedure makes a second pass through the data and figures out the
mean variance and the mean correlation.
{**************************************************************************}
procedure varcor(m, tipnumber:integer; mean: traits;
                             var vari:traits; var cor:real);
var
  i,j,k:integer;
  temp1 :traits;
  temp2,num1,num2 :real;
  ch :char;
begin
  temp1[1]:=0;
  temp1[2]:=0;
  temp2:=0;
  for i:=0 to (m-1) do
    begin
      vari[1]:=0.0;
      vari[2]:=0.0;
      cor:=0;
      for k:=1 to tipnumber do
        begin
            num1:= dat[i*tipnumber+k,1];
            num2:= dat[i*tipnumber+k,2];
            vari[1]:=vari[1] + sqr((num1 - mean[1]));
            vari[2]:=vari[2] + sqr((num2 - mean[2]));
            cor:=cor + (mean[1] - num1)*(mean[2] - num2);
          end;
      temp1[1]:=temp1[1]+vari[1]/tipnumber;
      temp1[2]:=temp1[2]+vari[2]/tipnumber;
      if (vari[1]>0) and (vari[2]>0) then   {avoid division by 0}
          temp2:=temp2 + cor/((tipnumber-1)*sqrt((vari[1]*vari[2])/
                          sqr(tipnumber)));
    end;
  vari[1]:=temp1[1]/m;
  vari[2]:=temp1[2]/m;
  cor:=temp2/m;
end;

{**************************************************************************}
{This procedure figures out the population of each slot in the histogram.
It then calls printhist to print the histogram given the slots.}
{**************************************************************************}
procedure hist(n:integer; min,max:traits);
var
  number1, number2: category;
  distance :traits;
  slot :integer;
  i :integer;

{**************************************************************************}
{This prints the histogram to the screen, given the slots in dist.
Note that the range is from the lower limit to the higher limit if limits
were used in the simulations; otherwise, the range is from the min min
to the max max of the simulated data.}
{**************************************************************************}
  procedure printhist(dist : category);
  var
    i,j:integer;
    maxnum: real;
    size :real;
    ch :char;
  begin
    maxnum:=dist[1];
    for i:=2 to 78 do
      if maxnum<dist[i] then
        maxnum:=dist[i];
    size:=maxnum/6.0;
        for j:=5 downto 0 do
          begin
            write('#');
            for i:=1 to 78 do
              if ((dist[i]/size) >= j) and (dist[i]<>0) then
                write('|')
              else
                write(' ');
            write('#');
          end;
  end;

{**************************************************************************}
begin
  distance[1]:=(max[1]-min[1])/78;
  distance[2]:=(max[2]-min[2])/78;
  for i:=1 to 80 do
    begin
      number1[i]:=0;
      number2[i]:=0;
    end;
  for i:=1 to n do
    begin
      if distance[1]<>0 then
        slot:= trunc((dat[i,1]-min[1])/distance[1])+1
      else
        slot:=1;
      number1[slot]:=number1[slot]+1;
      if distance[2]<>0 then
        slot:= trunc((dat[i,2]-min[2])/distance[2])+1
      else
        slot:=1;
      number2[slot]:=number2[slot]+1;
    end;
  number1[78]:=number1[78]+number1[79];
  number2[78]:=number2[78]+number2[79];
  writeln('Histogram of Trait 1 below. Scale from ',min[1],' to ',max[1],'.');
  if distance[1]<>0 then
    printhist(number1)
  else
    writeln('Variance too small for meaningfull histogram.');
  writeln('Histogram of Trait 2 below. Scale from ',min[2],' to ',max[2],'.');
  if distance[2]<>0 then
    printhist(number2)
  else
    writeln('Variance too small for meaningfull histogram.');
end;

{***********************************************************************}
begin
  readdat(infile,n);
  writeln;
  writeln('I''m Thinking..............................');
  m:=trunc(n/p.tipnumber);
  meanminmax(m,p.tipnumber,mean,min,max);
  clrscr;
  writeln('Here is some information on all the tip values of the simulated',
                   ' results.');
  writeln('Note: if there are more than 5,000 tips, this represents at most',
                                ' the first 5,000');
  writeln;
  writeln('Mean Mean     Trait1: ',mean[1],' Trait2: ',mean[2]);
  writeln('Max Max       Trait1: ',max[1],' Trait2: ',max[2]);
  writeln('Min Min       Trait1: ',min[1],' Trait2: ',min[2]);

  varcor(m, p.tipnumber, mean,vari,cor);
  writeln('Mean Variance Trait1: ',vari[1],' Trait2: ',vari[2]);
  writeln('Mean Correlation: ',cor);
  writeln;
  if p.bounded=0 then
    hist(n,min,max)
  else
    hist(n,p.lowerb,p.upperb);
  writeln;
  ret('Please Press Return to Continue.');
end;

end.
