{*************************************************************************}
{This is unit s_stat1.  This unit is used by PDSIMUL.  It does a statistical
analysis of the results of the simulations.  It reports the mean mean, mean
variance, min min, and max max.  It also displays a histogram for each
trait.  It works as follows.  It first reads in the tip data from the
file that was just written by s_simul1. It then calculates some statistics
from these data and dumps them to the screen.}
{*************************************************************************}
unit s_stat2;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

interface

uses s_util1;

procedure statme(var infile:text; p:parameter);

implementation

uses  crt,
      t_util1;
              
{**************************************************************************}
{This procedure, whose body is at the end of this file, uses the following
procedures to do what is described at the beginning of the file.}
{**************************************************************************}
procedure statme(var infile:text; p :parameter);
var
  mean,min,max,vari :traits;
  cor : real;                                                             
{************************************************************************}
{This procedure resets the file to the beginning, then reads in the text
at the beginning which we are not interested in here.}
{************************************************************************}
Procedure setupstream(var infile:text; tipnum:integer);
var
  st:string;
  i:integer;
begin
  reset(infile);
  for i:=1 to (tipnum+21) do
    readln(infile,st);
end;

{************************************************************************}
{This function reads in the tips from the next tree and returns them.  If
we are at the end of file, then it returns 0, else it returns 1 for success.
Oh, by the way, if there are more than 1000 tips this will not work.}
{************************************************************************}
Function readtips(var infile:text; tipnum:integer; var results:data): integer;
var
  i:integer;
  ch:char;
  dum:integer;
  dum2:real;
  dum3 :integer;
begin
  dum:=1;
  for i:=1 to tipnum do
    begin
      read(infile,ch);        
      if ch=Chr(26) then
        dum:=0;
      read(infile,ch);
      read(infile,ch);
      read(infile,ch);
      read(infile,results[i,1]);
      read(infile,results[i,2]);  
      read(infile,ch);
      read(infile,ch);
      read(infile,ch);
      read(infile,dum2);
      read(infile,dum2);
      read(infile,ch);   {so this won't choke on the sim #}
      read(infile,dum3); {read and toss the sim#; shouldn't this be readln?}
    end;
  readtips:=dum;
end;
  
{************************************************************************}
{This calculates the statistics for the data.  It reads in one trees worth
of tips at a time, then calculates the min, max, and mean.  It then
calculates the variance and correleation for that tree.  Finally, it
adds the means, variances, and correlation onto the counting variables
so it can find their mean when it is through.}
{************************************************************************}
Procedure dostats (var fl:text; tipnum:integer;
               var mean,min,max,vari:traits; var cor:real);
var
  runningmean,runningvar:etraits;
  temprunmean,temprunvar:etraits;
  runningcor,tempruncor:real;
  rawdata:data;
  i,j,k:integer;
  ch:char;
begin
  i:=0;
  min[1]:=1E38;
  max[1]:=-1E38;  
  min[2]:=1E38;
  max[2]:=-1E38;
  runningmean[1]:=0.0;
  runningmean[2]:=0; 
  runningvar[1]:=0;
  runningvar[2]:=0;
  runningcor:=0;
  setupstream(fl,tipnum);
  while(readtips(fl,tipnum,rawdata)=1) do
    begin
      for k:=1 to 2 do
        begin
          temprunmean[k]:=0;
          for j:=1 to tipnum do
            begin
              if (rawdata[j,k]>max[k]) then
                max[k]:=rawdata[j,k];
              if (rawdata[j,k]<min[k]) then
                min[k]:=rawdata[j,k];
              temprunmean[k]:=temprunmean[k]+rawdata[j,k];
            end;
          temprunmean[k]:=temprunmean[k]/tipnum;
          runningmean[k]:=runningmean[k]+temprunmean[k];
        end;
      temprunvar[1]:=0;
      temprunvar[2]:=0;
      tempruncor:=0;
        for j:=1 to tipnum do
          begin
            temprunvar[1]:=temprunvar[1]+sqr((rawdata[j,1]-temprunmean[1]));
            temprunvar[2]:=temprunvar[2]+sqr((rawdata[j,2]-temprunmean[2]));  
            tempruncor:=tempruncor+(temprunmean[1]-rawdata[j,1])*
                                   (temprunmean[2]-rawdata[j,2]);
          end;
     temprunvar[1]:=temprunvar[1]/(tipnum-1);
     temprunvar[2]:=temprunvar[2]/(tipnum-1);
     runningvar[1]:=runningvar[1]+temprunvar[1];
     runningvar[2]:=runningvar[2]+temprunvar[2];
     tempruncor:=tempruncor/(tipnum-1);
     if((temprunvar[1]>0) and (temprunvar[2]>0)) then
       tempruncor:=tempruncor/sqrt(temprunvar[1]*temprunvar[2])
     else
       tempruncor:=0;   {This stuff is to avoid division by zero.}
     runningcor:=runningcor+tempruncor;
     i:=i+1;
    end;
  for j:=1 to 2 do
    begin
      mean[j]:=runningmean[j]/i;
      vari[j]:=runningvar[j]/i;
    end;
  cor:=runningcor/i;
end;

{**************************************************************************}
{This procedure figures out the population of each slot in the histogram.
It then calls printhist to print the histogram given the slots.}
{**************************************************************************}
procedure hist(var fl:text; tipnum:integer; min,max:traits);
var
  number1, number2: category;
  distance :traits;
  slot :integer;
  i :integer;
  rawdata :data;

{**************************************************************************}
{This prints the histogram to the screen, given the slots in dist.
Note that the range is from the lower limit to the higher limit if limits
were used in the simulations; otherwise, the range is from the min min
to the max max of the simulated data.}
{**************************************************************************}
  procedure printhist(dist : category);
  var
    i,j:integer;
    maxnum: real;
    size :real;
    ch :char;
  begin
    maxnum:=dist[1];
    for i:=2 to 78 do
      if maxnum<dist[i] then
        maxnum:=dist[i];
    size:=maxnum/6.0;
        for j:=5 downto 0 do
          begin
            write('#');
            for i:=1 to 78 do
              if ((dist[i]/size) >= j) and (dist[i]<>0) then
                write('|')
              else
                write(' ');
            write('#');
          end;
  end;

{**************************************************************************}
begin
  distance[1]:=(max[1]-min[1])/78;
  distance[2]:=(max[2]-min[2])/78;
  for i:=1 to 80 do
    begin
      number1[i]:=0;
      number2[i]:=0;
    end; 
  setupstream(fl,tipnum);
  while(readtips(fl,tipnum,rawdata)=1) do
    begin
      for i:=1 to tipnum do
        begin
          if distance[1]<>0 then
            slot:= trunc((rawdata[i,1]-min[1])/distance[1])+1
          else
            slot:=1;
          number1[slot]:=number1[slot]+1;
          if distance[2]<>0 then
            slot:= trunc((rawdata[i,2]-min[2])/distance[2])+1
          else
            slot:=1;
          number2[slot]:=number2[slot]+1;
        end;
    end;
  number1[78]:=number1[78]+number1[79];
  number2[78]:=number2[78]+number2[79];
  write('Histogram of Trait 1 below. Scale from ');
  extendWrite(output,min[1]);
  write(' to ');
  extendWrite(output,max[1]);
  writeln('.');
  if distance[1]<>0 then
    printhist(number1)
  else
    writeln('Variance too small for meaningful histogram.');
  write('Histogram of Trait 2 below. Scale from ');
  extendWrite(output,min[2]);
  write(' to ');
  extendWrite(output,max[2]);
  writeln('.');
  if distance[2]<>0 then
    printhist(number2)
  else
    writeln('Variance too small for meaningful histogram.');
end;

{***********************************************************************}
begin
  writeln;
  writeln('I''m Thinking..............................');
   dostats(infile,p.tipnumber,mean,min,max,vari,cor);
  clrscr;
  writeln('Here is some information on all the tip values of the simulated',
                   ' results.');
  writeln;
  write('Mean Mean     Trait1: ');
  extendWrite(output,mean[1]);
  write(' Trait2: ');
  extendWrite(output,mean[2]);
  writeln;
  write('Max Max       Trait1: ');
  extendWrite(output,max[1]);
  write(' Trait2: ');
  extendWrite(output,max[2]);
  writeln;
  write('Min Min       Trait1: ');
  extendWrite(output,min[1]);
  write(' Trait2: ');
  extendWrite(output,min[2]);
  writeln;
  write('Mean Variance Trait1: ');
  extendWrite(output,vari[1]);
  write(' Trait2: ');
  extendWrite(output,vari[2]);
  writeln;
  write('Mean Correlation: ');
  extendWrite(output,cor);
  writeln;
  writeln;
  if p.bounded=0 then
    hist(infile,p.tipnumber,min,max)
  else
    hist(infile,p.tipnumber,p.lowerb,p.upperb);
  writeln;
  ret('Please Press Return to Continue.');
end;

end.


