{*************************************************************************}
{This unit, s_util1, has many of the type definitions for PDSIMUL as well
as all the global variables and constants and some procedures used widely
throughout the program.}
{*************************************************************************}
unit s_util1;

interface
{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 65520,0,655360} {maximum stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

uses t_util1;

   const
      background = 4;        {These are all colors}
      wback = 7;
      wborder = 0;
      tback = 0;
      tnormal = 0;
      tchange = 6;
      thigh = 12;
      tbhigh = 4;
      wbackhi = 15;
  type
      etraits = array[1..2] of real;
      category = array[1..80] of real;   {used for histogram in s_stat2}
      corner = array[1..2] of integer;
      boxes = array[1..10] of corner;    {used globally to keep track of
                                                     windows}
      data = array[1..1000] of traits;   {used to hold the tip info in
                                                    s_stat2}
      help = array[1..30,1..8] of string[80];  {used globally to hold help
                                                  information in s_menu1}
      datptr = ^bigdata;     {used by s_simul1 to keep simulation results
                              on the heap so it can save many simulations
                              at once.}
      bigdata = record
         name1    :namestring;
         dat1     :traits;
         name2    :namestring;
         dat2     :traits;
         simCount :integer;
         next     :datptr;
      end;
      parameter = record    {all the parameters of the simulation}
         filename    : flstring;      {Current Tree Name }
         contemp     : boolean;       {Contemporaneous Tips? }
         root        : nodeptr;       {Pointer to root of tree in memory}
         tipnumber   : integer;       {Total number of tips}
         height      : real;          {Height of highest tip}
         nordheim    : real;          {Nordheim number}
         nordvar     : traits;
         mean        : traits;        {Mean of actual tips in file}
         variance    : traits;        {Variance of actual tips in file}
         correlation : real;          {Correlation of actual tree}

         min         : traits;          {Minimum tip value}
         max         : traits;          {Maximum tip value}

         Seed        : longint;       {Random seed.  0=computer chooses}
         simnum      : integer;       {Number of simulations}

         starting    : traits;        {(I)nitial values of each trait}
         finalmean   : traits;        {Final Mean of tips for simulation}
         meanvariance: traits;        {Variance of 'mean movement'}
         selvariance : traits;        {Variance for simulated tips}
         selcorr     : real;          {Correlation for simulations}

         brownian    : integer;       {0=Brownian, 1=OU}
         peak        : traits;        {OU adaptive peaks initial value}
         decay       : traits;        {OU decay constant.}
         peakvelocity: traits;        {Velocity with which the peak moves}
         peakvar     : traits;        {Variance for peak movement}

         bounded     : integer;       {How bounded. See program menu.}
         UpperB      : traits;        {Upper bounds}
         LowerB      : traits;        {Lower bounds}

         gradual     : integer;       {0=Gradual, 1=Speciational, 2=Pun.}

         meanvelocity : traits;       {program calculated this in s_sim1}
         region       : traits;       {program calculated this in s_sim1}
      end;
   var           {Global Variables}
      mx,my:integer;    {Max coordinates of the graphics screen}
      box:boxes;        {Upper left hand coordinates of the windows}
      helparray :help;  {All the help information}

   Procedure rect(colorborder,colorfill,x1,y1,x2,y2,number:integer);
   Procedure rect2(color:integer);
   Procedure out(number,x,y:integer; st:string);
   procedure calculate(var p:parameter);

implementation

  uses graph;
{**************************************************************************}
{This procedure calculates the Nordheim number.  Each time a new tree is
entered and each time the "Gradual" parameter is changed this must be
called as the Nordheim number must be recalculated.
The Nordheim number is an estimation of the factor by which the variance we
want at the tips must be divided to get the correct variance of our input
distribution for each change up the tree.}
{**************************************************************************}
procedure calculate(var p:parameter);
  var
    kterm, span, ht, number :real;
    ch :char;
{**************************************************************************}
{This procedure finds the k term which is the summation of the following
for each branch (k=number of tips descending from the branch):
             k*(k-1)*branchlength.
Note that the branchlength is assumed to be 1 for speciational and
for punctuated equilibrium models.
{**************************************************************************}
  procedure findkterm(gradual: integer; rt,nd:nodeptr;var kterm:real);
    var
      number:integer;
    begin
      number:=0;
      numtps(nd,number);
      if (nd<>rt) then
        if (gradual=0) then
          kterm:=kterm+number*(number-1)*nd^.length
        else
          kterm:=kterm+number*(number-1);
      if (nd^.sibl<>nil) then
        begin
          findkterm(gradual, rt, nd^.sibL, kterm);
          findkterm(gradual, rt, nd^.sibR, kterm);
        end;
    end;

{**************************************************************************}
{This procedure finds the sum of all the branchlengths in the tree.  Again
note that the branch lengths are assumed to be 1 for speciational and for
punctuated equilibrium.}
{**************************************************************************}
  procedure findspan(gradual: integer; rt,nd:nodeptr;
                           ht:real; var span:real);
    begin
      if(nd<>rt) then
        if (gradual=0) then
          ht:=ht+nd^.length
        else
          ht:=ht+1;
      if(nd^.sibl<>nil) then
        begin
          findspan(gradual, rt, nd^.sibL, ht, span);
          findspan(gradual, rt, nd^.sibR, ht, span);
        end
      else
        span:=span+ht;
    end;

begin
  if p.root <> nil then
    begin
      span:=0;
      kterm:=0;
      ht:=0;
      findkterm(p.gradual, p.root, p.root, kterm);
      findspan(p.gradual, p.root, p.root, ht, span);
      number:=p.tipnumber;
      p.height:=span/number;  {Average height of tree.}
      p.nordheim:=sqrt((span/number - kterm/(number*(number - 1.0))));
          {This is what the Nordheim number equals}
    end
  else
    p.nordheim:=0.0;
end;

{**************************************************************************}
{This procedure creates a rectangle-window on the graphics screen.  It not
only draws it, but also registers the rectangle in the box global variable.
Note that the coordinates are given in terms of the space taken up by
characters on the screen.  The screen is 47 lines down and 80
characters across.}
{**************************************************************************}
Procedure rect(colorborder,colorfill,x1,y1,x2,y2,number:integer);
var
  top:boolean;
begin
  box[number,1]:=x1+1;
  box[number,2]:=y1+1;
  x1:=round((x1/80.0 + 1/160)*mx);
  x2:=round((x2/80.0 + 1/160)*mx);
  y1:=round((y1/47.0 + 1/90)*my);
  y2:=round((y2/47.0 + 1/90)*my);
  top:=true;
  setcolor(colorborder);
  setfillstyle(solidfill,colorfill);
  bar3d(x1,y1,x2,y2,2,top);
end;

{**************************************************************************}
{This is used to clear the screen by drawing an opaque rectangle over all
of it.}
{**************************************************************************}
Procedure rect2(color:integer);
var
  top:boolean;
begin
  top:=false;
  setcolor(color);
  setfillstyle(solidfill,color);
  bar3d(0,0,mx,my,0,top);
end;

{**************************************************************************}
{This is used to draw text to a rectangle.  Note that there is nothing
to prevent one from drawing outside the rectangle.  All this guarantees is
that the coordinates passed will be relative to the rectangle.}
{**************************************************************************}
Procedure out(number,x,y:integer; st:string);
begin
  x:=round((x+box[number,1])/80.0*mx);
  y:=round((y+box[number,2])/47.0*my);
  outtextxy(x,y,st);
end;

end.
