{This unit is in charge of all the branch manipulation features of the
program.  Externally it is one procedure that does it all.}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}

 
Unit te_brch1;

Interface
Uses t_util1;      {Type definitions, constants, and utility procedures.}
 
Procedure branchEdit(r,rt:nodeptr);
 
Implementation
Uses crt,             {Borland lo-res text screens}
     graph,           {Borland graphics libraries}
     te_util1;        {Utility procedures local to the editing module.}
 
{*************************************************************************}
{This procedure runs the branch manipulation editing features.}
{*************************************************************************}
Procedure branchEdit(r,rt:nodePtr);
   Var
      ch :char;
      mx,my :integer;
      top :boolean;
      height:extended;
      part :boolean;
   {*********************************************************************}
   {This procedure draws the top part of the screen for the branch editing
   module.}
   {*********************************************************************}
   Procedure drawtop(part: boolean;
                     rt:nodePtr;
                     top:boolean;
                     mx,my:integer;
                     Var height:extended);
      Var
         htScale :extended;
         dum :extended;
         t1,t2,ht :string;
         message :string[80];
         num, starNum, ratio:extended;
      Begin
         reDrawTree(rt, htscale, 0.5,mx,my);
         setColor(cnorm);
         setFillStyle(solidFill,0);
         bar3D(0,round(0.52*my),mx,round(0.62*my),1,top);
         setColor(chi);
         height:=0;
         dum:=0;
         getInfo(rt,num,starNum,ratio);
         extendStr(t1,num);
         extendStr(t2,starNum);
         extendStr(ht,ratio);
         message:='num='+t1+'  starnum='+t2+'  ratio='+ht;
         outTextXY(round(0.02*mx),round(0.53*my),message);
         If (part) Then Begin
            message:='-*** WARNING, you are only working with part of '+
               'the tree! ***-';
            outtextxy(round(0.02*mx),round(0.565*my),message);
         End; {then}
         findHeight(rt,rt,dum,height);
         extendStr(ht,height);
         message:='Total Height = ' + ht;
         OutTextXY(round(0.3*mx),round(0.6*my),message);
      End; {procedure drawTop}
   {***********************************************************************}
   {This procedure draws the bottom part of the graphics screen for the
   branch length editing module.}
   {***********************************************************************}
   Procedure drawBottom(top: boolean; mx,my :integer);
      Var
         message :string[80];
         xcoor :integer;
         mbase :integer;
         lineHeight : integer;
      {*******************************************************************}
      {This procedure simplifies maintenance of the menu that appears at
       the bottom of the branch length display, and may even save a little
       code space.                                                        }
      {*******************************************************************}
      Procedure menuItem(lineNumber : integer; text : string);
         Begin
            outTextXY(xcoor,mbase+lineNumber*lineHeight,text);
         End; {procedure menuItem}
      {*******************************************************************}
      Begin {procedure drawBottom}
         mbase := round(0.70*my);
         lineHeight := round(0.025*my);
         setcolor(chi);
         outTextXY(0,mbase,'From here you can:');
         message:='(S)cale:     rescale total height of tree.';
         xcoor:= round(0.1*mx);
         menuItem(1,message);
         message:='(E)xponent:  take all the branch lengths to a power.';
         menuItem(2,message);
         message:='(L)og:       take the log base 10 of all the'+
                                ' branch lengths.';
         menuItem(3,message);
         message:='(T)ransform: transform branch lengths by Grafen''s '+
                               '(1989) rho.';
         menuItem(4,message);
         message:='(C)onstant:  make all non-zero branch lengths equal'+
                                            ' to one.';
         menuItem(5,message);
         message:='(P)agel:     change branch lengths to Pagel''s arbitrary '+
                    'method.';
         menuItem(6,message);
         message:='(G)rafen:    change branch lengths to Grafen''s '+
                    'arbitrary method.';
         menuItem(7,message);
         message:='(N)ee        change branch lengths to Nee''s arbitrary '+
                   'method (log10).';
         menuItem(8,message);
         message:='(R)eturn:    return to the editing menu.';
         menuItem(9,message);
         message:='Please choose (S,E,L,T,C,P,G,N,R)';
         outTextXY(0,mbase+10*lineHeight,message);
         setColor(cnorm);
         setFillStyle(solidFill,0);
       {  bar3d(0,round(0.625*my),mx,round(0.74*my),1,top);} {bottom box here!}
      End; {procedure drawBottom}
      {*********************************************************************}
      {This recursive procedure does the work.  It is given the scaling
      factor, and then traverses the tree, multiplying all the branch
      lengths by the correct amount.}
      {*********************************************************************}
      Procedure scl(rt,nd:nodePtr; mult:extended);
         Begin
            If nd<>rt
            Then nd^.length:=nd^.length*mult;
            If nd^.sibL<>nil Then Begin
               scl(rt,nd^.sibL,mult);
               scl(rt,nd^.sibR,mult);
            End; {then}
         End; {scl}
   {***********************************************************************}
   {This procedure will scale the whole tree so that it is the desired
   height.  Note:  all the branch lengths stay proportional to one another.}
   {***********************************************************************}
   Procedure scale(part:boolean; r:nodePtr; top:boolean; mx,my:integer);
      Var
         desire, multiplier :extended;
         success: boolean;
         dumb:boolean;
         message: string[100];
         xcoor:integer;
         height:extended;
      Begin
         drawTop(part,r,top,mx,my,height);
         setColor(chi);
         message:='(S)cale:     rescale total height of tree.';
         xcoor:= round(0.1*mx);
         outTextXY(xcoor,round(0.775*my),message);
         Repeat
            setColor(cnorm);
            setFillStyle(solidfill,0);
            bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
            setColor(chi);
            xcoor:=round(0.02*mx);
            message:='To what value would you like the total height changed?'+
                       '  All branch lengths ';
            outTextXY(xcoor,round(0.65*my),message);
            message:='will be rescaled appropriately.  A 0.00000001 '+
                        '% rounding error may occur.';
            outTextXY(xcoor,round(0.675*my),message);
            message:= '     Please input a real number, or S for the same:';
            outTextXY(xcoor,round(0.7*my),message);
            desire:=height;
            gotoXY(54,22);
            readreal(desire,success,dumb);
         Until success;
         multiplier:=desire/height;
         scl(r,r,multiplier);
      End;  {procedure scale}
      {*********************************************************************}
      {This recursive procedure traverses the tree and raises all the branch
      lengths to a power.}
      {*********************************************************************}
      Procedure expMe(rt,nd:nodePtr; power:extended);
         Begin
            If ((nd<>rt) And (nd^.length <> 0))
            Then
               nd^.length:=exp(power*ln(nd^.length)) {There is no direct way}
            Else                                        {to do exponents.}
               if (nd<>rt)
               Then nd^.length:=0; {handle polytomies safely}
            If nd^.sibL<>nil Then Begin
                  expme(rt,nd^.sibL,power);
                  expme(rt,nd^.sibR,power);
            End; {then}
         End; {procedure expme}
   {*********************************************************************}
   {This procedure is like the one before, except it does not multiply,
    it raises the branch lengths to an exponent.  Note: the branch lengths
    do NOT stay proportional to one another.}
   {*********************************************************************}
   Procedure transform(part:boolean; rt:nodePtr; top:boolean; mx,my:integer);
      Var
         ch :char;
         power :extended;
         success: boolean;
         dum :boolean;
         message :string[80];
         xcoor : integer;
         height :extended;
      Begin
         drawTop(part,rt,top,mx,my,height);
         setColor(chi);
         xcoor:= round(0.1*mx);
         message:='(E)xponent:  take all the branch lengths to a power.';
         outTextXY(xcoor,round(0.8*my),message);
         xcoor:=round(0.02*mx);
         power:=1;
         Repeat
            setColor(cnorm);
            setFillStyle(solidFill,0);
            bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
            setColor(chi);
            message:='Please input the exponent to raise all the branch '+
                      'lengths to.';
            outTextXY(xcoor,round(0.665*my),message);
            message:='   Input a real number or S for 1.';
            outTextXY(xcoor,round(0.7*my),message);
            gotoXY(40,22);
            ReadReal(power,success,dum);
         Until success;
         expMe(rt,rt,power);
      End; {procedure transform}
   {***********************************************************************}
   {This procedure is again like the one before, except it does not
   multiply or raise to a power, but instead takes the log base 10
   of the branch lengths.}
   {***********************************************************************}
   Procedure logBranch(part: boolean; r:nodePtr; top:boolean; mx,my:integer);
      Var
         ch:char;
         success :boolean;
         message :string[80];
         xcoor:integer;
         height:extended;
      {*********************************************************************}
      {This procedure does the work here.  It goes through and logs all the
      branches.  Added feature to ignore polytomies PEM 14 Feb 1997.}
      {*********************************************************************}
      Procedure logMe(rt,nd:nodeptr);
         Begin
            If nd<>rt
            Then If (nd^.length <> 0)
                 Then nd^.length:=ln(nd^.length)/ln(10);
            If nd^.sibL<>nil Then Begin
               logMe(rt,nd^.sibL);
               logMe(rt,nd^.sibR);
            End; {then}
         End; {procedure logMe}
      {********************************************************************}
      {This recursive procedure checks to make sure that none of the
      branch lengths are less than 1.  If one of them is less than 1 then
      the log of it will be negative and we don't like negative lengths.
      This code now ignores polytomies (BL = 0); PEM 14 Feb 1997.}
      {********************************************************************}
      Procedure checkMe(rt,nd:nodePtr; Var succ: boolean);
         Begin
            If ((nd<>r) and (nd^.length<1) and (nd^.length <> 0))
            Then succ:=false
            Else
               If nd^.sibL<>nil Then Begin
                  checkme(rt,nd^.sibL,succ);
                  checkme(rt,nd^.sibR,succ);
               End {then}
            End; {procedure checkMe}
      {********************************************************************}
      Begin {logBranch}
         drawTop(part,rt,top,mx,my,height);
         xcoor:= round(0.1*mx);
         message:='(L)og:       take the log base 10 of all the'+
                                ' branch lengths.';
         outTextXY(xcoor,round(0.825*my),message);
         setColor(cnorm);
         setFillStyle(solidFill,0);
         bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
         setColor(chi);
         xcoor:= round(0.02*mx);
         success:=true;
         checkMe(r,r,success);
         If success Then Begin
            message:='Are you sure that you want to log the tree? (Y/N)';
            outTextXY(xcoor,round(0.68*my),message);
            Repeat
               ch:=upcase(readKey);
            Until (ch in ['Y','N']);
            If (ch='Y')
            Then logMe(r,r);
         End {then}
         Else Begin
            message:= 'Unable to perform transform.' +
                        '  Some or all of the branch lengths are less';
            outTextXY(xcoor,round(0.635*my),message);
            message:= 'than 1, which would yield negative branch lengths.'+
                        '  You must scale the tree';
            outTextXY(xcoor,round(0.66*my),message);
            message:= 'to make it taller, or lengthen the offending'+
                       ' branches individually.';
            outTextXY(xcoor,round(0.685*my),message);
            message:='Press any key to return to the branch editing menu.';
            outTextXY(xcoor,round(0.71*my),message);
            ch:=readKey;
         End; {else}
      End; {procedure logBranch}
      {********************************************************************}
      {This procedure sets all the non zero branch lengths equal to one.
       Note that it does not use its own transformation procedure like the
       previous procedures, but raises each length to the zero power with
       expMe.}
      {********************************************************************}
      Procedure constantOne(part:boolean;
                            rt: nodePtr;
                            top : boolean;
                            mx,my :integer);
         Var
            message : string[80];
            xcoor : integer;
            height : extended;
         Begin
            drawTop(part,rt,top,mx,my,height);
            setColor(chi);
            xcoor:= round(0.1*mx);
            message:='(C)onstant:  make all non-zero branch lengths equal to one.';
            outTextXY(xcoor,round(0.925*my),message);
            setColor(cnorm);
            setFillStyle(solidfill,0);
            bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
            setColor(chi);
            xcoor:= round(0.02*mx);
            message:='Are you sure that you want equal branch lengths? (Y/N)';
            outTextXY(xcoor,round(0.68*my),message);
            Repeat
               ch:=upCase(readKey);
            Until (ch in ['Y','N']);
            If (ch='Y') Then expMe(rt,rt,0);
         End; {procedure constantOne}
      {********************************************************************}
      {This procedure sets all the branch lengths to those dictated by
      Grafen's (1989, Phil. Trans. Royal Soc. London Ser. B 326:119-157)
      arbitrary method.  The distance from the tips to the current node is
      equal to one less than the number of tips descending from that node.}
      {********************************************************************}
      Procedure grafen(part:boolean;
                       rt: nodePtr;
                       top : boolean;
                       mx,my :integer);
         Var
            message : string[80];
            xcoor : integer;
            height : extended;
            dummy : integer;
         {******************************************************************}
         {This procedure computes the depths recursively.  The original
          Grafenme procedure was split into 2 pieces to support polytomy.
          PEM 9-May-95.                                                     }
         {******************************************************************}
         Procedure grafenMe1(rt,nd: nodePtr;
                             Var totalTips :integer);
            Var
               Rtips, Ltips :integer;
            Begin
               If (nd^.sibL=NIL) Then Begin
                  totaltips:=1; {this is a tip, so there is 1 in the subtree}
                  nd^.temp:=0 {set depth to 0}
               End {then}
               Else Begin
                  grafenMe1(rt,nd^.sibL, Ltips);
                  grafenMe1(rt,nd^.sibR, Rtips);
                  totalTips:=Ltips+Rtips;
                  nd^.temp := totalTips-1 {save number of tips-1 as depth}
               End {else}
            End; {Grafenme1}
         {******************************************************************}
         {This procedure finishes the transformation (taking polytomies into
          account.  It is also recursive. Added PEM 9-May-95.                }
         {******************************************************************}
         Procedure grafenMe2(rt,nd: nodePtr);
            Begin
               If (nd^.sibL <> NIL) Then Begin
                  If (nd^.sibL^.length <> 0) {test for polytomy}
                  Then nd^.sibL^.length := nd^.temp-nd^.sibL^.temp
                  Else nd^.sibL^.temp := nd^.temp; {"inherit" depth to child}
                  grafenMe2(rt,nd^.sibL);
                  If (nd^.sibR^.length <> 0) {test for polytomy}
                  Then nd^.sibR^.length := nd^.temp-nd^.sibR^.temp
                  Else nd^.sibR^.temp := nd^.temp; {"inherit" depth to child}
                  grafenme2(rt,nd^.sibR)
               End; {then}
            End; {procedure grafenMe2}
         {******************************************************************}
         Begin {grafen}
            drawTop(part,rt,top,mx,my,height);
            setColor(chi);
            xcoor:= round(0.1*mx);
            message:='(G)rafen:    change branch lengths to Grafen''s method.';
            outTextXY(xcoor,round(0.875*my),message);
            setColor(cnorm);
            setFillStyle(solidfill,0);
            bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
            setColor(chi);
            xcoor:= round(0.02*mx);
            message:='Are you sure that you want to Grafen transform the '+
                       'branch lengths? (Y/N)';
            outTextXY(xcoor,round(0.68*my),message);
            Repeat
               ch:=upCase(readKey);
            Until (ch in ['Y','N']);
            If (ch='Y') Then Begin
               grafenMe1(rt,rt,dummy);
               grafenMe2(rt,rt)
            End; {then}
         End; {procedure grafen}

      {********************************************************************}
      {This procedure sets all the branch lengths to those dictated by
      Nee's (pers. comm.) arbitrary method.  The distance from the tips
      to the current node is equal to the log10 of the number of tips
      descending from that node.  It is mostly copied from the code for
      Grafen's arbitrary transform.}
      {********************************************************************}
      Procedure Nee(part:boolean;
                       rt: nodePtr;
                       top : boolean;
                       mx,my :integer);
         Var
            message : string[80];
            xcoor : integer;
            height : extended;
            dummy : integer;
         {******************************************************************}
         {This procedure computes the depths recursively.  The original
          Grafenme procedure was split into 2 pieces to support polytomy.
          PEM 9-May-95.                                                     }
         {******************************************************************}
         Procedure NeeMe1(rt,nd: nodePtr;
                             Var totalTips :integer);
            Var
               Rtips, Ltips :integer;
            Begin
               If (nd^.sibL=NIL) Then Begin
                  totaltips:=1; {this is a tip, so there is 1 in the subtree}
                  nd^.temp:=0 {set depth to 0}
               End {then}
               Else Begin
                  NeeMe1(rt,nd^.sibL, Ltips);
                  NeeMe1(rt,nd^.sibR, Rtips);
                  totalTips:=Ltips+Rtips;
                  nd^.temp:=ln(totalTips)/ln(10); {save log10(tips) as depth}
               End {else}
            End; {Grafenme1}
         {******************************************************************}
         {This procedure finishes the transformation (taking polytomies into
          account.  It is also recursive. Added PEM 9-May-95.                }
         {******************************************************************}
         Procedure NeeMe2(rt,nd: nodePtr);
            Begin
               If (nd^.sibL <> NIL) Then Begin
                  If (nd^.sibL^.length <> 0) {test for polytomy}
                  Then nd^.sibL^.length := nd^.temp-nd^.sibL^.temp
                  Else nd^.sibL^.temp := nd^.temp; {"inherit" depth to child}
                  NeeMe2(rt,nd^.sibL);
                  If (nd^.sibR^.length <> 0) {test for polytomy}
                  Then nd^.sibR^.length := nd^.temp-nd^.sibR^.temp
                  Else nd^.sibR^.temp := nd^.temp; {"inherit" depth to child}
                  Neeme2(rt,nd^.sibR)
               End; {then}
            End; {procedure grafenMe2}
         {******************************************************************}
         Begin {Nee}
            drawTop(part,rt,top,mx,my,height);
            setColor(chi);
            xcoor:= round(0.1*mx);
            message:='(N)ee:    change branch lengths to Nee''s method.';
            outTextXY(xcoor,round(0.875*my),message);
            setColor(cnorm);
            setFillStyle(solidfill,0);
            bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
            setColor(chi);
            xcoor:= round(0.02*mx);
            message:='Are you sure that you want to Nee transform the '+
                       'branch lengths? (Y/N)';
            outTextXY(xcoor,round(0.68*my),message);
            Repeat
               ch:=upCase(readKey);
            Until (ch in ['Y','N']);
            If (ch='Y') Then Begin
               NeeMe1(rt,rt,dummy);
               NeeMe2(rt,rt)
            End; {then}
         End; {procedure Nee}


      {********************************************************************}
      {This procedure sets all the branch lengths to those dictated by
      Pagel's (1992, J. of Theoretical Biol 156:431-442) arbitrary method.
      The depth of a node is equal to the maximum of the number of
      bifurcation levels of the left child and the right child plus one.
      If a node is a polytomy, its depth is set to the maximum of the
      bifurcation levels of the offspring, but one is not added.  This
      results in polytomies being assigned the least depth possible.  This
      was an internal decision, as Pagel(1992, p. 441 paragraph 7) does
      not discuss how to assign a depth to a polytomy node.                }                
      {********************************************************************}
      Procedure pagel(part:boolean;
                      rt: nodeptr;
                      top : boolean;
                      mx,my :integer);
         Var
            message : string[80];
            xcoor : integer;
            height : extended;
            dummy : integer;
         {******************************************************************}
         {This procedure computes the depths recursively.  The original
          Pagelme procedure was split into 2 pieces to support polytomy.
          PEM 9-May-95.                                                      }
         {******************************************************************}
         Procedure pagelMe1(rt,nd: nodePtr;Var bifurcations :integer);
            Var
               Rbif, Lbif :integer;
            Begin
              If (nd^.sibL = NIL) Then Begin
                 bifurcations:=0;
                 nd^.temp := 0 {PEM 9-May-95}
              End {then}
              Else Begin
                 pagelme1(rt,nd^.sibL, Lbif);
                 pagelme1(rt,nd^.sibR, Rbif);
                 If (Lbif > Rbif) Then bifurcations:=Lbif +1
                 Else bifurcations:=Rbif +1;
                 If ((nd^.length = 0) And  { test for polytomy }
                     (nd <> rt))           { avoid unintended root polytomy }
                 Then bifurcations := bifurcations - 1; { don't add depth }
                                                     { for polytomies. PEM }
                 nd^.temp:=bifurcations;
              End; {else}
            End; {procedure pagelme1}
        {*******************************************************************}
        {This recursive procedure finishes the transformation (taking
         polytomies into account.  Note that it does its work while going
         back down the tree (toward the root), while pagelme2 does its work
         going up the tree (toward the tips).  Added PEM 9-May-95.          }
        {*******************************************************************}
        Procedure pagelMe2(rt,nd: nodePtr);
           Var
              Rbif, Lbif :integer;
           Begin
              If (nd^.sibL <> NIL) Then Begin
                 If (nd^.sibL^.length <> 0) {test for polytomy}
                 Then nd^.sibL^.length := nd^.temp-nd^.sibL^.temp
                 Else nd^.sibL^.temp := nd^.temp; {"inherit" depth to child}
                 pagelMe2(rt,nd^.sibL);
                 If (nd^.sibR^.length <> 0) {test for polytomy}
                 Then nd^.sibR^.length := nd^.temp-nd^.sibR^.temp
                 Else nd^.sibR^.temp := nd^.temp; {"inherit" depth to child}
                 pagelMe2(rt,nd^.sibR)
              End {then}
           End; {procedure pagelme2}
        {*******************************************************************}
        Begin {pagel}
           drawTop(part,rt,top,mx,my,height);
           setColor(chi);
           xcoor:= round(0.1*mx);
           message:='(P)agel:     change branch lengths to Pagel''s method.';
           outTextXY(xcoor,round(0.85*my),message);
           setColor(cnorm);
           setFillStyle(solidfill,0);
           bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
           setColor(chi);
           xcoor:= round(0.02*mx);
           message:='Are you sure that you want to Pagel transform the'+
                                    ' branch lengths? (Y/N)';
           outTextXY(xcoor,round(0.68*my),message);
           Repeat
              ch:=upcase(readkey);
           Until (ch in ['Y','N']);
           If (ch='Y') Then Begin
              pagelMe1(rt,rt,dummy);
              pagelMe2(rt,rt)
           End {then}
        End; {procedure pagel}
   {***********************************************************************}
   {This procedure transforms any set of branch lengths by Grafen's "rho"
   transform.  To do this it first prompts the user to enter the rho
   parameter.  It then checks the tree's height, and saves it.  It then
   scales the tree so its height is one.  It changes the branch lengths to
   be a measure of the distance from the top of the tree to each node.
   Note that the branch length field is being used for something other than
   storing branch lengths.  It then raises each distance (stored in branch
   length) to the power rho.  Finally it changes the distances back to
   branch lengths and rescales the tree.  Note that if the original tree
   had contemporaneous tips, then the transformed tree will also.  For a
   discussion of transforming branch lengths in the context of Felsenstein's
   (1985) method of phylogenetically independent contrasts, see Garland
   et al. (1992, Systematic Biology 41:18-32.}
   {***********************************************************************}
   Procedure grafen2(part:boolean; r:nodePtr; top:boolean; mx,my:integer);
      Var
         rho, multiplier :extended;
         success: boolean;
         dumb:boolean;
         message: string[100];
         xcoor:integer;
         height:extended;
      {**********************************************************************}
      {This recursive procedure transforms the tree so that the branch
      length field no longer holds the branch length, but holds the distance
      from the highest tip to itself.  Because the procedure assumes that
      the has already been transformed to a height of one, this distance is
      just 1 minus the height of the node.}
      {**********************************************************************}
      Procedure t1(rt,nd:nodePtr; ht:extended);
         Begin
            If (nd<>rt)
            Then ht:=ht+nd^.length;
            If nd^.sibL<>nil Then Begin
               t1(rt,nd^.sibL,ht);
               t1(rt,nd^.sibR,ht);
            End; {then}
            If (nd<>rt)
            Then nd^.length:=1-(ht);
 
            If (nd^.length<1e-18) {length must be due to roundoff error so}
            Then nd^.length:=0;   {zero it out; constant changed 12-27-99 PEM}
         End; {procedure t1}
      {*********************************************************************}
      {This recursive procedure transforms the tree so that the branch
      length field holds the branch length assuming that it previously held
      the distance from the highest tip to the node below previously.
      It also assumes that the height of the tree is one to start with.}
      {*********************************************************************}
      Procedure t2(rt,nd:nodeptr);
         Begin
            If nd^.sibL<>nil then
               begin
                  t2(rt,nd^.sibL);
                  t2(rt,nd^.sibR);
               end;
            If ((nd<>rt) and (nd<>rt^.sibR) and (nd<>rt^.sibL))
            Then
               nd^.length:=nd^.ancestor^.length-nd^.length
            Else
               If (nd<>rt) Then nd^.length:=1-nd^.length;
         End; {procedure t2}
      {*********************************************************************}
      Begin {grafen2}
         drawTop(part,r,top,mx,my,height);
         setColor(chi);
         xcoor:= round(0.1*mx);
         message:='(T)ransform: transform branch lenghts by Grafen''s rho.';
         outTextXY(xcoor,round(0.9*my),message);
         message:='WARNING!!!!!!  If one or more of the tip node''s '+
                   'lengths are less than';
         outTextXY(0,round(0.8*my),message);
         message:='1E-18 times the trees total height, this will give '+
                    'bad results!';
         outTextXY(xcoor,round(0.825*my),message);
         Repeat
            setColor(cnorm);
            setFillStyle(solidfill,0);
            bar3D(0,round(0.625*my),mx,round(0.74*my),1,top);
            setColor(chi);
            xcoor:=round(0.02*mx);
            message:='Please input the parameter rho such that 0 <= rho.';
            outTextXY(xcoor,round(0.65*my),message);
            message:='Enter 1 or S for no change to tree.';
            outTextXY(xcoor,round(0.675*my),message);
            message:= '     Please input a real number, or S for the same:';
            outTextXY(xcoor,round(0.7*my),message);
            rho:=1;
            gotoxy(54,22);
            readreal(rho,success,dumb);
            If (rho<0) Then success:=false; {loop until valid, positive rho}
         Until success;
         multiplier:=1/height;
         scl(r,r,multiplier);
         t1(r,r,0);
         expme(r,r,rho);
         t2(r,r);
         scl(r,r,height);
      end;
   {*********************************************************************}
   Begin {editBranch}
      top:=false;
      graphicsOn(mx,my);
      If (r<>rt)
      Then part:=true
      Else part:=false;
      Repeat
         drawTop(part,rt,top,mx,my,height);
         drawBottom(top,mx,my);
         Repeat
            ch:=upCase(readKey);
         Until (ch in ['S','E','L','P','G','T','C','N','R']);
         Case ch Of
            'S' : scale(part,rt,top,mx,my);
            'E' : transform(part,rt,top,mx,my);
            'L' : logBranch(part,rt,top,mx,my);
            'C' : constantOne(part,rt,top,mx,my);
            'G' : grafen(part,rt,top,mx,my);
            'N' : Nee(part,rt,top,mx,my);
            'P' : pagel(part,rt,top,mx,my);
            'T' : grafen2(part,rt,top,mx,my);
         End; {case ch}
      Until(ch='R');
      closeGraph;
   End; {procedure editBranch}
end. {unit te_brch1}
