{This unit has only one procedure in its external interface, but it
does quite a lot.  It manages all of the graphical editing.}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}



Unit te_see1;       

Interface

Uses t_util1;       {defines type nodePtr}

Procedure Seeit(var realRt,rt:nodePtr; var place:nodePtr);

Implementation

Uses graph,           {Borland graphics libraries}
     crt,             {Borland lo-res text screens}
     dos,             {Borland DOS interface (defines getTime)}
     te_util1;        {Local utility procedures.}

   {************************************************************************}
   {This procedure does the meat of the work of the tree editing.  It manages
   the keyboard input while the user graphically edits the tree.  It also
   contains the procedures that actually implement the editing routines.
   {************************************************************************}
   Procedure Seeit(Var realRt,rt:nodePtr; var place:nodePtr);
      Var
         ch :char;
         lengthMult :extended;
         mx,my:integer;
         height : extended; {changed to extended 3 Jun 1998}
         first :nodePtr; { added PEM 26 Aug 1996 }
         count :integer; { added PEM 26 Aug 1996 }
      {**********************************************************************}
      {This procedure draws a rectangle to outline the text portion below the
      drawn tree.}
      {**********************************************************************}
      Procedure drawRect(mx,my:integer);
         Var
            top :boolean;
         Begin
            top:=false;
            setColor(cnorm);
            setFillStyle(solidFill,0);
            bar3D(round(0.02*mx),round(my*0.8),round(mx*0.98),round(my*0.98),
                                 1,top);
         End; {procedure drawrect}
      {*********************************************************************}
      {This procedure draws a rectangle near the center of the screen
      for a dialogue box.}
      {**********************************************************************}
      Procedure drawRect2(bs:extended; mx,my:integer);
         Var
            top :boolean;
         Begin
            top:=false;
            setColor(cnorm);
            setFillStyle(solidFill,0);
            bar3D(round(0.02*mx),
                  round(bs),
                  round(mx*0.98),
                  round(bs+my*0.24),
                  1,
                  top);
         End; {procedure drawrect2}
      {*********************************************************************}
      {This procedure draws the lines necessary and the square for one node
      in the tree.}
      {**********************************************************************}
      Procedure drawNode(rt,nd :nodePtr;
                         mult:extended;
                         color:integer;
                         mx,my:integer);
         Begin
            setColor(color);
            drawSquare(nd);
            If nd<>rt
            Then lineTo(nd^.tempx,nd^.tempy+round(nd^.length*mult));
         End; {procedure drawNode}
      {*********************************************************************}
      {This procedure draws the text at the bottom of the graphics screen.
      The information changes as the user moves around the tree.  Each time
      they move from one node to another, this must be redrawn.  Note that
      calculating the depth can introduce errors if the height and length
      are equal or nearly so (e.g. at some or all of the tips).  To avoid
      printing questionable digits introduced by conversion errors, if
      the difference is less than 1E-10 * height, the depth is set to 0.}
      {**********************************************************************}
      Procedure drawText(rt,rrt,nd :nodePtr;
                         cnm,chi:integer;
                         fillColor:word;
                         mx,my:integer;
                         height:extended);
         Var
            message : string[80];
            t1,t2,bl : string;
            numStr : string[16];
            length : extended;
            depth  : extended;
            num, starNum, ratio: extended;
            xcoor : integer;
         Begin
            drawRect(mx,my);
            moveTo(round(0.02*mx),round(0.83*my));
            lineTo(round(0.98*mx),round(0.83*my));
            setColor(chi);
            getInfo(rt,num,starNum,ratio);
            extendStr(t1,num);
            extendStr(t2,starNum);
            extendStr(bl,ratio);
            message:='num='+t1+'  starnum='+t2+'  ratio='+bl;
            xcoor:=round(mx*0.04);
            outTextXY(xcoor,round(my*0.81),message);
            extendStr(bl,nd^.length);
            length:=0.0;
            findHtBack(rt,nd,length);
            extendStr(t1,length);
            depth:=height-length;
            If ((height = 0) Or (depth/height < 1E-10))
            Then depth := 0.0;
            extendStr(t2,depth);
            message:='Branch Length = '+bl+
                        '                      Node Name:  '+ nd^.name;
            outTextXY(xcoor,round(my*0.84),message);
            If (nd^.sibL = NIL) { added PEM 26 Aug 96}
            Then Begin
               message:='       Height = '+t1+
                          '                      Tip Number: ';
               str(nd^.tipNum,numStr);
               message := message+numStr;
            End {then}
            Else message:='       Height = '+t1;
            outTextXY(xcoor,round(my*0.87),message);
            message:='       Depth  = '+t2;
            outTextXY(xcoor,round(my*0.90),message);
            If rt<>rrt
            Then
               OutTextXY(xcoor,round(my*0.93),'         '+
                  '*!*!* You are only working with part of the tree *!*!*');
            If nd^.sibL=nil Then Begin
               extendStr(t1,nd^.dat[1]);
               extendStr(t2,nd^.dat[2]);
               message:='      Trait1= '+t1+'       Trait2= '+t2;
               outTextXY(xcoor,round(my*0.96),message);
            End {then}
         End; {procedure drawtext}
      {*********************************************************************}
      {This procedure figures out where the dialogue box should go so as not
      to obscure the current branch.}
      {**********************************************************************}
      Procedure getBr(Var bs:extended;
                      Var ry:integer;
                      nd :nodePtr;
                      mx,my:integer);
         Begin
            If nd^.tempy < my*0.5
            Then Begin
               ry:=22;
               bs:=my*0.56;
            End {then}
            Else Begin
               ry:=10;
               bs:=my*0.16;
            End {else}
         End; {procedure getBr}
      {*********************************************************************}
      {This procedure draws a horizontal line through the current node so
      the user can see which of the other nodes line up with it.}
      {**********************************************************************}
      Procedure lineNd(rt,nd:nodePtr; mx,my:integer);
         Var
            base :extended;
            ready :integer;
            lengthMult:extended;
            ans : char;
         Begin
            getBr(base,ready,nd,mx,my);
            drawRect2(base,mx,my);
            setColor(chi);
            outTextXY(round(mx*0.04),
                      round(base+my*0.08),
                      'Please press any key to continue.');
            moveTo(0,nd^.tempy);
            lineTo(mx,nd^.tempy);
            ans:=readKey;
            reDrawTree(rt,lengthMult,ht1,mx,my);
         End; {procedure linend}
      {*********************************************************************}
      {This procedure removes a node and its descendents from the tree.}
      {**********************************************************************}
      Procedure prune(Var nd,rt,rrt:nodePtr; Var lMult:extended; mx,my:integer);
         Var
            temp, temp2 :nodePtr;
            base : extended;
            textX : integer;
            ready :integer;
            ch: char;
            done : boolean;
         {*******************************************************************}
         {This recursive procedure removes all the children of the current
         node.}
         {*******************************************************************}
         Procedure trimming(nd:nodePtr);
            Begin
               If nd^.sibL <> nil
               Then Begin
                  trimming(nd^.sibL);
                  dispose(nd^.sibL);
                  nd^.sibL:=nil;
                  trimming(nd^.sibR);
                  dispose(nd^.sibR);
                  nd^.sibR:=nil
               End {then}
            End; {procedure trimming}
         {******************************************************************}
         Begin {prune}
            done:=false;
            If (nd<>rt)
            Then Begin
               trimming(nd);
               If Not((nd^.ancestor=rt) And
                      (nd^.ancestor^.sibL^.sibL=nil) And
                      (nd^.ancestor^.sibR^.sibL=nil))
               Then Begin
                  done:=true;
                  temp:=nd^.ancestor;
                  If temp^.sibL = nd Then Begin
                     dispose(temp^.sibL);
                     temp^.sibR^.length:=temp^.sibR^.length+temp^.length;
                     temp2:=temp^.sibR
                  End {then}
                  Else Begin
                     dispose(temp^.sibR);
                     temp^.sibL^.length:=temp^.sibL^.length+temp^.length;
                     temp2:=temp^.sibL
                  End; {else}
                  If temp^.ancestor<>nil
                  Then Begin
                     If temp^.ancestor^.sibL = temp
                     Then temp^.ancestor^.sibL:=temp2
                     Else temp^.ancestor^.sibR:=temp2;
                     nd:=temp^.ancestor;
                     If temp=rt Then Begin
                        rt:=temp2;
                        nd:=rt
                     End {then}
                  End {then}
                  Else Begin
                     rt:=temp2;
                     rrt:=rt;
                     rt^.length:=0;
                     nd:=rt;
                  End; {else}
                  temp2^.ancestor:=temp^.ancestor;
                  dispose(temp);
               End; {then}
            End; {then}
            If Not done
            Then Begin
               getBr(base,ready,nd,mx,my);
               drawRect2(base,mx,my);
               setColor(chi);
               textX := round(mx*0.04);
               outTextXY(textX,
                         round(base+my*0.04),
                         'Unable to complete the prune.  Too little tree left.');
               outTextXY(textX,
                         round(base+my*0.08),
                         'Please press any key to continue.');
               gotoXY(3,ready);
               ch:=readKey
            End; {then}
            reDrawTree(rt,lmult,ht1,mx,my);
         End; {procedure prune}

      {*********************************************************************}
      {This procedure lets the user change the name of a node.  It is not
       directly available from the menu, but is called by change trait
       and reroot.}
      {**********************************************************************}
      Procedure changeName(rrt,nd,rt : nodePtr;
                           lengthMult : extended;
                           mx,my: integer;
                           mesg1 : string;
                           mesg2 : string);
         Var
            dumPtr : nodeptr;
            i :integer;
            base : extended;
            textX : integer;
            ready :integer;
            name : string[2];
            success, good :boolean;
            dum :boolean;
         Begin
            getBr(base,ready,nd,mx,my);
            drawRect2(base,mx,my);
            setColor(chi);
            textX := round(mx*0.04);
            outTextXY(textX,
                      round(base+my*0.02),
                      mesg1+nd^.name);
            outTextXY(textX,
                      round(base +my*0.08),
                      mesg2);
            success:=true;
            good:=true;
            nd^.name:='  ';
            name:= '  ';   { definitely bad }
            Repeat
               setFillStyle(solidFill,0);
               bar(textX,
                   round(base+my*0.1),
                   mx-textX,            {was round((1-0.04)*mx)}
                   round(base+my*0.2)+2*textheight('The'));
               If success=false
               Then
                  outTextXY(textX,
                            round(base+my*0.1),
                            'The name '+
                            name+
                            ' is already being used.  Try Again.');
               If good=false
               Then
                  outTextXY(textX,
                            round(base+my*0.12),
                            'The name '+
                            name+
                            ' is bad.  There can be no spaces.');
               gotoXY(5,ready);
               readln(name);
               success:=true;
               good:=true;
               If ((length(name)<2) or
                   (ord(name[1])<33) or
                   (ord(name[2])<33) or
                   (name[1] = '''') or
                   (name[1] = '"') or
                   (name[1] = ',') or
                   (name[2] = '''') or
                   (name[2] = '"')  or
                   (name[2] = ','))
               Then good:=false
               Else checkname(rrt,name,success,dumptr);
            Until success and good;
            nd^.name:=name;
            reDrawTree(rt,lengthmult,ht1,mx,my);
      End;


      {*********************************************************************}
      {This procedure lets the user change the 2 traits of the current node.}
      {**********************************************************************}
      Procedure changeTrait(rrt,nd,rt :nodePtr;
                            lengthMult:extended;
                            mx,my:integer);
         Var
            dumPtr : nodeptr;
            num :extended;
            base : extended;
            textX : integer;
            ready :integer;
            i :integer;
            message :string[80];
            t1,t2 :string;
            name : string[2];
            success :boolean;
            dum :boolean;
         Begin 
            getBr(base,ready,nd,mx,my);
            If nd^.sibl=nil
            Then For i:=1 To 2 Do Begin
               drawRect2(base,mx,my);
               setColor(chi);
               textX := round(mx*0.04);
               extendStr(t1,nd^.dat[1]);
               extendStr(t2,nd^.dat[2]);
               message:='        Trait1= '+t1+' Trait2= '+t2;
               outTextXY(textX,round(base+my*0.02),message);
               str(i,t1);
               message:='Please input the new value for trait '+t1+
                           ' or S for same.';
               outTextXY(textX,round(base+my*0.08),message);
               Repeat
                  gotoXY(5,ready);
                  readReal(nd^.dat[i],success,dum);
               Until success;
            End; {then for}
            changeName(rrt,nd,rt,lengthMult,mx,my,
                       'Current Name is: ',
                       'Please input a new name for this node');
         End; {procedure changetrait}
      {*********************************************************************}
      {This procedure allows the user to add a node to the tree.}
      {**********************************************************************}
      Procedure add(var nd, rt, rrt:nodePtr;
                    var lengthMult:extended;
                    mx,my:integer);
         Var
            temp,temp2 :nodeptr;
            dum, height :extended;
            tempReal : extended;
            base :extended;
            textX :integer;
            ready :integer;
            message : string;
            ch :char;
            top:boolean;
            success :boolean;
            dumb :boolean;
         Begin
            getBr(base,ready,nd,mx,my);
            drawRect2(base,mx,my);
            setColor(chi);
            textX := round(mx*0.04);
            If nd=rt
            Then message := 'On which side of the root do you want' +
                                ' to add a branch, Left or Right.'
            Else message := 'On which side of the current branch do you want'+
                                ' the new one, Left or Right?';
            outTextXY(textX,round(base+my*0.02),message);
            outTextXY(textX,
                      round(base+my*0.06),
                      'Or you can Escape.  (L/R/E)');
            Repeat
               gotoXY(3,ready);
               ch:=UpCase(Readkey);
            Until ch in ['R','L','E'];
            If ch='E'
            Then Begin
               redrawtree(rt,lengthmult,ht1,mx,my);
               exit
            End; {then}
            new(temp);
            initme(temp);
            If (nd = rt) Then Begin
               temp^.ancestor := nil;
               nd^.ancestor := temp;
               rt := temp;
               rrt := temp;
            End {then}
            Else Begin
               temp^.ancestor:=nd^.ancestor;
               nd^.ancestor:=temp;
            End; {else}
            new(temp2);
            initme(temp2);
            temp2^.ancestor:=temp;
            temp2^.tipnum := -9;  {default until prepare tips is run}
            If (ch = 'R')
            Then Begin
               temp^.sibL:=nd;
               temp^.sibR:=temp2
            End
            Else Begin
               temp^.sibR:=nd;
               temp^.sibL:=temp2
            End; {else}
            If (temp <> rt)
            Then If temp^.ancestor^.sibL = nd
                 Then temp^.ancestor^.sibL:=temp
                 Else temp^.ancestor^.sibR:=temp;
            Repeat
               drawRect2(base,mx,my);
               setColor(chi);
               If (temp <> rt)
               Then message := 'How far up on the current branch do you'+
                                 ' want the node?'
               Else message := 'How far down do you want the new root?';
               outTextXY(textX,round(base+my*0.04),message);
               outTextXY(textX,
                         round(base+my*0.1),
                         'Please input a real number.');
               gotoXY(3,ready);
               readReal(tempReal,success,dumb);
               height := tempReal;
            Until success And ((height<=nd^.length) Or (rt=temp));
            If (temp<>rt) or (rt<>rrt)
            Then Begin
               temp^.length:=height;
               nd^.length:=nd^.length-height
            End
            Else Begin
               temp^.length:=0.0;
               nd^.length:=height
            End; {else}
            height:=0.0;
            findHeight(rt,nd,0.0,height);
            temp2^.length:=height;
            reDrawTree(rt,lengthMult,ht1,mx,my);
            nd:=nd^.ancestor;
            drawNode(rt,nd,lengthMult,chi,mx,my);
            changeTrait(rrt,nd,rt,lengthMult,mx,my);
            drawNode(rt,nd,lengthMult,cnorm,mx,my);
            nd:=temp2;
            drawNode(rt,nd,lengthmult,chi,mx,my);
            changeTrait(rrt,nd,rt,lengthMult,mx,my);
         End; {procedure add}
      {**********************************************************************}
      {This procedure finds the difference in height of the current node, and
      its sibling.}
      {**********************************************************************}
      Procedure findDelta(rt,place: nodePtr;
                          Var delta: extended);
         Var
            x,y,z :extended;
         Begin
            y:=0;                    {Initialize y}
            findHeight(rt,rt,0,y);
            x:=0;
            findHeight(rt,place,0,x);
            z:=0;
            findHtBack(rt,place,z);
            delta:=y-(x+z) + place^.length;
         End; {procedure findDelta}
      {********************************************************************}
      {This procedure allows the user to change the length of the current
      branch to whatever the user wants.  Note the special case of when the
      user enters "s" or "l", or when the current node is the root.  Look
      at the text that goes to the dialogue box for comments.}
      {**********************************************************************}
      Procedure lchange(n,r,rrt :nodePtr; Var lmult:extended; mx,my:integer);
         Var
            base :extended;
            value:extended;
            ready :integer;
            textX : integer;
            message : string[60];
            st : string;
            ans : char;
            success, l  :boolean;
         Begin
            getBr(base,ready,n,mx,my);
            drawRect2(base,mx,my);
            setColor(chi);
            textX := round(mx*0.04);
            If n=r
            Then Begin
               outTextXY(textX,
                         round(base+my*0.04),
                         'You cannot change the length of the root.');
               outTextXY(textX,
                         round(base+my*0.08),
                         'Press any key to continue.');
               ans:=readkey;
            End
            Else Begin
               extendStr(st,n^.length);
               message:='The current length of this branch is : '+st;
               outTextXY(textX,round(base+my*0.04),message);
               outTextXY(textX,
                         round(base+my*0.08),
                         'Input either the new length, S for the same,'+
                         ' or L to line up with the top.');
               value:=n^.length;
               Repeat
                  gotoXY(3,ready);
                  readReal(value,success,l);
               Until success or l;
               If l
               Then Begin
                  findDelta(r,n,value);
                  value:=value+n^.length;
               End; {then}
               If value<0
               Then Begin
                  outTextXY(textX,
                            round(base+my*0.14),
                            'This would give a negative branch length.');
                  outTextXY(textX,
                            round(base+my*0.16),
                            'Procedure Aborted.  Please Press any key.');
                  ans:=readKey
               End {then}
               Else
                  n^.length:=value
            End; {else}
            reDrawTree(r,lmult,ht1,mx,my)
         End; {lchange}
      {*******************************************************************}
      {This procedure adds a constant given it by the user.  It adds this
      constant to the current branch length.  It also adds this constant
      to all other branches that pass through this height.  The effect is
      to sort of cut the tree in half as if it were on a sheet of paper.
      The halves are pulled apart and the space in the middle is filled in
      with lines to connect the two halves.  You could also think of this
      as a stretching.
      {********************************************************************}
      Procedure scalarExt(n,r :nodePtr;
                          Var lmult:extended;
                          mx,my:integer);
         Var
            length: extended;
            add :extended;
            base :extended;
            dum:extended;
            textX : integer;
            ready :integer;
            message : string[60];
            st : string;
            success :boolean;
            ans:char;
            dumb:boolean;
         {******************************************************************}
         {This recursive procedure is initially passed the root, it then
         through recursive calls to itself goes up the tree and when it
         comes to the node of height >= the height to be tested, it checks
         if adding the scalar here will cause a negative branch length.  It
         returns false if it does.}
         {******************************************************************}
         Procedure firstCheck(nd: nodePtr;
                              l,scalar,height :extended;
                              Var succ:boolean);
            Begin
               height:=height+nd^.length;
               If height>=l
               Then Begin
                  If nd^.length+scalar < 0
                  Then succ:=false;
               End
               Else
                  If (nd^.sibR <> NIL)
                  Then Begin
                     firstCheck(nd^.sibR,l,scalar,height,succ);
                     firstCheck(nd^.sibL,l,scalar,height,succ);
                  End; {then}
            End; {procedure firstCheck}
         {*******************************************************************}
         {This actually does the work.  It walks up the tree like the
         previous procedure  until it comes to a branch that is at height
          >= to the height to be tested.  It then adds the scalar to this
          height.}
         {******************************************************************}
         Procedure addOn(nd: nodePtr;
                         l,scalar,height: extended);
            Begin
               height:=height+nd^.length;
               If height>=l
               Then nd^.length:=nd^.length+scalar
               Else
                  If (nd^.sibR<>NIL)
                  Then Begin
                     addOn(nd^.sibR,l,scalar,height);
                     addOn(nd^.sibL,l,scalar,height)
                  End {then}
            End; {procedure addOn}
         {*******************************************************************}
         Begin {scalarExt}
            getBr(base,ready,n,mx,my);
            drawRect2(base,mx,my);
            setColor(chi);
            textX := round(mx*0.04);
            If n=r
            Then Begin
               outTextXY(textX,
                         round(base+my*0.04),
                         'You cannot scalar extend from the root.');
               outTextXY(textX,
                         round(base+my*0.08),
                         'Press any key to continue.');
               ans:=readkey;
            End
            Else Begin
               length := 0;  {initialize length}
               findHtBack(r,n,length);
               extendStr(st,length);
               message:='The distance from root to this node is : '+st;
               outTextXY(textX,round(base+my*0.04),message);
               outTextXY(textX,
                         round(base+my*0.08),
                         'Please input the scalar you want added on, or S for 0.');
               success:=true;
               Repeat
                  If Not success
                  Then
                     outTextXY(textX,
                               round(base+my*0.11),
                               'That value would give a negative branch length.');
                  Repeat
                     gotoXY(3,ready);
                     add:=0.0;
                     readreal(add,success,dumb);
                  Until success;
                  dum:=0.0;
                  firstCheck(r,length,add,dum,success);
               Until success;
               dum:=0.0;
               addon(r,length,add,dum);
            End; {else}
            reDrawTree(r,lmult,ht1,mx,my);
         End; {procedure scaleExt}
      {*********************************************************************}
      {This procedure expands the scope of the current tree to be the
      whole tree.}
      {******************************************************************}
      Procedure pop(Var rt : nodePtr;  {the current root}
                    rrt : nodePtr;     {the "real" root}
                    Var lmult:extended;    {used by redraw}
                    mx,my:integer);    {screen size (for redraw)}
         Begin
            rt:=rrt;
            redrawtree(rt,lmult,ht1,mx,my);
         End; {pop}
      {******************************************************************}
      {This procedure restricts the scope of the current tree to just one
      branch specified by pl.  This is used in conjunction with pop.}
      {******************************************************************}
      Procedure edit(var pl,rt,rrt:nodeptr; var lmult:extended; mx,my:integer);
         Begin
            If pl^.sibL<>nil Then Begin {pl must be an internal node}
               rt:=pl;
               redrawtree(rt,lmult,ht1,mx,my);
            End; {then}
         End; {procedure edit}
      {*********************************************************************}
      {This procedure flashes the current branch so it can be seen on black
      and white monitors that don't distinguish shades of grey.}
      {*********************************************************************}
      Procedure flash(rt,place:nodePtr; mx,my: integer; lengthMult:extended);
         Var
            hour,minute,second,sec100 :word;
            new,current :integer;
            pressed :boolean;
         Begin
            Repeat
               pressed:=false;
               getTime(hour,minute,second,sec100);
               new:=(sec100+30) mod 100;
               Repeat
                  pressed:=keypressed;
                  getTime(hour,minute,second,sec100);
                  current:=sec100;
               Until (((current>new) and (current<(new + 70)))
                      Or (current < (new-30)) or (pressed));
               drawNode(rt,place,lengthmult,cback,mx,my);
               If Not pressed Then Begin
                  getTime(hour,minute,second,sec100);
                  new:=(sec100+30) mod 100;
                  Repeat
                     pressed:=keypressed;
                     getTime(hour,minute,second,sec100);
                     current:=sec100;
                  Until (((current>new) And (current<(new+70))) Or
                         (current < (new-30)) Or (pressed));
                  drawNode(rt,place,lengthmult,chi,mx,my);
               End;
            Until (pressed);
         End;
      {*********************************************************************}
      {This procedure allows the user to enter the name of a node or tip and
       goto it.}
      {*********************************************************************}
      Procedure GoToPlace(Var pl,rt,rrt :nodeptr;
                          Var lmult :extended;
                          mx,my :integer);
         Var spNode : nodePtr;
             base : extended;
             textX : integer;
             ready : integer;
             name : namestr;
             ans : char;
             noGood : boolean;
         Begin
            getBr(base,ready,pl,mx,my);
            drawRect2(base,mx,my);
            setColor(chi);
            textX := round(mx*0.04);
            outTextXY(textX,
                      round(base+my*0.08),
                      'Enter the name of the node or tip (case matters) you want to go to: ');
            gotoxy(5,ready);
            readln(name);
            nogood := true;
            While (noGood) Do Begin
               checkName(rt,name,noGood,spNode);
               If (noGood)
               Then Begin
                  drawRect2(base,mx,my);
                  setColor(chi);
                  outTextXY(textX,
                            round(base+my*0.08),
                            'No node with the name '+name+' could be found in the tree.');
                  outTextXY(textX,
                            round(base+my*0.10),
                            'Enter a new name, or carriage return to give up: ');
                  gotoxy(5,ready);
                  readln(name);
                  If (name = '')  Then Begin
                     noGood := false;
                     spNode := NIL;
                  End; {then}
               End; {then}
            End; {while}
            If spNode <> NIL Then pl := spNode;
            drawRect2(base,mx,my);
            setColor(chi);
            outTextXY(textX,
                      round(base+my*0.08),
                      'Press any key to continue.');
            ans:=readKey;
            reDrawTree(rt,lengthMult,ht1,mx,my);
         End; {procedure gotoplace}

      {*********************************************************************}
      {This procedure allows the user to swap the right and left subtrees of
       the node specified by place.}
      {*********************************************************************}
      Procedure rotatePlace(Var pl,rt,rrt :nodeptr;
                            Var lmult :extended;
                            mx,my :integer);
         Var temp : nodePtr;
         Begin
            If (place^.sibL <> NIL) Then Begin
               temp := place^.sibR;
               place^.sibR := place^.sibL;
               place^.sibL := temp;
            End; {then}
            reDrawTree(rt,lengthMult,ht1,mx,my);
         End; {procedure rotatePlace}

      {*********************************************************************}
      {This procedure allows the user to reroot the tree.                   }
      {*********************************************************************}
      Procedure reRoot(Var place,rt,rrt :nodePtr;
                       Var lmult :extended;
                       mx,my :integer);
         Var newRoot : nodePtr;
             dummyRoot : nodePtr;
             prev : nodePtr;
             ancestor : nodePtr;
             next : nodePtr;
             base : extended;
             temp : extended;
             savedLength : extended;
             ready : integer;
             userChar : char;
             doBranch : boolean;
         {***************************************************************}
         {This little procedure should save some space...                }
         {***************************************************************}
         Procedure lineout(offset : extended;
                           message : string);
            Begin
               outTextXY(round(mx*0.04),
                         round(base+my*offset),
                         message);
            End; {procedure lineout}
         {******************************************************************}
         {This local procedure asks the user whether to reroot at a node
          (false) or the midpoint of the branch to the node's ancestor
          (true).                                                           }
         {******************************************************************}
         Function branchOrNode(where :nodePtr) : boolean;
            Var charGood : boolean;
            Begin
               getBr(base,ready,place,mx,my);
               drawRect2(base,mx,my);
               setColor(chi);
               If (where^.sibR = Nil) Then Begin
                  lineout(0.08,'PDTREE cannot reroot at a tip, so the new root will be at');
                  lineout(0.10,'the midpoint of the branch to the tip''s ancestor.');
                  lineout(0.14,'Press any key to continue');
                  userChar := readKey;
                  branchOrNode := true;
               End
               Else If (where = rt) Then Begin
                       lineout(0.08,'You are already at the root');
                       lineout(0.14,'Press any key to continue');
                       userChar := readKey;
                       branchOrNode := false;
                    End
                    Else If (where^.ancestor <> rt) Or (where^.length > 0) Then Begin
                       charGood := false;
                       Repeat
                          setColor(chi);
                          lineout(0.08,'Do you want to reroot at the (N)ode or at');
                          lineOut(0.10,'the midpoint of the (B)branch (N=Enter/B)?');
                          gotoxy(5,ready);
                          userChar:=readKey;
                          write(userChar);
                          charGood := userChar in ['N','n','B','b',chr(13)];
                          If not(charGood) Then drawRect2(base,mx,my);
                       Until charGood;
                       branchOrNode := userChar in ['B','b'];
                    End;
            End; {function branchornode}
         {****************************************************************}
         Begin
            rt := rrt;    {always do a pop operation to be sure!}
            doBranch := branchOrNode(place);
            If (place <> rt) Then Begin
               If (place^.ancestor = rt) Then Begin
                  If place^.length > 0 Then Begin
                     If doBranch Then Begin
                        place^.length := place^.length*0.5;
                        If (rt^.sibR = place)
                        Then
                          rt^.sibL^.length := rt^.sibL^.length+place^.length
                        Else
                          rt^.sibR^.length := rt^.sibR^.length+place^.length
                     End
                     Else Begin
                        If (rt^.sibR = place)
                        Then
                          rt^.sibL^.length := rt^.sibL^.length+place^.length
                        Else
                          rt^.sibR^.length := rt^.sibR^.length+place^.length;
                        place^.length := 0;
                     End; { ~doBranch}
                  End
                  Else Begin {shuffle the polytomy}
                     If (rt^.sibR = place) Then Begin
                        rt^.sibR := place^.sibL;
                        rt^.sibR^.ancestor := rt;
                        place^.sibL := rt;
                     End
                     Else Begin {rt^.sibL = place}
                        rt^.sibL := place^.sibR;
                        rt^.sibL^.ancestor := rt;
                        place^.sibR := rt;
                     End; { ~ rt^.sibR = place}
                     rt^.ancestor := place;
                     place^.ancestor := nil;
                     rt := place;
                     rrt := rt;
                  End { ~ place^.length > 0}
               End {place^.ancestor = rt}
               Else Begin
                  dummyRoot := new(nodePtr);
                  dummyRoot^.sibL := rt^.sibL;
                  dummyRoot^.sibR := rt^.sibR;
                  dummyRoot^.sibL^.ancestor := dummyRoot;
                  dummyRoot^.sibR^.ancestor := dummyRoot;
                  dummyRoot^.ancestor := NIL;
                  dummyRoot^.name := '**';
                  dummyRoot^.length := 0.0;   {safety}
                  newRoot := rt;
                  If (place = place^.ancestor^.sibR)
                  Then Begin
                     newRoot^.sibR := place;
                     newRoot^.sibL := place^.ancestor;
                     place^.ancestor^.sibR := newRoot;
                  End
                  Else Begin
                     newRoot^.sibL := place;
                     newRoot^.sibR := place^.ancestor;
                     place^.ancestor^.sibL := newRoot;
                  End;
                  savedLength := place^.ancestor^.length;
                  If doBranch Then Begin
                     place^.ancestor^.length := place^.length*0.5;
                     place^.length := place^.length*0.5;
                  End
                  Else Begin
                     place^.ancestor^.length := place^.length;
                     place^.length := 0.0; {thus a polytomy}
                  End;
                  ancestor := place^.ancestor;
                  place^.ancestor := newRoot;
                  prev := newRoot;
                  next := ancestor^.ancestor;
                  Repeat
                     ancestor^.ancestor := prev;
                     {branch lengths}
                     If (ancestor^.sibL = prev) Then Begin
                        ancestor^.sibL := next;
                     End
                     Else Begin
                        ancestor^.sibR := next;
                     End; {else}
                     If (next^.ancestor <> nil) Then Begin
                        temp := next^.length;
                        next^.length := savedLength;
                        savedLength := temp;
                     End; {then}
                     prev := ancestor;
                     ancestor := next;
                     next := next^.ancestor
                  Until (next = nil); {thus ancestor = dummyRoot}
                  If (ancestor^.sibL = prev) Then Begin
                     ancestor^.sibR^.ancestor := prev;
                     ancestor^.sibR^.length :=
                             ancestor^.sibR^.length + savedLength;
                     If (prev^.sibR = ancestor)
                     Then prev^.sibR := ancestor^.sibR
                     Else prev^.sibL := ancestor^.sibR
                  End
                  Else Begin
                     ancestor^.sibL^.ancestor := prev;
                     ancestor^.sibL^.length :=
                             ancestor^.sibL^.length + savedLength ;
                     If (prev^.sibR = ancestor)
                     Then prev^.sibR := ancestor^.sibL
                     Else prev^.sibL := ancestor^.sibL
                  End; {else}
                  dispose(dummyRoot);
               End; {else (place^.ancestor <> rrt)}
               reDrawTree(rt,lengthMult,ht1,mx,my);
               changeName(rrt,rt,rt,lengthMult,mx,my,
                          'The previous root is named ',
                          'What would you like to name the new root? ');
            End; {else (place <> rrt)}
            place := rt;
            reDrawTree(rt,lengthMult,ht1,mx,my);
         End; {procedure reRoot}

   {*********************************************************************}
   Begin {see_it}
      graphicsOn(mx,my);
      redrawTree(rt,lengthmult,ht1,mx,my);
      place:=rt;
      prepareTips(rt,count,first);
      Repeat   
         height:=0.0;
         findHeight(rt,rt,0.0,height);
         If height = 0 Then height := 1;
         drawNode(rt,place,lengthmult,chi,mx,my);
         drawText(rt,realrt,place,cnorm,chi,cback,mx,my,height);
         flash(rt,place,mx,my,lengthmult);
         drawNode(rt,place,lengthmult,chi,mx,my);
         ch:=readkey;
         If ch = #0 Then Begin {read a special key (hopefully an arrow)}
            ch:=readkey; {read the second keycode}
            If (ord(ch)=75) And (place^.sibL<>nil) Then Begin {move left}
               drawNode(rt,place,lengthMult,cnorm,mx,my);
               place:=place^.sibL
            End {then}
            Else
               If (ord(ch)=77) And (place^.sibR<>nil) Then Begin {move right}
                  drawnode(rt,place,lengthmult,cnorm,mx,my);
                  place:=place^.sibR
               End {then}
               Else
                  If (ord(ch)=80) and (place<>rt) Then Begin {move down}
                     drawNode(rt,place,lengthmult,cnorm,mx,my);
                     place:=place^.ancestor;
                  End; {then}
         End {then for special key}
         Else Begin {ordinary key, hopefully an alphabetic one}
            ch:=upCase(ch);
            Case ch Of
               'T' : prune(place,rt,realrt,lengthmult,mx,my);
               'A' : add(place,rt,realrt,lengthmult,mx,my);
               'C' : changetrait(realrt,place,rt,lengthmult,mx,my);
               'D' : linend(rt,place,mx,my);
               'L' : lchange(place,rt,realrt,lengthmult,mx,my);
               'S' : scalarExt(place,rt,lengthmult,mx,my);
               'P' : pop(rt,realrt,lengthmult,mx,my);
               'E' : Edit(place,rt,realrt,lengthmult,mx,my);
               'G' : GoToPlace(place,rt,realrt,lengthmult,mx,my);
               'O' : rotatePlace(place,rt,realrt,lengthmult,mx,my);
               'V' : reRoot(place,rt,realrt,lengthmult,mx,my);
            End; {case ch}
         End; {else}
      Until ch ='R';
      closeGraph;
   End; {procedure see_it}

End. {unit te_see1}
