{This unit has many utility routines that were created for the PDTREE
editing unit.}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}


Unit te_util1;

Interface

Uses t_util1;         {type declarations and utility functions}

Procedure Graphicson(var mx,my:integer);
Procedure drawsquare(nd:nodeptr);
Procedure findheight(rt,nod:nodeptr;height:extended; var max:extended);
Procedure findhtback(rt,nd:nodeptr;var lth:extended);
Procedure drawtree(rt:nodeptr; var htscale: extended; ht:extended;mx,my:integer);
Procedure redrawtree(rt:nodeptr;var lmult:extended;h:extended;mx,my:integer);
Procedure drawPrintableTree(rt:nodePtr; var htscale: extended; ht:extended;mx,my:integer);
Procedure getinfo(rt: nodeptr; var num, starnum, ratio:extended);

Implementation

uses graph;          {Borland graphics package}

   {*********************************************************************}
   {This procedure simply turns on the graphics screen.  It returns the
   size of the screen in mx and my.}
   {*********************************************************************}
   Procedure graphicsOn(var mx,my:integer);
      Var
         graphDriver, graphMode, errorCode :integer;
      Begin
         graphDriver:=Detect;       {keyword to use whatever is avaiable}
         initGraph(GraphDriver, GraphMode,'');
         errorCode:=Graphresult;
         If ErrorCode<>grOk Then Begin
            writeln('Unable to recognize Graphics card');
            writeln('Press Return to Terminate program');
            readln;
            halt;
         End; {then}
         mx:=getmaxx;
         my:=getmaxy;
      End; {procedure graphicsOn}
   {**********************************************************************}
   {This procedure draws a tiny box that is used to show where lines
   connect.}
   {**********************************************************************}
   Procedure drawSquare(nd:nodeptr);
      Begin
         moveTo(nd^.tempx-1,nd^.tempy-1);
         lineTo(nd^.tempx+1,nd^.tempy-1);
         lineTo(nd^.tempx+1,nd^.tempy+1);
         lineTo(nd^.tempx-1,nd^.tempy+1);
         lineTo(nd^.tempx-1,nd^.tempy);
         lineTo(nd^.tempx,nd^.tempy);
      End;
   {*********************************************************************}
   {This recursive procedure finds the distance from the node it is
   given initially to the farthest tip that is a descendant of the initial
   node.  It includes the length of the initial node in the answer.
   It calls nothing but itself.}
   {*********************************************************************}
   Procedure findHeight(rt,nod:nodePtr;height:extended;var max:extended);
      Begin
         If nod<>rt
         Then height:=height+nod^.length;
         If nod^.sibL<> nil Then Begin
            findheight(rt,nod^.sibL,height,max);
            findheight(rt,nod^.sibR,height,max);
         End {then}
         Else
            If height>max Then max:=height;
      End; {procedure findHeight}
   {***********************************************************************}
   {This recursive procedure finds the distance from the node nd to the
   root rt.  It includes the length of nd.}
   {*********************************************************************}
   Procedure findHtBack(rt,nd:nodePtr;var lth:extended);
      Begin
         If nd<>rt Then Begin
            lth:=lth+nd^.length;
            findhtback(rt,nd^.ancestor,lth);
         End {then}
      End; {findHtBack}
   {***********************************************************************}
   {This procedure draws the tree with root rt to the screen.  It calls
   the recursive procedure draw to do most of the work.}
   {*********************************************************************}
   Procedure drawTree(rt:nodePtr; var htscale: extended; ht:extended;mx,my:integer);
      Var
         height  :extended;
         ntScale : extended;
         back : extended;
         ntps : integer;
         ntpsL, ntpsR :integer;
         zeroi : integer;
         zeror : extended;
      {*********************************************************************}
      {This recursive procedure draws the tree with root rt to the screen.
      It is only called by drawtree and relies on drawtree to figure out
      some parameters for it.}
      {*********************************************************************}
      Procedure draw(rt,nd:nodePtr;
                     hscl,nscl :extended;
                     total:extended;
                     Var numberLeaf:integer;
                     height:extended);
         Var
            Lptx,Lpty,Rptx,Rpty, bottom :integer;
         Begin
            If nd^.sibL <> nil Then Begin
               If (nd<>rt)
               Then height:=height+nd^.length;
               draw(rt,nd^.sibL,hscl,nscl,total,numberleaf,height);
               draw(rt,nd^.sibR,hscl,nscl,total,numberleaf,height);
               Lptx := nd^.sibL^.tempx;
               Rptx := nd^.sibR^.tempx;
               Lpty := nd^.sibL^.tempy;
               Rpty := nd^.sibR^.tempy;
               bottom := 3+ round(hscl*(total-height));
               moveTo(Lptx,Lpty);
               lineTo(Lptx,bottom);
               lineTo(Rptx,bottom);
               lineTo(Rptx,Rpty);
               nd^.tempy:=bottom;
               nd^.tempx:=round(0.5*(Rptx-Lptx) + Lptx);
               drawSquare(nd);
            End {then}
            Else Begin
               height:= height+ nd^.length;
               nd^.tempy := 3+ round(hscl*(total-height));
               nd^.tempx := round(nscl*numberleaf + nscl*0.5);
               drawSquare(nd);
               numberLeaf:=numberleaf+1;
            End; {else}
         End; {procedure draw}
   {***********************************************************************}
   Begin {drawtree}
      height:=0.0;
      findHeight(rt,rt,0.0,height);
      If height=0.0
      Then
         htscale:=0.0
      Else
         htscale:= ht*my/height;
      ntps:=0;
      numTps(rt,ntps);
      ntScale:=0.98*mx/ntps;
      ntpsL:=0;
      numTps(rt^.sibL,ntpsL);
      zeror := 0.0;
      zeroi :=0;
      draw(rt,rt,htscale,ntscale,height,zeroi,zeror);
   End; {procedure drawtree}
   {***********************************************************************}
   {This procedure clears the screen then draws the tree with drawtree.}
   {***********************************************************************}
   Procedure reDrawTree(rt:nodePtr;var lmult:extended;h:extended;mx,my:integer);
      Begin
         clearViewPort;
         setbkColor(cback);
         setColor(cnorm);
         drawTree(rt,lmult,h,mx,my);
      End; {procedure reDrawTree}
   {***********************************************************************}
   {This procedure is a modified version of drawTree used for prettier     }
   {screen dumps.  It does not print the little squares at nodes, but does }
   {a small "stem" at the root.                                            }
   {***********************************************************************}
   Procedure drawPrintableTree(rt:nodePtr; var htscale: extended; ht:extended;mx,my:integer);
      Var
         height  :extended;
         ntScale : extended;
         back : extended;
         ntps : integer;
         ntpsL, ntpsR :integer;
         zeroi : integer;
         zeror : extended;
      {*********************************************************************}
      {This recursive procedure draws the tree with root rt to the screen.
      It is only called by drawtree and relies on drawtree to figure out
      some parameters for it.}
      {*********************************************************************}
      Procedure draw(rt,nd:nodePtr;
                     hscl,nscl :extended;
                     total:extended;
                     Var numberLeaf:integer;
                     height:extended);
         Var
            Lptx,Lpty,Rptx,Rpty, bottom :integer;
         Begin
            If nd^.sibL <> nil Then Begin
               If (nd<>rt)
               Then height:=height+nd^.length;
               draw(rt,nd^.sibL,hscl,nscl,total,numberleaf,height);
               draw(rt,nd^.sibR,hscl,nscl,total,numberleaf,height);
               Lptx := nd^.sibL^.tempx;
               Rptx := nd^.sibR^.tempx;
               Lpty := nd^.sibL^.tempy;
               Rpty := nd^.sibR^.tempy;
               bottom := 3+ round(hscl*(total-height));
               moveTo(Lptx,Lpty);
               lineTo(Lptx,bottom);
               lineTo(Rptx,bottom);
               lineTo(Rptx,Rpty);
               nd^.tempy:=bottom;
               nd^.tempx:=round(0.5*(Rptx-Lptx) + Lptx);
            End {then}
            Else Begin
               height:= height+ nd^.length;
               nd^.tempy := 3+ round(hscl*(total-height));
               nd^.tempx := round(nscl*numberleaf + nscl*0.5);
               numberLeaf:=numberleaf+1;
            End; {else}
         End; {procedure draw}
   {***********************************************************************}
   Begin {drawprintabletree}
      height:=0.0;
      findHeight(rt,rt,0.0,height);
      If height=0.0
      Then
         htscale:=0.0
      Else
         htscale:= ht*my/height;
      ntps:=0;
      numTps(rt,ntps);
      ntScale:=0.98*mx/ntps;
      ntpsL:=0;
      numTps(rt^.sibL,ntpsL);
      zeror := 0.0;
      zeroi :=0;
      draw(rt,rt,htscale,ntscale,height,zeroi,zeror);
      moveTo(rt^.tempx,rt^.tempy);
      lineTo(rt^.tempx,rt^.tempy+4);
   End; {procedure drawtree}


   {*************************************************************************}
   {This procedure figures out 3 numbers and returns them.  First it figures
   out the sum of all the branch lengths and the total number of tips.
   Second it multiplies the number of tips by the height of the tree to get
   the sum of all the branch lengths if the tree were a star with all the tips
   lining up on top.  Finally it reports these 2 lengths as well as their
   ratio.}
   {*************************************************************************}
   Procedure getinfo(rt: nodeptr; var num, starnum, ratio:extended);
      Var
         numTips : integer;
         height : extended;
         dum :extended;
      {********************************************************************}
      {This recursive procedure traverses the tree pointed to by rt and
      figures out the height, number of tips, and the sum of all the branch
      lengths in that tree.}
      {********************************************************************}
      Procedure traverse(nd: nodePtr;
                         Var numTips:integer;
                         Var sum, max: extended;
                         height: extended);
         Begin
            height:=height+nd^.length;
            sum:=sum+nd^.length;
            If (nd^.SibR = NIL) Then Begin
               inc(numTips);
               If (height > max) Then max:=height;
            End {then}
            Else Begin
               traverse(nd^.sibL,numTips,sum,max,height);
               traverse(nd^.sibR,numTips,sum,max,height)
            End {else}
         End; {procedure traverse}
      {********************************************************************}
      Begin {getinfo}
         height:=0;
         dum:=0;
         numtips:=0;
         num:=0;
         traverse(rt,numTips, num, height, dum);
         starNum:=height*numTips;
         If starnum<>0
         Then ratio:=num/starnum
         Else ratio:=0;
      End; {procedure getinfo}

End. {unit te_util1}
