{**************************************************************************}
{ TP_IO.pas }
{ This program copyright 1993,1994 ???}
{ Written by Peter E. Midford, with extensive copying from PDTREE.PAS}
{ Last modified 24 Feb 1994.}
{**************************************************************************}
{This unit has the input and output procedures.  It provides for I/O of a
tree to and from disk as well as the creation of a tree.}

Unit tp_io;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Interface

Uses tp_utils, {type definitions, constants, and utility functions.}
     dos;      {access to filename types}


Procedure openSimFile(Var flName :pathstr;
                      Var ext :typString;
                      Var simFile :text;
                      Var succ :boolean); 

Procedure scanTree(Var tipCount :integer; Var simFile :text);

Procedure readDataSet(Var simData :nodeDataPtr;
                          tipCount : integer;
                      Var simFile :text);

Procedure openStatFile(fleName :pathStr;
                       Var statFile :text;
                       Var statFileName :pathStr);

Procedure openBoxFile(flename :pathStr;
                      Var boxFile :text;
                      Var boxFileName :pathStr);

Procedure newSimExt(simName : pathStr;
                    Var newExt : extStr;
                    doBCStats : boolean;
                    doCategories : boolean;
                    doRanks : boolean);

Procedure openNewSimFile(fleName :pathStr;
                         Var sitFile :text;
                         Var newSimFileName : pathStr;
                         doBCStats : boolean;
                         doCategories : boolean;
                         doRanks : boolean);

Procedure writeStatistics(SimStats :statPak;
                          setNumber :integer;
                          Var statFile :text);

Procedure writeFullStatistics(simStats :statPak;
                              setNumber :integer;
                              Var statFile :text);
 
Procedure writeBCStats(simStats :statPak;
                       bcStats :statPak;
                       setNumber :integer;
                       Var boxFile :text);

Procedure copySimHeader(Var simFile :text;
                        Var newSimFile :text;
                        simFileName :pathStr);

Procedure writeQSet(Var qSimFile :text;
                    tipData1 :tipArray;
                    tipData2 :tipArray;
                    data :nodeDataPtr);

Procedure extendWrite(Var outFile :text; num :extended);

Procedure scanForLimits(inputFLName : pathStr;
                        Var globalMin1 : extended;
                        Var globalMax1 : extended;
                        Var globalMin2 : extended;
                        Var globalMax2 : extended;
                        tipCount : integer;
                        userLimit : integer);

Implementation

Uses crt;     {single key input}

Const textBuffSize = 65500; {size of file I/O buffers}

Type textBuff = array[1..textBuffSize] of char;
     textBuffPtr = ^textBuff;
Var inBuff :textBuffPtr;    {buffer for input (SIM) file}
    outBuff :textBuffPtr;   {buffer for TPS file}
    outBuff2 :textBuffPtr;  {buffer for BOX file}
    outBuff3 :textBuffPtr;  {buffer for qSIM file}


{**************************************************************************}
{This procedure asks the user for the name of the input SIM file,
 validates the name (right syntax, file exists), and opens the file.}
{**************************************************************************}

Procedure openSimFile(Var flName :pathStr;
                      Var ext :typString;
                      Var SimFile :text;
                      Var succ :boolean);

   {************************************************************************}
   {This procedure reads in a file name.  That is simple, but the task gets
   more complex when you require that the program not crash when invalid
   junk is typed in by the user.  That is the reason that this routine is
   a bit of a mess.  Hopefully you won't have to mess with it.}
   {************************************************************************}
   Procedure getInFile(Var fleName :pathStr;
                       Var exten :typString;
                       Var succ :boolean);
      Var
         ch :char;
         place :byte;
         exists :boolean;
         i :integer;
         fl :text;
      Begin
         exists := true;
         Repeat
            If not exists Then
               writeln(fleName,' does not exist.');
            exists := true;
            writeln('Please input the name of the input file including the extension');
            writeln('   Or "Q" to quit program');
            readln(fleName);
            writeln;
            If ((fleName = 'Q') or (fleName = 'q')) Then halt;
            {$I-}               {turn off I/O error checking}
            assign(fl,fleName);
            reset(fl);
            close(fl);
            {$I+}               {turn I/O checking back on}
            exists := (IOResult=0) and (fleName<>'')
         Until (exists or (fleName ='R') or (fleName='r'));
         If (fleName ='R') or (fleName='r') Then   {If user wants to abort}
            Begin
               succ := false;
               halt            {Bail out here!}
            End {if}
      End; {procedure getinfile}

   Begin {procedure opensimfile}
       succ := true;
       getinfile(flName,ext,succ);
       If succ Then Begin
          new(inBuff);
          assign(simFile,flName);
          setTextBuf(simFile,inBuff^)
       End {then}
   End; {procedure OpenSimFile}


{***************************************************************************}
{Ok, here we go, this procedure is long, but it only reads in a tree.  It
passes back a pointer to the root, the filename, the extension of the
filename, and succ if it did input a file.}
{***************************************************************************}
Procedure scanTree(Var tipCount :integer; Var simFile :text);
   Var
      ch :char;
      fail :boolean;
      ok,good :boolean;
      name :nameString;
      ch2 :string[2];
      chhh :char;
      ch4 :string[4];
      tempInt : integer;
      currentValues :nodeDataPtr;
      trait1,trait2 :extended;
      setDone :boolean;
   {***********************************************************************}
   begin
      reset(simFile);
      Repeat
         Read(SimFile,ch4);
         Readln(SimFile)
      Until ((ch4='(Nod') or eof(simFile));
      If eof(simFile) Then Begin
         writeln('Input file is not recognizable as a simulation or randomization file.');
         writeln('Please check the file or run the program with a different file.');
         writeln;
         writeln('Exiting...');
         halt;
      End;
      tipCount := 0;
      Repeat
         read(simFile,ch4);
         readln(simFile);
         inc(tipCount);
      Until ch4='SIMU';
      Repeat
         read(simFile,ch);
         readln(simFile)
      Until ch='_'
   End; {procedure readtree}


{***************************************************************************}
{This procedure reads one set of data from a data file and returns a
 linked list of name-value pairs.
{***************************************************************************}
Procedure readDataSet(Var simData :nodeDataPtr;
                          tipCount : integer;
                      Var simFile :text);
  Var
    currentValues :nodeDataPtr;
    count :integer;
    setDone :boolean;
    ch :char;
  Begin
      simData := nil;
      setDone := false;
      count := 0;
      While ((not setDone) and (not eof(simFile))) Do Begin
         count := count + 1;
         If (simData = nil) Then Begin
            new(currentValues);
            initNodeData(currentValues);
            simData := currentValues
         End {then}
         Else Begin
            new(currentValues^.next);
            initNodeData(currentValues^.next);
            currentValues := currentValues^.next           
         End; {else}
         read(simFile,currentValues^.name);
         read(simFile,currentValues^.data[1]);
         read(simFile,currentValues^.data[2]);
         read(simFile,ch);
         read(simFile,currentValues^.ansName);
         read(simFile,currentValues^.ansData[1]);
         read(simFile,currentValues^.ansData[2]);
         If (count = tipCount) Then setDone := true;
         write('');
         readln(simFile)
      End; {while}
   End; {procedure readdataset}
       
{***************************************************************************}
{This procedure opens the statistics file.  The file object is returned.}
{***************************************************************************}
Procedure openStatFile(flename :pathStr;
                       Var statFile :text;
                       Var statFileName :pathStr);
   Var place :integer;
       i : integer;
       root :pathStr;
       ext : pathStr;
   Begin
     place := pos('.',fleName);
     root := copy(fleName,1,place);
     ext := copy(fleName,place+1,3);
     For i := 1 to 3 Do ext[i] := upcase(ext[i]);
     If (ext = 'SIT') Then statFileName := root + 'TPT'
     Else If (ext = 'SCM') Then statFileName := root + 'TCS'
          Else If (ext = 'SCT') Then statFileName := root + 'TCT'
               Else statFileName := root + 'TPS';
     new(outBuff);
     assign(statFile,statFileName);
     setTextBuf(statFile,outBuff^);
     rewrite(statFile)
   End; {procedure openstatfile}

{***************************************************************************}
{This procedure opens the file containing skew and kurtosis data.  The file
 object is returned.}
{***************************************************************************}
Procedure openBoxFile(fleName :pathStr;
                       Var boxFile :text;
                       Var boxFileName : pathStr);
   Var place :integer;
       root :pathStr;
   Begin
      place := pos('.',flename);
      root := copy(fleName,1,place);
      boxFileName := root + 'BOX';
      new(outBuff2);
      assign(boxFile,boxFileName);
      setTextBuf(boxFile,outBuff2^);
      rewrite(boxFile)
   End; {procedure openboxfile}

{***************************************************************************}
{***************************************************************************}
Procedure newSimExt(simName : pathStr;
                    Var newExt : extStr;
                    doBCStats : boolean;
                    doCategories : boolean;
                    doRanks : boolean);
   Var
       sourceChar : char;
       simDir : dirStr;
       simFName : nameStr;
       simExt : extStr;
   Begin
      fsplit(simName,simDir,simFName,simExt);
      sourceChar := simExt[2];
      sourceChar := upCase(sourceChar);
      If doBCStats
      Then
         If doCategories
         Then newExt := '.' + sourceChar + 'CT'
         Else If doRanks
              Then newExt := '.' + sourceChar + 'RT'
              Else newExt := '.' + sourceChar + 'IT'
      Else
         If doCategories
         Then newExt := '.' + sourceChar + 'CM'
         Else newExt := '.' + sourceChar + 'RM'
   End; { procedure newSimExt}

{***************************************************************************}
{This procedure opens the modified simulation data file.  The file object is
 returned.}
{***************************************************************************}
Procedure openNewSimFile(fleName :pathStr;
                         Var sitFile :text;
                         Var newSimFileName : pathStr;
                         doBCStats : boolean;
                         doCategories : boolean;
                         doRanks : boolean);
   Var
       simDir : dirStr;
       simFName : nameStr;
       simExt : extStr;
   Begin
      fsplit(fleName,simDir,simFName,simExt);
      newSimExt(fleName,simExt,doBCStats,doCategories,doRanks);
      newSimFileName := simDir + simFName + simExt;
      new(outBuff3);
      assign(sitFile,newSimFileName);
      setTextBuf(sitFile,outBuff3^);
      rewrite(sitFile)
   End; {procedure opensitfile}

{***************************************************************************}
{This procedure writes coprocessor extended reals to a file without the
 extra exponent and mantessa digits.                                       }
{***************************************************************************}
Procedure extendWrite(Var outFile :text; num :extended);
   Var outStr : string[19];
   Begin
      str(num:19:-14,outStr);
      delete(outStr,16,2);
      write(outFile,outStr,' ')
   End; {procedure extendwrite}

{***************************************************************************}
{This procedure writes the statistics for one simulated data set on a line
 containing 17 columns.}
{***************************************************************************}
Procedure writeStatistics(simStats :statPak;
                          setNumber :integer;
                          Var statFile :text);
   Begin
      write(statFile,setNumber:5,' ');      {Column 1 is set number}
      With simStats Do Begin
         extendWrite(statFile,corr);        {Column 2 is correlation}
         extendWrite(statFile,rho);         {Column 3 is spearman rank correlation}
         extendWrite(statFile,min[1]);      {Column 4 is minimum of trait 1}
         extendWrite(statFile,min[2]);      {Column 5 is minimum of trait 2}
         extendWrite(statFile,max[1]);      {Column 6 is maximum of trait 1}
         extendWrite(statFile,max[2]);      {Column 7 is minimum of trait 2}
         extendWrite(statFile,mean[1]);     {Column 8 is mean of trait 1}
         extendWrite(statFile,mean[2]);     {Column 9 is mean of trait 2}
         extendWrite(statFile,variance[1]); {Column 10 is variance of trait 1}
         extendWrite(statFile,variance[2]); {Column 11 is variance of trait 2}
         extendWrite(statFile,f);           {Column 12 is F statistic}
         extendWrite(statFile,ls);          {Column 13 is OLS slope}
         extendWrite(statFile,lsy);         {Column 14 is OLS intercept}
         extendWrite(statFile,rma);         {Column 15 is RMA slope}
         extendWrite(statFile,rmay);        {Column 16 is RMA intercept}
         extendWrite(statFile,ma);          {Column 17 is MA slope}
         extendWrite(statFile,may);         {Column 18 is MA intercept}
      End; {with simstats}
      writeln(statFile)
   End; {procedure writestatistics}

{***************************************************************************}
{This procedure writes the full statistics for one simulated data set on a
 line containing 25 columns.}
{***************************************************************************}
Procedure writeFullStatistics(simStats :statPak;
                              setNumber :integer;
                              Var statFile :text);
   Begin
      write(statFile,setNumber:5,' ');      {Column 1 is set number}
      With simStats Do Begin
         extendWrite(statFile,corr);        {Column 2 is correlation}
         extendWrite(statFile,rho);         {Column 3 is spearman rank correlation}
         extendWrite(statFile,min[1]);      {Column 4 is minimum of trait 1}
         extendWrite(statFile,min[2]);      {Column 5 is minimum of trait 2}
         extendWrite(statFile,max[1]);      {Column 6 is maximum of trait 1}
         extendWrite(statFile,max[2]);      {Column 7 is minimum of trait 2}
         extendWrite(statFile,mean[1]);     {Column 8 is mean of trait 1}
         extendWrite(statFile,mean[2]);     {Column 9 is mean of trait 2}
         extendWrite(statFile,variance[1]); {Column 10 is variance of trait 1}
         extendWrite(statFile,variance[2]); {Column 11 is variance of trait 2}
         extendWrite(statFile,f);           {Column 12 is F statistic}
         extendWrite(statFile,ls);          {Column 13 is OLS slope}
         extendWrite(statFile,lsy);         {Column 14 is OLS intercept}
         extendWrite(statFile,rma);         {Column 15 is RMA slope}
         extendWrite(statFile,rmay);        {Column 16 is RMA intercept}
         extendWrite(statFile,ma);          {Column 17 is MA slope}
         extendWrite(statFile,may);         {Column 18 is MA intercept}
         extendWrite(statFile,median[1]);   {Column 19 is median of trait 1}
         extendWrite(statFile,median[2]);   {Column 20 is median of trait 2}
         extendWrite(statFile,ks[1]);       {Column 21 is KS p-value of trait 1}
         extendWrite(statFile,ks[2]);       {Column 22 is KS p-value of trait 2}
      End; {with simstats}
      writeln(statFile)
   End; {procedure writestatistics}


{***************************************************************************}
{This procedure writes before and after skew and kurtosis statistics for one
 simulated data set on a line containing 9 columns.}
{***************************************************************************}
Procedure writeBCStats(simStats :statPak;
                       bcStats :statPak;
                       setNumber :integer;
                       Var boxFile :text);
   Begin
      write(boxFile,setNumber:5,' ');              {Column 1 is set number}
      With simStats Do Begin
         extendWrite(boxFile,lambda[1]);    {Column 2 is BC lambda of raw data}
         extendWrite(boxFile,lambda[2]);    {Column 3 is BC lambda of raw data}
         extendWrite(boxFile,ks[1]);         {Column 4 is KS p-value of trait 1 after box-cox}
         extendWrite(boxFile,ks[2]);         {Column 5 is KS p-value of trait 2 after box-cox}
         extendWrite(boxFile,skewness[1]); {Column 6 is skewness of trait 1 before box-cox}
         extendWrite(boxFile,skewness[2]); {Column 7 is skewness of trait 2 before box-cox}
         extendWrite(boxFile,kurtosis[1]); {Column 8 is kurtosis of trait 1 before box-cox}
         extendWrite(boxFile,kurtosis[2]); {Column 9 is kurtosis of trait 2 before box-cox}
         extendWrite(boxFile,skewness[1]);  {Column 10 is skewness of trait 1 after box-cox}
         extendWrite(boxFile,skewness[2]);  {Column 11 is skewness of trait 2 after box-cox}
         extendWrite(boxFile,kurtosis[1]);  {Column 12 is kurtosis of trait 1 after box-cox}
         extendWrite(boxFile,kurtosis[2]);  {Column 13 is kurtosis of trait 2 after box-cox}
      End; {with simstats}
      writeln(boxFile)
   End; {procedure writebcstats}

{***************************************************************************}
{This procedure copies the header portion of the old sim file into the new
 sim file.  The header ends with a line of dashes, after which the trait
 values will be written by writeqset.  A string indicating that PDTIPS has
 modified the file is added to the second line which otherwise gives the
 date with PDSIMUL wrote the original file.  PDANOVA crashes if the number
 of lines in the header is changed.}
{***************************************************************************}
Procedure copySimHeader(Var simFile :text;
                        Var newSimFile :text;
                        simFileName :pathStr);

   Var simLine :string[255];
       modStr :string[255];
       tmpStr :string;
       DT :dateTime;
       sec100 :word;
       ignore :word;
   Begin
      {First build the string indicating when the file was modified.}

      getDate(DT.year,DT.month,DT.day,ignore);
      getTime(DT.hour,DT.min,DT.sec,sec100);
      modStr := ' Modified by PDTIPS: ';
      str(DT.day,tmpStr);
      modStr := modStr + tmpStr;
      Case DT.month Of
          1 : modStr := modStr + ' Jan ';
          2 : modStr := modStr + ' Feb ';
          3 : modStr := modStr + ' Mar ';
          4 : modStr := modStr + ' Apr ';
          5 : modStr := modStr + ' May ';
          6 : modStr := modStr + ' Jun ';
          7 : modStr := modStr + ' Jul ';
          8 : modStr := modStr + ' Aug ';
          9 : modStr := modStr + ' Sep ';
         10 : modStr := modStr + ' Oct ';
         11 : modStr := modStr + ' Nov ';
         12 : modStr := modStr + ' Dec ';
      End; {case}
      str(DT.year,tmpStr);
      modStr := modStr + tmpStr + ' at ';
      str(DT.hour,tmpStr);
      modStr := modStr + tmpStr + ':';
      str(DT.min,tmpStr);
      modStr := modStr + tmpStr + ':';
      str(DT.sec,tmpStr);
      modStr := modStr + tmpStr + '.';
      str(sec100,tmpStr);
      modStr := modStr + tmpStr;

      {Copy the old file into the new file until the line of dashes is
       encountered.  Append the modification string to the second line.}
      reset(simFile);
      rewrite(newSimFile);
      readln(simFile,simLine);
      writeln(newSimFile,simLine);
      readln(simFile,simLine);
      simLine := simLine + ModStr;
      writeln(newSimFile,simLine);
      Repeat
         readln(simFile,simLine);
         writeln(newSimFile,simLine)
      Until simLine[1] = '_'
   End; {procedure copysimheader}

{***************************************************************************}
{This procedure writes one simulation's worth of modified data.  The
 node names come from the tree, while the data values come from the tipdata
 arrays.}
{***************************************************************************}
Procedure writeQSet(Var qSimFile :text;
                    tipData1 :tipArray;
                    tipData2 :tipArray;
                    data :nodeDataPtr);
   Var i :integer;
       curNode :nodeDataPtr;
   Begin
     i := 0;
     curNode := data;
     While (curNode <> nil) Do Begin
        inc(i);
        write(qSimFile,curNode^.Name,' ');
        extendWrite(qSimFile,tipData1[i]);
        extendWrite(qSimFile,tipData2[i]);
        write(qSimFile,curNode^.ansName,' ');
        extendWrite(qSimFile,0.0);
        extendWrite(qSimFile,0.0);
        writeln(qSimFile);
        curNode := curNode^.next;
     End {while}
   End; {procedure writeqset}

{***************************************************************************}
{***************************************************************************}
Procedure scanForLimits(inputFLName : pathStr;
                        Var globalMin1 : extended;
                        Var globalMax1 : extended;
                        Var globalMin2 : extended;
                        Var globalMax2 : extended;
                        tipCount : integer;
                        userLimit : integer);
   Const smallValue = -9.999E99;
         largeValue = 9.999E99;
   Var inFile : text;
       i : longInt;
       lineLimit : longInt;
       ch4 : string[4];
       tempName : nameString;
       tempVal1 : extended;
       tempVal2 : extended;
       ch : char;
   Begin
      lineLimit := tipCount*userLimit;
      globalMin1 := largeValue;
      globalMax1 := smallValue;
      globalMin2 := largeValue;
      globalMax2 := smallValue;
      assign(inFile,inputFLName);
      reset(inFile);
      Repeat
         read(inFile,ch4);
         readln(inFile)
      Until ch4='(Nod';
      Repeat
         read(inFile,ch4);
         readln(inFile);
      Until ch4='SIMU';
      Repeat
         read(inFile,ch);
         readln(inFile)
      Until ch='_';
      i := 1;
      While ((not eof(inFile)) And (i <= lineLimit)) Do Begin
         read(inFile,tempName);
         read(inFile,tempVal1);
         read(inFile,tempVal2);
         readln(inFile);
         If (tempVal1 < globalMin1) Then globalMin1 := tempVal1;
         If (tempVal1 > globalMax1) Then globalMax1 := tempVal1;
         If (tempVal2 < globalMin2) Then globalMin2 := tempVal2;
         If (tempVal2 > globalMax2) Then globalMax2 := tempVal2;
         inc(i);
      End; {while}
      close(inFile);
   End; {procedure scanForLimits}

End.