{**************************************************************************}
{ tp_utils.pas }
{ This program copyright 1993,1994 ???}
{ Written by Peter E. Midford, with extensive copying from PDTREE.PAS}
{ Last modified 2 Apr 1994.}
{**************************************************************************}

Unit tp_utils;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}

{The following switch has no effect in units, but may vary between
 application programs in this suite.}

{$M 16384,0,655360} {medium stack size, maximum heap size}

{End compiler switch block}
{-------------------------------------------------------------------------}

Interface

   Const maxTips = 300;

   {Define all the types used in PDTIPS.}

   Type
      flString = string[30];             {File name string}
      typString = string[3];             {File extension string}
      nameString = string[2];            {Name of a node}
      traits = array[1..2] of extended;  {Real number trait pair}
      itraits = array[1..2] of integer;  {Integer trait pair}

      statPak = Record                   {Accumulated statistics}
         min      :traits;               {Minimum of trait values in tree}
         max      :traits;               {Maximum of trait values in tree}
         mean     :traits;               {Mean trait value}
         variance :traits;               {Variance of trait values}
         skewness :traits;               {Skewness of trait values}
         kurtosis :traits;               {Kurtosis of trait values}
         median   :traits;               {Median of trait values}
         ks       :traits;               {K-S statistic of trait values}
         lambda   :traits;               {Box-Cox lambda of trait values}
         rankMean :traits;               {used for calculating rho}
         rankVar  :traits;               {used for calculating rho}
         cov      :extended;             {Covariance of trait pairs}
         corr     :extended;             {Correlation of trait pairs}
         rho      :extended;             {Spearman rank correlation}
         rma      :extended;             {Reduced major axis slope}
         rmay     :extended;             {Reduced major y-intercept}
         ma       :extended;             {Major axis slope}
         may      :extended;             {Major axis y-intercept}
         ls       :extended;             {Least squares slope}
         lsy      :extended;             {Least squares y-intercept}
         r2       :extended;             {Least squares r-squared}
         ts       :extended;             {t-statistic of regression}
         f        :extended;             {F-statistic of regression}
         sumResid :extended;             {Sum of regression residuals}
         number   :integer;              {Number of nodes in tree}
         df       :integer;              {degrees of freedom for t statistic}
         conYPos  :integer;              {Number of trait 2 values>0}
         conY0    :integer;              {Number of trait 2 values=0}
         conYNeg  :integer               {Number of trait 2 values<0}
      End; {record}

      nodeDataPtr = ^nodeData;

      nodedata = Record                  {Simulation sets are load into lists}
         name     :nameString;           {The name of the node}
         data     :traits;               {The simulated data for that node}
         ansName  :nameString;
         ansData  :traits;
         rank     :traits;
         mark     :boolean;              {used in ranking process}
         next     :nodeDataPtr           {Next element in the list}
      End; {record}

      corrArray = array[1..11] of extended; {This is shared by main and i/o }

      tipArray = array[1..maxTips] of extended; {For sorting and normality checks}



Procedure initNodeData(thedata :nodeDataPtr);

Procedure killData(Var theData :nodeDataPtr);

Procedure finishStat(Var stpk :statPak);

Procedure finishStatorg(Var stpk :statPak);

Procedure tipStat(data :nodeDataPtr; Var stpk :statPak);

Procedure KolSmir(data :nodeDataPtr; Var stpk :statPak);

Procedure preBCStats(data :nodeDataPtr; Var stpk :statPak);

Procedure loadTipArray(data :nodeDataPtr;
                       traitnum :integer;
                       Var ar :tipArray;
                       Var n :integer);

Procedure arraySK(data :tipArray;
                  size :integer;
                  traitnum :integer;
                  Var stats :statPak);

Procedure boxCox(data :nodeDataPtr;
                 Var bestLambda1 :extended;
                 Var bestLambda2 :extended;
                 Var bestData1 :tipArray;
                 Var bestData2 :tipArray);

Procedure ksnorm(arrayData :tipArray;
                 n :integer;
                 traitnum :integer;
                 Var stat :statpak);


Implementation


{***********************************************************************}
{This initializes some of the fields of the element holding simulated
 data from a single tree node.  It should be called whenever one of these
 elements is allocated.  It calls nothing.}
{***********************************************************************}
Procedure initNodeData(theData :nodeDataPtr);
   Begin
      thedata^.next := nil;
      thedata^.data[1] := 0.0;
      thedata^.data[2] := 0.0;
      thedata^.name := '  '
   End; {procedure initnodedata}

{***********************************************************************}
{This recursive procedure deallocates the Simulated data list.
 This is done to keep the heap neat.  It calls nothing.}
{***********************************************************************}
Procedure killData(Var theData :nodeDataPtr);
   Var temp :nodeDataPtr;
   Begin
      While (theData <> nil) Do Begin
         temp := theData^.next;
         dispose(theData);
         theData := temp
      End {while}
   End; {Procedure KillData}

{*************************************************************************}
{This calculates the rma and the y intercept for rma called rmay, ma and
 the y intercept for the ma called may, r squared called r2, t statistic
 called ts,the f statistic, and the degrees of freedom called df.}
{*************************************************************************}
  Procedure finishStat(Var stpk :statPak);
     Begin
        If stpk.variance[1]>0 Then
           stpk.rma := sqrt(stpk.variance[2]/stpk.variance[1])
        Else stpk.rma := 0;
        If ((stpk.corr>0) and (stpk.rma<0)) or
            ((stpk.corr<0) and (stpk.rma>0)) Then
           stpk.rma := stpk.rma*-1;
        stpk.rmay := stpk.mean[2]-stpk.rma*stpk.mean[1];

        stpk.ma := sqr(stpk.variance[1]+stpk.variance[2]) -
            4*(stpk.variance[2]*stpk.variance[1] - sqr(stpk.cov));
        stpk.ma := 0.5*(stpk.variance[2]+stpk.variance[1]+
            sqrt(stpk.ma));
        If (stpk.ma-stpk.variance[2]<>0) Then
          stpk.ma := stpk.cov/(stpk.ma-stpk.variance[2])
        Else stpk.ma := 0;
        stpk.may := stpk.mean[2]-stpk.ma*stpk.mean[1];

        stpk.r2 := sqr(stpk.corr);
        If (stpk.r2 = 1.0) Then Begin
           stpk.ts := 9.9999E99;
           stpk.f := 9.9999E99;
        End {then}
        Else Begin
             stpk.ts := stpk.corr*sqrt((stpk.number-2)/(1-stpk.r2));
             stpk.f := sqr(stpk.ts);
        End; {else}
        stpk.df := stpk.number-2
     End; {procedure finishstat}

{*************************************************************************}
{This is similar to the procedure above, except the rma and ma are forced
 to go through the origin.
 This calculates the rma and the y intercept for rma called rmay, ma and
 the y intercept for the ma called may, r squared called r2, t statistic
 called ts, f statistic, and the degrees of freedom called df.}
{*************************************************************************}
  Procedure finishStatOrg(Var stpk :statPak);
    Begin
       stpk.rma := stpk.ls/stpk.corr;
       If ((stpk.corr>0) and (stpk.rma<0)) or
           ((stpk.corr<0) and (stpk.rma>0)) Then
         stpk.rma:=stpk.rma*-1;
       stpk.rmay := 0.0;

       stpk.ma := sqr(stpk.variance[1]+stpk.variance[2]) -
           4*(stpk.variance[2]*stpk.variance[1] - sqr(stpk.cov));
       stpk.ma := 0.5*(stpk.variance[2]+stpk.variance[1]+
           sqrt(stpk.ma));
       stpk.ma := stpk.cov/(stpk.ma-stpk.variance[2]);
       stpk.may := 0.0;

       stpk.r2 := sqr(stpk.corr);

       stpk.ts := stpk.corr*sqrt((stpk.number-1)/(1-stpk.r2));
       stpk.f := sqr(stpk.ts);
       stpk.df := stpk.number-2
    End; {procedure finishstatorg}

{*******************************************************************}
{This procedure calculates the statistics for the tip data.  This
 procedure is laid out just like the ones in T_Felsen.

 getbounds determines the max and min.

 getmean summs up each variable and counts them.  From this one can
   easily deterime the mean.

 getvar summs up the variance and covaraince deviations fromwhich the
   varaiance and covariance can easily be determined.

 getsumms gets the summs needed for the least squares regression.
   Here regressions are not done through the origin.}
{*******************************************************************}
Procedure tipStat(data : nodeDataPtr; Var stpk :statPak);
  Var
    sy,sx,sx2,sxy,d :extended;
    dum :integer;
  {*******************************************************************}
  Procedure getBounds(data : nodeDataPtr; Var min,max :traits);
    Var curNode : nodeDataPtr;
    Begin
      curNode := data;
      While(curNode <> nil) Do Begin
         If (curNode^.data[1]>max[1]) Then max[1] := curNode^.data[1];
         If (curNode^.data[1]<min[1]) Then min[1] := curNode^.data[1];
         If (curNode^.data[2]>max[2]) Then max[2] := curNode^.data[2];
         If (curNode^.data[2]<min[2]) Then min[2] := curNode^.data[2];
         curNode := curNode^.next;
      End; {while}
    End;{procedure getbounds}
  {*******************************************************************}
  Procedure getMean(data :nodeDataPtr;
                    Var n :integer;
                    Var m :traits;
                    Var s :traits);
    Var curNode : nodeDataPtr;
    Begin
      curNode := data;
      While (curNode <> nil) Do Begin
         inc(n);
         m[1] := m[1]+curNode^.data[1];
         m[2] := m[2]+curNode^.data[2];
         s[1] := s[1]+curNode^.rank[1];
         s[2] := s[2]+curNode^.rank[2];
         curNode := curNode^.next;
      End; {while}
    End; {procedure getmean}
  {*******************************************************************}
  Procedure getVar(data :nodeDataPtr; Var s :statPak);
    Var curNode :nodeDataPtr;
    Begin
      curNode := data;
      While (curNode <> nil) Do Begin
         s.variance[1] := s.variance[1]+sqr(s.mean[1]-curNode^.data[1]);
         s.variance[2] := s.variance[2]+sqr(s.mean[2]-curNode^.data[2]);
         s.rankVar[1] := s.rankVar[1]+sqr(s.rankMean[1]-curNode^.rank[1]);
         s.rankVar[2] := s.rankVar[2]+sqr(s.rankMean[2]-curNode^.rank[2]);
         s.corr := s.corr +
                    (s.mean[1]-curNode^.data[1]) *
                    (s.mean[2]-curNode^.data[2]);
         s.rho := s.rho +
                    (s.rankMean[1]-curNode^.rank[1]) *
                    (s.rankMean[2]-curNode^.rank[2]);
         curNode := curNode^.next;
      End; {while}
    End; {procedure getvar}
  {*******************************************************************}
  Procedure getSumms(data :nodeDataPtr; Var sx,sy,sx2,sxy :extended);
    Var curNode : nodeDataPtr;
    Begin
      curNode := data;
      While (curNode <> nil) Do Begin
        sx := sx + curNode^.data[1];
        sx2 := sx2 + sqr(curNode^.data[1]);
        sy := sy + curNode^.data[2];
        sxy := sxy + curNode^.data[1]*curNode^.data[2];
        curNode := curNode^.next;
      End; {while}
    End; {procedure getsumms}   
  {*******************************************************************}
  Begin {procedure tipstat}
    stpk.max[1] := data^.data[1];
    stpk.min[1] := stpk.max[1];
    stpk.max[2] := data^.data[2];
    stpk.min[2] := stpk.max[2];
    getBounds(data,stpk.min,stpk.max);
    stpk.number := 0;
    stpk.mean[1] := 0;
    stpk.mean[2] := 0;
    stpk.rankMean[1] := 0;
    stpk.rankMean[2] := 0;
    getMean(data,stpk.number,stpk.mean,stpk.rankMean);
    stpk.mean[1] := stpk.mean[1]/stpk.number;
    stpk.mean[2] := stpk.mean[2]/stpk.number;
    stpk.rankMean[1] := stpk.rankMean[1]/stpk.number;
    stpk.rankMean[2] := stpk.rankMean[2]/stpk.number;
    stpk.variance[1] := 0;
    stpk.variance[2] := 0;
    stpk.rankVar[1] := 0;
    stpk.rankVar[2] := 0;
    stpk.corr := 0;
    stpk.rho := 0;
    getVar(data,stpk);
    stpk.variance[1] := stpk.variance[1]/(stpk.number-1);
    stpk.variance[2] := stpk.variance[2]/(stpk.number-1);
    stpk.rankVar[1] := stpk.rankVar[1]/(stpk.number-1);
    stpk.rankVar[2] := stpk.rankVar[2]/(stpk.number-1);
    stpk.corr := stpk.corr/(stpk.number-1);
    stpk.cov := stpk.corr;
    If ((stpk.variance[1]>0) And (stpk.variance[2]>0)) Then
       stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
    Else stpk.corr := 0;
    stpk.rho := stpk.rho/(stpk.number-1);
    If ((stpk.rankVar[1]>0) And (stpk.rankVar[2]>0)) Then
       stpk.rho:=stpk.rho/sqrt(stpk.rankVar[1]*stpk.rankVar[2]);
    sx := 0;
    sy := 0;
    sx2 := 0;
    sxy := 0;
    getsumms(data,sx,sy,sx2,sxy);
    If (stpk.number*sx2 = sqr(sx))
    Then d := 0
    Else d := 1/(stpk.number*sx2-sqr(sx));
    stpk.ls := d*(stpk.number*sxy-sx*sy);
    stpk.lsy := d*(sx2*sy-sx*sxy);
    finishStat(stpk)
  End; {procedure tipstat}

{*****************************************************************}
{This procedure computes Kolmogorov-Smirnov normality statistics
 for tip values in the data list and returns the results in stpk. }
{*****************************************************************}
Procedure KolSmir(data :nodeDataPtr; Var stpk :statPak);
   Var tips1,tips2 :tipArray;
       n :integer;
   Begin
      loadTipArray(data,1,tips1,n);
      ksNorm(tips1,n,1,stpk);
      loadTipArray(data,2,tips2,n);
      ksNorm(tips2,n,2,stpk)
   End; {procedure getks}


{*******************************************************************}
{This procedure calculates the extended statistics for the tip data.
  This procedure is laid out just like the ones in T_Felsen.
{*******************************************************************}
Procedure preBCStats(data :nodeDataPtr; Var stpk :statPak);
   Var tips1,tips2 :tipArray;
       n :integer;
  Begin
    loadTipArray(data,1,tips1,n);
    stpk.number := n;
    arraySK(tips1,n,1,stpk);
    loadTipArray(data,2,tips2,n);
    arraySK(tips2,n,2,stpk);
  End; {procedure preBCStats}


{********************************************************************}
{Quick and dirty procedure for loading the array for the Box-Cox and
 K-S tests.                                                         }
{********************************************************************}
Procedure loadTipArray(data :nodeDataPtr;
                       traitNum :integer;
                       Var ar :TipArray;
                       Var n :integer);
   Var current : nodeDataPtr;
   Begin
      n := 0;
      current := data;
      While (current<>nil) Do Begin
         n := n+1;
         ar[n] := current^.data[traitNum];
         current := current^.next
      End {while begin}
   End; {procedure loadtiparray}


{********************************************************************}
{Computes Skewness and Kurtosis for an array of data.               }
{********************************************************************}
Procedure arraySK(data :tipArray;
                  size :integer;
                  traitnum :integer;
                  Var stats :statPak);
   Var i :integer;
       s :extended;
       p :extended;
       mean :extended;
       svarn :extended;
       svar :extended;
       skew :extended;
       kurt :extended;
   Begin
      mean := 0.0;
      svar := 0.0;
      skew := 0.0;
      kurt := 0.0;
      For i := 1 To size Do mean := mean + data[i];
      mean := mean/size;
      For i := 1 to size Do Begin
         s := data[i] - mean;
         p := s*s;
         svar := svar+p;
         p := p*s;
         skew := skew+p;
         p := p*s;
         kurt := kurt+p;
      End; {for}
      svar := svar/(size-1);
      svarn := ((size-1)/size)*svar;
      stats.mean[traitnum] := mean;
      stats.variance[traitnum] := svar;
      If (svarn=0) Then Begin  {if there are too few bins, this CAN happen}
         stats.skewness[traitnum] := 0;
         stats.kurtosis[traitnum] := 0
      End {then}
      Else Begin
         stats.skewness[traitnum] := skew/(size*svarn*sqrt(svarn));
         stats.kurtosis[traitnum] := kurt/(size*sqr(svarn))-3.0
      End {else}
   End; {procedure arraySK}
 

{********************************************************************}
{Finds best transform using Box-Cox statistic and stepping through
 values of lambda from -2 to 2 with a step size of 0.1.
{********************************************************************}
Procedure boxCox(data :nodeDataPtr;
                 Var bestLambda1 :extended;
                 Var bestLambda2 :extended;
                 Var bestData1 :tipArray;
                 Var bestData2 :tipArray);
    Var size :integer;
        j :integer;
        tipData :tipArray;
        xFormData :tipArray;
        lnSum :extended;
        tipVar :extended;
        lambda10 :integer;
        lambda :extended;
        lScore :extended;
        bestLScore :extended;
        lambda1 :extended;
        lambda2 :extended;
        found0 :boolean;
    {********************************************************************}
    {Computes the variance of the data in the tipArray.}
    {********************************************************************}
    Function tipVariance(size :integer; data :tipArray) :extended;
       Var i :integer;
           mean :extended;
           summ :extended;
       Begin
          mean := 0;
          summ := 0;
          For i := 1 To size Do mean := mean + data[i];
          mean := mean/size;
          For i := 1 To size Do summ := summ + sqr(data[i] - mean);
          tipVariance := summ/(size-1)
       End; {function tipvariance}
    {********************************************************************}
    {Applies the Box-Cox transform with parameter lambda to the source
     array yielding values stored in the destination array.}
    {********************************************************************}
    Procedure xFormTipArray(lambda :extended;
                            size :integer;
                            source :tipArray;
                            Var destination :tipArray);
       Var i :integer;
       Begin
          If (lambda=0.0) Then
             For i := 1 To size Do destination[i] := ln(source[i])
          Else For i := 1 To size Do
               destination[i] := (exp(lambda*ln(source[i]))-1)/lambda
       End; {procedure xformtiparray}
   {********************************************************************}
    Begin {procedure boxcox}
       loadTipArray(data,1,tipData,size);
       found0 := false;
       For j := 1 To size Do If (tipData[j]<=0) Then found0 := true;
       tipVar := tipVariance(size,tipData);
       If (found0 Or (tipVar = 0))
       Then Begin  {if any data < 0, we lose}
          bestLambda1 := -999.0;
          loadTipArray(data,1,bestData1,size)  {copy starting data so .SIT file works}
       End {then}  
       Else Begin
          xFormTipArray(-3.0,size,tipData,xFormData);
          lnSum := 0.0;
          tipVar := tipVariance(size,xFormData);
          For j := 1 To size Do lnSum := lnSum + ln(tipData[j]);
          bestLScore := ((1-size)/2)*ln(tipVar) + (-3.0-1.0)*((size-1)/size)*lnSum;
          bestLambda1 := -3.0;
          For lambda10 := -29 To 30 Do Begin
             lambda := lambda10/10.0;
             xFormTipArray(lambda,size,tipData,xFormData);
             lnSum := 0.0;
             tipVar := tipVariance(size,xFormData);
             For j := 1 To size Do lnSum := lnSum + ln(tipData[j]);
             lScore := ((1-size)/2)*ln(tipVar) + (lambda-1.0)*((size-1)/size)*lnSum;
             If (lScore > bestLScore) Then Begin
                bestLScore := lScore;
                bestLambda1 := lambda
             End {then}
          End; {for}
          xFormTipArray(bestLambda1,size,tipData,bestData1)
       End; {else}
       loadTipArray(data,2,tipData,size);
       found0 := false;
       For j := 1 To size Do If tipData[j] <= 0 Then found0 := true;
       tipVar := tipVariance(size,tipData);
       If (found0 Or (tipVar = 0))
       Then Begin  {if any data < 0, we lose}
          bestLambda2 := -999.0;
          loadTipArray(data,2,bestData2,size)  {copy starting data so .SIT file works}
       End {then}  
       Else Begin
          xFormTipArray(-3.0,size,tipData,xFormData);
          lnSum := 0.0;
          tipVar := tipVariance(size,xFormData);
          For j := 1 To size Do lnSum := lnSum + ln(tipData[j]);
          bestLScore := ((1-size)/2)*ln(tipVar) + (-3.0-1.0)*((size-1)/size)*lnSum;
          bestLambda2 := -3.0;
          For lambda10 := -29 To 30 Do Begin
             lambda := lambda10/10.0;
             xFormTipArray(lambda,size,tipData,xFormData);
             lnSum := 0.0;
             tipVar := tipVariance(size,xFormData);
             For j := 1 To size Do lnSum := lnSum + ln(tipData[j]);
             lScore := ((1-size)/2)*ln(tipVar) + (lambda-1.0)*((size-1)/size)*lnSum;
             If (lScore > bestLScore) Then Begin
                bestLScore := lScore;
                bestLambda2 := lambda
             End {then}
          End; {for}
          xFormTipArray(bestLambda2,size,tipData,bestData2)
       End {else}
    End; {procedure boxcox}


{********************************************************************}
{Sort procedure for use in Kolmogorov-Smirnov test; sorts the array
 into ascending order using heapsort algorithm.  From Numerical
 Receipes in Pascal p. 259.}
{********************************************************************}
Procedure sort(n :integer;
               Var ra :tipArray);
   Label 99;
   Var i,l,j,ir :integer;
       rra :extended;
   Begin
      l := (n div 2)+1;
      ir := n;
      While true Do Begin
         If (l > 1) Then Begin
            l := l-1;
            rra := ra[l]
         End {while}
         Else Begin
            rra := ra[ir];
            ra[ir] := ra[1];
            ir := ir-1;
            If ir =1 Then Begin
               ra[1] := rra;
               Goto 99
            End {then}
         End; {else}
         i := l;
         j := l+l;
         While (j<=ir) Do Begin
            If (j<ir) Then
               If (ra[j]<ra[j+1]) Then j := j+1;
            If (rra<ra[j]) Then Begin
               ra[i] := ra[j];
               i := j;
               j := j+j
            End {then}
            Else j := ir+1
         End;
         ra[i] := rra
      End; {while}
      99:
   End; {procedure sort}

{********************************************************************}
{Support function for ksNorm computes the p-value corresponding to
 the ks parameter alam.}
{********************************************************************}
Function probKS(alam :extended): extended;
   Label 99;
   Const
      eps1 = 0.001;
      eps2 = 1.0e-8;
   Var
      a2,fac,sum,term,termbf :extended;
      j :integer;
   Begin
      a2 := -2.0*alam*alam;
      fac := 2.0;
      sum := 0.0;
      termbf := 0.0;
      For j := 1 To 100 Do Begin
         term := fac*exp(a2*sqr(j));
         sum := sum+term;
         If (abs(term) <= eps1*termbf) Or (abs(term) <= eps2*sum) Then Begin
            probks := sum;
            Goto 99
         End {then begin}
         Else Begin
            fac := -fac;
            termbf := abs(term)
         End {else}
      End; {for}
      probks := 1.0;
      99:
   End; {function probks}

{********************************************************************}
{ This function computes the cumulative z (Normal) distribution      }
{********************************************************************}
Function zDistribution(x :extended;
                       mean :extended;
                       variance :extended) :extended;
   Var t :extended;
   {********************************************************************}
   { Returns the complementary error function erfc(x) with fractional}
   { error everywhere less than 1.2E-7 (Numerical recipes p. 185) }
   {********************************************************************}
   Function erfcc(x :extended) :extended;
      Var t,z,ans :extended;
      Begin
         z := abs(x);
         t := 1.0/(1.0+0.5*z);
         ans := t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+
                 t*(0.09678418 +t*(-0.18628806+t*(0.27886807+
                 t*(-1.13520398+t*(1.48851587+t*(-0.82215223+
                 t*0.17087277)))))))));
         If (x>=0.0) Then erfcc := ans Else erfcc := 2.0-ans
      End; {function erfcc}
   {********************************************************************}
   { Converts erfcc to normal density distribution }
   {********************************************************************}
   Begin {function zdistribution}
      If (variance = 0)
      Then zdistribution := 0  {dubious, but non-crashing answer}
      Else Begin
         t := (x-mean)/(sqrt(variance*2)); {Not sure why sqrt(2) is required}
         zdistribution := 1-erfcc(t)/2
      End {else}
   End; {function zdistribution}

{********************************************************************}
{ Given the data array data1[1..n], this procedure will compute the
  KS statistic d, and return the significance level prob for the null
  hypothesis that set data1 is drawn from the normal distribution.
  The p-value is returned in the ks field of the statPak.
  Small values of prob show that the cumulative distribution function
  of data1 is significantly different from normal.  Data1 is modified
  by being sorted into ascending order.  For computing the normal
  density, the mean and variance are taken from statpak.  This
  routine is modified from pp. 521 of Numerical Receipes in
  Pascal.  Note that since the normal distribution has effectively
  infinite degrees of freedom, the p-value doesn't need to be adjusted
  for d.f..
{********************************************************************}
Procedure ksNorm(arrayData :tipArray;
                 n :integer;
                 traitnum :integer;
                 Var stat :statPak);
   Var i,j :integer;
       fo,fn,ff,en,d,dt :extended;
       datacopy :tipArray;
   Begin
      If (stat.variance[traitnum] = 0)
      Then stat.ks[traitnum] := 1.0
      Else Begin
         For j := 1 To n Do dataCopy[j] := arrayData[j];
         sort(n,dataCopy);
         en := n;
         d := 0.0;
         fo := 0.0;
         For j := 1 To n Do Begin
            fn := j/en;
            ff := zdistribution(dataCopy[j],stat.mean[traitnum],stat.variance[traitnum]);
            If (abs(fo-ff)>abs(fn-ff)) Then dt := abs(fo-ff)
            Else dt := abs(fn-ff);
            If (dt>d) Then d := dt;
            fo := fn
         End; {for}
         stat.ks[traitnum] := probks(sqrt(en)*d) {Compute significance}
      End {else}
   End; {procedure ksnorm}


End.
