{                              T_FELS_D.PAS                              }
{************************************************************************}
{This module displays 8 different X-Y diagnostic plots. Each has separate
procedures to calculate statistics and display the screen.  All are
similar, but need to be slightly different because the screens plot
contrasts against various other characteristics of the node they are
associated with, such as height, length of the daughter branches, etc.
The statistics procedure is called by the corresponding plot procedure
which then uses procedures from t_util1 and te_util1 to do the plotting.
The statistics procedures get called on their own when the user requests
the corresponding .FIC file which contains all the statistics from this
module as well as those produced by T_FELS_R.

Maybe one procedure with functional/procedural arguments for the
statistics would work (and save space), it could get really complicated.

The various diagnostic screens  are discussed in:

Garland, T., Jr., P. H. Harvey, and A. R. Ives. 1992. Procedures for the
     analysis of comparative data using phylogenetically independent
     contrasts. Systematic Biology 41:18-32.
Diaz-Uriarte, R., and T. Garland, Jr. 1996. Testing hypotheses of
     correlated evolution using phylogenetically independent contrasts:
     sensitivity to deviations from Brownian motion.
     Systematic Biology 45:27-47.

}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}


Unit t_fels_d;

Interface

uses t_util1;

Procedure computecon(nd:nodeptr);

Procedure statStdVsD(nd:nodePtr;t:integer;Var stpk:statPak);

Procedure plotStdVsD(rt:nodePtr;
                     t,n:integer;
                     fl:flString;
                     spNode : nodePtr;
                     clade : boolean;
                     Var bail:boolean);

Procedure statNodeCon(nd:nodePtr;t:integer;Var stpk:statPak);

Procedure plotNodeCon(rt:nodePtr;
                      t,n:integer;
                      fl:flString;
                      spNode : nodePtr;
                      clade : boolean;
                      Var bail:boolean);

Procedure statheightnode(nd:nodeptr;t:integer;var stpk:statpak);

Procedure statHeightCon(nd:nodePtr;t:integer;Var stpk:statPak);

Procedure plotHeightNode(rt:nodePtr;    
                         t,n:integer;
                         fl:flString;
                         spNode :nodePtr;
                         clade : boolean;
                         Var bail:boolean);

Procedure plotheightcon(rt:nodeptr;
                        t,n:integer;
                        fl:flstring;
                        spNode : nodePtr;
                        clade : boolean;
                        var bail:boolean);

Procedure showDiagnostics(nd : nodePtr; fln : flString);

Implementation

Uses crt,
     graph,
     te_util1;

  
{***************************************************************************}
{This procedure (formerly dumpText1) now prompts for the option of ploting
 a node or an entire clade in a different character.}
{***************************************************************************}
Procedure plotOptions(rt : nodePtr;
                      Var spNode : nodePtr;
                      Var wholeClade : boolean);
   Var
     xPos, yPos : integer;
     userChar : char;
     ndName : nameString;
     noGood : boolean;
   Begin
      spNode := nil;
      wholeClade := false;
      clrscr;
      gotoxy(78,25);
      write('0');
      gotoxy(1,1);
      writeln('    Screens 1-8 are diagnostics, helpful for checking whether your ');
      writeln('branch lengths are doing an adequate job, i.e., properly standardizing');
      writeln('the contrasts.');
      writeln('    In general, you hope to see no trends in any of these plots (e.g., see');
      writeln('Garland, Harvey, and Ives, 1992, Systematic Biology 31:18-32).  You should');
      writeln('also check for possible trends within subclades of you phylogeny');
      writeln('(e.g., Garland, 1992, American Naturalist 140:509-519).');
      writeln('    Remember that you may need different branch length transformations for');
      writeln('each trait in an analysis.  This program, however, only allows you to use');
      writeln('and view a single set of branch lengths at a time, the same for both traits.');
      writeln;
      writeln('By default, the root node contrast will be displayed as an ''X'', all other');
      writeln('contrasts will be displayed as closed circles.');
      writeln;
      writeln('Would you like to have a single contrast displayed with an open circle?');
      write('(Y/N/Any other character = N): ');
      xPos := whereX;
      yPos := whereY;
      writeln;
      writeln('(This feature is useful for testing a priori hypotheses about particular');
      writeln('contrasts.)');
      gotoxy(xPos,yPos);
      userChar := readKey;
      gotoxy(1,yPos+3);
      If ((userChar = 'Y') or (userChar = 'y')) Then Begin
         wholeClade := false;
         write('What is the two character name of the node (case matters)? ');
         readln(ndName);
         nogood := true;
         While (noGood) Do Begin
            checkName(rt,ndName,noGood,spNode);
            If (noGood = false) Then Begin
               If (spNode^.sibL = NIL) Then Begin
                  writeln('The node ',ndName,' is a tip, not a contrast.');
                  noGood := true;
               End {then}
            End
            Else
               writeln('No node with the name ',ndName,' could be found in the tree.');
            If (noGood)
            Then Begin
               write('Enter a new name, or carriage return to give up: ');
               readln(ndName);
               If (ndName = '')  Then Begin
                  noGood := false;
                  spNode := NIL;
               End; {then}
            End; {then}
         End; {while}
      End {then}
      Else Begin
         wholeClade := true;
         writeln;
         writeln('Would you like to see all the contrasts for a clade displayed with an open');
         write('circle? (Y/N/Any other character = N): ');
         userChar := readKey;
         writeln;
         If ((userChar = 'Y') or (userChar = 'y')) Then Begin
            writeln('What is the two character name of the clade''s basal ');
            write('(root) node (case matters)? ');
            readln(ndName);
            nogood := true;
            While (noGood) Do Begin
               checkName(rt,ndName,noGood,spNode);
               If (noGood = false) Then Begin
                  If (spNode^.sibL = NIL) Then Begin
                     writeln('The node ',ndName,' is a tip, so it is not the base of a clade.');
                     noGood := true;
                  End {then}
               End
               Else
                  writeln('No node with the name ',ndName,' could be found in the tree.');
               If (noGood) Then Begin
                  write('Enter a new name, or carriage return to give up: ');
                  readln(ndName);
                  If (ndName = '')  Then Begin
                     noGood := false;
                     spNode := NIL;
                     wholeClade := false;
                  End; {then}
               End; {then}
            End; {while}
         End; {then}
      End; {else}
      writeln;                          { might prevent screen overwrites }
      ret('Press any key to continue.'); 
   End; {procedure dumptext1}


{**************************************************************************}
{This procedure computes the standardized contrasts for each internal node.}
{It also computes the predicted trait values for those nodes by            }
{Felsenstein's method.  It stores these values in the tree.                }
{**************************************************************************}
Procedure computecon(nd:nodeptr);
   var
      v1,v2 :extended;
   Begin
      If nd^.sibL=nil
      Then nd^.variance:=nd^.length
      Else Begin
         computecon(nd^.sibL);
         computecon(nd^.sibR);
         v1:=nd^.sibL^.variance;
         v2:=nd^.sibR^.variance;
         nd^.variance:=nd^.length+(v1*v2/(v1+v2));
         nd^.sdcon:=sqrt(v1+v2);
         nd^.contrast[1]:=(nd^.sibL^.dat[1]-nd^.sibR^.dat[1])/nd^.sdcon;
         nd^.contrast[2]:=(nd^.sibL^.dat[2]-nd^.sibR^.dat[2])/nd^.sdcon;
         nd^.dat[1]:=(v1*nd^.sibR^.dat[1] + v2*nd^.sibL^.dat[1])/(v1+v2);
         nd^.dat[2]:=(v1*nd^.sibR^.dat[2] + v2*nd^.sibL^.dat[2])/(v1+v2);
      End; {else}
   End; {procedure computecon}

{**************************************************************************}
{This procedure is intended to prevent calling computecon with a tree that }
{will cause it to blowing up.  Unfortunately, it doesn't work.             }
{**************************************************************************}
{Procedure contrastCheck(nd:nodeptr;var ok:boolean);
   Var
      v1,v2 :extended;
      ok1,ok2 : boolean;
      tmp : extended;
   Begin
      If (nd^.sibL = nil)
      Then Begin
         nd^.variance := nd^.length;
         ok := true;
      End
      Else Begin
         contrastCheck(nd^.sibL,ok1);
         contrastCheck(nd^.sibR,ok2);
         If (ok1 and ok2) Then Begin
            v1 := nd^.sibL^.variance;
            v2 := nd^.sibR^.variance;
            tmp := v1 + v2;
            write(nd^.name,' ',tmp,' ');
            If (tmp = 0)
            Then ok := false
            Else ok := true;
         End
         Else ok := false;
      End;
   End;} {procedure contrastCheck}

{***************************************************************************}
{This procedure computes the statistics comparing the height of a node      }
{against one of the traits.  The trait is specified by the variable t.      }
{***************************************************************************}
Procedure statheightnode(nd:nodeptr;t:integer;var stpk:statpak);
   Var
      sy,sx,sx2,sxy,d:extended;
   {**********************************************************************}
   {This recursive procedure gets the max and min values for the data.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.  Note that none of
   the local procedures look at tip values.  Getbounds used to, but was
   changed 5 Dec 1999, PEM.}
   {**********************************************************************}
   Procedure getbounds(var min,max:traits; nd:nodePtr;t:integer);
      Begin
         If nd^.sibL<>nil Then Begin
            If nd^.temp3>max[1] Then max[1]:=nd^.temp3;
            If nd^.temp3<min[1] Then min[1]:=nd^.temp3;
            If nd^.dat[t]>max[2] Then max[2]:=nd^.dat[t];
            If nd^.dat[t]<min[2] Then min[2]:=nd^.dat[t];
            getbounds(min,max,nd^.sibL,t);
            getbounds(min,max,nd^.sibR,t);
         End; {then}
      End; {procedure getbounds}      
   {************************************************************************}
   {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getMean(t:integer;nd:nodePtr;var n:integer;var m: traits);
      Begin
         If nd^.sibL<>nil Then Begin
            inc(n);
            m[1]:=m[1]+nd^.temp3;
            m[2]:=m[2]+nd^.dat[t];
            getMean(t,nd^.sibL,n,m);
            getMean(t,nd^.sibR,n,m);
         End; {then}
      End; {procedure getmean}
   {************************************************************************}
   {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getVar(t:integer;nd:nodePtr;var s:statPak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(s.mean[1]-nd^.temp3);
            s.variance[2]:=s.variance[2]+sqr(s.mean[2]-nd^.dat[t]);
            s.corr:=s.corr+(s.mean[1]-nd^.temp3)*(s.mean[2]-nd^.dat[t]);
            getVar(t,nd^.sibL,s);
            getVar(t,nd^.sibR,s)
         End {then}
      End; {procedure getvar}
   {************************************************************************}
   {This recursive procedure does the first step in finding the least squares
   regression for the data.  It finds the sums of squares and products.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getSums(t:integer;var sx,sy,sx2,sxy:extended; nd:nodePtr);
      Begin
         If nd^.sibL<>nil Then Begin  
            sx:=sx+nd^.temp3;
            sx2:=sx2+sqr(nd^.temp3);
            sy:=sy+nd^.dat[t];
            sxy:=sxy+nd^.temp3*nd^.dat[t];
            getSums(t,sx,sy,sx2,sxy,nd^.sibL);
            getSums(t,sx,sy,sx2,sxy,nd^.sibR)
         End {then}
      End; {procedure getSums}
   {**************************************************************************}
   Begin  {statheightnode}
      stpk.n1:=1;
      stpk.n2:=2;
      stpk.max[1]:=nd^.temp3;
      stpk.max[2]:=nd^.dat[t];
      stpk.min[1]:=stpk.max[1];
      stpk.min[2]:=stpk.max[2];
      getBounds(stpk.min,stpk.max,nd,t);  
      stpk.number:=0;
      stpk.mean[1]:=0;
      stpk.mean[2]:=0;
      getMean(t,nd,stpk.number,stpk.mean);
      stpk.mean[1]:=stpk.mean[1]/stpk.number;
      stpk.mean[2]:=stpk.mean[2]/stpk.number;
      stpk.variance[1]:=0;
      stpk.variance[2]:=0;
      stpk.corr:=0;
      getVar(t,nd,stpk);
      stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);
      stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);
      stpk.corr:=stpk.corr/(stpk.number-1);
      stpk.cov:=stpk.corr;
      If(stpk.variance[1]>0) and (stpk.variance[2]>0)
      Then
         stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr:=0;
      sx:=0;
      sy:=0;
      sx2:=0;
      sxy:=0;
      getSums(t,sx,sy,sx2,sxy,nd);
      If ((stpk.number*sx2-sqr(sx))<>0) Then Begin
         stpk.lsregok:=true;
      d:=1/(stpk.number*sx2-sqr(sx));
      End {then}
      Else Begin
         stpk.lsregok:=false;
         d:=0;
      End; {else}
      stpk.ls:=d*(stpk.number*sxy-sx*sy);
      stpk.lsy:=d*(sx2*sy-sx*sxy);
      finishStat(stpk)
   End; {procedure statheightnode}

{****************************************************************************}
{This procedure sets up the graphics screen and plots out the data points
appropriate for statheightnode.  In addition to using statheightnode, it
also calls several procedures from the T_UTIL1 unit.}
{****************************************************************************}
Procedure plotHeightNode(rt:nodePtr;    { this really is the root }
                         t,n:integer;
                         fl:flString;
                         spNode :nodePtr;
                         clade : boolean;
                         Var bail:boolean);
   Var
      scalingFactor:traits;
      mx,my,temp:integer;
      stpk:statpak; 
      origin:itraits;
      s:string[10];
      st:string[80];
      coor:itraits;
   {*************************************************************************}
   {This recursive procedure plots data points onto the screen.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {*************************************************************************}
   Procedure dumpData(min:traits;
                      orgn:itraits;
                      sf:traits;
                      nd:nodePtr;
                      special : boolean;
                      t:integer);
      Var
         coor:itraits;
      Begin    
         coor[1]:=orgn[1]+round(sf[1]*(nd^.temp3-min[1]));
         coor[2]:=orgn[2]-round(sf[2]*(nd^.dat[t]-min[2]));
         If (nd^.sibL <> nil) Then Begin
            If (nd = rt) Then xSymbol(coor)
            Else
               If ((nd = spNode) or (special = true)) Then openCircle(coor)
               Else star(coor);
            If ((nd = spNode) and clade) Then Begin
               dumpData(min,orgn,sf,nd^.sibL,true,t);
               dumpData(min,orgn,sf,nd^.sibR,true,t)
            End {then}
            Else Begin
               dumpData(min,orgn,sf,nd^.sibL,special,t);
               dumpData(min,orgn,sf,nd^.sibR,special,t)
            End
         End {then}
      {   Else star2(coor) }   {no longer display tips}
      End; {procedure dumpdata}
   {*************************************************************************}
   Begin {plotheightnode}
      statHeightNode(rt,t,stpk);
      graphicsOn(mx,my);
      initPlot(1,2,stpk.min,stpk.max,fl,scalingFactor,origin,mx,my);
      outTextXY(0,0,'Plot of estimated base node');
      outTextXY(0,oneline,'value versus its corrected');
      str(t,s);
      outTextXY(0,2*oneline,'height for Trait '+s+'.');

      outTextXY(0,6*oneline,'d.f. may be fewer if tree');
      outTextXY(0,7*oneline,'   contains polytomies.');

      st:='Estimated nodal value';
      temp:=round((origin[2]-length(st)*oneline)/2);
      If temp<1
      Then temp:=1;
      dumpDown(round(0.9524*origin[1])-8,temp,st);
      st:='Corrected height of base node';
      temp:=round((mx-origin[1]-textwidth(st))/2)+origin[1];
      If temp< origin[1]+1
      Then temp:=origin[1]+1;
      outTextXY(temp,round(1.05*origin[2])+2,st);
      {Commented out notion for tip points, as these are no longer displayed}
{      coor[1]:=origin[1]+8*oneline;
      coor[2]:=round(1.05*origin[2]+2+2*oneline);
      outTextXY(coor[1]+oneline,coor[2],'indicates values at tips');
      star2(coor);  }
      str(n,s);
      outTextXY(mx-2*oneline,my-oneline,s);
      dumpData(stpk.min,origin,scalingFactor,rt,false,t);
      dumpStat(stpk,mx,my,bail,false)
   End; {procedure plotheightnode}
{***************************************************************************}
{This procedure computes the statistics pertaining to node values and       }
{contrasts.                                                                 }
{***************************************************************************}
Procedure statNodeCon(nd:nodePtr;t:integer;Var stpk:statPak);
   Var
      sy,sx,sx2,sxy,d:extended;
   {************************************************************************}
   {This recursive procedure gets the max and min values for the data.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getBounds(var min,max:traits; nd:nodeptr;t:integer);
      Var
         standard:extended;
      Begin
         If nd^.sibL<>nil Then Begin
            If nd^.dat[t]>max[1] Then max[1]:=nd^.dat[t];
            If nd^.dat[t]<min[1] Then min[1]:=nd^.dat[t];
            standard:=abs(nd^.contrast[t]);
            If standard>max[2] Then max[2]:=standard;
            If standard<min[2] Then min[2]:=standard;
            getBounds(min,max,nd^.sibL,t);
            getBounds(min,max,nd^.sibR,t)
         End {then}
      End; {procedure getBounds}
   {************************************************************************}
   {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getMean(t:integer;nd:nodeptr;var n:integer;var m: traits);
      Begin
         If nd^.sibL<>nil Then Begin 
            inc(n);
            m[1]:=m[1]+nd^.dat[t];
            m[2]:=m[2]+abs(nd^.contrast[t]);
            getMean(t,nd^.sibL,n,m);
            getMean(t,nd^.sibR,n,m)
         End  {getMean}
      End; {procedure getMean}
   {************************************************************************}
   {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getVar(t:integer;nd:nodeptr;var s:statpak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(s.mean[1]-nd^.dat[t]);
            s.variance[2]:=s.variance[2]+sqr(s.mean[2]-abs(nd^.contrast[t]));
            s.corr:=s.corr+(s.mean[1]-nd^.dat[t])*(s.mean[2]-
                             abs(nd^.contrast[t]));
            getVar(t,nd^.sibL,s);
            getVar(t,nd^.sibR,s);
         End; {then}
      End; {procedure getVar}
   {************************************************************************}
   {This recursive procedure does the first step in finding the least squares
   regression for the data.  It finds the sums needed.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getSums(t:integer;var sx,sy,sx2,sxy:extended; nd:nodeptr);
      Begin
         If nd^.sibL<>nil Then Begin
            sx:=sx+nd^.dat[t];
            sx2:=sx2+sqr(nd^.dat[t]);
            sy:=sy+abs(nd^.contrast[t]);
            sxy:=sxy+nd^.dat[t]*abs(nd^.contrast[t]);
            getSums(t,sx,sy,sx2,sxy,nd^.sibL);
            getSums(t,sx,sy,sx2,sxy,nd^.sibR);
         End; {then}
      End; {procedure getSums}
   {************************************************************************}
   Begin {statNodeCon}
      stpk.n1:=1;
      stpk.n2:=2;
      stpk.max[1]:=nd^.dat[t];
      stpk.max[2]:=abs(nd^.contrast[t]);
      stpk.min[1]:=stpk.max[1];
      stpk.min[2]:=stpk.max[2];
      getBounds(stpk.min,stpk.max,nd,t);  
      stpk.number:=0;
      stpk.mean[1]:=0;
      stpk.mean[2]:=0;
      getMean(t,nd,stpk.number,stpk.mean);
      stpk.mean[1]:=stpk.mean[1]/stpk.number;
      stpk.mean[2]:=stpk.mean[2]/stpk.number;
      stpk.variance[1]:=0;
      stpk.variance[2]:=0;
      stpk.corr:=0;
      getVar(t,nd,stpk);
      stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);
      stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);
      stpk.corr:=stpk.corr/(stpk.number-1);
      stpk.cov:=stpk.corr;
      If (stpk.variance[1]>0) and (stpk.variance[2]>0)
      Then
         stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr:=0;
      sx:=0;
      sy:=0;
      sx2:=0;
      sxy:=0;
      getSums(t,sx,sy,sx2,sxy,nd);
      If ((stpk.number*sx2-sqr(sx))<>0) Then Begin
         stpk.lsregok:=true;
         d:=1/(stpk.number*sx2-sqr(sx));
      End {then}
      Else Begin
         stpk.lsregok:=false;
         d:=0;
      End; {else}
      stpk.ls:=d*(stpk.number*sxy-sx*sy);
      stpk.lsy:=d*(sx2*sy-sx*sxy);
      finishStat(stpk);
   End; {procedure statNodeCon}    
{***************************************************************************}
{This procedure displays the node value and contrast statistics calculated
by statnodecon (above).  It calls several display procedures from the
T_UTIL1 unit. (Screens 3 and 4)                                             } 
{***************************************************************************}
Procedure plotNodeCon(rt:nodePtr;
                      t,n:integer;
                      fl:flString;
                      spNode : nodePtr;
                      clade : boolean;
                      Var bail:boolean);
   Var
      fakeMin,scalingFactor:traits;
      mx,my,temp:integer;
      stpk:statpak;
      origin:itraits;
      s:string[10];
      st:string[80];
   {************************************************************************}
   {This recursive procedure plots data points onto the screen.
    Notice that it is local to a parent procedure, and therefore, can coexist
    in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure dumpData(min:traits;
                      orgn:itraits;
                      sf:traits;
                      nd:nodePtr;
                      t:integer;
                      special : boolean);
      Var
         coor:itraits;
      Begin
         If(nd^.sibL<>nil) Then Begin
            coor[1]:=orgn[1]+round(sf[1]*(nd^.dat[t]-min[1]));
            coor[2]:=orgn[2]-round(sf[2]*(ABS(nd^.contrast[t])-min[2]));
            If (nd = rt) Then xSymbol(coor)
            Else
               If ((nd = spNode) or (special = true)) Then openCircle(coor)
               Else star(coor);
            If ((nd = spNode) and clade) Then Begin
               dumpData(min,orgn,sf,nd^.sibL,t,true);
               dumpData(min,orgn,sf,nd^.sibR,t,true);
            End {then}
            Else Begin
               dumpData(min,orgn,sf,nd^.sibL,t,special);
               dumpData(min,orgn,sf,nd^.sibR,t,special);
            End;
         End; {then}
      End; {procedure dumpData}
   {**************************************************************************}
   Begin {plotNodeCon}
      statNodeCon(rt,t,stpk);
      fakeMin[1]:=stpk.min[1];
      fakeMin[2]:=0.0;
      graphicsOn(mx,my);
      initPlotX0(1,2,fakemin,stpk.max,fl,scalingFactor,origin,mx,my);
      outTextXY(0,0,'Plot of absolute value of');
      outTextXY(0,oneline,'standardized contrasts versus');
      outTextXY(0,2*oneline,'the estimated nodal values');
      str(t,s);
      outTextXY(0,3*oneline,'for Trait '+s+'.');
      outTextXY(0,6*oneline,'d.f. may be fewer if tree');
      outTextXY(0,7*oneline,'   contains polytomies.');
      st:='Absolute value of standardized contrast';
      temp:=round((origin[2]-length(st)*oneline)/2);
      If temp<1
      Then temp:=1;
      dumpDown(round(0.9524*origin[1])-8,temp,st);
      st:='Estimated value of base node';
      temp:=round((mx-origin[1]-textwidth(st))/2)+origin[1];
      If temp< origin[1]+1
      Then temp:=origin[1]+1;
      outTextXY(temp,origin[2]+3,st);
      str(n,s);
      outTextXY(mx-2*oneline,my-oneline,s);
      dumpData(fakeMin,origin,scalingFactor,rt,t,false);
      dumpStat(stpk,mx,my,bail,true)
   End; {procedure plotNodeCon}

{***************************************************************************}
{This procedure computes the statistics comparing the height of a node with
its contrast for one of the traits.  The trait is specified by the argument
t.} 
{***************************************************************************}
Procedure statHeightCon(nd:nodePtr;t:integer;Var stpk:statPak);
   Var
      sy,sx,sx2,sxy,d:extended;
   {************************************************************************}
   {This recursive procedure gets the max and min values for the data.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getBounds(Var min,max:traits; nd:nodePtr;t:integer);
      Var
         standard:extended;
      Begin
         If nd^.sibL<>nil
         Then Begin
            If nd^.temp3>max[1] Then max[1]:=nd^.temp3;
            If nd^.temp3<min[1] Then min[1]:=nd^.temp3;
            standard:=ABS(nd^.contrast[t]);
            If standard>max[2] Then max[2]:=standard;
            If standard<min[2] Then min[2]:=standard;
            getBounds(min,max,nd^.sibL,t);
            getBounds(min,max,nd^.sibR,t)
         End {then}
      End; {procedure getBounds}     
   {************************************************************************}
   {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getMean(t:integer;nd:nodePtr;Var n:integer;Var m: traits);
      Begin
         If nd^.sibL<>nil Then Begin
            inc(n);
            m[1]:=m[1]+nd^.temp3;
            m[2]:=m[2]+abs(nd^.contrast[t]);
            getMean(t,nd^.sibL,n,m);
            getMean(t,nd^.sibR,n,m)
         End {then}
      End; {procedure getMean}
   {************************************************************************}
   {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getVar(t:integer;nd:nodePtr;Var s:statPak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(s.mean[1]-nd^.temp3);
            s.variance[2]:=s.variance[2]+sqr(s.mean[2]-abs(nd^.contrast[t]));
            s.corr:=s.corr+(s.mean[1]-nd^.temp3)*(s.mean[2]-
                                          abs(nd^.contrast[t]));
            getVar(t,nd^.sibL,s);
            getVar(t,nd^.sibR,s)
         End {then}
      End; {procedure getVar}
  {************************************************************************}
  {This recursive procedure does the first step in finding the least squares
  regression for the data.  It finds the sums needed.
  Notice that it is local to a parent procedure, and therefore, can coexist
  in a file with other procedures by the same name.}
  {************************************************************************}
  Procedure getSums(t:integer;Var sx,sy,sx2,sxy:extended; nd:nodePtr);
     Begin
        If nd^.sibL<>nil Then Begin
           sx:=sx+nd^.temp3;
           sx2:=sx2+sqr(nd^.temp3);
           sy:=sy+abs(nd^.contrast[t]);
           sxy:=sxy+nd^.temp3*abs(nd^.contrast[t]);
           getSums(t,sx,sy,sx2,sxy,nd^.sibL);
           getSums(t,sx,sy,sx2,sxy,nd^.sibR)
        End {then}
     End; {procedure getsums}
  {**************************************************************************}
  Begin {statHeightCon}
     stpk.n1:=1;
     stpk.n2:=2;
     stpk.max[1]:=nd^.temp3;
     stpk.max[2]:=abs(nd^.contrast[t]);
     stpk.min[1]:=stpk.max[1];
     stpk.min[2]:=stpk.max[2];
     getBounds(stpk.min,stpk.max,nd,t);  
     stpk.number:=0;
     stpk.mean[1]:=0;
     stpk.mean[2]:=0;
     getMean(t,nd,stpk.number,stpk.mean);
     stpk.mean[1]:=stpk.mean[1]/stpk.number;
     stpk.mean[2]:=stpk.mean[2]/stpk.number;
     stpk.variance[1]:=0;
     stpk.variance[2]:=0;
     stpk.corr:=0;
     getVar(t,nd,stpk);
     stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);
     stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);
     stpk.corr:=stpk.corr/(stpk.number-1);
     stpk.cov:=stpk.corr;
     If(stpk.variance[1]>0) and (stpk.variance[2]>0)
     Then stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
     Else stpk.corr:=0;
     sx:=0;
     sy:=0;
     sx2:=0;
     sxy:=0;
     getSums(t,sx,sy,sx2,sxy,nd);
     If ((stpk.number*sx2-sqr(sx))<>0) Then Begin
        stpk.lsregok:=true;
        d:=1/(stpk.number*sx2-sqr(sx));
     End {then}
     Else Begin
        stpk.lsregok:=false;
        d:=0;
     End; {else}
     stpk.ls:=d*(stpk.number*sxy-sx*sy);
     stpk.lsy:=d*(sx2*sy-sx*sxy);
     finishstat(stpk)
  End; {procedure statheightcon}   
{***************************************************************************}
{This procedure displays the node height and contrast statistics calculated
by statheightcon (above).  It calls several display procedures from the
T_UTIL1 unit. (Screens 3 and 4)                                             } 
{***************************************************************************}
Procedure plotheightcon(rt:nodeptr;
                        t,n:integer;
                        fl:flstring;
                        spNode : nodePtr;
                        clade : boolean;
                        var bail:boolean);
   Var
      fakeMin,scalingFactor:traits;
      mx,my,temp:integer;
      stpk:statpak; 
      origin:itraits;
      s:string[10];
      st:string[80];  
   {************************************************************************}
   {This recursive procedure plots data points onto the screen.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure dumpdata(min:traits;
                      orgn:itraits;
                      sf:traits;
                      nd:nodePtr;
                      t:integer;
                      special:boolean);
      Var
         coor:itraits;
      Begin
         If(nd^.sibL<>nil) Then Begin
            coor[1]:=orgn[1]+round(sf[1]*(nd^.temp3-min[1]));
            coor[2]:=orgn[2]-round(sf[2]*(ABS(nd^.contrast[t])-min[2]));
            If (nd = rt) Then xSymbol(coor)
            Else 
               If ((nd = spNode) or (special = true)) Then openCircle(coor)
               Else star(coor);
            If ((nd = spNode) and clade) Then Begin
               dumpData(min,orgn,sf,nd^.sibL,t,true);
               dumpData(min,orgn,sf,nd^.sibR,t,true);
            End {then}
            Else Begin
               dumpData(min,orgn,sf,nd^.sibL,t,special);
               dumpData(min,orgn,sf,nd^.sibR,t,special);
            End
         End {then}
      End; {procedure dumpdata}
   {*************************************************************************}
   Begin {plotheightcon}
      statHeightCon(rt,t,stpk);
      fakeMin[1]:=stpk.min[1];
      fakeMin[2]:=0.0;
      graphicsOn(mx,my);
      initPlotx0(1,2,fakemin,stpk.max,fl,scalingfactor,origin,mx,my);
      outTextXY(0,0,'Plot of absolute value of');
      outTextXY(0,oneline,'standardized contrasts versus');
      outTextXY(0,2*oneline,'its base node''s height');
      str(t,s);
      outTextXY(0,3*oneline,'for Trait '+s+'.');
      outTextXY(0,6*oneline,'d.f. may be fewer if tree');
      outTextXY(0,7*oneline,'   contains polytomies.');

      st:='Absolute value of standardized contrast';
      temp:=round((origin[2]-length(st)*oneline)/2);
      If temp<1
      Then temp:=1;
      dumpDown(round(0.9524*origin[1])-8,temp,st);
      st:='Corrected height of base node of contrast';
      temp:=round((mx-origin[1]-textwidth(st))/2)+origin[1];
      If temp< origin[1]+1
      Then temp:=origin[1]+1;
      outTextXY(temp,origin[2]+3,st);
      str(n,s);
      outTextXY(mx-2*oneline,my-oneline,s);

      dumpData(fakeMin,origin,scalingFactor,rt,t,false);
      dumpStat(stpk,mx,my,bail,true)
   End; {procedure plotheightcon}
{***************************************************************************}
{This procedure computes the statistics for the absolute value of
{standardized contrasts versus their standard deviations.          }
{***************************************************************************}
Procedure statStdVsD(nd:nodePtr;t:integer;Var stpk:statPak);
   Var
      sy,sx,sx2,sxy,d:extended;
   {************************************************************************}
   {This recursive procedure gets the max and min values for the standard
   deviation (in trait 1) and the absolute value of the standardized contrast
   (in trait2).  The t argument specifies which *data* trait provides the
   source of the contrasts.
   Notice that it is local to statStdVsD, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getbounds(Var min,max:traits; nd:nodePtr;t:integer);
      Var
         standard:extended;
      Begin
         If nd^.sibL<>nil
         Then Begin
            If nd^.sdcon>max[1] Then max[1]:=nd^.sdcon;
            If nd^.sdcon<min[1] Then min[1]:=nd^.sdcon;
            standard:=ABS(nd^.contrast[t]);
            If standard>max[2] Then max[2]:=standard;
            If standard<min[2] Then min[2]:=standard;
            getBounds(min,max,nd^.sibL,t);
            getBounds(min,max,nd^.sibR,t)
         End {then}
      End; {procedure getbounds}      
   {************************************************************************}
   {This recursive procedure sums and counts the standard deviations and
   absolute values of standardized contrasts.  Again, the first trait in the
   statpak holds the sum of the deviations and the second holds the sum of
   the absolute values of contrasts.  T specfies which data trait the
   procedure is using for contrasts.
   Notice that it is local to statStdVsD, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getMean(t:integer;nd:nodePtr;Var n:integer;var m: traits);
      Begin
         If nd^.sibL<>nil Then Begin 
            inc(n);
            m[1]:=m[1]+nd^.sdcon;
            m[2]:=m[2]+abs(nd^.contrast[t]);
            getMean(t,nd^.sibL,n,m);
            getMean(t,nd^.sibR,n,m)
         End {then}
      End; {procedure getmean}
        
   {************************************************************************}
   {This recursive procedure calculates the sums of squares and cross
   products of deviations.  These are used to calculate variances and
   correlations between standard deviations and absolute values of
   standardized contrasts.
   Notice that it is local to statStdVsD, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getVar(t:integer;nd:nodePtr;Var s:statPak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(s.mean[1]-nd^.sdcon);
            s.variance[2]:=s.variance[2]+sqr(s.mean[2]-abs(nd^.contrast[t]));
            s.corr:=s.corr+(s.mean[1]-nd^.sdcon)*(s.mean[2]-
                                          abs(nd^.contrast[t]));
            getVar(t,nd^.sibL,s);
            getVar(t,nd^.sibR,s)
         End {then}
      End; {procedure getvar}
   {************************************************************************}
   {This recursive procedure does the first step in finding the least squares
   regression for the data.  It finds the sums needed.
   Notice that it is local to statStdVsD, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getSums(t:integer;Var sx,sy,sx2,sxy:extended; nd:nodePtr);
      Begin
         If nd^.sibL<>nil Then Begin
            sx:=sx+nd^.sdcon;
            sx2:=sx2+sqr(nd^.sdcon);
            sy:=sy+abs(nd^.contrast[t]);
            sxy:=sxy+nd^.sdcon*abs(nd^.contrast[t]);
            getSums(t,sx,sy,sx2,sxy,nd^.sibL);
            getSums(t,sx,sy,sx2,sxy,nd^.sibR)
         End {then}
      End; {procedure getsums}
   {**************************************************************************}
   Begin  {statstdvsd}
      stpk.n1:=1;
      stpk.n2:=2;
      stpk.max[1]:=nd^.sdcon;
      stpk.max[2]:=abs(nd^.contrast[t]);
      stpk.min[1]:=stpk.max[1];
      stpk.min[2]:=stpk.max[2];
      getBounds(stpk.min,stpk.max,nd,t);
      stpk.number:=0;
      stpk.mean[1]:=0;
      stpk.mean[2]:=0;
      getMean(t,nd,stpk.number,stpk.mean);
      stpk.mean[1]:=stpk.mean[1]/stpk.number;
      stpk.mean[2]:=stpk.mean[2]/stpk.number;
      stpk.variance[1]:=0;
      stpk.variance[2]:=0;
      stpk.corr:=0;
      getVar(t,nd,stpk);
      stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);
      stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);
      stpk.corr:=stpk.corr/(stpk.number-1);
      stpk.cov:=stpk.corr;
      If (stpk.variance[1]>0) And (stpk.variance[2]>0)
      Then stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr:=0;
      sx:=0;
      sy:=0;
      sx2:=0;
      sxy:=0;
      getSums(t,sx,sy,sx2,sxy,nd);
      If ((stpk.number*sx2-sqr(sx))<>0) Then Begin
         stpk.lsregok:=true;
         d:=1/(stpk.number*sx2-sqr(sx));
         stpk.ls:=d*(stpk.number*sxy-sx*sy);
         stpk.lsy:=d*(sx2*sy-sx*sxy);
      End {then}
      Else Begin
         stpk.lsregok:=false;
         d:=0
      End; {else}
      finishStat(stpk)
   End; {procedure statstdvsd}
{***************************************************************************}
{This procedure displays the absolute value of standardized contrasts versus
their standard deviations statistics calculated by statstdvsd (above).  It
calls several display procedures from the T_UTIL1 unit. (Screens 1 and 2)                                             }
{***************************************************************************}
Procedure plotStdVsD(rt:nodePtr;
                     t,n:integer;
                     fl:flString;
                     spNode : nodePtr;
                     clade : boolean;
                     Var bail:boolean);
   Var
      fakeMin,scalingFactor:traits;
      mx,my,temp:integer;
      stpk:statPak; 
      origin:itraits;
      s:string[10];
      st:string[80];    
   {************************************************************************}
   {This recursive procedure plots data points onto the screen.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure dumpData(min:traits;
                      orgn:itraits;
                      sf:traits;
                      nd:nodePtr;
                      t:integer;
                      special :boolean);
      Var
         coor:itraits;
      Begin
         If(nd^.sibL<>nil) Then Begin
            coor[1]:=orgn[1]+round(sf[1]*(nd^.sdcon-min[1]));
            coor[2]:=orgn[2]-round(sf[2]*(ABS(nd^.contrast[t])-min[2]));
            If (nd = rt) Then xSymbol(coor)
            Else
               If ((nd = spNode) or (special = true)) Then openCircle(coor)
               Else star(coor);
            If ((nd = spNode) and clade) Then Begin
               dumpData(min,orgn,sf,nd^.sibL,t,true);
               dumpData(min,orgn,sf,nd^.sibR,t,true)
            End {then}
            Else Begin
               dumpData(min,orgn,sf,nd^.sibL,t,special);
               dumpData(min,orgn,sf,nd^.sibR,t,special)
            End
         End {then}
      End; {procedure dumpdata}
   {**************************************************************************}
   Begin {plotstdvsd}
      statStdVsD(rt,t,stpk);
      fakeMin[1]:=stpk.min[1];
      fakeMin[2]:=0.0;
      graphicsOn(mx,my);
      initPlotX0(1,2,fakeMin,stpk.max,fl,scalingFactor,origin,mx,my);
      outTextXY(0,0,'Plot of absolute value of');
      outTextXY(0,oneLine,'standardized contrasts versus');
      outTextXY(0,2*oneLine,'their standard deviations for');
      str(t,s);
      outTextXY(0,3*oneLine,'Trait '+s+'.');
      outTextXY(0,6*oneLine,'d.f. may be fewer if tree');
      outTextXY(0,7*oneLine,'   contains polytomies.');
      st:='Absolute value of standardized contrast';
      temp:=round((origin[2]-length(st)*oneline)/2);
      if temp<1
      Then temp:=1;
      dumpDown(round(0.9524*origin[1])-8,temp,st);
      st:='Square root of sum of corrected branch lengths';
      temp:=round((mx-origin[1]-textwidth(st))/2)+origin[1];
      If temp< origin[1]+1
      Then temp:=origin[1]+1;
      outTextXY(temp,origin[2]+3,st);
      str(n,s);
      outTextXY(mx-2*oneline,my-oneline,s);
      dumpData(fakeMin,origin,scalingFactor,rt,t,false);
      dumpStat(stpk,mx,my,bail,true)
   End; {procedure plotstdvsd}

{**************************************************************************}
{This procedure explains why we are skipping plots 5, 6, 7, and 8.}
{**************************************************************************}
Procedure explainSkip;
   Begin
      writeln('The height of the internal nodes shows no variation.');
      writeln;
      writeln;
      writeln('Your tree must be a star!');
      writeln;
      writeln;
      writeln('Therefore, plots 5, 6, 7, and 8 will be skipped.');
      writeln;
      writeln;
      ret('Please press any key to continue.')
   End; {procedure explainskip}

{**************************************************************************}
{This procedure writes text to the screen that explains that a branch of
 the tree is longer than one billion.                                      }
{**************************************************************************}
Procedure explainSkip2;
   Begin
      writeln('Your tree has some branch lengths greater than 1,000,000,000.');
      writeln('This is too large for some of the procedures.');
      writeln;
      writeln;
      writeln('Therefore, plots 5, 6, 7, and 8 will be skipped.');
      writeln;
      writeln;
      writeln('One way to avoid this problem is to rescale the total height ',
                 'of the tree.');
      ret('Please press any key to continue.')
   End; {explainskip2}

{**************************************************************************}
{This procedure runs the screen displays for screens 1 - 8.                }
{**************************************************************************}
Procedure showDiagnostics(nd : nodePtr; fln : flString);
   Var
      specialNode : nodePtr;
      wholeClade : boolean;
      bail : boolean;
      allZero : boolean;
      tooLong : boolean;
   Begin
      plotOptions(nd,specialNode,wholeClade);
      computeCon(nd);
      plotStdVsD(nd,1,1,fln,specialNode,wholeClade,bail);
      If Not(bail)
      Then plotStdVsD(nd,2,2,fln,specialNode,wholeClade,bail);
      If Not(bail)
      Then plotNodeCon(nd,1,3,fln,specialNode,wholeClade,bail);
      If Not(bail)
      Then plotNodeCon(nd,2,4,fln,specialNode,wholeClade,bail);
      If Not(bail)
      Then Begin
         temp3CorrectedHeight(nd,nd,0);
         allZero:=true;
         checkAllZero(nd,allzero);
         If allZero
         Then explainSkip
         Else Begin
            tooLong:=false;
           { checkBranchTooLong(nd,tooLong); }
            If Not(tooLong)
            Then Begin
               plotHeightCon(nd,1,5,fln,specialNode,wholeClade,bail);
               If Not(bail)
               Then plotHeightCon(nd,2,6,fln,specialNode,wholeClade,bail);
               temp3CorrectedHeight(nd,nd,0);
               If Not(bail)
               Then plotHeightNode(nd,1,7,fln,specialNode,wholeClade,bail);
               If Not(bail)
               Then plotHeightNode(nd,2,8,fln,specialNode,wholeClade,bail);
            End {then}
            Else explainSkip2
         End {else}
      End {then}
   End; {procedure showDiagnostics}

End. {unit T_FELS_D}