{************************************************************************}
{                              T_FELS_R.PAS                              }
{************************************************************************}
{This module displays 3 different X-Y plots related to regression of
contrasts between the two traits.  It also does some statistics related
to reconstructed root values and optionally provides some confidence and
prediction interval values.  Like T_FELS_D, each of the 3 X-Y plot screens
(numbered 9, 10, 11) has separate procedures to calculate statistics and
display the screen.  After plot screen #9, there are three text screens
9A, 9B, and 9C that elaborate on the regression displaye in 9.  Each of
these latter screens can generate a separate text file.


Maybe one procedure with functional/procedural arguments for the
statistics would work (and save space), it could get really complicated.

The various regression related screens are discussed in:


}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}



Unit T_FELS_R;

Interface

Uses t_util1;

Procedure dumpstat9A(Var dest: text; s :statpak);

Procedure dumpStat9B(Var dest : text;
                     stpk : statPak;
                     nd : nodePtr);


Procedure checkConstantTips(rt:nodePtr;
                            Var index : integer);

Procedure PlotRoot(nd:nodePtr;
                   fln : flString;
                   conSlope : extended;
                   conT1,conT2 : integer;
                   dfReduce : integer);

Procedure tempResidual(nd:nodePtr; m,y0:extended;t,t2:integer);

Procedure plotresidualX(rt:nodePtr;
                        t:integer;
                        fl:flString;
                        spNode : nodePtr;
                        clade : boolean;
                        Var bail:boolean);

Procedure tempAbsResidual(nd:nodePtr; m,y0:extended;t,t2:integer);

Procedure statResidualHeight(nd:nodePtr;Var stpk:statPak);

Procedure plotResidualHeight(rt : nodePtr;
                             fl : flString;
                             spNode : nodePtr;
                             dfReduce : integer;
                             clade : boolean;
                             Var bail : boolean);


Procedure statCon(nd:nodePtr;Var stpk:statPak);

Procedure plotCon(rt:nodePtr;
                  Var spNode : nodePtr;
                  fl:flString;
                  Var m,y0:extended;
                  Var t,t2:integer;
                  Var dfReduce : integer;
                  Var wholeclade : boolean;
                  Var bail:boolean);

Procedure statRoot(rt : nodePtr;
                   Var stpk : statPak;
                   conSlope : extended;
                   n1,n2 : integer);

Procedure statresidualX(nd:nodePtr;t:integer;Var stpk:statPak);

Procedure fileContrasts(nd:nodePtr;Var fl:text);

Procedure showRegressionStats(nd : nodePtr;fln : flString);


implementation
uses crt,
     graph,
     te_util1;

{************************************************************************}
{These constants control the width of the confidence/prediction intervals}
{used in the .RT and .CI files spit out by plotRoot.                     }
{************************************************************************}
   Const defaultWidth1 = 0.95;
         defaultWidth2 = 0.90;

{***************************************************************************}
{This procedure outputs text to the text screen, then prompts the user to
choose which trait goes along the X axis.}
{***************************************************************************}
Procedure dumpText2(Var n1,n2:integer);
   Var
      ch:char;
   Begin
      clrscr;
      writeln;
      writeln('The following screen, number 9, will be a scattergram of');
      writeln('     standardized independent contrasts for your two traits.');
      writeln;
      writeln('The trait that you choose to have plotted on the horizontal');
      writeln('     or X axis will be "positivized" as described in');
      writeln('     Garland, Harvey, and Ives (1992) Syst. Biol. 31:18-32.');
      writeln('     Note that if the trait you choose to have plotted on the');
      writeln('     X axis includes some contrasts that are exactly zero, then');
      writeln('     the sign of the corresponding Y variable contrasts will be');
      writeln('     ambiguous.  However, this graphical ambiguity does not');
      writeln('     affect the statistics, e.g., correlation, slopes, F.');
      writeln;
      writeln;
      write('Which trait would you like along the horizontal axis (1 (=Enter) or 2)? ');
      ch:=readKey;
      writeln;
      If ch<>'2' Then n1:=1 Else n1:=2;
      If n1=1 Then n2:=2 Else n2:=1;
      writeln;
   End; {procedure dumpText2}

   {***************************************************************************}
   {Just text for the screen.}
   {***************************************************************************}
   Procedure dumptext3;
      Begin
         clrscr;
         writeln;
         writeln('The remaining plots are scattergrams of residuals, i.e., ',
                  'vertical');
         writeln('deviations from the independent contrast least-squares linear',
                  ' regression');
         writeln('through the origin shown on screen 9.');
         writeln;
         writeln;
         writeln('Because this is regression through the origin,');
         writeln('residuals do not necessarily sum to zero.');
         writeln;
         writeln;
         writeln('See Grafen (1989, Phil. Trans. Royal Soc. London B. 326:119-157',
                   ', especially');
         writeln('p. 146) for some discussion of the plot on screen 11.');
         ret('Press any key to continue.');
      End; {procedure dumptext3}

   {***************************************************************************}
   {This queries the user for the length of a branch attached to the root for a}
   {hypothetical added species, to calculate prediction intervals for the      }
   {species.                                                                   }
   {***************************************************************************}
   Procedure dumpText4(Var Vh : extended; contempTips : boolean);
      Var userChar : char;
      Begin
         writeln;
         writeln('The calculations for prediction intervals effectively assume that');
         writeln('the species to be predicted is attached at the root (basal node) of');
         writeln('the phylogenetic tree (see Garland and Ives, 2000).  By default, the');
         writeln('branch length (Vh) leading to the predicted species is assumed to equal');
         writeln('the average length of the (uncorrected) branches from root to tips.');
         writeln;
         write('Vh is ');
         extendWrite(output,vh);
         If contempTips
         Then writeln('.  All tips are contemporaneous in the uncorrected tree.')
         Else writeln('.  The tips are not contemporaneous in the uncorrected tree.');
         writeln;
         write('Do you wish to change this value? (Y or anything else to not change) ');
         readln(userChar);
         If (userChar in ['Y','y']) Then Begin
            write('Enter the value for Vh: ');
            readln(Vh)
         End; {then}
         writeln;
      End; {procedure dumpText4}
  {**********************************************************************}
  {This procedure counts the polytomies in the tree.  This value will be
   used to calculate the minimum and maximum degrees of freedom.         }
  {**********************************************************************}
   Procedure countPolytomies(nd:nodePtr;Var count : integer);
      Begin
         If nd^.sibL<>nil
         Then Begin
            If ((nd^.sibL^.length = 0) And (nd^.sibL^.sibL <> nil))
            Then inc(count);
            countPolytomies(nd^.sibL,count);
         End; {then}
         If nd^.sibR<>nil
         Then Begin
            If ((nd^.sibR^.length = 0) and (nd^.sibR^.sibL <> nil))
            Then inc(count);
            countPolytomies(nd^.sibR,count);
         End; {then}
      End; {procedure countPolytomies}

  {***************************************************************************}
  {This procedure outputs text to the text screen, then prompts the user to
  choose which trait goes along the X axis.}
  {***************************************************************************}
   Procedure reduceDF(rt : nodePtr; Var s : statpak);
      Var 
          temp : longInt;
          polytomies : integer;
          valCode : integer;
          tips : integer;
          userStr : string[16];
      {*********************************************************************}
      {Fast, local version of numtps.
      {*********************************************************************}
      Procedure numtps(nod:nodeptr; var num:integer);
         Begin
            If nod^.sibL<>nil Then Begin
               numtps(nod^.sibL,num);
               numtps(nod^.sibR,num);
            End
            Else inc(num);
         End;
      {*********************************************************************}
      Begin
         polytomies := 0;
         countpolytomies(rt,polytomies);
         tips := 0;
         numtps(rt,tips);

         writeln('The following screens will report, among other things, confidence intervals');
         writeln('for the traits at the root node, as described in:');
         writeln;
         writeln('Garland, T., Jr., P. E. Midford, and A. R. Ives. 1999. An introduction to');
         writeln('     phylogenetically based statistical methods, with a new method for');
         writeln('     confidence intervals on ancestral values. American Zoologist 39:374-388.');
         writeln;
         writeln('You will also have the option of saving confidence intervals and prediction');
         writeln('intervals for new observations, as described in:');
         writeln;
         writeln('Garland, T., Jr., and A. R. Ives. 2000. Using the past to predict the present:');
         writeln('     confidence intervals for regression equations in phylogenetic comparative');
         writeln('     methods. American Naturalist 155:346-364.');
         writeln;
         writeln('For all of the foregoing confidence (and prediction) intervals, degrees of');
         writeln('freedom must be specified, hence the following option.');
         writeln;
         writeln('Your tree contains ',polytomies,' internal branches of zero length.  If some of these');
         writeln('indicate soft polytomies, then you may want to reduce degrees of freedom');
         writeln('(see Purvis and Garland, 1993, Syst. Biol. 42:569-575; Garland and');
         writeln('Diaz-Uriarte, 1999, Syst. Biol. 48:547-558).');
         writeln;
         writeln('Would you like to subtract some degrees of freedom for confidence and');
         writeln('prediction intervals in screens 9A, 9B, and 9C (Y/number/N{=Enter})? ');
         readln(userStr);
         val(userStr,temp,valCode);
         If (valCode = 0) Or
            ((userStr <> '') And (upcase(userStr[1]) = 'Y')) Then Begin
            If (valCode <> 0)
            Then Begin
               write('Enter the number to reduce degrees of freedom by: ');
               readln(userStr);
               val(userStr,temp,valCode);
            End;
            While ((valCode<>0) Or
                   (temp<0) Or
                   (temp>=tips-2)) Do Begin
               If (valCode<>0)
               Then write('Non-numeric string entered.')
               Else write('Your value is out of the range 0 - ',tips-3);
               writeln;
               write('Enter the number to reduce degrees of freedom by (0 = no reduction): ');
               readln(userStr);
               writeln;       {so number doesn't get overwritten by Finished message}
               val(userStr,temp,valCode);
            End; {while}
         End;{then}
         If valCode=0
         Then s.dfReduce := temp
         Else s.dfReduce := 0;
      End; {procedure reduceDF}

{***********************************************************************}
{This procedure displays the columns of the .CI (confidence and         }
{prediction interval) file.  The variable list appears on Screen 9C.    }
{***********************************************************************}
Procedure showCIFileColumns(name :flString);
   Var
      w1 : integer;
      w2 : integer;
   Begin
      w1 := round(width1*100);
      w2 := round(width2*100);
      clrscr;
      writeln('Your (Bivariate) Confidence/Prediction Interval file (',name,') will');
      writeln('contain the following columns:');
      writeln;
      writeln('     1.  Tip Name');
      writeln('     2.  X Value');
      writeln('     3.  Observed Y Value');
      writeln('     4.  Predicted Y Value (Yhat)');
      writeln('     5.  Lower ',w1:2,'% Confidence Interval');
      writeln('     6.  Upper ',w1:2,'% Confidence Interval');
      writeln('     7.  Lower ',w1:2,'% Prediction Interval');
      writeln('     8.  Upper ',w1:2,'% Prediction Interval');
      writeln('     9.  Lower ',w2:2,'% Confidence Interval');
      writeln('    10.  Upper ',w2:2,'% Confidence Interval');
      writeln('    11.  Lower ',w2:2,'% Prediction Interval');
      writeln('    12.  Upper ',w2:2,'% Prediction Interval');
      writeln;
      writeln('Writing data to disk...');
      gotoxy(78,24);
      write('9C');
   End; {procedure showCIFileColumns}

{*************************************************************************}
{This procedure dumps the statistics given to it to the graphics screen.
It is used to add the statistics output to the graphs, and also writes the
word 'zero' in for the means and the intercepts.}
{*************************************************************************}  
  Procedure dumpStat9(stpk:statpak;
                      mx,my:integer;
                      Var bail : boolean);
     Var
        t:string;
        ch:char;
     {********************************************************************}
     {This simple procedure writes a line of text to the graphics screen.
     {********************************************************************}
     Procedure showLine(line: integer;strtext : string);
        Begin
           outtextxy(0,line*oneline,strtext);
        End; {procedure showLine}
     {********************************************************************}
     {This procedure writes an extended value to the graphics screen.  It
      sort of combines the functionality of showLine (above) and
      extendWrite.                                                        }
     {********************************************************************}
     Procedure showValue(line: integer; dataLabel : string;num : extended);
        Begin
           extendStr(t,num);
           outtextxy(0,line*oneline,datalabel+t);
        End; {procedure showValue}
     {********************************************************************}
     Begin {procedure dumpStat9}
        str(stpk.number,t);
        showLine(8,'Number of contrasts: '+t);
        showValue(10,'Min X:',stpk.min[stpk.n1]);
        showValue(11,'Min Y:',stpk.min[stpk.n2]);
        showValue(13,'Max X:',stpk.max[stpk.n1]);
        showValue(14,'Max Y:',stpk.max[stpk.n2]);
        showLine(16,'Mean X,Y: zero,zero');
        { showLine(17,'Mean Y: zero'); }
        showValue(18,'Var X:',stpk.variance[stpk.n1]);
        showValue(19,'Var Y:',stpk.variance[stpk.n2]);
        showValue(20,'Cov:',stpk.cov);
        showValue(22,'Corr:',stpk.corr);
        showLine(24,'Reduced Major Axis.');
        If (stpk.lsregok) Then Begin
           showValue(25,'     Slope:',stpk.rma);
           showLine(26,' Intercept: zero');
        End {then}
        Else Begin
           showLine(25,'     Slope: undefined');
           showLine(26,' Intercept: undefined');
        End; {else}
        showLine(28,'Major Axis.');
        showValue(29,'     Slope:',stpk.ma);
        showLine(30,' Intercept: zero');
        showLine(32,'Least Squares Regression.');
        If (stpk.lsregok)
        Then Begin
           showValue(33,'     Slope:',stpk.ls);
           showLine(34,' Intercept: zero');
           showValue(35,' R squared:',stpk.r2);
        End {then}
        Else Begin
           showLine(33,'     Slope: undefined');
           showLine(34,' Intercept: undefined');
           showLine(35,' R squared: undefined');
        End; {else}
        If (abs(stpk.corr)<>1)
        Then Begin
           showValue(37,'t: ',stpk.ts);
           showValue(38,'F: ',stpk.f);
        End {then}
        Else Begin
           showLine(37,'t: undefined');
           showLine(38,'F: undefined');
        End; {else}
        str(stpk.df,t);
        showLine(39,'d.f.: '+t);
        str(stpk.conypos,t);
        showLine(41,'# Y contrasts > 0 : '+t);
        str(stpk.conyneg,t);
        showLine(42,'# Y contrasts < 0 : '+t);
        str(stpk.cony0,t);
        showLine(43,'# Y contrasts = 0 : '+t);
        showLine(44,'Sign-test (2 tailed): '); 
        showValue(45,'      ',stpk.signPvalue);
        ch:=readkey;
        If ((ch = 'R') or (ch = 'r'))
        Then bail := true
        Else bail := false;
        Closegraph;
     end; {procedure dumpStat9}

{**************************************************************************}
{This procedure calculates the statistics for the contrasts vs. contrasts
plot.  It uses several recursive procedures to help calculate the stats just
like in the procedures above.  The main difference is that it calculates
the stats through the origin so it artificially sets the means to zero.} 
{**************************************************************************}
Procedure statCon(nd:nodePtr;Var stpk:statPak);
Var
   sx2,sxy,d:extended;
   {************************************************************************}
   {This procedure finds max and min values for the contrasts on both traits}
   {************************************************************************}
   Procedure getBounds(Var min,max:traits; nd:nodePtr);
      Begin
         If nd^.sibL<>nil Then Begin
            If nd^.contrast[1]>max[1] Then max[1]:=nd^.contrast[1];
            If nd^.contrast[1]<min[1] Then min[1]:=nd^.contrast[1];
            If nd^.contrast[2]>max[2] Then max[2]:=nd^.contrast[2];
            If nd^.contrast[2]<min[2] Then min[2]:=nd^.contrast[2];
            getBounds(min,max,nd^.sibL);
            getBounds(min,max,nd^.sibR);
         End; {then}
      End; {procedure getbounds}
   {************************************************************************}
   {This procedure counts the interior nodes of the tree, which equals the
   number of contrasts.                                                    }
   {************************************************************************}
   Procedure getNum(nd:nodePtr;Var n:integer);
     Begin
         If nd^.sibL<>nil Then Begin
            inc(n);
            getNum(nd^.sibL,n);
            getNum(nd^.sibR,n);
         End; {then}
     End; {procedure getnum}
   {************************************************************************}
   {This procedure sums the squares and cross products of the contrasts. }
   {************************************************************************}
   Procedure getVar(nd:nodePtr; Var s:statPak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(nd^.contrast[1]);
            s.variance[2]:=s.variance[2]+sqr(nd^.contrast[2]);
            s.corr:=s.corr+(nd^.contrast[1])*(nd^.contrast[2]);
            getVar(nd^.sibL,s);
            getVar(nd^.sibR,s);
         End; {then}
      End; {procedure getvar}      
   {************************************************************************}
   {This procedure counts the number of contrasts that are greater than,
    less than, or equal to zero (updating the appropriate fields of the
    statpak.)                                                               }
   {************************************************************************}
   Procedure getNumCon(nd:nodeptr; Var s:statPak; n2:integer);
      Begin
         If nd^.sibL<>nil Then Begin
            If nd^.contrast[n2]>0
            Then inc(s.conypos)
            Else
               If nd^.contrast[n2]<0
               Then inc(s.conyneg)
               Else inc(s.cony0);
            getNumCon(nd^.sibL,s,n2);
            getNumCon(nd^.sibR,s,n2);
         End; {then}
      End; {procedure getnumcon}
   {************************************************************************}
   {This procedure calculates a sign-test p-value for the distribution of
    contrast signs calculated in the previous procedure.  The version in
    t_util1 is one-tailed, so that value is doubled here.  Note that any
    contrasts equal to zero are ignored, as indicated by Sokal and Rohlf
    (1995, Biometry 3rd ed. p.444).}
   {************************************************************************}
   Procedure calcSignTest(Var s:statPak);
      Begin
         s.signPvalue := signTest(s.conypos,s.conyneg);
         If (s.signPvalue > 0.5)
         Then s.signPvalue := 1.0
         Else s.signPvalue := 2*s.signPvalue;
      End; {procedure calcSignTest}

   {************************************************************************}
   {This procedure calculates sums of squares and cross products of
   contrasts.  Unlike getVar, it explicitly returns values, rather than
   writing fields in a statPak.}
   {************************************************************************}
   Procedure getSums(n1,n2:integer; Var sxy,sx2:extended; nd:nodePtr);
      Begin
         If(nd^.sibL<>nil) Then Begin
            sxy:=sxy+ nd^.contrast[n1]*nd^.contrast[n2];
            sx2:=sx2+ sqr(nd^.contrast[n1]);
            getSums(n1,n2,sxy,sx2,nd^.sibL);
            getSums(n1,n2,sxy,sx2,nd^.sibR);
         End; {then}
      End; {procedure getSums}
   {************************************************************************}
   {This procedure deserves some mention.  It "positivizes" the contrasts, as
   explained in Garland, Harvey, and Ives (1992) Syst. Biol. 41:18-32.  If
   the contrast on the X axes is negative, then it flips its sign as well as
   the sign of the other contrast at that node.}
   {************************************************************************}
   Procedure positivise(n1,n2:integer; nd:nodePtr);
      Begin
         If (nd^.sibL<>nil) Then Begin
            If (nd^.contrast[n1]<0) Then Begin
               nd^.contrast[1]:=nd^.contrast[1]*-1;
               nd^.contrast[2]:=nd^.contrast[2]*-1;
            End; {then}
            positivise(n1,n2,nd^.sibL);
            positivise(n1,n2,nd^.sibR);
         End; {then}
      End; {procedure postivise}
   {*************************************************************************}
   Begin {statCon}
      dumpText2(stpk.n1,stpk.n2);
      positivise(stpk.n1,stpk.n2,nd);
      stpk.max[1]:=nd^.contrast[1];
      stpk.max[2]:=nd^.contrast[2];
      stpk.min[1]:=stpk.max[1];
      stpk.min[2]:=stpk.max[2];
      getBounds(stpk.min,stpk.max,nd);
      stpk.number:=0;
      stpk.mean[1]:=0;
      stpk.mean[2]:=0;
      getNum(nd,stpk.number);
      stpk.variance[1]:=0;
      stpk.variance[2]:=0;
      stpk.corr:=0;
      getVar(nd,stpk);
      stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);
      stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);
      stpk.corr:=stpk.corr/(stpk.number-1);
      stpk.cov:=stpk.corr;
      If(stpk.variance[1]<>0) And (stpk.variance[2]<>0)
      Then
         stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else
         stpk.corr:=0;
      sxy:=0;
      sx2:=0;
      getSums(stpk.n1,stpk.n2,sxy,sx2,nd);
      If (sx2<>0) Then Begin
         stpk.lsregok:=true;
         stpk.ls:=sxy/sx2;
         stpk.lsy:=0.0;
      End {then}
      Else
         stpk.lsregok:=false;
      finishStatOrg(stpk);
      stpk.df:=stpk.df+1;
      stpk.conypos:=0;
      stpk.conyneg:=0;
      stpk.cony0:=0;
      getNumCon(nd,stpk,stpk.n2);
      calcSignTest(stpk);
   End; {procedure statCon}
{**************************************************************************}
{This procedure plots the positivized standardized contrasts one against the
other.  Again, it is similar to the diagnostic plot procedures in
T_FELS_D. (Screen 9) }
{**************************************************************************}
Procedure plotCon(rt:nodePtr;
                  Var spNode : nodePtr;
                  fl:flString;
                  Var m,y0:extended;
                  Var t,t2:integer;
                  Var dfReduce : integer;
                  Var wholeclade : boolean;
                  Var bail:boolean);
   Var
      fakeMin,scalingFactor:traits;
      origin:itraits;
      mx,my:integer;
      stpk:statPak;
      valCode : integer;
      userStr : string[16];
      ch:char;
   {************************************************************************}
   Procedure dumpdata(orgn:itraits;
                      sf,min:traits;
                      nd:nodePtr;
                      n1,n2:integer;
                      special :boolean);
      Var
         coor:itraits;
      Begin
         If(nd^.sibL<>nil) Then Begin
            coor[1]:=orgn[1]+round(sf[1]*(nd^.contrast[n1]-min[n1]));
            coor[2]:=orgn[2]-round(sf[2]*(nd^.contrast[n2]-min[n2]));
            If (nd^.contrast[n1]=0) Then Begin
               outTextXY(0,6*oneLine,'Note:  X contrast is 0, so');
               outTextXY(0,7*oneLine,'   positivization is ambiguous.');
            End;
            If (nd = rt) Then xSymbol(coor)
            Else
               If ((nd = spNode) or (special = true)) Then openCircle(coor)
               Else star(coor);
            If ((nd = spNode) and wholeClade) Then Begin
               dumpData(orgn,sf,min,nd^.sibL,n1,n2,true);
               dumpData(orgn,sf,min,nd^.sibR,n1,n2,true);
            End {then}
            Else Begin
               dumpData(orgn,sf,min,nd^.sibL,n1,n2,special);
               dumpData(orgn,sf,min,nd^.sibR,n1,n2,special);
            End;
         End; {then}
      End; {procedure dumpData}
   {************************************************************************}
   Procedure GetOptions(rt : nodePtr;
                        Var spNode : nodePtr;
                        Var wholeClade : boolean);
      Var
         xPos : integer;
         yPos : integer;
         userChar : char;
         ndName : nameString;
         noGood : boolean;
      Begin
         writeln('By default, the root node contrast will be displayed as an ''X'', all other');
         writeln('contrasts will be displayed as closed circles.');
         writeln;
         writeln('Would you like to have a single contrast displayed with an open circle?');
         write('(Y/N/Any other character = N): ');
         xPos := whereX;
         yPos := whereY;
         writeln;
         writeln('(This feature is useful for testing a priori hypotheses about particular');
         writeln('contrasts.)');
         gotoxy(xPos,yPos);
         userChar := readKey;
         gotoxy(1,yPos+3);
         If ((userChar = 'Y') or (userChar = 'y')) Then Begin
            wholeClade := false;
            write('What is the two character name of the node (case matters)? ');
            readln(ndName);
            nogood := true;
            While (noGood) Do Begin
               checkName(rt,ndName,noGood,spNode);
               If (noGood = false) Then Begin
                  If (spNode^.sibL = NIL) Then Begin
                     writeln('The node ',ndName,' is a tip, not a contrast.');
                     noGood := true;
                  End {then}
               End
               Else
                  writeln('No node with the name ',ndName,' could be found in the tree.');
               If (noGood)
               Then Begin
                  write('Enter a new name, or carriage return to give up: ');
                  readln(ndName);
                  If (ndName = '')  Then Begin
                     noGood := false;
                     spNode := NIL;
                  End; {then}
               End; {then}
            End; {while}
         End {then}
         Else Begin
            wholeClade := true;
            writeln;
            writeln('Would you like to see all the contrasts for a clade displayed with an open');
            write('circle? (Y/N/Any other character = N): ');
            userChar := readKey;
            writeln;
            If ((userChar = 'Y') or (userChar = 'y')) Then Begin
               writeln('What is the two character name of the clade''s basal ');
               write('(root) node (case matters)? ');
               readln(ndName);
               nogood := true;
               While (noGood) Do Begin
                  checkName(rt,ndName,noGood,spNode);
                  If (noGood = false) Then Begin
                     If (spNode^.sibL = NIL) Then Begin
                        writeln('The node ',ndName,' is a tip, so it is not the base of a clade.');
                        noGood := true;
                     End {then}
                  End
                  Else
                     writeln('No node with the name ',ndName,' could be found in the tree.');
                  If (noGood) Then Begin
                     write('Enter a new name, or carriage return to give up: ');
                     readln(ndName);
                     If (ndName = '')  Then Begin
                        noGood := false;
                        spNode := NIL;
                        wholeClade := false;
                     End; {then}
                  End; {then}
               End; {while}
            End; {then}
         End; {else}
      End; {procedure getoptions}
   {************************************************************************}
   Begin {plotCon}
      statCon(rt,stpk);
      getOptions(rt,SpNode,WholeClade);
      fakeMin[stpk.n1]:=0;
      fakeMin[stpk.n2]:=stpk.min[stpk.n2];
      graphicsOn(mx,my);
      initPlotY0(stpk.n1,
                 stpk.n2,
                 fakeMin,
                 stpk.max,
                 fl,
                 scalingFactor,
                 origin,
                 mx,
                 my);
      showAxes(origin,stpk.n1,stpk.n2,mx,my);
      outTextXY(0,0,'Plot of standardized contrasts');
      outTextXY(0,oneline,'   with x axis positivized.');
      outTextXY(0,3*oneline,'Regression through (0,0).');
      stpk.polytomies:= 0;
      countPolytomies(rt,stpk.polytomies);
{      If ((stpk.polytomies>0) And (stpk.dfreduce=0)) Then Begin
        outTextXY(0,4*oneline,'d.f. may be overestimated');
        outTextXY(0,5*oneline,'   as tree contains polytomies.');
      End; }
      outTextXY(0,4*oneline,'d.f. may be fewer if tree');
      outTextXY(0,5*oneline,'   contains polytomies.');
      dumpData(origin,scalingfactor,fakemin,rt,stpk.n1,stpk.n2,false);
      setLineStyle(3,0,1);
      plotLine(0,0,fakemin,stpk.max,scalingfactor,origin,stpk.n1,stpk.n2);
      setLineStyle(0,0,1);
      plotZeroLabel(fakeMin,scalingFactor,origin,stpk.n2);
      If (stpk.lsregok) Then Begin
         plotLine(stpk.ls,
                  stpk.lsy,
                  fakeMin,
                  stpk.max,
                  scalingFactor,
                  origin,
                  stpk.n1,
                  stpk.n2);
         outTextXY(origin[1],
                   my-2*oneLine,
                   ' Solid line = least squares reg. through origin');
      End; {then}
      outTextXY(mx-2*oneLine,my-oneLine,'9');
      dumpStat9(stpk,mx,my,bail);
      m:=stpk.ls;
      y0:=stpk.lsy;
      t:=stpk.n1;
      t2:=stpk.n2;
      dfReduce := stpk.dfReduce;
   End;{procedure plotCon}

{*************************************************************************}
{This procedure writes the statistics related to the estimated root values
 and the regression line based on those values. (Screen 9A)               }
{*************************************************************************}
   Procedure dumpstat9A(var dest : text;s :statpak);
      Var
         ch : char;
      Begin
         writeln(dest,'Number of tips:  ',s.number+1);
         writeln(dest,'Number of independent contrasts:  ',s.number);
         writeln(dest,'Number of branches of 0 length:  ',s.polytomies);
         writeln(dest,'User-specified degrees of freedom reduction for computing C.I.''s:  ',s.dfReduce);
         writeln(dest);
         s.df := s.df-s.dfReduce;
         write(dest,'Slope (least squares):            ');
         extendWrite(dest,s.ls);
         writeln(dest);
         write('');
         write(dest,' Standard Error:                  ');
         write('');
         extendWrite(dest,s.slopeSE);
         writeln(dest);
         write(dest,' Lower ',round(width1*100):2,'% Confidence Interval:   ');
         extendWrite(dest,s.ls-s.slopeCI);
         writeln(dest);
         write(dest,' Upper ',round(width1*100):2,'% Confidence Interval:   ');
         extendWrite(dest,s.ls+s.slopeCI);
         writeln(dest);
         write(dest,'Intercept (least squares):        ');
         extendWrite(dest,s.lsy);
         writeln(dest);
         write(dest,' Standard Error:                  ');
         extendWrite(dest,s.interSE);
         writeln(dest,'   / as mapped back onto \');
         write(dest,' Lower ',round(width1*100):2,'% Confidence Interval:   ');
         extendWrite(dest,s.lsy-s.interCI);
         writeln(dest,'  |  the original data    |');
         write(dest,' Upper ',round(width1*100):2,'% Confidence Interval:   ');
         extendWrite(dest,s.lsy+s.interCI);
         writeln(dest,'   \ space               /');
         writeln(dest);
         writeln(dest,'                                            X = Trait ',s.n1:1,
                       '       Y = Trait ',s.n2:1);
         write(dest,'Estimated Root Node Values:              ');
         extendWrite(dest,s.mean[s.n1]);
         write(dest,' ');
         extendWrite(dest,s.mean[s.n2]);
         writeln(dest);
         write(dest,'Rates of Evolution:                      ');
         extendWrite(dest,s.variance[s.n1]);
         write(dest,' ');
         extendWrite(dest,s.variance[s.n2]);
         writeln(dest);
         write(dest,'Standard Errors of Root Node Estimates:  ');
         extendWrite(dest,s.se[s.n1]);
         write(dest,' ');
         extendWrite(dest,s.se[s.n2]);
         writeln(dest);
         write(dest,'Lower ',round(width1*100):2,'% Confidence Intervals:          ');
         extendWrite(dest,s.mean[s.n1]-s.ci[s.n1]);
         write(dest,' ');
         extendWrite(dest,s.mean[s.n2]-s.ci[s.n2]);
         writeln(dest);
         write(dest,'Upper ',round(width1*100):2,'% Confidence Intervals:          ');
         extendWrite(dest,s.mean[s.n1]+s.ci[s.n1]);
         write(dest,' ');
         extendWrite(dest,s.mean[s.n2]+s.ci[s.n2]);
         writeln(dest);
         s.df := s.df+s.dfReduce;  {avoid accidental side effects}
         writeln(dest);
      End; {procedure dumpstat9A}

{************************************************************************}
{This function calculates the mean height of all the tree's tips.  This  }
{provides a default value for Vh.  Vh is the assumed height of a         }
{tip when calculating a prediction interval.  The program reports whether}
{tips are contemporaneous, using a tolerance of 1E-10, the same tolerance}
{used for deciding whether a tip has a depth of zero.                    }
{************************************************************************}
Procedure meanHeight(rt : nodePtr;
                     first : nodePtr;
                     count : integer;
                     Var mean : extended;
                     Var contemporaneous : boolean);
   Var sum : extended;
       currentHeight : extended;
       oldHeight : extended;
       current : nodePtr;
   {*****************************************************************}
   {Support function to quickly calculate the height of a tip.       }
   {*****************************************************************}
   Function quickHeight (nd : nodePtr;rt : nodePtr ) : extended;
       Var
          tmp : nodePtr;
          height : extended;
       Begin
          tmp := nd;
          height := tmp^.length;
          While(tmp <> rt) Do Begin
             tmp := tmp^.ancestor;
             If tmp<>rt
             Then height := height + tmp^.length
          End; {while}
          quickHeight := height;
       End; {function quickHeight}
   {*****************************************************************}
   Begin {function meanHeight}
      sum := 0.0;
      current := first;
      contemporaneous := true;
      oldHeight := quickHeight(first,rt);
      Repeat
         currentHeight := quickHeight(current,rt);
         sum := sum + currentHeight;
         If abs((oldHeight-currentHeight)/currentHeight) > 1E-10
         Then contemporaneous := false;
         oldHeight := currentHeight;
         current := current^.tipR
      Until current = NIL;
      mean := sum/count;
   End;{function meanHeight}

{*******************************************************************}
{This calculates "single variable" Prediction intervals for         }
{hypothetical additions to a tree, as per equation A10 of           }
{Garland, Midford, and Ives (1999) American Zoologist 39:374-388.   }
{*******************************************************************}
   Procedure calc1PI(rt : nodePtr;
                     stpk : statPak;
                     Vh : extended;
                     Var PIt1w1 : extended;
                     Var PIt1w2 : extended;
                     Var PIt2w1 : extended;
                     Var PIt2w2 : extended);
      Var
         tWidth1 : extended;
         tWidth2 : extended;
         w1 : extended;
         w2 : extended;
         v1 : extended;
         v2 : extended;
      Begin
         v1 := rt^.sibL^.variance;
         v2 := rt^.sibR^.variance;
         tWidth1 := sqrt(stpk.variance[1]*(Vh+(v1*v2)/(v1+v2)));
         tWidth2 := sqrt(stpk.variance[2]*(Vh+(v1*v2)/(v1+v2)));
         w1 := tCI(stpk.number-stpk.dfreduce,width1);
         w2 := tCI(stpk.number-stpk.dfreduce,width2);
         PIt1w1 := w1 * tWidth1;
         PIt2w1 := w1 * tWidth2;
         PIt1w2 := w2 * tWidth1;
         PIt2w2 := w2 * tWidth2;
      End; {calc1PI}

{*******************************************************************}
{This writes the single variable Prediction intervals to the screen }
{or a file.                                                         }
{*******************************************************************}
   Procedure write1PI(stpk     : statPak;
                      Vh       : extended;
                      PIt1w1   : extended;
                      PIt1w2   : extended;
                      PIt2w1   : extended;
                      PIt2w2   : extended;
                      Var fl   : text);
      Var w1,w2 : integer;
          strBuf : string;
      Begin
         writeln(fl);
         writeln(fl);
         write(fl,'For a tip on a branch of length ');
         extendWrite(fl,Vh);
         writeln(fl);
         writeln(fl);
         With stpk Do Begin
            w1 := trunc(100*(width1));
            w2 := trunc(100*(width2));
            extendStr(strBuf,mean[n1]-PIt1w1);
            writeln(fl,'Lower ',w1,'% Prediction interval for Trait 1: ',strBuf);
            extendStr(strBuf,mean[n1]+PIt1w1);
            writeln(fl,'Upper ',w1,'% Prediction interval for Trait 1: ',strBuf);
            extendStr(strBuf,mean[n1]-PIt1w2);
            writeln(fl,'Lower ',w2,'% Prediction interval for Trait 1: ',strBuf);
            extendStr(strBuf,mean[n1]+PIt1w2);
            writeln(fl,'Upper ',w2,'% Prediction interval for Trait 1: ',strBuf);
            writeln;   {add a gap between the traits PEM 3-7-2003}
            extendStr(strBuf,mean[n2]-PIt2w1);
            writeln(fl,'Lower ',w1,'% Prediction interval for Trait 2: ',strBuf);
            extendStr(strBuf,mean[n2]+PIt2w1);
            writeln(fl,'Upper ',w1,'% Prediction interval for Trait 2: ',strBuf);
            extendStr(strBuf,mean[n2]-PIt2w2);
            writeln(fl,'Lower ',w2,'% Prediction interval for Trait 2: ',strBuf);
            extendStr(strBuf,mean[n2]+PIt2w2);
            writeln(fl,'Upper ',w2,'% Prediction interval for Trait 2: ',strBuf);
         End; {with stpk}
      End; {procedure write1PI}

{*************************************************************************}
{This procedure writes the statistics related to the single variable
 prediction interval estimates.                                            }
{*************************************************************************}
   Procedure dumpStat9B(Var dest : text;
                        stpk : statPak;
                        nd : nodePtr);
      Var
         first       : nodePtr;
         Vh          : extended;
         PIt1w1      : extended;
         PIt1w2      : extended;
         PIt2w1      : extended;
         PIt2w2      : extended;
         count       : integer;
         contempTips : boolean;
      Begin
         prepareTips(nd,count,first);
         meanHeight(nd,first,count,Vh,contempTips);
         dumpText4(Vh,contempTips);
         write(dest,'Mean Tip height is ');
         extendWrite(dest,Vh);
         writeln(dest);
         If contempTips
         Then writeln(dest,'Tips are contemporaneous.')
         Else writeln(dest,'Tips are not contemporaneous.');
         writeln(dest);
         calc1PI(nd,stpk,Vh,PIt1w1,PIt1w2,PIt2w1,PIt2w2);
         write1PI(stpk,Vh,PIt1w1,PIt1w2,PIt2w1,PIt2w2,dest);
      End; {procedure dumpStat9B}

{*************************************************************************}
{This procedure calculates statistics for the estimated root values and the
regression line (including the intercept).  The statistics are values and
confidence intervals.  PEM 29 Aug 1996.  Root C.I.'s are calculated using
equation A9 of Garland, Midford, and Ives (1999).  Intercept C.I.'s
are calculated according to equation A13 of Garland and Ives (2000).      }
{*************************************************************************}
Procedure statRoot(rt : nodePtr;
                   Var stpk : statPak;
                   conSlope : extended;
                   n1,n2 : integer);
   Var
      v1 : extended;
      v2 : extended;
      width : extended;
      se : traits;
   {**********************************************************************}
   {This procedure counts the number of nodes in the tree.}
   {**********************************************************************}
   Procedure getNum(nd:nodePtr;Var n:integer);
      Begin
         If nd^.sibL<>nil Then Begin
            inc(n);
            getNum(nd^.sibL,n);
            getNum(nd^.sibR,n);
         End; {then}
      End; {procedure getnum}
   {**********************************************************************}
   {This procedure calculates the sums required to compute variances and
    correlations among contrasts.}
   {**********************************************************************}
   Procedure getVar(nd:nodePtr; Var s:statPak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(nd^.contrast[1]);
            s.variance[2]:=s.variance[2]+sqr(nd^.contrast[2]);
            s.corr:=s.corr+(nd^.contrast[1])*(nd^.contrast[2]);
            getVar(nd^.sibL,s);
            getVar(nd^.sibR,s);
         End; {then}
      End; {procedure getvar}
   {*******************************************************************}
   {This (recursive) procedure is used by slopeinterCI to sum the      }
   {squares of the standardized contrasts.                             }
   {*******************************************************************}
   Procedure getSums(n1,n2:integer;
                     Var sx2 :extended;
                     Var sy2 :extended;
                     b :extended;
                     nd:nodePtr);
      Begin
         If (nd^.sibL<>nil) Then Begin
            sx2:=sx2+ sqr(nd^.contrast[n1]);
            sy2:=sy2+ sqr(nd^.contrast[n2]-b*nd^.contrast[n1]);
            getSums(n1,n2,sx2,sy2,b,nd^.sibL);
            getSums(n1,n2,sx2,sy2,b,nd^.sibR);
         End; {then}
      End; {procedure getsums}
   {**********************************************************************}
   {This function calculates the CI for the intercept.  Width specifies   }
   {the level of confidence (usually 0.90 or 0.95).  This code implements }
   {equations 14 (for slope CI) and 25 (for yintercept CI) from the       }
   {appendix to Garland, T., Jr., and A. R. Ives. 2000. Using the past to }
   {predict the present: Confidence intervals for regression equations in }
   {phylogenetic comparative methods. American Naturalist 155:346-364.    }
   {**********************************************************************}
   Procedure slopeInterCI(rt :nodePtr; Var s :statPak; width : extended);
      Var
         SEy : extended;
         SEb : extended;
         Xz : extended;
         v1 : extended;
         v2 : extended;
         ysum : extended;
         xsum : extended;
         sigma2 : extended;
         rawCI :extended;
      Begin
         v1 := rt^.sibL^.variance;
         v2 := rt^.sibR^.variance;
         ysum := 0;
         xsum := 0;
         getSums(s.n1,s.n2,xsum,ysum,s.ls,rt);
         If (xsum <> 0)
         Then Begin
            sigma2 := ysum/(s.number);   {Note that s.number = # of tips - 1}
            SEb := sqrt(sigma2*s.number/(xsum*(s.number-1)));
            s.slopeSE := SEb;
            Xz := stpk.mean[s.n1];       {Xz only, since Xi = 0}
            SEy := sqrt((sigma2*s.number/(s.number-1))*((sqr(Xz)/xsum)+(v1*v2)/(v1+v2)));
            s.interSE := SEy;
            rawCI := tCI(s.number-1-s.dfReduce,width);
            s.interCI := rawCI*SEy;
            s.slopeCI := rawCI*SEb;
         End {Then}
         Else Begin
            s.interCI := -99999;
            s.slopeCI := -99999;
         End { Else}
      End; {procedure interceptCI}
   {**********************************************************************}
   Begin {statroot}
      width1 := defaultWidth1;
      width2 := defaultWidth2;
      stpk.number:=0;
      stpk.n1 := n1;
      stpk.n2 := n2;
      stpk.mean[1]:=rt^.dat[1];
      stpk.mean[2]:=rt^.dat[2];
      getNum(rt,stpk.number);  { number of contrasts = number of tips - 1 }
      stpk.polytomies := 0;
      countPolytomies(rt,stpk.polytomies);
      reduceDF(rt,stpk);
      stpk.variance[1]:=0;
      stpk.variance[2]:=0;
      stpk.corr:=0;
      getVar(rt,stpk);
      stpk.variance[1] := stpk.variance[1]/(stpk.number);  {rate of evolution of trait 1}
      stpk.variance[2] := stpk.variance[2]/(stpk.number);  {rate of evolution of trait 2}
      v1 := rt^.sibL^.variance;
      v2 := rt^.sibR^.variance;
      stpk.se[1] := sqrt(stpk.variance[1]*((v1*v2)/(v1+v2)));
      stpk.se[2] := sqrt(stpk.variance[2]*((v1*v2)/(v1+v2)));
      width := tCI(stpk.number-stpk.dfreduce,defaultWidth1);
      stpk.ci[1] := width*stpk.se[1]; {CI for root values}
      stpk.ci[2] := width*stpk.se[2];
      stpk.ls := conSlope;
      stpk.lsy := stpk.mean[n2]-stpk.ls*stpk.mean[n1];
      slopeInterCI(rt,stpk,defaultwidth1); {CI for intercept}
   End; {procedure StatRoot}


{****************************************************************************}
{This procedure displays the columns of the .RSD (residual) file.  Screen 9D.}
{Moved from T_FELSEN.PAs                                                     }
{****************************************************************************}
Procedure showRSDFileColumns(name :FlString);
   Begin
      clrscr;
      writeln('Your "Residual" file (',name,') will contain the following columns:');
      writeln;
      writeln('     1.  Tip Name');
      writeln('     2.  X Value');
      writeln('     3.  Observed Y Value');
      writeln('     4.  Predicted Y Value (Yhat)');
      writeln('     5.  Residual (Y - Yhat)');
      writeln('     6.  Standardized Residual (using actual SD of residuals, not MSE)');
      writeln('     7.  Leverage of X Value (using root value as Xmean)');
      writeln('     8.  Studentized Residual (Neter et al., 1989, p. 398)');
      gotoxy(78,24);
      write('9D');
   End; {procedure showRSDFileColumns}


{*************************************************************************}
{This large procedure displays statistics calculated by statRoot (above)  }
{as well as writing the .RT, .CI, and .RSD files.  (Screens 9A,9B,9C,9D)  }
{*************************************************************************}
Procedure PlotRoot(nd:nodePtr;
                   fln : flString;
                   conSlope : extended;
                   conT1,conT2 : integer;
                   dfReduce : integer);
   Var
      rtFile : text;
      PI1File : text;
      first : nodePtr;
      stpk : statPak;
      minTip,maxTip : extended;
      x2Sum : extended;
      y2Sum : extended;
      sigma2y : extended;
      contrast2x : extended;
      slopeCI : extended;
      interCI : extended;
      residMean : extended;
      Vh : extended;
      PI11 : extended;
      PI12 : extended;
      PI21 : extended;
      PI22 : extended;
      mx : integer;
      my : integer;
      count : integer;
      { dotPos : integer; }
      flStem : flstring;
      CIflName : flstring;
      RTflName : flstring;
      PI1flName : flstring;
      RSDflName : flstring;
      saveStr : string;
      userChar : char;
      contempTips : boolean;
      Vhin9B : boolean;
   {***********************************************************************}
   {This procedure calculates the minimum and maximum tip values.}
   {***********************************************************************}
   Procedure minMaxTips(first : nodeptr;
                        count : integer;
                        index : integer;
                        Var minTip : extended;
                        Var maxTip : extended);
      Var
         current : nodeptr;
      Begin
         minTip := first^.dat[index];
         maxTip := first^.dat[index];
         current := first^.tipR;
         While (current <> nil) Do Begin
            If (current^.dat[index] < minTip)
            Then minTip := current^.dat[index];
            If (current^.dat[index] > maxTip)
            Then maxTip := current^.dat[index];
            current := current^.tipR;
         End; {while}
      End; {procedure minMaxTips}
   {************************************************************************}
   {This procedure calculates the sums of squares used in calcCI (below).   }
   {************************************************************************}
   Procedure getSums(n1,n2:integer;
                     Var sx2 :extended;
                     Var sy2 :extended;
                     b :extended;
                     nd:nodePtr);
      Begin
         If (nd^.sibL<>nil) Then Begin
            sx2:=sx2+ sqr(nd^.contrast[n1]);
            sy2:=sy2+ sqr(nd^.contrast[n2]-b*nd^.contrast[n1]);
            getSums(n1,n2,sx2,sy2,b,nd^.sibL);
            getSums(n1,n2,sx2,sy2,b,nd^.sibR);
         End; {then}
      End; {procedure getsums}
   {*******************************************************************}
   {This function calculates the CI for a particular point value.      }
   {Width specifies the level of confidence (usually 0.90 or 0.95).    }
   {These confidence intervals for regression points (new 9C) and      }
   {CI file.                                                           }
   {*******************************************************************}
   Function calcCI(rt : nodePtr;
                   s :statPak;
                   Xi :extended;
                   index1 :integer;
                   index2 :integer;
                   x2Sum :extended;
                   sigma2y :extended;
                   width :extended) : extended;
      Var
         SEy : extended;
         deltaX : extended;
         v1 : extended;
         v2 : extended;
         rawCI :extended;
      Begin
         v1 := rt^.sibL^.variance;
         v2 := rt^.sibR^.variance;
         deltaX := Xi-stpk.mean[index1];
         SEy := sqrt(sigma2y*((sqr(deltaX)/x2sum)+(v1*v2)/(v1+v2)));
         rawCI := tCI(s.number-1-s.dfReduce,width);
         calcCI := rawCI*SEy
      End; {procedure calcCI}
   {*******************************************************************}
   {This function calculates the PI for a particular point value.      }
   {Width specifies the level of confidence (usually 0.90 or 0.95).    }
   {These prediction intervals for regression points (new 9C) and      }
   {CI file.                                                           }
   {*******************************************************************}
   Function calcPI(rt :nodePtr;
                   s  :statPak;
                   Xi :extended;
                   index1 :integer;
                   index2 :integer;
                   x2Sum :extended;
                   sigma2y :extended;
                   width :extended;
                   Vh :extended) : extended;
      Var
         SEyNew :extended;
         deltaX :extended;
         v1 :extended;
         v2 :extended;
      Begin
         v1 := rt^.sibL^.variance;
         v2 := rt^.sibR^.variance;
         deltaX := Xi-stpk.mean[index1];
         SEyNew := sqrt(sigma2y*(Vh+(sqr(deltaX)/x2sum)+(v1*v2)/(v1+v2)));
         calcPI := tCI(s.number-1-s.dfReduce,width)*SEyNew
      End; {procedure calcPI}
   {*******************************************************************}
   {This procedure writes out the CI file.                             }
   {*******************************************************************}
   Procedure dumpCIFile(rt :nodePtr;
                        stpk :statPak;
                        first :nodePtr;
                        minTip : extended;
                        maxTip : extended;
                        index1 : integer;
                        index2 : integer;
                        x2Sum : extended;
                        sigma2y : extended;
                        Vh : extended;
                        name : flString);
      Var
         CIfile : text;
         current : nodePtr;
         xvalue : extended;
         yhat : extended;
         leverage : extended;
         i : integer;
         code : integer;
         numStr : string;

      {****************************************************************}
      { This procedure mostly exists to save space.                    }
      {****************************************************************}
      Procedure quickWrite(x : extended);
         Begin
            extendWrite(CIfile,x);
            write(CIFile,' ');
         End; {local procedure quickWrite}
      {****************************************************************}
      {This procedure the common (final) portion of each line in the   }
      {CI/PI file.                                                     }
      {****************************************************************}
      Procedure finishLine(xvalue,yhat,x2Sum,sigma2y,Vh :extended);
         Var
            tmp :extended; {holds the intermediate results}
         Begin
            tmp := calcCI(rt,stpk,xvalue,index1,index2,x2Sum,sigma2y,width1);
            quickWrite(yhat-tmp);
            quickWrite(yhat+tmp);
            tmp := calcPI(rt,stpk,xvalue,index1,index2,x2Sum,sigma2y,width1,Vh);
            quickWrite(yhat-tmp);
            quickWrite(yhat+tmp);
            tmp := calcCI(rt,stpk,xvalue,index1,index2,x2Sum,sigma2y,width2);
            quickWrite(yhat-tmp);
            quickWrite(yhat+tmp);
            tmp := calcPI(rt,stpk,xvalue,index1,index2,x2Sum,sigma2y,width2,Vh);
            quickWrite(yhat-tmp);
            quickWrite(yhat+tmp);
            writeln(CIfile);
         End; {procedure FinishLine}   
      {*********************************************************************}
      Begin {dumpCIfile}
         assign(CIfile,name);
         rewrite(CIfile);
         current := first;
         x2sum := 0;
         y2sum := 0;
         getsums(index1,index2,x2Sum,y2sum,stpk.ls,rt);
         sigma2y:=y2sum/(stpk.number-1);
         While (current <> nil) Do Begin {C/PI's for the tips (in tip order)}
            write(CIfile,current^.name,'  ');
            quickWrite(current^.dat[index1]);
            quickWrite(current^.dat[index2]);
            yhat := stpk.lsy+current^.dat[index1]*stpk.ls;
            quickWrite(yhat);
            FinishLine(current^.dat[index1],yhat,x2Sum,sigma2y,Vh);
            current := current^.tipR;
         End; {while}
         For i := 10 downto 1 Do Begin {C/PI's for 1% - 10% from largest}
            write(CIFile,'-',i,' ');
            If (i<10)
            Then write(CIFile,' '); 
            xvalue := minTip - minTip*i/100.0;
            extendWrite(CIfile,xvalue);
            write(CIfile,' -99999            ');
            yhat := stpk.lsy+xvalue*stpk.ls;
            quickWrite(yhat);
            FinishLine(xvalue,yhat,x2Sum,sigma2y,Vh);
         End; {for}
         For i := 1 to 10 Do Begin {C/PI's for -1% - -10% from smallest}
            write(CIFile,'+',i,' ');
            If (i<10)
            Then write(CIFile,' '); 
            xvalue := maxTip + maxTip*i/100.0;
            extendWrite(CIfile,xvalue);
            write(CIfile,' -99999            ');
            yhat := stpk.lsy+xvalue*stpk.ls;
            quickWrite(yhat);
            FinishLine(xvalue,yhat,x2Sum,sigma2y,Vh);
         End; {for}
         write(CIFile,'00  ');
         xvalue := 0;             {then the C/PI's at the intercept}
         extendWrite(CIfile,xvalue);
         write(CIfile,' -99999            ');
         yhat := stpk.lsy+xvalue*stpk.ls;
         quickWrite(yhat);
         FinishLine(xvalue,yhat,x2Sum,sigma2y,Vh);
         gotoxy(1,20);               {finally, optional user vals}
         writeln('Do you want to add particular values to the CI file?');
         write('       (Y or anything else to not add values) ');
         readln(userChar);
         If ((userChar = 'Y') or (userChar = 'y')) Then Begin
            clrscr;
            Repeat
               write('Enter an X value (or press enter to quit): ');
               readln(numstr);
               Val(numstr,xvalue,code);
               If (code = 0) Then Begin
                  write(CIFile,'usr ');
                  extendWrite(CIfile,xvalue);
                  write(CIfile,' -99999            ');
                  yhat := stpk.lsy+xvalue*stpk.ls;
                  quickWrite(yhat);
                  FinishLine(xvalue,yhat,x2Sum,sigma2y,Vh);
               End; {then}
            Until (length(numstr) = 0);
         End; {then}
         close(CIfile);
      End; {dumpCIFile}
   {************************************************************************}
   {This calculates some simple statistics on residuals.  These statistics  }
   {can be used to normalize residuals.                                     }
   {************************************************************************}
   Procedure residualStats(first : nodePtr;
                           s : statPak;
                           Var residMean :extended;
                           Var residSD : extended);
      Var r : extended;   {residudual corresponding to each tip}
          rsum : extended;
          r2sum : extended;
          current : nodePtr;
          count : integer;
      Begin
         count := 0;
         rsum := 0;
         r2sum := 0;
         current := first;
         While (current <> nil) Do Begin
            r := current^.dat[s.n2]-(s.lsy+current^.dat[s.n1]*s.ls);
            inc(count);
            rsum := rsum+r;
            r2sum := r2sum+sqr(r);
            current := current^.tipR;
         End; {while}
         { writeln('Sum of Residuals is: ',rsum); }
         residMean := rsum/count;
         residSD := sqrt((r2sum-sqr(rsum)/count)/(count-1));
         { write('Resid SD is: ',residSD); }
      End; {procedure residualStats}

   {********************************************************************}
   {This calculates sums of the independent tip variable and its square.}
   {********************************************************************}
   Procedure tipSums(first : nodePtr;
                     index1 : integer;
                     Var xTipSum : extended;
                     Var xTip2Sum : extended);
      Var current : nodePtr;
      Begin
         current := first;
         While (current <> NIL) Do Begin
            xTipSum := xTipSum + current^.dat[index1];
            xTip2Sum := xTip2Sum + sqr(current^.dat[index1]);
            current := current^.tipR;
         End; {while}
      End; {procedure tipsums}

   {********************************************************************}
   {This procedure writes the RSD (residual) file.                      }
   {********************************************************************}
   Procedure dumpRSDFile(rt : nodePtr;
                         first : nodePtr;
                         stpk :statPak;
                         index1 : integer;
                         index2 : integer;
                         RSDFlName :flString);
      Var
         RSDFile : text;
         current : nodeptr;
         x2sum : extended;
         y2sum : extended;
         tipSum : extended;
         tip2Sum : extended;
         residSD : extended;
         yhat : extended;
         leverage : extended;
         rootMean : extended;
         rawResid : extended;
      {*****************************************************************}
      {This procedure mostly exists to save space.                      }
      {*****************************************************************}
      Procedure quickWrite(x : extended);
         Begin
            extendWrite(RSDFile,x);
            write(RSDFile,' ');
         End; {local procedure quickWrite}
      {*****************************************************************}
      Begin {dumpRSDFile}
         assign(RSDFile,RSDFlName);
         rewrite(RSDFile);
         current := first;
         x2sum := 0;
         y2sum := 0;
         getSums(index1,index2,x2Sum,y2sum,stpk.ls,rt);
         residualStats(first,stpk,residMean,residSD);
         rootMean := stpk.mean[index1];
         tip2Sum := 0;
         tipSum := 0;
         tipSums(first,index1,tipSum,tip2Sum);
         While (current <> nil) Do Begin {Residuals for the tips (in tip order)}
            write(RSDFile,current^.name,'  ');
            quickWrite(current^.dat[index1]);
            quickWrite(current^.dat[index2]);
            yhat := stpk.lsy+current^.dat[index1]*stpk.ls;
            quickWrite(yhat);
            rawResid := current^.dat[index2]-yhat; 
            quickWrite(rawResid);
            If (residSD <> 0)
            Then extendwrite(RSDFile,(rawResid-residMean)/residSD)
            Else write(RSDFile,'-9               ');
            write(RSDFile,' ');
            If (tip2Sum <> sqr(rootmean)*(stpk.number+1))
            Then Begin
               leverage :=
                 1/(stpk.number+1) +
                    sqr(current^.dat[index1]-rootMean)/(tip2Sum-sqr(rootmean)*(stpk.number+1));
               extendWrite(RSDFile,leverage)
            End {then}
            Else write(RSDFile,'-9              ');
            write(RSDFile,' ');            {Standardized residual}
            If ((residSD <> 0) And (leverage<1))
            Then extendWrite(RSDFile,(rawResid-residMean)/sqrt(sqr(residSD)*(1-leverage)))
            Else write(RSDFile,'-9               ');
            writeln(RSDFile);
            current := current^.tipR;
         End; {while}
         close(RSDFile);
      End; {procedure dumpRSDFile}
   {*********************************************************************}
   Begin {RootPlot}
      width1 := defaultWidth1;
      width2 := defaultWidth2;
      statRoot(nd,stpk,conSlope,conT1,conT2);
      clrscr;
      gotoXY(78,24);
      write('9A');
      gotoXY(1,1);
      flStem := copy(fln,1,pos('.',fln));
      RTflname := concat(flStem,'RT');
      saveStr := 'INDEPENDENT CONTRASTS STATISTICS (optionally saved as ' +
                   RTflname + ')';
      writeln(saveStr);
      dumpStat9A(output,stpk);
      writeln('Would you like to save the above information to a file?');
      write('(Enter Y to save, anything else to not save): ');
      readln(userChar);
      If ((userChar = 'Y') Or (userChar = 'y'))
      Then Begin
         writeln;
         writeln('Screen data will be saved to ',RTflname);
         assign(rtfile,RTflname);
         rewrite(rtfile);
         writeln(rtfile,saveStr);
         dumpStat9A(rtFile,stpk);
         close(rtFile);
      End; {then}
      prepareTips(nd,count,first);
      minMaxTips(first,count,conT1,minTip,maxTip);
      clrscr;
      writeln('A univariate prediction interval can be calculated for a new species added');
      writeln('on a branch extending from the root.  If you first reroot the tree');
      writeln('either at a node or anywhere along a branch, as described in Garland et al.');
      writeln('(1999) and Garland and Ives (2000), then you can place the branch leading to');
      writeln('the species to be predicted anywhere in the tree.  The predicted trait value');
      writeln('will equal the reconstructed root value, as shown on Screen 9A, so only the');
      writeln('prediction intervals will be displayed for each the two traits in your file.');
      writeln('These prediction intervals will vary in width depending on the height of the');
      writeln('branch (designated Vh) leading to the new tip, which you will be asked to');
      writeln('specify.');
      writeln;
      writeln('Would you like to perform this calculation? ');
      write('(Y to calculate, anything else to skip) ');
      readln(userChar);
      writeln;
      {new 9B goes here}
      If ((userChar = 'Y') or (userChar = 'y')) Then Begin
         {Vhin9B := true;}
         If (dfReduce > 0) Then Begin
            writeln;
            writeln('As you requested, ',stpk.dfReduce,' degrees of freedom will be removed');
            writeln('in the calculation of the regression confidence and prediction intervals.');
         End; {then}
         meanHeight(nd,first,count,Vh,contempTips);
         dumpText4(Vh,contempTips);
         clrscr;
         PI1flname := concat(flStem,'1PI');
         saveStr := 'Single variable confidence intervals ' +
                     '(optionally saved as ' + PI1flname + ')';
         writeln(saveStr);
         calc1PI(nd,stpk,Vh,PI11,PI12,PI21,PI22);
         write1PI(stpk,Vh,PI11,PI12,PI21,PI22,output);
         gotoxy(78,24);
         write('9B');
         gotoxy(1,20);
         writeln('Would you like to save the above information to a file?');
         write('(Enter Y to save, anything else to not save): ');
         readln(userChar);
         If ((userChar = 'Y') Or (userChar = 'y'))
         Then Begin
            writeln;
            writeln('Screen data will be saved to ',PI1flname);
            assign(PI1file,PI1flname);
            rewrite(PI1file);
            writeln(PI1file,saveStr);
            write1PI(stpk,Vh,PI11,PI12,PI21,PI22,PI1File);
            close(PI1File);
         End; {then}
         writeln('Press the enter key to continue.');
         readln;
      End; {if userchar}
      Vhin9B := false;
      clrscr;
      gotoXY(78,24);
      write('9C');
      gotoXY(1,1);
      writeln('The regression equation listed on a previous screen (9A), estimated by');
      writeln('independent contrasts, can be mapped back onto the original data space.');
      writeln('It can also be used to compute confidence and prediction intervals in the');
      writeln('original data space (Garland and Ives, 2000).  These are convenient for');
      writeln('making heuristic graphs.');
      If (dfReduce > 0) Then Begin
         writeln;
         writeln('As you requested, ',stpk.dfReduce,' degrees of freedom will be removed');
         writeln('in the calculation of the regression confidence and prediction intervals.');
      End; {then}
      writeln;
      writeln('Would you like to save regression confidence and prediction interval');
      write('data to a file? ');
      write('(Y to save, anything else to not save): ');
      readln(userChar);
      writeln;
      If ((userChar = 'Y') or (userChar = 'y')) Then Begin
         If not(Vhin9B)
         Then meanHeight(nd,first,count,Vh,contempTips);
         dumpText4(Vh,contempTips);
         writeln('The current (2-tailed) P-value widths are: 0.',
                  round(100*width1),
                  ' and 0.',
                  round(100*width2));
         write('Do you want to change these widths? (Y or anything else to not change) ');
         readln(userChar);
         If ((userChar = 'Y') or (userChar = 'y'))
         Then Begin
            write('Enter the first P-value width: ');
            readln(width1);
            writeln;
            write('Enter the second P-value width: ');
            readln(width2);
         End {then}
         Else Begin
            width1 := defaultWidth1;
            width2 := defaultWidth2;
         End; {else}
         CIflname := concat(flStem,'CI');
         showCIFileColumns(CIflName);
         dumpCIFile(nd,
                    stpk,
                    first,
                    minTip,
                    maxTip,
                    conT1,
                    conT2,
                    x2Sum,
                    sigma2y,
                    Vh,
                    CIflName);
      End; {then}
      clrscr;
      writeln('The foregoing regression equation, estimated by independent contrasts but');
      writeln('mapped back onto the original data space, can also be used to compute');
      writeln('"residuals" (vertical deviations).  These "residuals" are NOT free of');
      writeln('phylogenetic effects and cannot be used in conventional statistical');
      writeln('procedures, but they may be useful for making heuristic graphs.');
      writeln;
      writeln('Warning: This output file is still a test version.  Use only with Caution!');
      writeln;
      writeln('Would you like to save regression "residual" data to a file?');
      write('(Enter Y to save, anything else to not save): ');
      readln(userChar);
      If ((userChar = 'Y') or (userChar = 'y')) Then Begin
         writeln;
         writeln('Writing data to disk...');
         RSDFlName := concat(flStem,'RSD');
         dumpRSDFile(nd,first,stpk,conT1,conT2,RSDFlName);
         showRSDFileColumns(RSDFlName);
         gotoxy(1,20);
         writeln('Press any key to continue.');
         userChar := readkey;
      End; {then}
   End; {procedure RootPlot}
{**************************************************************************}
{This recursive procedure is passed a node, and the least squares parameters.
It then fills in the temp field with the residual.}
{**************************************************************************}
Procedure tempResidual(nd:nodePtr; m,y0:extended;t,t2:integer);
   Begin
      If nd^.sibL<>nil Then Begin
         nd^.temp:=nd^.contrast[t2]-(m*nd^.contrast[t]+y0);
         tempResidual(nd^.sibL,m,y0,t,t2);
         tempResidual(nd^.sibR,m,y0,t,t2);
      End; {then}
   End; {procedure tempResidual} 
{**************************************************************************}
{This recursive procedure is passed a node, and the least squares parameters.
It then fills in the temp field with the absolute value of the residual.
(PEM 13-May-1994) 
{**************************************************************************}
Procedure tempAbsResidual(nd:nodePtr; m,y0:extended;t,t2:integer);
   Begin
      If nd^.sibL<>nil Then Begin
         nd^.temp:=abs(nd^.contrast[t2]-(m*nd^.contrast[t]+y0));
         tempAbsResidual(nd^.sibL,m,y0,t,t2);
         tempAbsResidual(nd^.sibR,m,y0,t,t2);
      End; {then}
   End; {procedure tempAbsResidual} 

{*************************************************************************}
{This procedure dumps the statistics given to it to the graphics screen.
It is used to add the statistics output to the graphs, and it writes
the word 'zero' in for mean of Y and the intercepts.}
{*************************************************************************}  
  Procedure dumpstat11(stpk:statpak;mx,my:integer; var bail : boolean);
     Var
        t:string;
        ch:char;
     {********************************************************************}
     {This simple procedure writes a line of text to the graphics screen.
     {********************************************************************}
     Procedure showLine(line: integer;strtext : string);
        Begin
           outtextxy(0,line*oneline,strtext);
        End; {procedure showLine}
     {********************************************************************}
     {This procedure writes an extended value to the graphics screen.  It
      sort of combines the functionality of showLine (above) and
      extendWrite.                                                        }
     {********************************************************************}
     Procedure showValue(line: integer; dataLabel : string;num : extended);
        Begin
           extendStr(t,num);
           outtextxy(0,line*oneline,datalabel+t);
        End; {procedure showValue}
     {********************************************************************}
    Begin
       str(stpk.number,t);
       showLine(9,'Number of contrasts: '+t);
       showValue(11,'Min X:',stpk.min[stpk.n1]);
       showValue(12,'Min Y:',stpk.min[stpk.n2]);
       showValue(14,'Max X:',stpk.max[stpk.n1]);
       showValue(15,'Max Y:',stpk.max[stpk.n2]);
       showValue(17,'Mean X:',stpk.mean[stpk.n1]);
       showValue(18,'Mean Y:',stpk.mean[stpk.n2]);
       showValue(20,'Var X:',stpk.variance[stpk.n1]);
       showValue(21,'Var Y:',stpk.variance[stpk.n2]);
       showValue(23,'Cov:',stpk.cov);
       showValue(24,'Corr:',stpk.corr);
       showLine(26,'Reduced Major Axis.');
       If (stpk.lsregok) Then Begin
          showValue(27,'     Slope:',stpk.rma);
          showValue(28,' Intercept:',stpk.rmay);
       End {then}
       Else Begin
          showLine(27,'     Slope: undefined');
          showLine(28,' Intercept: undefined');
       End; {else}
       showLine(30,'Major Axis.');
       showValue(31,'     Slope:',stpk.ma);
       showValue(32,' Intercept:',stpk.may);
       showLine(34,'Least Squares Regression.');
       If (stpk.lsregok) Then Begin
          showValue(35,'     Slope:',stpk.ls);
          showValue(36,' Intercept:',stpk.lsy);
          showValue(37,' R squared:',stpk.r2);
       End
       Else Begin   
          showLine(35,'     Slope: undefined');
          showLine(36,' Intercept: undefined');
          showLine(37,' R squared: undefined');
       End; {else}
       If (abs(stpk.corr)<>1) Then Begin
          showValue(39,'t: ',stpk.ts);
          showValue(40,'F: ',stpk.f);
       End {then}
       Else Begin
          showLine(39,'t: undefined');
          showLine(40,'F: undefined');
       End; {else}        
       str(stpk.df,t);
       showLine(41,'d.f.: '+t);
       ch:=readkey;
       If ((ch = 'R') or (ch = 'r'))
       Then bail := true
       Else bail := false;
       Closegraph;
    End; {procedure dumpstat11}

{***************************************************************************}
{This procedure computes the statistics comparing the residuals from a      }
{regression of the two traits against the height of the node.  The X axis   }
{trait is specified by stpk's n1 value and was previously chosen by the     }
{user.  These statistics will be displayed by plotresidualheight on         }
{Screen 11.                                                                 }
{***************************************************************************}
Procedure statResidualHeight(nd:nodePtr;Var stpk:statPak);
   Var
      sx,sy,sx2,sxy,d:extended;
   {************************************************************************}
   {This recursive procedure gets the max and min values for the data.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getBounds(Var min,max:traits; nd:nodePtr);
      Begin
         If nd^.sibL<>nil Then Begin
            If nd^.temp3>max[1] Then max[1]:=nd^.temp3;
            If nd^.temp3<min[1] Then min[1]:=nd^.temp3;
            If nd^.temp>max[2] Then max[2]:=nd^.temp;
            If nd^.temp<min[2] Then min[2]:=nd^.temp;
            getBounds(min,max,nd^.sibL);
            getBounds(min,max,nd^.sibR);
         End; {then}
      End; {procedure getbounds}      
   {************************************************************************}
   {This recursive procedure does the first step in finding the mean values
   for the data.  It finds the sums and the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getMean(nd:nodeptr;Var n:integer;Var m: traits);
      Begin
         If nd^.sibL<>nil Then Begin 
            inc(n);
            m[1]:=m[1]+nd^.temp3;
            m[2]:=m[2]+nd^.temp;
            getMean(nd^.sibL,n,m);
            getMean(nd^.sibR,n,m);
         End; {then}
      End; {procedure getmean}
        
   {************************************************************************}
   {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getVar(nd:nodeptr;Var s:statPak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(s.mean[1]-nd^.temp3);
            s.variance[2]:=s.variance[2]+sqr(s.mean[2]-nd^.temp);
            s.corr:=s.corr+(s.mean[1]-nd^.temp3)*(s.mean[2]-nd^.temp);
            getVar(nd^.sibL,s);
            getVar(nd^.sibR,s);
         End; {then}
      End; {procedure getvar}
   {************************************************************************}
   {This recursive procedure does the first step in finding the least squares
   regression for the data.  It finds the sums of squares and products.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getSums(Var sx,sy,sx2,sxy:extended; nd:nodePtr);
      Begin
         If nd^.sibL<>nil Then Begin
            sx:=sx+nd^.temp3;
            sx2:=sx2+sqr(nd^.temp3);
            sy:=sy+nd^.temp;
            sxy:=sxy+nd^.temp3*nd^.temp;
            getSums(sx,sy,sx2,sxy,nd^.sibL);
            getSums(sx,sy,sx2,sxy,nd^.sibR);
         End; {then}
      End; {procedure getsums}
   {**************************************************************************}
   Begin {statresidualheight}
      stpk.n1:=1;
      stpk.n2:=2;
      stpk.max[1]:=nd^.temp3;
      stpk.max[2]:=nd^.temp;
      stpk.min[1]:=stpk.max[1];
      stpk.min[2]:=stpk.max[2];
      getBounds(stpk.min,stpk.max,nd);  
      stpk.number:=0;
      stpk.mean[1]:=0;
      stpk.mean[2]:=0;
      getMean(nd,stpk.number,stpk.mean);
      stpk.mean[1]:=stpk.mean[1]/stpk.number;
      stpk.mean[2]:=stpk.mean[2]/stpk.number;
      stpk.variance[1]:=0;
      stpk.variance[2]:=0;
      stpk.corr:=0;
      getVar(nd,stpk);
      stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);
      stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);
      stpk.corr:=stpk.corr/(stpk.number-1);
      stpk.cov:=stpk.corr;
      If(stpk.variance[1]>0) and (stpk.variance[2]>0)
      Then stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr:=0;
      sx:=0;
      sy:=0;
      sx2:=0;
      sxy:=0;
      getSums(sx,sy,sx2,sxy,nd);
      If ((stpk.number*sx2-sqr(sx))<>0)
      Then Begin
         stpk.lsregok:=true;
         d:=1/(stpk.number*sx2-sqr(sx));
      End {then}
      Else Begin
         stpk.lsregok:=false;
         d:=0;
      End; {else}
      stpk.ls:=d*(stpk.number*sxy-sx*sy);
      stpk.lsy:=d*(sx2*sy-sx*sxy);
      finishStat(stpk);
   End; {procedure statresidualheight}
{***************************************************************************}
{This procedure sets up the graphics screen and plots out the data points.
It uses statresidualheight to do the statistics and also calls several
procedures from the T_UTIL1 unit. (Screen 11)}
{***************************************************************************}
Procedure plotResidualHeight(rt : nodePtr;
                             fl : flString;
                             spNode : nodePtr;
                             dfReduce : integer;
                             clade : boolean;
                             Var bail : boolean);
   Var
      scalingFactor:traits;
      mx,my,temp:integer;
      stpk:statPak; 
      origin:itraits;
      s:string[10];
      st:string[80]; 
   {************************************************************************}
   {This recursive procedure plots data points onto the screen.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure dumpData(min:traits;
                      orgn:itraits;
                      sf:traits;
                      nd:nodePtr;
                      special : boolean);
      Var
         coor:itraits;
      Begin
         If(nd^.sibL<>nil) Then Begin
            coor[1]:=orgn[1]+round(sf[1]*(nd^.temp3-min[1]));
            coor[2]:=orgn[2]-round(sf[2]*(nd^.temp-min[2]));
            If (nd = rt) Then xSymbol(coor)
            Else
               If ((nd = spNode) or (special = true))
               Then openCircle(coor)
               Else star(coor);
            If ((nd = spNode) and clade) Then Begin
               dumpData(min,orgn,sf,nd^.sibL,true);
               dumpData(min,orgn,sf,nd^.sibR,true);
            End {then}
            Else Begin
               dumpData(min,orgn,sf,nd^.sibL,special);
               dumpData(min,orgn,sf,nd^.sibR,special);
            End;
         End; {then}
      End; {procedure dumpdata}
   {*************************************************************************}
   Begin {plotresidualheight}
      stpk.dfReduce := dfReduce;  {new statpak, so copy reduced d.f.}
      statResidualHeight(rt,stpk);
      graphicsOn(mx,my);
      initPlotx0(1,2,stpk.min,stpk.max,fl,scalingfactor,origin,mx,my);
      outTextXY(0,0,'Plot of absolute value of');
      outTextXY(0,oneline,'residuals versus height of');
      outTextXY(0,2*oneline,'their X-variable base nodes.');
      outTextXY(0,4*oneline,'d.f. may be fewer if tree');
      outTextXY(0,5*oneline,'   contains polytomies.');
      st:='Absolute Value of Residual';
      temp:=round((origin[2]-length(st)*oneline)/2);
      If temp<1
      Then temp:=1;
      dumpDown(round(0.9524*origin[1])-8,temp,st);
      st:='Corrected height of base node';
      temp:=round((mx-origin[1]-textwidth(st))/2)+origin[1];
      If temp< origin[1]+1
      Then temp:=origin[1]+1;
      outTextXY(temp,round(origin[2])+3,st);
      setLineStyle(0,0,1);
      outTextXY(mx-2*oneline,my-oneline,'11');
      dumpData(stpk.min,origin,scalingfactor,rt,false);
      dumpStat11(stpk,mx,my,bail);
   End; {procedure plotresidualheight}

{**************************************************************************}
{This procedure traverses the tree and writes out one line of numbers per
node.  They are preceded by the daughter node names, and the current node
name.  The numbers are: the trait 1 unstandardized contrast, the trait 2
unstandardized contrast, the standard deviation calculated in the
computecon procedure above, the trait 1 predicted value, the trait 2
predicted value, the node height, and the average height of the daughter
nodes.
{**************************************************************************}
Procedure fileContrasts(nd:nodePtr;Var fl:text);
   {***********************************************************************}
   { This procedure mostly exists to save space.                           }
   {***********************************************************************}
   Procedure quickWrite(x : extended);
      Begin
         extendWrite(fl,x);
         write(fl,' ');
      End; {local procedure quickWrite}
   {***********************************************************************}
   Begin
      If (nd^.sibL<>nil) Then Begin      
         write(fl,nd^.sibL^.name,'-',nd^.sibR^.name,'    ',nd^.name,'   ');
         quickWrite(nd^.contrast[1]*nd^.sdcon);
         quickWrite(nd^.contrast[2]*nd^.sdcon);
         quickWrite(nd^.sdcon);
         quickWrite(nd^.dat[1]);
         quickWrite(nd^.dat[2]);
         quickWrite(nd^.temp2);                       {raw height}
         quickWrite(nd^.temp3);                       {corrected height}
         quickWrite(nd^.temp);                        {raw daughter height}
         quickWrite(nd^.temp4);                       {corrected daughter ht}
         write(fl,nd^.sibL^.name,'-',nd^.name,'    ');
         extendWrite(fl,nd^.sibL^.variance);
         write(fl,'   ');
         write(fl,nd^.sibR^.name,'-',nd^.name,'    ');
         extendWrite(fl,nd^.sibR^.variance);
         writeln(fl);
         fileContrasts(nd^.sibL,fl);
         fileContrasts(nd^.sibR,fl)
      End {then begin}
   End; {procedure filecontrasts}

{**************************************************************************}
{This procedure checks that neither of the traits is constant at the tips. }
{This condition makes it impossible to calculate confidence intervals on   }
{screens 9A-9C and can cause slope calculations to blow up as well.        }
{**************************************************************************}
Procedure checkConstantTips(rt:nodePtr;
                            Var index : integer);
   Var first : nodePtr;
       curNode : nodePtr;
       checkConstant : traits;
       count : integer;
       oneEqual : boolean;
       twoEqual : boolean;
   Begin
      prepareTips(rt,count,first);
      curNode := first;
      oneEqual := true;
      twoEqual := true;
      checkConstant := first^.dat;
      While ((curNode^.tipR <> nil) And (oneEqual Or TwoEqual)) Do Begin
         curNode := curNode^.tipR;
         If curNode^.dat[1] <> checkConstant[1]
         Then oneEqual := false;
         If curNode^.dat[2] <> checkConstant[2]
         Then twoEqual := false;
      End; {while}
      If oneEqual
      Then index := 1
      Else If twoEqual Then index := 2 Else index := 0;
   End; {procedure checkconstanttips}

{**************************************************************************}
{This procedure explains why we are skipping plot 11 }
{**************************************************************************}
Procedure explainSkip;
   Begin
      writeln('The height of the internal nodes shows no variation.');
      writeln;
      writeln;
      writeln('Your tree must be a star!');
      writeln;
      writeln;
      writeln('Therefore, plot 11 will be skipped.');
      writeln;
      writeln;
      ret('Please press any key to continue.');
   End; {procedure explainskip}

{**************************************************************************}
{This procedure writes text to the screen that explains that a branch of
 the tree is longer than one billion.                                      }
{**************************************************************************}
{Procedure explainSkip2;
   Begin
      writeln('Your tree has some branch lengths greater than 1,000,000,000.');
      writeln('This is too large for some of the procedures.');
      writeln;
      writeln;
      writeln('Therefore, plot 11 will be skipped.');
      writeln;
      writeln;
      writeln('One way to avoid this problem is to rescale the total height ',
                 'of the tree.');
      ret('Please press any key to continue.');
   End;} {explainskip2}  {commented out 14 Dec 1999 PEM; we shouldn't
                          need this particular check anymore, if numbers
                          get too big, extendWrite will catch them, and
                          it won't complain until 9E99.}
{***************************************************************************}
{This procedure writes a screen message when constant tips are found.       }
{***************************************************************************}
Procedure explainSkip3(index : integer);
   Begin
      writeln('The tip values for trait ',index,' of your tree is constant.');
      writeln('This is makes it impossible to calculate confidence intervals');
      writeln('and may cause problems with slope calculations performed on');
      writeln('tip data.');  
      writeln;
      writeln; 
      writeln('Therefore, screens 9A,9B,9C,10, and 11 will be skipped.');
      writeln;
      writeln;
      ret('Please press any key to continue.');
   End;

{*************************************************************************}
{This procedure dumps the statistics given to it to the graphics screen.
It is used to add the statistics output to the graphs.  It also writes the
word 'zero' in for the means, and omits the RMA, MA, least squares reg.,
t, F, and d.f.  However, it does add the sum of the residuals.}
{*************************************************************************}  
   Procedure dumpstat10(stpk:statpak;mx,my:integer; var bail : boolean);
      Var
         t:string;
         ch:char;
     {********************************************************************}
     {This simple procedure writes a line of text to the graphics screen.
     {********************************************************************}
     Procedure showLine(line: integer;strtext : string);
        Begin
           outtextxy(0,line*oneline,strtext);
        End; {procedure showLine}
     {********************************************************************}
     {This procedure writes an extended value to the graphics screen.  It
      sort of combines the functionality of showLine (above) and
      extendWrite.                                                        }
     {********************************************************************}
     Procedure showValue(line: integer; dataLabel : string;num : extended);
        Begin
           extendStr(t,num);
           outtextxy(0,line*oneline,datalabel+t);
        End; {procedure showValue}
     {********************************************************************}
     Begin
        str(stpk.number,t);
        showLine(9,'Number of contrasts: '+t);
        showValue(11,'Min X:',stpk.min[stpk.n1]);
        showValue(12,'Min Y:',stpk.min[stpk.n2]);
        showValue(14,'Max X:',stpk.max[stpk.n1]);
        showValue(15,'Max Y:',stpk.max[stpk.n2]);
        showLine(17,'Mean X: zero');
        showValue(18,'Mean Y:',stpk.mean[2]);
        showValue(20,'Var X:',stpk.variance[stpk.n1]);
        showValue(21,'Var Y:',stpk.variance[stpk.n2]);
        showValue(23,'Cov:',stpk.cov);
        showValue(24,'Corr:',stpk.corr);
        showLine(26,'Covariance and correlation');
        showLine(27,'are computed through X');
        showLine(28,'origin, but not through Y');
        showLine(29,'origin.');
        showLine(31,'Note that residuals do not');
        showLine(32,'necessarily sum to zero');
        showLine(33,'because regressions of');
        showLine(34,'contrasts are computed');
        showLine(35,'through the origin');
        showLine(36,'(e.g., screen 9).');
        ch:=readkey;
        If ((ch = 'R') or (ch = 'r'))
        Then bail := true
        Else bail := false;
        Closegraph;
     End; {procedure dumpstat10}

{***************************************************************************}
{This procedure computes the statistics comparing residual values to one of
the trait values.  The trait is selected by the argument t.} 
{***************************************************************************}
Procedure statresidualX(nd:nodePtr;t:integer;Var stpk:statPak);
   {************************************************************************}
   {This recursive procedure gets the max and min values for the data.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getBounds(Var min,max:traits; nd:nodePtr;t:integer);
      Begin
         If nd^.sibL<>nil Then Begin
            If nd^.contrast[t]>max[1] Then max[1]:=nd^.contrast[t];
            If nd^.contrast[t]<min[1] Then min[1]:=nd^.contrast[t];
            If nd^.temp>max[2] Then max[2]:=nd^.temp;
            If nd^.temp<min[2] Then min[2]:=nd^.temp;
            getBounds(min,max,nd^.sibL,t);
            getBounds(min,max,nd^.sibR,t);
         End; {then}
      End; {procedure getbounds}     
   {************************************************************************}
   {This recursive procedure finds the number of data points.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getNum(nd:nodeptr;Var n:integer);
      Begin
         If nd^.sibL<>nil Then Begin 
            inc(n);
            getNum(nd^.sibL,n);
            getNum(nd^.sibR,n);
         End; {then}
      End; {procedure getnum}
   {************************************************************************}
   {This recursive procedure finds the mean of the second trait.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getMean(nd:nodeptr;Var sum :extended);
      Begin
         If nd^.sibL<>nil Then Begin 
            sum:=sum+nd^.temp;
            getMean(nd^.sibL,sum);
            getMean(nd^.sibR,sum);
         End; {then}
      End; {procedure getnum}
        
   {************************************************************************}
   {This recursive procedure does the first step in finding the variance and
   covariance for the data.  It finds the sums of the deviations.
   Notice that it is local to a parent procedure, and therefore, can coexist
   in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure getVar(t:integer;nd:nodePtr;Var s:statPak);
      Begin
         If nd^.sibL<>nil Then Begin
            s.variance[1]:=s.variance[1]+sqr(s.mean[1]-nd^.contrast[t]);
            s.variance[2]:=s.variance[2]+sqr(s.mean[2]-nd^.temp);
            s.corr:=s.corr+(s.mean[1]-nd^.contrast[t])*(s.mean[2]-nd^.temp);
            getVar(t,nd^.sibL,s);
            getVar(t,nd^.sibR,s);
         End; {then}
      End; {procedure getvar}
   {**************************************************************************}
   Begin
      stpk.n1:=1;
      stpk.n2:=2;
      stpk.max[1]:=nd^.contrast[t];
      stpk.max[2]:=nd^.temp;
      stpk.min[1]:=stpk.max[1];
      stpk.min[2]:=stpk.max[2];
      getBounds(stpk.min,stpk.max,nd,t);  
      stpk.number:=0;
      getNum(nd,stpk.number);
      stpk.mean[1]:=0;
      stpk.mean[2]:=0;
      getMean(nd,stpk.mean[2]);
      stpk.mean[2] :=stpk.mean[2]/stpk.number;
      stpk.variance[1]:=0;
      stpk.variance[2]:=0;
      stpk.corr:=0;
      getVar(t,nd,stpk);
      stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);
      stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);
      stpk.corr:=stpk.corr/(stpk.number-1);
      stpk.cov:=stpk.corr;
      If(stpk.variance[1]>0) and (stpk.variance[2]>0)
      Then stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
      Else stpk.corr:=0;
   End; {procedure statresidualX}

{***************************************************************************}
{This procedure sets up the graphics screen and plots out the data points.
It uses the statistics calculating procedure above, and several procedures
from the T_UTIL1 unit. (Screen 10)                                          } 
{***************************************************************************}
Procedure plotresidualX(rt:nodePtr;
                        t:integer;
                        fl:flString;
                        spNode : nodePtr;
                        clade : boolean;
                        Var bail:boolean);
   Var
      scalingFactor:traits;
      mx,my,temp:integer;
      stpk:statPak;
      origin:itraits;
      s:string[10];
      st:string[80];
   {************************************************************************}
   {This recursive procedure plots data points onto the screen.
    Notice that it is local to a parent procedure, and therefore, can coexist
    coexist in a file with other procedures by the same name.}
   {************************************************************************}
   Procedure dumpData(min:traits;
                      orgn:itraits;
                      sf:traits;
                      nd:nodePtr;
                      t:integer;
                      special : boolean);
      Var
         coor:itraits;
      Begin
         If(nd^.sibL<>nil) Then Begin
            coor[1]:=orgn[1]+round(sf[1]*(nd^.contrast[t]-min[1]));
            coor[2]:=orgn[2]-round(sf[2]*(nd^.temp-min[2]));
            If (nd = rt) Then xSymbol(coor)
            Else
               If ((nd = spNode) or (special = true)) Then openCircle(coor)
               Else star(coor);
            If ((nd = spNode) and clade) Then Begin
               dumpData(min,orgn,sf,nd^.sibL,t,true);
               dumpData(min,orgn,sf,nd^.sibR,t,true);
            End {then}
            Else Begin
               dumpData(min,orgn,sf,nd^.sibL,t,special);
               dumpData(min,orgn,sf,nd^.sibR,t,special);
            End;
         End; {then}
      End; {procedure dumpdata}
   {*************************************************************************}
   Begin {plotResidualX}
      statResidualX(rt,t,stpk);
      graphicsOn(mx,my);
      initPlot(1,2,stpk.min,stpk.max,fl,scalingfactor,origin,mx,my);
      outTextXY(0,0,'Plot of residuals versus');
      outTextXY(0,oneline,'positivized standardized');
      str(t,s);
      outTextXY(0,2*oneline,'contrasts for Trait '+s+'.');

      st:='Residual';
      temp:=round((origin[2]-length(st)*oneline)/2);
      If temp<1
      Then temp:=1;
      dumpDown(round(0.9524*origin[1])-8,temp,st);
      If t=1
      Then st:='Trait 1 = X'
      Else st:='Trait 2 = X';
      temp:=round((mx-origin[1]-textwidth(st))/2)+origin[1];
      If temp< origin[1]+1
      Then temp:=origin[1]+1;
      outTextXY(temp,round(1.05*origin[2])+2,st);
      setLineStyle(3,0,1);
      plotLine(0,0,stpk.min,stpk.max,scalingfactor,origin,stpk.n1,stpk.n2);
      setLineStyle(0,0,1);
      outTextXY(origin[1],my-2*oneline,
                 '        Dashed line = residual of zero');
      outTextXY(mx-2*oneline,my-oneline,'10');
      dumpData(stpk.min,origin,scalingfactor,rt,t,false);
      dumpStat10(stpk,mx,my,bail);
   End; {procedure plotResidualX}

{**********************************************************************}
{This procedure runs the screen displays for screens 9, 9A, 9B, 9C, 10
 and 11                                                                }
{**********************************************************************}
Procedure showRegressionStats(nd : nodePtr;fln : flString);
   Var
      specialNode : nodePtr;
      m,y0:extended;
      t,t2:integer;
      constantIndex : integer;
      dfReduce : integer;
      wholeClade : boolean;
      bail : boolean;
      tooLong : boolean;
      allZero : boolean;
   Begin
      checkConstantTips(nd,constantIndex);
      If (constantIndex=0) Then Begin
         plotCon(nd,specialNode,fln,m,y0,t,t2,dfReduce,wholeClade,bail);
         If Not(bail) Then Begin
            plotRoot(nd,fln,m,t,t2,dfReduce);
            dumpText3;
            tempResidual(nd,m,y0,t,t2);
         End; {then}
         If Not(bail)
         Then plotResidualX(nd,t,fln,specialNode,wholeClade,bail);
         tooLong := false;
         allZero := false;
         checkAllZero(nd,allzero);
       {  checkBranchTooLong(nd,tooLong); }
         If (Not(bail) And
            Not(allZero) And
            Not(tooLong))
         Then Begin
            tempAbsResidual(nd,m,y0,t,t2);
            temp3CorrectedHeight(nd,nd,0);
            plotResidualHeight(nd,fln,specialNode,dfReduce,wholeClade,bail)
         End {then}
         Else If (Not(bail)) Then Begin
            If allZero
            Then explainSkip
            {Else explainSkip2;}
         End; {else/if}
      End
      Else explainSkip3(constantIndex);
   End; {procedure showRegressionStats}
 

End. {unit t_fels_r}