{This unit has the input and output procedures.  It provides for I/O of a
tree to and from disk as well as the creation of a tree.}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}


Unit t_io1;

Interface

Uses t_util1;  {type definitions, constants, and utility functions.}

Procedure InputTree(Var rt: nodeptr;
                    Var flname :flstring;
                    Var ext :typstring;
                    Var succ:boolean);
Procedure SaveTree(rt:nodeptr;
                   Var flename:flstring;
                   Var typ:typstring;
                   ok: boolean);
Procedure CreateTree(Var rt:nodeptr;
                     Var have:boolean;
                     Var flname:flstring;
                     Var typ:typstring);

implementation

uses crt;      {Borland low-res text screen utils}

{***************************************************************************}
{Ok, here we go, this procedure is long, but it only reads in a tree.  It
passes back a pointer to the root, the filename, the extension of the
filename, and succ if it did input a file.}
{***************************************************************************}
Procedure InputTree(Var rt: nodePtr;
                    Var flname :flString;
                    Var ext :typString;
                    Var succ:boolean);
   Var
      fl : text;
      ch : char;
      me : nodePtr;
      dumptr : nodePtr;
      name :nameString;
      ok, good : boolean;
   {************************************************************************}
   {This procedure reads in a file name.  That is simple, but the task gets
   more complex when you require that the program not crash when invalid
   junk is typed in by the user.  That is the reason that this routine is
   a bit of a mess.  Hopefully you won't have to mess with it.}
   {************************************************************************}
   Procedure getInFile(Var fleName: flstring;
                       Var exten:typstring;
                       Var succ:boolean);
      Var
         i :integer;
         ch :char;
         place : byte;
         exists, valid:boolean;
         returnToMenu : boolean;
      Begin
         exists:=true;
         valid:=true;
         Repeat
            If not exists
            Then writeln(fleName,' does not exist.');
            If not valid
            Then writeln(fleName,' is not of a valid type.');
            exists:=true;
            valid:=true;
            writeln('Please input the name of the input file with ');
            writeln('the .INP, .PDI, or .BRK ending.');
            writeln('   Or "R" for return to Main Menu');
            readln(flename);
            returnToMenu := ((fleName ='R') or (fleName='r'));
            place:=pos('.',fleName);
            If place = 0          {If no . is found in the filename}
            Then valid:= false    
            Else Begin
               exten:=copy(flename,place+1,3);
               For i:=1 To 3 Do
                  exten[i]:=upcase(exten[i]);
               If (exten<>'INP') And (exten<>'PDI') And (exten<>'BRK')
               Then valid := false              {If extension bad}
               Else Begin                 {check to see if the file exists}
                  {$I-}
                  Assign(fl,flename);
                  Reset(Fl);
                  Close(Fl);
                  {$I+}
                  exists:= (IOResult=0) and (flename<>'');
               End; {else}
            End; {else}
         Until (exists And valid) Or returnToMenu;
         If returnToMenu Then Begin {If user wants to abort}
            succ:=false;
            ret('Please press any key to return to the Main Menu.');
         End; {then}
      End; {procedure getInFile}

   {************************************************************************}
   {This small function handles part of the error reporting for the three
    verification procedures.  PEM 2 Nov 1999.                               }
   {************************************************************************}
   Function badFileReport(typStr,firstLine,secondLine : string) : boolean;
     Var userChar : char;
     Begin
        writeln('This does not seem to be a valid ' + typStr + ' file.');
        writeln('The first line was:');
        writeln(firstLine);
        writeln;
        writeln('The second line was:');
        writeln(secondLine);
        writeln;
        writeln('Do you want to try to read it as a ' + typStr + ' file anyway (Y/N=Enter)? ');
        userChar := readKey;
        badFileReport := userChar in ['Y','y']
     End; {function badFileReport}

   {************************************************************************}
   {The following three functions check the veracity of the input file
    extension.  They operate by opening the file, reading the first couple
    of lines and comparing the lines with the format expected.

    verifyPDI validates a presumed PDI file.  It expects an integer on the
    first line, followed by a blank line.                                   }
   {************************************************************************}
   Function verifyPDI(Var fl : text) : boolean;
      Var i : integer;
          firstLine : string;
          secondLine : string;
          userChar : char;
          badVal : boolean;
      Begin
         reset(fl);
         readln(fl,firstLine);
         readln(fl,secondLine);
         i := 1;
         badVal := false;
         Repeat
           If ((firstLine[i] < '0') or
               (firstLine[i] > '9'))
           Then badVal := true
           Else inc(i);
         Until badVal or (i > length(firstLine));
         If badVal or (length(secondLine)>0)
         Then verifyPDI := badFileReport('PDI',firstLine,secondLine)
         Else verifyPDI := True;
      End; {procedure verifyPDI}

   {************************************************************************}
   {verifyINP validates a presumed INP file.  It expects the first line to
    start with the word 'This' and the second line to start with two spaces
    followed by the word 'comparative'                                      }
   {************************************************************************}
   Function verifyINP(Var fl : text) : boolean;
      Const goodFirst =
             'This is a file of descriptive and input data to be used in the';
            goodSecond = 
             '  comparative method analysis programs CMSINGLE, CMSIMUL, CMTPANAL';
      Var firstLine : string;
          secondLine : string;
          userChar : char;
      Begin
         reset(fl);
         readln(fl,firstLine);
         readln(fl,secondLine);
         If eof(fl) Then Begin
            writeln('Error The file appears to be empty.  No tree loaded.');
            verifyINP := false
         End
         Else Begin
            If (firstLine <> goodFirst) or
               (secondLine <> goodSecond)
            Then verifyINP := badFileReport('INP',firstLine,secondLine)
            Else verifyINP := True
         End {else}
      End; {procedure verifyINP}

   {************************************************************************}
   {verifyBRK validates a presumed BRK file.  Since there is no "header" for
    BRK files, all we can do is check for the presence of matching
    parenthesis and hope this is not a file of LISP S-expressions.          }
   {************************************************************************}
   Function verifyBRK(Var fl: text) : boolean;
      Var parenCount : integer;
          firstLine : string;
          secondLine : string;
          dummy : boolean;
          testChar : char;
      Begin
         parenCount := 0;
         reset(fl);
         Repeat
           read(fl,testChar);
           If (testChar = '(') Then inc(parenCount);
           If (testChar = ')') Then dec(parenCount)
         Until eof(fl);
         close(fl);
         reset(fl);
         readln(fl,firstLine);
         if not(eof(fl))
         Then readln(fl,secondLine)
         Else secondLine := '';
         If (parenCount <> 0)
         Then Begin           {this is rather bogus, but it prevents hanging!}
            dummy := badFileReport('BRK',firstLine,secondLine);
            If dummy Then Begin
               writeln;
               writeln('Error - Unbalanced parenthesis detected.  No tree loaded.');
            End; {then}
            verifyBRK := False;
         End
         Else verifyBRK := True
      End; {procedure verifyBRK}

   {************************************************************************}
   {This procedure will read the tree into memory if it is in a PDI file.
    It requires the file text variable and returns rt, the pointer to the
    root of the tree it just read in.  It also returns fail=true if it
    failed.  Unless you messed with the PDI format you shouldn't need to
    mess with this.  If you want to see what it does, print out a PDI file
    and see how all the reads correspond to the data.
    Also keep in mind that there are conditional checks done on the file.
    Because this code reads extended precision values, it will read values
    as small as 1E-99.

    This is a new,improved tokenizing version of the PDI file reader,
    coded by PEM 11 Nov 1999.  It allows any number of spaces (no tabs)
    between tip names, and accepts names longer than 2 characters.  We
    could read and save names longer than 2 characters, but there is a
    space tradoff between tree size and the amount of information in each
    node (e.g. extra characters in the names).  Remember that the heap
    is limited by the size of code (currently >200K) and the stack (32K).
    A (partial) way round the heap limitation would be allocating name
    strings dynamically (with different sizes of names).  This approach
    would reduce the impact of longer names, but not avoid the problem.  It
    would also be a messy enhancement, and time would be better spent
    porting this program to a runtime environment that supports a larger
    heap.
   {************************************************************************}
   
   Procedure ReadPDI(Var fl :text;Var rt :nodePtr;var fail:boolean);
      Const tabChar = chr(9);
      Var
         me :nodeptr;
         valCode : integer;
         i : integer;
         nextName : string;
         ancestorName : string;
         lengthStr : string;
         dataStr : string;

      {****************************************************************}
      {This procedure reads a tab or space delimited token from the    }
      {source.  The token is returned in tok, which is set to a        }
      {maximum length of tlength.                                      }
      {****************************************************************}
      Procedure getToken(Var source:text; Var tok:string; tlength:byte);
         Var
            count : byte;
            curChar : char;
            done : boolean;
         Begin
            count := 0;
            done := eoln(source);
            While not(done) Do Begin
               read(source,curChar);
               If ((curChar = ' ') Or (curChar = tabChar)) Then Begin
                  If (count > 0) Then Begin {we have a token, finish}
                     done := true;
                  End
               End
               Else Begin
                  inc(count);
                  tok[count]:= curChar;
               End;
               done := done Or eoln(source);
            End; {while}
            If count<tlength
            Then tok[0] := chr(count)
            Else tok[0] := chr(tlength);
         End; {local procedure getToken}
      {****************************************************************}
      {This procedure checks that each node in the tree is either an   }
      {interior node with two daughters or a tip with no daughters.    }
      {It is possible to create nodes with only one daugther when      }
      {editing a file with the text editor.  PEM 7 Jan 2001            }
      {****************************************************************}
      Function checkDaughters(nod: nodePtr) : boolean;
         Begin
            If (nod^.sibL = NIL) And (nod^.sibR = NIL)
            Then checkDaughters := True
            Else Begin
               If (nod^.sibL = NIL) Or (nod^.sibR = NIL)
               Then Begin
                  writeln('Node ',nod^.name,' has only one offspring.  Please');
                  writeln('modify your tree in a text editor so that node ',nod^.name);
                  writeln('has only two offspring.  (Once the corrected tree');
                  writeln('file is loaded back into PDTREE you can trim this');
                  writeln('extraneous daughter.)  ');
                  checkDaughters := False;
               End
               Else checkDaughters := checkDaughters(nod^.sibL) And
                                       checkDaughters(nod^.sibR);
            End; {else}
         End; {function checkDaughters}
      {********************************************************************}
      Begin
         If verifyPDI(fl) Then Begin
            reset(fl);
            readln(fl);
            readln(fl);
            initme(rt);
            rt^.ancestor:=nil;
            rt^.length:=0.0;
            readln(fl,rt^.name);       {only read the root name, rest is trash}
            getToken(fl,nextName,2);
            While nextName<>'' Do Begin    {reading the tree block}
               getToken(fl,ancestorName,2);
               fail:=true;
               checkName(rt,ancestorName,fail,me);
               If fail Then Begin
                  gotoxy(1,21);
                  writeln('Ancestor name "'+ancestorName+'" not found in tree block. ');
                  writeln('File is invalid.');
                  killTree(rt);
                  rt := nil;   {dispose doesn't set the pointer to NIL!}
                  exit;
               End; {then}
               If me^.sibL=nil Then Begin
                  new(me^.sibL);
                  initMe(me^.sibL);
                  me^.sibL^.ancestor:=me;
                  me:=me^.sibL;
               End {then}
               Else Begin
                  new(me^.sibR);
                  initMe(me^.sibR);
                  me^.sibR^.ancestor:=me;
                  me:=me^.sibR;
               End; {else}
               me^.name:=nextName;
               me^.length := 0.0;
               getToken(fl,lengthStr,32);  {length doesn't really matter here}
               val(lengthStr,me^.length,valCode);
               If valCode <> 0 Then Begin
                  writeln('Error - Bad length for branch leading to ', me^.name, '.');
                  writeln('        Length string read was: ',lengthStr);
                  writeln('        Will set branchlength to zero; you should fix the .PDI file');
                  writeln;
                  me^.length := 0;
               End; {then}
               readln(fl);              {flush the rest of the line}
               getToken(fl,nextName,2);
            End; {while}
            If not(fail)  {check that each node has 0 or 2 daughters}
            Then fail := not(checkDaughters(rt));
            If not(fail) Then Begin
               readln(fl);
               getToken(fl,nextName,2);
               While nextName<>'' Do Begin {reading the tip data block}
                  fail:=true;
                  checkName(rt,nextName,fail,me);
                  If fail Then Begin
                     gotoxy(1,21);
                     writeln('Tip name "'+nextName+'" not found in tip value block. ');
                     writeln('File is invalid.');
                     killTree(rt);
                     rt := nil;   {dispose doesn't set the pointer to NIL!}
                     exit;
                  End;
                  {error checking, obviously this isn't fool proof, tips that}
                  {actually are set to 0,0 or are just uninitialized won't}
                  {have duplicates caught by this code.  PEM 7-12-01}
                  If ((me^.dat[1] <> 0.0) Or (me^.dat[2] <> 0.0)) Then Begin
                     writeln('Warning - Tip ',me^.name,' was previously initialized.');
                     writeln('          Recommend checking PDI file for duplicate tip names.');
                     writeln;
                  End; {then}
                  For i := 1 to 2 Do Begin
                     me^.dat[i] := 0.0;
                     getToken(fl,dataStr,20);  {length won't matter here}
                     val(dataStr,me^.dat[i],valCode);
                     If valCode <> 0 Then Begin
                        writeln('Error - Bad data in trait ',i:1,' for tip ', me^.name, '.');
                        writeln('        Data string read was: "',dataStr,'".');
                        writeln('        PTREE will set trait value to zero; you should fix the .PDI file');
                        writeln;
                        me^.dat[i] := 0
                     End {then}
                  End; {for i}
                  readln(fl);  {flush the rest of the line}
                  getToken(fl,nextName,2);  {on to the next token}
               End; {while}
            End {then}
            Else Begin
               If (rt <> nil) Then Begin
                  killTree(rt);
                  rt := nil; {dispose doesn't set the pointer to NIL!}
               End; {then}
            End; {else}
         End {then}
         Else fail := true;
         close(fl)
      End; {procedure readPDI}
      {$V+} {Turn string type checking back on}
   {***********************************************************************}
   {This Procedure is similar to the last in the results it gives, but it
   speaks another language, so to speak.  It deals with INP files which have
   the same information, but stored in a different format.  One thing I did
   do a little differently is to have a few procedures in here.  They are
   just to avoid some duplicate code -- nothing special about them.  Again,
   to get a feel for what this procedure does, print up a .INP file and
   see how this procedures reads correspond to the file it is reading.}
   {***********************************************************************}
   Procedure readINP(Var fl :text;
                     Var rt :nodePtr;
                     Var fail: boolean);
      Var
         me : nodePtr;
         ch : char;
         ch2 :string[2];
         chhh : char;
         ch4 : string[4];
         tempint : integer;
      {*********************************************************************}
      {This just performs the menial task of reading in the data on a
      node.}
      {*********************************************************************}
      Procedure readSib(var fl: text; sib,me:nodePtr);
         Begin
            sib^.ancestor:=me;
            read(fl,sib^.name);
            read(fl,ch);
            read(fl,sib^.length);
         End; {procedure readsib}
      {*********************************************************************}
      {This does something similar to the previous procedure, except that
      it checks the tree for duplicate names, then reads in one real piece
      of data.}
      {*********************************************************************}
      Procedure readTrait(i:integer;
                          rt:nodeptr;
                          Var fl:text;
                          ch2:string;
                          Var fail:boolean);
         Var
            me :nodeptr;
            tempReal :extended;
         Begin
            fail:=true;
            checkName(rt,ch2,fail,me);
            If fail then exit;
            read(fl,tempreal);
            me^.dat[i]:=tempreal;
         End; {readTrait}
   {***********************************************************************}
      Begin
         If verifyINP(fl) Then Begin
            reset(fl);
            Repeat
               read(fl,ch4);
               readln(fl);
               If eof(fl) Then Begin
                  writeln('Error - Keyword search failed. No tree loaded.');  
                  exit;
               End; {then}
            Until ch4='  of';
            initme(rt);
            rt^.ancestor:=nil;
            rt^.length:=0.0;
            read(fl,ch2);
            rt^.name:=ch2;
            While not eof(fl) Do Begin
               fail:=true;
               checkname(rt,ch2,fail,me);
               If fail Then Begin
                  ret('Tree in file is invalid(1).  Please press any key '+
                        'to return to the Main Menu.');
                  killtree(rt);
                  exit;
               End; {then}

               new(me^.sibL);
               initme(me^.sibL); 
               read(fl,ch);
               readsib(fl,me^.sibL,me);
               read(fl,tempint);
               read(fl,ch);
               new(me^.sibR);
               initme(me^.sibR);
               readsib(fl,me^.sibR,me);
               readln(fl);
               read(fl,ch2);
            End; {while}
            close(fl);
            reset(fl);
            Repeat
               read(fl,ch);
               readln(fl);
            Until ch='-';
            read(fl,ch2);
            While Not eoln(fl) Do Begin
               readtrait(1,rt,fl,ch2,fail);
               If fail Then Begin   
                  ret('Tree in file is invalid(2).  Please press any key '+
                     'to return to the Main Menu.');
                  exit;
               End; {then}
               read(fl,ch);
               read(fl,ch2);
               read(fl,ch2);
               readtrait(2,rt,fl,ch2,fail);
               If fail Then Begin
                  ret('Tree in file is invalid(3).  Please press any key '+
                       'to return to the Main Menu.');
                  exit;
               End; {then}
               readln(fl);
               read(fl,ch2);
            End; {while}
         End {then}
         Else fail := true;
         close(fl);
      End; {procedure readINP}
   {***********************************************************************}
   {This procedure reads in a tree of the .BRK format.  It has 2 major
   problems.  The .BRK format does not necessarily have a space after a
   number so I had to write the procedure readnum (below) to read a number
   character by character.  Second, while this program must have names
   associated with internal nodes, the .BRK format does not save them.
   therefore, I had to write the procedure getname (below) that simply
   gives me a 2 character name.  A lesser problem is that .BRK supports
   tip names longer than 2 characters while PDTREE and PDSIMUL support only
   2 characters.  This procedure is much different in character than the last
   two because it simply parses the input, reading it like a set of
   instructions one at a time.  Each of the characters ( ) : , is
   viewed as an order to do something so my algorithm simply inputs
   characters one at a time doing what each character says.  It ignores
   white space, and will skip over any number of lines in the beginning until
   a ( is found at the start of a line.  To get a feel for how this format
   works, print out a .BRK file, the tree from the edit module and see how
   they correleate.

   The .BRK format has 2 files that are saved along with it.  The .TIP file
   has the Tip data, and the .NAM file has the tip names followed by zeroes.
   If you do not see a need for the .NAM file ignore it, it is produced as a
   convenience to people who use MacClade.  This program only reads in the
   .BRK and the .TIP files.

   In July 1999, I added some code so that PDTREE could directly read the
   trees embedded in NEXUS files produced by MacClade.  The most important
   part of this was defaulting all branch lengths to 1, unless the branch
   was interior to a polytomy.  Although this is usually not "correct", it
   is consistant with the approach to branch lengths taken in the MacClade
   book.  See section 8 of PDINSTR.DOC for a description of the steps
   required for getting files from MacClade to PDTREE. PEM
   }
   {***********************************************************************}
   Procedure readBRK(Var fl:text;Var rt:nodeptr; Var fail:boolean);
      Var
         nd,temp:nodeptr;
         badNode : nodePtr;
         ch:char;
         nm:namestring;
         newNodeName : nameString;
         namenum,n:integer;
         pleaseread,ok:boolean;
         {logFile : text;}
         foo : integer;
      {*********************************************************************}
      {This procedure reads in characters one at a time, skipping white
       space until a number is input, then it collects the characters of the number
       until a non number symbol is input.  It finally, converts the
       number's characters into the number and returns it in the variable
       num and returns the last character input.                            }
      {*********************************************************************}
      Procedure readNum(Var fl:text; Var num:extended; Var ch:char);
         Var
            st:string;
            n,dum:integer;
            done:boolean;
         Begin
            st:='';
            Repeat
               read(fl,ch);
               n:=ord(ch);
            Until((ch<>' ') and (n<>9) and (n<>10) and (n<>13)); {whitespace}
            done:=false;
            Repeat
               If (((n>47) And (n<58)) Or    {chr(47)=0, chr(58)=9}
                   (ch='.') Or (ch='+') Or (ch='-') Or
                   (ch='e') or (ch='E'))
               Then Begin
                  st:=st+ch;
                  read(fl,ch);
                  n:=ord(ch);
               End {then}
               Else done:=true;
            Until done;
            val(st,num,dum);
         End; {procedure readnum}
      {**********************************************************************}
      {This procedure returns a 2 character node name.  It only uses upper
      case characters and the digits 0-9.  It is deterministic and simply
      returns the first unused name after the "num"th name in a sequence
      that starts with A and ends with 9.  It increments num as necessary so
      it points to an unexplored name when the procedure returns.}
      {***********************************************************************}
      Procedure genName(Var nm:namestring; rt:nodeptr; Var num:integer);
         Var
            n1,n2:integer;
            where:nodeptr;
            ch:char;
            nm2:namestring;
         Begin
            Repeat
               n1:=num mod 26;
               n2:=num div 26;
               If n2<=8
               Then nm2:=chr(49+n2)+chr(65+n1)
               Else nm2 := chr(40+n2)+chr(97+n1);
               inc(num);
               ok:=true;
               CheckName(rt,nm2,ok,where);
            Until ok;
            nm:=nm2;
         End; {genName}
      {*********************************************************************}
      {This procedure reads a single taxon name.}
      {*********************************************************************}
      Procedure readTaxon(Var nm : nameString;Var ch : char);
         Const expectedTerminals = [':',',',')'];
         Var
            gname : nameString;
            n : integer;  {holds the ascii value of the current character}
            quoted : boolean;
            peeked : boolean;
         Begin
            n:=ord(ch);
            nm:='';
            quoted := false;
            Repeat
               If (ch = '''') Then Begin
                  If (length(nm) = 0)
                  Then quoted := true;
               End
               Else Begin
                  If (length(nm)<2)
                  Then nm:=nm+ch;
               End;
               read(fl,ch);
               n:=ord(ch);
            Until((quoted And (ch = '''')) Or
                  (not(quoted) And (ch in expectedTerminals)));
            If quoted Then Begin
               read(fl,ch);
               n:=ord(ch);
            End; {then}
            {Now force the name to behave itself}
            If (length(nm)=0) Then genName(nm,rt,namenum);
            If (length(nm)=1) Then nm := ' '+nm;
            If (nm[1] = '_') Then nm[1] := ' ';
            If ((length(nm)>1) And (nm[2] = '_')) Then nm[2] := ' ';
            checkName(rt,nm,ok,badNode);
            If (not ok) Then Begin
               genName(gname,rt,nameNum);
               badNode^.name := gname;
            End; {then}
         End; {procedure readTaxon}

      Procedure report(nd:nodeptr);
         Begin
            If (nd <> nil) Then Begin
               writeln('Debug: name ',nd^.name,' length ',nd^.length);
               report(nd^.sibL);
               report(nd^.sibR);
            End;
         End; { procedure report }
      {*********************************************************************}
      {This procedure checks for the presence of unbranched (internal)
       nodes.  Such nodes have been generated in Newick trees generated by
       bugs in some (unnamed) programs, and are actually supported by
       Mesquite.  They are currently incompatible with the representation
       used by PDTREE, so this procedure will flag them.  In theory it would
       be possible to just splice them out and throw them away, probably
       requiring the user to save the tree and restart PDTREE.
       20 Feb 2002 PEM}
      {*********************************************************************}
      Procedure checkForUnbranchedNodes(nd : nodePtr;Var ok : boolean);
         Begin
            If ok Then Begin   {Don't do any extra work}
               If ((nd^.sibL = NIL) And
                   (nd^.sibR = NIL))
               Then ok := true   {really a nop}
               Else If ((nd^.sibL = NIL) Or
                        (nd^.sibR = NIL))
                    Then Begin
                      writeln('Error: Node ',nd^.name,' has only one daughter');
                      writeln('       Such nodes are not supported by PDTREE');
                      writeln('       Please check your .BRK file and the program');
                      writeln('       that created it.');
                      writeln;
                      ok := false;
                    End {else}
                    Else Begin
                       checkForUnbranchedNodes(nd^.sibL,ok);
                       checkForUnbranchedNodes(nd^.sibR,ok);
                    End; {else}
            End; {then}
         End; {procedure checkForUnbranchedNodes}
      {*********************************************************************}
      Begin {readBRK}
         If verifyBRK(fl) Then Begin
            {assign(logfile,'brk.log');
            rewrite(logfile);           }
            initMe(rt);
            nameNum:=0;
            pleaseRead:=true;
            reset(fl); 
            read(fl,ch);
            If ch<>'(' Then Repeat  {Pass by all the meaningless lines in the beginning.}
               If (ord(ch)<>13) And (ord(ch)<>10)
               Then readln(fl);         {if line has characters left}
               read(fl,ch);
            Until((ch='(') or (ord(ch)=26));
            If (ord(ch)=26) Then Begin               {chr(26)= end of file}
               fail:=true;
               ret('Tree file is invalid. Please press any key');
               exit;
            End; {then}
            rt^.ancestor:=nil;
            rt^.length:=0;
            rt^.name:='rt';
            nd:=rt;     {nd is used to point to the node that we are working on.}
            Repeat
               If pleaseRead
               Then read(fl,ch)
               Else pleaseRead:=true;
               If ch='(' Then Begin {Add interior node above the current node, nd.}
                  If nd^.sibL=nil Then Begin {Do the left one unless occupied.}
                     {writeln(logfile,'add interior node as left daughter');}
                     new(nd^.sibL);
                     initMe(nd^.sibL);
                     nd^.sibL^.ancestor:=nd;
                     nd:=nd^.sibL;
                  End {then}
                  Else Begin {If the left is occupied, then the right one is free.}
                     {writeln(logfile,'add interior node as left daughter');}
                     new(nd^.sibR);
                     initme(nd^.sibR);
                     nd^.sibR^.ancestor:=nd;
                     nd:=nd^.sibR;
                  End; {else}
                  {common code now that nd is updated}
                  genName(nd^.name,rt,namenum);
                  nd^.length:=1.0;   {read MacClade files w/o branch lengths, July 1999 PEM}
               End {then}
               Else If ch=','
                  Then Begin    {Drop down one node in the tree.}
                     nd:=nd^.ancestor;
                     If nd^.sibR<>nil Then  {If there is a polytomy or more.}
                       If nd=rt Then Begin    {Add new root, put tree in sibL.}
                          {writeln(logfile,'creating polytomy at root'); }
                          new(nd^.ancestor);
                          initMe(nd^.ancestor);
                          nd^.ancestor^.sibL:=nd;
                          nd:=nd^.ancestor;
                          rt:=nd;
                          genName(nd^.sibL^.name,rt,nameNum);
                          nd^.name:='rt';
                          nd^.length:=0.0; {set length to zero here! PEM July 1999}
                          nd^.ancestor:=nil;
                       End {then}
                       Else Begin {Put new node in between nodes, and put clade in sibL.}
                          {writeln(logfile,'polytomy elsewhere'); }
                          new(temp);
                          initMe(temp);
                          temp^.ancestor:=nd^.ancestor;
                          If(nd^.ancestor^.sibL=nd)
                          Then nd^.ancestor^.sibL:=temp
                          Else nd^.ancestor^.sibR:=temp;
                          nd^.ancestor:=temp;
                          temp^.sibL:=nd;
                          temp^.length:=nd^.length;
                          nd^.length:=0.0;  {set length to zero here! PEM July 1999}
                          genName(temp^.name,rt,nameNum);
                          nd:=temp;
                       End; {else}
                  End {then}
                  Else If ch=')' Then Begin {Drop down a level.}
                     If nd^.ancestor<>nil
                     Then nd:=nd^.ancestor;
               End {then}
               Else If ch=':' Then Begin  {input a number and interpret as branch length}
                  {writeln(logfile,'reading a branch length');}
                  readnum(fl,nd^.length,ch);
                  pleaseread:=false;
               End {then}
               Else Begin
                  n:=ord(ch);
                  If (n<>9) And (n<>10) And (n<>13) And (n<>32) And (ch<>';')
                  Then Begin  {If not whitespace, must be the name of a tip node}
                     pleaseRead:=false;
                     readTaxon(nm,ch);
                     If nd^.sibL=nil Then Begin
                     {Do the left one unless it is already occupied.}
                        {writeln(logfile,'adding a tip ',nm,' to left daughter'); }
                        new(nd^.sibL);
                        initMe(nd^.sibL);
                        nd^.sibL^.ancestor:=nd;
                        nd:=nd^.sibL;
                        nd^.name:=nm;
                     End {then}
                     Else Begin
                     {If the left is not free, then the right one is.}
                        {writeln(logfile,'adding a tip ',nm,' to right daughter');}
                        new(nd^.sibR);
                        initMe(nd^.sibR);
                        nd^.sibR^.ancestor:=nd;
                        nd:=nd^.sibR;
                        nd^.name:=nm;
                     End; {else}
                     If (ch = ':')
                     Then nd^.length:=0 {this will be overwritten by the following number}
                     Else nd^.length:=1.0;  {default for trees w/o branch lengths}
                  End; {then}
               End; {else}
            Until ch=';';
            If nameNum > 467 Then Begin
               writeln('Error: loaded tree is too big; loaded terminated');
               killtree(rt);
               fail := true;
            End {then}
            Else fail:=false;
         End {then}
         Else fail := true;
         close(fl);
         ok := true;
         checkForUnbranchedNodes(rt,ok);
         fail := fail Or not(ok);
         {writeln(logfile);
          close(logfile); }
         {report(rt);
          numtps(rt,foo);}
      End; {readBRK}
   {***********************************************************************}
   {This procedure attempts to read in the .TIP file to get the tip data
   for the tree just read in in the .BRK format.  If the file does not exist
   a warning is issued and all the data remains zero.}
   {***********************************************************************}
   Procedure tryReadTIP(flnm:flstring; rt:nodeptr);
      Var
         name:nameString;
         fl : text;
         i : integer;
         ch : char;
         ch2 : char;
         good : boolean;
      {*********************************************************************}
      {This procedure is given a tip name.  It searches the tree for that
      tip.  When it finds it, it reads in that tips data, and sets good to
      true.  It then stops recursing. }
      {*********************************************************************}
      Procedure inputData(nd:nodePtr;
                          name:nameString;
                          Var fl:text;
                          Var good:boolean);
         Begin
            If ((nd^.sibL=nil) and ((nd^.name[1]=name[1])
                                     And (nd^.name[2] = name[2]))) Then Begin
               good:=true;
               readln(fl,nd^.dat[1],nd^.dat[2]);
            End; {then}
            If (nd^.sibL<>nil) Then Begin
               inputData(nd^.sibL,name,fl,good);
               If Not good
               Then inputData(nd^.sibR,name,fl,good);
            End {then}
         End; {procedure inputData}
      {*********************************************************************}
      Begin
         flnm:=flnm+'TIP';
         {$I-}     {check to see if the file exists}
         assign(fl,flnm);
         reset(Fl);
         close(Fl);
         {$I+}
         If ((IOResult=0) and (flnm<>'')) Then Begin
            reset(fl);
            Repeat
              read(fl,ch);
              If ch = ''''
              Then Begin
                 read(fl,ch);
                 read(fl,ch2);
                 name := ch+ch2;
                 read(fl,ch);    {kill second '}
              End
              Else Begin
                 read(fl,ch2);
                 name := ch+ch2;
              End;
              good:=false;
              inputData(rt,name,fl,good);
              If Not good Then Begin   {If the tip not found, data not read}
                 readln(fl);
                 writeln('Tip ',name,' not found.  Its data was skipped.');
              End; {then}
            Until eof(fl);
         End {then}
         Else Begin
            writeln('The .TIP file was not found.  The .TIP file should have',
                      ' had the tip data.');
            writeln('Therefore, all the tip data of the tree just read in ',
                      'remains zero.');

         End {else}
      End; {procedure tryreadTIP}

   {***********************************************************************}
   {The idea here is to make sure that no two nodes in the tree have the same
   name.  If the input file had this problem, the user is told about it,
   and forced to input a new name for the duplicate nodes.  At this point,
   if there were duplicate nodes, the new tree may be 'incorrect'.  I mean
   it may not be an accurate representation of what was intended by the
   file.  However, in all other aspects after it passes here it is valid.}
   {***********************************************************************}
   Procedure doubleCheckNames(rt,nd:nodePtr;
                              Var notFound:boolean;
                              Var bdnd:nodePtr);
      Var
         nm : string[2];
         dumptr :nodePtr;
      Begin
         nm:=nd^.name;
         nd^.name:='  ';
         checkName(rt,nm,notfound,dumptr);
         nd^.name:=nm;
         If Not notfound
         Then bdnd:=nd;
         If notFound And (nd^.sibL<>nil)
         Then doubleCheckNames(rt,nd^.sibL,notfound,bdnd);
         If notFound And (nd^.sibR<>nil)
         Then doubleCheckNames(rt,nd^.sibR,notfound,bdnd)
      End; {doubleCheckNames}
   {*************************************************************************}
   {Here we are at the body of the input procedure.  Again, if it were assumed
   that the user did not do dumb things this would be a lot shorter and simpler,
   but we try to make sure that noone does anything they don't intend!
   This checks for such things as a file already in memory, and deals with
   duplicate names of nodes.}
   {*************************************************************************}
   Begin
      clrscr;
      writeln('******************* Input Module ******************');
      writeln;
      writeln;
      ch:='Y';
      If succ=true Then Begin
         Repeat
            writeln('You already have a tree in memory.');
            writeln('It is called: ',flname);
            writeln('If you choose to input a different tree, then ',flname);
            writeln('will be cleared from memory.');
            writeln;
            writeln('Do you want to input a new tree and remove the ',
                      'current');
            writeln('tree from memory? (Y/N)');
            ch:=UpCase(readkey);
            writeln;
         Until ch In ['Y','N'];
         If ch='Y' Then Begin    {added 13-May-95 to fix minor bug PEM}
            killTree(rt);        {make sure the tree really is returned}
            rt:=nil              {to the heap}
         End {then}
      End; {then}
      If ch='Y' Then Begin
         succ:=true;
         getinfile(flname,ext,succ);
         If succ Then Begin
            Assign(fl,flname);
            new(rt);
            initMe(rt);
            If ext='INP'
            Then readINP(fl,rt,succ)
            Else
               If ext='PDI'
               Then readPDI(fl,rt,succ)
               Else Begin
                  readBRK(fl,rt,succ);
                  If not(succ) Then Begin
                     tryReadTIP(copy(flname,1,length(flname)-3),rt);
                     writeln;
                     writeln('NOTE: the .BRK format does not have the ',
                              'internal node names stored.  They are');
                     writeln('   generated anew each time the tree is ',
                              'read in.');
                  End; {then}
               End; {else}
               succ:=not succ;  {succ actually flags failure here, so invert}
               If succ Then Begin
                  writeln(ext,' file read.');
                  Repeat
                     succ:=true;
                     doubleCheckNames(rt,rt,succ,me);
                     If Not(succ) Then Begin
                        writeln;
                        writeln('A duplicate node name was found.');
                        writeln('Please input a new name for: ',
                                    me^.name,'.');
                        ok:=true;
                        good := true;
                        Repeat
                           If Not ok
                           Then writeln(name,' is not unique.  ','Try again.');
                           If Not good
                           Then writeln(name,' is not 2 characters.  ',
                                           'Please try again.');
                           readln(name);
                           ok:=true;
                           good:=true;
                           If length(name)<2 Then good:=false;
                           checkName(rt,name,ok,dumptr);
                        Until good and ok;
                        me^.name:=name;
                     End; {then}
                  Until succ;
                  whiteCheck(rt,rt);
                  writeln;
                  writeln;
                  writeln('File successfully loaded.');
               End; {then}
            End;
         End;
         If not(succ) and (rt<>nil) Then Begin
            dispose(rt);
            rt := nil;
         End; {then}
         ret('Please press any key to return to the Main Menu.');
   End;

{***************************************************************************}
{This procedure again is long, but only saves the tree to disk.  It requires
the root, the current filename, the current extension, and ok which is true
if there is a tree to save.  It is only called by the main program.}
{***************************************************************************}
Procedure saveTree(rt:nodePtr;
                   Var fleName:flString;
                   Var typ:typString;
                   ok: boolean);
   Var
      fl         : text;
      num        : extended;
      ans        : char;
      userChar   : char;
      doFloats   : boolean;
      writeError : boolean;
   {************************************************************************}
   {This procedure gets the filename from the user.  That's all.  It is
   longer and messier than necessary because we try not to let the program
   crash when the user does silly things.}
   {************************************************************************}
   Procedure getFlnm(Var flnm:flstring; Var typ:typString);
      Var
         good, valid, again : boolean;
         ch:char;
         i :integer;  
         place :byte;
      Begin
         writeln('The current name for this tree is: ',flnm);
         writeln;
         writeln('Would you like to change its name before it is saved');
         writeln('so that the old file will not be destroyed? (Y/N)');
         Repeat
            ch:=upCase(readkey);
         Until ch In ['Y','N'];
         If ch = 'N' Then Begin
            {$I-}
            Assign(fl,flnm);
            Rewrite(fl);
            close(fl);
            {$I+}
            If (IOresult<>0) or (flnm='') Then good:=false;
         End
         Else good:=true;
         valid:=true;
         again:=false;
         If ((ch='Y') or not(good)) Then Repeat
            writeln;
            If Not valid
            Then writeln(flnm,' is not of a valid type.');
            If Not good
            Then writeln(flnm,' is write locked or not a legal filename.');
            writeln('What shall be the new name?  Do not forget the ',
                     '.INP, .PDI, or .BRK extension.');
            writeln('Remember that the extension determines ',
                     'the format in which the tree is saved!');
            writeln;
            writeln('Be aware that .BRK does not save',
                     ' non-tip node names, and puts the tip');
            writeln('data it its own file with a .TIP extension.');
            writeln('In addition, the .BRK format has a third file ',
                      'with a .NAM extention that is');
            writeln('rather special purpose.  Just be aware that it',
                      ' exists.');
            readln(flnm);
            good:=true;
            valid:=true;
            again:=false;
            place:=pos('.',flnm);
            If place = 0
            Then valid:=false
            Else Begin
               typ:=copy(flnm,place+1,3);
               For i:=1 To 3 Do
                  typ[i]:=upcase(typ[i]);
               If (typ<>'INP') And (typ<>'PDI') And (typ<>'BRK')
               Then valid:=false
               Else Begin
                  {$I-}
                  assign(fl,flnm);
                  reset(fl);
                  close(fl);
                  {$I+}
                  If (IOresult=0) and (flnm<>'')
                  Then Begin
                     writeln;
                     writeln(flnm,' already exits.');
                     writeln('Do you want to delete it and ',
                               'save the current tree in its ',
                               'place? (Y/N)');
                     Repeat
                        ch:=UpCase(readkey);
                     Until ch in['Y','N'];
                     If ch='N'
                     Then again:=true
                     Else Begin
                        {$I-}
                        Assign(fl,flnm);
                        Rewrite(fl);
                        close(fl);
                        {$I+}
                        If (IOresult<>0) or (flnm='') Then good:=false;
                     End; {else}
                  End {then}
                  Else Begin
                     {$I-}
                     Assign(fl,flnm);
                     Rewrite(fl);
                     close(fl);
                     {$I+}
                     If (IOresult<>0) or (flnm='') Then good:=false;
                  End; {else}
               End; {else}
            End; {else}
         Until good And valid And Not again
      End; {procedure getflnm}
   {********************************************************************}
   {This function sweeps through the tree, checking branch lengths and  }
   {tip values for values out of bounds.  It returns false if it finds  }
   {an out of bounds value.                                             }
   {********************************************************************}
   Function boundsSweep(nd,rt : nodePtr) : boolean;
      Var temp : boolean;
      Begin
         temp := false;
         If nd^.sibL = nil Then Begin
            If ((abs(nd^.dat[1]) >= 1E15) Or
                ((abs(nd^.dat[1]) <= 1E-15) And
                 (abs(nd^.dat[1]) > 1E-99)))
            Then temp := true;
            If ((abs(nd^.dat[2]) >= 1E15) Or
                ((abs(nd^.dat[2]) <= 1E-15) And
                 (abs(nd^.dat[2]) > 1E-99)))
            Then temp := true;
         End {then}
         Else Begin
            If nd<>rt Then Begin
               If ((abs(nd^.length)>1E15) Or
                   ((abs(nd^.length)<1E-15) And
                    (abs(nd^.length) > 1E-99)))
               Then temp := true;
            End; {then}
            If not(temp) Then temp := boundsSweep(nd^.sibL,rt);
            If not(temp) Then temp := boundsSweep(nd^.sibR,rt);
         End; {else}
         boundsSweep := temp;
      End; {function boundsSweep}

   {********************************************************************}
   {This procedure writes an extended value to the output file.
    Depending on the value of doFloats, either an extended precision
    value (with exponents) or a floating point number will be written.  }
   {********************************************************************}
   Procedure ew(Var outFile :text;num :extended);
      Var outStr :string;
      Begin
         If doFloats
         Then Begin
            If ((abs(num)<1E15) And
                ((abs(num)>1E-15) Or
                 (abs(num)=0))) Then Begin
               str(num:17:15,outStr);
               If (length(outStr)>16)
               Then delete(outStr,16,length(outStr)-16);
               write(outFile,' ',outStr);
            End {then}
            Else writeln('Value out of bounds')
         End {then}
         Else extendwrite(outFile,num);
      End; {procedure ew}

   {*********************************************************************}
   {This procedure writes the tips data to the file in the PDI format
   recursively.}
   {*********************************************************************}
   Procedure wrttips(nd:nodePtr;Var fl:text);
     Begin
        If nd^.sibL=nil Then Begin   {this is a tip}
           write(fl,nd^.name,'  ');
           ew(fl,nd^.dat[1]);
           write(fl,'  ');
           ew(fl,nd^.dat[2]);
           writeln(fl)
        End {then}
        Else Begin                 {this is not a tip, ascend recursively}
           wrttips(nd^.sibL,fl);
           wrttips(nd^.sibR,fl);
        End {else}
     End; {procedure wrttips}
   {************************************************************************}
   {You may have noticed that this series of procedures is analogous to the
   ones in the read file block.  This one is in charge of writing the tree
   to the file.  One difference is that this uses recursive procedures of
   its own to do a lot of the work.}
   {************************************************************************}
   Procedure writePDI(rt:nodeptr;flname:flstring);
      Var
         fl : text;
         num: integer;
      {*********************************************************************}
      {This procedure writes the tree structure to a file in the PDI form
      recursively.}
      {*********************************************************************}
      Procedure wrtTree(nd:nodeptr; var fl:text);
         Begin
            write(fl,nd^.name,'  ');
            If nd^.ancestor=nil
            Then write(fl,'$$')                {write dummy ancestor for root}
            Else write(fl,nd^.ancestor^.name);
            write(fl,'  ');
            ew(fl,nd^.length); {write length of branch leading to nd}
            writeln(fl);
            If nd^.sibL<>nil Then Begin
                  wrttree(nd^.sibL,fl); {ascend tree recursively}
                  wrttree(nd^.sibR,fl);
            End
         End;
      {*********************************************************************}
      Begin
         assign(fl,flname);
         rewrite(fl);
         num:=0;
         numtps(rt,num);          {count the tips}
         writeln(fl,num);  {write tip count at top of file}
         writeln(fl);
         wrttree(rt,fl);
         writeln(fl);
         wrttips(rt,fl);
         close(fl)
      End; {writePDI}
   {************************************************************************}
   {This one writes the tree in the INP format.  Again, it uses two recursive
   procedures to help it.}
   {************************************************************************}
   Procedure writeINP(rt: nodePtr;flName:flString);
      Var num :integer;
      {********************************************************************}
      {This recursive procedure writes the tips data for INP format.}
      {********************************************************************}
      Procedure writeTips(nd:nodePtr;var fl:text);
         Begin
            If nd^.sibl=nil Then Begin
               write(fl,nd^.name,' ');
               ew(fl,nd^.dat[1]);
               write(fl,'   ');
               write(fl,nd^.name,' ');
               ew(fl,nd^.dat[2]);
               writeln(fl)
            End {then}
            Else Begin
               writetips(nd^.sibL,fl);
               writetips(nd^.sibR,fl);
            End
         End; {writetips}
      {********************************************************************}
      {This one writes the tree structure in the INP format.  Yeah, it is
      recursive.}
      {********************************************************************}
      Procedure writeTree(nd:nodeptr;var fl:text);
         Var
            num :integer;
         Begin
            If nd^.sibL<>nil Then Begin
               write(fl,nd^.name,' ',nd^.sibL^.name,' ');
               ew(fl,nd^.sibL^.length);
               num:=0;
               numtps(nd^.sibL,num);
               write(fl,' ',num,' ',nd^.sibR^.name,' ');
               ew(fl,nd^.sibR^.length);
               write(fl,' ');
               num:=0;
               numtps(nd^.sibR,num);
               write(fl,round(num),' ');
               If nd^.ancestor<>nil Then Begin
                   write(fl,nd^.ancestor^.name,' ');
                   ew(fl,nd^.length);
                   writeln(fl)
               End {then}
               Else writeln(fl,'**  0.0000000000E+00');
               writetree(nd^.sibL,fl);
               writetree(nd^.sibR,fl);
            End; {then}
         End; {procedure writetree}
      {*********************************************************************}
      Begin {procedure writeINP}
         assign(fl,flname);
         rewrite(fl);
         writeln(fl,'This is a file of descriptive and input data to',
                                    ' be used in the');
         writeln(fl,'  comparative method analysis programs CMSINGLE,',
                              ' CMSIMUL, CMTPANAL');
         writeln(fl,'  CMFLANAL, CMMEANAL, and CMTRANAL.');
         num:=0;
         numtps(rt,num);
         Writeln(fl,'There are ',num,' tip values on this phylogeny.',
                           '  Their');
         Writeln(fl,'  measured values are :');
         Writeln(fl,'    Trait A             Trait B');
         Writeln(fl,'Name     Value      Name     Value');
         Writeln(fl,'-------------------------------------------');
         writetips(rt,fl);
         writeln(fl);
         Writeln(fl,'The phylogeny structure may be described by',
                      ' the following sequences');
         Writeln(fl,'  of tips, nodes and branch lengths:');
         writeTree(rt,fl);
         close(fl)
      End; {procedure writeINP}

   {************************************************************************}
   {This procedure writes node names into Nexus files after performing      }
   {several checks and conversions.  It is also used in writing PDAP        }
   {formats derived from the Nexus specification (Maddison et. al. 1997)    }
   {(e.g.,BRK,TIP,NAM).                                                     }
   {************************************************************************}
   Procedure writeNexusName(Var fl : text; nam : nameString);
      Var
         i : integer;
         dummy : nameString;
         quotesNeeded : boolean;
      Begin
         quotesNeeded := false;
         For i := 1 To 2 Do Begin
            If nam[i] in ['[',']','(',')','{','}','=',',',';',':','-','\',
                          '/','*','"','''']
            Then quotesNeeded := true;
         End; {for i}
         dummy := nam;
         If (nam[1] = ' ') Then dummy[1] := '_';
         If ((length(nam) > 1) And (nam[2] = ' '))
         Then dummy[2] := '_';
         If quotesNeeded
         Then write(fl,'''',dummy,'''')
         Else write(fl,dummy);
      End; {procedure writeNexusName}

   {************************************************************************}
   {This function takes a name and returns the number of characters required}
   {to write the name according to the rules implemented by the preceeding  }
   {procedure.                                                              }
   {************************************************************************}
   Function nexusNameLength(nam : nameString) : integer;
      Var
         i : integer;
         quotesNeeded : boolean;
         result : integer;
      Begin
         quotesNeeded := false;
         For i := 1 To 2 Do Begin
            If nam[i] in ['[',']','(',')','{','}','=',',',';',':','-','\',
                          '/','*','"','''']
            Then quotesNeeded := true;
         End; {for i}
         If quotesNeeded
         Then nexusNameLength := length(nam)+3
         Else nexusNameLength := length(nam)+1;
      End; {function nexusnamelength}

   {************************************************************************}
   {This procedure writes the .BRK file out to disk.  (See the documentation
   on the readBRK procedure above).  It is much easier to write than to read.}
   {************************************************************************}
   Procedure writeBRK(rt:nodeptr; flname:flstring);
      Var
         fl:text;
         lineCount : integer;

      {********************************************************************}
      {This procedure adds a newline when necessary.                       }
      {********************************************************************}
      Procedure checkLength(charCount : integer; Var fl:text);
         Begin
            lineCount := lineCount + charCount;
            If lineCount > 75 Then Begin
               writeln(fl);
               lineCount := charCount;
            End; {then}
         End; {procedure checkLength}

      {********************************************************************}
      {This recursive procedure does all the work.  It traverses the tree,
      and writes out the .BRK representation of its topology with branch
      lengths and tip node names.}
      {********************************************************************}
      Procedure writeme2(nd,rt:nodeptr; var fl:text);
         Begin
            If nd^.sibL=nil Then Begin
               checkLength(nexusNameLength(nd^.name),fl);
               writeNexusName(fl,nd^.name);
               write(fl,':');
               checkLength(16,fl);
               ew(fl,nd^.length);
            End {then}
            Else Begin
               If ((nd^.length<>0) Or (nd=rt)) Then Begin
                  checkLength(1,fl);
                  write(fl,'(');
                  writeme2(nd^.sibL,rt,fl);
                  checkLength(1,fl);
                  write(fl,',');
                  writeme2(nd^.sibR,rt,fl);
                  checkLength(2,fl);
                  write(fl,'):');
                  checkLength(16,fl);
                  ew(fl,nd^.length);
               End {then}
               Else Begin
                  writeme2(nd^.sibL,rt,fl);
                  checkLength(1,fl);
                  write(fl,',');
                  writeme2(nd^.sibR,rt,fl);
               End; {else}
            End; {else}
        End; {procedure writeme2}
      {********************************************************************}
      Begin {writeBRK}
         lineCount := 0;
         assign(fl,flname);    {open the file}
         rewrite(fl);
         writeme2(rt,rt,fl);   {call writeme2}
         writeln(fl,';');      {add a semicolon}
         close(fl);            {done}
      End; {writeBRK}
   {************************************************************************}
   {This procedure writes the tips names to the file in the PDI format.
   It follows the names with a tab and a zero}
   {************************************************************************}
   Procedure writeNAM(rt:nodeptr;flname:flstring);
      Var
         fl:text;
         place:integer;
      {*********************************************************************}
      {This recursive procedure helps the one it is in do its job.}
      {*********************************************************************}
      Procedure wrtName(nd:nodeptr;var fl:text);
         Begin
            If nd^.sibL=nil                      {if a tip}
            Then Begin
               writeNexusName(fl,nd^.name); {the name followed by...}
               writeln(fl,chr(9),'0');      {a tab character and '0'}
            End {then}
            Else Begin
               wrtName(nd^.sibL,fl);
               wrtName(nd^.sibR,fl);
            End {else}              
         End; {procedure wrtname}
     {********************************************************************}
     Begin {writeNAM}
        place:=pos('.',flname);
        flname:=copy(flname,1,place)+'NAM';
        assign(fl,flName);
        rewrite(fl);
        wrtName(rt,fl);
        close(fl);
      End; {procedure writeNAM}
   {************************************************************************}
   {This procedure writes the .TIP file that is saved along with the .BRK and
   .NAM files.  The .TIP file holds the tip data in the same format as the
   .PDI file saves the tip data.}
   {************************************************************************}
   Procedure writeTIP(rt:nodeptr;flname:flstring);
     Var
       fl:text;
       place:integer;
      {*********************************************************************}
      {This procedure is very similar to the wrttips procedure for PDI files}
      {except that it quotes the tip names as needed.  Note: I resisted the
      {temptation to call this writeQTips PEM 16 Sept 2000.}
      {*********************************************************************}
      Procedure wrtQuotedTips(nd:nodePtr;Var fl:text);
         Begin
           If nd^.sibL=nil Then Begin   {this is a tip}
              writeNexusName(fl,nd^.name);
              write(fl,'  ');
              ew(fl,nd^.dat[1]);
              write(fl,'  ');
              ew(fl,nd^.dat[2]);
              writeln(fl)
           End {then}
           Else Begin                 {this is not a tip, ascend recursively}
              wrtQuotedTips(nd^.sibL,fl);
              wrtQuotedTips(nd^.sibR,fl);
           End {else}
        End; {procedure wrtQuotedTips}
     {*********************************************************************}
     Begin {procedure writeTIP}
       place:=pos('.',flname);
       flname:=copy(flname,1,place)+'TIP';
       assign(fl,flname);
       rewrite(fl);
       wrtQuotedTips(rt,fl);
       close(fl);
     End; {procedure writeTIP}
   {************************************************************************}
   Begin {saveTree}
      writeError := false;
      clrscr;
      writeln('********************* Save Module *******************');
      writeln;
      writeln;
      If not ok
      Then writeln('You must have a tree in memory to save it to disk.')
      Else Begin
         getflnm(flename,typ);
         writeln;
         writeln('Would you like to write values in floating point format (no exponents)?');
         write('(Y/N=Enter): ');
         userChar := readKey;
         writeln;
         doFloats := userChar in ['Y','y'];
         If doFloats Then Begin
            writeError := boundsSweep(rt,rt);
            If writeError Then Begin
               writeln('The tree contains values that are too large or small to write in floating');
               writeln('point format.');
               writeln;
               write('Would you like save the tree using exponential format (Y/N)? ');
               userChar := readKey;
               writeln;
               If userChar in ['Y','y'] Then Begin
                  doFloats := false;
                  writeError := false;
               End; {then}
            End; {then}
         End; {then}
         If doFloats Then Begin
            writeln;
            writeln('Warning - Values in your PDI file may include roundoff errors.  These will not');
            writeln('          affect the values used internally when the file is reloaded.');
            writeln;
         End; {then}
         If not(writeError) Then Begin
            If typ='INP'
            Then writeINP(rt,flename);
            If typ='PDI'
            Then writePDI(rt,flename);
            If typ='BRK'
            Then Begin
               writeNAM(rt,flename);
               writeTIP(rt,flename);
               writeBRK(rt,flename);
            End; {then}
            writeln('The Save was successful.');
         End; {then}
      End; {else}
      ret('Please press any key to return to the Main Menu.');
   End; {procedure saveTree}
{***************************************************************************}
{This procedure creates a simple tree with a root and two branches.  It
requires the root and name of any tree in memory, and a boolean that is
true if there is a tree in memory.  If there is, it asks the user if it
can write over that tree.  If the user replies yes then it does and changes
the root and the filename.
{***************************************************************************}
Procedure createTree(Var rt:nodePtr;
                     Var have:boolean;
                     Var flName:flString;
                     Var typ:typString);
   Var
      ans : char;
      num :extended;
      done : boolean;
      dum : boolean;
   Begin
      clrscr;
      writeln('************* Create Module *************');
      writeln;
      writeln;
      ans:='Y';
      If have Then Repeat
         writeln;
         writeln('You already have a tree in memory.');
         writeln('It is called : ',flname);
         writeln('If you create a tree now, you will remove ');
         writeln('the resident tree from memory.');
         writeln;
         writeln('Do you still want to create a tree? (Y/N)');
         ans:=UpCase(readkey);
      Until ans in ['Y','N'];
      If ans='Y' Then Begin
         If rt <> nil Then killtree(rt);
         flname:='new.inp';
         typ:='INP';
         Repeat
            writeln;
            writeln('How tall would you like your new tree?');
            writeln('   Please enter a real number or S for 50 million.');
            num:=5E7;
            done:=true;
            readReal(num,done,dum);
         Until done;
         have:=true;
         new(rt);
         initme(rt);
         rt^.length:=0.0;
         rt^.ancestor:=nil;
         rt^.name:='rt';
         new(rt^.sibL);
         initme(rt^.sibL);
         rt^.sibL^.ancestor:=rt;
         rt^.sibL^.length:=num;
         rt^.sibL^.name:='lf';
         new(rt^.sibR);
         initme(rt^.sibR);
         rt^.sibR^.ancestor:=rt;
         rt^.sibR^.length:=num;
         rt^.sibR^.name:='ri';
         writeln;
         writeln('There, your new tree is made.');
         writeln('  Go to Edit to view and add to this',
                       ' marvelous creation.');
      End; {then}
      ret('Please press any key to return to the Main Menu.');
   End; {procedure createTree}

End. {unit T_IO1}
