{*************************************************************************}
{This unit has all the type definitions of the program, as well as some   }
{procedures used widely throughout the program PDTREE.  If two different  }
{units wish to share a procedure, they keep it in here.                   }
{I have added the t statistic routines here as well.  [PEM 3 Sept. 1996]  }
{*************************************************************************}

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
{This block defines the compiler switch settings}
{$A+} {Always align variables on a word boundary}
{$B-} {use short-circuit boolean evaluation}
{$D+} {include debugging information}
{$E-} {shouldn't need to include 80X87 emulation (assumed >= 80486 CPU)}
{$F-} {no need for far calls in this application}
{$G+} {always use 80286 instructions (assumed >=80486 CPU)}
{$I+} {keep I/O checking enabled (unless locally necessary)}
{$L+} {local symbol information (helps in debugging)}
{$N+} {always assume a coprocessor is present}
{$O-} {overlays are not used in this application}
{$R-} {range checking is off for this application}
{$S+} {leave stack checking on since a lot of recursive calls are used}
{$V+} {strict checking of strings passed as Var procedure parameters}
{$X-} {always accept the value of a function call}
{End compiler switch block}
{-------------------------------------------------------------------------}


Unit t_util1;

Interface

   Const
      cback = 0;       {Black}
      cnorm = 9;        {Blue}
      white = 1;
      chi   = 10;          {Green}
      ht1 = 0.76;

      sizex = 0.6;          {xy plot size}
      sizey = 0.88;          {xy plot size}
      oneline = 10;         {height of one line of text}

   Type
      flstring = string[30];
      typstring = string[3];
      namestring = string[2];
      traits = array[1..2] of extended;
      itraits = array[1..2] of integer;
      statpak = Record
         n1          : integer;   {Equals 1 if array[1] is on x axis where
                         array[1] holds data to be plotted i.e. nd^.dat[n1]}
         n2          : integer;
         min         : traits;  
         max         : traits;
         number      : integer;     {total number of data points}
         mean        : traits;
         variance    : traits;
         ci          : traits;
         se          : traits;
         cov         : extended;
         corr        : extended;
         rma         : extended;    {reduced major axis}
         rmay        : extended;    {reduced major axis y intercept}
         ma          : extended;    {major axis}
         may         : extended;    {major axis y intercept}
         ls          : extended;    {least squares slope}
         lsy         : extended;    {least square y intercept}
         r2          : extended;    {r squared value for least squares regression}
         ts          : extended;    {T statistic}
         f           : extended;    {F statistic}
         sumresid    : extended;    {sum of regression residuals}
         slopeCI     : extended;    {Confidence interval around ls}
         interCI     : extended;    {Confidence interval around lsy}
         slopeSE     : extended;
         interSE     : extended;
         signPvalue  : extended;    {Holds the p-value for the sign test}
         df          : integer;     {nominal degrees of freedom}
         polytomies  : integer;     {number of polytomies in the tree}
         dfReduce    : integer;     {User specified amount to reduce degrees of freedom}
         conypos     : integer;     {number of y axis contrasts > 0}
         cony0       : integer;     {number of y axis contrasts = 0}
         conyneg     : integer;     {number of y axis contrasts < 0}
         lsregok     : boolean;     {Equals TRUE if it is ok to do regression}
      end; {record statpak}
      nodeptr = ^node;

      {The tree contains 2*N-1 node records (N=# of tips).  Thus anything
       that makes this record larger reduces the maximum size of the
       tree.  That's why (at present) names are limited to two characters.}

      node = Record
         Ancestor : nodeptr;
         dat      : traits;  {Tip data}
         contrast : traits;  {As calculated by Felsensteins Method}
         temp     : extended;
         temp2    : extended;
         temp3    : extended;
         temp4    : extended;
         sdcon    : extended;    {Standard deviation of contrasts}
         variance : extended;    {Usually equal to branch length}
         length   : extended;
         SibL     : nodeptr;   {Children of node}
         SibR     : nodeptr;
         tipL     : nodeptr;   {Pointer to nodes within tree.  Originally
                                      tip to tip.}           
         tipR     : nodeptr;
         tipnum   : integer;  {Number assigned to tip node.  Upper Left
                             tip gets 1, next tip in traversal get 2 etc.}
         name     : namestring;
         tempx    : integer;   {graphics stuff}
         tempy    : integer;
      end; {record node}

   Var width1,width2 :extended;

   Procedure preparetips(rt:nodeptr; var total:integer; var c:nodeptr);
   Procedure tipstat(nd:nodeptr;
                     var stpk:statpak);
   Procedure dumpdown(x,y:integer;st:string);
   Procedure tempHeight(nd:nodePtr; ht:extended);
   Procedure temp2Height(nd:nodePtr; ht:extended);
   Procedure tempAveHeight(nd:nodePtr; ht:extended);
   Procedure temp3CorrectedHeight(nd:nodePtr; rt: nodePtr; ht:extended);
   Procedure temp4AveCorrectedHeight(nd:nodePtr; rt: nodePtr; ht:extended);
   Procedure checkAllZero(nd:nodePtr; Var allZero:boolean);
   Procedure checkBranchTooLong(nd:nodePtr;Var tooLong:boolean);
   Procedure plotline(m,y0       :extended;
                      min,max,sf :traits;
                      origin     :itraits;
                      n1,n2      :integer);
   Procedure plotZeroLabel(min    : traits;
                           sf     : traits;
                           origin : itraits;
                           axis   : integer);
   Procedure showaxes(origin:itraits;n1,n2,mx,my:integer);
   Procedure getnumber(var n1,n2:integer);
   Procedure star2(coor:itraits);
   Procedure star(coor:itraits);
   Procedure openCircle(coor:itraits);
   Procedure xSymbol(coor:itraits);
   Procedure initplot(n1,n2:integer;min,max:traits;filename:flstring;
                      var scalingfactor:traits;var origin:itraits;
                      var mx,my: integer);
   Procedure initplotx0(n1,n2:integer;min,max:traits;filename:flstring;
                        var scalingfactor:traits;var origin:itraits;
                        var mx,my: integer);
   Procedure initploty0(n1,n2:integer;min,max:traits;filename:flstring;
                        var scalingfactor:traits;var origin:itraits;
                        var mx,my: integer);
   Procedure finishstat(var stpk:statpak);     
   Procedure finishstatorg(var stpk:statpak);
   Procedure dumpstat(stpk:statpak;
                      mx,my:integer;
                      Var bail : boolean;
                      contrasts : boolean);
   Procedure initme(me:nodeptr);
   Procedure killtree(var nod:nodeptr);
   Procedure ret(message:string);
   Procedure CheckName(nod:nodeptr; name:namestring;var notfound:boolean;
                                var want:nodeptr);
   Procedure whiteCheck(nd,rt: nodePtr);
   Procedure numtps(nod:nodeptr; Var num:integer);
   Procedure ReadReal(var num: extended;var succ,l:boolean);
   Procedure input1(var mul1:extended; posx,posy:integer);
   Procedure readInteger(var I1: integer; posx,posy: integer);
   Procedure loadgraphics;
   Function binomial050(N,R : integer) : extended;
   Function signTest(N1,N2: integer) : extended;
   Function tvalue(t :extended; df : integer) : extended;
   Function tCI(df : integer; CI : extended) : extended;
   Procedure extendWrite(Var outFile :text;num :extended);
   Procedure extendStr(Var textBuf :string; num :extended);

Implementation

Uses crt,
     graph,
     dos;


{**********************************************************************}
{This links in the egavga driver.                                      }
{**********************************************************************}
procedure EgaVgaDriverProc; external;
{$L EGAVGA.OBJ }

{**********************************************************************}
{This procedure loads in the graphics driver information and registers }
{the egavga graphics driver.                                           }
{**********************************************************************}
Procedure loadgraphics;
   Begin
      If (RegisterBGIdriver(@EgaVgaDriverProc) < 0) Then Begin
         clrscr;
         writeln('Unable to register the device drivers.');
         writeln('When you do something that uses the graphics screen,');
         writeln('the program will crash.');
         writeln;
         writeln;
         ret('Press any key to continue.');
      End; {then}
   End; {procedure loadgraphics}

{*************************************************************************}
{This procedure initializes some of the fields of a node to 0.0 or nil.   }
{It is called just about whenever I 'new' a pointer.  It calls nothing.   }
{*************************************************************************}
Procedure initMe(me:nodeptr);
   Begin
      me^.dat[1]:=0.0;
      me^.dat[2]:=0.0;
      me^.variance := -9;
      me^.sibL:=nil;
      me^.sibR:=nil;
   End; {initme}  
{*************************************************************************}
{This recursive procedure deallocates the linked list implemented tree.   }
{This is done to keep the heap neat.  It calls nothing but itself.        }
{*************************************************************************}
Procedure killTree(var nod:nodeptr);
   Begin
      If nod^.sibL<>nil
      Then killtree(nod^.sibL);
      If nod^.sibR<>nil
      Then killtree(nod^.sibR);
      dispose(nod);
   End; {killTree}
{*************************************************************************}
{This just prints a message to the bottom of the screen and then          }
{inputs a keypress.  It is called from all over, and calls nothing.       }
{*************************************************************************}
Procedure ret(message:string);
   Var
      ans:char;
   Begin
      gotoxy(1,23);
      writeln;
      writeln(message);
      ans:=readkey;
   End; {procedure ret}
{*************************************************************************}
{This recursive procedure checks a name that is passed with the rest of   }
{the tree to see if it is already in the tree.  If it is, it sets notfound}
{to false.  It calls nothing but itself.                                  }
{*************************************************************************}
Procedure CheckName(nod:nodeptr; name:namestring;var notfound:boolean;
                                var want:nodeptr);
   Begin
      If nod^.name=name Then Begin
         notfound:=false;
         want:=nod;
      End {then}
      Else Begin
         If (nod^.sibL<>nil)
         Then checkname(nod^.sibL,name,notfound,want);
         If (nod^.sibR<>nil) and notfound
         Then checkname(nod^.sibR,name,notfound,want);
      End; {else}
   End; {procedure checkname}
{*************************************************************************}
{This procedure checks that a name does not contain whitespace.  If if    }
{does, it prompts the user to enter a new name.  This needs to be called  }
{after the tree is loaded, so that whiteCheck can call checkName to avoid }
{duplications.                                                            }
{*************************************************************************}
Procedure whiteCheck(nd,rt: nodePtr);
   Var want : nodeptr;
       notFound : boolean;
       newName : nameString;
   Begin
      If ((ord(nd^.name[1])<=32) Or
          (ord(nd^.name[2])<=32) Or
          (length(nd^.name)<2)) Then Begin
         writeln;
         If length(nd^.name)=2 Then Begin
            writeln('Error:  The two letter node name "',nd^.name,'" contains');
            writeln('non-printing characters.  Please enter a new name that does');
            write('not contain spaces or other non-printing characters:  ');
         End {then}
         Else Begin
            writeln('Error:  The node name "',nd^.name,'" is too short.  Please');
            writeln('enter a new name that contains exactly 2 printable characters:  ');
         End; {else}
         readln(newName);
         notFound := true;
         checkName(rt,newName,notFound,want);
         While (not(notFound) Or
                (ord(newName[1])<=32) Or
                (ord(newName[2])<=32) Or
                (length(newName)<2)) Do Begin
            If not(notFound)
            Then writeln('The name you entered is in use by another node.')
            Else
              If length(newName)<2
              Then writeln('The name you entered is too short.')
              Else writeln('The name you entered has a non-printing character.');
            write('Please enter a new name: ');   
            readln(newName);
            notFound := true;
            checkName(rt,newName,notFound,want);
         End; {while}
         nd^.name := newName; {assign the new name to the node}
      End; {then}
      If (nd^.sibL<> nil)
      Then whiteCheck(nd^.sibL,rt);
      If (nd^.sibR<> nil)
      Then whiteCheck(nd^.sibR,rt);
   End; {procedure whiteCheck}
{*************************************************************************}
{This recursive procedure returns the number of tips descended from the   }
{node it is given.  It calls nothing but itself.                          }
{*************************************************************************}
Procedure numtps(nod:nodeptr; Var num:integer);
   Begin
      If nod^.sibL<>nil Then Begin
         numtps(nod^.sibL,num);
         numtps(nod^.sibR,num);
      End
      Else Begin
         inc(num);
      End;
   End; {procedure numtps}

{*************************************************************************}
{This procedure tries to read in a real number.  If a valid real number is}
{read then num is changed and num gets the number and succ is true.       }
{If "s" is input then num is not changed, l is false, and succ is true.   }
{If "l" is input then succ=true, l=true, num is not changed.  So if       }
{"s", "l", or a real number are entered by the user, the procedure says   }
{it was successful.  It calls nothing.                                    }
{*************************************************************************}
Procedure readReal(var num: extended;var succ,l:boolean);
   Var
      s :string;
      code: integer;
   Begin
      succ:=true;
      l:=false;
      Readln(s);
      If (s<>'s') and (s<>'S') and (s<>'l') and (s<>'L') Then Begin
         val(s,num,code);
         If code<>0
         Then succ:=false
      End; {then}
      If (s='l') or (s='L')
      Then l:=true;
   End; {procedure readReal}
{*************************************************************************}
{This procedure inputs one real number from the spot posx,posy on the
keyboard.}
{*************************************************************************}
Procedure input1(var mul1:extended; posx,posy:integer);
  Var
    succ,l :boolean;
  Begin
     Repeat
        gotoxy(posx,posy);
        clreol;
        gotoxy(posx,posy);
        readreal(mul1,succ,l);
     Until((succ) and not(l));
  End; {procedure input1}

{**************************************************************************}
{This procedure moves the cursor to posx, posy on the screen.  It then
 attempts to read one integer value from the keyboard (echoing to the
 screen).  If the conversion is successful, I1 will contain the converted
 value.  Otherwise, the echoed string is erased and the loop reads again. }
{ **************************************************************************}
Procedure readInteger(var I1: integer; posx,posy: integer);
   Var
      s : string;
      valCode : integer;
   Begin
      valCode := 1;
      Repeat
         gotoxy(posx,posy);
         clreol;
         readln(s);
         val(s,I1,valCode);
      Until(valCode = 0)
   End; {procedure readInteger}

{*************************************************************************}
{This procedure prompts the user to choose which trait goes along the X
axis for graphing procedures.  It then reads in a number 1 or 2.}
{*************************************************************************}
  Procedure getNumber(var n1,n2:integer);
     Var
        ch:char;
     Begin
        write('Which trait would you like along the horizontal axis (1 (=Enter) or 2)? ');
        ch:=readkey;
        writeln;
        If ch<>'2' Then n1:=1 Else n1:=2;
        If n1=1 Then n2:=2 Else n2:=1;
     End; {procedure getnumber}

{*************************************************************************}
{This procedure draws the labels along the X and Y axis for the graphing
procedures.}
{*************************************************************************}
  Procedure showAxes(origin:itraits;n1,n2,mx,my:integer);
     Var
        dum:itraits; {dummy location ???}
     Begin
        dum[1]:=round((mx-origin[1]-50)/2)+origin[1];
        dum[2]:=round((origin[2]-7*oneline)/2);
        If n1=1 Then Begin
           outtextxy(dum[1],round(1.05*origin[2])+3,'Trait 1 = X');
           dumpdown(round(0.9524*origin[1])-8,dum[2],'Trait 2 = Y');
        End {then}
        Else Begin
           outtextxy(dum[1],round(1.05*origin[2])+3,'Trait 2 = X');
           dumpdown(round(0.952*origin[1])-8,dum[2],'Trait 1 = Y');
        End; {else}
     End; {procedure showAxes}
{*************************************************************************}
{This procedure outputs the tree name, the date, and the time.  It figures
out the date and time from dos calls.}
{*************************************************************************}
   Procedure dateAndTime(filename:flstring);
      Var
         yr,mo,day,dow,h,m,s,s100:word;
         yr2,day2,h2,m2,s2:string[5];
         mo2:string[10];
      Begin
         outtextxy(0,46*oneline,'Tree name: '+filename);
         getdate(yr,mo,day,dow);
         gettime(h,m,s,s100);
         str(yr,yr2);
         Case mo of
            1: mo2:='January';
            2: mo2:='February';
            3: mo2:='March';
            4: mo2:='April';
            5: mo2:='May';
            6: mo2:='June';
            7: mo2:='July';
            8: mo2:='August';
            9: mo2:='September';
           10: mo2:='October';
           11: mo2:='November';
           12: mo2:='December';
         End; {case mo}
         str(day,day2);
         str(h,h2);
         str(m,m2);
         If m<10 Then m2 := '0'+m2;
         str(s,s2);
         If s<10 Then s2 := '0'+s2;
         outtextxy(0,47*oneline,day2+' '+mo2+' '+yr2+' '+h2+':'+
                       m2+':'+s2);
      End; {procedure dataAndTime}
{*************************************************************************}
{This is the first of 3 procedures that all do the same thing, except this
one assumes that the x axis is at zero, the next assumes that the y axis is
zero, and the third has no such assumptions.

They figure out the scaling factor, the origin, then draw the axes, and
call the date-time procedure.}
{*************************************************************************}
  Procedure initPlotx0(n1,n2:integer;
                       min,max:traits;
                       filename:flstring;
                       Var scalingfactor:traits;
                       Var origin:itraits;
                       Var mx,my: integer);
     Begin
        If (max[1]-min[1])>0
        Then scalingfactor[1]:=mx*sizex/(max[n1]-min[n1])
        Else scalingfactor[1]:=1;
        If (max[2]-min[2])>0
        Then scalingfactor[2]:=my*sizey/(max[n2]-min[n2])
        Else scalingfactor[2]:=1;
        origin[1]:=round(mx*(1-sizex))-3;
        origin[2]:=round(my*sizey)+3;

        moveto(round(0.9524*origin[1]),round(origin[2]));
        lineto(round(0.9524*origin[1]),0);
        moveto(round(0.9524*origin[1]),round(origin[2]));
        lineto(mx,round(origin[2]));
        outtextxy(0,46*oneline,'Tree name: '+filename);
        moveto(round(0.9524*origin[1])-9,origin[2]);
        lineto(round(0.9524*origin[1]),origin[2]);
        outtextxy(round(0.9524*origin[1])-16,origin[2]-round(0.4*oneline),'0');
        dateandtime(filename);
     End; {procedure initPlotX0}
{*************************************************************************}
{(see initplotx0 above)}
{*************************************************************************}
  Procedure initPlotY0(n1,n2:integer;
                       min,max:traits;
                       filename:flstring;
                       Var scalingfactor:traits;
                       Var origin:itraits;
                       Var mx,my: integer);
     Begin
        If (max[1]-min[1])>0
        Then scalingfactor[1]:=mx*sizex/(max[n1]-min[n1])
        Else scalingfactor[1]:=1;
        If (max[2]-min[2])>0
        Then scalingfactor[2]:=my*sizey/(max[n2]-min[n2])
        Else scalingfactor[2]:=1;
        origin[1]:=round(mx*(1-sizex))-3;
        origin[2]:=round(my*sizey)+3;

        moveto(round(origin[1]),round(1.05*origin[2]));
        lineto(round(origin[1]),0);
        moveto(round(origin[1]),round(1.05*origin[2]));
        lineto(mx,round(1.05*origin[2])); 
        dateandtime(filename); 
     End; {procedure initPlotY0}
{*************************************************************************}
{(see initplotx0 above)}
{*************************************************************************}
  Procedure initPlot(n1,n2:integer;
                     min,max:traits;
                     filename:flstring;
                     Var scalingfactor:traits;
                     Var origin:itraits;
                     Var mx,my: integer);
     Begin
        If (max[n1]-min[n1])>0
        Then scalingfactor[1]:=mx*sizex/(max[n1]-min[n1])
        Else scalingfactor[1]:=1;
        If (max[n2]-min[n2])>0
        Then scalingfactor[2]:=my*sizey/(max[n2]-min[n2])
        Else scalingfactor[2]:=1;
        origin[1]:=round(mx*(1-sizex))-3;
        origin[2]:=round(my*sizey)+3;

        moveto(round(0.9524*origin[1]),round(1.05*origin[2]));
        lineto(round(0.9524*origin[1]),0);
        moveto(round(0.9524*origin[1]),round(1.05*origin[2]));
        lineto(mx,round(1.05*origin[2]));
        dateandtime(filename);
     End; {initPlot}
{*************************************************************************}
{This calculates the rma and the y intercept for rma called rmay, ma and
the y intercept for the ma called may, r squared called r2, t statistic
called ts, the f statistic, and the degrees of freedom called df.  Most of
this is based on Chapters 14-15 of Sokal and Rolf 1981. }
{*************************************************************************}
  Procedure finishStat(var stpk:statpak);
     Var
        test1 :extended;
        d : extended;
        lambda1 : extended;
     Begin
        With stpk Do Begin
           If variance[n1]>0 Then Begin
              rma := sqrt(variance[n2]/variance[n1]);
              If ((corr>0) and (rma<0)) or
                  ((corr<0) and (rma>0))
              Then rma := rma*-1;
              rmay := mean[n2]-rma*mean[n1];
           End; {then}
           d := sqrt(sqr(variance[n1] + variance[n2]) -
                  4*(variance[n1]*variance[n2]-sqr(cov)));
           lambda1 := (variance[n1] + variance[n2]+d)/2;
           {careful, Sokal and Rolf 1981 set Y1 for vertical axis, for
            example on 595-596.  PDTREE uses n1 for horizontal axis.
            The difference turns up in these expressions. PEM 25 Sept 2001}
           If (lambda1 = variance[n2])  {divide by zero check}
           Then If (cov > 0)
                Then ma := 9.99E99   {Your favorite estimate of infinity here}
                Else ma := -9.99E99
           Else ma := cov/(lambda1-variance[n2]);
           stpk.may := mean[n2] - ma*mean[n1];
           r2:=sqr(corr);
           If (r2<1) Then Begin
              ts := corr*sqrt((number-2)/(1-r2));
              f := sqr(ts);
           End {then}
           Else Begin
              ts := 9.99E99;  {estimate of infinity}
              f := 9.99E99;
           End; {else}
           df := number-2;
        End {with stpk}
     End; {procedure finishStat}    

{*************************************************************************}
{This is similar to the procedure above, except the rma and ma are forced
to go through the origin.
This calculates the rma and the y intercept for rma called rmay, ma and
the y intercept for the ma called may, r squared called r2, t statistic
called ts, f statistic, and the degrees of freedom called df.}
{*************************************************************************}
  Procedure finishStatOrg(var stpk:statpak);
     Var
        d : extended;
        lambda1 : extended;
     Begin
        With stpk Do Begin
           If corr=0 Then lsregok:=false;
           If lsregok=true Then rma := ls/corr;
           If ((corr>0) and (rma<0)) Or
               ((corr<0) and (rma>0))
           Then rma := rma*-1;
           rmay:=0.0;
           {These lines are the old ma code.}
           {ma := sqr(variance[n1]+variance[n2]) -
                    4*(variance[n2]*variance[n1] - sqr(cov));
           ma := 0.5*(variance[n2]+variance[n1]+sqrt(ma));
           ma := cov/(ma-variance[n2]);}
           {The variances here are calculated with the mean set to zero.
            This constitutes the difference between this ma and that
            calculated by finishstat.  The formula from Sokal and Rolf
            is used unchanged.  The intercept is just set to zero.  }
           d := sqrt(sqr(variance[n1] + variance[n2]) -
                  4*(variance[n1]*variance[n2]-sqr(cov)));
           lambda1 := (variance[n1] + variance[n2]+d)/2;
           If (lambda1 = variance[n2])  {divide by zero check}
           Then If (cov > 0)
                Then ma := 9.99E99   {Your favorite estimate of infinity here}
                Else ma := -9.99E99
           Else ma := cov/(lambda1-variance[n2]);
           may := 0.0;
           r2 := sqr(corr);
           If (r2<>1) Then Begin
              ts := corr*sqrt((number-1)/(1-r2));
              f := sqr(ts);
           End {then}
           Else Begin
              ts := 9.99E99;  {estimate of infinity}
              f := 9.99E99;
           End; {else}
           df := number-2;
        End {with stpk}
     End; {procedure finishStatOrg}
{*************************************************************************}
{This procedure dumps the statistics given to it to the graphics screen.
It is used to add the statistics output to the graphs.}
{*************************************************************************}  
  Procedure dumpstat(stpk:statpak;
                     mx,my:integer;
                     var bail : boolean;
                     contrasts : boolean);
     Var
        t:string;
        ch:char;
     Begin
        str(stpk.number,t);
        If contrasts
        Then outtextxy(0,9*oneline,'Number of contrasts: '+t)
        Else outtextxy(0,9*oneline,'Number of data points: '+t);
        extendStr(t,stpk.min[stpk.n1]);
        outtextxy(0,11*oneline,'Min X:'+t);
        extendStr(t,stpk.min[stpk.n2]);
        outtextxy(0,12*oneline,'Min Y:'+t);
        extendStr(t,stpk.max[stpk.n1]);
        outtextxy(0,14*oneline,'Max X:'+t);
        extendStr(t,stpk.max[stpk.n2]);
        outtextxy(0,15*oneline,'Max Y:'+t);
        extendStr(t,stpk.mean[stpk.n1]);
        outtextxy(0,17*oneline,'Mean X:'+t);
        extendStr(t,stpk.mean[stpk.n2]);
        outtextxy(0,18*oneline,'Mean Y:'+t);
        extendStr(t,stpk.variance[stpk.n1]);
        outtextxy(0,20*oneline,'Var X:'+t);
        extendStr(t,stpk.variance[stpk.n2]);
        outtextxy(0,21*oneline,'Var Y:'+t);
        extendStr(t,stpk.cov);
        outtextxy(0,23*oneline,'Cov:'+t);
        extendStr(t,stpk.corr);
        outtextxy(0,24*oneline,'Corr:'+t);
        outtextxy(0,26*oneline,'Reduced Major Axis.');
        If (stpk.lsregok) Then Begin
           extendStr(t,stpk.rma);
           outtextxy(0,27*oneline,'     Slope:'+t);
           extendStr(t,stpk.rmay);
           outtextxy(0,28*oneline,' Intercept:'+t);
        End {then}
        Else Begin
           outtextxy(0,27*oneline,'     Slope: undefined');
           outtextxy(0,28*oneline,' Intercept: undefined');
        End; {else}
        outtextxy(0,30*oneline,'Major Axis.');
        extendStr(t,stpk.ma);
        outtextxy(0,31*oneline,'     Slope:'+t);
        extendStr(t,stpk.may);
        outtextxy(0,32*oneline,' Intercept:'+t);
        outtextxy(0,34*oneline,'Least Squares Regression.');
        If (stpk.lsregok) Then Begin
           extendStr(t,stpk.ls);
           outtextxy(0,35*oneline,'     Slope:'+t);
           extendStr(t,stpk.lsy);
           outtextxy(0,36*oneline,' Intercept:'+t);
           extendStr(t,stpk.r2);
           outtextxy(0,37*oneline,' R squared:'+t);
        End {then}
        Else Begin
           outtextxy(0,35*oneline,'     Slope: undefined');
           outtextxy(0,36*oneline,' Intercept: undefined');
           outtextxy(0,37*oneline,' R squared: undefined');
        End; {else}
        If (abs(stpk.corr)<1)
        Then Begin
           extendStr(t,stpk.ts);
           outtextxy(0,39*oneline,'t: '+t);
           extendStr(t,stpk.f);
           outtextxy(0,40*oneline,'F: '+t);
        End {then}
        Else Begin
           outtextxy(0,39*oneline,'t: undefined');
           outtextxy(0,40*oneline,'F: undefined');
        End; {else}
        str(stpk.df,t);
        outtextxy(0,41*oneline,'d.f.: '+t);
        ch:=readkey;
        If ((ch = 'R') or (ch = 'r'))
        Then bail := true
        Else bail := false;
        Closegraph;
     End; {procedure dumpstat}
{***************************************************************************}
{This procedure outputs a star shape to the graphics screen.  It is used
by the graphing features of PDTree to put a data point on the screen.}
{***************************************************************************}
Procedure star(coor:itraits);
   Begin
      line(coor[1]-1,coor[2]-2,coor[1]+1,coor[2]-2);
      line(coor[1]-2,coor[2]-1,coor[1]+2,coor[2]-1);
      line(coor[1]-2,coor[2],coor[1]+2,coor[2]);
      line(coor[1]-2,coor[2]+1,coor[1]+2,coor[2]+1);
      line(coor[1]-1,coor[2]+2,coor[1]+1,coor[2]+2);
   End; {procedure star}
{***************************************************************************}
{This procedure outputs a filled circle to the graphics screen.  It is used
by the graphing features of PDTree to put a data point on the screen.}
{***************************************************************************}
Procedure openCircle(coor:itraits);
   Begin
      line(coor[1]-1,coor[2]+3,coor[1]-3,coor[2]+1);
      line(coor[1]-3,coor[2]+1,coor[1]-3,coor[2]-1);
      line(coor[1]-3,coor[2]-1,coor[1]-1,coor[2]-3);
      line(coor[1]-1,coor[2]-3,coor[1]+1,coor[2]-3);
      line(coor[1]+1,coor[2]-3,coor[1]+3,coor[2]-1);
      line(coor[1]+3,coor[2]-1,coor[1]+3,coor[2]+1);
      line(coor[1]+3,coor[2]+1,coor[1]+1,coor[2]+3);
      line(coor[1]+1,coor[2]+3,coor[1]-1,coor[2]+3);
   End; {procedure star}

{***************************************************************************}
{This procedure outputs an 'X' shape to the graphics screen.  It is used
by the graphing features of PDTree to put a data point on the screen.}
{***************************************************************************}
  Procedure xSymbol(coor:itraits);
    Begin
       setLineStyle(SolidLn,0,ThickWidth);
       line(coor[1]-3,coor[2]-3,coor[1]+3,coor[2]+3);
       line(coor[1]-3,coor[2]+3,coor[1]+3,coor[2]-3);
       setLineStyle(SolidLn,0,NormWidth);
    End; {procedure star}

{*************************************************************************}
{This procedure draws a cross or plus shape to the graphics screen.  It
 is used by the graphing features of PDTree to put a data point on the
 screen.}
{*************************************************************************}
  Procedure star2(coor:itraits);
     Begin
        line(coor[1]-3,coor[2],coor[1]+3,coor[2]);
        line(coor[1],coor[2]-3,coor[1],coor[2]+3);
        line(coor[1]-1,coor[2]-1,coor[1]-1,coor[2]+1);
        line(coor[1]+1,coor[2]-1,coor[1]+1,coor[2]+1);
     End; {procedure star2}

{*************************************************************************}
{This procedure plots the line given to it in the box on the graphics screen.
The line it is given is in the form slope "m" and y intercept "y0".
Min and max denote the bounds of the data points that fit on the screen's
window.  sf is a scaling factor to translate from the real valued data points
to pixels on screen.  Origin is the pixel coordinates of the lower left
hand corner of the display window.  n1 and n2 determine which data dimension
goes along which axis.}
{*************************************************************************}
  Procedure plotLine(m,y0:extended;
                     min,max,sf:traits;
                     origin:itraits;
                     n1,n2:integer);
     Var
        x1,x2,y1,y2:extended;
        success,one:boolean;
        dum1,dum2:extended;
     Begin
        success:=false;
        one:=false;
        x1:=min[n1];
        y1:=m*x1+y0;
        If (y1<max[n2]) and (y1>min[n2])
        Then one:=true;
        dum1:=max[n1];
        dum2:=m*dum1+y0;
        If (dum2<max[n2]) And (dum2>min[n2])
        Then Begin
           If one Then Begin
              success:=true;
              x2:=dum1;
              y2:=dum2;
           End {then}
           Else Begin
              one:=true;
              x1:=dum1;
              y1:=dum2;
           End; {else}
        End; {then}
        If (m<>0) And (Not success)
        Then Begin
           dum2:=min[n2];
           dum1:=(dum2-y0)/m;
           If (dum1<max[n1]) And (dum1>min[n1]) Then Begin
              If one Then Begin
                 success:=true;
                 x2:=dum1;
                 y2:=dum2;
              End {then}
              Else Begin
                 one:=true;
                 x1:=dum1;
                 y1:=dum2;
              End; {else}
           End; {then}
           dum2:=max[n2];
           dum1:=(dum2-y0)/m;
           If (dum1<max[n1]) And (dum1>min[n1]) And one Then Begin
              success:=true;
              x2:=dum1;
              y2:=dum2;
           End; {then}
        End; {then}
        If success Then Begin
           moveto(origin[1]+round(sf[1]*(x1-min[n1])),
                   origin[2]-round(sf[2]*(y1-min[n2])));
           lineto(origin[1]+round(sf[1]*(x2-min[n1])),
                   origin[2]-round(sf[2]*(y2-min[n2])));
        End; {then}
     End; {procedure plotLine}

{***********************************************************************}
{This procedure plots a '0' at the zero point on the vertical axis for  }
{screen 9.                                                              }
{***********************************************************************}
  Procedure plotZeroLabel(min    : traits;
                          sf     : traits;
                          origin : itraits;
                          axis   : integer);
     Begin
        outtextxy(origin[1]-8,(origin[2]+round(sf[2]*min[axis]))-4,'0');
     End; {procedure plotLine}

{***********************************************************************}
{This procedure writes the text on the screen vertically starting at the
x,y coordinates given.}
{***********************************************************************}
  Procedure dumpdown(x,y:integer;st:string);
     Begin
        While (length(st)>0) Do Begin
           outtextxy(x,y,copy(st,1,1));
           delete(st,1,1);
           y:=y+oneline;
        End; {while}
     End; {procedure dumpdown}

{**************************************************************************}
{This recursive procedure is passed a node, and the height of its ancestor.
It then fills in the height of the node and goes further up the tree until
all the nodes have their height put in the temp field.}
{**************************************************************************}
Procedure tempHeight(nd:nodePtr; ht:extended);
   Begin
      nd^.temp:=ht+nd^.length;
      If nd^.sibL<>nil Then Begin
         tempHeight(nd^.sibL,nd^.temp);
         tempHeight(nd^.sibR,nd^.temp);
      End; {then}
   End; {procedure tempheight}
{**************************************************************************}
{This recursive procedure is passed a node, and the height of its ancestor.
It then fills in the height of the node and goes further up the tree until
all the nodes have their height put in the temp field.}
{**************************************************************************}
Procedure temp2Height(nd:nodePtr; ht:extended);
   Begin
      nd^.temp2:=ht+nd^.length;
      If nd^.sibL<>nil Then Begin
         temp2Height(nd^.sibL,nd^.temp2);
         temp2Height(nd^.sibR,nd^.temp2);
      End; {then}
   End; {procedure temp2height}

{**************************************************************************}
{This recursive procedure is passed a node, and the height of its ancestor.
It then fills in the temp field with the average height of the siblings of
this node.  If the siblings of this node has siblings, then it recurses up
the tree.}
{**************************************************************************}
Procedure tempAveHeight(nd:nodePtr; ht:extended);
   Begin
      ht:=ht+nd^.length;
      nd^.temp:=ht+(nd^.sibL^.length+nd^.sibR^.length)/2;
      If nd^.sibL^.sibL<>nil
      Then tempAveHeight(nd^.sibL,ht);
      If nd^.sibR^.sibR<>nil
      Then tempAveHeight(nd^.sibR,ht);
   End; {procedure tempaveheight}


{*************************************************************************}
{The following two procedures are like the preceeding height procedureds,
 except that they fill in "contrast corrected" heights, rather than simple
 heights based on lengths. }
{*************************************************************************}

{*************************************************************************}
{The corrected length version of tempHeight.  Like tempHeight, the sum of
 corrected lengths of nodes leading back to the root is assigned to the
 node's temp field.}
{*************************************************************************}
Procedure temp3CorrectedHeight(nd:nodePtr; rt: nodePtr; ht:extended);
   Begin
      If nd <> rt
      Then nd^.temp3 := ht+nd^.variance
      Else Begin
         nd^.temp3 := 0;
      End;
      If nd^.sibL<>nil Then Begin
         temp3CorrectedHeight(nd^.sibL,rt,nd^.temp3);
         temp3CorrectedHeight(nd^.sibR,rt,nd^.temp3);
      End; {then}
   End; {procedure tempCorrectedHeight}


{*************************************************************************}
{The corrected length version of tempAveHeight.  Like
 tempAve2Height, the average of the heights of the node's two daughters
 is assigned to the node's temp2 field.}
{*************************************************************************}
Procedure temp4AveCorrectedHeight(nd:nodePtr; rt: nodePtr; ht:extended);
   Begin
      If nd <> rt
      Then ht:=ht+nd^.variance
      Else Begin
         ht:= 0;
      End;
      nd^.temp4:=ht+(nd^.sibL^.variance+nd^.sibR^.variance)/2;
      If nd^.sibL^.sibL<>nil
      Then temp4AveCorrectedHeight(nd^.sibL,rt,ht);
      If nd^.sibR^.sibR<>nil
      Then temp4AveCorrectedHeight(nd^.sibR,rt,ht);
   End; {procedure tempaveheight}


{**************************************************************************}
{This procedure checks if all the internal branch lengths are zero or not.}
{**************************************************************************}
Procedure checkAllZero(nd:nodePtr; Var allZero:boolean);
   Begin
      If nd^.sibL<>nil Then Begin
         If nd^.length=0 Then Begin
            checkAllZero(nd^.sibL,allZero);
            checkAllZero(nd^.sibR,allZero);
         End {then}
         Else allZero:=false;
      End; {then}
   End; {procedure checkAllZero}

{**************************************************************************}
{This check for any branch lengths greater than 1e9}
{**************************************************************************}
Procedure checkBranchTooLong(nd:nodePtr;Var tooLong:boolean);
   Begin
      If (nd^.length>1e9) Then toolong:=true;
      If nd^.sibL<>nil Then Begin
         checkbranchtoolong(nd^.sibL,toolong);
         checkbranchtoolong(nd^.sibR,toolong);
      End {then}
   End; {procedure checkbranchtoolong}

{*******************************************************************}
{This procedure calculates the statistics for the tip data.  This
procedure is laid out just like the ones in T_Felsen.
{*******************************************************************}
  Procedure tipStat(nd:nodeptr;
                    var stpk:statpak);
     Var
        sy,sx,sx2,sxy,d :extended;
        rt:nodeptr;
        dum:integer;
     {*******************************************************************}
     {getbounds determines the max and min.                              }
     {*******************************************************************}
     Procedure getbounds(var min,max:traits; nd:nodeptr);
        Begin
           If nd^.dat[1]>max[1] Then max[1]:=nd^.dat[1];
           If nd^.dat[1]<min[1] Then min[1]:=nd^.dat[1];
           If nd^.dat[2]>max[2] Then max[2]:=nd^.dat[2];
           If nd^.dat[2]<min[2] Then min[2]:=nd^.dat[2];
           If nd^.tipR<>nil Then getbounds(min,max,nd^.tipR);
        end; {getbounds}
     {*******************************************************************}
     {getmean summs up each variable and counts them.  From this one can
     {easily determine the mean.                                         }
     {*******************************************************************}
     Procedure getmean(nd:nodeptr;var n:integer;var m:traits);
        Begin
           inc(n);
           m[1]:=m[1]+nd^.dat[1];
           m[2]:=m[2]+nd^.dat[2];
           If nd^.tipR<>nil
           Then getmean(nd^.tipR,n,m);
        End; {getmean}
     {*******************************************************************}
     {getvar summs up the variance and covaraince deviations from which
     the variance and covariance can easily be determined.               }
     {*******************************************************************}
     Procedure getvar(nd:nodeptr; var s:statpak);
        Begin
           s.variance[1]:=s.variance[1]+sqr(s.mean[1]-nd^.dat[1]);
           s.variance[2]:=s.variance[2]+sqr(s.mean[2]-nd^.dat[2]);
           s.corr:=s.corr+(s.mean[1]-nd^.dat[1])*(s.mean[2]-nd^.dat[2]);
           If nd^.tipR<>nil
           Then getvar(nd^.tipR,s);
        End; {getvar}
     {*******************************************************************}
     {getsumms gets the summs needed for the least squares regression.
      Here regressions are not done through the origin.}
     {*******************************************************************}
     Procedure getsumms(n1,n2:integer; var sx,sy,sx2,sxy:extended;nd:nodeptr);
        Begin
           sx:=sx + nd^.dat[n1];
           sx2:=sx2 + sqr(nd^.dat[n1]);
           sy:=sy + nd^.dat[n2];
           sxy:=sxy + nd^.dat[n1]*nd^.dat[n2];
           If nd^.tipR<>nil
           Then getsumms(n1,n2,sx,sy,sx2,sxy,nd^.tipR);
        End; {getsumms}   
  {*******************************************************************}
     Begin
        If nd^.sibL<>nil Then Begin
           rt:=nd;
           preparetips(rt,dum,nd);
        End; {then}
        stpk.max[1]:=nd^.dat[1];
        stpk.min[1]:=stpk.max[1];
        stpk.max[2]:=nd^.dat[2];
        stpk.min[2]:=stpk.max[2];
        getbounds(stpk.min,stpk.max,nd);
        getnumber(stpk.n1,stpk.n2);
        stpk.number:=0;
        stpk.mean[1]:=0;
        stpk.mean[2]:=0;
        getmean(nd,stpk.number,stpk.mean); 
        stpk.mean[1]:=stpk.mean[1]/stpk.number;   
        stpk.mean[2]:=stpk.mean[2]/stpk.number;      
        stpk.variance[1]:=0;
        stpk.variance[2]:=0;
        stpk.corr:=0;
        getvar(nd,stpk);
        stpk.variance[1]:=stpk.variance[1]/(stpk.number-1);   
        stpk.variance[2]:=stpk.variance[2]/(stpk.number-1);  
        stpk.corr:=stpk.corr/(stpk.number-1);
        stpk.cov:=stpk.corr;
        If(stpk.variance[1]>0) and (stpk.variance[2]>0)
        Then
           stpk.corr:=stpk.corr/sqrt(stpk.variance[1]*stpk.variance[2])
        Else stpk.corr:=0;
        sx:=0;
        sy:=0;
        sx2:=0;
        sxy:=0;
        getsumms(stpk.n1,stpk.n2,sx,sy,sx2,sxy,nd);
        If ((stpk.number*sx2-sqr(sx))<>0) Then Begin
           stpk.lsregok:=true;
           d:=1/(stpk.number*sx2-sqr(sx));
        End {then}
        Else Begin
           stpk.lsregok:=false;
           d:=0;
        End; {else}
        stpk.ls:=d*(stpk.number*sxy-sx*sy);
        stpk.lsy:=d*(sx2*sy-sx*sxy);
        finishstat(stpk);
     End; {procedure tipStat}
{*****************************************************************}
{This procedure prepares the tree for the tip data editing.  It
sets the pointers from tip to tip, and also sets the tipnum field.}
{******************************************************************}
  Procedure prepareTips(rt:nodeptr; var total:integer; var c:nodeptr);
     var
        num :integer;
        previous,first :nodeptr;
  {****************************************************************}
  {This recursive procedure traverses the tree and sets the tip
   pointers tipR,tipL, and the tip number tipnum.}
  {****************************************************************}
     Procedure traverse(nd:nodeptr;
                        var prev,first:nodeptr;
                        var num:integer);
        Begin
           If (nd^.SibL <> NIL) Then Begin
              traverse(nd^.SibL,prev,first,num);
              traverse(nd^.SibR,prev,first,num);
           End {then}
           Else Begin
              inc(num);
              If (num=1)
              Then first:= nd;
              nd^.tipnum:=num;
              nd^.tipL:=prev;
              If (prev<>NIL)
              Then prev^.tipR:=nd;
              prev:=nd;
           End; {else}
        End; {procedure traverse}
  {****************************************************************}
     Begin
        num:=0;
        previous:=NIL;
        traverse(rt,previous,first,num);
        previous^.tipR:=NIL;
        total:= previous^.tipnum;
        c:=first
     End; {prepareTips}


{*************************************************************************}
{This procedure calculates symmetric (p = 0.5) binomial terms.  Note that }
{this symmetry simplifies two-tailed calculations and the pq power        }
{expressions, which are constant over the loop.  There are probably more  }
{expression optimizations in the C(n,r) code, but this is a little better }
{than calculating all three factorials outright. PEM 17 Mar 2003          }
{*************************************************************************}
   Function binomial050(N,R : integer) : extended;
      Function perm(N,R : integer) : extended;
         Var
            result: extended;
            count : integer;
         Begin
            result := 1;
            For count := N Downto R+1 Do
               result := result * count;
            perm := result;
         End; {local function perm}
      Var
         result : extended;
         log05 : extended;
         comb : extended;
         bino : extended;
         pnc : extended;
         f : extended;
         pwr : extended;
         count : integer;
      Begin
        result := 0;
        pwr := exp(N*ln(0.5));
        For count := 0 To R Do Begin
           pnc := perm(N,count);
           f := perm(N-count,1);
           comb := pnc/f;
           bino := pwr*comb;
           result := result + bino;
        End; {for}
        binomial050 := result
      End; {function binomial050}

{*************************************************************************}
{*************************************************************************}
   Function signTest(N1,N2: integer) : extended;
      Var
         count : integer;
      Begin
        If (N1 > N2)            {calculate the smaller tail value}
        Then signTest := binomial050(N1+N2,N2)
        Else signTest := binomial050(N1+N2,N1);
      End; {function signtest}



{*************************************************************************}
{Statistics (t and someday F) code below this line.  Much of this code is }
{taken from Press, W. H., Flannery, B. P., Teukolsky, S. A., and          }
{Vetterling, W. T. 1989. Numerical Recipes in Pascal. Cambridge, Cambridge}
{University Press.  The results produced by these routines differ         }
{from the values produced by SPSS/PC+ Version 5.0.  The values usually    }
{differ in the fourth decimal place.  We also compared results from the   }
{the routines used herein (from Numerical Recipes) with those produced by }
{an HP-21S calculator; they agree to more than 7 decimal places.          }
{*************************************************************************}


{************************************************************************}
{This function calculates the incomplete beta function.  It comes from   }
{page 188 of Numerical Recipes in Pascal.                                }
{************************************************************************}
   Function betai(a,b,x : extended) : extended;

      Var
         bt : extended;

      {*********************************************************************}
      {This function computes the natural log of the gamma function.  It    }
      {comes from page 177 of Numerical Recipes in Pascal.  The authors     }
      {claim the results from this function are good to 2e-10.              }
      {*********************************************************************}
      Function gammln(xx : extended): extended;
         Const
            stp = 2.50662827465;

         Var
            x,tmp,ser : extended;
         Begin
            x := xx-1.0;
            tmp := x+5.5;
            tmp := (x+0.5)*ln(tmp)-tmp;
            ser := 1.0+76.18009173/(x+1.0)-86.50532033/(x+2.0)+
                    24.01409822/(x+3.0)-1.231739516/(x+4.0)+
                    0.120858003e-2/(x+5.0)-0.536382e-5/(x+6.0);
            gammln := tmp+ln(stp*ser);
         End; {gammln}
      {*********************************************************************}
      {This function is computes the continued fraction helper function for }
      {the incomplete beta function.  It comes from pages 188-189 of        }
      {Numerical Recipes in Pascal.                                         }
      {*********************************************************************}
      Function betacf(a,b,x : extended) : extended;
         Label 99;
         Const
            itmax = 100;
            eps = 3.0e-7;
         Var
            tem,qap,qam,qab,em,d : extended;
            bz,bpp,bp,bm,az,app : extended;
            am,aold,ap: extended;
            m: integer;
         Begin
            am := 1.0;
            bm := 1.0;
            az := 1.0;
            qab := a+b;
            qap := a+1.0;
            qam := a-1.0;
            bz := 1.0-qab*x/qap;
            For m := 1 To itmax Do Begin
               em := m;
               tem := em+em;
               d := em*(b-m)*x/((qam+tem)*(a+tem));
               ap := az+d*am;
               bp := bz+d*bm;
               d := -(a+em)*(qab+em)*x/((a+tem)*(qap+tem));
               app := ap+d*az;
               bpp := bp+d*bz;
               aold := az;
               am := ap/bpp;
               bm := bp/bpp;
               az := app/bpp;
               bz := 1.0;
               If abs(az-aold) < eps*abs(az) Then Goto 99
            End; {for}
            writeln('pause in BETACF');
            writeln('a or b too big, or itmax too small');
            readln;
         99:
            betacf := az;
         End;{function betacf}
      {********************************************************************}
      Begin
         If ((x < 0.0) Or (x > 1.0)) Then Begin
            writeln('pause in routine BETAI');
            readln
         End;
         If ((x = 0.0) Or (x = 1.0)) Then bt := 0.0
         Else
            bt := exp(gammln(a+b)-gammln(a)-gammln(b)
                   +a*ln(x)+b*ln(1.0-x));
         If (x < (a+1.0)/(a+b+2.0))
         Then
            betai := bt*betacf(a,b,x)/a
         Else
            betai := 1.0-bt*betacf(b,a,1.0-x)/b
      End;{function betai} 

{*************************************************************************}
{This function calculates the p-value for the t-statistic t with df       }
{degrees of freedom.  It follows from equation 6.3.9 on page 189 of       }
{Numerical Recipes in Pascal.                                             }
{*************************************************************************}
   Function tvalue(t :extended; df : integer) : extended;
      Begin
         tvalue := 1 - betai(df/2,0.5,df*1.0/(df*1.0+t*t));
      End;


{**********************************************************************}
{This function calculates confidence intervals for df degrees of       }
{freedom with the width CI.  The approach I took was to solve for an x }
{whose t value equaled the desired CI width.  To solve the equation I  }
{used the zbrent algorithm listed on pp 285-256 of Numerical Recipes in}
{Pascal.  I "cheated" and hardcoded calls to tvalue in the appropriate }
{places in the algorithm, while the original passed in a function.     }
{This might cause a problem when this module is extended to handle F   }
{statistic calculations.                                               }
{**********************************************************************}

   Function tCI(df : integer;CI : extended) : extended;
      {*******************************************************************}
      {This function finds the roots of fx in the interval [x1,x2] within }
      {the tolerance tol.                                                 }
      {*******************************************************************}
      Function zbrent(x1,x2,tol: extended;ci : extended) : extended;
         Label 99;
         Const
            itmax = 100;
            eps = 3.0e-8;
         Var
            a,b,c,d,e: extended;
            min1,min2,min: extended;
            fa,fb,fc,p,q,r: extended;
            s,tol1,xm:extended;
            iter: integer;
         Begin
            a := x1;
            b := x2;
            fa := tvalue(a,df)-ci; { fix me !}
            fb := tvalue(b,df)-ci; { me too !}
            If fb*fa > 0.0 Then Begin
               writeln('pause in routine zbrent');
               writeln('root must be bracketed');
               readln;
            End; {then}
            fc := fb;
            For iter := 1 To itmax Do Begin
               If fb*fc > 0.0 Then Begin
                  c := a;
                  fc := fa;
                  d := b-a;
                  e := d;
               End; {then}
               If abs(fc) < abs(fb) Then Begin
                  a := b;
                  b := c;
                  c := a;
                  fa := fb;
                  fb := fc;
                  fc := fa;
               End; {then}
               tol1 := 2.0*eps*abs(b)+0.5*tol; { Convergence check }
               xm := 0.5*(c-b);
               If ((abs(xm) <= tol1) And (Abs(fa) > Abs(fb))) Then Begin
                  zbrent := b;
                  Goto 99
               End;
               If (abs(e) >= tol1) And (abs(fa) > abs(fb)) Then Begin
                  s := fb/fa;   {Attempt inverse quadratic interpolation}
                  If a = c Then Begin
                     p := 2.0*xm*s;
                     q := 1.0 - s;
                  End {then}
                  Else Begin
                     q := fa/fc;
                     r := fb/fc;
                     p := s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
                     q := (q-1.0)*(r-1.0)*(s-1.0);
                  End; {else}
                  If p > 0.0 Then q := -q;
                  p := abs(p);
                  min1 := 3.0*xm*q-abs(tol1*q);
                  min2 := abs(e*q);
                  If min1 < min2 Then min := min1 Else min := min2;
                  If 2.0*p < min Then Begin
                     e := d;                {Accept interpolation}
                     d := p/q;
                  End {then}
                  Else Begin
                     d := xm;           {Interpolation failed, use bisection}
                     e := d;
                  End; {else}
               End {then}
               Else Begin
                  d := xm;      {Bounds decreasing too slowly, use bisection}
                  e := d;
               End; {else}
               a := b;          {Move last best guess to a}
               fa := fb;
               If abs(d) > tol1 Then b := b+d
               Else Begin
                  If xm >= 0 Then b := b + abs(tol1)
                  Else b := b-abs(tol1)
               End; {else}
               fb := tvalue(b,df)-ci;     {fix me}
            End;
            writeln('pause in routine ZBrent');
            writeln('maximum number of iterations exceeded');
            readln;
            zbrent := b;
         99:
         End; {function zbrent}
      {*********************************************************************}
      Begin
         tCI := zbrent(0,100,0.00000001,CI);
      End;

   {************************************************************************}
   {This procedure writes coprocessor extended reals to a file without the
    extra exponent and mantessa digits.}
   {************************************************************************}
   Procedure extendWrite(Var outFile :text;num :extended);
      Var outStr :string[19];
          userChar : char;
      Begin
         If (abs(num) < 9.9E99) Then Begin
            If ((abs(num) > 1E-99) Or (Abs(num) = 0)) Then Begin
               str(num:19:-11,outStr);
               delete(outStr,16,2);
               write(outFile,outStr);
            End
            Else Begin
               clrscr;
               writeln('Error:  The program tried to write a number with an absolute value smaller');
               writeln('        than 1E-99.  The program will now exit.  You should go into');
               writeln('        PDTREE and rescale your tree, for example, if the height of your');
               writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
               writeln('        you may need to rescale tip values.');
               writeln;
               write('Press a key to exit the program');
               userChar := readkey;
               halt
            End; {else}
         End
         Else Begin
            clrscr;
            writeln('Error:  The program tried to write a number with an absolute value larger than');
            writeln('        9.9999999999E+99.  The program will now exit.  You should go into');
            writeln('        PDTREE and rescale your tree, for example, if the height of your');
            writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
            writeln('        you may need to rescale tip values.');
            writeln;
            write('Press a key to exit the program');
            userChar := readkey;
            halt
         End; {else}
      End; {procedure extendwrite}

   {************************************************************************}
   {This procedure writes coprocessor extended reals to a string buffer
    without the extra exponent and mantessa digits.}
   {************************************************************************}
   Procedure extendStr(Var textBuf :string; num :extended);
      Var userChar : char;
      Begin
         If (abs(num) < 9.9E99) Then Begin
            If ((abs(num) > 1E-99) Or (Abs(num) = 0)) Then Begin
               str(num:19:-11,textBuf);
               delete(textBuf,16,2);
            End
            Else Begin
               restorecrtmode;
               writeln('Error:  The program tried to display a number with an absolute value smaller');
               writeln('        than 1E-99.  The program will now exit.  You should go into');
               writeln('        PDTREE and rescale your tree, for example, if the height of your');
               writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
               writeln('        you may need to rescale tip values.');
               writeln;
               write('Press a key to exit the program');
               userChar := readkey;
               halt
            End; {else}
         End
         Else Begin
            restorecrtmode;
            writeln('Error:  The program tried to display a number with an absolute value larger');
            writeln('        than 9.9999999999E+99.  The program will now exit.  You should go into');
            writeln('        PDTREE and rescale your tree, for example, if the height of your');
            writeln('        tree is 70 million, you could rescale it to 70.  As an alternative,');
            writeln('        you may need to rescale tip values.');
            writeln;
            write('Press a key to exit the program');
            userChar := readkey;
            halt
         End; {else}
      End; {procedure extendwrite}


End. {Module T_UTIL1}

